/* aggregator_ep - configurable aggregator event processor */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
08aug14,tle  Fix sendtoself action
11mar14,sdt  Fix 'Checked Return' warnings
24feb14,sdt  Disable unexpected NO_EFFECT warnings

DESCRIPTION
 * This event processor is dynamically configurable using settings to
 * define how to aggregate data from events.  The controllable settings
 * are:
 * 1. Add types
 * 2. Add/delete tables
 * 3. Add/delete actions
 *
 * TYPES
 * -----
 *
 * Types are defined as list of value type 2 (VT2) strings.  This
 * should be used to define structures, natures and typedef for use in
 * tables and actions.
 *
 * TABLES
 * ------
 *
 * Tables are defined as structures with zero or more keys, and zero
 * or more fields of data.  Structures with keys are wrapped into a
 * dictionary.  Tables are addressed either by the structure or
 * typedef name.  They are modified by action statements described
 * below.  The aggregator EP recognize the following table natures:
 *
 * - cafe.PRIVATE
 *
 *   Private tables are not sent upstreams on snapshot actions.
 *
 * - cafe.PERSISTENT
 *
 *   Persistent tables are not cleared on snapshot actions.
 *
 * - cafe.TABLE_NAME(name)
 *
 *   The table name is use as trigger for certain action operations.
 *
 * ACTIONS
 * -------
 *
 * Actions perform some operation based on some trigger.  Actions
 * comprise 1) a trigger, 2) a filter, 3) a target, 4) an operation,
 * and 5) arguments.
 *
 * A trigger is either a comma separated list of fully qualified event
 * types, the first type is the triggering event and the remaining
 * types are orthogonal data required with the event, this is called
 * an event trigger, or a quoted string which is use to match table
 * names defined using the cafe.TABLE_NAME nature, this is called
 * table name triggers.
 *
 * A filter is either empty or a boolean expression.  The action is
 * performed when the trigger occures and the filter expression is
 * empty or evaluates to true, otherwise the action is ignored.  A
 * filter expression may refer to the trigger or tables.
 *
 * A target is either empty or an expression referencing a table,
 * table row or field in a table row, depending on the operation.  A
 * table expression may refer to the trigger or tables.
 *
 * An operation is a string naming the operation to be performed by
 * the action statement.  
 *
 * Available event triggered operations:
 *
 * - echo
 *
 *   Echo's the values of the actionable event to stdout.  No
 *   arguments.
 *
 * - count
 *
 *   Increments the value of target.  Target must be of type L or Q
 *   and have nature cafe.SUM.  No arguments.
 *
 * - set
 *
 *   Sets value of target to value of argument.  Target must be of
 *   type L, Q, d or ? and have nature cafe.SET. One argument.
 *
 * - sum
 *
 *   Adds value of argument to value of target.  Target must be of
 *   type L, Q or d and have nature cafe.SUM. One argument.
 *
 * - min
 *
 *   Sets target to min value of target and argument. Target must be
 *   of type cafe.setval_ q, Q or d and have nature cafe.MIN. One
 *   argument.
 *
 * - max
 *
 *   Sets target to max value of target and argument.  Target must be
 *   of type cafe.setval_ q, Q or d and have nature cafe.MAX. One
 *   argument.
 *
 * - avg
 *
 *   Updates average value in target with value of argument.  Target
 *   must be of type cafe.avgval_ q, Q or d and have nature
 *   cafe.AVG. One argument.
 *
 * - clear
 *
 *   Clear/deletes row specified by target.  Target must be a row in a
 *   dict.  No arguments.
 *
 * - delete
 *
 *   Clear/deletes rows in specified target.  Target must be a dict.
 *   One argument which is evaluated for each row.  The current row
 *   can be referenced in the argument expression using $1.  Only rows
 *   for which the argument evaluates to true are deleted.
 *
 * - reset
 *
 *   Clear/deletes all rows in all tables, including persistent tables
 *   and removes tables pending to be removed.  No target and no
 *   arguments.
 *
 * - sendtoself
 *
 *   Send data to this instance of the aggregator.  This is useful
 *   when aggregating data from a table that was aggregated by this
 *   aggregator instance.  For example when building history or
 *   filtering data from a larger table.  First argument is the event
 *   and subsequent arguments are orthogonal data.  Arguments must be
 *   of type struct or dict.  No target.
 *
 *   Note that to be able to reaggregate a table, it needs to sent as
 *   as an aggregate event and the table must have a name nature.
 *   Example:
 *
 *     add_table: "t[cafe_aggregator.aggregator]"
 *     add_action: "cafe_timer.timer,cafe.timestamp","","","sendtoself","cafe_aggregator.aggregator,details_table__274"
 *
 * - snapshot
 *
 *   Generate snapshot event of all public tables, i.e. tables without
 *   cafe.PRIVATE nature.  No target and no arguments.
 *
 * - mm_update
 *
 *   Add or remove memory map entry.  Target must be of type
 *   cafe.mm_table_entry or a typedef of this type. Arguments are
 *   <memory-space-id>, <address>, <length>, <page-offset>, <mapped>,
 *   <object-file>
 *
 * - mm_clone
 *
 *   Clone memory map.  Target must be of type cafe.mm_table_entry or
 *   a typedef of this type. Arguments are <parent-memory-space-id>,
 *   <new-memory-space-id>
 *
 * - mm_clear
 *
 *   Clear memory map.  Target must be of type cafe.mm_table_entry or
 *   a typedef of this type. Argument is <memory-space-id to be cleared>, and
 *   may optionally also provide a second argument: <memory-space-id of parent
 *   address space>.
 *
 * - mm_delete
 *
 *   Delete memory map.  Target must be of type cafe.mm_table_entry or
 *   a typedef of this type. Argument is <memory-space-id>
 *
 * Available table name triggered operations:
 *
 * - aggregate
 *
 *   Perform table aggregation.  Target must be a table name.
 *   Arguments are a list of expressions.  The first set of
 *   expressions map to keys in the output table and the remaining
 *   expressions are naturally aggregated in fields in the output
 *   table.  Pseudo code for the basic aggregate operation:
 *
 *     foreach $1 in input_table
 *         output_row = output_table[<key_arg1>,<key_arg2>,...]
 *         foreach field in output_table
 *             output_row.field <natural_aggregation> <field_argN>
 *         end
 *     end
 *
 *   Arguments in one of the key positions enclosed by the function
 *   "flatten()" allows an input row field of type dynamic array to be
 *   aggregated into one or more output rows with cumulative
 *   aggregation.  Pseudo code for the aggregate operation when
 *   flatten() is used:
 *
 *     foreach $1 in input_table
 *         evalue flatten_expr into flatten_table
 *         foreach $2 in flatten_table
 *             if have_seen_flatten_row($2)
 *                 continue
 *             output_row = output_table[$2.symbol_name,$2.symbol_offset]
 *             foreach field in output_table
 *                 if not first_flatten_row and not has_nature(field,cafe.CUMULATIVE)
 *                     continue
 *                 output_row.field <natural_aggregation> <field_argN>
 *             end
 *         end
 *     end
 *
 * - mm_delta
 *
 *   Update memory map table with memory map delta.  Target must be of
 *   type cafe.mm_table_entry or a typedef of this type.  Argument is
 *   delta table.  Typically this is the triggering table and
 *   therefore "$1".
 *
 * - stop_dispatch
 *
 *   Stops the event source if the trigger occur and the filter is
 *   true.  The filter is applied to each row in the table.  No target
 *   and no arguments.
 *
 * - address_resolve
 *
 *   Resolve callstack entries on matching incoming events and tables.
 *   Requires that the event or table row contains an address space ID, and
 *   memory map generation number in addition to the raw stack itself.  The
 *   action argument dictates the fields emitted in the resolved results.  This
 *   argument must be a structure value type containing some combination of
 *   attributes, limited to those specified by the cafe.mm_sym_resolve_fields
 *   type.  This action has two other effects, first it causes every incoming
 *   snapshot event to trigger an output snapshot, and second, it causes the
 *   aggregator to mirror any incoming table data in its own output.
 *
 * Arguments may be empty or a comma separated list of expressions.
 * The number and meaning of the arguments depend on the operation.
 *
 * See cafe/expr.h for information about expressions syntax and
 * expression functions.
 */

#include <string.h>
#include <errno.h>
#include <cafe/cafe.h>
#include <cafe/errors.h>
#include <cafe/ptr_array.h>
#include <cafe/stdio.h>
#include <cafe/expr.h>
#include <cafe/utils.h>
#include <cafe/ep/aggregator.h>
#include <cafe/ep/timer_ep.h>
#include <cafe/value_print.h>
#include "../src/strtoll.h"
#include "../src/cafe_assert.h"
#include "symresolve_actions.h"

#define EVENT_ACTION_HASH_SIZE 32

#define TABLE_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct table_inst, table_list)

#define ACTION_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct action_spec, action_list)

#define EAHASH_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct event_actions_map, eahash_list)

typedef enum {
	CAFE_AGGREGATOR_EP_OP_NONE,
	CAFE_AGGREGATOR_EP_OP_COUNT_L,
	CAFE_AGGREGATOR_EP_OP_COUNT_Q,
	CAFE_AGGREGATOR_EP_OP_SET_q,
	CAFE_AGGREGATOR_EP_OP_SET_Q,
	CAFE_AGGREGATOR_EP_OP_SET_d,
	CAFE_AGGREGATOR_EP_OP_SET_bool,
	CAFE_AGGREGATOR_EP_OP_SET_s,
	CAFE_AGGREGATOR_EP_OP_SET_b,
	CAFE_AGGREGATOR_EP_OP_SET_q2,
	CAFE_AGGREGATOR_EP_OP_SET_Q2,
	CAFE_AGGREGATOR_EP_OP_SET_d2,
	CAFE_AGGREGATOR_EP_OP_SET_bool2,
	CAFE_AGGREGATOR_EP_OP_SET_s2,
	CAFE_AGGREGATOR_EP_OP_SET_b2,
	CAFE_AGGREGATOR_EP_OP_SET_q3,
	CAFE_AGGREGATOR_EP_OP_SET_Q3,
	CAFE_AGGREGATOR_EP_OP_SET_d3,
	CAFE_AGGREGATOR_EP_OP_SET_bool3,
	CAFE_AGGREGATOR_EP_OP_SET_s3,
	CAFE_AGGREGATOR_EP_OP_SET_b3,
	CAFE_AGGREGATOR_EP_OP_SET_obj,
	CAFE_AGGREGATOR_EP_OP_SUM_q,
	CAFE_AGGREGATOR_EP_OP_SUM_Q,
	CAFE_AGGREGATOR_EP_OP_SUM_d,
	CAFE_AGGREGATOR_EP_OP_MIN_q,
	CAFE_AGGREGATOR_EP_OP_MIN_Q,
	CAFE_AGGREGATOR_EP_OP_MIN_d,
	CAFE_AGGREGATOR_EP_OP_MIN_q2,
	CAFE_AGGREGATOR_EP_OP_MIN_Q2,
	CAFE_AGGREGATOR_EP_OP_MIN_d2,
	CAFE_AGGREGATOR_EP_OP_MAX_q,
	CAFE_AGGREGATOR_EP_OP_MAX_Q,
	CAFE_AGGREGATOR_EP_OP_MAX_d,
	CAFE_AGGREGATOR_EP_OP_MAX_q2,
	CAFE_AGGREGATOR_EP_OP_MAX_Q2,
	CAFE_AGGREGATOR_EP_OP_MAX_d2,
	CAFE_AGGREGATOR_EP_OP_AVG_q,
	CAFE_AGGREGATOR_EP_OP_AVG_Q,
	CAFE_AGGREGATOR_EP_OP_AVG_d,
	CAFE_AGGREGATOR_EP_OP_AVG_q2,
	CAFE_AGGREGATOR_EP_OP_AVG_Q2,
	CAFE_AGGREGATOR_EP_OP_AVG_d2,
	CAFE_AGGREGATOR_EP_OP_CLEAR,
	CAFE_AGGREGATOR_EP_OP_DELETE,
	CAFE_AGGREGATOR_EP_OP_RESET,
	CAFE_AGGREGATOR_EP_OP_SENDTOSELF,
	CAFE_AGGREGATOR_EP_OP_SNAPSHOT,
	CAFE_AGGREGATOR_EP_OP_AGGREGATE,
	CAFE_AGGREGATOR_EP_OP_STOP_DISPATCH,
	CAFE_AGGREGATOR_EP_OP_MM_UPDATE,
	CAFE_AGGREGATOR_EP_OP_MM_DELTA,
	CAFE_AGGREGATOR_EP_OP_MM_CLONE,
	CAFE_AGGREGATOR_EP_OP_MM_CLEAR,
	CAFE_AGGREGATOR_EP_OP_MM_CLEAR2,
	CAFE_AGGREGATOR_EP_OP_MM_DELETE,
	CAFE_AGGREGATOR_EP_OP_RESOLVE,
	CAFE_AGGREGATOR_EP_OP_MIRROR,
	CAFE_AGGREGATOR_EP_OP_SKIP,
	CAFE_AGGREGATOR_EP_OP_ECHO
} cafe_aggregator_ep_op_tag;

struct operation_variant {
	cafe_aggregator_ep_op_tag tag;
	const char *result_vtstr;
	const char *args_vtstr;
};

struct operation_entry {
	const char *name;
	struct operation_variant *variants;
	const char *nature;
};

struct table_inst {
	bool is_deleted;
	bool is_private;
	bool is_persistent;
	bool is_wrapped;
	bool is_symresolver;
	struct cafe_value_type *type;
	struct cafe_value_data *value;
	struct cafe_list_head table_list;
};

struct action_spec {
	const char *trigger_str;
	const char *filter_str;
	const char *cell_str;
	const char *operation_str;
	const char *args_str;
	struct cafe_vm_expr *trigger_expr;
	const char *trigger_name;
	struct cafe_event_descr *trigger_type;
	struct cafe_vm_expr *filter_expr;
	struct cafe_vm_expr *cell_expr;
	struct cafe_vm_expr *args_expr;
	struct operation_entry *op_ent;
	struct cafe_list_head action_list;
};

struct event_action_inst {
	cafe_aggregator_ep_op_tag operation;
	struct cafe_vm_expr *filter;
	struct cafe_vm_expr *cell_obj;
	struct cafe_vm_attr *cell_attr;
	struct cafe_vm_expr *args;
	struct action_spec *action;
	int table_index;
	struct cafe_value_type *cell_dict;
	struct cafe_value_type *cell_type;
	struct cafe_vm_expr *flatten_expr;
	struct cafe_value_type *flatten_type;
	unsigned int flatten_arg_start;
	unsigned int flatten_arg_end;
	struct cafe_value_type *table_type;
	cafe_aggregator_ep_op_tag *subops;
	cafe_aggregator_ep_op_tag *subops2;
	struct cafe_event_action_info *event_info;
};

struct event_actions_map {
	struct cafe_list_head eahash_list;
	struct cafe_event_descr *event;
	struct event_action_inst *list;
	unsigned int max;
	unsigned int len;
	bool active;
};

struct seen_value_entry {
	cafe_hashval_t hash;
	void *value;
};

struct seen_value {
 	unsigned int size;
 	unsigned int count;
	struct cafe_value_type *type;
	struct seen_value_entry *list;
};


/* The aggregator EP maintains two output events (and their associated action
   lists), one for the current table set, and one for the current table set
   plus any tables whose deletion has been deferred */
enum {
	PENDING_DATA_EVENT = 0,
	SNAPSHOT_EVENT,
	NUM_AGGREGATOR_EVENTS
};


struct aggregator_inst {
	long change_count;
	unsigned long wrap_count;
	bool refresh_output_event_actions;
	bool pending_delete;
	struct cafe_event_descr *event[NUM_AGGREGATOR_EVENTS];
	cafe_eal_handle_t sink_actions[NUM_AGGREGATOR_EVENTS];
	struct cafe_event_action_info *start_event;
	struct cafe_event_action_info *stop_event;
	struct seen_value seen_value;
	struct symresolve_actions *symresolver;
	struct table_inst *symresolver_table;
	struct table_inst *sym_index_table;
	struct table_inst *lineinfo_index_table;
	struct cafe_value_type *mm_delta_type;
	struct cafe_value_type *mm_no_delta_type;
	struct cafe_value_type *symbol_index_type;
	struct cafe_value_type *symbol_index_ext_type;
	struct cafe_value_type *lineinfo_index_type;
	unsigned int temp_type_max;
	struct cafe_value_type **temp_type_list;
	unsigned int temp_arg_max;
	cafe_any_value *temp_arg_list;
	void *temp_event_data[2];    /* temporary event data for sendtoself */
	struct cafe_list_head table_list_head;
	struct cafe_list_head action_list_head;
	struct cafe_list_head eahash_list_head[EVENT_ACTION_HASH_SIZE];
	struct cafe_ptr_array mirror_list;
	struct event_actions_map *current_aggregator_eamap;
	unsigned int * ortho_dmap;
	unsigned int ortho_dmap_size;
};

struct workarea {
	struct cafe_vm_dynarray del_tables;
	struct cafe_vm_dynarray del_actions;
	struct cafe_vm_dynarray add_types;
	struct cafe_vm_dynarray add_tables;
	struct cafe_vm_dynarray add_actions;
};

static struct operation_variant count_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_COUNT_L, "L" },
	{ CAFE_AGGREGATOR_EP_OP_COUNT_Q, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant set_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_SET_q3, cafe_setval_q_VT_DEF, cafe_setval_q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_SET_q, cafe_setval_q_VT_DEF, "q" },
	{ CAFE_AGGREGATOR_EP_OP_SET_Q3, cafe_setval_Q_VT_DEF, cafe_setval_Q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_SET_Q, cafe_setval_Q_VT_DEF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_SET_d3, cafe_setval_d_VT_DEF, cafe_setval_d_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_SET_d, cafe_setval_d_VT_DEF, "d" },
	{ CAFE_AGGREGATOR_EP_OP_SET_bool3, cafe_setval_bool_VT_DEF, cafe_setval_bool_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_SET_bool, cafe_setval_bool_VT_DEF, "?" },
	{ CAFE_AGGREGATOR_EP_OP_SET_s3, cafe_setval_s_VT_DEF, cafe_setval_s_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_SET_s, cafe_setval_s_VT_DEF, "s" },
	{ CAFE_AGGREGATOR_EP_OP_SET_b3, cafe_setval_b_VT_DEF, cafe_setval_b_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_SET_b, cafe_setval_b_VT_DEF, "b" },
	{ CAFE_AGGREGATOR_EP_OP_SET_q2, "q", "q" },
	{ CAFE_AGGREGATOR_EP_OP_SET_Q2, "Q", "Q" },
	{ CAFE_AGGREGATOR_EP_OP_SET_d2, "d", "d" },
	{ CAFE_AGGREGATOR_EP_OP_SET_bool2, "?", "?" },
	{ CAFE_AGGREGATOR_EP_OP_SET_s2, "s", "s" },
	{ CAFE_AGGREGATOR_EP_OP_SET_b2, "b", "b" },
	{ CAFE_AGGREGATOR_EP_OP_SET_obj, "*", "*" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant sum_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_SUM_q, "q", "q" },
	{ CAFE_AGGREGATOR_EP_OP_SUM_Q, "Q", "Q" },
	{ CAFE_AGGREGATOR_EP_OP_SUM_d, "d", "d" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant min_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_MIN_q2, cafe_setval_q_VT_DEF, cafe_setval_q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_MIN_q, cafe_setval_q_VT_DEF, "q" },
	{ CAFE_AGGREGATOR_EP_OP_MIN_Q2, cafe_setval_Q_VT_DEF, cafe_setval_Q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_MIN_Q, cafe_setval_Q_VT_DEF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_MIN_d2, cafe_setval_d_VT_DEF, cafe_setval_d_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_MIN_d, cafe_setval_d_VT_DEF, "d" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant max_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_MAX_q2, cafe_setval_q_VT_DEF, cafe_setval_q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_MAX_q, cafe_setval_q_VT_DEF, "q" },
	{ CAFE_AGGREGATOR_EP_OP_MAX_Q2, cafe_setval_Q_VT_DEF, cafe_setval_Q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_MAX_Q, cafe_setval_Q_VT_DEF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_MAX_d2, cafe_setval_d_VT_DEF, cafe_setval_d_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_MAX_d, cafe_setval_d_VT_DEF, "d" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant avg_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_AVG_q2, cafe_avgval_q_VT_DEF, cafe_avgval_q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_AVG_q, cafe_avgval_q_VT_DEF, "q" },
	{ CAFE_AGGREGATOR_EP_OP_AVG_Q2, cafe_avgval_Q_VT_DEF, cafe_avgval_Q_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_AVG_Q, cafe_avgval_Q_VT_DEF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_AVG_d2, cafe_avgval_d_VT_DEF, cafe_avgval_d_VT_DEF },
	{ CAFE_AGGREGATOR_EP_OP_AVG_d, cafe_avgval_d_VT_DEF, "d" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant clear_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_CLEAR, "*" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant delete_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_DELETE, "*", "?" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant reset_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_RESET },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant sendtoself_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_SENDTOSELF },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant snapshot_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_SNAPSHOT },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant aggregate_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_AGGREGATE, "*" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant stop_dispatch_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_STOP_DISPATCH },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant mm_update_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_MM_UPDATE, CAFE_MM_DELTA_REF, "Q,Q,Q,Q,Q,s" },
	{ CAFE_AGGREGATOR_EP_OP_MM_UPDATE, CAFE_MM_NO_DELTA_REF, "Q,Q,Q,Q,Q,s" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant mm_delta_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_MM_DELTA, CAFE_MM_DELTA_REF, CAFE_MM_DELTA_REF },
	{ CAFE_AGGREGATOR_EP_OP_MM_DELTA, CAFE_MM_NO_DELTA_REF, CAFE_MM_DELTA_REF },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant mm_clone_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_MM_CLONE, CAFE_MM_DELTA_REF, "Q,Q" },
	{ CAFE_AGGREGATOR_EP_OP_MM_CLONE, CAFE_MM_NO_DELTA_REF, "Q,Q" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant mm_clear_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_MM_CLEAR, CAFE_MM_DELTA_REF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_MM_CLEAR2, CAFE_MM_DELTA_REF, "Q,Q" },
	{ CAFE_AGGREGATOR_EP_OP_MM_CLEAR, CAFE_MM_NO_DELTA_REF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_MM_CLEAR2, CAFE_MM_NO_DELTA_REF, "Q,Q" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant mm_delete_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_MM_DELETE, CAFE_MM_DELTA_REF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_MM_DELETE, CAFE_MM_NO_DELTA_REF, "Q" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant echo_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_ECHO },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_variant resolve_variants[] = {
	{ CAFE_AGGREGATOR_EP_OP_RESOLVE, "*" },
	{ CAFE_AGGREGATOR_EP_OP_MIRROR, "*" },
	{ CAFE_AGGREGATOR_EP_OP_NONE }
};

static struct operation_entry operations[] = {
	{ "count",	count_variants,		cafe_SUM_VT_REF },
	{ "set",	set_variants,		cafe_SET_VT_REF },
	{ "sum",	sum_variants,		cafe_SUM_VT_REF },
	{ "min",	min_variants,		cafe_MIN_VT_REF },
	{ "max",	max_variants,		cafe_MAX_VT_REF },
	{ "avg",	avg_variants,		cafe_AVG_VT_REF },
	{ "clear",	clear_variants,		NULL },
	{ "delete",	delete_variants,	NULL },
	{ "reset",	reset_variants,		NULL },
	{ "sendtoself",	sendtoself_variants,	NULL },
	{ "snapshot",	snapshot_variants,	NULL },
	{ "aggregate",	aggregate_variants,	NULL },
	{ "stop_dispatch",stop_dispatch_variants,NULL },
	{ "mm_update",	mm_update_variants,	NULL },
	{ "mm_delta",	mm_delta_variants,	NULL },
	{ "mm_clone",	mm_clone_variants,	NULL },
	{ "mm_clear",	mm_clear_variants,	NULL },
	{ "mm_delete",	mm_delete_variants,	NULL },
	{ "echo",	echo_variants,		NULL },
	{ "address_resolve",	resolve_variants,	NULL },
	{ NULL,		NULL,			NULL }
};


/* debug aid:
   1 : Report errors from process_settings
   2 : Report all delete and adds from process_settings (types, tables, actions)
   4 : More detailed warnings from symbol resolution (e.g. address found in unmapped region)
*/

long cafe_aggregator_debug = 0;

static int update_output_event_actions(
	struct cafe_ep_inst *inst);
static int update_output_event(
	struct cafe_ep_inst *inst);
static void table_inst_delete(
	struct cafe_ep_inst *inst,
	struct table_inst *ti,
	int deferred);
static int action_resolve(
	struct cafe_ep_inst *inst,
	struct action_spec *action,
	struct cafe_event_descr *event,
	unsigned int *odmap,
	int table_index,
	struct event_actions_map *eamap,
	struct cafe_error_report *e);
static void aggregator_ep_delete(
	struct cafe_ep_inst *inst);
static int update_type_max(
	struct cafe_ep_inst *inst,
	unsigned int needed);

static const struct symresolve_functions *symresolve_funcs;

static struct cafe_ep_spec spec;

static const char *wildcard_TYPE[] = {
	cafe_wildcard_VT_REF
};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(wildcard)
};

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(cafe_aggregator_aggregator_VT_REF)
};

static void seen_value_clear(
	struct seen_value *seen,
	struct cafe_value_type *type)
{
	seen->count = 0;
	seen->type = type;
	if (seen->list)
		memset(seen->list, 0, seen->size * sizeof *seen->list);
}

static bool seen_value_is_new(
	const struct cafe_allocator *allocator,
	struct seen_value *seen,
	void *value)
{
	cafe_hashval_t hash = cafe_hash_value(seen->type, value);
	unsigned int i;

	if (seen->list) {
		i = (unsigned int) (hash % seen->size);
		while (seen->list[i].value != NULL) {
			if (seen->list[i].hash == hash &&
			    cafe_same_value(seen->type, seen->list[i].value, value))
				return false;
			if (++i == seen->size)
				i = 0;
		}
	} else
		i = 0;

	if (seen->count*2 >= seen->size) {
		/* Re-hash */
		unsigned int new_size = seen->size*2 + 5;
		struct seen_value_entry *new_list = CAFE_ALLOC(
			allocator, new_size * sizeof *new_list);
		if (new_list == NULL)
			return false;
		memset(new_list, 0, new_size * sizeof *new_list);

		for (i = 0; i < seen->size; i++) {
			void *val;
			unsigned int j;

			if (seen->list == NULL)
				continue;
			val = seen->list[i].value;
			if (val == NULL)
				continue;
			j = (unsigned int) (seen->list[i].hash % new_size);
			while(new_list[j].value != NULL) {
				if(++j == new_size)
					j = 0;
			}
			new_list[j].hash = seen->list[i].hash;
			new_list[j].value = val;
		}

		CAFE_FREE(allocator, seen->list);
		seen->list = new_list;
		seen->size = new_size;

		i = (unsigned int) (hash % new_size);
		while (seen->list[i].value != NULL) {
			if(++i == seen->size)
				i = 0;
		}
	}
	assert(seen->list != NULL);
	seen->list[i].hash = hash;
	seen->list[i].value = value;
	seen->count++;
	return true;
}

static const char *parse_l(const char *fmt, long *result)
{
	char *str;

	errno = 0;
	*result = strtol(fmt, &str, 0);
	if (errno != 0 || fmt >= str)
		return NULL;
	return str;
}

static int is_source_stopped(const struct cafe_ep_inst *inst)
{
	assert(inst->input_ctx != CAFE_EXEC_CTX_ANY);
	return (inst->input_ctx == NULL ||
		inst->input_ctx == CAFE_EXEC_CTX_SESSION ||
		inst->input_ctx->state == CAFE_EP_INST_STATE_STOPPED);
}

static int table_name_eq(const struct cafe_value_type *type, const char *name)
{
	const struct cafe_value_nature_ref *nature;
	struct cafe_vm_expr *e;

	nature = cafe_find_nature_str(type->ident, cafe_TABLE_NAME_VT_REF);
	if (nature == NULL)
		return 0;

	e = nature->args;
	assert(e != NULL && e->kind == CAFE_EXPR_LIST);
	e = e->u.e;
	assert(e != NULL && e->kind == CAFE_EXPR_ELEMENT);
	e = e->u.elem.e;
	assert(e != NULL && e->kind == CAFE_EXPR_STRING);
	return strcmp(e->u.str.p, name) == 0;
}

static struct cafe_value_type *unwrap_table_type(
	struct cafe_value_type *type)
{
	struct cafe_value_type *base_type;
	struct cafe_vm_attr *attr;

	if (type == NULL ||
	    cafe_find_nature_str(type->ident, "N[AGGREGATOR_TABLE]") == NULL)
		return type;

	base_type = cafe_base_type(type);
	if (base_type->kind != CAFE_TYPE_STRUCT || base_type->elements != 1)
		return type;

	attr = cafe_type_attr_by_index(base_type, 0);
	if (attr->offset != 0)
		return type;

	return cafe_attr_type(attr);
}

static int matching_event(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_descr *trigger,
	unsigned int *odmap)
{
	unsigned int i;
	unsigned int k;
	struct cafe_value_type *type;

	assert(event);
	assert(event->num_types > 0);

	if (trigger == NULL || trigger->num_types < 1)
		return 0;

	type = trigger->types[0];
	if (!cafe_type_based_on(event->types[0], type) &&
	    type != inst->session->wildcard_type)
		return 0;

	i = 1;
	k = 0;
	if (odmap)
		odmap[k++] = 0;
	while (i < trigger->num_types) {
		unsigned int l;

		type = trigger->types[i++];
		for (l = 1; l < event->num_types; l++) {
			if (cafe_type_based_on(event->types[l], type))
				break;
		}
		if (l >= event->num_types || k >= event->num_types)
			return 0;
		if (odmap)
			odmap[k++] = l;
	}
	return 1;
}

static int matching_table_name(
	struct cafe_event_descr *event,
	const char *table_name,
	int idx)
{
	assert(idx >= 0);
	if (table_name == NULL)
		return -1;
	while (idx < (int)event->num_types) {
		struct cafe_value_type *type =
			unwrap_table_type(event->types[idx]);
                assert(type != NULL);
		if (table_name_eq(type, table_name))
			return idx;
		idx++;
	}
	return -1;
}

static int matching_table_type(
	const struct cafe_event_descr *event,
	const struct cafe_event_descr *trigger,
	int idx)
{
	struct cafe_value_type *trigger_type;

	assert(idx >= 0);
	if (trigger == NULL || trigger->num_types != 1)
		return -1;

	trigger_type = trigger->types[0];
	while (idx < (int)event->num_types) {
		struct cafe_value_type *type =
			unwrap_table_type(event->types[idx]);
		type = cafe_base_type(type);
                assert(type != NULL);
		if (type->kind == CAFE_TYPE_DICT) {
			type = cafe_base_type(cafe_element_type(type));
		}
		if (cafe_type_based_on(type, trigger_type))
			return idx;
		idx++;
	}
	return -1;
}


/**
 *  Returns an expr referencing the table matching the given type, if it
 * exists.
 */
static
int table_matching_type(struct cafe_ep_inst *inst,
			struct cafe_value_context *context,
			struct cafe_value_type *type,
			struct cafe_vm_expr **r)
{
	struct aggregator_inst *ai = inst->client_data;
	struct cafe_list_head *table_list_head = &ai->table_list_head;
	struct cafe_list_head *list_item;
	struct table_inst *ti;
	int rc;

	assert(type->kind == CAFE_TYPE_STRUCT);

	cafe_list_foreach(list_item, table_list_head) {
		ti = TABLE_LIST_NODE_ENTRY(list_item);

		if (ti->is_deleted)
			continue;
		if (!cafe_type_based_on(ti->type, type))
			continue;

		return cafe_expr_var(context, type, ti, r);
	}

	/* matching table not found, create one */

	rc = cafe_aggregator_ep_table_create(inst, type);
	if (rc != CAFE_OK) {
		(void)fprintf(stderr, "error creating table: %s\n",
			cafe_strerror(rc));
		return rc;
	}

	ti = TABLE_LIST_NODE_ENTRY(ai->table_list_head.prev);
	rc = cafe_expr_var(context, ti->type, ti, r);
	if (rc != CAFE_OK)
		(void)fprintf(stderr, "error creating generic resolve output expression: %s\n",
			cafe_strerror(rc));
		//td: more cleanup here

	return rc;
}


/**
 *  Given a struct type, return a new struct type which mirrors the input type,
 * except that the memory map generation number has been stripped, and the
 * callstack has been replaced with an array of structures defined by the
 * "resolved_type" parameter.
 */
static int resolved_struct_get(const struct cafe_value_type *type,
			       const struct cafe_vm_expr *resolved_expr,
			       const struct cafe_vm_attr *mm_attr,
			       const struct cafe_vm_attr *stack_attr,
			       const char *suffix,
			       const struct cafe_allocator *allocator,
			       struct cafe_value_type **new_type)
{
	struct cafe_value_type *resolved_types, *resolved_type_array;
	struct cafe_array attr_list;
	size_t len1, len2;
	char *name_buf;
	struct cafe_vm_attr *attr;
	int rc;

	resolved_types = cafe_base_type(resolved_expr->type);
	assert(resolved_types != NULL);
	assert(resolved_types->kind == CAFE_TYPE_STRUCT);
	rc = cafe_type_dynarray(resolved_types, &resolved_type_array);
	if (rc != CAFE_OK) {
		(void)fprintf(stderr, "Error wrapping resolved types as a "
			"dynamic array: %s\n", cafe_strerror(rc));
		return rc;
	}

	cafe_array_init(&attr_list, allocator, sizeof(*attr));

	cafe_type_attr_foreach(attr, type) {
		if (attr == mm_attr)
			continue;

		if (attr == stack_attr)
			rc = cafe_value_attr_list_add_ident(
				stack_attr->kind, resolved_type_array->ident,
				"resolved_stack", NULL, &attr_list);
		else
			rc = cafe_value_attr_list_add(attr, &attr_list);
		if (rc != CAFE_OK)
			goto fail;
	}

	len1 = strlen(type->ident->name);
	len2 = strlen(suffix);
	name_buf = CAFE_ALLOC(allocator, len1 + len2 + 1);
	if (name_buf == NULL) {
		rc = CAFE_NOMEM;
		goto fail;
	}
	memcpy(name_buf, type->ident->name, len1);
	memcpy(&name_buf[len1], suffix, len2);
	name_buf[len1 + len2] = 0;
	rc = cafe_value_struct_create(name_buf, type->ident->ns, &attr_list,
				      type->ident->natures, new_type);
	CAFE_FREE(allocator, name_buf);
	if (rc != CAFE_OK) {
		fprintf(stderr, "Error creating resolved structure definition:"
			" %s\n", cafe_strerror(rc));
		goto fail;
	}

	return CAFE_OK;

fail:
	cafe_array_release(&attr_list);
	return rc;
}

static void eainst_free(
	const struct cafe_allocator *a,
	struct event_action_inst *eainst)
{
	if (eainst->subops != NULL)
		CAFE_FREE(a, eainst->subops);
	if (eainst->filter != NULL)
		cafe_expr_free(a, eainst->filter);
	if (eainst->cell_obj)
		cafe_expr_free(a, eainst->cell_obj);
	if (eainst->args)
		cafe_expr_free(a, eainst->args);
	if (eainst->flatten_expr)
		cafe_expr_free(a, eainst->flatten_expr);
}

static void eamap_free(
	struct cafe_ep_inst *inst,
	struct event_actions_map *eamap)
{
	struct aggregator_inst *ai =
		(struct aggregator_inst *)inst->client_data;
	assert(!eamap->active);
	while (eamap->len-- > 0) {
		struct event_action_inst *eainst = eamap->list + eamap->len;
		eainst_free(inst->allocator, eainst);
	}

	if (eamap == ai->current_aggregator_eamap)
		ai->current_aggregator_eamap = NULL;

	CAFE_FREE(inst->allocator, eamap->list);
	CAFE_FREE(inst->allocator, eamap);
}

static void eahash_free(
	struct cafe_ep_inst *inst,
	struct aggregator_inst *ai)
{
	struct cafe_list_head *list_head;
	struct cafe_list_head *list_item;
	struct cafe_list_head *tmp;
	unsigned int i;

	for (i = 0; i < EVENT_ACTION_HASH_SIZE; i++) {
		list_head = &ai->eahash_list_head[i];
		cafe_list_foreach_safe(list_item, list_head, tmp) {
			struct event_actions_map *eamap =
				EAHASH_LIST_NODE_ENTRY(list_item);
			cafe_list_remove(&eamap->eahash_list);
			eamap_free(inst, eamap);
		}
	}
}

static struct event_action_inst *eainst_alloc(
	const struct cafe_allocator *allocator,
	struct event_actions_map *eamap)
{
	struct event_action_inst *eainst;
	unsigned int max = eamap->max;
	if (eamap->len == max) {
		struct event_action_inst *newlist;
		max = max ? max*2 : 1;
		newlist = CAFE_REALLOC(allocator, eamap->list,
				       max * sizeof *eamap->list);
		if (newlist == NULL)
			return NULL;
		eamap->list = newlist;
		eamap->max = max;
	}
	eainst = eamap->list + eamap->len;
	memset(eainst, 0, sizeof *eainst);
	return eainst;
}

static const char *trigger_table_name(
	struct cafe_vm_expr *trigger)
{
	if (trigger == NULL ||
	    trigger->kind != CAFE_EXPR_LIST ||
	    (trigger = trigger->u.e) == NULL)
		return NULL;
	assert(trigger->kind == CAFE_EXPR_ELEMENT);
	if (trigger->u.elem.e->kind != CAFE_EXPR_STRING)
		return NULL;
	return trigger->u.elem.e->u.str.p;
}

static int trigger_type(
	struct cafe_ep_inst *inst,
	struct cafe_vm_expr *trigger_expr,
	struct cafe_event_descr **event)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_vm_expr *trigger;
	struct cafe_value_type **types;
	unsigned int num_types;
	int rc;

	if (trigger_expr == NULL || trigger_expr->kind != CAFE_EXPR_LIST)
		return CAFE_ERROR;

	/* Validate and count types in trigger */
	num_types = 0;
	trigger = trigger_expr->u.e;
	while (trigger != NULL) {
		assert(trigger->kind == CAFE_EXPR_ELEMENT);
		if (trigger->u.elem.e->kind != CAFE_EXPR_TYPE)
			return CAFE_ERROR;
		num_types++;
		trigger = trigger->u.elem.next;
	}
	if (num_types == 0)
		return CAFE_ERROR;

	rc = update_type_max(inst, num_types);
	if (rc != CAFE_OK)
		return rc;
	types = ai->temp_type_list;
	num_types = 0;
	trigger = trigger_expr->u.e;
	while (trigger != NULL) {
		assert(trigger->kind == CAFE_EXPR_ELEMENT);
		assert(trigger->u.elem.e->kind == CAFE_EXPR_TYPE);
		types[num_types++] = trigger->u.elem.e->type;
		trigger = trigger->u.elem.next;
	}
	return cafe_cfg_get_event(inst->session, num_types, types, event);
}

static bool ptr_array_contains(struct cafe_ptr_array *list, const void *item)
{
	unsigned int i;

	for (i = 0; i < list->len; ++i)
		if (list->data[i] == item)
			return true;

	return false;
}

static void dispatch_status(struct cafe_error_report *err,
			    struct cafe_ep_inst *inst)
{
	struct cafe_byte_buffer buf;

	(void)cafe_bbuf_init(&buf, inst->allocator, 128);
	(void)cafe_error_bbuf_write(err, &buf, NULL, ": ");
	cafe_dispatch_status(inst, CAFE_LOG_WARNING, buf.data);
	cafe_bbuf_release(&buf);
	cafe_error_clear(err);
}

static void action_resolve_dispatch_status(struct cafe_error_report *err,
					   int rc, const struct action_spec *action,
					   struct cafe_ep_inst *inst)
{
	if (err->rc != CAFE_OK) {
		cafe_error_set(err, err->rc, "Action resolve failed <%s> <%s> <%s> <%s> <%s>",
			       action->trigger_str,
			       action->filter_str, action->cell_str,
			       action->operation_str, action->args_str);
		dispatch_status(err, inst);
	} else if (rc != CAFE_OK) {
		char *msg = cafe_mprintf(inst->allocator, "Action resolve failed <%s> <%s> <%s> <%s> <%s>: %s",
					 action->trigger_str,
					 action->filter_str, action->cell_str,
					 action->operation_str,
					 action->args_str, cafe_strerror(rc));
		assert(msg != NULL);
		cafe_dispatch_status(inst, CAFE_LOG_WARNING, msg);
		CAFE_FREE(inst->allocator, msg);
	}
}

static int eamap_build(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct event_actions_map **result)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct event_actions_map *eamap;
	struct cafe_list_head *list_item;
	struct cafe_error_report e;
	int rc;

	/* Add event to actions map for next time we get this event */
	if ((eamap = CAFE_ALLOC(inst->allocator, sizeof *eamap)) == NULL)
		return CAFE_NOMEM;
	memset(eamap, 0, sizeof *eamap);
	memset(&e, 0, sizeof(e));
	eamap->event = event;
	cafe_list_foreach(list_item, &ai->action_list_head) {
		struct action_spec *action = 
			ACTION_LIST_NODE_ENTRY(list_item);

		if (action->op_ent->variants == aggregate_variants ||
		    action->op_ent->variants == stop_dispatch_variants) {
			int table_index = 0;

			while ((table_index = matching_table_name(
					event, action->trigger_name, table_index)) >= 0) {
				rc = action_resolve(inst, action, event, NULL,
					       table_index, eamap, &e);
				action_resolve_dispatch_status(&e, rc, action,
							       inst);
				table_index++;
			}
		} else if (action->op_ent->variants == mm_delta_variants) {
			int table_index = 0;

			while ((table_index = matching_table_type(
					event, action->trigger_type, table_index)) >= 0) {
				rc = action_resolve(inst, action, event, NULL, table_index, eamap, &e);
				action_resolve_dispatch_status(&e, rc, action,
							       inst);
				table_index++;
			}
		} else if (action->op_ent->variants == resolve_variants) {
			unsigned int i;
			struct cafe_ptr_array new_mirror_list;

			if (!matching_event(inst, event, action->trigger_type,
					    NULL))
				continue;

			// Only consider aggregator snapshot events for now
			if (event->types[0] != inst->output_events[0])
				continue;

			cafe_ptr_array_init(&new_mirror_list, inst->allocator);

			CAFE_LOCK();
			for (i = 1; i < event->num_types; ++i) {
				/* Skip the delta tables; these are "consumed"
				   by the resolution action and do not need
				   mirroring. */
				if ((event->types[i] == ai->mm_delta_type) ||
				    (event->types[i] == ai->mm_no_delta_type))
					continue;
				rc = action_resolve(inst, action, event, NULL,
						    i, eamap, &e);
				action_resolve_dispatch_status(&e, rc, action,
							       inst);

				rc = cafe_ptr_array_append(&new_mirror_list,
							   eamap->list[eamap->len - 1].cell_obj->u.p);
				if (rc != CAFE_OK) {
					cafe_error_set(&e, rc, "Error adding new table to mirror list");
					dispatch_status(&e, inst);
				}
			}

			/* mark any no longer mirrored tables as deleted */
			for (i = 0; i < ai->mirror_list.len; ++i) {
				struct table_inst *ti = cafe_ptr_array_val(&ai->mirror_list, i);
				if (!ptr_array_contains(&new_mirror_list, ti))
					table_inst_delete(inst, ti, false);
			}
			cafe_ptr_array_release(&ai->mirror_list);
			ai->mirror_list = new_mirror_list;

			if (ai->current_aggregator_eamap != NULL) {
				cafe_list_remove(&ai->current_aggregator_eamap->eahash_list);
				eamap_free(inst, ai->current_aggregator_eamap);
			}
			ai->current_aggregator_eamap = eamap;

			rc = update_output_event(inst);
			if (rc != CAFE_OK) {
				dispatch_status(&cafe_error_cur, inst);
				CAFE_UNLOCK();
				return rc;
			}

			rc = update_output_event_actions(inst);
			if (rc != CAFE_OK) {
				dispatch_status(&cafe_error_cur, inst);
				CAFE_UNLOCK();
				return rc;
			}
			CAFE_UNLOCK();
		} else {
			if (event->num_types > ai->ortho_dmap_size) {
			    ai->ortho_dmap_size = event->num_types;
			    ai->ortho_dmap = CAFE_REALLOC(inst->allocator,
			                                  ai->ortho_dmap,
			                                  ai->ortho_dmap_size * sizeof(unsigned int));
			}
			assert(ai->ortho_dmap != NULL);
			if (!matching_event(inst, event, action->trigger_type, ai->ortho_dmap))
				continue;
			rc = action_resolve(inst, action, event, ai->ortho_dmap, -1, eamap, &e);
			action_resolve_dispatch_status(&e, rc, action, inst);
		}
	}
	*result = eamap;
	return CAFE_OK;
}

static int get_mm_gen(
	struct cafe_expr_eval_context *context,
	unsigned long long address_space_uid,
	unsigned long long *result)
{
	struct symresolve_actions *symresolver =
		(struct symresolve_actions *)context->mm_inst;
	unsigned long long current_gen;
	int rc;

	rc = symresolve_funcs->mm_current_gen(symresolver,
					      address_space_uid,
					      &current_gen);
	if (rc != CAFE_OK) {
		if (cafe_aggregator_debug & 1)
			(void)fprintf(stderr, "memory map current generation error %d\n", rc);
		return rc;
	}
	*result = current_gen;
	return CAFE_OK;
}

static int do_mm_sym_resolve(
	struct cafe_expr_eval_context *context,
	const struct cafe_vm_dynarray *stack,
	unsigned long long address_space_uid,
	unsigned long long current_gen,
	const struct cafe_value_type *result_type,
	struct cafe_vm_dynarray *result)
{
	struct symresolve_actions *symresolver =
		(struct symresolve_actions *)context->mm_inst;
	int rc;

	result_type = cafe_base_type(result_type);
        assert(result_type != NULL);
	assert(result_type->kind == CAFE_TYPE_DYNARRAY);

	rc = symresolve_funcs->callstack(symresolver,
					 address_space_uid,
					 current_gen,
					 stack,
					 result_type, result);
	if (rc != CAFE_OK)
		(void)fprintf(stderr, "symbol resolve error %d\n", rc);

	return rc;
}

static void print_event(
	struct cafe_event_action_info *event_info,
	void **data)
{
	unsigned int i;
	struct cafe_value_print_options options = CAFE_VALUE_PRINT_OPTIONS_INIT;

	options.type_name = true;
	options.field_names = true;
	options.radix = 16;

	for (i = 0; i < event_info->event->num_types; i++) {
		if (i != 0)
			(void)fputs_unlocked(", ", stdout);
		cafe_value_print(stdout, NULL, event_info->event->types[i],
				 data ? data[i] : NULL, &options);
	}

	(void)fputc_unlocked('\n', stdout);
}

static void action_event_handler(
	struct cafe_event_action_info *info,
	void **data)
{
	struct cafe_ep_inst *inst = info->inst;
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_event_descr *event = info->event;
	struct cafe_list_head *list_head;
	struct cafe_list_head *list_item;
	struct event_actions_map *eamap;
	struct event_action_inst *eainst;
	struct cafe_expr_eval_context context;
	int do_mm_flush = 0;
	int stop_dispatch = 0;
	void *curdata[2];
	unsigned int i;
	int rc;

	assert(!ai->refresh_output_event_actions);
	list_head = &ai->eahash_list_head[((unsigned long)event) % EVENT_ACTION_HASH_SIZE];
	cafe_list_foreach(list_item, list_head) {
		eamap = EAHASH_LIST_NODE_ENTRY(list_item);
		if (eamap->event == event) {
			break;
		}
	}
	if (list_item == list_head) {
		/* Add event to actions map for next time we get this event */
		rc = eamap_build(inst, event, &eamap);
		if (rc != CAFE_OK)
			return;
		cafe_list_add_last(&eamap->eahash_list, list_head);
		if (cafe_aggregator_debug & 2)
			fprintf(stderr, "action_event_handler called for %s, new %d\n",
				event->types[0]->ident->name,
				eamap->len);
	}

	/* Avoid infinite recursion */
	if (eamap->active)
		return;
	eamap->active = true;

	cafe_expr_eval_context_create(&context, inst->allocator);
	context.event = event;
	if (ai->symresolver) {
		context.mm_inst = (struct mm_inst_data *)ai->symresolver;
		context.mm_gen = get_mm_gen;
		context.mm_sym_resolve = do_mm_sym_resolve;
	}
	for (i = 0; i < eamap->len; i++) {
		cafe_any_value result;
		cafe_any_value *args = ai->temp_arg_list;
		struct cafe_dict_struct *dict;
		struct cafe_dict_iterator iter;
		struct cafe_value_data *value;
		struct cafe_vm_dynarray *flatten_dary = NULL;
		size_t flatten_index = 0;

		eainst = eamap->list + i;
		if (eainst->table_index < 0) {
			dict = NULL;
			context.event_data = data;
		} else {
			if (eainst->table_type &&
			    eainst->operation != CAFE_AGGREGATOR_EP_OP_MM_DELTA) {
				dict = (struct cafe_dict_struct *)data[eainst->table_index];
				value = cafe_dict_first(dict, &iter);
				if (value == NULL)
					continue;
				curdata[0] = value->u.data;
			} else {
				dict = NULL;
				curdata[0] = data[eainst->table_index];
			}
			context.event_data = curdata;
		}
		do {
			if (eainst->filter != NULL) {
				rc = cafe_expr_eval_value(&context, eainst->filter, &result);
				if (rc != CAFE_OK)
					continue;
				if (result.ll == 0)
					continue;
			}
			if (eainst->flatten_expr != NULL) {
				assert(eainst->operation == CAFE_AGGREGATOR_EP_OP_AGGREGATE);
				rc = cafe_expr_eval_value(&context, eainst->flatten_expr, &result);
				if (rc != CAFE_OK)
					break;
				flatten_dary = (struct cafe_vm_dynarray *)result.p;
				if (flatten_dary->l == 0)
					continue;
				flatten_index = 0;
				curdata[1] = flatten_dary->p;
			}
			if (eainst->cell_obj != NULL) {
				assert(eainst->cell_attr != NULL ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_CLEAR ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_RESOLVE ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_MIRROR ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_DELETE ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_AGGREGATE ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_MM_DELTA ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_MM_UPDATE ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_MM_CLONE ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_MM_CLEAR ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_MM_CLEAR2 ||
				       eainst->operation == CAFE_AGGREGATOR_EP_OP_MM_DELETE);
				rc = cafe_expr_eval_addr(&context, eainst->cell_obj, &result);
				if (rc != CAFE_OK)
					continue;
				if (result.p == NULL &&
				    eainst->operation != CAFE_AGGREGATOR_EP_OP_CLEAR &&
				    eainst->operation != CAFE_AGGREGATOR_EP_OP_SNAPSHOT)
					continue;
			}
			if (eainst->args != 0 &&
			    eainst->operation != CAFE_AGGREGATOR_EP_OP_DELETE) {
				struct cafe_vm_expr *e = eainst->args;
				unsigned int j = 0;
				assert(e->kind == CAFE_EXPR_LIST);
				for (e = e->u.e; e != NULL; e = e->u.elem.next) {
					assert(j < ai->temp_arg_max);
					if (e->u.elem.e->kind == CAFE_EXPR_TYPE) {
						/* arg[0] of sendtoself is a type and cannot be evaluated */
						args[j].p = NULL;
					} else {
						rc = cafe_expr_eval_value(&context, e->u.elem.e, &args[j]);
						if (rc != CAFE_OK)
							break;
					}
					j++;
				}
				if (e != NULL)
					continue;
			}
			switch (eainst->operation) {
			case CAFE_AGGREGATOR_EP_OP_COUNT_L:
				assert(eainst->cell_obj != NULL);
				*(unsigned long *)result.p += 1;
				break;

			case CAFE_AGGREGATOR_EP_OP_COUNT_Q:
				assert(eainst->cell_obj != NULL);
				*(unsigned long long *)result.p += 1;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_q3: {
				struct cafe_setval_q *s =
					(struct cafe_setval_q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				((struct cafe_setval_q *)result.p)->is_set = 1;
				((struct cafe_setval_q *)result.p)->value = s->value;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_Q3: {
				struct cafe_setval_Q *s =
					(struct cafe_setval_Q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				((struct cafe_setval_Q *)result.p)->is_set = 1;
				((struct cafe_setval_Q *)result.p)->value = s->value;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_d3: {
				struct cafe_setval_d *s =
					(struct cafe_setval_d *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				((struct cafe_setval_d *)result.p)->is_set = 1;
				((struct cafe_setval_d *)result.p)->value = s->value;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_bool3: {
				struct cafe_setval_bool *s =
					(struct cafe_setval_bool *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				((struct cafe_setval_bool *)result.p)->is_set = 1;
				((struct cafe_setval_bool *)result.p)->value = s->value;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_s3: {
				struct cafe_setval_s *s =
					(struct cafe_setval_s *)args[0].p;
				struct cafe_setval_s *p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				p = (struct cafe_setval_s *)result.p;
				if (p->value)
					CAFE_FREE(inst->allocator, p->value);
				p->is_set = 1;
				p->value = s->value;
				rc = cafe_tmp_string_persist(&context, &p->value);
				if (rc != CAFE_OK)
					p->value = NULL;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_b3: {
				struct cafe_setval_b *s =
					(struct cafe_setval_b *)args[0].p;
				struct cafe_setval_b *p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				p = (struct cafe_setval_b *)result.p;
				if (p->value.p)
					CAFE_FREE(inst->allocator, p->value.p);
				p->is_set = 1;
				p->value = s->value;
				rc = cafe_tmp_blob_persist(&context, &p->value);
				if (rc != CAFE_OK) {
					p->value.l = 0;
					p->value.p = NULL;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_q:
				assert(eainst->cell_obj != NULL);
				((struct cafe_setval_q *)result.p)->is_set = 1;
				((struct cafe_setval_q *)result.p)->value = args[0].ll;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_Q:
				assert(eainst->cell_obj != NULL);
				((struct cafe_setval_Q *)result.p)->is_set = 1;
				((struct cafe_setval_Q *)result.p)->value = args[0].ull;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_d:
				assert(eainst->cell_obj != NULL);
				((struct cafe_setval_d *)result.p)->is_set = 1;
				((struct cafe_setval_d *)result.p)->value = args[0].d;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_bool:
				assert(eainst->cell_obj != NULL);
				((struct cafe_setval_bool *)result.p)->is_set = 1;
				((struct cafe_setval_bool *)result.p)->value = (bool) args[0].ll;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_s: {
				struct cafe_setval_s *p;
				assert(eainst->cell_obj != NULL);
				p = (struct cafe_setval_s *)result.p;
				if (p->value)
					CAFE_FREE(inst->allocator, p->value);
				p->is_set = 1;
				p->value = args[0].str;
				rc = cafe_tmp_string_persist(&context, &p->value);
				if (rc != CAFE_OK)
					p->value = NULL;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_b: {
				struct cafe_setval_b *p;
				assert(eainst->cell_obj != NULL);
				p = (struct cafe_setval_b *)result.p;
				if (p->value.p)
					CAFE_FREE(inst->allocator, p->value.p);
				p->is_set = 1;
				p->value = args[0].blob;
				rc = cafe_tmp_blob_persist(&context, &p->value);
				if (rc != CAFE_OK) {
					p->value.l = 0;
					p->value.p = NULL;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_q2:
				assert(eainst->cell_obj != NULL);
				*(long long *)result.p = args[0].ll;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_Q2:
				assert(eainst->cell_obj != NULL);
				*(unsigned long long *)result.p = args[0].ull;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_d2:
				assert(eainst->cell_obj != NULL);
				*(double *)result.p = args[0].d;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_bool2:
				assert(eainst->cell_obj != NULL);
				*(bool *)result.p = (bool) args[0].ll;
				break;

			case CAFE_AGGREGATOR_EP_OP_SET_s2: {
				void *p;
				assert(eainst->cell_obj != NULL);
				p = *(char **)result.p;
				if (p)
					CAFE_FREE(inst->allocator, p);
				*(const char **)result.p = args[0].str;
				rc = cafe_tmp_string_persist(&context, result.p);
				if (rc != CAFE_OK)
					*(const char **)result.p = NULL;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_b2: {
				void *p;
				assert(eainst->cell_obj != NULL);
				p = ((struct cafe_vm_blob *)result.p)->p;
				if (p)
					CAFE_FREE(inst->allocator, p);
				*(struct cafe_vm_blob *)result.p = args[0].blob;
				rc = cafe_tmp_blob_persist(&context, result.p);
				if (rc != CAFE_OK) {
					((struct cafe_vm_blob *)result.p)->l = 0;
					((struct cafe_vm_blob *)result.p)->p = NULL;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SET_obj: {
				const cafe_any_value *arg_val = &args[0];
				struct cafe_value_type * val_type;

				val_type = eainst->cell_obj->type;
				if (val_type->kind == CAFE_TYPE_DYNARRAY)
					arg_val = arg_val->p;
				assert(eainst->cell_obj != NULL);
				rc = cafe_value_store_type(val_type,
							   result.p,
							   arg_val,
							   inst->allocator);
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SUM_q:
				assert(eainst->cell_obj != NULL);
				*(long long *)result.p += args[0].ll;
				break;

			case CAFE_AGGREGATOR_EP_OP_SUM_Q:
				assert(eainst->cell_obj != NULL);
				*(unsigned long long *)result.p += args[0].ull;
				break;

			case CAFE_AGGREGATOR_EP_OP_SUM_d:
				assert(eainst->cell_obj != NULL);
				*(double *)result.p += args[0].d;
				break;

			case CAFE_AGGREGATOR_EP_OP_MIN_q2: {
				struct cafe_setval_q *s =
					(struct cafe_setval_q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				if (!((struct cafe_setval_q *)result.p)->is_set) {
					((struct cafe_setval_q *)result.p)->is_set = 1;
					((struct cafe_setval_q *)result.p)->value = s->value;
				} else if (((struct cafe_setval_q *)result.p)->value > s->value) {
					((struct cafe_setval_q *)result.p)->value = s->value;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_MIN_Q2: {
				struct cafe_setval_Q *s =
					(struct cafe_setval_Q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				if (!((struct cafe_setval_Q *)result.p)->is_set) {
					((struct cafe_setval_Q *)result.p)->is_set = 1;
					((struct cafe_setval_Q *)result.p)->value = s->value;
				} else if (((struct cafe_setval_Q *)result.p)->value > s->value) {
					((struct cafe_setval_Q *)result.p)->value = s->value;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_MIN_d2: {
				struct cafe_setval_d *s =
					(struct cafe_setval_d *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				if (!((struct cafe_setval_d *)result.p)->is_set) {
					((struct cafe_setval_d *)result.p)->is_set = 1;
					((struct cafe_setval_d *)result.p)->value = s->value;
				} else if(!((struct cafe_setval_d *)result.p)->value > s->value) {
					((struct cafe_setval_d *)result.p)->value = s->value;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_MIN_q:
				assert(eainst->cell_obj != NULL);
				if (!((struct cafe_setval_q *)result.p)->is_set) {
					((struct cafe_setval_q *)result.p)->is_set = 1;
					((struct cafe_setval_q *)result.p)->value = args[0].ll;
				} else if (((struct cafe_setval_q *)result.p)->value > args[0].ll) {
					((struct cafe_setval_q *)result.p)->value = args[0].ll;
				}
				break;

			case CAFE_AGGREGATOR_EP_OP_MIN_Q:
				assert(eainst->cell_obj != NULL);
				if (!((struct cafe_setval_Q *)result.p)->is_set) {
					((struct cafe_setval_Q *)result.p)->is_set = 1;
					((struct cafe_setval_Q *)result.p)->value = args[0].ull;
				} else if (((struct cafe_setval_Q *)result.p)->value > args[0].ull) {
					((struct cafe_setval_Q *)result.p)->value = args[0].ull;
				}
				break;

			case CAFE_AGGREGATOR_EP_OP_MIN_d:
				assert(eainst->cell_obj != NULL);
				if (!((struct cafe_setval_d *)result.p)->is_set) {
					((struct cafe_setval_d *)result.p)->is_set = 1;
					((struct cafe_setval_d *)result.p)->value = args[0].d;
				} else if(!((struct cafe_setval_d *)result.p)->value > args[0].d) {
					((struct cafe_setval_d *)result.p)->value = args[0].d;
				}
				break;

			case CAFE_AGGREGATOR_EP_OP_MAX_q2: {
				struct cafe_setval_q *s =
					(struct cafe_setval_q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				if (!((struct cafe_setval_q *)result.p)->is_set) {
					((struct cafe_setval_q *)result.p)->is_set = 1;
					((struct cafe_setval_q *)result.p)->value = s->value;
				} else if (((struct cafe_setval_q *)result.p)->value < s->value) {
					((struct cafe_setval_q *)result.p)->value = s->value;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_MAX_Q2: {
				struct cafe_setval_Q *s =
					(struct cafe_setval_Q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				if (!((struct cafe_setval_Q *)result.p)->is_set) {
					((struct cafe_setval_Q *)result.p)->is_set = 1;
					((struct cafe_setval_Q *)result.p)->value = s->value;
				} else if (((struct cafe_setval_Q *)result.p)->value < s->value) {
					((struct cafe_setval_Q *)result.p)->value = s->value;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_MAX_d2: {
				struct cafe_setval_d *s =
					(struct cafe_setval_d *)args[0].p;
				assert(eainst->cell_obj != NULL);
				if (!s || !s->is_set)
					break;
				if (!((struct cafe_setval_d *)result.p)->is_set) {
					((struct cafe_setval_d *)result.p)->is_set = 1;
					((struct cafe_setval_d *)result.p)->value = s->value;
				} else if(!((struct cafe_setval_d *)result.p)->value < s->value) {
					((struct cafe_setval_d *)result.p)->value = s->value;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_MAX_q:
				assert(eainst->cell_obj != NULL);
				if (!((struct cafe_setval_q *)result.p)->is_set) {
					((struct cafe_setval_q *)result.p)->is_set = 1;
					((struct cafe_setval_q *)result.p)->value = args[0].ll;
				} else if (((struct cafe_setval_q *)result.p)->value < args[0].ll) {
					((struct cafe_setval_q *)result.p)->value = args[0].ll;
				}
				break;

			case CAFE_AGGREGATOR_EP_OP_MAX_Q:
				assert(eainst->cell_obj != NULL);
				if (!((struct cafe_setval_Q *)result.p)->is_set) {
					((struct cafe_setval_Q *)result.p)->is_set = 1;
					((struct cafe_setval_Q *)result.p)->value = args[0].ull;
				} else if (((struct cafe_setval_Q *)result.p)->value < args[0].ull) {
					((struct cafe_setval_Q *)result.p)->value = args[0].ull;
				}
				break;

			case CAFE_AGGREGATOR_EP_OP_MAX_d:
				assert(eainst->cell_obj != NULL);
				if (!((struct cafe_setval_d *)result.p)->is_set) {
					((struct cafe_setval_d *)result.p)->is_set = 1;
					((struct cafe_setval_d *)result.p)->value = args[0].d;
				} else if(!((struct cafe_setval_d *)result.p)->value < args[0].d) {
					((struct cafe_setval_d *)result.p)->value = args[0].d;
				}
				break;

			case CAFE_AGGREGATOR_EP_OP_AVG_q2: {
				struct cafe_avgval_q *s =
					(struct cafe_avgval_q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				((struct cafe_avgval_q *)result.p)->count += s->count;
				((struct cafe_avgval_q *)result.p)->total += s->total;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_AVG_Q2: {
				struct cafe_avgval_Q *s =
					(struct cafe_avgval_Q *)args[0].p;
				assert(eainst->cell_obj != NULL);
				((struct cafe_avgval_Q *)result.p)->count += s->count;
				((struct cafe_avgval_Q *)result.p)->total += s->total;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_AVG_d2: {
				struct cafe_avgval_d *s =
					(struct cafe_avgval_d *)args[0].p;
				assert(eainst->cell_obj != NULL);
				((struct cafe_avgval_d *)result.p)->count += s->count;
				((struct cafe_avgval_d *)result.p)->total += s->total;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_AVG_q:
				assert(eainst->cell_obj != NULL);
				((struct cafe_avgval_q *)result.p)->count++;
				((struct cafe_avgval_q *)result.p)->total += args[0].ll;
				break;

			case CAFE_AGGREGATOR_EP_OP_AVG_Q:
				assert(eainst->cell_obj != NULL);
				((struct cafe_avgval_Q *)result.p)->count++;
				((struct cafe_avgval_Q *)result.p)->total += args[0].ull;
				break;

			case CAFE_AGGREGATOR_EP_OP_AVG_d:
				assert(eainst->cell_obj != NULL);
				((struct cafe_avgval_d *)result.p)->count++;
				((struct cafe_avgval_d *)result.p)->total += args[0].d;
				break;

			case CAFE_AGGREGATOR_EP_OP_CLEAR:
				assert(eainst->cell_obj != NULL);
				break;

			case CAFE_AGGREGATOR_EP_OP_DELETE:
				assert(eainst->cell_obj != NULL);
				assert(eainst->args != 0);
				assert(dict == NULL);

				context.event_data = curdata;
				dict = (struct cafe_dict_struct *)result.p;
				value = cafe_dict_first(dict, &iter);
				while (value != NULL) {
					struct cafe_vm_expr *e = eainst->args;

					curdata[0] = value->u.data;
					assert(e->kind == CAFE_EXPR_LIST);
					e = e->u.e;
					assert(e != NULL);
					assert(e->u.elem.next == NULL);
					rc = cafe_expr_eval_value(&context, e->u.elem.e, &args[0]);
					if (rc != CAFE_OK)
						break;
					if (args[0].ll)
						cafe_dict_remove(dict, value);
					value = cafe_dict_next(dict, &iter);
				}
				dict = NULL;
				break;

			case CAFE_AGGREGATOR_EP_OP_RESET:{
				struct cafe_list_head *tmp;

				assert(eainst->cell_obj == NULL);
				cafe_list_foreach_safe(list_item, &ai->table_list_head, tmp) {
					struct table_inst *ti = TABLE_LIST_NODE_ENTRY(list_item);
					if (ti->is_deleted) {
						assert(ai->pending_delete);
						table_inst_delete(inst, ti, 0);
					} else {
						if (ti->is_symresolver) {
							assert(ai->symresolver);
							symresolve_funcs->actions_reset(
								ai->symresolver);
						}
						cafe_value_clear(ti->value);
					}
				}
				ai->pending_delete = false;
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SENDTOSELF: {
				unsigned int i;
				void **event_data = ai->temp_event_data;

				assert(eainst->event_info->event->num_types <= sizeof(ai->temp_event_data) / sizeof(void *));
				for (i = 0; i < eainst->event_info->event->num_types; i++)
					event_data[i] = args[i].p;

				/* Dispatch event recursively. */
				action_event_handler(eainst->event_info, event_data);
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_ECHO: {
				print_event(info, data);
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_SNAPSHOT: {
				struct cafe_list_head *tmp;
				unsigned int num_data;
				unsigned int idx = ai->pending_delete ?
					PENDING_DATA_EVENT: SNAPSHOT_EVENT;
				void **event_data = (void **)ai->temp_type_list;

				assert(eainst->cell_obj == NULL);
				assert(ai->event[idx]->num_types <= ai->temp_type_max);
				if (ai->symresolver_table)
					symresolve_funcs->mm_gen_delta(ai->symresolver, ai->symresolver_table->value->u.data);
				num_data = 0;
				event_data[num_data++] = NULL;
				cafe_list_foreach(list_item, &ai->table_list_head) {
					struct table_inst *ti = TABLE_LIST_NODE_ENTRY(list_item);
					if (!ti->is_private) {
						event_data[num_data++] = ti->value->u.data;
					}
				}
				assert(num_data == ai->event[idx]->num_types);
				cafe_dispatch_event(&ai->sink_actions[idx], event_data);
				if (ai->symresolver_table)
					symresolve_funcs->mm_clear_delta(ai->symresolver, ai->symresolver_table->value->u.data);
				cafe_list_foreach_safe(list_item, &ai->table_list_head, tmp) {
					struct table_inst *ti = TABLE_LIST_NODE_ENTRY(list_item);
					if (ti->is_deleted) {
						assert(ai->pending_delete);
						table_inst_delete(inst, ti, 0);
						num_data--;
					} else if (!ti->is_persistent) {
						cafe_value_clear(ti->value);
					}
				}
				ai->pending_delete = false;
				assert(num_data == ai->event[SNAPSHOT_EVENT]->num_types);
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_AGGREGATE: {
				cafe_aggregator_ep_op_tag *subops;
				cafe_aggregator_ep_op_tag *subops2;
				cafe_any_value *arg;
				unsigned int j;
				bool do_cumulative_rows = false;
				void *row;

				assert(eainst->cell_obj != NULL);
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "aggregate operation\n");
				subops = eainst->subops;
				subops2 = eainst->subops2;

			next_cumulative:
				arg = args;
				if (eainst->cell_dict) {
					struct cafe_value_data *val =
						cafe_dict_get(result.p, arg);
					if (val == NULL)
						continue;
					arg += eainst->cell_dict->elements;
					row = val->u.data;
				} else {
					row = result.p;
				}
				for (j = 0; j < eainst->cell_type->elements; j++) {
					struct cafe_vm_attr *attr =
						eainst->cell_type->attributes + j;
					bool cell_set = false;
					void *cell;
					if (attr->kind != CAFE_ATTR_VAR) {
						assert(attr->kind == CAFE_ATTR_VAR ||
						       attr->kind == CAFE_ATTR_KEY ||
						       attr->kind == CAFE_ATTR_CONST ||
						       attr->kind == CAFE_ATTR_PARAM);
						continue;
					}
					cell = (char *)row + attr->offset;
					switch (*subops++) {
					case CAFE_AGGREGATOR_EP_OP_SET_q3: {
						struct cafe_setval_q *s =
							(struct cafe_setval_q *)arg->p;
						if (!s || !s->is_set)
							break;
						((struct cafe_setval_q *)cell)->is_set = 1;
						((struct cafe_setval_q *)cell)->value = s->value;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_Q3: {
						struct cafe_setval_Q *s =
							(struct cafe_setval_Q *)arg->p;
						if (!s || !s->is_set)
							break;
						((struct cafe_setval_Q *)cell)->is_set = 1;
						((struct cafe_setval_Q *)cell)->value = s->value;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_d3: {
						struct cafe_setval_d *s =
							(struct cafe_setval_d *)arg->p;
						if (!s || !s->is_set)
							break;
						((struct cafe_setval_d *)cell)->is_set = 1;
						((struct cafe_setval_d *)cell)->value = s->value;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_bool3: {
						struct cafe_setval_bool *s =
							(struct cafe_setval_bool *)arg->p;
						if (!s || !s->is_set)
							break;
						((struct cafe_setval_bool *)cell)->is_set = 1;
						((struct cafe_setval_bool *)cell)->value = s->value;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_s3: {
						struct cafe_setval_s *s =
							(struct cafe_setval_s *)arg->p;
						struct cafe_setval_s *p;
						if (!s || !s->is_set)
							break;
						p = (struct cafe_setval_s *)cell;
						if (p->value)
							CAFE_FREE(inst->allocator, p->value);
						p->is_set = 1;
						p->value = s->value;
						rc = cafe_tmp_string_persist(&context, &p->value);
						if (rc != CAFE_OK)
							p->value = NULL;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_b3: {
						struct cafe_setval_b *s =
							(struct cafe_setval_b *)arg->p;
						struct cafe_setval_b *p;
						if (!s || !s->is_set)
							break;
						p = (struct cafe_setval_b *)cell;
						if (p->value.p)
							CAFE_FREE(inst->allocator, p->value.p);
						p->is_set = 1;
						p->value = s->value;
						rc = cafe_tmp_blob_persist(&context, &p->value);
						if (rc != CAFE_OK) {
							p->value.l = 0;
							p->value.p = NULL;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_q:
						((struct cafe_setval_q *)cell)->is_set = 1;
						((struct cafe_setval_q *)cell)->value = arg->ll;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_Q:
						((struct cafe_setval_Q *)cell)->is_set = 1;
						((struct cafe_setval_Q *)cell)->value = arg->ull;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_d:
						((struct cafe_setval_d *)cell)->is_set = 1;
						((struct cafe_setval_d *)cell)->value = arg->d;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_bool:
						((struct cafe_setval_bool *)cell)->is_set = 1;
						((struct cafe_setval_bool *)cell)->value = (bool) arg->ll;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_s: {
						struct cafe_setval_s *p;
						p = (struct cafe_setval_s *)cell;
						if (p->value)
							CAFE_FREE(inst->allocator, p->value);
						p->is_set = 1;
						p->value = arg->str;
						rc = cafe_tmp_string_persist(&context, &p->value);
						if (rc != CAFE_OK)
							p->value = NULL;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_b: {
						struct cafe_setval_b *p;
						p = (struct cafe_setval_b *)cell;
						if (p->value.p)
							CAFE_FREE(inst->allocator, p->value.p);
						p->is_set = 1;
						p->value = arg->blob;
						rc = cafe_tmp_blob_persist(&context, &p->value);
						if (rc != CAFE_OK) {
							p->value.l = 0;
							p->value.p = NULL;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_q2:
						*(long long *)cell = arg->ll;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_Q2:
						*(unsigned long long *)cell = arg->ull;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_d2:
						*(double *)cell = arg->d;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_bool2:
						*(bool *)cell = (bool) arg->ll;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SET_s2: {
						void *p;
						p = *(char **)cell;
						if (p)
							CAFE_FREE(inst->allocator, p);
						*(const char **)cell = arg->str;
						rc = cafe_tmp_string_persist(&context, cell);
						if (rc != CAFE_OK)
							*(const char **)cell = NULL;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SET_b2: {
						void *p;
						p = ((struct cafe_vm_blob *)cell)->p;
						if (p)
							CAFE_FREE(inst->allocator, p);
						*(struct cafe_vm_blob *)cell = arg->blob;
						rc = cafe_tmp_blob_persist(&context, cell);
						if (rc != CAFE_OK) {
							((struct cafe_vm_blob *)cell)->l = 0;
							((struct cafe_vm_blob *)cell)->p = NULL;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_SUM_q:
						if (arg->ll == 0)
							break;
						*(long long *)cell += arg->ll;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SUM_Q:
						if (arg->ull == 0)
							break;
						*(unsigned long long *)cell += arg->ull;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SUM_d:
						if (arg->d == 0)
							break;
						*(double *)cell += arg->d;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_MIN_q2: {
						struct cafe_setval_q *s =
							(struct cafe_setval_q *)arg->p;
						if (!s || !s->is_set)
							break;
						if (!((struct cafe_setval_q *)cell)->is_set) {
							((struct cafe_setval_q *)cell)->is_set = 1;
							((struct cafe_setval_q *)cell)->value = s->value;
						} else if (((struct cafe_setval_q *)cell)->value > s->value) {
							((struct cafe_setval_q *)cell)->value = s->value;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_MIN_Q2: {
						struct cafe_setval_Q *s =
							(struct cafe_setval_Q *)arg->p;
						if (!s || !s->is_set)
							break;
						if (!((struct cafe_setval_Q *)cell)->is_set) {
							((struct cafe_setval_Q *)cell)->is_set = 1;
							((struct cafe_setval_Q *)cell)->value = s->value;
						} else if (((struct cafe_setval_Q *)cell)->value > s->value) {
							((struct cafe_setval_Q *)cell)->value = s->value;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_MIN_d2: {
						struct cafe_setval_d *s =
							(struct cafe_setval_d *)arg->p;
						if (!s || !s->is_set)
							break;
						if (!((struct cafe_setval_d *)cell)->is_set) {
							((struct cafe_setval_d *)cell)->is_set = 1;
							((struct cafe_setval_d *)cell)->value = s->value;
						} else if(!((struct cafe_setval_d *)cell)->value > s->value) {
							((struct cafe_setval_d *)cell)->value = s->value;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_MIN_q:
						if (!((struct cafe_setval_q *)cell)->is_set) {
							((struct cafe_setval_q *)cell)->is_set = 1;
							((struct cafe_setval_q *)cell)->value = arg->ll;
						} else if (((struct cafe_setval_q *)cell)->value > arg->ll) {
							((struct cafe_setval_q *)cell)->value = arg->ll;
						}
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_MIN_Q:
						if (!((struct cafe_setval_Q *)cell)->is_set) {
							((struct cafe_setval_Q *)cell)->is_set = 1;
							((struct cafe_setval_Q *)cell)->value = arg->ull;
						} else if (((struct cafe_setval_Q *)cell)->value > arg->ull) {
							((struct cafe_setval_Q *)cell)->value = arg->ull;
						}
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_MIN_d:
						if (!((struct cafe_setval_d *)cell)->is_set) {
							((struct cafe_setval_d *)cell)->is_set = 1;
							((struct cafe_setval_d *)cell)->value = arg->d;
						} else if(!((struct cafe_setval_d *)cell)->value > arg->d) {
							((struct cafe_setval_d *)cell)->value = arg->d;
						}
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_MAX_q2: {
						struct cafe_setval_q *s =
							(struct cafe_setval_q *)arg->p;
						if (!s || !s->is_set)
							break;
						if (!((struct cafe_setval_q *)cell)->is_set) {
							((struct cafe_setval_q *)cell)->is_set = 1;
							((struct cafe_setval_q *)cell)->value = s->value;
						} else if (((struct cafe_setval_q *)cell)->value < s->value) {
							((struct cafe_setval_q *)cell)->value = s->value;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_MAX_Q2: {
						struct cafe_setval_Q *s =
							(struct cafe_setval_Q *)arg->p;
						if (!s || !s->is_set)
							break;
						if (!((struct cafe_setval_Q *)cell)->is_set) {
							((struct cafe_setval_Q *)cell)->is_set = 1;
							((struct cafe_setval_Q *)cell)->value = s->value;
						} else if (((struct cafe_setval_Q *)cell)->value < s->value) {
							((struct cafe_setval_Q *)cell)->value = s->value;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_MAX_d2: {
						struct cafe_setval_d *s =
							(struct cafe_setval_d *)arg->p;
						if (!s || !s->is_set)
							break;
						if (!((struct cafe_setval_d *)cell)->is_set) {
							((struct cafe_setval_d *)cell)->is_set = 1;
							((struct cafe_setval_d *)cell)->value = s->value;
						} else if(!((struct cafe_setval_d *)cell)->value < s->value) {
							((struct cafe_setval_d *)cell)->value = s->value;
						}
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_MAX_q:
						if (!((struct cafe_setval_q *)cell)->is_set) {
							((struct cafe_setval_q *)cell)->is_set = 1;
							((struct cafe_setval_q *)cell)->value = arg->ll;
						} else if (((struct cafe_setval_q *)cell)->value < arg->ll) {
							((struct cafe_setval_q *)cell)->value = arg->ll;
						}
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_MAX_Q:
						if (!((struct cafe_setval_Q *)cell)->is_set) {
							((struct cafe_setval_Q *)cell)->is_set = 1;
							((struct cafe_setval_Q *)cell)->value = arg->ull;
						} else if (((struct cafe_setval_Q *)cell)->value < arg->ull) {
							((struct cafe_setval_Q *)cell)->value = arg->ull;
						}
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_MAX_d:
						if (!((struct cafe_setval_d *)cell)->is_set) {
							((struct cafe_setval_d *)cell)->is_set = 1;
							((struct cafe_setval_d *)cell)->value = arg->d;
						} else if(!((struct cafe_setval_d *)cell)->value < arg->d) {
							((struct cafe_setval_d *)cell)->value = arg->d;
						}
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_AVG_q2: {
						struct cafe_avgval_q *s =
							(struct cafe_avgval_q *)arg->p;
						if (s->count == 0)
							break;
						((struct cafe_avgval_q *)cell)->count += s->count;
						((struct cafe_avgval_q *)cell)->total += s->total;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_AVG_Q2: {
						struct cafe_avgval_Q *s =
							(struct cafe_avgval_Q *)arg->p;
						if (s->count == 0)
							break;
						((struct cafe_avgval_Q *)cell)->count += s->count;
						((struct cafe_avgval_Q *)cell)->total += s->total;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_AVG_d2: {
						struct cafe_avgval_d *s =
							(struct cafe_avgval_d *)arg->p;
						if (s->count == 0)
							break;
						((struct cafe_avgval_d *)cell)->count += s->count;
						((struct cafe_avgval_d *)cell)->total += s->total;
						cell_set = true;
						break;
					}

					case CAFE_AGGREGATOR_EP_OP_AVG_q:
						((struct cafe_avgval_q *)cell)->count++;
						((struct cafe_avgval_q *)cell)->total += arg->ll;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_AVG_Q:
						((struct cafe_avgval_Q *)cell)->count++;
						((struct cafe_avgval_Q *)cell)->total += arg->ull;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_AVG_d:
						((struct cafe_avgval_d *)cell)->count++;
						((struct cafe_avgval_d *)cell)->total += arg->d;
						cell_set = true;
						break;

					case CAFE_AGGREGATOR_EP_OP_SKIP:
						break;

					case CAFE_AGGREGATOR_EP_OP_COUNT_L:
					case CAFE_AGGREGATOR_EP_OP_COUNT_Q:
					case CAFE_AGGREGATOR_EP_OP_SET_obj:
					case CAFE_AGGREGATOR_EP_OP_CLEAR:
					case CAFE_AGGREGATOR_EP_OP_DELETE:
					case CAFE_AGGREGATOR_EP_OP_RESET:
					case CAFE_AGGREGATOR_EP_OP_SENDTOSELF:
					case CAFE_AGGREGATOR_EP_OP_SNAPSHOT:
					case CAFE_AGGREGATOR_EP_OP_AGGREGATE:
					case CAFE_AGGREGATOR_EP_OP_STOP_DISPATCH:
					case CAFE_AGGREGATOR_EP_OP_MM_UPDATE:
					case CAFE_AGGREGATOR_EP_OP_MM_DELTA:
					case CAFE_AGGREGATOR_EP_OP_MM_CLONE:
					case CAFE_AGGREGATOR_EP_OP_MM_CLEAR:
					case CAFE_AGGREGATOR_EP_OP_MM_CLEAR2:
					case CAFE_AGGREGATOR_EP_OP_MM_DELETE:
					case CAFE_AGGREGATOR_EP_OP_RESOLVE:
					case CAFE_AGGREGATOR_EP_OP_MIRROR:
					case CAFE_AGGREGATOR_EP_OP_ECHO:
					case CAFE_AGGREGATOR_EP_OP_NONE:
					default:
						/* coverity[array_null] */
						assert(!"unexpected sub action operation");
					}
					if (subops2 != NULL &&
					    *subops2++ != CAFE_AGGREGATOR_EP_OP_SKIP)
						do_cumulative_rows |= cell_set;
					arg++;
				}
				if (do_cumulative_rows && flatten_dary &&
				    ++flatten_index < flatten_dary->l) {
					struct cafe_vm_expr *e;

					if (flatten_index == 1) {
						seen_value_clear(&ai->seen_value, eainst->flatten_type);
						seen_value_is_new(inst->allocator, &ai->seen_value, curdata[1]);
					}
					do {
						curdata[1] = (char *)curdata[1] + eainst->flatten_type->size;
					} while (!seen_value_is_new(inst->allocator, &ai->seen_value, curdata[1]) &&
						 ++flatten_index < flatten_dary->l);
					if (flatten_index >= flatten_dary->l)
						break;

					/* Reevaluate keys for the new row of the flatten table */
					for (e = eainst->args->u.e, j = 0;
					     e != NULL; e = e->u.elem.next, j++) {
						if (j < eainst->flatten_arg_start)
							continue;
						if (j >= eainst->flatten_arg_end)
							break;
						assert(j < ai->temp_arg_max);
						rc = cafe_expr_eval_value(&context, e->u.elem.e, &args[j]);
						if (rc != CAFE_OK)
							break;
					}
					if (j < eainst->flatten_arg_end)
						continue;
					subops = eainst->subops2;
					subops2 = NULL;
					goto next_cumulative;
				}
				break;
			}

			case CAFE_AGGREGATOR_EP_OP_STOP_DISPATCH:
				stop_dispatch = true;
				break;

			case CAFE_AGGREGATOR_EP_OP_MM_UPDATE:
				assert(eainst->cell_obj != NULL);
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "memory map update operation\n");
				if (ai->symresolver == NULL)
					continue;
				rc = symresolve_funcs->mm_update(
					ai->symresolver,
					args[0].ull, /* address_space_uid */
					args[1].ull, /* start_addr */
					args[2].ull, /* size */
					args[3].ull, /* file_offset */
					args[5].str, /* file_name */
					(bool)args[4].ull); /* map */
				if (rc != CAFE_OK && (cafe_aggregator_debug & 1))
					fprintf(stderr, "memory map update operation error %d\n", rc);
				break;

			case CAFE_AGGREGATOR_EP_OP_MM_DELTA:
				assert(eainst->cell_obj != NULL);
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "memory map delta operation\n");
				if (ai->symresolver == NULL)
					continue;
				rc = symresolve_funcs->mm_delta(
					ai->symresolver, args[0].p);
				if (rc != CAFE_OK && (cafe_aggregator_debug & 1))
					fprintf(stderr, "memory map delta operation error %d\n", rc);
				do_mm_flush = 1;
				break;

			case CAFE_AGGREGATOR_EP_OP_MM_CLONE:
				assert(eainst->cell_obj != NULL);
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "memory map clone operation\n");
				if (ai->symresolver == NULL)
					continue;
				rc = symresolve_funcs->address_space_clone(
					ai->symresolver, args[0].ull, args[1].ull);
				if (rc != CAFE_OK && (cafe_aggregator_debug & 1))
					fprintf(stderr, "memory map clone operation error %d\n", rc);
				break;

			case CAFE_AGGREGATOR_EP_OP_MM_CLEAR:
				assert(eainst->cell_obj != NULL);
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "memory map clear operation\n");
				if (ai->symresolver == NULL)
					continue;
				rc = symresolve_funcs->address_space_clear(
					ai->symresolver, args[0].ull, 0);
				if (rc != CAFE_OK && (cafe_aggregator_debug & 1))
					fprintf(stderr, "memory map clear operation error %d\n", rc);
				break;

			case CAFE_AGGREGATOR_EP_OP_MM_CLEAR2:
				assert(eainst->cell_obj != NULL);
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "memory map clear operation\n");
				if (ai->symresolver == NULL)
					continue;
				rc = symresolve_funcs->address_space_clear(
					ai->symresolver, args[0].ull, args[1].ull);
				if (rc != CAFE_OK && (cafe_aggregator_debug & 1))
					fprintf(stderr, "memory map clear operation error %d\n", rc);
				break;

			case CAFE_AGGREGATOR_EP_OP_MM_DELETE:
				assert(eainst->cell_obj != NULL);
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "memory map delete operation\n");
				if (ai->symresolver == NULL)
					continue;
				rc = symresolve_funcs->address_space_delete(
					ai->symresolver, args[0].ull);
				if (rc != CAFE_OK && (cafe_aggregator_debug & 1))
					fprintf(stderr, "memory map delete operation error %d\n", rc);
				break;

			case CAFE_AGGREGATOR_EP_OP_RESOLVE: {
				struct cafe_value_data *val;
				struct cafe_vm_attr *attr;
				const cafe_any_value *arg;
				struct table_inst *dest_table = result.p;
				struct cafe_dict_struct *dest_dict = (struct cafe_dict_struct *)
					dest_table->value->u.data;

				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "resolve operation: %s\n",
						dest_table->type->ident->name);

				val = cafe_dict_get(dest_dict, args);
				if (val == NULL) {
					fprintf(stderr, "Failed to get table row when resolving symbols\n");
					continue;
				}

				arg = args + dest_dict->type->elements;
				cafe_type_attr_foreach(attr,
						       eainst->cell_obj->type)
				{
					void *cell;
					const cafe_any_value *arg_val = arg;
					struct cafe_value_type * val_type;
					if (attr->kind != CAFE_ATTR_VAR)
						continue;

					cell = val->u.data + attr->offset;
					val_type = cafe_attr_type(attr);
					if (val_type->kind == CAFE_TYPE_DYNARRAY)
						arg_val = arg_val->p;
					rc = cafe_value_store_type(val_type,
								   cell, arg_val,
								   inst->allocator);
					if (rc != CAFE_OK)
						fprintf(stderr, "Error copying data into symresolved row: %s\n", cafe_strerror(rc));

					++arg;
				}

				break;
			}

			case CAFE_AGGREGATOR_EP_OP_MIRROR:
			{
				struct table_inst *dest_table = result.p;
				if (cafe_aggregator_debug & 2)
					fprintf(stderr, "mirroring into %s\n", dest_table->type->ident->name);
				rc = cafe_value_store_type(dest_table->value->type,
							   dest_table->value->u.data, args,
							   inst->allocator);
				if (rc != CAFE_OK)
					fprintf(stderr, "Error copying data into mirrored data: %s\n", cafe_strerror(rc));

				break;
			}
			case CAFE_AGGREGATOR_EP_OP_NONE:
			case CAFE_AGGREGATOR_EP_OP_SKIP:
				break;

			default:
				/* coverity[array_null] */
				assert(!"unexpected action operation");
			}
		} while (dict != NULL &&
			 (value = cafe_dict_next(dict, &iter)) != NULL &&
			 (curdata[0] = value->u.data, 1));
	}
	if (do_mm_flush)
		symresolve_funcs->mm_flush(ai->symresolver);
	if (stop_dispatch &&
	    inst->input_ctx != NULL &&
	    inst->input_ctx != CAFE_EXEC_CTX_ANY &&
	    inst->input_ctx != CAFE_EXEC_CTX_SESSION &&
	    inst->input_ctx->spec->stop_dispatch != NULL)
		inst->input_ctx->spec->stop_dispatch(inst->input_ctx);
	cafe_expr_eval_context_delete(&context);
	eamap->active = false;
}

static int process_settings(
	struct cafe_ep_inst *inst)
{
	struct workarea *w = (struct workarea *)inst->workarea->u.data;
	struct cafe_value_type *type;
	unsigned int i;
	int rc;

	if (w->del_tables.l) {
		char **del_tables = w->del_tables.p;
		assert(del_tables != NULL);
		for (i = 0; i < w->del_tables.l; i++) {
			rc = cafe_type_parse(inst->session->ns,
					     del_tables[i], &type);
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc, "Invalid type for table delete <%s>",
					       del_tables[i]);
				return rc;
			}

			rc = cafe_aggregator_ep_table_delete(inst, type);
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc, "Table delete failed <%s>",
					       del_tables[i]);
				return rc;
			}
			if (cafe_aggregator_debug & 2)
				fprintf(stderr, "Delete table <%s>\n",
					del_tables[i]);
		}
	}

	if (w->del_actions.l) {
		struct cafe_vm_dynarray *del_actions = w->del_actions.p;
		assert(del_actions != NULL);
		for (i = 0; i < w->del_actions.l; i++) {
			char **del_action;
			if (del_actions[i].l != 5)
				return CAFE_ERROR;
			del_action = del_actions[i].p;
			assert(del_action != NULL);
			rc = cafe_aggregator_ep_action_del(
				inst,
				del_action[0],  /* trigger */
				del_action[1],  /* filter */
				del_action[2],  /* cell */
				del_action[3],  /* operation */
				del_action[4]); /* args */
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc, "Action delete failed <%s> <%s> <%s> <%s> <%s>",
					       del_action[0], del_action[1], del_action[2],
					       del_action[3], del_action[4]);
				return rc;
			}

			if (cafe_aggregator_debug & 2)
				(void)fprintf(stderr, "Action delete <%s> <%s> <%s> <%s> <%s>\n",
					del_action[0], del_action[1], del_action[2],
					del_action[3], del_action[4]);
		}
	}

	if (w->add_types.l) {
		char **add_types = w->add_types.p;
		assert(add_types != NULL);
		for (i = 0; i < w->add_types.l; i++) {
			rc = cafe_type_parse(inst->session->ns,
					     add_types[i], &type);
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc, "Add type failed <%s>",
					       add_types[i]);
				return rc;
			}
			if (cafe_aggregator_debug & 2)
				fprintf(stderr, "Add type <%s>\n", add_types[i]);
		}
	}

	if (w->add_tables.l) {
		char **add_tables = w->add_tables.p;
		assert(add_tables != NULL);
		for (i = 0; i < w->add_tables.l; i++) {
			rc = cafe_type_parse(inst->session->ns,
					     add_tables[i], &type);
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc, "Invalid type for table add <%s>",
					       add_tables[i]);
				return rc;
			}

			rc = cafe_aggregator_ep_table_create(inst, type);
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc, "Table add failed <%s>",
					       add_tables[i]);
				return rc;
			}
			if (cafe_aggregator_debug & 2)
				fprintf(stderr, "Table add <%s>\n",
					add_tables[i]);
		}
	}

	if (w->add_actions.l) {
		struct cafe_vm_dynarray *add_actions = w->add_actions.p;
		assert(add_actions != NULL);
		for (i = 0; i < w->add_actions.l; i++) {
			char **add_action;
			if (add_actions[i].l != 5) {
				cafe_error_set(NULL, CAFE_ERROR, "add_actions.length = %"PRIsize_t"", add_actions[i].l);
				return CAFE_ERROR;
			}
			add_action = add_actions[i].p;
			assert(add_action != NULL);

			rc = cafe_aggregator_ep_action_add(
				inst,
				add_action[0],  /* trigger */
				add_action[1],  /* filter */
				add_action[2],  /* cell */
				add_action[3],  /* operation */
				add_action[4]); /* args */
			if (rc != CAFE_OK) {
				cafe_error_set(NULL, rc, "Action add failed: <%s> <%s> <%s> <%s> <%s>",
						  add_action[0], add_action[1], add_action[2],
						  add_action[3], add_action[4]);
				return rc;
			}
			if (cafe_aggregator_debug & 2)
				(void)fprintf(stderr, "Action add: <%s> <%s> <%s> <%s> <%s>\n",
					add_action[0], add_action[1], add_action[2],
					add_action[3], add_action[4]);
		}
	}

	return CAFE_OK;
}

static int aggregator_ep_create(
	struct cafe_ep_inst *inst)
{
	struct aggregator_inst *ai;
	struct cafe_value_type *type;
	struct cafe_event_descr *event;
	unsigned int i;
	int rc;
	char *debug;

	if ((debug = getenv("CAFE_AGGREGATOR_DEBUG")) != NULL)
		parse_l(debug, &cafe_aggregator_debug);

	if ((ai = CAFE_ALLOC(inst->allocator, sizeof *ai)) == NULL)
		return CAFE_NOMEM;
	memset(ai, 0, sizeof *ai);
	ai->sink_actions[SNAPSHOT_EVENT] = CAFE_EAL_HANDLE_NULL;
	ai->sink_actions[PENDING_DATA_EVENT] = CAFE_EAL_HANDLE_NULL;
	inst->client_data = ai;

	/* Create start event info */
	rc = cafe_type_parse(inst->session->ns,
			     cafe_start_VT_DEF, &type);
	if (rc != CAFE_OK)
		goto error;
	rc = cafe_cfg_get_event(inst->session, 1, &type, &event);
	if (rc != CAFE_OK)
		goto error;
	rc = cafe_ep_inst_event_action_info_get2(
		inst, event, event, NULL, &ai->start_event);
	if (rc != CAFE_OK)
		goto error;

	/* Create stop event info */
	rc = cafe_type_parse(inst->session->ns,
			     cafe_stop_VT_DEF, &type);
	if (rc != CAFE_OK)
		goto error;
	rc = cafe_cfg_get_event(inst->session, 1, &type, &event);
	if (rc != CAFE_OK)
		goto error;
	rc = cafe_ep_inst_event_action_info_get2(
		inst, event, event, NULL, &ai->stop_event);
	if (rc != CAFE_OK)
		goto error;

	/* Type of memory map delta table */
	rc = cafe_type_parse(inst->session->ns,
			     CAFE_MM_DELTA_REF, &ai->mm_delta_type);
	if (rc != CAFE_OK)
		goto error;

	/* Type of memory map delta table */
	rc = cafe_type_parse(inst->session->ns,
			     CAFE_MM_NO_DELTA_REF, &ai->mm_no_delta_type);
	if (rc != CAFE_OK)
		goto error;

	rc = cafe_type_parse(inst->session->ns, CAFE_NORMALIZED_SYMBOL_REF,
			     &ai->symbol_index_type);
	if (rc != CAFE_OK)
		goto error;

	rc = cafe_type_parse(inst->session->ns, CAFE_NORMALIZED_SYMBOL_EXT_REF,
			     &ai->symbol_index_ext_type);
	if (rc != CAFE_OK)
		goto error;

	rc = cafe_type_parse(inst->session->ns, CAFE_NORMALIZED_LINEINFO_REF,
			     &ai->lineinfo_index_type);
	if (rc != CAFE_OK)
		goto error;

	ai->temp_type_max = 1;
	ai->temp_type_list = CAFE_ALLOC(
		inst->allocator, ai->temp_type_max * sizeof *ai->temp_type_list);
	if (ai->temp_type_list == NULL) {
		rc = CAFE_NOMEM;
		goto error;
	}

	ai->temp_arg_max = 1;
	ai->temp_arg_list = CAFE_ALLOC(
		inst->allocator, ai->temp_arg_max * sizeof *ai->temp_arg_list);
	if (ai->temp_arg_list == NULL) {
		rc = CAFE_NOMEM;
		goto error;
	}

	cafe_list_init(&ai->table_list_head);
	cafe_list_init(&ai->action_list_head);
	for (i = 0; i < EVENT_ACTION_HASH_SIZE; i++)
		cafe_list_init(&ai->eahash_list_head[i]);
	cafe_ptr_array_init(&ai->mirror_list, inst->allocator);

	rc = process_settings(inst);
	if (rc != CAFE_OK)
		goto error2;

	if (ai->event[PENDING_DATA_EVENT] == NULL) {
		rc = update_output_event(inst);
		if (rc != CAFE_OK)
			goto error2;
	}
	return CAFE_OK;

error:
	CAFE_FREE(inst->allocator, ai);
	return rc;

error2:
	aggregator_ep_delete(inst);
	return rc;
}

static void action_inst_delete(
	struct cafe_ep_inst *inst,
	struct action_spec *action)
{
	if (action->trigger_expr != NULL)
		cafe_expr_free(inst->allocator, action->trigger_expr);
	if (action->filter_expr != NULL)
		cafe_expr_free(inst->allocator, action->filter_expr);
	if (action->cell_expr != NULL)
		cafe_expr_free(inst->allocator, action->cell_expr);
	if (action->args_expr != NULL)
		cafe_expr_free(inst->allocator, action->args_expr);
	if (action->action_list.next)
		cafe_list_remove(&action->action_list);
	CAFE_FREE(inst->allocator, action);
}

static int action_inst_create(
	struct cafe_ep_inst *inst,
	const char *trigger,
	const char *filter,
	const char *cell,
	const char *operation,
	const char *args,
	struct action_spec **action)

{
	size_t trigger_sz;
	size_t filter_sz;
	size_t cell_sz;
	size_t operation_sz;
	size_t args_sz;
	char *p;
	struct cafe_value_context *context;
	struct cafe_vm_expr *e;
	struct cafe_vm_expr *e2;
	struct cafe_vm_expr *e3;
	struct operation_entry *op_ent;
	int rc;

	if (trigger == NULL || operation == NULL) {
		cafe_error_set(NULL, CAFE_ERROR, "action requires a trigger and an operation");
		return CAFE_ERROR;
	}
	trigger_sz = strlen(trigger) + 1;
	filter_sz = strlen(filter) + 1;
	cell_sz = strlen(cell) + 1;
	operation_sz = strlen(operation) + 1;
	args_sz = strlen(args) + 1;
	*action = CAFE_ALLOC(inst->allocator, sizeof **action + 	
			     trigger_sz + filter_sz + cell_sz +
			     operation_sz + args_sz);
	if (*action == NULL)
		return CAFE_NOMEM;
	memset(*action, 0, sizeof(**action));

	p = (char *)(*action + 1);
	(*action)->trigger_str = p;
	memcpy(p, trigger, trigger_sz);
	p += trigger_sz;

	(*action)->filter_str = p;
	memcpy(p, filter, filter_sz);
	p += filter_sz;

	(*action)->cell_str = p;
	memcpy(p, cell, cell_sz);
	p += cell_sz;

	(*action)->operation_str = p;
	memcpy(p, operation, operation_sz);
	p += operation_sz;

	(*action)->args_str = p;
	memcpy(p, args, args_sz);
	p += args_sz;

	context = cafe_value_context_create(inst->allocator);
	assert(context != NULL);
	context->ns = inst->session->ns;

	rc = cafe_expr_list_parse(context, &trigger, &e);
	if (rc != CAFE_OK) {
		cafe_error_set(NULL, rc, "error parsing trigger expression");
		goto error;
	}
	if (*trigger != '\0') {
		cafe_expr_free(inst->allocator, e);
		rc = CAFE_SYNTAX;
		goto error;
	}
	assert(e->kind == CAFE_EXPR_LIST);

	context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
	for (e2 = e->u.e; e2 != NULL; e2 = e2->u.elem.next) {
		assert(e2->kind == CAFE_EXPR_ELEMENT);
		e2->u.elem.e = e3 = cafe_expr_resolve(context, e2->u.elem.e);
		assert(e3 != NULL);
		if (e3->type == NULL) {
			if (e3->kind == CAFE_EXPR_IDENT)
				cafe_error_set(NULL, rc, "error resolving trigger expression entry: %s", e3->u.str);
			else if (e3->kind == CAFE_EXPR_FIELD)
				cafe_error_set(NULL, rc, "error resolving trigger expression entry \"%s\" into a type", e3->u.bin.r->u.str.p);
			else
				cafe_error_set(NULL, rc, "error resolving trigger expression entry %d", e3->kind);
			cafe_expr_free(inst->allocator, e);
			rc = CAFE_ERROR;
			goto error;
		}
	}
	(*action)->trigger_expr = e;

	(*action)->trigger_name = trigger_table_name((*action)->trigger_expr);
	if ((*action)->trigger_name == NULL) {
		rc = trigger_type(inst, (*action)->trigger_expr,
				  &(*action)->trigger_type);
		if (rc != CAFE_OK) {
			cafe_error_set(NULL, rc, "error resolving trigger type");
			goto error;
		}
	}

	for (op_ent = operations; op_ent->name != NULL; op_ent++) {
		if (strcmp(op_ent->name, operation) == 0)
			break;
	}
	if (op_ent->name == NULL) {
		rc = CAFE_SYNTAX;
		cafe_error_set(NULL, rc, "Unknown action operation \"%s\"",
			       operation);
		goto error;
	}
	(*action)->op_ent = op_ent;

	if (*filter != '\0') {
		rc = cafe_expr_parse(context, &filter, &(*action)->filter_expr);
		if (rc != CAFE_OK) {
			cafe_error_set(NULL, rc, "error parsing filter expression");
			goto error;
		}
	}

	if (*cell != '\0') {
		rc = cafe_expr_parse(context, &cell, &(*action)->cell_expr);
		if (rc != CAFE_OK) {
			cafe_error_set(NULL, rc, "error parsing output cell expression");
			goto error;
		}
	}

	if (*args != '\0') {
		rc = cafe_expr_list_parse(context, &args,
					  &(*action)->args_expr);
		if (rc != CAFE_OK) {
			cafe_error_set(NULL, rc, "error parsing argument list");
			goto error;
		}
		if (*args != '\0') {
			rc = CAFE_SYNTAX;
			goto error;
		}
		assert((*action)->args_expr->kind == CAFE_EXPR_LIST);
	}

	cafe_value_context_delete(context);
	return CAFE_OK;

error:
	action_inst_delete(inst, *action);
	cafe_value_context_delete(context);
	return rc;
}

static void table_inst_delete(
	struct cafe_ep_inst *inst,
	struct table_inst *ti,
	int deferred)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;

	if (ti->is_symresolver)
		ai->symresolver_table = NULL;

	if (deferred) {
		ti->is_deleted = true;
		return;
	}
	if (ti->table_list.next)
		cafe_list_remove(&ti->table_list);
	cafe_value_delete(ti->value);
	CAFE_FREE(inst->allocator, ti);
}

static void aggregator_ep_delete(
	struct cafe_ep_inst *inst)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_list_head *list_item;
	struct cafe_list_head *tmp;

	CAFE_FREE(inst->allocator, ai->seen_value.list);

	if (ai->temp_type_list)
		CAFE_FREE(inst->allocator, ai->temp_type_list);

	if (ai->temp_arg_list)
		CAFE_FREE(inst->allocator, ai->temp_arg_list);

	cafe_event_action_list_swap(&ai->sink_actions[SNAPSHOT_EVENT],
				    CAFE_EAL_HANDLE_NULL);
	cafe_event_action_list_swap(&ai->sink_actions[PENDING_DATA_EVENT],
				    CAFE_EAL_HANDLE_NULL);

	cafe_ptr_array_release(&ai->mirror_list);
	cafe_list_foreach_safe(list_item, &ai->table_list_head, tmp) {
		struct table_inst *ti = TABLE_LIST_NODE_ENTRY(list_item);
		table_inst_delete(inst, ti, 0);
	}

	cafe_list_foreach_safe(list_item, &ai->action_list_head, tmp) {
		struct action_spec *action = 
			ACTION_LIST_NODE_ENTRY(list_item);
		action_inst_delete(inst, action);
	}

	eahash_free(inst, ai);

	if (ai->symresolver != NULL)
		symresolve_funcs->actions_delete(ai->symresolver);

        if (ai->ortho_dmap)
            CAFE_FREE(inst->allocator, ai->ortho_dmap);

	CAFE_FREE(inst->allocator, ai);
}

static int aggregator_ep_start(
	struct cafe_ep_inst *inst)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;

 	cafe_ep_inst_state_started(inst);
	action_event_handler(ai->start_event, NULL);

	return CAFE_OK;
}

static int aggregator_ep_stop(
	struct cafe_ep_inst *inst)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;

	action_event_handler(ai->stop_event, NULL);
 	cafe_ep_inst_state_stopped(inst);

	return CAFE_OK;
}

static int get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions)
{
	int rc;
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_event_action_info *event_info;
	struct cafe_list_head *list_item;

	cafe_list_foreach(list_item, &ai->action_list_head) {
		struct action_spec *action = 
			ACTION_LIST_NODE_ENTRY(list_item);
		if (action->op_ent->variants == aggregate_variants ||
		    action->op_ent->variants == stop_dispatch_variants) {
			if (matching_table_name(event, action->trigger_name, 0) < 0)
				continue;
		} else if (action->op_ent->variants == mm_delta_variants) {
			if (matching_table_type(event, action->trigger_type, 0) < 0)
				continue;
		} else if (!matching_event(inst, event, action->trigger_type, NULL))
			continue;

		rc = cafe_ep_inst_event_action_info_get2(
			inst, event, event, NULL, &event_info);
		if (rc != CAFE_OK)
			return rc;

		assert(event_info != NULL);
		rc = cafe_add_event_action(
			inst, actions, action_event_handler, event_info);
		if (rc != CAFE_OK)
			return rc;
		/* By break-ing we ensure that triggers (by type) get
		 * connected to an action handler only once. */
		break;
	}
	return CAFE_OK;
}

static int update_output_event_actions(
	struct cafe_ep_inst *inst)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_event_action_list *sink_actions;
	cafe_eal_handle_t sink_actions_handle;
	unsigned int idx;
	int rc;

	if (ai->change_count == inst->session->output_event_action_change_count &&
	    !ai->refresh_output_event_actions)
		return CAFE_OK;
	ai->change_count = inst->session->output_event_action_change_count;
	ai->refresh_output_event_actions = false;

	for (idx = 0; idx < NUM_AGGREGATOR_EVENTS; ++idx) {
		rc = cafe_event_action_list_alloc(inst, ai->event[idx],
						  &sink_actions,
						  &sink_actions_handle);
		if (rc != CAFE_OK)
			goto error;

		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			goto error;
		}
		cafe_event_action_list_swap(&ai->sink_actions[idx], sink_actions_handle);
	}
	return CAFE_OK;

error:
	for (idx = 0; idx < NUM_AGGREGATOR_EVENTS; ++idx)
		cafe_event_action_list_swap(&ai->sink_actions[idx],
					    CAFE_EAL_HANDLE_NULL);
	return rc;
}

static int wrap_table_type(
	struct cafe_ep_inst *inst,
	struct cafe_value_type *type,
	struct cafe_value_type **wrapped_type)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	char *type_ref;
	char *buf;
	size_t bufsz;
	int rc;

	type_ref = cafe_type_ref_str(type);
	if (type_ref == NULL)
		return CAFE_NOMEM;
	bufsz = strlen(type_ref) + strlen(type->ident->name) + 100;
	buf = CAFE_ALLOC(inst->allocator, bufsz);
	if (buf == NULL)
		return CAFE_NOMEM;
	(void)snprintf(buf, bufsz, "S[aggregator_table__%s__%lu,v[table,D[%s]],n[AGGREGATOR_TABLE]]",
		 type->ident->name, ai->wrap_count++, type_ref);
	rc = cafe_type_parse(inst->session->ns, buf, wrapped_type);
	if (rc != CAFE_OK)
		cafe_error_set(NULL, rc, "Failed to parse wrapped table type \"%s\"", buf);
	CAFE_FREE(inst->allocator, buf);
	CAFE_FREE(inst->allocator, type_ref);
	return rc;
}

static int update_output_event(
	struct cafe_ep_inst *inst)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_event_descr *event;
	struct cafe_event_descr *event2;
	struct cafe_list_head *list_item;
	struct cafe_value_type **types = ai->temp_type_list;
	struct cafe_value_type **types2 = types + ai->temp_type_max / 2;
	unsigned int num_types;
	unsigned int num_types2;
	int rc;

	num_types = 0;
	num_types2 = 0;
	types[num_types++] = inst->output_events[0];
	types2[num_types2++] = inst->output_events[0];
	cafe_list_foreach(list_item, &ai->table_list_head) {
		struct table_inst *ti = TABLE_LIST_NODE_ENTRY(list_item);
		if (ti->is_private)
			continue;
		types[num_types++] = ti->value->type;
		if (ti->is_deleted)
			continue;
		types2[num_types2++] = ti->value->type;
	}
	rc = cafe_cfg_get_event(inst->session, num_types, types, &event);
	if (rc != CAFE_OK)
		return rc;
	rc = cafe_cfg_get_event(inst->session, num_types2, types2, &event2);
	if (rc != CAFE_OK)
		return rc;
	if (ai->event[PENDING_DATA_EVENT] != event
	    || ai->event[SNAPSHOT_EVENT] != event2)
	{
		ai->event[PENDING_DATA_EVENT] = event;
		ai->event[SNAPSHOT_EVENT] = event2;
		if (event != event2)
			ai->pending_delete = true;
		ai->refresh_output_event_actions = true;
	}
	return CAFE_OK;
}

static int update_type_max(
	struct cafe_ep_inst *inst,
	unsigned int needed)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;

	if (needed > ai->temp_type_max) {
		unsigned int new_type_max = ai->temp_type_max*2;
		struct cafe_value_type **new_type_list;

		while (needed > new_type_max)
			new_type_max *= 2;
		new_type_list = CAFE_REALLOC(
			inst->allocator, ai->temp_type_list,
			new_type_max * sizeof *new_type_list);
		if (new_type_list == NULL) {
			return CAFE_NOMEM;
		}
		ai->temp_type_max = new_type_max;
		ai->temp_type_list = new_type_list;
	}
	return CAFE_OK;
}

static int update_arg_max(
	struct cafe_ep_inst *inst,
	unsigned int needed)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;

	if (needed > ai->temp_arg_max) {
		unsigned int new_arg_max = ai->temp_arg_max*2;
		cafe_any_value *new_arg_list;

		while (needed > new_arg_max)
			new_arg_max *= 2;
		new_arg_list = CAFE_REALLOC(
			inst->allocator, ai->temp_arg_list,
			new_arg_max * sizeof *new_arg_list);
		if (new_arg_list == NULL) {
			return CAFE_NOMEM;
		}
		ai->temp_arg_max = new_arg_max;
		ai->temp_arg_list = new_arg_list;
	}
	return CAFE_OK;
}

static void symresolve_errhandler(const char *msg, void *arg)
{
	struct cafe_ep_inst *inst = arg;
	if (inst->session->status_notifier)
		inst->session->status_notifier(inst, CAFE_LOG_WARNING, msg);
	else {
		(void)fputs_unlocked("WARNING: ", stdout);
		(void)fputs_unlocked(msg, stdout);
		(void)fputc_unlocked('\n', stdout);
	}
}

int cafe_aggregator_ep_table_create(
	struct cafe_ep_inst *inst,
	struct cafe_value_type *type)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct table_inst *ti;
	struct cafe_value_type *base_type;
	struct cafe_vm_attr *attr;
	struct cafe_list_head *list_item;
	unsigned int num_types;
	int has_keys;
	int rc;

	if (inst->spec != &spec)
		return CAFE_ERROR;

	base_type = cafe_base_type(type);
	if ((base_type == NULL) || (base_type->kind != CAFE_TYPE_STRUCT))
		return CAFE_ERROR;

	num_types = 1;
	cafe_list_foreach(list_item, &ai->table_list_head) {
		struct table_inst *ti = TABLE_LIST_NODE_ENTRY(list_item);
		if (ti->type == type && !ti->is_deleted)
			/* Already have table of this type */
			return CAFE_OK;
		if (!ti->is_private)
			num_types++;
	}
	rc = update_type_max(inst, num_types*2 + 1);
	if (rc != CAFE_OK)
		return rc;

	has_keys = 0;
	cafe_type_attr_foreach(attr, base_type) {
		assert(attr->kind == CAFE_ATTR_VAR ||
		       attr->kind == CAFE_ATTR_KEY ||
		       attr->kind == CAFE_ATTR_CONST ||
		       attr->kind == CAFE_ATTR_PARAM);
		if (attr->kind == CAFE_ATTR_KEY) {
			has_keys = 1;
			break;
		}
	}

	/* Create new table instance */
	if ((ti = CAFE_ALLOC(inst->allocator, sizeof *ti)) == NULL)
		return CAFE_NOMEM;
	memset(ti, 0, sizeof *ti);
	ti->is_private = cafe_find_nature_str(type->ident, cafe_PRIVATE_VT_DEF) != NULL;
	ti->is_persistent = cafe_find_nature_str(type->ident, cafe_PERSISTENT_VT_DEF) != NULL;
	ti->type = type;
	if (has_keys) {
		rc = wrap_table_type(inst, type, &type);
		if (rc != CAFE_OK) {
			CAFE_FREE(inst->allocator, ti);
			return rc;
		}
		ti->is_wrapped = true;
	}
	rc = cafe_value_create(type, &ti->value);
	if (rc != CAFE_OK) {
		CAFE_FREE(inst->allocator, ti);
		return rc;
	}

	/* Create symbol resolve instance */
	if (symresolve_funcs != NULL)
	{
		if (base_type == ai->mm_delta_type ||
		    base_type == ai->mm_no_delta_type)
		{
			if (ai->symresolver == NULL) {
				rc = symresolve_funcs->actions_create(
						base_type, inst->allocator, &ai->symresolver);
				if (rc != CAFE_OK) {
					cafe_value_delete(ti->value);
					CAFE_FREE(inst->allocator, ti);
					return rc;
				}
				symresolve_funcs->set_errhandler(ai->symresolver,
				                                 symresolve_errhandler,
				                                 inst);
			}

			if (base_type == ai->mm_delta_type) {
				assert(!ti->is_private);
				ai->symresolver_table = ti;
			}
			ti->is_symresolver = true;
		} else if (base_type == ai->symbol_index_type) {
			ai->sym_index_table = ti;
			symresolve_funcs->set_symindex(ai->symresolver,
			                               (struct cafe_dict_struct *)ti->value->u.data);
		} else if (base_type == ai->symbol_index_ext_type) {
			ai->sym_index_table = ti;
			symresolve_funcs->set_symindex(ai->symresolver,
			                               (struct cafe_dict_struct *)ti->value->u.data);
		} else if (base_type == ai->lineinfo_index_type) {
			ai->lineinfo_index_table = ti;
			symresolve_funcs->set_lineinfoindex(ai->symresolver,
			                                    (struct cafe_dict_struct *)ti->value->u.data);
		}
	}

	cafe_list_add_last(&ti->table_list, &ai->table_list_head);
	return update_output_event(inst);
}

int cafe_aggregator_ep_table_delete(
	struct cafe_ep_inst *inst,
	struct cafe_value_type *type)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_list_head *list_item;
	struct table_inst *ti;

	if (inst->spec != &spec)
		return CAFE_ERROR;

	if (!is_source_stopped(inst))
		return CAFE_INVALID_STATE;

	cafe_list_foreach(list_item, &ai->table_list_head) {
		ti = TABLE_LIST_NODE_ENTRY(list_item);
		if (ti->type == type && !ti->is_deleted) {
			/* Defer deletion of non-private tables until
			 * next snapshot. */
			table_inst_delete(inst, ti, !ti->is_private);
			inst->session->output_event_action_change_count++;
			break;
		}
	}
	return update_output_event(inst);
}

static struct cafe_vm_expr *resolve_expr_type(
	struct cafe_value_context *context,
	struct cafe_vm_expr *e)
{
	struct cafe_ep_inst *inst = context->inst;
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct cafe_event_descr *trigger_type = context->trigger_type;
	struct cafe_list_head *list_item;
	struct cafe_vm_expr *r;
	unsigned int idx;
	int rc;

	assert(e->kind == CAFE_EXPR_TYPE);
	assert(e->type);
	cafe_list_foreach(list_item, &ai->table_list_head) {
		struct table_inst *ti = TABLE_LIST_NODE_ENTRY(list_item);
		if (ti->is_deleted)
			continue;
		if (!cafe_type_based_on(ti->type, e->type))
			continue;
		if (ti->is_wrapped) {
			struct cafe_value_type *type = ti->value->type;
			assert(type->kind == CAFE_TYPE_STRUCT);
			assert(type->elements == 1);
			assert(type->attributes->kind == CAFE_ATTR_VAR);
			assert(type->attributes->offset == 0);
			rc = cafe_expr_var(context, cafe_attr_type(type->attributes),
					   ti->value->u.data, &r);
		} else {
			assert(ti->value->type == ti->type);
			rc = cafe_expr_var(context, ti->value->type, ti->value->u.data, &r);
		}
		if (rc != CAFE_OK)
			return e;
		cafe_expr_free(context->allocator, e);
		return r;
	}
	if (trigger_type && context->expr_resolve_mode == CAFE_EXPR_RESOLVE_LOAD) {
		struct cafe_event_descr *event = context->event;
		assert(event);
		assert(context->odmap);
		for (idx = 0; idx < trigger_type->num_types; idx++) {
			unsigned int odidx;
			if (!cafe_type_based_on(trigger_type->types[idx], e->type))
				continue;
			odidx = context->odmap[idx];
			assert(odidx < event->num_types);
			rc = cafe_expr_arg(context, event->types[odidx], odidx, &r);
			if (rc != CAFE_OK)
				return e;
			cafe_expr_free(context->allocator, e);
			return r;
		}
	}
	return e;
}

static struct cafe_vm_expr *resolve_expr_arg(
	struct cafe_value_context *context,
	struct cafe_vm_expr *e)
{
	struct cafe_event_descr *event = context->event;

	assert(e->kind == CAFE_EXPR_ARG);
	if (event != NULL && e->u.ull < event->num_types) {
		if (context->odmap) {
			e->u.ull = context->odmap[e->u.ull];
			assert(e->u.ull  < event->num_types);
		}
		e->type = event->types[e->u.ull];
	} else if (context->curobj_type && e->u.ull < 1) {
		e->type = context->curobj_type;
	}
	return e;
}

static bool valid_mm_sym_resolve_type(
	const struct cafe_value_context *context,
	const struct cafe_value_type *type)
{
	unsigned int i;

	type = cafe_base_type(type);
	if (type == NULL || type->kind != CAFE_TYPE_STRUCT)
		return false;
	for (i = 0; i < type->elements; i++) {
		struct cafe_vm_attr *attr = type->attributes + i;
		struct cafe_value_type *t;
		if (attr->kind != CAFE_ATTR_VAR)
			return false;
		t = cafe_base_type(cafe_attr_type(attr));
		if (strcmp(attr->ident->name, "object_name") == 0 ||
		    strcmp(attr->ident->name, "symbol_name") == 0 ||
		    strcmp(attr->ident->name, "source_name") == 0) {
			if ((t != NULL) && (t->kind != CAFE_TYPE_STRING))
				return false;
		} else if (strcmp(attr->ident->name, "object_address") == 0 ||
			   strcmp(attr->ident->name, "symbol_offset") == 0 ||
			   strcmp(attr->ident->name, "symbol_index") == 0 ||
			   strcmp(attr->ident->name, "lineinfo_index") == 0 ||
			   strcmp(attr->ident->name, "stack_index") == 0 ||
			   strcmp(attr->ident->name, "original_address") == 0 ||
			   strcmp(attr->ident->name, "line_number") == 0 ||
			   strcmp(attr->ident->name, "column_number") == 0)
		{
			if ((t != NULL) && (t->kind != CAFE_TYPE_ULONGLONG))
				return false;
		} else {
			return false;
		}
	}
	return true;
}

static struct action_spec *find_action(
	struct aggregator_inst *ai,
	const char *trigger,
	const char *filter,
	const char *cell,
	const char *operation,
	const char *args)
{
	struct cafe_list_head *list_item;

	cafe_list_foreach(list_item, &ai->action_list_head) {
		struct action_spec *action = ACTION_LIST_NODE_ENTRY(list_item);
		if (strcmp(action->trigger_str, trigger) == 0 &&
		    strcmp(action->filter_str, filter) == 0 &&
		    strcmp(action->cell_str, cell) == 0 &&
		    strcmp(action->operation_str, operation) == 0 &&
		    strcmp(action->args_str, args) == 0) {
			return action;
		}
	}
	return NULL;
}

static int safesamestr(const char *s1, const char *s2)
{
	return s1 == s2 || (s1 != NULL && s2 != NULL && strcmp(s1, s2) == 0);
}

static int safesamestrn(const char *s1, const char *s2, size_t n)
{
	return s1 == s2 || (s1 != NULL && s2 != NULL && strncmp(s1, s2, n) == 0);
}

static struct operation_variant *select_first_aggr_op(
	struct cafe_ep_inst *inst,
	struct operation_variant *op_var,
	struct cafe_value_type *type)
{
	struct cafe_value_type *t;
	int rc;

	if (op_var == NULL)
		return NULL;
	while (op_var->tag != CAFE_AGGREGATOR_EP_OP_NONE) {
		if (op_var->result_vtstr == NULL) {
			if (type != NULL) {
				return NULL;
			}
			return op_var;
		}
		if (strcmp(op_var->result_vtstr, "*") == 0)
			return op_var;
		rc = cafe_type_parse(inst->session->ns,
				     op_var->result_vtstr, &t);
		if (rc != CAFE_OK)
			return NULL;
		if (cafe_type_based_on(type, t))
			return op_var;
		op_var++;
	}
	return NULL;
}

static struct operation_variant *select_next_aggr_op(
	struct operation_variant *op_var,
	size_t args_len)
{
	if (op_var != NULL &&
	    safesamestr(op_var->result_vtstr, op_var[1].result_vtstr) &&
	    safesamestrn(op_var->args_vtstr, op_var[1].args_vtstr, args_len)) {
		return op_var + 1;
	}
	return NULL;
}

static int flatten_action_arg(
	struct event_action_inst *eainst,
	struct cafe_value_context *context,
	struct cafe_vm_expr *arg,
	struct cafe_value_type *type,
	unsigned int argno)
{
	struct cafe_value_type *type2;
	struct cafe_vm_expr *e;
	struct cafe_vm_expr *e2;
	unsigned int i;
	int rc;

	assert(arg != NULL);
	assert(arg->kind == CAFE_EXPR_ELEMENT);

	e = arg->u.elem.e;
	if (e == NULL ||
	    e->kind != CAFE_EXPR_FLATTEN ||
	    eainst->flatten_type != NULL ||
	    (e2 = e->u.e) == NULL ||
	    (type2 = cafe_base_type(e2->type)) == NULL ||
	    type2->kind != CAFE_TYPE_DYNARRAY)
		return CAFE_ERROR;
	type2 = cafe_base_type(cafe_element_type(type2));
	eainst->flatten_type = type2;
	eainst->flatten_expr = e2;
	eainst->flatten_arg_start = argno;
	e->u.e = NULL;
	cafe_expr_free(context->allocator, e);
	arg->u.elem.e = NULL;

	/* Use whole element if it is not a struct or the expected
	 * argument type */
	assert(type2 != NULL);
	if (type2->kind != CAFE_TYPE_STRUCT || type == type2) {
		eainst->flatten_arg_end = argno + 1;
		return cafe_expr_arg(context, type2, 1, &arg->u.elem.e);
	}
	if (type2->elements == 0)
		return CAFE_ERROR;

	/* Use each struct field as argument */
	for (i = 0; i < type2->elements; i++) {
		struct cafe_vm_attr *attr =
			type2->attributes + i;
		if (attr->kind != CAFE_ATTR_VAR)
			return CAFE_ERROR;
		rc = cafe_expr_arg(context, type2, 1, &e);
		if (rc != CAFE_OK)
			return rc;
		rc = cafe_expr_field(context, e, attr, &e);
		if (rc != CAFE_OK)
			return rc;
		if (i == 0) {
			arg->u.elem.e = e;
		} else {
			e = cafe_expr_new_element(context->allocator, e);
			e->u.elem.next = arg->u.elem.next;
			arg->u.elem.next = e;
			arg = e;
		}
	}
	eainst->flatten_arg_end = argno + i;
	return CAFE_OK;
}

/**
 *  Construct an expression representing a field within a struct.  Both the
 * structure type and the desired field are specified through the "attr"
 * argument.
 */
static
int expr_event_struct_field(struct cafe_value_context *context,
			    struct cafe_vm_attr *attr,
			    struct cafe_vm_expr **result)
{
	struct cafe_vm_expr *table_e;
	int rc;

	rc = cafe_expr_arg(context, attr->type, 0, &table_e);
	if (rc != CAFE_OK) {
		cafe_error_set(NULL, rc, "Error building resolve lookup"
			       " key table expression");
		return rc;
	}
	rc = cafe_expr_field(context, table_e, attr, result);
	if (rc != CAFE_OK) {
		cafe_error_set(NULL, rc, "Error building resolve lookup"
			       " key expression");
		return rc;
	}

	return CAFE_OK;
}

static int action_resolve(
	struct cafe_ep_inst *inst,
	struct action_spec *action,
	struct cafe_event_descr *event,
	unsigned int *odmap,
	int table_index,
	struct event_actions_map *eamap,
	struct cafe_error_report *err)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	int rc;
	struct cafe_value_context *context;
	struct cafe_vm_expr *e;
	struct cafe_vm_expr *e2;
	struct cafe_vm_expr *e3;
	struct cafe_value_type *type = NULL;
	struct operation_entry *op_ent = action->op_ent;
	struct operation_variant *op_var;
	struct operation_variant *op_var2;
	struct event_action_inst *eainst;
	const char *s;

	assert(table_index < 0 || table_index < (int)event->num_types);
	eainst = eainst_alloc(inst->allocator, eamap);
	if (eainst == NULL)
		return CAFE_NOMEM;
	eainst->action = action;
	eainst->table_index = table_index;

	context = cafe_value_context_create(inst->allocator);
	assert(context != NULL);
	context->ns = inst->session->ns;
	context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
	context->inst = inst;
	if (table_index < 0) {
		assert(action->trigger_type != NULL);
		context->event = event;
		context->trigger_type = action->trigger_type;
		context->odmap = odmap;
	} else {
		type = unwrap_table_type(event->types[table_index]);
		type = cafe_base_type(type);
		assert(type != NULL);
		if (type->kind == CAFE_TYPE_DICT) {
			eainst->table_type = type;
			type = cafe_base_type(cafe_element_type(type));
		}
		assert(type != NULL);
		if (type->kind != CAFE_TYPE_STRUCT) {
			rc = CAFE_ERROR;
			goto error;
		}
		context->curobj_type = type;
	}
	context->resolve_expr_type = resolve_expr_type;
	context->resolve_expr_arg = resolve_expr_arg;
	context->valid_mm_sym_resolve_type = valid_mm_sym_resolve_type;

	if (action->filter_expr != NULL) {
		rc = cafe_builtin_type(inst->session->ns, CAFE_TYPE_BOOL, &type);
		if (rc != CAFE_OK) {
			cafe_error_set(err, rc, "failed to reference the BOOL builtin type");
			goto error;
		}
		e = cafe_expr_copy(inst->allocator, action->filter_expr);
		context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
		eainst->filter = e = cafe_expr_resolve_for_type(context, e, type);
		if (e == NULL || e->type == NULL) {
			cafe_error_set(err, CAFE_ERROR, "bad filter <%s>",
				       action->filter_str);
			rc = CAFE_ERROR;
			goto error;
		}
	}

	if ((action->cell_expr != NULL) &&
	    (op_ent->variants != resolve_variants))
	{
		if (op_ent->variants == clear_variants) {
			context->expr_resolve_mode = CAFE_EXPR_RESOLVE_CLEAR;
		} else if (op_ent->variants == delete_variants) {
			context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
		} else {
			context->expr_resolve_mode = CAFE_EXPR_RESOLVE_STORE;
		}
		e = cafe_expr_copy(inst->allocator, action->cell_expr);
		eainst->cell_obj = e = cafe_expr_resolve(context, e);
		if (e == NULL || e->type == NULL) {
			cafe_error_set(err, CAFE_ERROR, "bad cell <%s>",
				       action->cell_str);
			rc = CAFE_ERROR;
			goto error;
		}
		if (op_ent->variants == clear_variants) {
			if (e->kind != CAFE_EXPR_DICT_CLEAR) {
				rc = CAFE_ERROR;
				goto error;
			}
			assert(op_ent->nature == NULL);
		} else if (op_ent->variants == delete_variants) {
			struct cafe_value_type *cell_type =
				cafe_base_type(eainst->cell_obj->type);
			assert(cell_type != NULL);
			if (cell_type->kind != CAFE_TYPE_DICT) {
				rc = CAFE_ERROR;
				goto error;
			}
			cell_type = cafe_base_type(cafe_element_type(cell_type));
			eainst->cell_type = cell_type;

			/* This is a bit of a kludge, we change the
			 * meaning of $1 for arguments to refer to the
			 * current row.  For the filter and cell
			 * expression $n refers to the trigger
			 * event. */
			context->event = NULL;
			context->trigger_type = NULL;
			context->odmap = NULL;
			context->curobj_type = cell_type;
		} else if (op_ent->variants == aggregate_variants ||
			   op_ent->variants == mm_update_variants ||
			   op_ent->variants == mm_delta_variants ||
			   op_ent->variants == mm_clone_variants ||
			   op_ent->variants == mm_clear_variants ||
			   op_ent->variants == mm_delete_variants) {
			if (e->kind != CAFE_EXPR_VAR) {
				cafe_error_set(err, CAFE_ERROR, "output field doesn't reference a declared table <%s> (%u)",
					       action->cell_str, e->kind);
				rc = CAFE_ERROR;
				goto error;
			}
		} else {
			while (e->kind == CAFE_EXPR_INDEX_ARRAY ||
			       e->kind == CAFE_EXPR_INDEX_DYNARRAY)
				e = e->u.bin.l;
			if (e->kind != CAFE_EXPR_FIELD ||
			    e->u.bin.r->kind != CAFE_EXPR_ATTR) {
				rc = CAFE_ERROR;
				goto error;
			}
			eainst->cell_attr = e->u.bin.r->u.attr;
			if (op_ent->nature &&
			    !cafe_find_nature_str(eainst->cell_attr->ident, op_ent->nature)) {
				rc = CAFE_ERROR;
				goto error;
			}
		}
	}

	op_var = select_first_aggr_op(inst, op_ent->variants,
				      eainst->cell_obj ?
				      eainst->cell_obj->type : NULL);
	if (op_var == NULL) {
		cafe_error_set(err, CAFE_ERROR, "no supported operation available <%s>", op_ent->name);
		rc = CAFE_ERROR;
		goto error;
	}

	if (op_ent->variants == clear_variants) {
		assert(op_var->args_vtstr == NULL);
		if (action->args_expr != NULL) {
			rc = CAFE_ERROR;
			goto error;
		}
	} else if (op_ent->variants == aggregate_variants) {
		struct cafe_value_type *cell_type;
		struct cafe_vm_attr *attr;
		unsigned int i;
		cafe_aggregator_ep_op_tag *subops;
		cafe_aggregator_ep_op_tag *subops2;
		int has_cumulative;

		assert(op_var->args_vtstr == NULL);
		if (action->args_expr == NULL) {
			rc = CAFE_ERROR;
			goto error;
		}
		if (eainst->cell_obj == NULL) {
			rc = CAFE_ERROR;
			goto error;
		}
		cell_type = cafe_base_type(eainst->cell_obj->type);
		eainst->args = e = cafe_expr_copy(inst->allocator, action->args_expr);
		assert(e != NULL);
		assert(e->kind == CAFE_EXPR_LIST);
		assert(cell_type != NULL);
		e2 = e->u.e;
		if (cell_type->kind == CAFE_TYPE_DICT) {
			eainst->cell_dict = cell_type;
			for (i = 0; i < cell_type->elements; i++) {
				attr = cell_type->keymap[i];
				assert(attr->kind == CAFE_ATTR_KEY);
			retry_aggregate_arg:
				if (e2 == NULL) {
					rc = CAFE_ERROR;
					goto error;
				}
				assert(e2->kind == CAFE_EXPR_ELEMENT);

				type = cafe_base_type(cafe_attr_type(attr));
				context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
				e2->u.elem.e = e3 = cafe_expr_resolve_for_type(
					context, e2->u.elem.e, type);
				if (e3 == NULL || e3->type == NULL) {
					rc = flatten_action_arg(eainst, context, e2, type, i);
					if (rc != CAFE_OK)
						goto error;
					goto retry_aggregate_arg;
				}
				e2 = e2->u.elem.next;
			}
			cell_type = cafe_base_type(cafe_element_type(cell_type));
		}
		assert (cell_type != NULL);
		eainst->cell_type = cell_type;
		eainst->subops = subops = CAFE_ALLOC(
			inst->allocator, cell_type->elements * sizeof *subops * 2);
		if (subops == NULL) {
			rc = CAFE_NOMEM;
			goto error;
		}
		eainst->subops2 = subops2 = subops + cell_type->elements;
		has_cumulative = 0;
		for (i = 0; i < cell_type->elements; i++) {
			attr = cell_type->attributes + i;
			if (attr->kind != CAFE_ATTR_VAR) {
				assert(attr->kind == CAFE_ATTR_VAR ||
				       attr->kind == CAFE_ATTR_KEY ||
				       attr->kind == CAFE_ATTR_CONST ||
				       attr->kind == CAFE_ATTR_PARAM);
				continue;
			}
			if (e2 == NULL) {
				rc = CAFE_ERROR;
				goto error;
			}
			assert(e2->kind == CAFE_EXPR_ELEMENT);

			if (cafe_find_nature_str(attr->ident, cafe_SUM_VT_REF)) {
				op_var2 = sum_variants;
			} else if (cafe_find_nature_str(attr->ident, cafe_SET_VT_REF)) {
				op_var2 = set_variants;
			} else if (cafe_find_nature_str(attr->ident, cafe_MIN_VT_REF)) {
				op_var2 = min_variants;
			} else if (cafe_find_nature_str(attr->ident, cafe_MAX_VT_REF)) {
				op_var2 = max_variants;
			} else if (cafe_find_nature_str(attr->ident, cafe_AVG_VT_REF)) {
				op_var2 = avg_variants;
			} else {
				op_var2 = NULL;
			}

			type = cafe_base_type(cafe_attr_type(attr));
			op_var2 = select_first_aggr_op(inst, op_var2, type);
			if (op_var2 == NULL) {
				rc = CAFE_ERROR;
				goto error;
			}
			rc = cafe_type_parse(inst->session->ns,
					     op_var2->args_vtstr,
					     &type);
			if (rc != CAFE_OK)
				goto error;

			context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
			e2->u.elem.e = e3 = cafe_expr_resolve_for_type(
				context, e2->u.elem.e, type);
			if (e3 == NULL || e3->type == NULL) {
				if (e3 != NULL) {
					cafe_expr_free(context->allocator, e3);
					e2->u.elem.e = NULL;
				}
				rc = cafe_expr_unset(context, type, &e2->u.elem.e);
				if (rc != CAFE_OK)
					goto error;
				*subops++ = CAFE_AGGREGATOR_EP_OP_SKIP;
				*subops2++ = CAFE_AGGREGATOR_EP_OP_SKIP;
				e2 = e2->u.elem.next;
				continue;
			}
			while (cafe_base_type(e3->type) != cafe_base_type(type)) {
				op_var2 = select_next_aggr_op(op_var2, 0);
				if (op_var2 == NULL) {
					rc = CAFE_ERROR;
					goto error;
				}
				rc = cafe_type_parse(inst->session->ns,
						     op_var2->args_vtstr,
						     &type);
				if (rc != CAFE_OK)
					goto error;
			}

			*subops++ = op_var2->tag;
			if (cafe_find_nature_str(attr->ident, cafe_CUMULATIVE_VT_REF)) {
				*subops2++ = op_var2->tag;
				has_cumulative = 1;
			} else {
				*subops2++ = CAFE_AGGREGATOR_EP_OP_SKIP;
			}

			e2 = e2->u.elem.next;
		}
		if (!has_cumulative)
			eainst->subops2 = NULL;
		if (e2 != NULL) {
			cafe_error_set(err, CAFE_SYNTAX, "aggregate action, more parameters given than columns (%u found, looking for \"%s\")", i, op_var->args_vtstr);
			rc = CAFE_SYNTAX;
			goto error;
		}
		rc = update_arg_max(inst, cell_type->elements);
		if (rc != CAFE_OK)
			goto error;
	} else if (op_ent->variants == sendtoself_variants) {
		struct cafe_value_type **types;
		struct cafe_event_descr *evt;
		unsigned int num_args;

		eainst->args = e = cafe_expr_copy(inst->allocator, action->args_expr);
		assert(e != NULL);
		assert(e->kind == CAFE_EXPR_LIST);

		num_args = 0;
		for (e2 = e->u.e; e2 != NULL; e2 = e2->u.elem.next) {
			assert(e2->kind == CAFE_EXPR_ELEMENT);
			num_args++;
		}
		rc = update_type_max(inst, num_args);
		if (rc != CAFE_OK)
			goto error;
		rc = update_arg_max(inst, num_args);
		if (rc != CAFE_OK)
			goto error;

		num_args = 0;
		types = ai->temp_type_list;
		for (e2 = e->u.e; e2 != NULL; e2 = e2->u.elem.next) {
			assert(e2->kind == CAFE_EXPR_ELEMENT);
			context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
			e2->u.elem.e = e3 = cafe_expr_resolve(context, e2->u.elem.e);
			if (e3 == NULL || e3->type == NULL) {
				rc = CAFE_ERROR;
				goto error;
			}
			type = cafe_base_type(e3->type);
			assert(type != NULL);
			if (type->kind == CAFE_TYPE_STRUCT) {
				type = e3->type;
			} else if (type->kind == CAFE_TYPE_DICT) {
				rc = wrap_table_type(inst, cafe_element_type(type), &type);
				if (rc != CAFE_OK)
					goto error;
			} else {
				rc = CAFE_SYNTAX;
				goto error;
			}
			types[num_args++] = type;
		}
		CAFE_LOCK();
		rc = cafe_cfg_get_event(inst->session, num_args, types, &evt);
		if (rc != CAFE_OK) {
			CAFE_UNLOCK();
			goto error;
		}
		rc = cafe_ep_inst_event_action_info_get2(
			inst, evt, evt, NULL, &eainst->event_info);
		CAFE_UNLOCK();
		if (rc != CAFE_OK)
			return rc;

	} else if (op_ent->variants == resolve_variants) {
		struct cafe_vm_attr *mm_gen_attr, *raw_stack_attr, *addr_space_attr;
		struct cafe_vm_expr *list, *element_e;
		struct cafe_vm_expr **trav;
		struct cafe_vm_attr *attr;
		struct cafe_vm_expr *raw_stack_arg;
		struct cafe_vm_expr *address_space_arg;
		struct cafe_vm_expr *mm_generation_arg;
		struct cafe_value_type *element_t, *result_t = NULL;
		unsigned int n_args = 0;

		list = cafe_expr_new_list(inst->allocator);
		if (list == NULL) {
			rc = CAFE_NOMEM;
			goto error;
		}
		trav = &list->u.e;
		eainst->args = list;

		assert(type != NULL);
		element_t = (type->kind == CAFE_TYPE_DICT) ?
			cafe_element_type(eainst->table_type) : type;

		/* locate arguments necessary for symresolve call */
		mm_gen_attr = NULL;
		raw_stack_attr = NULL;
		addr_space_attr = NULL;
		cafe_type_attr_foreach(attr, element_t) {
			if (cafe_find_nature_str(attr->ident, cafe_memory_map_generation_VT_DEF) != NULL) {
				mm_gen_attr = attr;
				rc = expr_event_struct_field(context, attr,
							     &mm_generation_arg);
				if (rc != CAFE_OK)
					goto error;
			} else if (cafe_find_nature_str(attr->ident, cafe_callstack_VT_DEF) != NULL) {
				raw_stack_attr = attr;
				rc = expr_event_struct_field(context, attr,
							     &raw_stack_arg);
				if (rc != CAFE_OK)
					goto error;
			} else if (cafe_find_nature_str(attr->ident, cafe_address_space_id_VT_DEF) != NULL) {
				addr_space_attr = attr;
				rc = expr_event_struct_field(context, attr,
							     &address_space_arg);
				if (rc != CAFE_OK)
					goto error;
			}
		}

		if (mm_gen_attr == NULL) {
			char *msg = cafe_mprintf(inst->allocator, "%s lacks memory map generation info, not resolving symbols", element_t->type_str);
			assert(msg != NULL);
			cafe_dispatch_status(inst, CAFE_LOG_INFO, msg);
			CAFE_FREE(inst->allocator, msg);
			goto no_translation;
		}

		if (addr_space_attr == NULL) {
			char *msg = cafe_mprintf(inst->allocator, "%s lacks address space identifiers, not resolving symbols", element_t->type_str);
			assert(msg != NULL);
			cafe_dispatch_status(inst, CAFE_LOG_INFO, msg);
			CAFE_FREE(inst->allocator, msg);
			goto no_translation;
		}

		if (raw_stack_attr == NULL) {
			char *msg = cafe_mprintf(inst->allocator, "%s lacks callstacks, not resolving symbols", element_t->type_str);
			assert(msg != NULL);
			cafe_dispatch_status(inst, CAFE_LOG_INFO, msg);
			CAFE_FREE(inst->allocator, msg);
			goto no_translation;
		}

		/* parse/resolve the object-resolved output struct type */
		action->args_expr->u.e->u.elem.e = cafe_expr_resolve(context, action->args_expr->u.e->u.elem.e);
		if (action->args_expr->u.e->u.elem.e == NULL ||
		    action->args_expr->u.e->u.elem.e->type == NULL)
		{
			cafe_error_set(err, CAFE_ERROR,
				       "Error parsing resolve argument");
			rc = CAFE_ERROR;
			goto error;
		}

		/* copy result values into output table, maintaining order of
		   resolved stack with respect to the raw stack attribute of
		   the incoming type. */

    		/* add keys */
		cafe_type_attr_foreach(attr, element_t) {
			struct cafe_vm_expr *field_e;

			if ((attr == mm_gen_attr) ||
			    (attr->kind != CAFE_ATTR_KEY))
			    continue;

			if (attr == raw_stack_attr) {
				rc = cafe_expr_symresolve(context,
							  raw_stack_arg,
							  cafe_expr_copy(inst->allocator, address_space_arg),
							  mm_generation_arg,
							  action->args_expr->u.e->u.elem.e->type,
							  &field_e);
				if (rc != CAFE_OK)
					goto error;
			} else if (attr == addr_space_attr) {
				field_e = address_space_arg;
			} else {
				rc = expr_event_struct_field(context, attr,
							     &field_e);
				if (rc != CAFE_OK)
					goto error;
			}

			element_e = cafe_expr_new_element(inst->allocator,
							  field_e);

			*trav = element_e;
			trav = &(*trav)->u.elem.next;
			++n_args;
		}

		/* add values */
		cafe_type_attr_foreach(attr, element_t) {
			struct cafe_vm_expr *field_e;

			if ((attr == mm_gen_attr) ||
			    (attr->kind != CAFE_ATTR_VAR))
			continue;

			if (attr == raw_stack_attr) {
				rc = cafe_expr_symresolve(context,
							  raw_stack_arg,
							  cafe_expr_copy(inst->allocator, address_space_arg),
							  mm_generation_arg,
							  action->args_expr->u.e->u.elem.e->type,
							  &field_e);
				if (rc != CAFE_OK)
					goto error;
			} else if (attr == addr_space_attr) {
				field_e = address_space_arg;
			} else {
				rc = expr_event_struct_field(context, attr,
							     &field_e);
				if (rc != CAFE_OK)
					goto error;
			}

			element_e = cafe_expr_new_element(inst->allocator,
							  field_e);
			*trav = element_e;
			trav = &(*trav)->u.elem.next;
			++n_args;
		}

		/* calculate the output destination; cell_obj */
		rc = resolved_struct_get(element_t,
					 action->args_expr->u.e->u.elem.e,
					 mm_gen_attr, raw_stack_attr, action->cell_str,
					 inst->allocator, &result_t);
		if (rc != CAFE_OK) {
			cafe_error_set(err, rc,
				       "Error creating resolved structure "
				       "definition");
			goto error;
		}

	no_translation:
		if (list->u.e == NULL) {
			/* set up for mirroring */
			rc = cafe_expr_arg(context, type, table_index,
					   &element_e);
			if (rc != CAFE_OK)
				goto error;

			*trav = cafe_expr_new_element(inst->allocator,
						      element_e);
			eainst->table_index = -1;
			result_t = element_t;
			n_args = 1;
			++op_var;
		}
		if (result_t == NULL) {
			rc = CAFE_ERROR;
			goto error;
		}
		rc = table_matching_type(inst, context, result_t,
					 &eainst->cell_obj);
		if (rc != CAFE_OK)
			goto error;

		rc = update_arg_max(inst, n_args);
		if (rc != CAFE_OK)
			goto error;

	} else if (action->args_expr != NULL) {
		unsigned int num_args;
		eainst->args = e = cafe_expr_copy(inst->allocator, action->args_expr);
		assert(e != NULL);
		assert(e->kind == CAFE_EXPR_LIST);
	try_next_variant:
		num_args = 0;
		s = op_var->args_vtstr;
		e2 = e->u.e;
		while (e2 != NULL) {
			size_t args_len = s - op_var->args_vtstr;
			num_args++;
			assert(e2->kind == CAFE_EXPR_ELEMENT);
			if (*s == '*') {
				s++;
				type = cafe_base_type(eainst->cell_obj->type);
			} else {
				rc = cafe_type_parse2(inst->session->ns,
						      &s, ",", &type);
				if (rc != CAFE_OK) {
					cafe_error_set(err, rc, "failed to parse argument list <%s>", s);
					goto error;
				}
			}
			context->expr_resolve_mode = CAFE_EXPR_RESOLVE_LOAD;
			e2->u.elem.e = e3 = cafe_expr_resolve_for_type(
				context, e2->u.elem.e, type);
			if (e3 == NULL || e3->type == NULL) {
				rc = CAFE_ERROR;
				goto error;
			}
			while (cafe_base_type(e3->type) != cafe_base_type(type)) {
				op_var = select_next_aggr_op(op_var, args_len);
				if (op_var == NULL) {
					rc = CAFE_ERROR;
					goto error;
				}
				s = op_var->args_vtstr + args_len;
				rc = cafe_type_parse2(inst->session->ns,
						      &s, ",", &type);
				if (rc != CAFE_OK)
					goto error;
			}
			e2 = e2->u.elem.next;
			if (*s != ',')
				break;
			s++;
		}

		if (e2 != NULL || *s != '\0') {
			struct operation_variant *next_var = op_var + 1;

			if ((next_var->tag != CAFE_AGGREGATOR_EP_OP_NONE) &&
			    (strcmp(next_var->result_vtstr, op_var->result_vtstr) == 0)) {
				op_var = next_var;
				goto try_next_variant;
			}
			cafe_error_set(err, CAFE_SYNTAX, "more parameters given than columns (%u found, looking for \"%s\")", num_args, op_var->args_vtstr);
			rc = CAFE_SYNTAX;
			goto error;
		}
		rc = update_arg_max(inst, num_args);
		if (rc != CAFE_OK)
			goto error;
	} else if (op_var->args_vtstr != NULL) {
		rc = CAFE_ERROR;
		goto error;
	}
	eainst->operation = op_var->tag;

	cafe_value_context_delete(context);
	eamap->len++;
	return CAFE_OK;

error:
	cafe_value_context_delete(context);
	eainst_free(inst->allocator, eainst);
	return rc;
}

int cafe_aggregator_ep_action_add(
	struct cafe_ep_inst *inst,
	const char *trigger,
	const char *filter,
	const char *cell,
	const char *operation,
	const char *args)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	int rc;
	struct action_spec *action;
	
	assert(inst->spec == &spec);

	if (filter == NULL)
		filter = "";

	if (cell == NULL)
		cell = "";

	if (args == NULL)
		args = "";

	if (!is_source_stopped(inst))
		return CAFE_INVALID_STATE;

add_snapshot:
	action = find_action(ai, trigger, filter, cell, operation, args);
	if (action != NULL) {
		/* Duplicate action, move it to expected position */
		cafe_list_remove(&action->action_list);
		cafe_list_add_last(&action->action_list, &ai->action_list_head);
		return CAFE_OK;
	}

	rc = action_inst_create(inst, trigger, filter, cell, operation, args,
				&action);
	if (rc != CAFE_OK)
		return rc;

	cafe_list_add_last(&action->action_list, &ai->action_list_head);

	if (action->op_ent->variants == resolve_variants) {
		trigger = "cafe_aggregator.aggregator";
		cell = "";
		operation = "snapshot";
		args = "";
		++inst->session->output_event_action_change_count;
		goto add_snapshot;
	}

	inst->session->output_event_action_change_count++;
	return CAFE_OK;
}

int cafe_aggregator_ep_action_del(
	struct cafe_ep_inst *inst,
	const char *trigger,
	const char *filter,
	const char *cell,
	const char *operation,
	const char *args)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	struct action_spec *action;

	if (inst->spec != &spec)
		return CAFE_ERROR;

	if (!is_source_stopped(inst))
		return CAFE_INVALID_STATE;

	action = find_action(ai, trigger ? trigger : "",
			     filter ? filter : "", cell ? cell : "",
			     operation ? operation : "", args ? args : "");
	if (action) {
		action_inst_delete(inst, action);
		inst->session->output_event_action_change_count++;
	}

	return CAFE_OK;
}

static int aggregator_ep_settings_commit(
	struct cafe_ep_inst *inst,
	unsigned int num_settings,
	const struct cafe_cfg_setting *settings)
{
	unsigned int i;
	int rc;

	cafe_value_clear(inst->workarea);
	for (i = 0; i < num_settings; ++i) {
		rc = cafe_value_store(inst->workarea_type,
				      inst->workarea->u.data,
				      settings[i].name,
				      &settings[i].value,
				      inst->allocator);
		if (rc != CAFE_OK)
			return rc;
	}

	return process_settings(inst);
}

static void aggregator_ep_input_state_change(
	struct cafe_ep_inst *inst)
{
	struct aggregator_inst *ai = (struct aggregator_inst *)inst->client_data;
	int rc;

	CAFE_ASSERT_LOCKED();
	assert(inst->input_ctx != CAFE_EXEC_CTX_ANY);
	if (inst->input_ctx == NULL ||
	    inst->input_ctx == CAFE_EXEC_CTX_SESSION)
		return;
	switch (inst->input_ctx->state) {
	case CAFE_EP_INST_STATE_STARTED:
		eahash_free(inst, ai);
		rc = update_output_event_actions(inst);
		if (rc != CAFE_OK) {
			cafe_dispatch_error(inst, rc);
			break;
		}
		break;

	case CAFE_EP_INST_STATE_STOPPED:
		rc = update_output_event_actions(inst);
		if (rc != CAFE_OK) {
			cafe_dispatch_error(inst, rc);
			break;
		}
		break;

	default:
		/* coverity[array_null] */
		assert(!"unexpected state");
	}
}

#define WORKAREA_DEF "S[" CAFE_AGGREGATOR_EP_NAME ".workarea,"	\
	"v[del_tables,a[s]," cafe_setting_VT_NAT "]"		\
	"v[del_actions,a[a[s]]," cafe_setting_VT_NAT "]"	\
	"v[add_types,a[s]," cafe_setting_VT_NAT "]"		\
	"v[add_tables,a[s]," cafe_setting_VT_NAT "]"		\
	"v[add_actions,a[a[s]]," cafe_setting_VT_NAT "],"	\
	cafe_aggregator_BUCKETING_VT_NAT			\
	cafe_aggregator_SYMRESOLVE_VT_NAT			\
	"]"

#define WORKAREA_REF "t[" CAFE_AGGREGATOR_EP_NAME ".workarea]"

/*  Describes the natures necessary to trigger symbol resolution when a
    "resolve" action is specified. */
#define GENERIC_SYMRESOLVE_FIELDS					\
	cafe_callstack_VT_DEF						\
	cafe_address_space_id_VT_DEF					\
	cafe_memory_map_generation_VT_DEF				\
	"S[" CAFE_AGGREGATOR_EP_NAME ".required_generic_symresolve_natures,," \
	cafe_callstack_VT_NAT						\
	cafe_address_space_id_VT_NAT					\
	cafe_memory_map_generation_VT_NAT "]"

#define ALL_TYPES				\
	TIMER_DEF				\
	cafe_start_VT_DEF			\
	cafe_stop_VT_DEF			\
	cafe_setting_VT_DEF			\
	cafe_AGGREGATE_TYPE_VT_DEF		\
	cafe_SUM_VT_DEF				\
	cafe_SET_VT_DEF				\
	cafe_MIN_VT_DEF				\
	cafe_MAX_VT_DEF				\
	cafe_AVG_VT_DEF				\
	cafe_CUMULATIVE_VT_DEF			\
	cafe_setval_q_VT_DEF			\
	cafe_setval_Q_VT_DEF			\
	cafe_setval_d_VT_DEF			\
	cafe_setval_bool_VT_DEF			\
	cafe_setval_s_VT_DEF			\
	cafe_setval_b_VT_DEF			\
	cafe_avgval_q_VT_DEF			\
	cafe_avgval_Q_VT_DEF			\
	cafe_avgval_d_VT_DEF			\
	cafe_aggregator_aggregator_VT_DEF	\
	cafe_wildcard_VT_DEF			\
	cafe_PRIVATE_VT_DEF			\
	cafe_PERSISTENT_VT_DEF			\
	"N[AGGREGATOR_TABLE]"			\
	cafe_TABLE_NAME_VT_DEF			\
	cafe_ADDRESS_VT_DEF			\
	cafe_CAPABILITY_VT_DEF			\
	cafe_aggregator_BUCKETING_VT_DEF	\
	cafe_aggregator_SYMRESOLVE_VT_DEF	\
	CAFE_MM_DELTA_ENTRY_DEF			\
	CAFE_MM_DELTA_DEF			\
	CAFE_MM_NO_DELTA_DEF			\
	CAFE_NORMALIZED_SYMBOL_DEF		\
	CAFE_NORMALIZED_SYMBOL_EXT_DEF		\
	CAFE_NORMALIZED_LINEINFO_DEF		\
	"S[cafe.mm_sym_resolve_fields,"		\
	"v[object_name,s]"			\
	"v[object_address,Q]"			\
	"v[symbol_name,s]"			\
	"v[symbol_offset,Q]"			\
	"v[symbol_index,Q]"			\
	"v[lineinfo_index,Q]"			\
	"v[stack_index,Q]"			\
	"v[original_address,Q]"			\
	"v[source_name,s]"			\
	"v[line_number,Q]"			\
	"v[column_number,Q]]"			\
	GENERIC_SYMRESOLVE_FIELDS		\
	WORKAREA_DEF

static struct cafe_ep_spec spec = {
	CAFE_AGGREGATOR_EP_NAME,
	aggregator_ep_create,		/* create_hook */
	aggregator_ep_delete,		/* delete_hook */
	aggregator_ep_start,		/* start */
	aggregator_ep_stop,		/* stop */
	get_input_event_actions,
	update_output_event_actions,
	NULL,				/* inject event */
	ALL_TYPES,
	WORKAREA_REF,
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events,
	0,
	NULL,
	aggregator_ep_settings_commit,
	aggregator_ep_input_state_change
};

static bool cfg_registered = false;

void cafe_aggregator_ep_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}

int cafe_aggregator_ep_register(void)
{
	int rc = CAFE_OK;

	if (!cfg_registered) {
		rc = cafe_ep_spec_register(&spec);
		if (rc == CAFE_OK)
			cfg_registered = true;
	}

	return rc;
}

void aggregator_symresolve_register(const struct symresolve_functions *funcs)
{
	assert(symresolve_funcs == NULL);
	symresolve_funcs = funcs;
}
