/* realtime_ep - delay replay of events to realtime */

/* Copyright (c) 2010, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

/*
 * This event processor restores time spacing of events to realtime
 * when events include timestamp information.  A frequency percentage
 * may be applied to the time spacing to increase or decrease playback
 * rate.
 */

#include "../src/cafe_os.h"
#include <string.h>
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/value_print.h>
#include <cafe/ep/realtime_ep.h>
#include "../src/cafe_assert.h"
#include "../src/time_utils.h"


static void timestamp_freq_act_handler(struct cafe_event_action_info *, void **);
static void wildcard_with_ts_act_handler(struct cafe_event_action_info *, void **);
static void wildcard_act_handler(struct cafe_event_action_info *, void **);

struct realtime_inst {
	unsigned long long clock_frequency;
	struct cafe_timestamp last_timestamp;
	int has_last_timestamp;
	struct timespec last_realtime_event;
};

struct workarea {
	unsigned long frequency_percent;
};

static const char *wildcard1_TYPE[] = {
	cafe_wildcard_VT_REF
};

static const char *wildcard2_TYPE[] = {
	cafe_wildcard_VT_REF,
	cafe_timestamp_VT_REF
};

static const char *timestamp_freq_TYPE[] = {
	cafe_timestamp_config_VT_REF
};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT2(timestamp_freq, timestamp_freq_act_handler),
	CAFE_INPEV_INIT2(wildcard2, wildcard_with_ts_act_handler),
	CAFE_INPEV_INIT2(wildcard1, wildcard_act_handler),
};

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};

static void timestamp_freq_act_handler(struct cafe_event_action_info *info,
				       void **data)
{
	const struct workarea *w;
	struct realtime_inst *i;
	const struct cafe_timestamp_config *e;

	i = (struct realtime_inst *)info->inst->client_data;
	w = (const struct workarea *)info->inst->workarea->u.data;
	e = (struct cafe_timestamp_config *)data[0];

	i->clock_frequency = e->frequency * w->frequency_percent / 100;
}

static void wildcard_with_ts_act_handler(struct cafe_event_action_info *info,
					 void **data)
{
	struct realtime_inst *i;
	struct cafe_timestamp *timestamp_event;
	unsigned long long delta;
	struct timespec timenow;
	struct timespec delay;

	assert(info->event->num_types > 1);

	i = (struct realtime_inst *)info->inst->client_data;

	timestamp_event = data[info->odmap[1]];

	if (i->clock_frequency == 0 ||
	    timestamp_event->timestamp < i->last_timestamp.timestamp)
		return;

	if (!i->has_last_timestamp) {
		i->has_last_timestamp = 1;
		i->last_timestamp.timestamp = timestamp_event->timestamp;
		if (clock_gettime(CLOCK_REALTIME, &i->last_realtime_event) != 0) {
			i->has_last_timestamp = 0;
			i->clock_frequency = 0;
		}
		return;
	}

	delta = timestamp_event->timestamp - i->last_timestamp.timestamp;
	i->last_timestamp.timestamp = timestamp_event->timestamp;

	timespec_ticks_add((unsigned long)delta, (unsigned long)i->clock_frequency,
	                   &i->last_realtime_event);
	if (clock_gettime(CLOCK_REALTIME, &timenow) < 0) {
		i->has_last_timestamp = 0;
		i->clock_frequency = 0;
		return;
	}

	/* Delay propagation if we received the event too early. */
	if (timespec_cmp(&i->last_realtime_event, &timenow) > 0) {
		delay.tv_sec = i->last_realtime_event.tv_sec - timenow.tv_sec;
		delay.tv_nsec = i->last_realtime_event.tv_nsec - timenow.tv_nsec;
		if (i->last_realtime_event.tv_nsec < timenow.tv_nsec) {
			delay.tv_sec -= 1;
			delay.tv_nsec += 1000000000;
		}
		(void)nanosleep(&delay, NULL);
	}
}

static void wildcard_act_handler(struct cafe_event_action_info *info,
				 void **data)
{
}

static int realtime_ep_create(struct cafe_ep_inst *inst)
{
	struct realtime_inst *i;

	if ((i = CAFE_ALLOC(inst->allocator, sizeof *i)) == NULL)
		return CAFE_NOMEM;

	memset(i, 0, sizeof *i);

	inst->client_data = i;
	return CAFE_OK;
}

static void realtime_ep_delete(struct cafe_ep_inst *inst)
{
	CAFE_FREE(inst->allocator, inst->client_data);
}

static int get_input_event_actions(struct cafe_ep_inst *inst,
				   struct cafe_event_descr *event,
				   struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);
	if (rc != CAFE_OK)
		return rc;

	if (event_info == NULL)
		return CAFE_OK;

	rc = cafe_add_event_action(inst, actions,
				   (CAFE_ACTION_HANDLER)event_info->impl_data,
				   event_info);
	if (rc != CAFE_OK)
		return rc;

	/* Ask the sink if it is interested in this event type. */
	return cafe_ep_inst_get_sink_event_actions(inst, actions);
}

#define EP_NAME "cafe_realtime"

#define WORKAREA_DEF "S[" EP_NAME ".workarea,v[frequency_percent,L," cafe_setting_VT_NAT ",100]]"
#define WORKAREA_REF "t[" EP_NAME ".workarea]"

#define ALL_TYPES				\
	cafe_timestamp_config_VT_DEF		\
	cafe_timestamp_VT_DEF			\
	cafe_wildcard_VT_DEF			\
	cafe_setting_VT_DEF			\
	WORKAREA_DEF

static struct cafe_ep_spec spec = {
	EP_NAME,
	realtime_ep_create,		/* create_hook */
	realtime_ep_delete,		/* delete_hook */
	NULL,				/* start */
	NULL,				/* stop */
	get_input_event_actions,
	NULL,
	NULL,				/* inject event */
	ALL_TYPES,
	WORKAREA_REF,
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events
};

static bool cfg_registered = false;

void cafe_realtime_ep_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}

int cafe_realtime_ep_register(void)
{
	int rc = CAFE_OK;

	if (!cfg_registered) {
		rc = cafe_ep_spec_register(&spec);
		if (rc == CAFE_OK)
			cfg_registered = true;
	}

	return rc;
}
