/* sysinfo.c - cafe configuration for target information. */

/* Copyright (c) 2009-2010, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
26aug14,tle  Register system_info EP for wrlinux targets (US30550)
05jun14,sdt  Fix conversion to int from size_t warning.
26may14,sdt  Checked returned value to avoid dereferencing null pointer
15may14,sdt  Don't set release field that is meaningless in VxWorks 7 (WB4-3418)
             Improved sysinfo data freeing
02apr14,tle  Add getCapabilities command providing system information (US30548)

DESCRIPTION

This cafe configuration gathers BSP and target specific information as
a list of key-value pairs.

*/

#include <string.h>
#include <stdio.h>
#include "../src/cafe_assert.h"
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/value_type.h>
#include <cafe/dict.h>

#if defined(__VXWORKS__)
#include <vxWorks.h>
#include <version.h>
#include <logLib.h>
#include <sysLib.h>  /* for sysModel() */
#include <bootLib.h>  /* for BOOT_PARAMS */
#else
#include <sys/utsname.h>
#endif

#include <cafe/ep/sysinfo.h>

/* forward declarations */
static int sysinfo_cfg_start(struct cafe_ep_inst *inst);
static int sysinfo_cfg_update_output_event_actions(struct cafe_ep_inst *inst);
static int sysinfo_inst_create(struct cafe_ep_inst *inst);
static void sysinfo_inst_delete(struct cafe_ep_inst *inst);
static int sysinfo_cfg_get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions);


#if defined(__VXWORKS__)
static const char *RTPSUPPORT_KEY = "RTPSupport";
static const char *BSPNAME_KEY = "BSP Name";
static const char *IP_ADDR_KEY = "IP addr";
static const char *BOOTFILE_KEY = "Bootfile";
static const char *GATEWAY_KEY = "Gateway";
static const char *HOST_ADDR_KEY = "Host";
#endif

#define SYSINFO_SNAPSHOT_EVENT_DEF  "S[system_info.snapshot]"

const char *snapshot_event_TYPE[] = {SYSINFO_SNAPSHOT_EVENT_REF};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(snapshot_event)
};

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT("t[system_info.workarea]")
};


#define WORKAREA_VT							\
	"S[sysinfo_row,k[key,s]v[value,s]]"				\
	"S[system_info.workarea,v[sysname,s]v[release,s]v[version,s]v[properties,D[t[sysinfo_row]]]]"


static struct cafe_ep_spec spec = {
	"system_info",
	sysinfo_inst_create,
	sysinfo_inst_delete,
	sysinfo_cfg_start,
	NULL, /* stop, */
	sysinfo_cfg_get_input_event_actions,
	sysinfo_cfg_update_output_event_actions,
	NULL,
	cafe_wildcard_VT_DEF
	SYSINFO_SNAPSHOT_EVENT_DEF
	WORKAREA_VT,  /* types */
	"t[system_info.workarea]",
	CAFE_NELEMENTS(input_events),
	input_events,  /* input_events */
	CAFE_NELEMENTS(output_events),
	output_events
};


struct sysinfo_inst_data {
	unsigned long output_event_action_change_count;
	cafe_eal_handle_t sink_actions;
};


static
int sysinfo_inst_create(struct cafe_ep_inst *inst)
{
	struct sysinfo_inst_data *inst_data;

	if ((inst_data = CAFE_ALLOC(inst->allocator, sizeof *inst_data)) == NULL)
		return CAFE_NOMEM;

	inst_data->sink_actions = CAFE_EAL_HANDLE_NULL;
	inst_data->output_event_action_change_count =
		inst->session->output_event_action_change_count - 1;
	inst->client_data = inst_data;

	return CAFE_OK;
}


static
void sysinfo_inst_delete(struct cafe_ep_inst *inst)
{
	struct sysinfo_inst_data *inst_data = inst->client_data;

	cafe_event_action_list_release(&inst_data->sink_actions);
	CAFE_FREE(inst->allocator, inst_data);
}


static
int sysinfo_cfg_update_output_event_actions(struct cafe_ep_inst *inst)
{
	struct sysinfo_inst_data *inst_data = inst->client_data;
	int rc;
	unsigned int i;

	if (inst_data->output_event_action_change_count ==
	    inst->session->output_event_action_change_count)
		return CAFE_OK;
	inst_data->output_event_action_change_count =
		inst->session->output_event_action_change_count;

	for (i = 0; i < inst->enabled_output_events.ind; ++i) {
		struct cafe_event_descr *event =
			inst->enabled_output_events.list[i].event;
		struct cafe_event_action_list *sink_actions;
		cafe_eal_handle_t sink_actions_handle;

		rc = cafe_event_action_list_alloc(inst, event,
						  &sink_actions,
						  &sink_actions_handle);
		if (rc != CAFE_OK)
			return rc;
		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			return rc;
		}

		cafe_event_action_list_swap(&inst_data->sink_actions,
					    sink_actions_handle);
	}
	return CAFE_OK;
}


#if defined(__VXWORKS__)
/**
 * Create a key/value pair in dictionary. Returns true on success
 */
static
bool sysinfo_dict_add(struct cafe_dict_struct *dict, const char *key,
		      const char *value)
{
	cafe_any_value keys[1];
	struct cafe_value_data *val;
	struct cafe_sysconfig_property *row;

	keys[0].str = (char *) key;

	val = cafe_dict_get(dict, keys);
	if (val == NULL)
		return false;

	row = (struct cafe_sysconfig_property *)val->u.data;
	row->value = CAFE_STRDUP(val->type->ident->ns->allocator, value);

	return true;
}
#endif

#if defined(_WRS_KERNEL)

//td: enable cpu-specific properties
#if 0
#if (_VX_CPU_FAMILY == _VX_I80X86)
#include "ep/arch/i86/cpu_info.h"
bool sysinfo_pentium(struct cafe_sysconfig_info *sysinfo)
{
	unsigned int n;
	char buff[500];
	char name[100];
	int c = 0;

	cafe_cpu_info_get_vendor_id(buff, sizeof(buff));
	cafe_cpu_info_processor_signature();

	if (*buff != '\0')
		if (sysinfo_dict_add(&sysinfo->properties,
					"Vendor", buff) != 0)
			return false;

	cafe_cpu_info_get_cpu_family(buff, sizeof(buff));
	if (*buff != '\0')
		if (sysinfo_dict_add(&sysinfo->properties,
					"CPU Family", buff) != 0)
			return false;

	cafe_cpu_info_get_model(buff, sizeof(buff));
	if (*buff != '\0')
		if (sysinfo_dict_add(&sysinfo->properties,
					"CPU Model", buff) != 0)
			return false;

	cafe_cpu_info_get_stepping(buff, sizeof(buff));
	if (*buff != '\0')
		if (sysinfo_dict_add(&sysinfo->properties,
					"Stepping", buff) != 0)
			return false;

	cafe_cpu_info_get_cpu_freq(buff, sizeof(buff));
	if (*buff != '\0') {
		if (sysinfo_dict_add(&sysinfo->properties,
					"CPU Freq", buff) != 0)
			return false;
	} else {
		cafe_cpu_info_get_model_name(buff, sizeof(buff));
		if (*buff != '\0')
			if (sysinfo_dict_add(&sysinfo->properties,
						"Model name", buff) != 0)
				return false;

	}

	for (n = 0; n < 16; ++n) {
		cafe_cpu_info_cache_descriptors(buff, sizeof(buff), n);
		if (*buff != '\0') {
			sprintf(name, "Cache/TLB(%c)", 'a' + c);
			c++;
			if (sysinfo_dict_add(&sysinfo->properties,
					     name, buff) != 0)
				return false;
		}
	}

	return true;
}
#endif
#endif  /* if 0 */
#endif  /* _WRS_KERNEL */


static inline
void sysinfo_emit_snapshot(struct cafe_ep_inst *inst)
{
	struct sysinfo_inst_data *inst_data = inst->client_data;
	void *args[1];
	args[0] = inst->workarea->u.data;

	cafe_dispatch_event(&inst_data->sink_actions, args);
}


static
void sysinfo_snapshot_request_handler(
	struct cafe_event_action_info *event_info, void *event_data[])
{
	sysinfo_emit_snapshot(event_info->inst);
}

/**
 * Fill given struct with sysinfo properties.
 */
int cafe_sysinfo_get(struct cafe_sysconfig_info *sysinfo)
{
	const struct cafe_allocator *allocator;

	assert(sysinfo->properties.type);
	allocator = sysinfo->properties.type->ident->ns->allocator;
	assert(allocator);

#if defined(__VXWORKS__)
	BOOT_PARAMS bootStruct;
#if (_WRS_VXWORKS_MAJOR > 6)
	/* release field is meaningless for VxWorks 7 */
	{
	    size_t len = strlen(runtimeName) +
			strlen(RUNTIME_QUALIFIER_1 RUNTIME_QUALIFIER_2) + 1;
	    sysinfo->sysname = CAFE_ALLOC(allocator, len);
	    if (sysinfo->sysname) {
		(void)snprintf(sysinfo->sysname, len, "%s%s", runtimeName,
		         RUNTIME_QUALIFIER_1 RUNTIME_QUALIFIER_2);
		sysinfo->sysname[len] = 0;
	    }
	    sysinfo->release = NULL;
	}
#else
	sysinfo->sysname = CAFE_STRDUP(allocator, runtimeName);
	sysinfo->release = CAFE_STRDUP(allocator, runtimeVersion);
#endif /* (_WRS_VXWORKS_MAJOR > 6) */
	sysinfo->version = CAFE_STRDUP(allocator, creationDate);

#if defined(_WRS_CONFIG_RTP)
	/* RTP support */
	if (!sysinfo_dict_add(&sysinfo->properties, RTPSUPPORT_KEY, "true"))
		goto error;
#endif
	/* BSP name */
	if (!sysinfo_dict_add(&sysinfo->properties, BSPNAME_KEY,
	                      (sysModel() != NULL) ? sysModel() : "null"))
		goto error;

	/* result of bootStringToStruct() points to EOS on success */
	if (*bootStringToStruct(sysBootLine, &bootStruct) == '\0') {
		/* IP addr */
		if (!sysinfo_dict_add(&sysinfo->properties,
				     IP_ADDR_KEY, bootStruct.ead))
			goto error;

		/* Host IP addr */
		if (!sysinfo_dict_add(&sysinfo->properties,
				     HOST_ADDR_KEY, bootStruct.had))
			goto error;

		/* Gateway addr */
		if (!sysinfo_dict_add(&sysinfo->properties,
				     GATEWAY_KEY, bootStruct.gad))
			goto error;

		/* bootFile */
		if (!sysinfo_dict_add(&sysinfo->properties,
				     BOOTFILE_KEY, bootStruct.bootFile))
			goto error;
	}

//td: enable cpu-specific properties
#if 0
#if (_VX_CPU_FAMILY == _VX_I80X86)
	if (!sysinfo_pentium(sysinfo))
		goto error;
#endif
#endif

#else
	struct utsname name;
	(void)uname(&name);
	sysinfo->sysname = CAFE_STRDUP(allocator, name.sysname);
	sysinfo->release = CAFE_STRDUP(allocator, name.release);
	sysinfo->version = CAFE_STRDUP(allocator, name.version);
#endif
	return CAFE_OK;

#if defined(__VXWORKS__)
error:
	(void)logMsg("Error adding to dictionary\n");
	return CAFE_ERROR;
#endif
}

/**
 * Released allocated field of sysinfo struct
 */
void cafe_sysinfo_free(struct cafe_sysconfig_info *sysinfo)
{
	const struct cafe_allocator *allocator;

	if (sysinfo == NULL)
		return;

	assert(sysinfo->properties.type);
	allocator = sysinfo->properties.type->ident->ns->allocator;
	assert(allocator);
	CAFE_FREE(allocator, sysinfo->sysname);
	CAFE_FREE(allocator, sysinfo->release);
	CAFE_FREE(allocator, sysinfo->version);
}

/**
 *
 */
static
int sysinfo_cfg_start(struct cafe_ep_inst *inst)
{
	struct cafe_sysconfig_info *sysinfo =
		(struct cafe_sysconfig_info *)inst->workarea->u.data;
	int rc;

	cafe_ep_inst_state_started(inst);
	rc = cafe_sysinfo_get(sysinfo);
	if (rc == CAFE_OK)
		sysinfo_emit_snapshot(inst);
	return rc;
}


/**
 *
 */
static
int sysinfo_cfg_get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);
	if (rc != CAFE_OK || event_info == NULL)
		return rc;
	return cafe_add_event_action(
		inst, actions, sysinfo_snapshot_request_handler, event_info);
}


static bool cfg_initialized = false;

int sysinfo_cfg_register(void)
{
	int rc;

	if (cfg_initialized)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		cfg_initialized = true;

	return rc;
}

int sysinfo_cfg_unregister(void)
{
	int rc = CAFE_OK;

	if (cfg_initialized) {
		rc = cafe_ep_spec_unregister(&spec);
		if (rc == CAFE_OK)
			cfg_initialized = false;
	}
	return rc;
}
