/* Copyright (c) 2011, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
04nov14,tle  Fix linux perf skipcount test (WB4-4454)
08jul14,sdt  Used CUT_PRINTF/CUT_PRINT_ERR macro to make clear output
*/

#include "../../src/cafe_os.h"
#include <string.h>
#include <assert.h>
#include <sys/swap.h>
#include <test/cafe_CUnitTest.h>
#include <cafe/atomic.h>
#include <cafe/ep/linux_trace.h>
#include <cafe/autogenerated/cafe.type.h>

#define EP_NAME "linux_perf_skipcount_test"

#define TEST_EVENT_REF  "t[syscalls.sys_enter_swapon]"

#define ALL_TYPES				\
	TEST_EVENT_REF


static const unsigned int NTHREADS = 10;
static const unsigned int NEVENTS = 1000;

static struct cafe_instrument *instrument = NULL;
static struct cafe_ep_inst *inst[2];
static bool input_event_enabled = false;

static const char *test_event_TYPE[] = {
	TEST_EVENT_REF
};

static const char *skipcount_default[] = {
	TEST_EVENT_REF,
	cafe_timestamp_VT_REF,
	cafe_skipcount_VT_REF
};

static const char *skipcount_as_2[] = {
	TEST_EVENT_REF,
	cafe_timestamp_VT_REF,
	"P[cafe.skipcount,2]"
};

static const char *skipcount_as_20[] = {
	TEST_EVENT_REF,
	cafe_timestamp_VT_REF,
	"P[cafe.skipcount,20]"
};


static struct {
	cafe_atomic_val action_handler_count;
	unsigned long skipcount_value;
	cafe_atomic_val event_count;
} event_data;


static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(test_event)
};


static
void action_handler(struct cafe_event_action_info *info, void *data[])
{
	assert(input_event_enabled);
	cafe_atomic_val_inc(&event_data.action_handler_count);

	if (info->event->num_types == 3) {
		const struct cafe_skipcount *skipcount = data[2];
		event_data.skipcount_value = skipcount->value;
		cafe_atomic_val_add(&event_data.event_count, skipcount->value);
	}
}


static
int get_input_event_actions(struct cafe_ep_inst *inst,
			    struct cafe_event_descr *event,
			    struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *info;

	if (!input_event_enabled)
		return CAFE_OK;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &info);

	if (rc != CAFE_OK || info == NULL)
		return rc;

	return cafe_add_event_action(inst, actions, action_handler, info);
}


static struct cafe_ep_spec spec = {
	EP_NAME,
	NULL,			/* create */
	NULL,			/* delete */
	NULL,			/* start */
	NULL,			/* stop */
	get_input_event_actions,
	NULL,			/* update outputs */
	NULL,			/* inject event */
	ALL_TYPES,
	NULL,
	CAFE_NELEMENTS(input_events),
	input_events
};


static bool skipcount_ep_registered = false;


static
void skipcount_ep_unregister(void)
{
	if (skipcount_ep_registered) {
		cafe_ep_spec_unregister(&spec);
		skipcount_ep_registered = false;
	}
}


static
int skipcount_ep_register(void)
{
	int rc = CAFE_OK;

	if (!skipcount_ep_registered) {
		rc = cafe_ep_spec_register(&spec);
		if (rc == CAFE_OK)
			skipcount_ep_registered = true;
	}

	return rc;
}


static
int build_instrument(void)
{
	unsigned int i;
	struct cafe_ep_spec *spec[2];

	CAFE_LOCK();

	CUT_ASSERT_CAFE_OK(linux_trace_register());
	CUT_ASSERT_CAFE_OK(skipcount_ep_register());

	CUT_ASSERT_CAFE_OK(cafe_instrument_create(EP_NAME, &instrument));

	spec[0] = cafe_ep_spec_find("linux_trace");
	spec[1] = cafe_ep_spec_find(EP_NAME);
	CUT_ASSERT_NOT_NULL(spec[0]);
	CUT_ASSERT_NOT_NULL(spec[1]);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument, spec[0], 0, NULL,
					       &inst[0]));

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_create(instrument, spec[1], 0, NULL,
					       &inst[1]));

	for (i = 0; i < spec[1]->n_input_events; ++i) {
		struct cafe_event_descr *const t = inst[1]->input_events[i];

		CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst[0], t));
	}

	/* Bypass threading compatibility checks. */
	inst[0]->flags |= CAFE_EP_INST_FLAG_OUTPUT_SESSION_THREAD;

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_connect(inst[0], inst[1]));

	input_event_enabled = false;
	CUT_ASSERT_CAFE_OK(cafe_instrument_refresh_output_event_actions(instrument));

	CUT_ASSERT_CAFE_OK(cafe_instrument_start(instrument));

	CAFE_UNLOCK();
	return 0;
}


static
int delete_instrument(void)
{
	if (instrument != NULL) {
		CAFE_LOCK();
		CUT_ASSERT_CAFE_OK(cafe_instrument_delete(instrument));
		CAFE_UNLOCK();
		instrument = NULL;
	}
	CAFE_LOCK();
	skipcount_ep_unregister();
	linux_trace_unregister();
	CAFE_UNLOCK();
	return 0;
}


static
void dispatch_event(unsigned int count)
{
	unsigned int i;

	for (i = 0; i < count; ++i) {
		swapon("/non/existant/path", 0);
	}
	sleep(3);
}


static
int test_with_invalid_limit(void)
{
	int rc;
	unsigned int i, j;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;
	static const char *type1_str[] = {
		TEST_EVENT_REF,
		cafe_timestamp_VT_REF,
		"P[cafe.skipcount,4294967296]",
	};
	static const char *type2_str[] = {
		TEST_EVENT_REF,
		cafe_timestamp_VT_REF,
		"P[cafe.skipcount,0]"
	};
	const char **args[2] = { type1_str, type2_str };

	CUT_ASSERT_NOT_NULL(inst[0]);
	CUT_ASSERT_NOT_NULL(inst[1]);

	CAFE_LOCK();

	/* skips the max limit test */
	for (i = 1; i < CAFE_NELEMENTS(args); ++i) {
		for (j = 0; j < 3; ++j)
			CUT_ASSERT_CAFE_OK(cafe_type_parse(instrument->ns,
							   args[i][j],
							   &types[j]));

		CUT_ASSERT_EQUAL(3, j);
		CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, j, types,
						      &event));
		CUT_ASSERT_EQUAL(3, event->num_types);
		CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst[0],
								    event));
		input_event_enabled = true;
		rc = cafe_instrument_refresh_output_event_actions(instrument);
		CUT_ASSERT_EQUAL(CAFE_ERROR, rc);
		CUT_ASSERT_CAFE_OK(cafe_ep_inst_disable_output_event(inst[0],
								     event));
		input_event_enabled = false;
		rc = cafe_instrument_refresh_output_event_actions(instrument);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CAFE_UNLOCK();
	return 0;
}


static
void setup(void)
{
	memset(&event_data, 0, sizeof(event_data));
}


static
int test_skipcount(void)
{
	int rc;
	unsigned int i;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;
	cafe_atomic_val local_count;

	CAFE_LOCK();

	/* Test with default skipcount value. */
	for (i = 0; i < CAFE_NELEMENTS(skipcount_default); ++i) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_default[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(CAFE_NELEMENTS(skipcount_default), event->num_types);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst[0], event));
	input_event_enabled = true;
	CUT_ASSERT_CAFE_OK(cafe_instrument_refresh_output_event_actions(instrument));
	setup();
	dispatch_event(5);
	CUT_ASSERT_EQUAL(5, event_data.action_handler_count);
	CUT_ASSERT_EQUAL(1, event_data.skipcount_value);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disable_output_event(inst[0], event));
	input_event_enabled = false;
	CUT_ASSERT_CAFE_OK(cafe_instrument_refresh_output_event_actions(instrument));

	/* Test with skipcount = 2. */
	for (i = 0; i < CAFE_NELEMENTS(skipcount_as_2); ++i) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_as_2[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(CAFE_NELEMENTS(skipcount_as_2), event->num_types);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst[0], event));
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	/* skipcount does not work immediately, so dispatch a few events before the measurment */
	setup();
	dispatch_event(4);
	setup();
	dispatch_event(5);
	CUT_ASSERT_EQUAL(2, event_data.action_handler_count);
	dispatch_event(2);
	CUT_ASSERT_EQUAL(2, event_data.skipcount_value);
	CUT_ASSERT_EQUAL(3, event_data.action_handler_count);
	rc = cafe_ep_inst_disable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = false;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);

	/* Test with skipcount = 20. */
	for (i = 0; i < CAFE_NELEMENTS(skipcount_as_20); ++i) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_as_20[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}

	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst[0], event));
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	/* skipcount does not work immediately, so dispatch a few events before the measurment */
	setup();
	dispatch_event(30);
	setup();
	dispatch_event(19);
	CUT_ASSERT_EQUAL(20, event_data.skipcount_value);
	local_count = cafe_atomic_set(&event_data.action_handler_count, 0);
	CUT_ASSERT(3 > local_count);
	dispatch_event(20);
	local_count = cafe_atomic_set(&event_data.action_handler_count, 0);
	CUT_ASSERT_EQUAL(20, event_data.skipcount_value);
	CUT_ASSERT_EQUAL(1, local_count);
	setup();
	for (i = 0; i < 200 / 20; ++i) {
		dispatch_event(20);
		CUT_ASSERT_EQUAL(i + 1 % 20, event_data.action_handler_count);
		CUT_ASSERT_EQUAL(20, event_data.skipcount_value);
	}
	CUT_ASSERT_EQUAL(200 / 20, event_data.action_handler_count);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disable_output_event(inst[0], event));
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);

	CAFE_UNLOCK();

	return 0;
}

struct multithreaded_context {
	unsigned int n_events;
	cafe_atomic_val setup_complete;
};


static
void *thread_worker(void *arg)
{
	struct multithreaded_context *ctx = arg;

	while (!cafe_atomic_get(&ctx->setup_complete)) {}
	dispatch_event(ctx->n_events);
	return NULL;
}


static
bool join_all_threads(const pthread_t tids[], unsigned int n_tids)
{
	unsigned int i;
	bool ret = true;

	for (i = 0; i < n_tids; ++i) {
		int rc = pthread_join(tids[i], NULL);
		if (rc != 0) {
			CUT_PRINTF_ERR("error joining thread: %s",
			               strerror(rc));
			ret = false;
		}
	}

	return ret;
}


static
int test_multithreaded_with_skipcount_using_defaults(void)
{
	int rc;
	unsigned int i;
	pthread_t tid[NTHREADS];
	struct multithreaded_context ctx;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;

	/* Test with skipcount default limit. */
	setup();
	CAFE_LOCK();

	for (i = 0; i < CAFE_NELEMENTS(skipcount_default); ++i) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_default[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst[0], event));
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	ctx.n_events = NEVENTS;
	ctx.setup_complete = 0;

	for (i = 0; i < CAFE_NELEMENTS(tid); ++i) {
		rc = pthread_create(&tid[i], NULL, thread_worker, &ctx);
		CUT_ASSERT_EQUAL(0, rc);
	}

	while (!cafe_atomic_cas(&ctx.setup_complete, 0, 1)) {}

	CUT_ASSERT_TRUE(join_all_threads(tid, CAFE_NELEMENTS(tid)));

	CUT_ASSERT_TRUE(event_data.event_count >= 1);
	//CUT_ASSERT_TRUE(event_data.event_count <= (ctx.n_events * NTHREADS) + NTHREADS * NUM_ROOT_INSTANCES);
	CUT_ASSERT_EQUAL(1, event_data.event_count / event_data.action_handler_count);

	CAFE_LOCK();
	rc = cafe_ep_inst_disable_output_event(inst[0], event);
	CUT_ASSERT_CAFE_OK(rc);
	input_event_enabled = false;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	return 0;
}


static
int test_multithreaded_with_skipcount_set_as_twenty(void)
{
	int rc;
	unsigned int i;
	pthread_t tid[NTHREADS];
	struct multithreaded_context ctx;
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event;

	setup();
	CAFE_LOCK();

	for (i = 0; i < CAFE_NELEMENTS(skipcount_as_20); ++i) {
		rc = cafe_type_parse(instrument->ns,
				     skipcount_as_20[i],
				     &types[i]);
		CUT_ASSERT_CAFE_OK(rc);
	}
	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(instrument, i, types, &event));
	CUT_ASSERT_EQUAL(3, event->num_types);
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst[0], event));
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	ctx.n_events = NEVENTS;
	ctx.setup_complete = 0;

	for (i = 0; i < CAFE_NELEMENTS(tid); ++i) {
		rc = pthread_create(&tid[i], NULL, thread_worker, &ctx);
		CUT_ASSERT_EQUAL(0, rc);
	}

	while (!cafe_atomic_cas(&ctx.setup_complete, 0, 1)) {}

	CUT_ASSERT_TRUE(join_all_threads(tid, CAFE_NELEMENTS(tid)));

	CUT_ASSERT(event_data.event_count >= ctx.n_events * (NTHREADS - 1));
	CUT_ASSERT_DOUBLES_EQUAL(20, event_data.event_count / event_data.action_handler_count, 1);
	CAFE_LOCK();
	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disable_output_event(inst[0], event));
	input_event_enabled = true;
	rc = cafe_instrument_refresh_output_event_actions(instrument);
	CUT_ASSERT_CAFE_OK(rc);
	CAFE_UNLOCK();

	return 0;
}


CUT_BEGIN_SUITE(linux_perf_skipcount_suite)
	CUT_RUN_TEST(build_instrument);
	CUT_RUN_TEST(test_with_invalid_limit);
	CUT_RUN_TEST(test_skipcount);
	CUT_RUN_TEST(test_multithreaded_with_skipcount_using_defaults);
	CUT_RUN_TEST(test_multithreaded_with_skipcount_set_as_twenty);
	CUT_RUN_TEST(delete_instrument);
CUT_END_SUITE
