/* verification_cfg.c - used to verify an incoming event stream */

/* Copyright (c) 2010, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/cafe_print.h>
#include <cafe/value_type.h>
#include "../src/cafe_trace.h"
#include "verify_cfg.h"


static bool value_data_equal(const struct cafe_value_type *type,
			     const void *d1, const void *d2);


static
int verify_cfg_start(struct cafe_ep_inst *inst)
{
	struct verify_cfg_workarea *w =
		(struct verify_cfg_workarea *) inst->workarea->u.data;

	assert(inst->workarea != NULL);
	TRACE(("Verifying %"PRIsize_t" events\n", w->events.l));
	w->failed = 0;
	w->sequence = 0;
	cafe_ep_inst_state_started(inst);
	return CAFE_OK;
}


static
int verify_cfg_stop(struct cafe_ep_inst *inst)
{
	struct verify_cfg_workarea *w =
		(struct verify_cfg_workarea *) inst->workarea->u.data;

	if (w->events.l != (w->sequence + w->failed)) {
		printf("verification failed: expected %"PRIsize_t" events, received"
		       " %u\n", w->events.l, w->sequence + w->failed);
		++w->failed;
		return CAFE_ERROR;
	}

	if (w->failed != 0)
		return CAFE_ERROR;
	cafe_ep_inst_state_stopped(inst);
	return CAFE_OK;
}


/**
 *  Given two values of the same type, compare them
 */
static
bool value_data_equal(const struct cafe_value_type *type,
		      const void *d1, const void *d2)
{
	switch(type->kind) {
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_USHORT:
	case CAFE_TYPE_INT:
	case CAFE_TYPE_UINT:
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_ULONG:
	case CAFE_TYPE_LONGLONG:
	case CAFE_TYPE_ULONGLONG:
		if (memcmp(d1, d2, type->size))
			return false;
		return true;

	case CAFE_TYPE_FLOAT: {
		float f1 = *(float *)d1;
		float f2 = *(float *)d2;
		return f1 == f2;
	}

	case CAFE_TYPE_DOUBLE: {
		double f1 = *(double *)d1;
		double f2 = *(double *)d2;
		return f1 == f2;
	}

	case CAFE_TYPE_STRUCT:
	{
		const struct cafe_vm_attr *attr;

		cafe_type_attr_foreach(attr, type) {
			if (!value_data_equal(cafe_attr_type(attr),
					      (char *)d1 + attr->offset,
					      (char *)d2 + attr->offset))
				return false;
		}
		return true;
	}

	case CAFE_TYPE_BOOL:
	{
		bool b1 = *(bool *) d1;
		bool b2 = *(bool *) d2;

		if ((b1 && !b2) || (!b1 && b2))
			return false;
		return true;
	}

	case CAFE_TYPE_STRING:
	{
		const char *s1 = *(char **) d1;
		const char *s2 = *(char **) d2;

		if (strcmp(s1, s2) == 0)
			return true;
		return false;
	}

	case CAFE_TYPE_BLOB:
	{
		const struct cafe_vm_blob *b1 = (struct cafe_vm_blob *) d1;
		const struct cafe_vm_blob *b2 = (struct cafe_vm_blob *) d2;

		if (b1->l != b2->l)
			return false;

		if (memcmp(b1, b2, b1->l) == 0)
			return true;
		return false;
	}

	case CAFE_TYPE_DICT:
	{
		struct cafe_dict_struct *dict1 = (struct cafe_dict_struct *) d1;
		struct cafe_dict_struct *dict2 = (struct cafe_dict_struct *) d2;
		struct cafe_dict_iterator iter1, iter2;
		struct cafe_value_data *val1, *val2;

		if (cafe_dict_count_rows(dict1) != cafe_dict_count_rows(dict2))
			return false;

		/* test each row in turn */
		val1 = cafe_dict_first(dict1, &iter1);
		val2 = cafe_dict_first(dict2, &iter2);
		for (;(val1 != 0) && (val2 != 0);
		     val1 = cafe_dict_next(dict1, &iter1),
			     val2 = cafe_dict_next(dict2, &iter2)) {
			if (!value_data_equal(cafe_element_type(type),
					      val1->u.data, val2->u.data))
				return false;
		}
		return true;
	}

	case CAFE_TYPE_TYPE:
	default:
		return false;
	}
	return true;
}


/**
 * Returns true iff the actual type and data match the expected type and data.
 */
bool validate_type_data(struct cafe_value_type *actual_type,
			const char *expected_type,
			const void *event_data,
			const void *expected_data,
			size_t data_size)
{
	/* test expected value type */
	if (strcmp(actual_type->type_str, expected_type) != 0) {
		printf("verification failed, event doesn't match expected "
		       "value (expected \"%s\", received \"%s\")\n",
		       expected_type, actual_type->type_str);
		return false;
	}

	/* test expected value data */
	if (!value_data_equal(actual_type, event_data, expected_data)) {
		printf("verification failed, event data doesn't match expected "
		       "value\n");
		printf("got:");
		(void)cafe_event_data_print(actual_type, event_data);
		printf("expected:");
		(void)cafe_event_data_print(actual_type, expected_data);
		return false;
	}
	return true;
}


static
void verify_event(struct cafe_event_action_info *event_info, void **event_data)
{
	struct cafe_ep_inst *inst = event_info->inst;
	struct verify_cfg_workarea *w =
		(struct verify_cfg_workarea *) inst->workarea->u.data;
	struct event_verify_spec *event;
	const char *verification_data;
	const char **verification_types;
	unsigned int i;

	TRACE(("Verifying... (%u/%"PRIsize_t")\n", w->sequence, w->events.l));
	if (w->events.l <= w->sequence) {
		printf("verification failed, not expecting event %u\n",
		       w->sequence);
		goto fail;
	}

	event = (struct event_verify_spec *)w->events.p
		+ w->sequence + w->failed;

	if (event_info->event->num_types < event->types.l) {
		printf("expected %"PRIsize_t" types, only got %u\n", event->types.l,
		       event_info->event->num_types);
		goto fail;
	}

	/* Make sure we receive at least the expected types and data with this
	   event.  This includes the core event type/data, as well as
	   orthogonal and contributed types */
	verification_data = event->data.p;
	verification_types = event->types.p;
	for (i = 0; i < event->types.l; ++i) {
		if (!validate_type_data(event_info->event->types[i],
					verification_types[i],
					event_data[i],
					verification_data,
					event_info->event->types[i]->size))
			goto fail;
		verification_data += event_info->event->types[i]->size;
	}

	++w->sequence;
	cafe_dispatch_event(&event_info->sink_actions, event_data);
	return;

fail:
	++w->failed;
	cafe_dispatch_error(event_info->inst, CAFE_ERROR);
}


static
int verify_cfg_get_input_event_actions(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);
	if (rc != CAFE_OK || event_info == NULL)
		return rc;
	return cafe_add_event_action(inst, actions, verify_event, event_info);
}


const char *verify_event_TYPE[] = {cafe_wildcard_VT_REF};
static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT(verify_event)
};


static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT(cafe_wildcard_VT_REF)
};


static struct cafe_ep_spec spec = {
	"verify_cfg",
	NULL,
	NULL,
	verify_cfg_start,
	verify_cfg_stop,
	verify_cfg_get_input_event_actions,
	NULL,
	NULL,
	cafe_input_any_ctx_VT_DEF
	cafe_wildcard_VT_DEF
	"S[verify.event_spec,v[types,a[s]]v[events,b]]",
	/* WRS_FIXME: This ep is not thread or interrupt safe and
	 * therefore should not have cafe_input_any_ctx_VT_NAT set.
	 * However, has long as we don't count on in being thread
	 * safe, setting this nature makes it possible to use with EPs
	 * that needs to connect to safe EPs. */
	"S[verify_cfg.workarea,v[sequence,I]v[failed,I]v[events,a[t[verify.event_spec]]]," cafe_input_any_ctx_VT_NAT "]",
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events
};


static bool cfg_registered = false;


void verify_cfg_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}


int verify_cfg_register(void)
{
	int rc;

	if (cfg_registered)
		return CAFE_OK;

	rc = cafe_ep_spec_register(&spec);
	if (rc == CAFE_OK)
		cfg_registered = true;

	return rc;
}
