/* vxworks_eventtest_host.c - host-only parts of vxWorks event test */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
30jun14,sdt  Used new TEST_ENABLE_SET macro to handle test_enabled flag

DESCRIPTION
*/

#include <string.h>		/* memset */
#include <test/cafe_CUnitTest.h>
#include <cafe/autogenerated/vxworks.type.h>
#include "../vxworks_events.h"
#include "../vxworks_kernel_host.h"
#include "vxworks_eventtest.h"

/*
  The POST_EVENT_BEGIN() macro creates the top half of a function to emit the
  event of interest
  The POST_EVENT_END() macro emits the bottom half. In between these macros,
  the event structure, e, can be populated.
  The _NO_STRUCT variants are for events which do not have an event struct
*/


#define POST_EVENT_BEGIN(name)					\
static int post_##name(void)					\
{								\
	struct vxWorks_##name *e;				\
	e = malloc(sizeof(*e));					\
	if (e == NULL)						\
		return CAFE_NOMEM; \

#define POST_EVENT_END(evname)					\
	cafe_cfg_vxworks_ev_process_id(evname, e, NULL);	\
	free(e);						\
	return CAFE_OK;						\
}

#define POST_EVENT_BEGIN_NO_STRUCT(name)			\
static int post_##name(void)					\
{								\

#define POST_EVENT_END_NO_STRUCT(evname)			\
	cafe_cfg_vxworks_ev_process_id(evname, NULL, NULL);	\
	return CAFE_OK;						\
}

/*
  The TEST_EVENT() macro calls the function to generate the event, and then
  verifies that the event was received correctly. It calls a function which
  returns true if the event structure contents are correct
  The _NO_STRUCT variant is for events which do not have an event structure
*/

#define TEST_EVENT(funcname)					\
int funcname##_test(struct cafe_instrument *session)		\
{								\
	unsigned int i;						\
	int result = 0;						\
	struct vxWorks_##funcname *e;				\
	reset_logged_events();					\
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, funcname##_TYPE[0], \
					&expected_type));	\
	TEST_ENABLE_SET(1);					\
	CUT_ASSERT_CAFE_OK(post_##funcname());			\
	TEST_ENABLE_SET(0);					\
	for (i = 0; i < logged_event_count; i++) {		\
		if (logged_events[i].num_types != 2)		\
			continue;				\
		if (logged_events[i].types[0] != expected_type) \
			continue;				\
		e = &logged_events[i].event_data[0].u.funcname; \
		if (funcname##_test_result(e)) {		\
			result = 1;				\
			break;					\
		}						\
	}							\
	CUT_ASSERT_NOT_EQUAL_MESSAGE(#funcname " event", 0, result); \
	return 0;						\
}

#define TEST_EVENT_NO_STRUCT(funcname)				\
int funcname##_test(struct cafe_instrument *session)		\
{								\
	unsigned int i;						\
	int result = 0;						\
	reset_logged_events();					\
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, funcname##_TYPE[0], \
					&expected_type));	\
	TEST_ENABLE_SET(1);					\
	CUT_ASSERT_CAFE_OK(post_##funcname());			\
	TEST_ENABLE_SET(0);					\
	for (i = 0; i < logged_event_count; i++) {		\
		if (logged_events[i].num_types != 2)		\
			continue;				\
		if (logged_events[i].types[0] != expected_type) \
			continue;				\
		result = 1;					\
		break;						\
	}							\
	CUT_ASSERT_NOT_EQUAL_MESSAGE(#funcname " event", 0, result); \
	return 0; 						\
}

POST_EVENT_BEGIN(taskName)
	e->state = 1;
	e->priority = 2;
	e->lockCnt = 3;
	e->tid = 12345;
	e->entryPt = 0x2000;
	e->rtp = 0x3000;
	e->affinity = 4;
	e->name = (char *)"made-up-task";
POST_EVENT_END(EVENT_TASKNAME)

static bool taskName_test_result(struct vxWorks_taskName *e)
{
	return (e->state == 1) &&
		(e->priority == 2) &&
		(e->lockCnt == 3) &&
		(e->tid == 12345) &&
		(e->entryPt == 0x2000) &&
		(e->rtp == 0x3000) &&
		(e->affinity == 4) &&
		(strcmp(e->name, "made-up-task") == 0);
}

int taskName_test (struct cafe_instrument *session)
{
	struct vxWorks_taskName *e;
	unsigned int i;
	int result = 0;

	reset_logged_events();

	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, taskName_TYPE[0],
					   &expected_type));

	TEST_ENABLE_SET(1);

	post_taskName();

	TEST_ENABLE_SET(0);

	for (i = 0; i < logged_event_count; i++) {
		e = &logged_events[i].event_data[0].u.taskName;
		/* taskname doesn't have timestamp */
		if (logged_events[i].num_types != 1)
			continue;
		if (taskName_test_result(e)) {
			result = 1;
			break;
		}
	}
	CUT_ASSERT_NOT_EQUAL_MESSAGE("taskName event", 0, result);
	return 0;
}

POST_EVENT_BEGIN(taskDelay)
	e->objId = 1234;
	e->classId = 0x8000;
	e->ticks = 10;
POST_EVENT_END(EVENT_TASKDELAY)

static bool taskDelay_test_result(struct vxWorks_taskDelay *e)
{
	return ((e->objId == 1234) &&
		(e->classId == 0x8000) &&
		(e->ticks == 10));
}

TEST_EVENT(taskDelay)

POST_EVENT_BEGIN(memAlloc)
	e->objId = 5678;
	e->classId = 0x9000;
	e->nBytesRequested = 987;
POST_EVENT_END(EVENT_MEMALLOC)

static bool memAlloc_test_result(struct vxWorks_memAlloc *e)
{
	return (e->objId == 5678) &&
		(e->classId == 0x9000) &&
		(e->nBytesRequested == 987);
}

TEST_EVENT(memAlloc)

POST_EVENT_BEGIN(memRealloc)
	e->objId = 5678;
	e->classId = 0x9000;
	e->partId = 0xa000;
	e->oldBlockSize = 987;
	e->pOldBlock = 123456;
	e->pNewBlock = 987654;
POST_EVENT_END(EVENT_MEMREALLOC)

static bool memRealloc_test_result(struct vxWorks_memRealloc *e)
{
	return ((e->objId == 5678) &&
		(e->classId == 0x9000) &&
		(e->partId == 0xa000) &&
		(e->oldBlockSize == 987) &&
		(e->pOldBlock == 123456) &&
		(e->pNewBlock == 987654));
}

TEST_EVENT(memRealloc)

POST_EVENT_BEGIN(memPartCreate)
	e->objId = 5678;
	e->classId = 0x9000;
	e->partId = 0xa000;
	e->poolSize = 123456;
POST_EVENT_END(EVENT_MEMPARTCREATE)

static bool memPartCreate_test_result(struct vxWorks_memPartCreate *e)
{
	return ((e->objId == 5678) &&
		(e->classId == 0x9000) &&
		(e->partId == 0xa000) &&
		(e->poolSize == 123456));
}

TEST_EVENT(memPartCreate)

POST_EVENT_BEGIN(memFree)
	e->objId = 5678;
	e->classId = 0x9000;
POST_EVENT_END(EVENT_MEMFREE)

static bool memFree_test_result(struct vxWorks_memFree *e)
{
	return ((e->objId == 5678) &&
		(e->classId == 0x9000));
}

TEST_EVENT(memFree)


POST_EVENT_BEGIN(windDelete)
	e->taskId = 5678;
POST_EVENT_END(EVENT_WINDDELETE)

static bool windDelete_test_result(struct vxWorks_windDelete *e)
{
	return 5678 == e->taskId;
}

TEST_EVENT(windDelete)

static int post_windExitDispatch(int pi)
{
	struct vxWorks_windExitDispatch *e;
	e = malloc(sizeof (*e));
	if (e == NULL)
		return CAFE_NOMEM;
	e->taskId = 1234;
	e->priority = 37;

	if (pi)
		cafe_cfg_vxworks_ev_process_id(EVENT_WIND_EXIT_DISPATCH_PI, e, NULL);
	else
		cafe_cfg_vxworks_ev_process_id(EVENT_WIND_EXIT_DISPATCH, e, NULL);

	free(e);
	return CAFE_OK;
}

int windExitDispatch_test(struct cafe_instrument *session, int pi)
{
	struct vxWorks_windExitDispatch *e;
	int result = 0;
	unsigned int i;

	reset_logged_events();
	if (pi)
		CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, windExitDispatchPI_TYPE[0],
						   &expected_type));
	else
		CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, windExitDispatch_TYPE[0],
						   &expected_type));

	TEST_ENABLE_SET(1);

	post_windExitDispatch(pi);
	TEST_ENABLE_SET(0);

	for (i = 0; i < logged_event_count; i++) {
		e = &logged_events[i].event_data[0].u.windExitDispatch;
		if (logged_events[i].num_types != 2)
			continue;
		if ((1234 == e->taskId) &&
		    (37 == e->priority)) {
			result = 1;
			break;
		}
	}

	if (pi)
		CUT_ASSERT_NOT_EQUAL_MESSAGE("windExitDispatchPI event", 0, result);
	else
		CUT_ASSERT_NOT_EQUAL_MESSAGE("windExitDispatch event", 0, result);
	return 0;
}

POST_EVENT_BEGIN(taskSpawn)
	e->taskId = 1234;
	e->priority = 37;
	e->entryPoint = 0x1234;
POST_EVENT_END(EVENT_TASKSPAWN)

static bool taskSpawn_test_result(struct vxWorks_taskSpawn *e)
{
	return (1234 == e->taskId) &&
		(37 == e->priority) &&
		(0x1234 == e->entryPoint);
}

TEST_EVENT(taskSpawn)


POST_EVENT_BEGIN(intEnter)
	e->interrupt = 1234;
POST_EVENT_END(EVENT_INT_ENTER)

static bool intEnter_test_result(struct vxWorks_intEnter *e)
{
	return e->interrupt == 1234;
}

TEST_EVENT(intEnter)


int taskLockUnlock_test(struct cafe_instrument *session, int lock)
{
	unsigned int i;
	int result = 0;

	reset_logged_events();
	if (lock)
		CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, taskLock_TYPE[0],
						   &expected_type));
	else
		CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, taskUnlock_TYPE[0],
						   &expected_type));

	TEST_ENABLE_SET(1);

	if (lock)
		cafe_cfg_vxworks_ev_process_id(EVENT_TASKLOCK, NULL, 0);
	else
		cafe_cfg_vxworks_ev_process_id(EVENT_TASKUNLOCK, NULL, 0);

	TEST_ENABLE_SET(0);

	for (i = 0; i < logged_event_count; i++) {
		if (logged_events[i].num_types != 2)
			continue;
		if (logged_events[i].types[0] != expected_type)
			continue;

		result = 1;
	}

	if (lock)
		CUT_ASSERT_NOT_EQUAL_MESSAGE("taskLock event", 0, result);
	else
		CUT_ASSERT_NOT_EQUAL_MESSAGE("taskUnlockevent", 0, result);
	return 0;
}


int intExit_test(struct cafe_instrument *session, int k)
{
	unsigned int i;
	int result = 0;

	reset_logged_events();
	if (k)
		CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, intExitK_TYPE[0],
						   &expected_type));
	else
		CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, intExit_TYPE[0],
						   &expected_type));

	TEST_ENABLE_SET(1);

	if (k)
		cafe_cfg_vxworks_ev_process_id(EVENT_INT_EXIT_K, NULL, 0);
	else
		cafe_cfg_vxworks_ev_process_id(EVENT_INT_EXIT, NULL, 0);

	TEST_ENABLE_SET(0);

	for (i = 0; i < logged_event_count; i++) {
		if (logged_events[i].num_types != 2)
			continue;
		if (logged_events[i].types[0] != expected_type)
			continue;

		result = 1;
	}

	CUT_ASSERT_NOT_EQUAL_MESSAGE("intExit event", 0, result);
	return 0;
}

POST_EVENT_BEGIN_NO_STRUCT(windExitIdle)
POST_EVENT_END_NO_STRUCT(EVENT_WIND_EXIT_IDLE)

TEST_EVENT_NO_STRUCT(windExitIdle)


POST_EVENT_BEGIN_NO_STRUCT(tickAnnounce)
POST_EVENT_END_NO_STRUCT(EVENT_TICKANNOUNCE)

TEST_EVENT_NO_STRUCT(tickAnnounce)

int timestampConfig_test (struct cafe_instrument *session,
			  struct cafe_ep_inst *vxworks)
{
	unsigned int i;
	int result = 0;
	struct cafe_timestamp_config *e;

	reset_logged_events();
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, timestampConfig_TYPE[0],
					   &expected_type));

	CAFE_LOCK();
	cafe_ep_inst_stop(vxworks);

	TEST_ENABLE_SET(1);
	cafe_ep_inst_start(vxworks);
	TEST_ENABLE_SET(0);
	CAFE_UNLOCK();

	for (i = 0; i < logged_event_count; i++) {
		if (logged_events[i].types[0] != expected_type)
			continue;
		if (logged_events[i].num_types != 1)
			continue;
		e = &logged_events[i].event_data[0].u.timestampConfig;
		if (e->frequency == 12345678) {
			result = 1;
			break;
		}

		result = 1;
	}
	CUT_ASSERT_NOT_EQUAL_MESSAGE("timestampConfig event", 0, result);
	return 0;
}

int taskNameAtStart_test (struct cafe_instrument *session,
			  struct cafe_ep_inst *vxworks)
{
	unsigned int i;
	int result = 0;
	struct vxWorks_taskName *e;

	reset_logged_events();
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, taskName_TYPE[0],
					   &expected_type));

	CAFE_LOCK();
	cafe_ep_inst_stop(vxworks);

	TEST_ENABLE_SET(1);
	cafe_ep_inst_start(vxworks);
	TEST_ENABLE_SET(0);
	CAFE_UNLOCK();

	for (i = 0; i < logged_event_count; i++) {
		if (logged_events[i].types[0] != expected_type)
			continue;
		if (logged_events[i].num_types != 1)
			continue;
		e = &logged_events[i].event_data[0].u.taskName;
		if ((e->state == 1) &&
		    (e->priority == 2) &&
		    (e->lockCnt == 3) &&
		    (e->tid == 12345) &&
		    (e->entryPt == 0x2000) &&
		    (e->rtp == 0x3000) &&
		    (e->affinity == 4) &&
		    (strcmp(e->name, "start-up-task") == 0)) {
			result = 1;
			break;
		}

		result = 1;
	}
	CUT_ASSERT_NOT_EQUAL_MESSAGE("taskName event at start", 0, result);
	return 0;
}

int numCores_test(struct cafe_instrument *session,
		  struct cafe_ep_inst *vxworks)
{
	unsigned int i;
	int result = 0;
	struct cafe_num_cores *e;

	reset_logged_events();
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, numCores_TYPE[0],
					   &expected_type));

	CAFE_LOCK();
	cafe_ep_inst_stop(vxworks);

	TEST_ENABLE_SET(1);
	cafe_ep_inst_start(vxworks);
	TEST_ENABLE_SET(0);
	CAFE_UNLOCK();

	for (i = 0; i < logged_event_count; i++) {
		if (logged_events[i].num_types != 1)
			continue;
		if (logged_events[i].types[0] != expected_type)
			continue;
		e = &logged_events[i].event_data[0].u.numCores;
		if (e->num_cores == 2) {
			result = 1;
			break;
		}
	}
	CUT_ASSERT_NOT_EQUAL_MESSAGE("numCores event", 0, result);
	return 0;
}

void ioTests_treadown (void)
{
    return;
}

POST_EVENT_BEGIN(ioOpen)
	e->fd = 123;
	e->classId = 0x1234;
	e->fdEntry = 0x5678;
	e->flags = 0x55;
	e->filename.p = (char *)"filename";
	e->filename.l = strlen(e->filename.p);
POST_EVENT_END(EVENT_IOOPEN)

static bool ioOpen_test_result(struct vxWorks_ioOpen *e)
{
	return 	e->fd == 123 &&
		e->classId == 0x1234 &&
		e->fdEntry == 0x5678 &&
		e->flags == 0x55;
}

TEST_EVENT(ioOpen)

POST_EVENT_BEGIN(ioCreat)
	e->fd = 123;
	e->classId = 0x1234;
	e->fdEntry = 0x5678;
	e->flags = 0x55;
	e->filename.p = (char *)"filename";
	e->filename.l = strlen(e->filename.p);
POST_EVENT_END(EVENT_IOCREAT)

static bool ioCreat_test_result(struct vxWorks_ioCreat *e)
{
	return 	e->fd == 123 &&
		e->classId == 0x1234 &&
		e->fdEntry == 0x5678 &&
		e->flags == 0x55;
}

TEST_EVENT(ioCreat)


POST_EVENT_BEGIN(ioRead)
	e->fd = 123;
	e->classId = 0x1234;
	e->fdEntry = 0x5678;
	e->result = 987;
	e->buffer = 0x56789;
	e->data.p = (char *)"hello";
	e->data.l = strlen("hello");
POST_EVENT_END(EVENT_IOREAD)

static bool ioRead_test_result(struct vxWorks_ioRead *e)
{
	return ((e->fd == 123) &&
		(e->classId == 0x1234) &&
		(e->fdEntry == 0x5678) &&
		(e->result == 987) &&
		(e->buffer == 0x56789) &&
		(strncmp("hello", e->data.p, e->data.l) == 0));
}

TEST_EVENT(ioRead)

POST_EVENT_BEGIN(ioWrite)
	e->fd = 123;
	e->classId = 0x1234;
	e->fdEntry = 0x5678;
	e->nBytes = 987;
	e->buffer = 0x56789;
	e->data.p = (char *)"event text";
	e->data.l = strlen("event text");
POST_EVENT_END(EVENT_IOWRITE)


static bool ioWrite_test_result(struct vxWorks_ioWrite *e)
{
	return ((e->fd == 123) &&
		(e->classId == 0x1234) &&
		(e->fdEntry == 0x5678) &&
		(e->nBytes == 987) &&
		(e->buffer == 0x56789) &&
		(strncmp("event text", e->data.p, e->data.l) == 0));
}

TEST_EVENT(ioWrite)

POST_EVENT_BEGIN(ioClose)
	e->fd = 123;
	e->classId = 0x1234;
	e->fdEntry = 0x5678;
POST_EVENT_END(EVENT_IOCLOSE)

static bool ioClose_test_result(struct vxWorks_ioClose *e)
{
	return e->fd == 123 &&
		e->classId == 0x1234 &&
		e->fdEntry == 0x5678;
}

TEST_EVENT(ioClose)

POST_EVENT_BEGIN(ioIoctl)
	e->fd = 123;
	e->arg = 99;
	e->cmd = 77;
	e->classId = 0x1234;
	e->fdEntry = 0x5678;
POST_EVENT_END(EVENT_IOIOCTL)

static bool ioIoctl_test_result(struct vxWorks_ioIoctl *e)
{
	return ((e->fd == 123) &&
		(e->classId == 0x1234) &&
		(e->fdEntry == 0x5678) &&
		(e->arg == 99) &&
		(e->cmd == 77));
}

TEST_EVENT(ioIoctl)

POST_EVENT_BEGIN(ioDelete)
e->fileName.p = (char *)"some_file";
	e->fileName.l = strlen("some_file");
POST_EVENT_END(EVENT_IODELETE)

static bool ioDelete_test_result(struct vxWorks_ioDelete *e)
{
	return (strncmp("some_file", e->fileName.p, e->fileName.l) == 0);
}

TEST_EVENT(ioDelete)

POST_EVENT_BEGIN(windPrioritySetRaise)
	e->taskId = 5678;
	e->oldPri = 789;
	e->newPri = 234;
POST_EVENT_END(EVENT_WINDPRIORITYSETRAISE)

static bool windPrioritySetRaise_test_result(struct vxWorks_windPrioritySetRaise *e)
{
	return ((5678 == e->taskId) &&
		(789 == e->oldPri) &&
		(234 == e->newPri));
}

TEST_EVENT(windPrioritySetRaise)

POST_EVENT_BEGIN(windPrioritySetLower)
	e->taskId = 5678;
	e->oldPri = 123;
	e->newPri = 456;
POST_EVENT_END(EVENT_WINDPRIORITYSETLOWER)

static bool windPrioritySetLower_test_result(struct vxWorks_windPrioritySetLower *e)
{
	return 	e->taskId == 5678 &&
		e->oldPri == 123 &&
		e->newPri == 456;
}

TEST_EVENT(windPrioritySetLower)


void semTests_treadown (void)
{
    return;
}

POST_EVENT_BEGIN(semBCreate)
	e->classId = 1234;
	e->objId = 4567;
	e->semId = 5432;
	e->owner = 9876;
	e->options = 8901;
POST_EVENT_END(EVENT_SEMBCREATE)

static bool semBCreate_test_result(struct vxWorks_semBCreate *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->semId == 5432) &&
		(e->owner == 9876) &&
		(e->options == 8901));
}

TEST_EVENT(semBCreate)

POST_EVENT_BEGIN(semMCreate)
	e->classId = 1234;
	e->objId = 4567;
	e->semId = 5432;
	e->owner = 9876;
	e->options = 8901;
POST_EVENT_END(EVENT_SEMMCREATE)

static bool semMCreate_test_result(struct vxWorks_semMCreate *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->semId == 5432) &&
		(e->owner == 9876) &&
		(e->options == 8901));
}

TEST_EVENT(semMCreate)

POST_EVENT_BEGIN(semCCreate)
	e->classId = 1234;
	e->objId = 4567;
	e->semId = 5432;
	e->initialCount = 9876;
	e->options = 8901;
POST_EVENT_END(EVENT_SEMCCREATE)

static bool semCCreate_test_result(struct vxWorks_semCCreate *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->semId == 5432) &&
		(e->initialCount == 9876) &&
		(e->options == 8901));
}

TEST_EVENT(semCCreate)

POST_EVENT_BEGIN(semDelete)
	e->classId = 1234;
	e->objId = 4567;
	e->semId = 5432;
POST_EVENT_END(EVENT_SEMDELETE)

static bool semDelete_test_result(struct vxWorks_semDelete *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->semId == 5432));
}

TEST_EVENT(semDelete)

POST_EVENT_BEGIN(semTake)
	e->classId = 1234;
	e->objId = 4567;
	e->semId = 5432;
	e->owner = 9876;
	e->recurse = 8901;
POST_EVENT_END(EVENT_SEMTAKE)

static bool semTake_test_result(struct vxWorks_semTake *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->semId == 5432) &&
		(e->owner == 9876) &&
		(e->recurse == 8901));
}

TEST_EVENT(semTake)

POST_EVENT_BEGIN(semGive)
	e->classId = 1234;
	e->objId = 4567;
	e->semId = 5432;
	e->owner = 9876;
	e->recurse = 8901;
POST_EVENT_END(EVENT_SEMGIVE)

static bool semGive_test_result(struct vxWorks_semGive *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->semId == 5432) &&
		(e->owner == 9876) &&
		(e->recurse == 8901));
}

TEST_EVENT(semGive)


POST_EVENT_BEGIN(semFlush)
	e->classId = 1234;
	e->objId = 4567;
	e->semId = 5432;
	e->qHead = 9876;
	e->recurse = 8901;
POST_EVENT_END(EVENT_SEMFLUSH)

static bool semFlush_test_result(struct vxWorks_semFlush *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->semId == 5432) &&
		(e->qHead == 9876) &&
		(e->recurse == 8901));
}

TEST_EVENT(semFlush)


POST_EVENT_BEGIN(msgQSend)
	e->classId = 1234;
	e->objId = 4567;
	e->msgQId = 5432;
	e->buffer = 9876;
	e->nBytes = 321;
	e->priority = 423;
POST_EVENT_END(EVENT_MSGQSEND)

static bool msgQSend_test_result(struct vxWorks_msgQSend *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->msgQId == 5432) &&
		(e->buffer == 9876) &&
		(e->priority == 423) &&
		(e->nBytes == 321));
}

TEST_EVENT(msgQSend)

POST_EVENT_BEGIN(msgQRecv)
	e->classId = 1234;
	e->objId = 4567;
	e->msgQId = 5432;
	e->buffer = 9876;
	e->maxBytes = 321;
POST_EVENT_END(EVENT_MSGQRECEIVE)

static bool msgQRecv_test_result(struct vxWorks_msgQRecv *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->msgQId == 5432) &&
		(e->buffer == 9876) &&
		(e->maxBytes == 321));
}

TEST_EVENT(msgQRecv)


POST_EVENT_BEGIN(msgQCreate)
	e->classId = 1234;
	e->objId = 4567;
	e->msgQId = 5432;
	e->maxMsgs = 12;
	e->maxMsgLength = 34;
	e->options = 56;
POST_EVENT_END(EVENT_MSGQCREATE)

static bool msgQCreate_test_result(struct vxWorks_msgQCreate *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->msgQId == 5432) &&
		(e->maxMsgs == 12) &&
		(e->maxMsgLength == 34) &&
		(e->options == 56));
}

TEST_EVENT(msgQCreate)

POST_EVENT_BEGIN(msgQDelete)
	e->classId = 1234;
	e->objId = 4567;
	e->msgQId = 5432;
POST_EVENT_END(EVENT_MSGQDELETE)

static bool msgQDelete_test_result(struct vxWorks_msgQDelete *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->msgQId == 5432));
}

TEST_EVENT(msgQDelete)

POST_EVENT_BEGIN(wdCreate)
	e->classId = 1234;
	e->objId = 4567;
	e->wdId = 5432;
POST_EVENT_END(EVENT_WDCREATE)

static bool wdCreate_test_result(struct vxWorks_wdCreate *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->wdId == 5432));
}

TEST_EVENT(wdCreate)

POST_EVENT_BEGIN(wdDelete)
	e->classId = 1234;
	e->objId = 4567;
	e->wdId = 5432;
POST_EVENT_END(EVENT_WDDELETE)

static bool wdDelete_test_result(struct vxWorks_wdDelete *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->wdId == 5432));
}

TEST_EVENT(wdDelete)

POST_EVENT_BEGIN(wdStart)
	e->classId = 1234;
	e->objId = 4567;
	e->wdId = 5432;
	e->delay = 123;
POST_EVENT_END(EVENT_WDSTART)

static bool wdStart_test_result(struct vxWorks_wdStart *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->wdId == 5432) &&
		(e->delay == 123));
}

TEST_EVENT(wdStart)

POST_EVENT_BEGIN(wdCancel)
	e->classId = 1234;
	e->objId = 4567;
	e->wdId = 5432;
POST_EVENT_END(EVENT_WDCANCEL)

static bool wdCancel_test_result(struct vxWorks_wdCancel *e)
{
	return ((e->classId == 1234) &&
		(e->objId == 4567) &&
		(e->wdId == 5432));
}

TEST_EVENT(wdCancel)


#if 0
static int postObjSemEvent(int test)
{
	struct vxWorks_objSemGive *e0;
	struct vxWorks_objSemTake *e1;
	struct vxWorks_objSemRTake *e2;
	struct vxWorks_objSemWTake *e3;

	switch(test) {
	case 0:
		e0 = malloc(sizeof(*e0));
		CUT_ASSERT(e0);
		e0->semId = 1234;
		cafe_cfg_vxworks_ev_process_id(EVENT_OBJ_SEMGIVE, e0, NULL);
		free(e0);
		break;
	case 1:
		e1 = malloc(sizeof(*e1));
		CUT_ASSERT(e1);
		e1->semId = 1234;
		cafe_cfg_vxworks_ev_process_id(EVENT_OBJ_SEMTAKE, e1, NULL);
		free(e1);
		break;
	case 2:
		e2 = malloc(sizeof(*e2));
		CUT_ASSERT(e2);
		e2->semId = 1234;
		cafe_cfg_vxworks_ev_process_id(EVENT_OBJ_SEMRTAKE, e2, NULL);
		free(e2);
		break;
	case 3:
		e3 = malloc(sizeof(*e3));
		CUT_ASSERT(e3);
		e3->semId = 1234;
		cafe_cfg_vxworks_ev_process_id(EVENT_OBJ_SEMWTAKE, e3, NULL);
		free(e3);
		break;

	}
	return 0;
}


int objSem_test(struct cafe_instrument *session, int test)
{
	int result = 0;
	unsigned int i;
	const char *type_str;

	const char *msg[4] = {
		"objSemGive event",
		"objSemTake event",
		"objSemRTake event",
		"objSemWTake event"
	};

	switch (test) {
	case 0: type_str = objSemGive_TYPE[0];
		break;
	case 1: type_str = objSemTake_TYPE[0];
		break;
	case 2: type_str = objSemRTake_TYPE[0];
		break;
	case 3: type_str = objSemWTake_TYPE[0];
		break;
	default:
		CUT_ASSERT_MESSAGE("Invalid test type", 0);
	}

	reset_logged_events();
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, type_str, &expected_type));

	TEST_ENABLE_SET(1);
	if (postObjSemEvent(test) != 0)
		return 0;

	TEST_ENABLE_SET(0);

	for (i = 0; i < logged_event_count; i++) {
		unsigned long long objId;
		switch (test) {
		case 0:
			objId = logged_events[i].event_data[0].u.objSemGive.semId;
			break;
		case 1:
			objId = logged_events[i].event_data[0].u.objSemTake.semId;
			break;
		case 2:
			objId = logged_events[i].event_data[0].u.objSemRTake.semId;
			break;
		case 3:
			objId = logged_events[i].event_data[0].u.objSemWTake.semId;
			break;
		default:
			CUT_ASSERT_MESSAGE("Invalid test type", 0);
		}

		if (logged_events[i].num_types != 2)
			continue;
		if (objId == 1234) {
			result = 1;
			break;
		}
	}

	CUT_ASSERT_NOT_EQUAL_MESSAGE(msg[test], 0, result);

	return 0;
}

#endif


POST_EVENT_BEGIN(objMsgReceive)
	e->msgQId = 9876;
POST_EVENT_END(EVENT_OBJ_MSGRECEIVE)

static bool objMsgReceive_test_result(struct vxWorks_objMsgReceive *e)
{
	return (e->msgQId == 9876);
}

TEST_EVENT(objMsgReceive)

POST_EVENT_BEGIN(objMsgSend)
	e->msgQId = 5555;
POST_EVENT_END(EVENT_OBJ_MSGSEND)

static bool objMsgSend_test_result(struct vxWorks_objMsgSend *e)
{
	return (e->msgQId == 5555);
}

TEST_EVENT(objMsgSend)

int numCoresOnEnableDisable_test (struct cafe_instrument *session,
				  struct cafe_ep_inst *inst,
				  int enable)
{
	struct cafe_num_cores *e;
	unsigned int i;
	int result = 0;
	struct cafe_event_descr *evt;

	reset_logged_events();

	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, numCores_TYPE[0],
					   &expected_type));

	CAFE_LOCK();

	CUT_ASSERT_CAFE_OK(cafe_cfg_get_event(session, 1, &expected_type, &evt));

	if (enable == 0)
		TEST_ENABLE_SET(1);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_disable_output_event(inst, evt));
	CUT_ASSERT_CAFE_OK(cafe_instrument_refresh_output_event_actions(session));

	if (enable == 0)
		TEST_ENABLE_SET(0);

	CUT_ASSERT_CAFE_OK(cafe_ep_inst_enable_output_event(inst, evt));

	if (enable == 1)
		TEST_ENABLE_SET(1);

	CUT_ASSERT_CAFE_OK(cafe_instrument_refresh_output_event_actions(session));

	if (enable == 1)
		TEST_ENABLE_SET(0);

	CAFE_UNLOCK();

	for (i = 0; i < logged_event_count; i++) {
		if (logged_events[i].num_types != 1)
			continue;
		e = &logged_events[i].event_data[0].u.numCores;
		if (e->num_cores == 2) {
			result = 1;
			break;
		}
	}

	if (enable == 0)
		CUT_ASSERT_EQUAL_MESSAGE("numCores event on event disable", 0, result);
	else
		CUT_ASSERT_NOT_EQUAL_MESSAGE("numCores event on event enable", 0, result);

	return 0;
}

POST_EVENT_BEGIN(windSuspend)
	e->taskId = 1234;
POST_EVENT_END(EVENT_WINDSUSPEND)

static bool windSuspend_test_result(struct vxWorks_windSuspend *e)
{
	return (e->taskId == 1234);
}

TEST_EVENT(windSuspend)

POST_EVENT_BEGIN(taskSuspend)
	e->taskId = 1234;

POST_EVENT_END(EVENT_TASKSUSPEND)

static bool taskSuspend_test_result(struct vxWorks_taskSuspend *e)
{
	return (e->taskId == 1234);
}

TEST_EVENT(taskSuspend);


POST_EVENT_BEGIN(taskPrioritySet)
	e->taskId = 1234;
	e->oldPri = 12;
	e->newPri = 34;
POST_EVENT_END(EVENT_TASKPRIORITYSET)

static bool taskPrioritySet_test_result(struct vxWorks_taskPrioritySet *e)
{
	return (e->taskId == 1234 &&
		e->oldPri == 12 &&
		e->newPri == 34);
}

TEST_EVENT(taskPrioritySet)

POST_EVENT_BEGIN(edrErrorInject)
    e->kind = 1;
    e->filename = (char *)"edrfilename";
    e->linenumber = 3;
    e->address = 4;
    e->message = (char *)"edr message";
POST_EVENT_END(EVENT_EDR_ERROR_INJECT)

static bool edrErrorInject_test_result(struct vxWorks_edrErrorInject *e)
{
	return true;
}

TEST_EVENT(edrErrorInject)


/* --- Test isrInfo event --- */
POST_EVENT_BEGIN(isrInfo)
    e->isrId = 1;
    e->handler = 2;
    e->name = (char *)"isr name";
POST_EVENT_END(EVENT_ISR_INFO)

static bool isrInfo_test_result(struct vxWorks_isrInfo *e)
{
	return true;
}

TEST_EVENT(isrInfo)

/* --- Test rtpName event --- */
POST_EVENT_BEGIN(rtpName)
    e->options = 1;
    e->owner = 2;
    e->rtpId = 3;
    e->name = (char *)"rtpName";
POST_EVENT_END(EVENT_RTPNAME)

static bool rtpName_test_result(struct vxWorks_rtpName *e)
{
	return true;
}

TEST_EVENT(rtpName)

/* --- Test exception event --- */
POST_EVENT_BEGIN(exception)
POST_EVENT_END(EVENT_EXCEPTION)

static bool exception_test_result(struct vxWorks_exception *e)
{
	return true;
}

TEST_EVENT(exception)

/* --- Test windSpawn event --- */
POST_EVENT_BEGIN(windSpawn)
POST_EVENT_END(EVENT_WINDSPAWN)

static bool windSpawn_test_result(struct vxWorks_windSpawn *e)
{
	return true;
}

TEST_EVENT(windSpawn)

/* --- Test windResume event --- */
POST_EVENT_BEGIN(windResume)
POST_EVENT_END(EVENT_WINDRESUME)

static bool windResume_test_result(struct vxWorks_windResume *e)
{
	return true;
}

TEST_EVENT(windResume)

/* --- Test windSemDelete event --- */
POST_EVENT_BEGIN(windSemDelete)
POST_EVENT_END(EVENT_WINDSEMDELETE)

static bool windSemDelete_test_result(struct vxWorks_windSemDelete *e)
{
	return true;
}

TEST_EVENT(windSemDelete)

/* --- Test windTickAnnounceTmrSlc event --- */
POST_EVENT_BEGIN_NO_STRUCT(windTickAnnounceTmrSlc)
POST_EVENT_END_NO_STRUCT(EVENT_WINDTICKANNOUNCETMRSLC)

TEST_EVENT_NO_STRUCT(windTickAnnounceTmrSlc)

/* --- Test windTickAnnounceTmrWd event --- */
POST_EVENT_BEGIN(windTickAnnounceTmrWd)
POST_EVENT_END(EVENT_WINDTICKANNOUNCETMRWD)

static bool windTickAnnounceTmrWd_test_result(struct vxWorks_windTickAnnounceTmrWd *e)
{
	return true;
}

TEST_EVENT(windTickAnnounceTmrWd)

/* --- Test windDelay event --- */
POST_EVENT_BEGIN(windDelay)
POST_EVENT_END(EVENT_WINDDELAY)

static bool windDelay_test_result(struct vxWorks_windDelay *e)
{
	return true;
}

TEST_EVENT(windDelay)

/* --- Test windUndelay event --- */
POST_EVENT_BEGIN(windUndelay)
POST_EVENT_END(EVENT_WINDUNDELAY)

static bool windUndelay_test_result(struct vxWorks_windUndelay *e)
{
	return true;
}

TEST_EVENT(windUndelay)

/* --- Test windWdStart event --- */
POST_EVENT_BEGIN(windWdStart)
POST_EVENT_END(EVENT_WINDWDSTART)

static bool windWdStart_test_result(struct vxWorks_windWdStart *e)
{
	return true;
}

TEST_EVENT(windWdStart)

/* --- Test windWdCancel event --- */
POST_EVENT_BEGIN(windWdCancel)
POST_EVENT_END(EVENT_WINDWDCANCEL)

static bool windWdCancel_test_result(struct vxWorks_windWdCancel *e)
{
	return true;
}

TEST_EVENT(windWdCancel)

/* --- Test windPendQGet event --- */
POST_EVENT_BEGIN(windPendQGet)
POST_EVENT_END(EVENT_WINDPENDQGET)

static bool windPendQGet_test_result(struct vxWorks_windPendQGet *e)
{
	return true;
}

TEST_EVENT(windPendQGet)

/* --- Test windPendQFlush event --- */
POST_EVENT_BEGIN(windPendQFlush)
POST_EVENT_END(EVENT_WINDPENDQFLUSH)

static bool windPendQFlush_test_result(struct vxWorks_windPendQFlush *e)
{
	return true;
}

TEST_EVENT(windPendQFlush)

/* --- Test windPendQTerminate event --- */
POST_EVENT_BEGIN(windPendQTerminate)
POST_EVENT_END(EVENT_WINDPENDQTERMINATE)

static bool windPendQTerminate_test_result(struct vxWorks_windPendQTerminate *e)
{
	return true;
}

TEST_EVENT(windPendQTerminate)

/* --- Test windTickUndelay event --- */
POST_EVENT_BEGIN(windTickUndelay)
POST_EVENT_END(EVENT_WINDTICKUNDELAY)

static bool windTickUndelay_test_result(struct vxWorks_windTickUndelay *e)
{
	return true;
}

TEST_EVENT(windTickUndelay)

/* --- Test objSemGive event --- */
POST_EVENT_BEGIN(objSemGive)
POST_EVENT_END(EVENT_OBJ_SEMGIVE)

static bool objSemGive_test_result(struct vxWorks_objSemGive *e)
{
	return true;
}

TEST_EVENT(objSemGive)

/* --- Test objSemTake event --- */
POST_EVENT_BEGIN(objSemTake)
POST_EVENT_END(EVENT_OBJ_SEMTAKE)

static bool objSemTake_test_result(struct vxWorks_objSemTake *e)
{
	return true;
}

TEST_EVENT(objSemTake)

/* --- Test objSemFlush event --- */
POST_EVENT_BEGIN(objSemFlush)
POST_EVENT_END(EVENT_OBJ_SEMFLUSH)

static bool objSemFlush_test_result(struct vxWorks_objSemFlush *e)
{
	return true;
}

TEST_EVENT(objSemFlush)

/* --- Test objMsgDelete event --- */
POST_EVENT_BEGIN(objMsgDelete)
POST_EVENT_END(EVENT_OBJ_MSGDELETE)

static bool objMsgDelete_test_result(struct vxWorks_objMsgDelete *e)
{
	return true;
}

TEST_EVENT(objMsgDelete)

/* --- Test objSigPause event --- */
POST_EVENT_BEGIN(objSigPause)
POST_EVENT_END(EVENT_OBJ_SIGPAUSE)

static bool objSigPause_test_result(struct vxWorks_objSigPause *e)
{
	return true;
}

TEST_EVENT(objSigPause)

/* --- Test objSigSuspend event --- */
POST_EVENT_BEGIN(objSigSuspend)
POST_EVENT_END(EVENT_OBJ_SIGSUSPEND)

static bool objSigSuspend_test_result(struct vxWorks_objSigSuspend *e)
{
	return true;
}

TEST_EVENT(objSigSuspend)

/* --- Test objSigKill event --- */
POST_EVENT_BEGIN(objSigKill)
POST_EVENT_END(EVENT_OBJ_SIGKILL)

static bool objSigKill_test_result(struct vxWorks_objSigKill *e)
{
	return true;
}

TEST_EVENT(objSigKill)

/* --- Test windStop event --- */
POST_EVENT_BEGIN(windStop)
POST_EVENT_END(EVENT_WINDSTOP)

static bool windStop_test_result(struct vxWorks_windStop *e)
{
	return true;
}

TEST_EVENT(windStop)

/* --- Test windCont event --- */
POST_EVENT_BEGIN(windCont)
POST_EVENT_END(EVENT_WINDCONT)

static bool windCont_test_result(struct vxWorks_windCont *e)
{
	return true;
}

TEST_EVENT(windCont)

/* --- Test windTickTimeout event --- */
POST_EVENT_BEGIN(windTickTimeout)
POST_EVENT_END(EVENT_WINDTICKTIMEOUT)

static bool windTickTimeout_test_result(struct vxWorks_windTickTimeout *e)
{
	return true;
}

TEST_EVENT(windTickTimeout)

/* --- Test objSigWait event --- */
POST_EVENT_BEGIN(objSigWait)
POST_EVENT_END(EVENT_OBJ_SIGWAIT)

static bool objSigWait_test_result(struct vxWorks_objSigWait *e)
{
	return true;
}

TEST_EVENT(objSigWait)

/* --- Test objEventSend event --- */
POST_EVENT_BEGIN_NO_STRUCT(objEventSend)
POST_EVENT_END_NO_STRUCT(EVENT_OBJ_EVENTSEND)

TEST_EVENT_NO_STRUCT(objEventSend)

/* --- Test objEventReceive event --- */
POST_EVENT_BEGIN_NO_STRUCT(objEventReceive)
POST_EVENT_END_NO_STRUCT(EVENT_OBJ_EVENTRECEIVE)

TEST_EVENT_NO_STRUCT(objEventReceive)

/* --- Test objWindRotate event --- */
POST_EVENT_BEGIN(objWindRotate)
POST_EVENT_END(EVENT_OBJ_WINDROTATE)

static bool objWindRotate_test_result(struct vxWorks_objWindRotate *e)
{
	return true;
}

TEST_EVENT(objWindRotate)

/* --- Test objSemRTake event --- */
POST_EVENT_BEGIN(objSemRTake)
POST_EVENT_END(EVENT_OBJ_SEMRTAKE)

static bool objSemRTake_test_result(struct vxWorks_objSemRTake *e)
{
	return true;
}

TEST_EVENT(objSemRTake)

/* --- Test objSemWTake event --- */
POST_EVENT_BEGIN(objSemWTake)
POST_EVENT_END(EVENT_OBJ_SEMWTAKE)

static bool objSemWTake_test_result(struct vxWorks_objSemWTake *e)
{
	return true;
}

TEST_EVENT(objSemWTake)

/* --- Test taskDestroy event --- */
POST_EVENT_BEGIN(taskDestroy)
POST_EVENT_END(EVENT_TASKDESTROY)

static bool taskDestroy_test_result(struct vxWorks_taskDestroy *e)
{
	return true;
}

TEST_EVENT(taskDestroy)

/* --- Test taskResume event --- */
POST_EVENT_BEGIN(taskResume)
POST_EVENT_END(EVENT_TASKRESUME)

static bool taskResume_test_result(struct vxWorks_taskResume *e)
{
	return true;
}

TEST_EVENT(taskResume)

/* --- Test taskSafe event --- */
POST_EVENT_BEGIN(taskSafe)
POST_EVENT_END(EVENT_TASKSAFE)

static bool taskSafe_test_result(struct vxWorks_taskSafe *e)
{
	return true;
}

TEST_EVENT(taskSafe)

/* --- Test taskUnsafe event --- */
POST_EVENT_BEGIN(taskUnsafe)
POST_EVENT_END(EVENT_TASKUNSAFE)

static bool taskUnsafe_test_result(struct vxWorks_taskUnsafe *e)
{
	return true;
}

TEST_EVENT(taskUnsafe)

/* --- Test semMGiveForce event --- */
POST_EVENT_BEGIN(semMGiveForce)
POST_EVENT_END(EVENT_SEMMGIVEFORCE)

static bool semMGiveForce_test_result(struct vxWorks_semMGiveForce *e)
{
	return true;
}

TEST_EVENT(semMGiveForce)

/* --- Test signal event --- */
POST_EVENT_BEGIN(signal)
POST_EVENT_END(EVENT_SIGNAL)

static bool signal_test_result(struct vxWorks_signal *e)
{
	return true;
}

TEST_EVENT(signal)

/* --- Test sigSuspend event --- */
POST_EVENT_BEGIN(sigSuspend)
POST_EVENT_END(EVENT_SIGSUSPEND)

static bool sigSuspend_test_result(struct vxWorks_sigSuspend *e)
{
	return true;
}

TEST_EVENT(sigSuspend)

/* --- Test pause event --- */
POST_EVENT_BEGIN(pause)
POST_EVENT_END(EVENT_PAUSE)

static bool pause_test_result(struct vxWorks_pause *e)
{
	return true;
}

TEST_EVENT(pause)

/* --- Test kill event --- */
POST_EVENT_BEGIN(kill)
POST_EVENT_END(EVENT_KILL)

static bool kill_test_result(struct vxWorks_kill *e)
{
	return true;
}

TEST_EVENT(kill)

/* --- Test sigWrapper event --- */
POST_EVENT_BEGIN(sigWrapper)
POST_EVENT_END(EVENT_SIGWRAPPER)

static bool sigWrapper_test_result(struct vxWorks_sigWrapper *e)
{
	return true;
}

TEST_EVENT(sigWrapper)
/* --- Test memAddToPool event --- */
POST_EVENT_BEGIN(memAddToPool)
POST_EVENT_END(EVENT_MEMADDTOPOOL)

static bool memAddToPool_test_result(struct vxWorks_memAddToPool *e)
{
	return true;
}

TEST_EVENT(memAddToPool)

/* --- Test taskStop event --- */
POST_EVENT_BEGIN(taskStop)
POST_EVENT_END(EVENT_TASKSTOP)

static bool taskStop_test_result(struct vxWorks_taskStop *e)
{
	return true;
}

TEST_EVENT(taskStop)

/* --- Test taskCont event --- */
POST_EVENT_BEGIN(taskCont)
POST_EVENT_END(EVENT_TASKCONT)

static bool taskCont_test_result(struct vxWorks_taskCont *e)
{
	return true;
}

TEST_EVENT(taskCont)

/* --- Test eventSend event --- */
POST_EVENT_BEGIN(eventSend)
POST_EVENT_END(EVENT_EVENTSEND)

static bool eventSend_test_result(struct vxWorks_eventSend *e)
{
	return true;
}

TEST_EVENT(eventSend)

/* --- Test eventReceive event --- */
POST_EVENT_BEGIN(eventReceive)
POST_EVENT_END(EVENT_EVENTRECEIVE)

static bool eventReceive_test_result(struct vxWorks_eventReceive *e)
{
	return true;
}

TEST_EVENT(eventReceive)

/* --- Test ownerSet event --- */
POST_EVENT_BEGIN(objOwnerSet)
POST_EVENT_END(EVENT_OWNERSET)

static bool objOwnerSet_test_result(struct vxWorks_objOwnerSet *e)
{
	return true;
}

TEST_EVENT(objOwnerSet)

/* --- Test isrDispatcherInvoke event --- */
POST_EVENT_BEGIN(isrDispatcherInvoke)
POST_EVENT_END(EVENT_ISR_DISPATCHER_INVOKE)

static bool isrDispatcherInvoke_test_result(struct vxWorks_isrDispatcherInvoke *e)
{
	return true;
}

TEST_EVENT(isrDispatcherInvoke)

/* --- Test isrDispatcherDone event --- */
POST_EVENT_BEGIN(isrDispatcherDone)
POST_EVENT_END(EVENT_ISR_DISPATCHER_DONE)

static bool isrDispatcherDone_test_result(struct vxWorks_isrDispatcherDone *e)
{
	return true;
}

TEST_EVENT(isrDispatcherDone)

/* --- Test isrCreate event --- */
POST_EVENT_BEGIN(isrCreate)
POST_EVENT_END(EVENT_ISR_CREATE)

static bool isrCreate_test_result(struct vxWorks_isrCreate *e)
{
	return true;
}

TEST_EVENT(isrCreate)

/* --- Test isrDestroy event --- */
POST_EVENT_BEGIN(isrDestroy)
POST_EVENT_END(EVENT_ISR_DESTROY)

static bool isrDestroy_test_result(struct vxWorks_isrDestroy *e)
{
	return true;
}

TEST_EVENT(isrDestroy)

/* --- Test syscallReturnOk event --- */
POST_EVENT_BEGIN(syscallReturnOk)
POST_EVENT_END(EVENT_SYSCALL_RETURN_OK)

static bool syscallReturnOk_test_result(struct vxWorks_syscallReturnOk *e)
{
	return true;
}

TEST_EVENT(syscallReturnOk)

/* --- Test syscallReturnError event --- */
POST_EVENT_BEGIN(syscallReturnError)
POST_EVENT_END(EVENT_SYSCALL_RETURN_ERR)

static bool syscallReturnError_test_result(struct vxWorks_syscallReturnError *e)
{
	return true;
}

TEST_EVENT(syscallReturnError)

/* --- Test syscallBadId event --- */
POST_EVENT_BEGIN(syscallBadId)
POST_EVENT_END(EVENT_SYSCALL_BAD_ID)

static bool syscallBadId_test_result(struct vxWorks_syscallBadId *e)
{
	return true;
}

TEST_EVENT(syscallBadId)

/* --- Test rtpSpawn event --- */
POST_EVENT_BEGIN(rtpSpawn)
POST_EVENT_END(EVENT_RTP_SPAWN)

static bool rtpSpawn_test_result(struct vxWorks_rtpSpawn *e)
{
	return true;
}

TEST_EVENT(rtpSpawn)

/* --- Test rtpDelete event --- */
POST_EVENT_BEGIN(rtpDelete)
POST_EVENT_END(EVENT_RTP_DELETE)

static bool rtpDelete_test_result(struct vxWorks_rtpDelete *e)
{
	return true;
}

TEST_EVENT(rtpDelete)

/* --- Test rtpInitTask event --- */
POST_EVENT_BEGIN(rtpInitTask)
POST_EVENT_END(EVENT_RTP_INIT_TASK)

static bool rtpInitTask_test_result(struct vxWorks_rtpInitTask *e)
{
	return true;
}

TEST_EVENT(rtpInitTask)

/* --- Test rtpLoad event --- */
POST_EVENT_BEGIN(rtpLoad)
    e->name = (char *)"rtpName";
POST_EVENT_END(EVENT_RTP_LOAD)

static bool rtpLoad_test_result(struct vxWorks_rtpLoad *e)
{
	return true;
}

TEST_EVENT(rtpLoad)

/* --- Test rtpKill event --- */
POST_EVENT_BEGIN(rtpKill)
POST_EVENT_END(EVENT_RTPKILL)

static bool rtpKill_test_result(struct vxWorks_rtpKill *e)
{
	return true;
}

TEST_EVENT(rtpKill)

/* --- Test rtpSignal event --- */
POST_EVENT_BEGIN(rtpSignal)
POST_EVENT_END(EVENT_RTPSIGNAL)

static bool rtpSignal_test_result(struct vxWorks_rtpSignal *e)
{
	return true;
}

TEST_EVENT(rtpSignal)

/* --- Test rtpSuspend event --- */
POST_EVENT_BEGIN(rtpSuspend)
POST_EVENT_END(EVENT_RTPSUSPEND)

static bool rtpSuspend_test_result(struct vxWorks_rtpSuspend *e)
{
	return true;
}

TEST_EVENT(rtpSuspend)

/* --- Test rtpPause event --- */
POST_EVENT_BEGIN(rtpPause)
POST_EVENT_END(EVENT_RTPPAUSE)

static bool rtpPause_test_result(struct vxWorks_rtpPause *e)
{
	return true;
}

TEST_EVENT(rtpPause)

/* --- Test rtpTaskKill event --- */
POST_EVENT_BEGIN(rtpTaskKill)
POST_EVENT_END(EVENT_RTPTASKKILL)

static bool rtpTaskKill_test_result(struct vxWorks_rtpTaskKill *e)
{
	return true;
}

TEST_EVENT(rtpTaskKill)

/* --- Test rtpSigQueue event --- */
POST_EVENT_BEGIN(rtpSigQueue)
POST_EVENT_END(EVENT_RTPSIGQUEUE)

static bool rtpSigQueue_test_result(struct vxWorks_rtpSigQueue *e)
{
	return true;
}

TEST_EVENT(rtpSigQueue)

/* --- Test rtpTaskSigQueue event --- */
POST_EVENT_BEGIN(rtpTaskSigQueue)
POST_EVENT_END(EVENT_RTPTASKSIGQUEUE)

static bool rtpTaskSigQueue_test_result(struct vxWorks_rtpTaskSigQueue *e)
{
	return true;
}

TEST_EVENT(rtpTaskSigQueue)

/* --- Test rtpChildWait event --- */
POST_EVENT_BEGIN(rtpChildWait)
POST_EVENT_END(EVENT_RTPCHILDWAIT)

static bool rtpChildWait_test_result(struct vxWorks_rtpChildWait *e)
{
	return true;
}

TEST_EVENT(rtpChildWait)

/* --- Test rtpSigTimedWait event --- */
POST_EVENT_BEGIN(rtpSigTimedWait)
POST_EVENT_END(EVENT_RTPSIGTIMEDWAIT)

static bool rtpSigTimedWait_test_result(struct vxWorks_rtpSigTimedWait *e)
{
	return true;
}

TEST_EVENT(rtpSigTimedWait)

/* --- Test rtpSigWrapper event --- */
POST_EVENT_BEGIN(rtpSigWrapper)
POST_EVENT_END(EVENT_RTPSIGWRAPPER)

static bool rtpSigWrapper_test_result(struct vxWorks_rtpSigWrapper *e)
{
	return true;
}

TEST_EVENT(rtpSigWrapper)

/* --- Test rtpSigWrapperDone event --- */
POST_EVENT_BEGIN(rtpSigWrapperDone)
POST_EVENT_END(EVENT_RTPSIGWRAPPER_DONE)

static bool rtpSigWrapperDone_test_result(struct vxWorks_rtpSigWrapperDone *e)
{
	return true;
}

TEST_EVENT(rtpSigWrapperDone)

/* --- Test rtpSigProcMask event --- */
POST_EVENT_BEGIN(rtpSigProcMask)
POST_EVENT_END(EVENT_RTPSIGPROCMASK)

static bool rtpSigProcMask_test_result(struct vxWorks_rtpSigProcMask *e)
{
	return true;
}

TEST_EVENT(rtpSigProcMask)

/* --- Test rtpSigPending event --- */
POST_EVENT_BEGIN(rtpSigPending)
POST_EVENT_END(EVENT_RTPSIGPENDING)

static bool rtpSigPending_test_result(struct vxWorks_rtpSigPending *e)
{
	return true;
}

TEST_EVENT(rtpSigPending)

/* --- Test taskRotate event --- */
POST_EVENT_BEGIN(taskRotate)
POST_EVENT_END(EVENT_TASKROTATE)

static bool taskRotate_test_result(struct vxWorks_taskRotate *e)
{
	return true;
}

TEST_EVENT(taskRotate)

/* --- Test sdCreate event --- */
POST_EVENT_BEGIN(sdCreate)
    e->objId = 1;
    e->classId = 2;
    e->nameLength = 3;
    e->name = (char *)"sd name";
    e->sdId = 5;
    e->options = 6;
    e->size = 7;
    e->attr = 8;
    e->virtAddr = 9;
POST_EVENT_END(EVENT_SD_CREATE)

static bool sdCreate_test_result(struct vxWorks_sdCreate *e)
{
	return true;
}

TEST_EVENT(sdCreate)

/* --- Test sdDelete event --- */
POST_EVENT_BEGIN(sdDelete)
POST_EVENT_END(EVENT_SD_DELETE)

static bool sdDelete_test_result(struct vxWorks_sdDelete *e)
{
	return true;
}

TEST_EVENT(sdDelete)

/* --- Test sdMap event --- */
POST_EVENT_BEGIN(sdMap)
POST_EVENT_END(EVENT_SD_MAP)

static bool sdMap_test_result(struct vxWorks_sdMap *e)
{
	return true;
}

TEST_EVENT(sdMap)

/* --- Test sdUnmap event --- */
POST_EVENT_BEGIN(sdUnmap)
POST_EVENT_END(EVENT_SD_UNMAP)

static bool sdUnmap_test_result(struct vxWorks_sdUnmap *e)
{
	return true;
}

TEST_EVENT(sdUnmap)

/* --- Test semRWCreate event --- */
POST_EVENT_BEGIN(semRWCreate)
POST_EVENT_END(EVENT_SEMRWCREATE)

static bool semRWCreate_test_result(struct vxWorks_semRWCreate *e)
{
	return true;
}

TEST_EVENT(semRWCreate)

/* --- Test semRTake event --- */
POST_EVENT_BEGIN(semRTake)
POST_EVENT_END(EVENT_SEMRTAKE)

static bool semRTake_test_result(struct vxWorks_semRTake *e)
{
	return true;
}

TEST_EVENT(semRTake)

/* --- Test semWTake event --- */
POST_EVENT_BEGIN(semWTake)
POST_EVENT_END(EVENT_SEMWTAKE)

static bool semWTake_test_result(struct vxWorks_semWTake *e)
{
	return true;
}

TEST_EVENT(semWTake)

/* --- Test taskCpuAffinitySet event --- */
POST_EVENT_BEGIN(taskCpuAffinitySet)
POST_EVENT_END(EVENT_TASK_CPU_AFFINITY_SET)

static bool taskCpuAffinitySet_test_result(struct vxWorks_taskCpuAffinitySet *e)
{
	return true;
}

TEST_EVENT(taskCpuAffinitySet)

/* --- Test dprintf event --- */
POST_EVENT_BEGIN(dprintf)
    e->message = (char *)"message";
    e->length = strlen(e->message);
POST_EVENT_END(EVENT_DPRINTF_LOG)

static bool dprintf_test_result(struct vxWorks_dprintf *e)
{
	return true;
}

TEST_EVENT(dprintf)

/* --- Test timerRollover event --- */
POST_EVENT_BEGIN_NO_STRUCT(timerRollover)
POST_EVENT_END_NO_STRUCT(EVENT_TIMER_ROLLOVER)

TEST_EVENT_NO_STRUCT(timerRollover)

/* --- Test moduleLoad event --- */
POST_EVENT_BEGIN(moduleLoad)
	e->name = (char *)"module";
	e->moduleId = 1234;
	e->textAddr = 1;
	e->textSize = 2;
	e->dataAddr = 3;
	e->dataSize = 4;
	e->bssAddr = 5;
	e->bssSize = 6;
	e->rtpId = 7;
POST_EVENT_END(EVENT_CAFE_MODULE_LOAD)

static bool moduleLoad_test_result(struct vxWorks_moduleLoad *e)
{
	return strcmp(e->name, "module") == 0 &&
		e->moduleId == 1234 &&
		e->textAddr == 1 &&
		e->textSize == 2 &&
		e->dataAddr == 3 &&
		e->dataSize == 4 &&
		e->bssAddr == 5 &&
		e->bssSize == 6 &&
		e->rtpId == 7;
}

TEST_EVENT(moduleLoad)

/* --- Test moduleUnload event --- */
POST_EVENT_BEGIN(moduleUnload)
	e->moduleId = 123;
POST_EVENT_END(EVENT_CAFE_MODULE_UNLOAD)

static bool moduleUnload_test_result(struct vxWorks_moduleUnload *e)
{
	return e->moduleId == 123;
}

TEST_EVENT(moduleUnload)

/* --- Test rtpUnload event --- */
POST_EVENT_BEGIN(rtpUnload)
	e->rtpId = 123;
POST_EVENT_END(EVENT_CAFE_RTP_UNLOAD)

static bool rtpUnload_test_result(struct vxWorks_rtpUnload *e)
{
	return e->rtpId == 123;
}

TEST_EVENT(rtpUnload)

/* test kernel RTP at start */

int kernelOnStart_test (struct cafe_instrument *session,
			struct cafe_ep_inst *vxworks)
{
	unsigned int i;
	int result = 0;
	struct vxWorks_rtpLoad *e;

	reset_logged_events();
	CUT_ASSERT_CAFE_OK(cafe_type_parse(session->ns, rtpLoad_TYPE[0],
					   &expected_type));

	CAFE_LOCK();
	cafe_ep_inst_stop(vxworks);

	TEST_ENABLE_SET(1);
	cafe_ep_inst_start(vxworks);
	TEST_ENABLE_SET(0);
	CAFE_UNLOCK();

	for (i = 0; i < logged_event_count; i++) {
		if (logged_events[i].types[0] != expected_type)
			continue;

		e = &logged_events[i].event_data[0].u.rtpLoad;
		if ((e->dataSize == 1) &&
		    (e->dataAddr == 2) &&
		    (e->textSize == 3) &&
		    (e->textAddr == 4) &&
		    (e->entryPoint == 5) &&
		    (e->rtpId == 6) &&
		    (strcmp(e->name, "vxWorks") == 0)) {
			result = 1;
			break;
		}

		result = 1;
	}
	CUT_ASSERT_NOT_EQUAL_MESSAGE("kernelId at start", 0, result);
	return 0;
}
