/* timer.c - timer event processor */

/* Copyright (c) 2010-2011, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

/*
 * This EP acts as a configurable timer.  The timer is enabled by
 * changing the "interval" setting.  Units are in milliseconds and
 * zero will disable the timer.  When any timer expires, a
 * "cafe.timer" event is sent to participating event processors and
 * the timer restarts.  This processor also supports timestamp as
 * orthogonal data.
 */

#include <string.h>
#include <cafe/cafe.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/value_print.h>
#include <cafe/ep/timer_ep.h>
#include "../src/cafe_assert.h"

enum {
	ACTION_TIMER,
	ACTION_INTERVAL_UPDATED,
	ACTION_DEBUG_UPDATED,
	NUM_ACTIONS
};

static void timestamp_freq_act_handler(struct cafe_event_action_info *, void **);
static void timestamp_freq_with_ts_act_handler(struct cafe_event_action_info *, void **);
static void wildcard_with_ts_act_handler(struct cafe_event_action_info *, void **);
static void timer_set_interval_handler(struct cafe_event_action_info *, void **);
static void timer_set_debug_handler(struct cafe_event_action_info *, void **);

struct timestamp_event {
	unsigned long long tick;
};

struct timestamp_freq_event {
	unsigned long long frequency;
	unsigned long long period;
};

struct timer_inst {
	long long elapsed_ticks;
	unsigned long long clock_frequency;
	cafe_eal_handle_t sink_actions[NUM_ACTIONS];
	long change_count;
	struct timestamp_event last_timestamp;
	int first_time;
	unsigned int overall_dispatch_count;
};

struct workarea {
	/* interval before firing event; 0 = disabled. */
	long long interval;
	unsigned int debug;
};

static const char *wildcard2_TYPE[] = {
	cafe_wildcard_VT_REF,
	cafe_timestamp_VT_REF
};

static const char *timestamp_freq_TYPE[] = {
	cafe_timestamp_config_VT_REF
};

static const char *timestamp_freq_with_ts_TYPE[] = {
	cafe_timestamp_config_VT_REF,
	cafe_timestamp_VT_REF
};

static const char *timer_set_interval_TYPE[] = {
	TIMER_SET_INTERVAL_REF
};

static const char *timer_set_debug_TYPE[] = {
	TIMER_SET_DEBUG_REF
};

static struct cafe_ep_spec_input_evt input_events[] = {
	CAFE_INPEV_INIT2(timestamp_freq, timestamp_freq_act_handler),
	CAFE_INPEV_INIT2(timestamp_freq_with_ts, timestamp_freq_with_ts_act_handler),
	CAFE_INPEV_INIT2(timer_set_interval, timer_set_interval_handler),
	CAFE_INPEV_INIT2(timer_set_debug, timer_set_debug_handler),
	CAFE_INPEV_INIT2(wildcard2, wildcard_with_ts_act_handler),
};

static struct cafe_ep_spec_output_evt output_events[] = {
	CAFE_OUTEV_INIT2(TIMER_REF, (void *)ACTION_TIMER),
	CAFE_OUTEV_INIT2(TIMER_INTERVAL_CHANGED_REF, (void *)ACTION_INTERVAL_UPDATED),
	CAFE_OUTEV_INIT2(TIMER_DEBUG_CHANGED_REF, (void *)ACTION_DEBUG_UPDATED)
};

static const char *orthogonal_data[] = {
	cafe_timestamp_VT_REF
};

static long long time2tick(unsigned long long milliseconds,
			   unsigned long long frequency)
{
	return milliseconds * frequency / 1000;
}

static void timer_set_interval_handler(struct cafe_event_action_info *info,
				       void **data)
{
	struct workarea *w;
	void *output_event_data[2];
	struct timer_inst *i = (struct timer_inst *)info->inst->client_data;

	w = (struct workarea *)info->inst->workarea->u.data;
	if (w->interval < 0)
		w->interval = 0;
	output_event_data[0] = &w->interval;
	output_event_data[1] = &i->last_timestamp.tick;
	cafe_dispatch_event(&i->sink_actions[ACTION_INTERVAL_UPDATED],
			    output_event_data);
}

static void timer_set_debug_handler(struct cafe_event_action_info *info,
				    void **data)
{
	struct workarea *w;
	void *output_event_data[2];
	struct timer_inst *i = (struct timer_inst *)info->inst->client_data;

	w = (struct workarea *)info->inst->workarea->u.data;
	output_event_data[0] = &w->debug;
	output_event_data[1] = &i->last_timestamp.tick;
	cafe_dispatch_event(&i->sink_actions[ACTION_DEBUG_UPDATED],
			    output_event_data);
}

static void timestamp_freq_with_ts_act_handler(struct cafe_event_action_info *info,
					       void **data)
{
	struct timer_inst *i;
	const struct timestamp_freq_event *e;
	const struct timestamp_event *t;

	i = (struct timer_inst *)info->inst->client_data;
	e = (struct timestamp_freq_event *)data[0];
	t = (struct timestamp_event *)data[1];
	i->clock_frequency = e->frequency;
	if (i->first_time) {
		i->first_time = 0;
		i->last_timestamp.tick = t->tick;
	}
}

static void timestamp_freq_act_handler(struct cafe_event_action_info *info,
				       void **data)
{
	struct timer_inst *i;
	const struct timestamp_freq_event *e;

	i = (struct timer_inst *)info->inst->client_data;
	e = (struct timestamp_freq_event *)data[0];
	i->clock_frequency = e->frequency;
}

static void wildcard_with_ts_act_handler(struct cafe_event_action_info *info,
					 void **data)
{
	struct workarea *w;
	struct timer_inst *i;
	struct timestamp_event *timestamp_event;
	unsigned long long delta;
	long long reset_ticks;
	unsigned int dispatch_count = 0;

	i = (struct timer_inst *)info->inst->client_data;
	w = (struct workarea *)info->inst->workarea->u.data;

	assert(info->event->num_types > 1);
	assert(w->interval >= 0);

	if (w->interval == 0)
		return;		/* no timer set! */

	if (i->clock_frequency == 0)
		return;

	timestamp_event = data[info->odmap[1]];

	/* Ignore incoming events where time goes backwards. */

	if (timestamp_event->tick < i->last_timestamp.tick)
		return;

	if (i->first_time) {
		i->last_timestamp.tick = timestamp_event->tick;
		i->first_time = 0;
		return;
	}

	delta = timestamp_event->tick - i->last_timestamp.tick;
	i->last_timestamp.tick = timestamp_event->tick;
	i->elapsed_ticks += delta;
	reset_ticks = time2tick(w->interval, i->clock_frequency);

	if (i->elapsed_ticks >= reset_ticks)
		i->overall_dispatch_count++;

	while (i->elapsed_ticks >= reset_ticks) {
		struct timestamp_event now;
		void *timer_data[2];
		timer_data[0] = NULL;
		timer_data[1] = &now;
		now.tick = timestamp_event->tick - i->elapsed_ticks + reset_ticks;
		if (w->debug) {
			(void)printf("%s[%p] [%u/%u] now=%llu "
			       "interval=%llu tick_period=%llu "
			       "elapsed_ticks=%lld\n",
			       info->inst->spec->name,
			       info->inst,
			       i->overall_dispatch_count,
			       ++dispatch_count,
			       now.tick,
			       w->interval,
			       time2tick(w->interval, i->clock_frequency),
			       i->elapsed_ticks);
		}
		cafe_dispatch_event(&i->sink_actions[ACTION_TIMER], timer_data);
		i->elapsed_ticks -= reset_ticks;
	}
}

static int timer_ep_create(struct cafe_ep_inst *inst)
{
	struct timer_inst *i;
	struct workarea *w;
	int idx;

	if ((i = CAFE_ALLOC(inst->allocator, sizeof *i)) == NULL)
		return CAFE_NOMEM;

	memset(i, 0, sizeof *i);

	for (idx = 0; idx < CAFE_NELEMENTS(i->sink_actions); idx++)
		i->sink_actions[idx] = CAFE_EAL_HANDLE_NULL;

	inst->client_data = i;
	w = (struct workarea *)inst->workarea->u.data;
	if (w->interval < 0)
		w->interval = 0;
	i->first_time = 1;
	return CAFE_OK;
}

static void timer_ep_delete(struct cafe_ep_inst *inst)
{
	struct timer_inst *i = (struct timer_inst *)inst->client_data;
	int idx;

	for (idx = 0; idx < CAFE_NELEMENTS(i->sink_actions); idx++)
		cafe_event_action_list_release(&i->sink_actions[idx]);

	CAFE_FREE(inst->allocator, inst->client_data);
}

static int get_input_event_actions(struct cafe_ep_inst *inst,
				   struct cafe_event_descr *event,
				   struct cafe_event_action_list *actions)
{
	int rc;
	struct cafe_event_action_info *event_info;
	struct cafe_event_action_list *sink_actions;
	cafe_eal_handle_t sink_actions_handle;

	rc = cafe_ep_inst_event_action_info_get(inst, event, &event_info);

	if (rc != CAFE_OK)
		return rc;

	if (event_info == NULL)
		return CAFE_OK;

	rc = cafe_event_action_list_alloc(inst, event,
					  &sink_actions, &sink_actions_handle);

	if (rc != CAFE_OK)
		return rc;

	/* Ask the sink if it is interested in this event type. */

	rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);

	if (rc != CAFE_OK) {
		cafe_event_action_list_release(&sink_actions_handle);
		return rc;
	}

	cafe_event_action_list_swap(&event_info->sink_actions,
				    sink_actions_handle);

	return cafe_add_event_action(inst, actions,
				     (CAFE_ACTION_HANDLER)event_info->impl_data,
				     event_info);
}

static int update_output_event_actions(struct cafe_ep_inst *inst)
{
	int rc;
	unsigned int idx;
	struct timer_inst *i = (struct timer_inst *)inst->client_data;

	if (i->change_count == inst->session->output_event_action_change_count)
		return CAFE_OK;

	i->change_count = inst->session->output_event_action_change_count;

	for (idx = 0; idx < inst->enabled_output_events.ind; idx++) {
		struct cafe_event_descr *event;
		struct cafe_event_action_list *sink_actions;
		cafe_eal_handle_t sink_actions_handle = CAFE_EAL_HANDLE_NULL;
		long id = (long)inst->enabled_output_events.list[idx].impl_data;

		event = inst->enabled_output_events.list[idx].event;
		assert(event);

		rc = cafe_event_action_list_alloc(inst, event,
						  &sink_actions,
						  &sink_actions_handle);

		if (rc != CAFE_OK)
			return rc;

		rc = cafe_ep_inst_get_sink_event_actions(inst, sink_actions);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sink_actions_handle);
			return rc;
		}

		cafe_event_action_list_swap(&i->sink_actions[id],
					    sink_actions_handle);
	}

	return CAFE_OK;
}

#define EP_NAME "cafe_timer"

#define WORKAREA_DEF "S[" EP_NAME ".workarea,v[interval,q," cafe_setting_VT_NAT "]v[debug,I," cafe_setting_VT_NAT "]]"
#define WORKAREA_REF "t[" EP_NAME ".workarea]"

#define ALL_TYPES				\
	cafe_timestamp_config_VT_DEF		\
	cafe_timestamp_VT_DEF			\
	cafe_wildcard_VT_DEF			\
	cafe_setting_VT_DEF			\
	TIMER_DEF				\
	TIMER_SET_INTERVAL_DEF			\
	TIMER_SET_DEBUG_DEF			\
	TIMER_INTERVAL_CHANGED_DEF		\
	TIMER_DEBUG_CHANGED_DEF			\
	WORKAREA_DEF

static struct cafe_ep_spec spec = {
	EP_NAME,
	timer_ep_create,		/* create_hook */
	timer_ep_delete,		/* delete_hook */
	NULL,				/* start */
	NULL,				/* stop */
	get_input_event_actions,
	update_output_event_actions,
	NULL,				/* inject event */
	ALL_TYPES,
	WORKAREA_REF,
	CAFE_NELEMENTS(input_events),
	input_events,
	CAFE_NELEMENTS(output_events),
	output_events,
	CAFE_NELEMENTS(orthogonal_data),
	orthogonal_data
};

static bool cfg_registered = false;

void cafe_timer_ep_unregister(void)
{
	if (cfg_registered) {
		(void)cafe_ep_spec_unregister(&spec);
		cfg_registered = false;
	}
}

int cafe_timer_ep_register(void)
{
	int rc = CAFE_OK;

	if (!cfg_registered) {
		rc = cafe_ep_spec_register(&spec);
		if (rc == CAFE_OK)
			cfg_registered = true;
	}

	return rc;
}
