/* cafe_CUnitTest.h - Cafe-specific CUT macros */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
08jul14,sdt  Fix CUnitest conflicts with other components - WB4-3658
08jul14,sdt  Add CUT_PRINTF/CUT_PRINT_ERR macros to make clear output
*/

#ifndef __CAFE_CUNITTEST_H__
#define __CAFE_CUNITTEST_H__

/*
 * Overwrite CUT_xxx global variables to be able build code when other
 * components / VxWorks layers also define CUnittests.
 */


#define CUT_tests_run			CUT_cafe_tests_run
#define CUT_tests_failed		CUT_cafe_tests_failed
#define CUT_suite_tests_run		CUT_cafe_suite_tests_run
#define CUT_suite_tests_failed		CUT_cafe_suite_tests_failed
#define CUT_assertions_count		CUT_cafe_assertions_count
#define CUT_curr_suite			CUT_cafe_curr_suite
#ifdef __VXWORKS__
    #define CUT_fixture_setup		CUT_cafe_fixture_setup
    #define CUT_fixture_teardown	CUT_cafe_fixture_teardown
#endif

/* now include standard CUniTest header file */

#include <test/CUnitTest.h>
#include <cafe/cafe.h>  /* for CAFE_OK and cafe_strerror() */
#include <cafe/errors.h>

/*!
 @define     CUT_PRINTF
 @abstract   printf macro.
 @param      formater string.
 @param      args...
 @discussion Take care of flushing stdout before printing message on stdout.
 */

#if defined(_MSC_VER)
    #define CUT_PRINTF(fmt, ...) \
    do { \
	fflush(stdout); \
	fprintf(stdout, fmt, __VA_ARGS__); \
    } while (0)
#else
    #define CUT_PRINTF(fmt, args...) \
    do { \
	fflush(stdout); \
	fprintf(stdout, fmt, ##args); \
    } while (0)
#endif /* _MSC_VER */

/*!
 @define     CUT_PRINTF_ERR
 @abstract   printf macro.
 @param      formater string.
 @param      args...
 @discussion Take care of flushing stdout AND stderr before printing message
	     on stderr.
 */

#if defined(_MSC_VER)
    #define CUT_PRINTF_ERR(fmt, ...) \
    do { \
	fflush(stdout); \
	fflush(stderr); \
	fprintf(stderr, fmt, __VA_ARGS__); \
    } while (0)
#else
    #define CUT_PRINTF_ERR(fmt, args...) \
    do { \
	fflush(stdout); \
	fflush(stderr); \
	fprintf(stderr, fmt, ##args); \
    } while (0)
#endif /* _MSC_VER */

/*!
    @define     CUT_ASSERT
    @abstract   Basic assertion macro.
    @param      condition An expression that should evaluate to anything
                non-false on success and false on failure.
    @discussion The basic CUnitTest assertion macro prints the file, line
                number, and reason for failure if the passed expression is
                false.
*/
#define CUT_ASSERT_CAFE_OK(condition)					\
	do {								\
		int CUT_rc;						\
		++CUT_assertions_count;					\
		CUT_rc = (condition);					\
		if (cafe_islocked() && (cafe_error_cur.rc != CAFE_OK)) { \
			fprintf(stderr,					\
				"%s:%d: error: Test %s%s%s failed : ",	\
				__FILE__,				\
				__LINE__,				\
				CUT_curr_suite,				\
				(CUT_curr_suite[0]=='\0' ? "":":"),	\
				__func__);				\
			cafe_error_print(NULL, stderr, ": ");		\
			fputc('\n', stderr);				\
			cafe_error_clear(NULL);				\
			return 1;					\
		} else if (CUT_rc != CAFE_OK) {				\
			fprintf(stderr,					\
				"%s:%d: error: Test %s%s%s failed : "	\
				"%s (%s)\n",				\
				__FILE__,				\
				__LINE__,				\
				CUT_curr_suite,				\
				(CUT_curr_suite[0]=='\0' ? "":":"),	\
				__func__,				\
				#condition,				\
				cafe_strerror(CUT_rc));			\
			return 1;					\
		}							\
	} while (0)

/*!
 @define     CUT_RUN_TEST_F
 @abstract   Test runner with fixtures.
 @param      test The name of the test function to run. The function is assumed
             to have the following signature: <code>int test_name(void)</code>.
             The function should return 0.
 @param      setup Name of the fixture setup function.
 @param      teardown Name of the fixture teardown function.
 */
/**/

#define CUT_RUN_TEST_F(test, _setup, _teardown)		\
do { 							\
    void (*CUT_test_setup)(void) = _setup; 		\
    void (*CUT_test_teardown)(void) = _teardown; 	\
    if (CUT_test_setup) CUT_test_setup();		\
    CUT_RUN_TEST(test);					\
    if (CUT_test_teardown) CUT_test_teardown();		\
} while (0)


#if defined(__VXWORKS__)
    extern void cafe_cunit_test_init(void);
    extern char * cafe_cunit_logRootDirGet(void);

    #define CAFE_BEGIN_SUITE_HARNESS  CUT_BEGIN_SUITE
    #define CAFE_END_SUITE_HARNESS    CUT_END_SUITE
#else
    #define CAFE_BEGIN_SUITE_HARNESS  CUT_BEGIN_TEST_HARNESS
    #define CAFE_END_SUITE_HARNESS    CUT_END_TEST_HARNESS
#endif

#endif /* __CAFE_CUNITTEST_H__ */
