/* Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <cafe/stdio.h>
#include <cafe/cafe.h>
#include <cafe/ptr_array.h>
#include "dirname.h"
#include "bailout_allocator.h"
#include "compiler.h"
#include "strbuf.h"

#define QUOTEME_(x) #x
#define QUOTEME(x) QUOTEME_(x)

/* WRS_TODO: consider walking the tree as opposed to directly iterating
 * over it when printing. */

static const char *const progname = "cafe-type2h";
static int compile_only = 0;
static const char *cli_prefix = "";

#ifdef __GNUC__
static int xfprintf(FILE *stream, const char *fmt, ...) __attribute__((format(printf, 2, 3)));
static void die(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
#endif

static int xfprintf(FILE *stream, const char *fmt, ...)
{
	int rc;
	va_list ap;

	va_start(ap, fmt);
	rc = vfprintf(stream, fmt, ap);
	va_end(ap);

	if (rc < 0) {
		fprintf(stderr, "write failed: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	return rc;
}

static void die(const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s: error: ", progname);
	fflush(stderr);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
	exit(EXIT_FAILURE);
}

#define fprintf USE_xfprintf_instead

static char *get_namespace(const char *filename)
{
	char *tmp_str;
	char *dname;
	struct cafe_byte_buffer sb = STRBUF_INIT(bailout_allocator);

	/* WRS_XXX or use namespace from parser. */

	tmp_str = CAFE_STRDUP(bailout_allocator, filename);
	dname = xdirname(tmp_str);

	if (dname == NULL)
		die("dirname failed for '%s'", filename);

	if (*dname == '.' || *dname == '\0')
		strbuf_fmt(&sb, "%s", filename);
	else
		strbuf_fmt(&sb, "%s", &filename[1 + strlen(dname)]);

	CAFE_FREE(bailout_allocator, tmp_str);

	return strbuf_steal(&sb);
}

static char *generate_file_ident(const char *prefix, const char *str)
{
	struct cafe_byte_buffer sb = STRBUF_INIT(bailout_allocator);
	strbuf_fmt(&sb, "%s%s", prefix ? prefix : "", str);
	return strbuf_steal(&sb);
}

static void print_quoted_string(FILE *stream, const char *s)
{
	const char *q = strchr(s, '"');

	if (q == NULL) {
		fputs(s, stream);
	} else {
		while (*s != '\0') {
			switch (*s) {
			case '"':
				putc('\\', stream);
			default:
				putc(*s, stream);
				break;
			}
			s++;
		}
	}
}

static void print_id(FILE *stream, const char *s, size_t len, bool to_upper)
{
	size_t i;

	for (i = 0; i < len; i++) {
		switch (s[i]) {
		case '.':	/* oh! */
			putc('_', stream);
			break;
		case '@':	/* la! */
		case ':':	/* la! */
			putc('_', stream);
			break;
		default:
			if (to_upper)
				putc(toupper(s[i]), stream);
			else
				putc(s[i], stream);
			break;
		}
	}

}

static void print_str_escaped_toupper(FILE *stream, const char *id)
{
	size_t i;
	size_t len = strlen(id);

	for (i = 0; i < len; i++) {
		if (isalpha(id[i]) || isalnum(id[i]))
			putc(toupper(id[i]), stream);
		else
			putc('_', stream);
	}
}

static void print_node_str_with_prefix(FILE *stream, struct pt_node *n, bool to_upper)
{
	if (cli_prefix != NULL)
		xfprintf(stream, "%s", cli_prefix);

	print_id(stream, n->left->istr->str, n->left->istr->len, to_upper);
}

static void print_vt_def_macro(FILE *stream, struct pt_node *n)
{
	struct cafe_byte_buffer sb = STRBUF_INIT(bailout_allocator);

	compiler_value_type_str(n, &sb);
	xfprintf(stream, "#define ");
	print_node_str_with_prefix(stream, n, false);
	xfprintf(stream, "_VT_DEF \"");
	print_quoted_string(stream, sb.data);
	xfprintf(stream, "\"\n");
	strbuf_release(&sb);
}

static void print_vt_ref_macro(FILE *stream, struct pt_node *n)
{
	xfprintf(stream, "#define ");
	print_node_str_with_prefix(stream, n, false);
	xfprintf(stream, "_VT_REF \"t[%s]\"\n", n->left->istr->str);
}

static void print_nature_macro(FILE *stream, struct pt_node *n)
{
	xfprintf(stream, "#define ");
	print_node_str_with_prefix(stream, n, false);
	assert(n->type == PT_NATURE);
	xfprintf(stream, "_VT_NAT \"n[%s]\"\n", n->left->istr->str);
}

static void print_header_guard_begin(FILE *stream, const char *id)
{
	xfprintf(stream, "#ifndef ");
	print_str_escaped_toupper(stream, progname);
	xfprintf(stream, "_");
	print_str_escaped_toupper(stream, id);
	xfprintf(stream, "_autogenerated_h\n");
	xfprintf(stream, "#define ");
	print_str_escaped_toupper(stream, progname);
	xfprintf(stream, "_");
	print_str_escaped_toupper(stream, id);
	xfprintf(stream, "_autogenerated_h\n\n");
}

static void print_cpp_header_guard_begin(FILE *stream)
{
	xfprintf(stream, "#ifdef __cplusplus\n");
	xfprintf(stream, "extern \"C\" {\n");
	xfprintf(stream, "#endif\n\n");
}

static void print_cpp_header_guard_end(FILE *stream)
{
	xfprintf(stream, "#ifdef __cplusplus\n");
	xfprintf(stream, "}\n");
	xfprintf(stream, "#endif /* __cplusplus */\n\n");
}

static void print_header_guard_end(FILE *stream)
{
	xfprintf(stream, "#endif\n");
}

static const char *c_field_type(struct pt_node *n)
{
	switch (n->middle->type) {
	case PT_ARRAY:
		return "struct cafe_vm_dynarray";
	case PT_DICT:
		return "struct cafe_dict_struct";
	case PT_BLOB:
		return "struct cafe_vm_blob";
	case PT_BOOLEAN:
		return "bool";
	case PT_CHAR:
		return "char";
	case PT_DOUBLE:
		return "double";
	case PT_FLOAT:
		return "float";
	case PT_INT:
		return "int";
	case PT_LONG:
		return "long";
	case PT_LONGLONG:
		return "long long";
	case PT_REAL:
		return "double";
	case PT_SHORT:
		return "short";
	case PT_STRING:
		return "char *";
	case PT_UCHAR:
		return "unsigned char";
	case PT_UINT:
		return "unsigned int";
	case PT_ULONG:
		return "unsigned long";
	case PT_ULONGLONG:
		return "unsigned long long";
	case PT_USHORT:
		return "unsigned short";
	case PT_IDENTIFIER:
	case PT_ATTRIBUTE:
	case PT_CONST:
	case PT_NUMBER:
	case PT_AND:
	case PT_ASSIGNMENT:
	case PT_CONDITIONAL:
	case PT_DIVIDE:
	case PT_EQ:
	case PT_FALSE:
	case PT_GE:
	case PT_GT:
	case PT_KEY:
	case PT_LE:
	case PT_LIST:
	case PT_LT:
	case PT_MINUS:
	case PT_MOD:
	case PT_NATURE:
	case PT_NATURE_REF:
	case PT_NE:
	case PT_NOT:
	case PT_NULL:
	case PT_OR:
	case PT_PLUS:
	case PT_ROOT:
	case PT_SETTING:
	case PT_STRUCT:
	case PT_STRVAL:
	case PT_TIMES:
	case PT_TRUE:
	case PT_TYPEDEF:
		break;
	case PT_UNDEFINED:
		assert(0);
	}
	return QUOTEME(__FILE__) QUOTEME(__LINE__) "@WRS_FIXME";
}

static bool attribute_has_setting(struct pt_node *n)
{
	size_t i;

	assert(n->type == PT_ATTRIBUTE);

	if (n->left == NULL)
		return false;

	for (i = 0; i < n->left->children.count; i++) {
		if (n->left->children.list[i]->type == PT_SETTING)
			return true;
	}

	return false;
}

static size_t c_field_count(struct pt_node *n)
{
	size_t i;
	size_t count = 0;

	assert(n->type == PT_STRUCT);

	if (n->right != NULL && n->right->children.count > 0) {
		for (i = 0; i < n->right->children.count; i++) {
			if (!attribute_has_setting(n->right->children.list[i]))
				count++;
		}
	}

	return count;
}

static void print_c_struct(FILE *stream, struct pt_node *n)
{
	size_t i;

	if (n->arg4 != NULL)
		xfprintf(stream, "%s\n", n->arg4->istr->str);

	xfprintf(stream, "struct ");
	print_node_str_with_prefix(stream, n, false);

	if (n->right != NULL) {
		if (n->right->children.count > 0)
			xfprintf(stream, " {\n");

		for (i = 0; i < n->right->children.count; i++) {
			const char *sep = " ";
			const char *type = c_field_type(n->right->children.list[i]);
			size_t type_sz = strlen(type);
			assert(n->right->children.list[i]->type == PT_ATTRIBUTE);
			if (attribute_has_setting(n->right->children.list[i]))
				continue;
			if (type_sz > 0 && *(type + type_sz - 1) == '*')
				sep = "";
			xfprintf(stream, "\t%s%s%s;\n", c_field_type(n->right->children.list[i]), sep, n->right->children.list[i]->right->istr->str);
		}

		if (n->right->children.count > 0)
			xfprintf(stream, "}");
	}

	xfprintf(stream, ";\n");
}

static void print_copyright(FILE *stream)
{
	time_t tnow;
	char *date;
	char *year;

	time(&tnow);
	date = ctime(&tnow);
	date[24] = '\0';
	year = &date[20];

	xfprintf(stream, "\
/* Copyright (c) %s Wind River Systems, Inc.\n\
 *\n\
 * The right to copy, distribute, modify or otherwise make use\n\
 * of this software may be licensed only pursuant to the terms\n\
 * of an applicable Wind River license agreement.\n\
 */\n\
\n\
/* This file has been autogenerated by %s(1). */\n", year, progname);
	xfprintf(stream, "\n");
}

static int qsort_cmp(const void *a, const void *b)
{
#ifdef _WIN32
	const struct pt_node **x = a;
	const struct pt_node **y = b;
#else
	struct pt_node *const *x = a;
	struct pt_node *const *y = b;
#endif
	assert((*x)->type == PT_STRUCT || (*x)->type == PT_NATURE);
	assert((*y)->type == PT_STRUCT || (*y)->type == PT_NATURE);
	return strcmp((*x)->left->istr->str, (*y)->left->istr->str);
}

struct get_types_ctx {
	const char *filename;
	struct cafe_ptr_array *types;
};

static int walk_top_level_types(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct get_types_ctx *ctx = (struct get_types_ctx *)data;

	if (visit != PT_WALK_PREORDER)
		return 1;

	if (n->type == PT_STRUCT || n->type == PT_NATURE)
		cafe_ptr_array_append(ctx->types, n);

	return level < 2;
}

static void get_types_for_filename(const char *filename, struct cafe_type_compiler *compiler, struct cafe_ptr_array *types)
{
	struct cafe_list_head *item;
	struct get_types_ctx ctx;

	ctx.filename = filename;
	ctx.types = types;

	cafe_list_foreach(item, &compiler->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		if (filename == NULL || (p->root != NULL && strcmp(p->filename, filename)) == 0) {
			pt_walk(p->root, walk_top_level_types, &ctx);
		}
	}

	qsort(types->data, cafe_ptr_array_len(types), sizeof *types->data, qsort_cmp);
}

static void print_vt_macros(FILE *stream, const char *file_ident, struct cafe_ptr_array *types)
{
	size_t i;

	for (i = 0; i < cafe_ptr_array_len(types); i++) {
		struct pt_node *n = cafe_ptr_array_val(types, i);
		print_vt_def_macro(stream, n);
		print_vt_ref_macro(stream, n);
		if (n->type == PT_NATURE)
			print_nature_macro(stream, n);
		xfprintf(stream, "\n");
	}

	xfprintf(stream, "#define ");
	print_str_escaped_toupper(stream, file_ident);
	xfprintf(stream, "_ALL_VT_DEFS \\\n");

	for (i = 0; i < cafe_ptr_array_len(types); i++) {
		struct pt_node *n = cafe_ptr_array_val(types, i);
		xfprintf(stream, "\t");
		print_node_str_with_prefix(stream, n, false);
		xfprintf(stream, "_VT_DEF \\");
		xfprintf(stream, "\n");
	}

	if (cafe_ptr_array_len(types) > 0)
		xfprintf(stream, "\n");
}

static void print_struct_defs(FILE *stream, struct cafe_ptr_array *types)
{
	size_t i;

	for (i = 0; i < cafe_ptr_array_len(types); i++) {
		struct pt_node *n = cafe_ptr_array_val(types, i);

		if (n->type == PT_STRUCT && c_field_count(n) > 0) {
			print_c_struct(stream, n);
			xfprintf(stream, "\n");
		}
	}
}

static int print_imports(FILE *stream, struct cafe_type_compiler *compiler, const char *filename)
{
	size_t i;
	struct cafe_list_head *item;
	size_t count = 0;

	cafe_list_foreach_rev(item, &compiler->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		if (p->root != NULL && p->root->left != NULL) {
			for (i = 0; i < p->root->left->children.count; i++) {
				count++;
				xfprintf(stream, "#include <cafe/autogenerated/%s.h>\n", p->root->left->children.list[i]->istr->str);
			}
		}
	}

	return count;
}

static void usage(void)
{
	xfprintf(stderr, "Usage: %s [ <INPUT> [ <OUTPUT> ] ]\n", progname);
	xfprintf(stderr, "Converts Cafe Types (structs/natures) to an ANSI-C header file\n");
	xfprintf(stderr, "  -c -- compile only\n");
	exit(EXIT_FAILURE);
}

static int parse_opts(int argc, char *argv[])
{
	int opt;
	int ind = 1;

	while (ind < argc) {
		char *s = argv[ind];

		if (*s++ != '-')
			break;

		ind++;

		if (s[0] == '-' && s[1] == '\0')
			break;

		while ((opt = *s++) != '\0') {
			switch (opt) {
			case 'c':
				compile_only = 1;
				break;
			case 'P':
				if (!(ind < argc))
					die("missing argument for -P\n");
				cli_prefix = argv[ind++];
				break;
			case 'h':
			case '?':
			default:
				usage();
			}
		}
	}

	return ind;
}

int main(int argc, char *argv[])
{
	int rc;
	FILE *ostream = NULL;
	const char *filename = NULL;
	struct cafe_type_compiler compiler;
	struct cafe_ptr_array types;
	int filename_index;
	char *file_ident;
	char *ns;		/* namespace */

	filename_index = parse_opts(argc, argv);

	if (filename_index < argc)
		filename = argv[filename_index];

	if (filename_index + 1 >= argc)
		ostream = stdout;

	rc = compiler_init(&compiler, NULL, NULL, bailout_allocator);

	if (rc != CAFE_OK)
		die("cannot initialize compiler\n");

	rc = compiler_parsefile(&compiler, filename);

	if (rc != CAFE_OK) {
		compiler_release(&compiler);
		return EXIT_FAILURE;
	}

	rc = compiler_validate(&compiler);

	if (rc != CAFE_OK || compile_only) {
		compiler_release(&compiler);
		return rc == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
	}

	cafe_ptr_array_init(&types, bailout_allocator);
	get_types_for_filename(filename, &compiler, &types);

	if (filename == NULL)
		filename = "<stdin>";

	if (ostream == NULL) {
		ostream = fopen(argv[filename_index + 1], "wb");
		if (ostream == NULL) {
			int saved_errno = errno;
			cafe_ptr_array_release(&types);
			compiler_release(&compiler);
			die("cannot open %s: %s\n", argv[filename_index + 1], strerror(saved_errno));
		}
	}

	print_copyright(ostream);
	ns = get_namespace(filename);
	file_ident = generate_file_ident(cli_prefix, ns);
	print_header_guard_begin(ostream, file_ident);
	if (print_imports(ostream, &compiler, filename) > 0)
		xfprintf(ostream, "\n");
	print_cpp_header_guard_begin(ostream);

	if (cafe_ptr_array_len(&types) > 0) {
		print_vt_macros(ostream, file_ident, &types);
		print_struct_defs(ostream, &types);
	}

	print_cpp_header_guard_end(ostream);
	print_header_guard_end(ostream);

	if (ostream != stdout && ostream != NULL) {
		if (fclose(ostream) != 0) {
			int saved_errno = errno;
			cafe_ptr_array_release(&types);
			compiler_release(&compiler);
			die("cannot close %s: %s\n", argv[filename_index + 1], strerror(saved_errno));
		}
	}

	cafe_ptr_array_release(&types);
	compiler_release(&compiler);
	CAFE_FREE(bailout_allocator, ns);
	CAFE_FREE(bailout_allocator, file_ident);

	return EXIT_SUCCESS;
}
