/* Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <cafe/stdio.h>
#include <cafe/cafe.h>
#include <cafe/ptr_array.h>

#include "bailout_allocator.h"
#include "compiler.h"
#include "strbuf.h"

#define QUOTEME_(x) #x
#define QUOTEME(x) QUOTEME_(x)

static const char *const progname = "cafe-typepp";
static int compile_only = 0;

#ifdef __GNUC__
static int xfprintf(FILE *stream, const char *fmt, ...) __attribute__((format(printf, 2, 3)));
static void die(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
#endif

static int xfprintf(FILE *stream, const char *fmt, ...)
{
	int rc;
	va_list ap;

	va_start(ap, fmt);
	rc = vfprintf(stream, fmt, ap);
	va_end(ap);

	if (rc < 0) {
		fprintf(stderr, "write failed: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	return rc;
}

static void die(const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s: error: ", progname);
	fflush(stderr);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
	exit(EXIT_FAILURE);
}

#define fprintf USE_xfprintf_instead

static struct attribute_type {
	pt_type_t type;
	const char *tag;
} attribute_types[] = {
	{ PT_ARRAY,	"array" },
	{ PT_BLOB,	"blob" },
	{ PT_BOOLEAN,	"boolean" },
	{ PT_CHAR,	"char" },
	{ PT_DICT,	"dict" },
	{ PT_DOUBLE,	"double" },
	{ PT_FLOAT,	"float" },
	{ PT_INT,	"int" },
	{ PT_LONG,	"long" },
	{ PT_LONGLONG,	"longlong" },
	{ PT_SHORT,	"short" },
	{ PT_STRING,	"string" },
	{ PT_UCHAR,	"uchar" },
	{ PT_UINT,	"uint" },
	{ PT_ULONG,	"ulong" },
	{ PT_ULONGLONG, "ulonglong" },
	{ PT_USHORT,	"ushort" },
};

static const struct attribute_type *attribute_type(pt_type_t type)
{
	size_t i;

	for (i = 0; i < CAFE_NELEMENTS(attribute_types); i++) {
		if (attribute_types[i].type == type)
			return &attribute_types[i];
	}

	return NULL;
}

static int walk_qualifier_list(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	FILE *stream = (FILE *)data;

	if (visit != PT_WALK_PREORDER)
		return 1;

	if (n->type == PT_LIST)
		return 1;

	if (n->type == PT_KEY)
		xfprintf(stream, "key");
	else if (n->type == PT_SETTING)
		xfprintf(stream, "setting");
	else if (n->type == PT_CONST)
		xfprintf(stream, "const");

	if (!list_ctx->is_last)
		xfprintf(stream, " ");

	return 1;
}

static void print_expr(FILE *stream, struct pt_node *n)
{
	if (n == NULL)
		return;

	print_expr(stream, n->left);

	if (n->right != NULL)
		xfprintf(stream, " ");

	if (n->type == PT_STRVAL)
		xfprintf(stream, "\"%s\"", n->istr->str);
	else if (n->type == PT_TRUE)
		xfprintf(stream, "true");
	else if (n->type == PT_FALSE)
		xfprintf(stream, "false");
	else
		xfprintf(stream, "%s", n->istr->str);

	if (n->right != NULL)
		xfprintf(stream, " ");

	print_expr(stream, n->right);
}

static int walk_natures(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	FILE *stream = (FILE *)data;

	if (n->type == PT_NATURE_REF && visit == PT_WALK_PREORDER) {
		xfprintf(stream, ":%s", n->left->istr->str);
		if (n->middle != NULL) {
			size_t i;

			xfprintf(stream, "(");

			for (i = 0; i < n->middle->children.count; i++) {
				struct pt_node *c = n->middle->children.list[i];
				print_expr(stream, c);
				if (i + 1 != n->middle->children.count)
					xfprintf(stream, ", ");
			}

			xfprintf(stream, ")");
		}

		if (!list_ctx->is_last)
			xfprintf(stream, " ");
	}

	return 1;
}

static int encode_type(struct pt_node *n, struct cafe_byte_buffer *sb)
{
	switch (n->type) {
	case PT_ARRAY:
		strbuf_fmt(sb, "%s<", attribute_type(n->type)->tag);
		encode_type(n->left, sb);
		if (n->middle) {
			strbuf_fmt(sb, ","); /* dim */
			encode_type(n->middle, sb);
		}
		strbuf_fmt(sb, ">");
		break;
	case PT_DICT:
		strbuf_fmt(sb, "%s<", attribute_type(n->type)->tag);
		encode_type(n->left, sb);
		strbuf_fmt(sb, ">");
		break;
	case PT_BLOB:
	case PT_BOOLEAN:
	case PT_CHAR:
	case PT_DOUBLE:
	case PT_FLOAT:
	case PT_INT:
	case PT_LONG:
	case PT_LONGLONG:
	case PT_REAL:
	case PT_SHORT:
	case PT_STRING:
	case PT_UCHAR:
	case PT_UINT:
	case PT_ULONG:
	case PT_ULONGLONG:
	case PT_USHORT:
		strbuf_fmt(sb, "%s", attribute_type(n->type)->tag);
		break;
	case PT_IDENTIFIER:
		strbuf_fmt(sb, "%s", n->istr->str);
		break;
	case PT_ATTRIBUTE:
	case PT_CONST:
	case PT_NUMBER:
		strbuf_fmt(sb, "%s", n->istr->str);
		break;
	case PT_AND:
	case PT_ASSIGNMENT:
	case PT_CONDITIONAL:
	case PT_DIVIDE:
	case PT_EQ:
	case PT_FALSE:
	case PT_GE:
	case PT_GT:
	case PT_KEY:
	case PT_LE:
	case PT_LIST:
	case PT_LT:
	case PT_MINUS:
	case PT_MOD:
	case PT_NATURE:
	case PT_NATURE_REF:
	case PT_NE:
	case PT_NOT:
	case PT_NULL:
	case PT_OR:
	case PT_PLUS:
	case PT_ROOT:
	case PT_SETTING:
	case PT_STRUCT:
	case PT_STRVAL:
	case PT_TIMES:
	case PT_TRUE:
	case PT_TYPEDEF:
		strbuf_fmt(sb, "%s", QUOTEME(__FILE__) QUOTEME(__LINE__) "@WRS_FIXME");
		break;
	case PT_UNDEFINED:
		break;
	}

	return CAFE_OK;
}

static int walk_attributes(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	FILE *stream = (FILE *)data;

	if (n->type == PT_ATTRIBUTE && visit == PT_WALK_PREORDER) {
		struct cafe_byte_buffer attribute_type = STRBUF_INIT(bailout_allocator);
		xfprintf(stream, "\t");
		if (n->left != NULL) {
			pt_walk(n->left, walk_qualifier_list, data);
			xfprintf(stream, " ");
		}
		encode_type(n->middle, &attribute_type);
		xfprintf(stream, "%s %s", attribute_type.data,
			 n->right->istr->str);
		if (n->arg4 != NULL) {
			xfprintf(stream, " = ");
			print_expr(stream, n->arg4->left);
		}
		if (n->arg5 != NULL) {
			xfprintf(stream, " ");
			pt_walk(n->arg5, walk_natures, stream);
		}
		strbuf_release(&attribute_type);
	} else if (n->type == PT_ATTRIBUTE && visit == PT_WALK_POSTORDER) {
		xfprintf(stream, ";\n");
	}

	return 1;
}

static int walk_struct_or_nature_type(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	FILE *stream = (FILE *)data;

	if (n->type == PT_STRUCT || n->type == PT_NATURE) {
		if (visit == PT_WALK_PREORDER) {
			if (n->arg4 != NULL)
				xfprintf(stream, "%s\n", n->arg4->istr->str);
			xfprintf(stream, "%s %s ", n->type == PT_STRUCT ? "struct" : "nature", n->left->istr->str);
			pt_walk(n->middle, walk_natures, data);
			if (n->middle != NULL)
				xfprintf(stream, " ");
			xfprintf(stream, "{");
			if (n->right != NULL)
				xfprintf(stream, "\n");
			pt_walk(n->right, walk_attributes, data);
		} else if (visit == PT_WALK_POSTORDER) {
			xfprintf(stream, "}\n");
		}
	} else if (n->type == PT_TYPEDEF) {
		if (visit == PT_WALK_PREORDER) {
			struct cafe_byte_buffer type = STRBUF_INIT(bailout_allocator);
			encode_type(n->left, &type);
			xfprintf(stream, "typedef %s %s;\n", type.data,
				 n->middle->istr->str);
			strbuf_release(&type);
		}
	}

	return 1;
}

static int qsort_cmp(const void *a, const void *b)
{
#ifdef _WIN32
	const struct pt_node **x = a;
	const struct pt_node **y = b;
#else
	struct pt_node *const *x = a;
	struct pt_node *const *y = b;
#endif
	assert((*x)->type == PT_STRUCT || (*x)->type == PT_NATURE || (*x)->type == PT_TYPEDEF);
	assert((*y)->type == PT_STRUCT || (*y)->type == PT_NATURE || (*y)->type == PT_TYPEDEF);
	return strcmp((*x)->left->istr->str, (*y)->left->istr->str);
}

struct get_types_ctx {
	const char *filename;
	struct cafe_ptr_array *list;
};

static int walk_top_level_types(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct get_types_ctx *ctx = (struct get_types_ctx *)data;

	if (visit != PT_WALK_PREORDER)
		return 1;

	if (n->type == PT_STRUCT || n->type == PT_NATURE || n->type == PT_TYPEDEF)
		cafe_ptr_array_append(ctx->list, n);

	return 1;
}

static void get_types_for_filename(const char *filename, struct cafe_type_compiler *compiler, struct cafe_ptr_array *types)
{
	struct cafe_list_head *item;
	struct get_types_ctx ctx;

	ctx.filename = filename;
	ctx.list = types;

	cafe_list_foreach(item, &compiler->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		if (filename == NULL || (p->root != NULL && strcmp(p->filename, filename) == 0))
			pt_walk(p->root, walk_top_level_types, &ctx);
	}

	qsort(types->data, cafe_ptr_array_len(types), sizeof *types->data, qsort_cmp);
}

static void get_imports_for_filename(const char *filename, struct cafe_type_compiler *compiler, struct cafe_ptr_array *imports)
{
	size_t i;
	struct cafe_list_head *item;

	cafe_list_foreach(item, &compiler->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		if (filename == NULL || (p->root != NULL && strcmp(p->filename, filename) == 0)) {
			if (p->root->left == NULL)
				continue;
			for (i = 0; i < p->root->left->children.count; i++)
				cafe_ptr_array_append(imports, p->root->left->children.list[i]);
		}
	}
}

static void pretty_print(FILE *stream, struct cafe_ptr_array *types)
{
	size_t i;

	if (cafe_ptr_array_len(types) > 0)
		xfprintf(stream, "\n");

	for (i = 0; i < cafe_ptr_array_len(types); i++) {
		pt_walk(cafe_ptr_array_val(types, i), walk_struct_or_nature_type, stream);
		if (i + 1 < cafe_ptr_array_len(types))
			xfprintf(stream, "\n");
	}
}

static void usage(void)
{
	xfprintf(stderr, "Usage: %s [-c] [ <INPUT> [ <OUTPUT> ] ]\n", progname);
	xfprintf(stderr, "Pretty print Cafe Types (structs/natures)\n");
	xfprintf(stderr, "  -c -- compile only\n");
	exit(EXIT_FAILURE);
}

static int parse_opts(int argc, char *argv[])
{
	int opt;
	int ind = 1;

	while (ind < argc) {
		char *s = argv[ind];

		if (*s++ != '-')
			break;

		ind++;

		if (s[0] == '-' && s[1] == '\0')
			break;

		while ((opt = *s++) != '\0') {
			switch (opt) {
			case 'c':
				compile_only = 1;
				break;
			case 'h':
			case '?':
			default:
				usage();
			}
		}
	}

	return ind;
}

int main(int argc, char *argv[])
{
	size_t i;
	int rc;
	FILE *ostream = NULL;
	const char *filename = NULL;
	struct cafe_type_compiler compiler;
	struct cafe_ptr_array types;
	struct cafe_ptr_array imports;
	int filename_index;

	filename_index = parse_opts(argc, argv);

	if (filename_index < argc)
		filename = argv[filename_index];

	if (filename_index + 1 >= argc)
		ostream = stdout;

	rc = compiler_init(&compiler, NULL, NULL, bailout_allocator);

	if (rc != CAFE_OK)
		die("cannot initialize compiler\n");

	rc = compiler_parsefile(&compiler, filename);

	if (rc != CAFE_OK) {
		compiler_release(&compiler);
		return EXIT_FAILURE;
	}

	rc = compiler_validate(&compiler);

	if (rc != CAFE_OK || compile_only) {
		compiler_release(&compiler);
		return rc == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
	}

	cafe_ptr_array_init(&types, bailout_allocator);
	cafe_ptr_array_init(&imports, bailout_allocator);
	get_types_for_filename(filename, &compiler, &types);
	get_imports_for_filename(filename, &compiler, &imports);

	if (filename == NULL)
		filename = "<stdin>";

	if (ostream == NULL) {
		ostream = fopen(argv[filename_index + 1], "wb");
		if (ostream == NULL) {
			int saved_errno = errno;
			cafe_ptr_array_release(&types);
			compiler_release(&compiler);
			die("cannot open %s: %s\n", argv[filename_index + 1], strerror(saved_errno));
		}
	}

	xfprintf(ostream, "// -*- mode: java; c-file-style: \"linux\" -*-\n");
	xfprintf(ostream, "//\n");
	xfprintf(ostream, "// Note: Keeping natures on one line helps Emacs' indentation engine.\n");

	if (cafe_ptr_array_len(&imports) > 0)
		xfprintf(ostream, "\n");

	for (i = 0; i < cafe_ptr_array_len(&imports); i++) {
		struct pt_node *n = cafe_ptr_array_val(&imports, i);
		xfprintf(ostream, "#include \"%s\"\n", n->istr->str);
	}

	pretty_print(ostream, &types);
	cafe_ptr_array_release(&types);
	cafe_ptr_array_release(&imports);
	compiler_release(&compiler);

	if (ostream != stdout && ostream != NULL) {
		if (fclose(ostream) != 0)
			die("cannot close %s: %s\n", argv[filename_index + 1], strerror(errno));
	}

	return EXIT_SUCCESS;
}
