/*
 * Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include <cafe/cafe.h>
#include <cafe/hashtbl.h>
#include <cafe/allocator.h>
#include "grammar.h"
#include "symtbl.h"
#include "strbuf.h"
#include "istr.h"
#include "parser.h"
#include "compiler.h"
#include "../src/strtoll.h"

#define PT_DEBUG 0

#if PT_DEBUG
#define PT_TRACE(N, LEVEL, ID) do { trace(N, LEVEL, ID); } while (0)
#else
#define PT_TRACE(N, LEVEL, ID)
#endif

struct define_context {
	struct cafe_type_compiler *compiler;
	struct scope *current_scope;
};

struct resolve_context {
	struct cafe_type_compiler *compiler;
	struct symbol *symbol;
	struct pt_node *parent;
};

static int check_expr_is_constant(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data);

/* Defined in order from narrowest to widest. */

struct builtin_type {
	pt_type_t type;
	const char *name;
	struct istr *istr;
};

static struct builtin_type builtin_types[18] = {
	/*  0 */ { PT_UNDEFINED,	"void" },
	/*  1 */ { PT_ARRAY,		"array" },
	/*  2 */ { PT_DICT,		"dict" },
	/*  3 */ { PT_BLOB,		"blob" },
	/*  4 */ { PT_STRING,		"string" },
	/*  5 */ { PT_BOOLEAN,		"boolean" },
	/*  6 */ { PT_CHAR,		"char" },
	/*  7 */ { PT_UCHAR,		"uchar" },
	/*  8 */ { PT_SHORT,		"short" },
	/*  9 */ { PT_USHORT,		"ushort" },
	/* 10 */ { PT_INT,		"int" },
	/* 11 */ { PT_UINT,		"uint" },
	/* 12 */ { PT_LONG,		"long" },
	/* 13 */ { PT_ULONG,		"ulong" },
	/* 14 */ { PT_LONGLONG,		"longlong" },
	/* 15 */ { PT_ULONGLONG,	"ulonglong" },
	/* 16 */ { PT_FLOAT,		"float" },
	/* 17 */ { PT_DOUBLE,		"double" }
};

#define _ 0

/* Mapping for type1 <op> type2, with void as an illegal mapping.
 *
 * Read this table as: what is result type if I do:
 *
 *   ... (char)132 + (int)8765
 *   ... (char)132 / (int)8765
 *   ... (char)132 * (int)8765
 *   ... (char)132 - (int)8765
 *
 * Answer: arithmetic_result_tbl[6][10] = char
 */
static int arithmetic_result_tbl[18][18] = {
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*                  |    0 |     1 |    2 |    3 |      4 |       5 |    6 |     7 |     8 |      9 |  10 |   11 |   12 |    13 |       14 |        15 |    16 |     17 | */
	/*                  | void | array | dict | blob | string | boolean | char | uchar | short | ushort | int | uint | long | ulong | longlong | ulonglong | float | double | */
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*  0: void */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  1: array */     {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  2: dict */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  3: blob */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  4: string */    {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  5: boolean */   {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  6: char */      {    _,      _,     _,     _,       _,        _,     6,      6,      6,       6,    6,     6,     6,      6,         6,          6,      6,       6 },
	/*  7: uchar */     {    _,      _,     _,     _,       _,        _,     7,      7,      7,       7,    7,     7,     7,      7,         7,          7,      7,       7 },
	/*  8: short */     {    _,      _,     _,     _,       _,        _,     8,      8,      8,       8,    8,     8,     8,      8,         8,          8,      8,       8 },
	/*  9: ushort */    {    _,      _,     _,     _,       _,        _,     9,      9,      9,       9,    9,     9,     9,      9,         9,          9,      9,       9 },
	/* 10: int */       {    _,      _,     _,     _,       _,        _,    10,     10,     10,      10,   10,    10,    10,     10,        10,         10,     10,      10 },
	/* 11: uint */      {    _,      _,     _,     _,       _,        _,    11,     11,     11,      11,   11,    11,    11,     11,        11,         11,     11,      11 },
	/* 12: long */      {    _,      _,     _,     _,       _,        _,    12,     12,     12,      12,   12,    12,    12,     12,        12,         12,     12,      12 },
	/* 13: ulong */     {    _,      _,     _,     _,       _,        _,    13,     13,     13,      13,   13,    13,    13,     13,        13,         13,     13,      13 },
	/* 14: longlong */  {    _,      _,     _,     _,       _,        _,    14,     14,     14,      14,   14,    14,    14,     14,        14,         14,     14,      14 },
	/* 15: ulonglong */ {    _,      _,     _,     _,       _,        _,    15,     15,     15,      15,   15,    15,    15,     15,        15,         15,     15,      15 },
	/* 16: float */     {    _,      _,     _,     _,       _,        _,    16,     16,     16,      16,   16,    16,    16,     16,        16,         16,     16,      17 },
	/* 17: double */    {    _,      _,     _,     _,       _,        _,    17,     17,     17,      17,   17,    17,    17,     17,        17,         17,     17,      17 }
};

/* Mapping for type1 <op> type2, with void as an illegal mapping.
 *
 * Read this table as: what is result type if I do:
 *
 *  ... (char)132 < (int)8765.
 *
 * Answer: relational_result_tbl[6][10] = boolean
 */
static int relational_result_tbl[18][18] = {
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*                  |    0 |     1 |    2 |    3 |      4 |       5 |    6 |     7 |     8 |      9 |  10 |   11 |   12 |    13 |       14 |        15 |    16 |     17 | */
	/*                  | void | array | dict | blob | string | boolean | char | uchar | short | ushort | int | uint | long | ulong | longlong | ulonglong | float | double | */
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*  0: void */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  1: array */     {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  2: dict */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  3: blob */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  4: string */    {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  5: boolean */   {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  6: char */      {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/*  7: uchar */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/*  8: short */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/*  9: ushort */    {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 10: int */       {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 11: uint */      {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 12: long */      {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 13: ulong */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 14: longlong */  {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 15: ulonglong */ {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 16: float */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 17: double */    {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 }
};

/* Mapping for type1 <op> type2, with void as an illegal mapping.
 *
 * Read this table as: what is result type if I do:
 *
 *  ... (boolean)true == (boolean)false.
 *  ... (char)132 == (int)8765.
 *  ... (char)132 != (int)8765.
 *
 * Answer: equality_result_tbl[6][10] = boolean
 */
static int equality_result_tbl[18][18] = {
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*                  |    0 |     1 |    2 |    3 |      4 |       5 |    6 |     7 |     8 |      9 |  10 |   11 |   12 |    13 |       14 |        15 |    16 |     17 | */
	/*                  | void | array | dict | blob | string | boolean | char | uchar | short | ushort | int | uint | long | ulong | longlong | ulonglong | float | double | */
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*  0: void */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  1: array */     {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  2: dict */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  3: blob */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  4: string */    {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  5: boolean */   {    _,      _,     _,     _,       _,        5,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  6: char */      {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/*  7: uchar */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/*  8: short */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/*  9: ushort */    {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 10: int */       {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 11: uint */      {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 12: long */      {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 13: ulong */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 14: longlong */  {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 15: ulonglong */ {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 16: float */     {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 },
	/* 17: double */    {    _,      _,     _,     _,       _,        _,     5,      5,      5,       5,    5,     5,     5,      5,         5,          5,      5,       5 }
};

/* Truth table for type promotion (or widening): _ implies that no
 * promotion is required, it does not indicate an error. */

static int promote_from_to_tbl[18][18] = {
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*                  |    0 |     1 |    2 |    3 |      4 |       5 |    6 |     7 |     8 |      9 |  10 |   11 |   12 |    13 |       14 |        15 |    16 |     17 | */
	/*                  | void | array | dict | blob | string | boolean | char | uchar | short | ushort | int | uint | long | ulong | longlong | ulonglong | float | double | */
	/* ---------------  |------|-------|------|------|--------|---------|------|-------|-------|--------|-----|------|------|-------|----------|-----------|-------|--------| */
	/*  0: void */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  1: array */     {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  2: dict */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  3: blob */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  4: string */    {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  5: boolean */   {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ },
	/*  6: char */      {    _,      _,     _,     _,       _,        _,     _,      7,      8,       9,   10,    11,    12,     13,        14,         15,     16,      17 },
	/*  7: uchar */     {    _,      _,     _,     _,       _,        _,     _,      _,      8,       9,   10,    11,    12,     13,        14,         15,     16,      17 },
	/*  8: short */     {    _,      _,     _,     _,       _,        _,     _,      _,      _,       9,   10,    11,    12,     13,        14,         15,     16,      17 },
	/*  9: ushort */    {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,   10,    11,    12,     13,        14,         15,     16,      17 },
	/* 10: int */       {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,    11,    12,     13,        14,         15,     16,      17 },
	/* 11: uint */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,    12,     13,        14,         15,     16,      17 },
	/* 12: long */      {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,     13,        14,         15,     16,      17 },
	/* 13: ulong */     {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,        14,         15,     16,      17 },
	/* 14: longlong */  {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,         15,     16,      17 },
	/* 15: ulonglong */ {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,     16,      17 },
	/* 16: float */     {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,      17 },
	/* 17: double */    {    _,      _,     _,     _,       _,        _,     _,      _,      _,       _,    _,     _,     _,      _,         _,          _,      _,       _ }
};

#undef _

static struct pt_node *builtin_type_node[18];

static void print_scope(struct scope *scope)
{
	struct cafe_list_head *item;

	cafe_list_foreach(item, &scope->symbol_list_head) {
		struct symbol *symbol;
		symbol = SYMBOL_LIST_ENTRY(item);
		printf("'%s:%s'", symbol->name->str, pt_type_str(symbol->type));
		if (item->next != &scope->symbol_list_head)
			printf(" ");
	}
}

static void print_scope_tree(struct cafe_list_head *list_head)
{
	struct cafe_list_head *item;

	cafe_list_foreach(item, list_head) {
		struct scope *scope;
		scope = SCOPE_LIST_ENTRY(item);
		printf("scope: '%s' symbols(", scope->name);
		print_scope(scope);
		printf(")\n");
	}
}

#if PT_DEBUG
static void trace(struct pt_node *n, int level, const char *id)
{
	fprintf(stdout, "PT_TRACE: level: %d %s %s%s%s%s[%s%s%u:%u]\n",
		level,
		id,
		pt_type_str(n->type),
		n->istr ? " '" : " ",
		n->istr ? n->istr->str : "",
		n->istr ? "' " : "",
		n->loc.filename ? n->loc.filename : "",
		n->loc.filename ? ":" : "",
		n->loc.lineno,
		n->loc.column);
}
#endif

static pt_type_t numeric_type(const char *str,
			      /* True if the value is known not to have negative values. */
			      bool non_negative)
{
	assert(str);

/* WRS_XXX I know there's a better way than this.  This doesn't handle
 * trailing L, LL, UL, ULL indicators either. */

	if (isdigit(*str)) {
		unsigned long long val;
		char *endptr;

		errno = 0;
		val = strtoull(str, &endptr, 0);
		assert(errno == 0);

		if (val <= UCHAR_MAX)
			return PT_UCHAR;
		else if (val <= USHRT_MAX)
			return PT_USHORT;
		else if (val <= UINT_MAX)
			return PT_UINT;
		else if (val <= ULONG_MAX)
			return PT_ULONG;
		else
			return PT_ULONGLONG;
	} else {
		long long val;
		char *endptr;

		errno = 0;
		val = strtoll(str, &endptr, 0);
		assert(errno == 0);

		if (val >= SCHAR_MIN && val <= SCHAR_MAX)
			return PT_CHAR;
		else if (val >= SHRT_MIN && val <= SHRT_MAX)
			return PT_SHORT;
		else if (val >= INT_MIN && val <= INT_MAX)
			return PT_INT;
		else if (val >= LONG_MIN && val <= LONG_MAX)
			return PT_LONG;
		else
			return PT_LONGLONG;
	}
}

static bool can_assign_to(pt_type_t value_type, pt_type_t dest_type, pt_type_t promotion_type)
{
	return (value_type == dest_type || promotion_type == dest_type);
}

void compiler_error(struct cafe_type_compiler *c, const struct src_loc *loc, const char *fmt, ...)
{
	int rc;
	va_list ap;
	struct cafe_byte_buffer sb = STRBUF_INIT(c->allocator);

	c->compilation_errors++;
	va_start(ap, fmt);
	rc = strbuf_vfmt(&sb, fmt, ap);
	va_end(ap);
	assert(c->error_handler);

	if (rc != CAFE_OK)
		c->error_handler(loc, "out of memory!");
	else
		c->error_handler(loc, sb.data);

	strbuf_release(&sb);
}

void compiler_warning(struct cafe_type_compiler *c, const struct src_loc *loc, const char *fmt, ...)
{
	int rc;
	va_list ap;
	struct cafe_byte_buffer sb = STRBUF_INIT(c->allocator);

	c->compilation_warnings++;
	va_start(ap, fmt);
	rc = strbuf_vfmt(&sb, fmt, ap);
	va_end(ap);
	assert(c->warning_handler);

	if (rc != CAFE_OK)
		c->warning_handler(loc, "out of memory!");
	else
		c->warning_handler(loc, sb.data);

	strbuf_release(&sb);
}

void compiler_default_error_handler(const struct src_loc *loc, const char *msg)
{
	if (loc != NULL) {
		fprintf(stderr, "%s%s%d:%d: ",
			loc->filename ? loc->filename : "",
			loc->filename ? ":" : "",
			loc->lineno, loc->column);
		fprintf(stderr, "error: ");
	} else {
		fprintf(stderr, "error: ");
	}

	fputs(msg, stderr);
	fputc('\n', stderr);
}

void compiler_default_warning_handler(const struct src_loc *loc, const char *msg)
{
	if (loc != NULL) {
		fprintf(stderr, "%s%s%d:%d: ",
			loc->filename ? loc->filename : "",
			loc->filename ? ":" : "",
			loc->lineno, loc->column);
		fprintf(stderr, "warning: ");
	} else {
		fprintf(stderr, "warning: ");
	}

	fputs(msg, stderr);
	fputc('\n', stderr);
}

#define error compiler_error
#define warning compiler_warning

static int rank2type(int n)
{
	switch (n) {
	case 0:		return PT_UNDEFINED;
	case 1:		return PT_ARRAY;
	case 2:		return PT_DICT;
	case 3:		return PT_BLOB;
	case 4:		return PT_STRING;
	case 5:		return PT_BOOLEAN;
	case 6:		return PT_CHAR;
	case 7:		return PT_UCHAR;
	case 8:		return PT_SHORT;
	case 9:		return PT_USHORT;
	case 10:	return PT_INT;
	case 11:	return PT_UINT;
	case 12:	return PT_LONG;
	case 13:	return PT_ULONG;
	case 14:	return PT_LONGLONG;
	case 15:	return PT_ULONGLONG;
	case 16:	return PT_FLOAT;
	case 17:	return PT_DOUBLE;
	default:
		assert(0);
	}
	assert(0);
	return PT_UNDEFINED;
}

static int type_rank(pt_type_t type)
{
	switch (type) {
	case PT_UNDEFINED:	return 0;
	case PT_ARRAY:		return 1;
	case PT_DICT:		return 2;
	case PT_BLOB:		return 3;
	case PT_STRING:		return 4;
	case PT_BOOLEAN:	return 5;
	case PT_CHAR:		return 6;
	case PT_UCHAR:		return 7;
	case PT_SHORT:		return 8;
	case PT_USHORT:		return 9;
	case PT_INT:		return 10;
	case PT_UINT:		return 11;
	case PT_LONG:		return 12;
	case PT_ULONG:		return 13;
	case PT_LONGLONG:	return 14;
	case PT_ULONGLONG:	return 15;
	case PT_FLOAT:		return 16;
	case PT_DOUBLE:		return 17;
	case PT_AND:
	case PT_ASSIGNMENT:
	case PT_ATTRIBUTE:
	case PT_CONDITIONAL:
	case PT_CONST:
	case PT_NUMBER:
	case PT_DIVIDE:
	case PT_EQ:
	case PT_FALSE:
	case PT_GE:
	case PT_GT:
	case PT_IDENTIFIER:
	case PT_KEY:
	case PT_LE:
	case PT_LIST:
	case PT_LT:
	case PT_MINUS:
	case PT_MOD:
	case PT_NATURE:
	case PT_NATURE_REF:
	case PT_NE:
	case PT_NOT:
	case PT_NULL:
	case PT_OR:
	case PT_PLUS:
	case PT_REAL:
	case PT_ROOT:
	case PT_SETTING:
	case PT_STRUCT:
	case PT_STRVAL:
	case PT_TIMES:
	case PT_TRUE:
	case PT_TYPEDEF:
		break;
	}

	assert(0);
	return PT_UNDEFINED;
}

static struct symbol *basetype(struct symbol *symbol)
{
	assert(symbol);

	while (symbol->type == PT_TYPEDEF) {
		assert(symbol->node_def);
		assert(symbol->node_def_parent);
		symbol = scope_resolve(symbol->scope, symbol->node_def_parent->left->istr);
	}

	return symbol;
}

static pt_type_t get_expr_result_type(int type_table[][18], struct pt_node *left, struct pt_node *right, struct resolve_context *ctx)
{
	int a_idx = type_rank(left->eval_type);
	int b_idx = type_rank(right->eval_type);
	pt_type_t result;

	result = builtin_types[type_table[a_idx][b_idx]].type;

	if (result == PT_UNDEFINED) {
		error(ctx->compiler, &left->loc, "'%s', '%s' have incompatible types", pt_type_str(left->type), pt_type_str(right->type));
	} else {
		left->promote_to_type  = rank2type(promote_from_to_tbl[a_idx][b_idx]);
		right->promote_to_type = rank2type(promote_from_to_tbl[b_idx][a_idx]);
	}

	return result;
}

struct walk_expr_context {
	struct cafe_type_compiler *compiler;
	struct scope *scope;
};

static struct symbol *alloc_nature(struct cafe_type_compiler *c, struct scope *scope, struct pt_nature *n)
{
	struct symbol *symbol;

	assert(n->node->type == PT_NATURE);
	symbol = symbol_alloc(n->ident->istr, n->node->type, &n->ident->loc, scope, c->allocator);
	symbol->node_def = n->ident;
	symbol->node_def_parent = n->node;
	n->ident->symbol = symbol;

	return symbol;
}

static struct symbol *alloc_struct(struct cafe_type_compiler *c, struct scope *scope, struct pt_struct *n)
{
	struct symbol *symbol;

	assert(n->node->type == PT_STRUCT);
	symbol = symbol_alloc(n->ident->istr, n->node->type, &n->ident->loc, scope, c->allocator);
	symbol->node_def = n->ident;
	symbol->node_def_parent = n->node;
	n->ident->symbol = symbol;

	return symbol;
}

static struct symbol *alloc_attribute(struct cafe_type_compiler *c, struct scope *scope, struct pt_attribute *n)
{
	struct symbol *symbol;

	assert(n->node->type == PT_ATTRIBUTE);
	symbol = symbol_alloc(n->ident->istr, n->typeref->type, &n->ident->loc, scope, c->allocator);
	symbol->node_def = n->ident;
	symbol->node_def_parent = n->node;
	n->ident->symbol = symbol;

	return symbol;
}

static struct symbol *alloc_typedef(struct cafe_type_compiler *c, struct scope *scope, struct pt_typedef *n)
{
	struct symbol *symbol;

	assert(n->node->type == PT_TYPEDEF);
	symbol = symbol_alloc(n->ident->istr, n->node->type, &n->ident->loc, scope, c->allocator);
	symbol->node_def = n->ident;
	symbol->node_def_parent = n->node;
	n->ident->symbol = symbol;

	return symbol;
}

static bool aggregate_has_a_key(struct pt_node *n)
{
	size_t i;
	size_t j;

	assert(n->type == PT_STRUCT || n->type == PT_NATURE);

	if (n->right == NULL)
		return false;

	assert(n->right->type == PT_LIST);

	for (i = 0; i < n->right->children.count; i++) {
		struct pt_node *c = n->right->children.list[i];

		if (c->left != NULL && c->left->children.list != NULL) {
			for (j = 0; j < c->left->children.count; j++) {
				if (c->left->children.list[j]->type == PT_KEY)
					return true;
			}
		}
	}

	return false;
}

static int define_scope(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	assert(visit == PT_WALK_PREORDER);
	n->scope = (struct scope *)data;
	return 1;
}

static int compute_expr_type(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct resolve_context *ctx = (struct resolve_context *)data;
	unsigned int error_count = ctx->compiler->compilation_errors;

	PT_TRACE(n, level, "compute_expr_type");
	assert(ctx->parent);
	assert(visit == PT_WALK_POSTORDER);

	switch (n->type) {
		/* Binary Operators */
	case PT_DIVIDE:
	case PT_MINUS:
	case PT_MOD:
	case PT_PLUS:
	case PT_TIMES:
		n->eval_type = get_expr_result_type(arithmetic_result_tbl, n->left, n->right, ctx);
		break;

		/* Relational Operators */
	case PT_GE:
	case PT_GT:
	case PT_LE:
	case PT_LT:
		get_expr_result_type(relational_result_tbl, n->left, n->right, ctx);
		/* although the types may be different, these operators are always boolean. */
		n->eval_type = PT_BOOLEAN;
		break;

		/* Ternary Operators */
	case PT_CONDITIONAL:
		n->eval_type = get_expr_result_type(arithmetic_result_tbl, n->middle, n->right, ctx);
		break;

		/* Equality Operators */
	case PT_EQ:
	case PT_NE:
		get_expr_result_type(equality_result_tbl, n->left, n->right, ctx);
		/* although the types may be different, these operators are always boolean. */
		n->eval_type = PT_BOOLEAN;
		break;

		/* Boolean Operators */
	case PT_TRUE:
	case PT_FALSE:
		n->eval_type = PT_BOOLEAN;
		break;

		/* Logical Operators */
	case PT_AND:
	case PT_NOT:
	case PT_OR:
		n->eval_type = PT_BOOLEAN;
		break;

		/* Unary Operators. */

		/* Literals */
	case PT_NUMBER:
		n->eval_type = numeric_type(n->istr->str, false);
		break;

	case PT_STRVAL:
		n->eval_type = PT_STRING;
		break;

	case PT_REAL:
		n->eval_type = PT_FLOAT; /* WRS_XXX need to handle double. */
		break;

	case PT_ARRAY:
	case PT_ASSIGNMENT:
	case PT_ATTRIBUTE:
	case PT_BLOB:
	case PT_BOOLEAN:
	case PT_CHAR:
	case PT_CONST:
	case PT_DOUBLE:
	case PT_DICT:
	case PT_FLOAT:
	case PT_IDENTIFIER:
	case PT_INT:
	case PT_KEY:
	case PT_LIST:
	case PT_LONG:
	case PT_LONGLONG:
	case PT_NATURE:
	case PT_NATURE_REF:
	case PT_NULL:
	case PT_ROOT:
	case PT_SETTING:
	case PT_SHORT:
	case PT_STRING:
	case PT_STRUCT:
	case PT_TYPEDEF:
	case PT_UCHAR:
	case PT_UINT:
	case PT_ULONG:
	case PT_ULONGLONG:
	case PT_USHORT:
	case PT_UNDEFINED:
		return 1;
	}

	/* Give up walking the remainder of the tree if we generated an error. */
	return ctx->compiler->compilation_errors == error_count;
}

static int define_attributes(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct pt_attribute *attribute;
	struct symbol *symbol;
	struct define_context *ctx = (struct define_context *)data;

	PT_TRACE(n, level, "define_attributes");
	assert(visit == PT_WALK_PREORDER);

	if (n->type == PT_ATTRIBUTE) {
		attribute = &n->u.attribute;
		symbol = scope_resolve_field(ctx->current_scope, attribute->ident->istr);

		if (symbol != NULL) {
			error(ctx->compiler, &attribute->ident->loc, "duplicate member '%s'", attribute->ident->istr->str);
			error(ctx->compiler, symbol->loc, "previous declaration of '%s' was here", symbol->name->str);
		} else {
			pt_walk_descendants_preorder(attribute->node, define_scope, ctx->current_scope);
			symbol = alloc_attribute(ctx->compiler, ctx->current_scope, attribute);
			scope_add_symbol(ctx->current_scope, symbol);
		}

		return 0;
	}

	return 1;
}

static void define_struct(struct pt_node *n, struct define_context *ctx)
{
	struct symbol *symbol;
	struct pt_struct *ztruct;

	ztruct = &n->u.structure;
	symbol = cafe_hashtbl_lookup(ctx->current_scope->symbols, ztruct->ident->istr);

	if (symbol != NULL) {
		error(ctx->compiler, &ztruct->ident->loc, "redefinition of type '%s'", ztruct->ident->istr->str);
		error(ctx->compiler, symbol->loc, "previous declaration of '%s' was here", symbol->name->str);
	} else {
		pt_walk_preorder(ztruct->ident, define_scope, ctx->current_scope);
		pt_walk_preorder(ztruct->natures, define_scope, ctx->current_scope);
		symbol = alloc_struct(ctx->compiler, ctx->current_scope, ztruct);
		scope_add_symbol(ctx->current_scope, symbol);
		ctx->current_scope = scope_alloc(ctx->current_scope, ztruct->ident->istr->str, ctx->compiler->allocator);
		cafe_list_add_last(&ctx->current_scope->scope_list_item, &ctx->compiler->scope_list_head);
		pt_walk_preorder(ztruct->attributes, define_attributes, ctx);
		ctx->current_scope = ctx->current_scope->enclosing_scope;
	}
}

static void define_nature(struct pt_node *n, struct define_context *ctx)
{
	struct symbol *symbol;
	struct pt_nature *nature;

	nature = &n->u.nature;
	nature->ident->scope = ctx->current_scope;
	symbol = cafe_hashtbl_lookup(ctx->current_scope->symbols, nature->ident->istr);

	if (symbol != NULL) {
		error(ctx->compiler, &nature->ident->loc, "redefinition of type '%s'", nature->ident->istr->str);
		error(ctx->compiler, symbol->loc, "previous declaration of '%s' was here", symbol->name->str);
	} else {
		pt_walk_preorder(nature->natures, define_scope, ctx->current_scope);
		symbol = alloc_nature(ctx->compiler, ctx->current_scope, nature);
		scope_add_symbol(ctx->current_scope, symbol);
		ctx->current_scope = scope_alloc(ctx->current_scope, nature->ident->istr->str, ctx->compiler->allocator);
		cafe_list_add_last(&ctx->current_scope->scope_list_item, &ctx->compiler->scope_list_head);
		pt_walk_preorder(nature->attributes, define_attributes, ctx);
		ctx->current_scope = ctx->current_scope->enclosing_scope;
	}
}

static void define_typedef(struct pt_node *n, struct define_context *ctx)
{
	struct symbol *symbol;
	struct pt_typedef *ztypedef;

	assert(n->type == PT_TYPEDEF);
	ztypedef = &n->u.ztypedef;
	pt_walk_preorder(ztypedef->typeref, define_scope, ctx->current_scope);
	pt_walk_preorder(ztypedef->ident, define_scope, ctx->current_scope);
	symbol = cafe_hashtbl_lookup(ctx->current_scope->symbols, ztypedef->ident->istr);

	if (symbol != NULL) {
		error(ctx->compiler, &ztypedef->ident->loc, "redefinition of type '%s'", ztypedef->ident->istr->str);
		error(ctx->compiler, symbol->loc, "previous declaration of '%s' was here", symbol->name->str);
	} else {
		symbol = alloc_typedef(ctx->compiler, ctx->current_scope, ztypedef);
		scope_add_symbol(ctx->current_scope, symbol);
	}
}

static int define_types(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct define_context *ctx = (struct define_context *)data;

	PT_TRACE(n, level, "define_types");
	assert(visit == PT_WALK_PREORDER);

	if (n->type == PT_STRUCT) {
		define_struct(n, ctx);
		return 0;
	}  else if (n->type == PT_NATURE) {
		define_nature(n, ctx);
		return 0;
	} else if (n->type == PT_TYPEDEF) {
		define_typedef(n, ctx);
		return 0;
	}

	return 1;
}

static int resolve_typeref(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct symbol *symbol;
	struct resolve_context *ctx = (struct resolve_context *)data;

	PT_TRACE(n, level, "resolve_typeref");
	assert(visit == PT_WALK_PREORDER);
	assert(n->scope);
	assert(ctx->symbol);
	symbol = scope_resolve(n->scope, n->istr);

	if (symbol == NULL) {
		error(ctx->compiler, &n->loc, "unknown type '%s'", n->istr->str);
		return 0;
	} else if (symbol == ctx->symbol) {
		error(ctx->compiler, &n->loc, "unknown type '%s'", n->istr->str);
		return 0;
	}

	if (level > 0) {
		if (ctx->symbol->type == PT_DICT && basetype(symbol)->type != PT_STRUCT) {
			error(ctx->compiler, &n->loc, "dict type '%s' (%s) is not a struct", n->istr->str, pt_type_str(basetype(symbol)->type));
			return 0;
		}
		if (ctx->symbol->type == PT_DICT && !aggregate_has_a_key(symbol->node_def_parent)) {
			error(ctx->compiler, &n->loc, "dict type '%s' has no key", basetype(symbol)->name->str);
			return 0;
		}
	}

	ctx->symbol->symbol_type = symbol;
	ctx->symbol = symbol;

	return 1;		/* resolve other type references. */
}

static int resolve_natures(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct symbol *symbol;
	struct resolve_context *ctx = (struct resolve_context *)data;

	PT_TRACE(n, level, "resolve_natures");
	assert(visit == PT_WALK_PREORDER);

	if (n->type != PT_NATURE_REF)
		return 1;

	assert(n->left->scope);
	symbol = scope_resolve(n->left->scope, n->left->istr);

	if (symbol == NULL)
		error(ctx->compiler, &n->left->loc, "unknown nature '%s'", n->left->istr->str);
	else if (symbol->type != PT_NATURE) {
		if (basetype(symbol)->type != PT_NATURE) {
			error(ctx->compiler, &n->left->loc, "type '%s' is not a nature", n->left->istr->str);
			return 0;
		}
	} else {
		pt_walk_postorder(n->middle, compute_expr_type, ctx);
		pt_walk_descendants_postorder(n->middle, check_expr_is_constant, ctx);
	}

	return 0;
}

static int resolve_field(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct resolve_context *ctx = (struct resolve_context *)data;
	struct pt_attribute *attribute;

	PT_TRACE(n, level, "resolve_field");
	assert(visit == PT_WALK_PREORDER);

	if (n->type != PT_ATTRIBUTE)
		return 1;

	attribute = &n->u.attribute;
	assert(attribute->ident->symbol);
	assert(attribute->ident->symbol->node_def);
	assert(attribute->ident->symbol->node_def->type == PT_IDENTIFIER);
	ctx->symbol = attribute->ident->symbol;
	pt_walk_preorder(attribute->typeref, resolve_typeref, ctx);
	pt_walk_preorder(attribute->natures, resolve_natures, ctx);

	if (attribute->ident->symbol->symbol_type != NULL) {
		assert(ctx->parent->type == PT_STRUCT || ctx->parent->type == PT_NATURE);
		if (strcmp(ctx->parent->left->symbol->name->str, basetype(attribute->ident->symbol->symbol_type)->name->str) == 0) {
			error(ctx->compiler, &attribute->typeref->loc, "type '%s' is self-referential", attribute->typeref->istr->str);
			return 0;
		}
	}

	if (attribute->assignment_expr != NULL) {
		int texpr; /* type expression */
		int tdecl; /* type declaration */
		unsigned int error_count = ctx->compiler->compilation_errors;

		assert(ctx->parent);
		pt_walk_descendants_postorder(attribute->assignment_expr, compute_expr_type, ctx);

		if (error_count != ctx->compiler->compilation_errors)
			return 0;

		texpr = type_rank(attribute->assignment_expr->left->eval_type);
		tdecl = type_rank(basetype(attribute->ident->symbol->symbol_type)->eval_type);
		attribute->assignment_expr->left->promote_to_type = rank2type(promote_from_to_tbl[texpr][tdecl]);

		if (!can_assign_to(attribute->assignment_expr->left->eval_type,      /* right */
				   attribute->ident->symbol->symbol_type->eval_type, /* left */
				   attribute->assignment_expr->left->promote_to_type)) {
			error(ctx->compiler, &attribute->ident->loc, "incompatible types in assignment");
			return 0;
		}

		pt_walk_postorder(attribute->assignment_expr->left, check_expr_is_constant, ctx);
	}

	return 0;
}

static int resolve_types(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	struct resolve_context *ctx = (struct resolve_context *)data;

	PT_TRACE(n, level, "resolve_types");
	assert(visit == PT_WALK_PREORDER);

	if (n->type == PT_STRUCT) {
		struct pt_struct *ztruct = &n->u.structure;
		ctx->parent = n;
		pt_walk_preorder(ztruct->natures, resolve_natures, ctx);
		pt_walk_preorder(ztruct->attributes, resolve_field, ctx);
		return 0;
	} else if (n->type == PT_NATURE) {
		struct pt_nature *nature = &n->u.nature;
		ctx->parent = n;
		pt_walk_preorder(nature->natures, resolve_natures, ctx);
		pt_walk_preorder(nature->attributes, resolve_field, ctx);
		return 0;
	} else if (n->type == PT_TYPEDEF) {
		struct pt_typedef *ztypedef = &n->u.ztypedef;
		assert(ztypedef->ident->symbol);
		assert(ztypedef->ident->symbol->node_def);
		assert(ztypedef->ident->symbol->node_def->type == PT_IDENTIFIER);
		ctx->symbol = ztypedef->ident->symbol;
		pt_walk_preorder(ztypedef->typeref, resolve_typeref, ctx);
		return 0;
	}

	return 1;
}

static int resolve_nature_expressions(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	size_t i;
	struct symbol *symbol;
	struct resolve_context *ctx = (struct resolve_context *)data;

	PT_TRACE(n, level, "resolve_natures");
	assert(visit == PT_WALK_PREORDER);

	if (n->type != PT_NATURE_REF)
		return 1;

	if (n->middle == NULL)
		return 1;

	assert(n->left->scope);
	symbol = scope_resolve(n->left->scope, n->left->istr);
	assert(symbol);

	for (i = 0; i < n->middle->children.count; i++) {
		int texpr; /* type expression */
		int tdecl; /* type declaration */
		struct pt_node *node_decl;
		struct pt_node *node_expr;

		if (symbol->node_def_parent->u.nature.attributes == NULL) {
			error(ctx->compiler, &n->left->loc, "nature '%s' has 0 attributes, but %zu supplied",
			      symbol->name->str,
			      n->middle->children.count);
			return 0;
		}

		if (n->middle->children.count > symbol->node_def_parent->u.nature.attributes->children.count) {
			error(ctx->compiler, &n->left->loc, "nature '%s' has %zu attributes, but %zu supplied",
			      symbol->name->str,
			      symbol->node_def_parent->u.nature.attributes->children.count,
			      n->middle->children.count);
			return 0;
		}

		node_decl = symbol->node_def_parent->u.nature.attributes->children.list[i];
		node_expr = n->middle->children.list[i];
		texpr = type_rank(node_expr->eval_type);
		tdecl = type_rank(node_decl->u.attribute.ident->symbol->symbol_type->eval_type);

		if (!can_assign_to(node_expr->eval_type, /* right */
				   node_decl->u.attribute.ident->symbol->symbol_type->eval_type, /* left */
				   rank2type(promote_from_to_tbl[texpr][tdecl]))) {
			error(ctx->compiler, &node_expr->loc, "incompatible types when initializing attribute '%s.%s'",
			      symbol->name->str,
			      node_decl->u.attribute.ident->istr->str);
			return 0;
		}
	}

	return 1;
}

int compiler_validate(struct cafe_type_compiler *c)
{
	struct cafe_list_head *item;

	/* Pass 1.  Define symbols (i.e., identifiers). */

	cafe_list_foreach(item, &c->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		struct define_context ctx = { c, c->global_scope };

		pt_walk_preorder(p->root, define_types, &ctx);

		if (c->compilation_errors > 0)
			return CAFE_ERROR;
	}

	/* Pass 2.  Resolve identifiers, including expression types. */

	cafe_list_foreach(item, &c->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		struct resolve_context ctx = { c };

		pt_walk_preorder(p->root, resolve_types, &ctx);

		if (c->compilation_errors > 0)
			return CAFE_ERROR;
	}

	/* Pass 3.  Resolve nature expression usage. */

	cafe_list_foreach(item, &c->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		struct resolve_context ctx = { c };

		pt_walk_preorder(p->root, resolve_nature_expressions, &ctx);

		if (c->compilation_errors > 0)
			return CAFE_ERROR;
	}

	if (getenv("PRINT_SCOPE_TREE") != NULL)
		print_scope_tree(&c->scope_list_head);

	if (getenv("DUMP_PT_AS_DOT") != NULL) {
		FILE *stream;

		stream = fopen(getenv("DUMP_PT_AS_DOT"), "wb");

		if (stream == NULL)
			stream = stderr;

		pt_print_dot(c, stream);

		if (stream != stderr)
			fclose(stream);
	}

	/* Invoke with DUMP_PT=/dev/{stdout,stderr} if you don't want this going to a file. */

	if (getenv("DUMP_PT") != NULL) {
		FILE *stream;
		struct cafe_list_head *item;

		stream = fopen(getenv("DUMP_PT"), "wb");

		if (stream == NULL)
			stream = stderr;

		cafe_list_foreach(item, &c->parser_list_head) {
			struct parser *p = PARSER_LIST_ENTRY(item);
			pt_print_node(p->root, stream, 0);
		}

		if (stream != stderr)
			fclose(stream);
	}

	return c->compilation_errors == 0 ? CAFE_OK : CAFE_ERROR;
}

int compiler_parsefile(struct cafe_type_compiler *c, const char *filename)
{
	int rc = parser_parsefile(c, filename);

	return rc;
}

static void strtbl_delval(void *k, void *client_data)
{
	istr_free((struct istr *)k, (const struct cafe_allocator *)client_data);
}

static void scope_tree_release(struct cafe_type_compiler *c)
{
	struct cafe_list_head *item, *tmp;
	const struct cafe_allocator *allocator = c->allocator;

	cafe_list_foreach_safe(item, &c->scope_list_head, tmp) {
		struct scope *scope;
		struct cafe_list_head *item2, *tmp2;

		scope = SCOPE_LIST_ENTRY(item);

		cafe_list_foreach_safe(item2, &scope->symbol_list_head, tmp2) {
			struct symbol *symbol;
			symbol = SYMBOL_LIST_ENTRY(item2);
			symbol_free(symbol, allocator);
		}

		assert(cafe_list_is_empty(&scope->symbol_list_head));
		scope_free(scope, allocator);
	}
	assert(cafe_list_is_empty(&c->scope_list_head));
}

void compiler_release(struct cafe_type_compiler *compiler)
{
	size_t i;
	struct cafe_list_head *item, *tmp;

	cafe_hashtbl_delete(compiler->strtbl, NULL, NULL, strtbl_delval, (void *)compiler->allocator);

	cafe_list_foreach_safe(item, &compiler->parser_list_head, tmp) {
		struct parser *p = PARSER_LIST_ENTRY(item);
		CAFE_FREE(p->compiler->allocator, p->filename);
		pt_node_free(p->root);
		cafe_list_remove(&p->list_item);
		strbuf_release(&p->lexer.sb);
		CAFE_FREE(compiler->allocator, p);
	}

	for (i = 0; i < CAFE_NELEMENTS(builtin_type_node); i++)
		pt_node_free(builtin_type_node[i]);

	scope_tree_release(compiler);
}

int register_builtin_types(struct cafe_hashtbl *strtbl, const struct cafe_allocator *allocator)
{
	size_t i;

	for (i = 0; i < CAFE_NELEMENTS(builtin_types); i++)
		builtin_types[i].istr = istr_intern(strtbl, builtin_types[i].name, strlen(builtin_types[i].name), allocator);

	return CAFE_OK;
}

static int register_builtin_types_in_global_scope(struct cafe_type_compiler *c)
{
	size_t i;

	for (i = 0; i < CAFE_NELEMENTS(builtin_types); i++) {
		int rc;
		struct symbol *symbol;
		struct istr *istr = builtin_types[i].istr;

		symbol = symbol_alloc(istr, builtin_types[i].type, NULL, c->global_scope, c->allocator);

		if (symbol == NULL)
			return CAFE_NOMEM;

		symbol->eval_type = builtin_types[i].type;
		symbol->node_def = pt_node_alloc(builtin_types[i].type, builtin_types[i].istr, NULL);
		builtin_type_node[i] = symbol->node_def;
		symbol->node_def->scope = c->global_scope;
		symbol->node_def->eval_type = builtin_types[i].type;
		rc = scope_add_symbol(c->global_scope, symbol);

		if (rc != CAFE_OK) {
			symbol_free(symbol, c->allocator);
			return rc;
		}
	}

	return CAFE_OK;
}

int compiler_init(struct cafe_type_compiler *c, error_handler error_handler, warning_handler warning_handler, const struct cafe_allocator *allocator)
{
	int rc;

	c->allocator = allocator;
	c->strtbl = cafe_hashtbl_new(1, allocator, istr_hashtbl_hash, istr_hashtbl_equals);

	if (c->strtbl == NULL)
		return CAFE_NOMEM;

	cafe_hashtbl_autoresize_enable(c->strtbl, 75);
	cafe_list_init(&c->scope_list_head);
	c->global_scope = scope_alloc(NULL, "global", c->allocator);

	if (c->global_scope == NULL) {
		cafe_hashtbl_delete(c->strtbl, NULL, NULL, NULL, NULL);
		return CAFE_NOMEM;
	}

	cafe_list_add_last(&c->global_scope->scope_list_item, &c->scope_list_head);
	cafe_hashtbl_autoresize_enable(c->global_scope->symbols, 75);

	if (error_handler == NULL)
		c->error_handler = compiler_default_error_handler;
	else
		c->error_handler = error_handler;

	if (warning_handler == NULL)
		c->warning_handler = compiler_default_warning_handler;
	else
		c->warning_handler = warning_handler;

	c->compilation_errors = 0;
	c->compilation_warnings = 0;
	cafe_list_init(&c->parser_list_head);
	rc = register_builtin_types(c->strtbl, allocator);

	if (rc != CAFE_OK)
		compiler_release(c);

	rc = register_builtin_types_in_global_scope(c);

	if (rc != CAFE_OK)
		compiler_release(c);


	return rc;
}

static int check_expr_is_constant(struct pt_node *n, pt_walk_order_t visit, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	//struct resolve_context *ctx = (struct resolve_context *)data;

	PT_TRACE(n, level, "check_expr_is_constant");
	assert(visit == PT_WALK_POSTORDER);

	switch (n->type) {
	case PT_FALSE:
	case PT_NUMBER:
	case PT_REAL:
	case PT_STRVAL:
	case PT_TRUE:
		return 1;
	case PT_AND:
	case PT_ARRAY:
	case PT_ASSIGNMENT:
	case PT_ATTRIBUTE:
	case PT_BLOB:
	case PT_BOOLEAN:
	case PT_CHAR:
	case PT_CONDITIONAL:
	case PT_CONST:
	case PT_DICT:
	case PT_DIVIDE:
	case PT_DOUBLE:
	case PT_EQ:
	case PT_FLOAT:
	case PT_GE:
	case PT_GT:
	case PT_IDENTIFIER:
	case PT_INT:
	case PT_KEY:
	case PT_LE:
	case PT_LIST:
	case PT_LONG:
	case PT_LONGLONG:
	case PT_LT:
	case PT_MINUS:
	case PT_MOD:
	case PT_NATURE:
	case PT_NATURE_REF:
	case PT_NE:
	case PT_NOT:
	case PT_NULL:
	case PT_OR:
	case PT_PLUS:
	case PT_ROOT:
	case PT_SETTING:
	case PT_SHORT:
	case PT_STRING:
	case PT_STRUCT:
	case PT_TIMES:
	case PT_TYPEDEF:
	case PT_UCHAR:
	case PT_UINT:
	case PT_ULONG:
	case PT_ULONGLONG:
	case PT_USHORT:
	case PT_UNDEFINED:
//		error(ctx->compiler, &n->loc, "only constants are supported in assignment statements (WRS_TODO)");
		return 1;
	}

	return 1;
}
