/*
 * Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <errno.h>
#include <cafe/cafe.h>
#include <cafe/list.h>
#include "dirname.h"
#include "pt.h"
#include "compiler.h"
#include "parser.h"

#define error compiler_error

static bool imported_filename_p(struct cafe_type_compiler *c, const char *filename)
{
	struct cafe_list_head *item;

	cafe_list_foreach(item, &c->parser_list_head) {
		struct parser *p = PARSER_LIST_ENTRY(item);

		if (p->filename == NULL)
			continue;

		if (strcmp(p->filename, filename) == 0)
			return true;
	}

	return false;
}

static int parse_imports(struct cafe_type_compiler *c, const char *filename_context, struct pt_node *n)
{
	size_t i;
	const char *ctxt = filename_context ? filename_context : ".";

	if (n == NULL)
		return CAFE_OK;

	assert(n->type == PT_LIST);

	for (i = 0; i < n->children.count; i++) {
		int rc;
		char *tmp_filename;
		char *dname;
		struct cafe_byte_buffer sb = STRBUF_INIT(c->allocator);

		tmp_filename = CAFE_STRDUP(c->allocator, ctxt);

		if (tmp_filename == NULL)
			return CAFE_NOMEM;

		dname = xdirname(tmp_filename);

		if (dname == NULL) {
			error(c, &n->children.list[i]->loc, "dirname error on '%s': %s", ctxt, strerror(errno));
			CAFE_FREE(c->allocator, tmp_filename);
			return CAFE_ERROR;
		}

		if (strcmp(dname, ".") == 0)
			rc = strbuf_fmt(&sb, "%s", n->children.list[i]->istr->str);
		else
			rc = strbuf_fmt(&sb, "%s/%s", dname, n->children.list[i]->istr->str);

		CAFE_FREE(c->allocator, tmp_filename);

		if (rc != CAFE_OK) {
			strbuf_release(&sb);
			return rc;
		}

		if (imported_filename_p(c, sb.data)) {
			compiler_warning(c, &n->children.list[i]->loc, "already imported '%s'", sb.data);
			strbuf_release(&sb);
			continue;
		}

		rc = parser_parsefile(c, sb.data);
		strbuf_release(&sb);

		if (rc != CAFE_OK)
			return rc;
	}

	return CAFE_OK;
}

struct parser *parser_alloc(struct cafe_type_compiler *compiler, const char *filename)
{
	struct parser *p;

	if ((p = CAFE_ALLOC(compiler->allocator, sizeof *p)) == NULL)
		return NULL;

	if (filename != NULL) {
		p->filename = CAFE_STRDUP(compiler->allocator, filename);
		if (p->filename == NULL) {
			CAFE_FREE(compiler->allocator, p);
			return NULL;
		}
	} else {
		p->filename = NULL;
	}

	p->compiler = compiler;
	p->root = NULL;
	memset(&p->lexer, 0, sizeof p->lexer);
	cafe_list_init(&p->list_item);

	return p;
}

int parser_parsefile(struct cafe_type_compiler *c, const char *filename)
{
	FILE *stream;
	void *lemon;
	struct parser *parser;
	bool lexer_trace = getenv("S") != NULL;

#ifndef NDEBUG
	if (getenv("L") != NULL)
		cafe_type_parserTrace(stderr, (char *)"LEMON: ");
#endif
	if ((lemon = cafe_type_parserAlloc(malloc)) == NULL)
		return CAFE_NOMEM;

	if (filename != NULL) {
		stream = fopen(filename, "rb");
		if (stream == NULL) {
			error(c, NULL, "cannot open \"%s\": %s", filename, strerror(errno));
			cafe_type_parserFree(lemon, free);
			return CAFE_ERROR;
		}
	} else {
		stream = stdin;
	}

	parser = parser_alloc(c, filename);

	if (parser == NULL) {
		if (stream != stdin)
			fclose(stream);
		cafe_type_parserFree(lemon, free);
		return CAFE_NOMEM;
	}

	lexer_init(&parser->lexer, lexer_default_error_handler, parser->filename, stream, c->strtbl, c->allocator);

	while (1) {
		int tkid = lexer_next_token(&parser->lexer);

		if (tkid == 0 || tkid == TK_EOF || parser->lexer.lexer_errors > 0)
			break;

		assert(parser->lexer.token.id > 0);

		if (lexer_trace) {
			fprintf(stdout, "LEXER: %s: "
				"line: %d, "
				"column: %d, "
				"name: %s",
				parser->lexer.token.loc.filename,
				parser->lexer.token.loc.lineno,
				parser->lexer.token.loc.column,
				parser->lexer.token.name);
			if (parser->lexer.token.istr != NULL)
				fprintf(stdout, ", value: %s", parser->lexer.token.istr->str);
			fprintf(stdout, "\n");
		}

		cafe_type_parser(lemon, parser->lexer.token.id, parser->lexer.token, parser);
	}

	if (stream != stdin)
		fclose(stream);

	cafe_type_parser(lemon, 0, parser->lexer.token, parser);
	cafe_type_parserFree(lemon, free);
	cafe_list_add_first(&parser->list_item, &c->parser_list_head);

	if (c->compilation_errors == 0 && parser->root != NULL && parser->root->left != NULL)
		parse_imports(c, filename, parser->root->left);

	return c->compilation_errors == 0 ? CAFE_OK : CAFE_ERROR;
}
