/* Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <cafe/cafe.h>
#include <cafe/array.h>
#include <cafe/ptr_array.h>
#include "pt.h"
#include "symtbl.h"
#include "bailout_allocator.h"

#define PRINT_LOC(STRM, LOC)					\
	do {							\
		fprintf(STRM, "%s%s%u:%u",			\
			(LOC).filename ? (LOC).filename : "",	\
			(LOC).filename ? ":" : "",		\
			(LOC).lineno, (LOC).column);		\
	} while (0)

static struct src_loc pt_unknown_location = {
	NULL, 0, 0
};

const char *pt_type_str(pt_type_t type)
{
	switch (type) {
	case PT_AND:		return "AND";
	case PT_ARRAY:		return "ARRAY";
	case PT_ASSIGNMENT:	return "ASSIGNMENT";
	case PT_ATTRIBUTE:	return "ATTRIBUTE";
	case PT_BLOB:		return "BLOB";
	case PT_BOOLEAN:	return "BOOLEAN";
	case PT_CHAR:		return "CHAR";
	case PT_CONDITIONAL:	return "CONDITIONAL";
	case PT_CONST:		return "CONST";
	case PT_DICT:		return "DICT";
	case PT_DIVIDE:		return "DIVIDE";
	case PT_DOUBLE:		return "DOUBLE";
	case PT_EQ:		return "EQ";
	case PT_FALSE:		return "FALSE";
	case PT_FLOAT:		return "FLOAT";
	case PT_GE:		return "GE";
	case PT_GT:		return "GT";
	case PT_IDENTIFIER:	return "IDENTIFIER";
	case PT_INT:		return "INT";
	case PT_KEY:		return "KEY";
	case PT_LE:		return "LE";
	case PT_LIST:		return "LIST";
	case PT_LONG:		return "LONG";
	case PT_LONGLONG:	return "LONGLONG";
	case PT_LT:		return "LT";
	case PT_MINUS:		return "MINUS";
	case PT_MOD:		return "MOD";
	case PT_NATURE:		return "NATURE";
	case PT_NATURE_REF:	return "NATURE_REF";
	case PT_NE:		return "NE";
	case PT_NOT:		return "NOT";
	case PT_NULL:		return "NULL";
	case PT_NUMBER:		return "NUMBER";
	case PT_OR:		return "OR";
	case PT_PLUS:		return "PLUS";
	case PT_REAL:		return "REAL";
	case PT_ROOT:		return "ROOT";
	case PT_SETTING:	return "SETTING";
	case PT_SHORT:		return "SHORT";
	case PT_STRING:		return "STRING";
	case PT_STRUCT:		return "STRUCT";
	case PT_STRVAL:		return "STRVAL";
	case PT_TIMES:		return "TIMES";
	case PT_TRUE:		return "TRUE";
	case PT_TYPEDEF:	return "TYPEDEF";
	case PT_UCHAR:		return "UCHAR";
	case PT_UINT:		return "UINT";
	case PT_ULONG:		return "ULONG";
	case PT_ULONGLONG:	return "ULONGLONG";
	case PT_UNDEFINED:	return "<undefined>";
	case PT_USHORT:		return "USHORT";
	}

	assert(0);
	return "<unknown_pt_node_type>";
}

/* WRS_TODO: use an allocator and check for failures. */

static void *xmalloc(size_t n)
{
	void *p = malloc(n);

	if (p == NULL) {
		fputs("Out of memory!\n", stderr);
		exit(1);
	}

	return p;
}

static void *xrealloc(void *ptr, size_t n)
{
	void *p = realloc(ptr, n);

	if (p == NULL) {
		fputs("Out of memory!\n", stderr);
		exit(1);
	}

	return p;
}

static void node_mashup(struct pt_node *n)
{
	if (n->type == PT_ATTRIBUTE) {
		n->u.attribute.node = n;
		n->u.attribute.qualifiers = n->left;
		n->u.attribute.typeref = n->middle;
		n->u.attribute.ident = n->right;
		n->u.attribute.assignment_expr = n->arg4;
		n->u.attribute.natures = n->arg5;
	} else if (n->type == PT_STRUCT) {
		n->u.structure.node = n;
		n->u.structure.ident = n->left;
		n->u.structure.natures = n->middle;
		n->u.structure.attributes = n->right;
		n->u.structure.documentation = n->arg4;
	} else if (n->type == PT_NATURE) {
		n->u.nature.node = n;
		n->u.nature.ident = n->left;
		n->u.nature.natures = n->middle;
		n->u.nature.attributes = n->right;
		n->u.nature.documentation = n->arg4;
	} else if (n->type == PT_TYPEDEF) {
		n->u.ztypedef.node = n;
		n->u.ztypedef.typeref = n->left;
		n->u.ztypedef.ident = n->middle;
		n->u.ztypedef.documentation = n->right;
	}
}

struct pt_node *pt_node_alloc(pt_type_t type, struct istr *istr, struct src_loc *loc)
{
	struct pt_node *n;

	n = xmalloc(sizeof *n);
	n->type = type;
	n->loc = (loc != NULL) ? *loc : pt_unknown_location;
	n->istr = istr;
	n->left = NULL;
	n->middle = NULL;
	n->right = NULL;
	n->arg4 = NULL;
	n->arg5 = NULL;
	n->children.count = 0;
	n->children.capacity = 0;
	n->children.list = NULL;
	n->symbol = NULL;
	n->scope = NULL;
	n->eval_type = PT_UNDEFINED;
	n->promote_to_type = PT_UNDEFINED;

	return n;
}

struct pt_node *pt_node_alloc0(pt_type_t type, struct istr *istr, struct src_loc *loc)
{
	return pt_node_alloc(type, istr, loc);
}

struct pt_node *pt_node_alloc1(pt_type_t type, struct istr *istr, struct src_loc *loc, struct pt_node *left)
{
	struct pt_node *n = pt_node_alloc(type, istr, loc);
	n->left = left;
	node_mashup(n);
	return n;
}

struct pt_node *pt_node_alloc2(pt_type_t type, struct istr *istr, struct src_loc *loc, struct pt_node *left, struct pt_node *middle)
{
	struct pt_node *n = pt_node_alloc(type, istr, loc);
	n->left = left;
	n->middle = middle;
	node_mashup(n);
	return n;
}

struct pt_node *pt_node_alloc3(pt_type_t type, struct istr *istr, struct src_loc *loc, struct pt_node *left, struct pt_node *middle, struct pt_node *right)
{
	struct pt_node *n = pt_node_alloc(type, istr, loc);
	n->left = left;
	n->middle = middle;
	n->right = right;
	node_mashup(n);
	return n;
}

struct pt_node *pt_node_alloc4(pt_type_t type, struct istr *istr, struct src_loc *loc, struct pt_node *left, struct pt_node *middle, struct pt_node *right, struct pt_node *arg4)
{
	struct pt_node *n = pt_node_alloc(type, istr, loc);
	n->left = left;
	n->middle = middle;
	n->right = right;
	n->arg4 = arg4;
	node_mashup(n);
	return n;
}

struct pt_node *pt_node_alloc5(pt_type_t type, struct istr *istr, struct src_loc *loc, struct pt_node *left, struct pt_node *middle, struct pt_node *right, struct pt_node *arg4, struct pt_node *arg5)
{
	struct pt_node *n = pt_node_alloc(type, istr, loc);
	n->left = left;
	n->middle = middle;
	n->right = right;
	n->arg4 = arg4;
	n->arg5 = arg5;
	node_mashup(n);
	return n;
}

struct pt_node *pt_node_list(struct pt_node *n)
{
	struct pt_node *p = pt_node_alloc(PT_LIST, NULL, NULL);
	p->children.count = p->children.capacity = 1;
	p->children.list = xmalloc(sizeof *p->children.list);
	p->children.list[0] = n;
	return p;
}

struct pt_node *pt_node_list_unshift(struct pt_node *n, struct pt_node *child)
{
	if (n->children.count == n->children.capacity) {
		size_t new_size = n->children.capacity * 2;
		void *new_ptr;

		if (new_size == 0)
			new_size = 1;

		new_ptr = xrealloc(n->children.list,
				   new_size * sizeof *n->children.list);
		n->children.capacity = new_size;
		n->children.list = new_ptr;
	}

	if (n->children.count > 0)
		memmove(&n->children.list[1], &n->children.list[0], (n->children.count) * sizeof *n->children.list);
	n->children.count++;
	n->children.list[0] = child;

	return n;
}

struct pt_node *pt_node_list_shift(struct pt_node *n, struct pt_node *child)
{
	if (n->children.count == n->children.capacity) {
		size_t new_size = n->children.capacity * 2;
		void *new_ptr;

		new_ptr = xrealloc(n->children.list,
				   new_size * sizeof *n->children.list);
		n->children.capacity = new_size;
		n->children.list = new_ptr;
	}

	n->children.list[n->children.count++] = child;

	return n;
}

struct pt_node *pt_binary(pt_type_t type, struct pt_node *left, struct pt_node *right, struct src_loc *loc, struct istr *istr)
{
	struct pt_node *n = pt_node_alloc(type, NULL, loc);
	n->istr = istr;
	n->left = left;
	n->right = right;
	return n;
}

struct pt_node *pt_unary(pt_type_t type, struct pt_node *left, struct src_loc *loc, struct istr *istr)
{
	struct pt_node *n = pt_node_alloc(type, NULL, loc);
	n->istr = istr;
	n->left = left;
	return n;
}

struct pt_node *pt_ternary(struct pt_node *left, struct pt_node *middle, struct pt_node *right, struct src_loc *loc)
{
	struct pt_node *n = pt_node_alloc(PT_CONDITIONAL, NULL, loc);
	n->left = left;
	n->middle = middle;	/* this will break expr evaluation */
	n->right = right;
	return n;
}

void pt_node_free(struct pt_node *n)
{
	size_t i;

	if (n == NULL)
		return;

	if (n->left != NULL)
		pt_node_free(n->left);

	if (n->middle != NULL)
		pt_node_free(n->middle);

	if (n->right != NULL)
		pt_node_free(n->right);

	if (n->arg4 != NULL)
		pt_node_free(n->arg4);

	if (n->arg5 != NULL)
		pt_node_free(n->arg5);

	if (n->children.list != NULL) {
		for (i = 0; i < n->children.count; i++)
			pt_node_free(n->children.list[i]);
		free(n->children.list);
	}

	free(n);
}

static int pt_recurse(struct pt_node *n, int descendants_only, pt_walk_action preorder_action, pt_walk_action postorder_action, int level, struct pt_walk_list_ctx *list_ctx, void *data)
{
	size_t i;
	int rc = 1;

	if (n == NULL)
		return 1;

	if (preorder_action != NULL) {
		if (!descendants_only) {
			rc = preorder_action(n, PT_WALK_PREORDER, level, list_ctx, data);
		}

		if (rc != 1)
			return rc;
	}

	if (rc == 1 && n->left != NULL) {
		rc = pt_recurse(n->left, 0, preorder_action, postorder_action, level + 1, list_ctx, data);
	}

	if (rc == 1 && n->middle != NULL) {
		rc = pt_recurse(n->middle, 0, preorder_action, postorder_action, level + 1, list_ctx, data);
	}

	if (rc == 1 && n->right != NULL) {
		rc = pt_recurse(n->right, 0, preorder_action, postorder_action, level + 1, list_ctx, data);
	}

	if (rc == 1 && n->arg4 != NULL) {
		rc = pt_recurse(n->arg4, 0, preorder_action, postorder_action, level + 1, list_ctx, data);
	}

	if (rc == 1 && n->arg5 != NULL) {
		rc = pt_recurse(n->arg5, 0, preorder_action, postorder_action, level + 1, list_ctx, data);
	}

	if (rc == 1 && n->children.list != NULL) {
		list_ctx->count = n->children.count;
		list_ctx->is_first = true;

		for (i = 0; i < n->children.count; i++) {
			list_ctx->is_last = i + 1 >= n->children.count;
			pt_recurse(n->children.list[i], 0, preorder_action, postorder_action, level + 1, list_ctx, data);
		}
	}

	if (rc == 1 && postorder_action != NULL) {
		if (!descendants_only) {
			rc = postorder_action(n, PT_WALK_POSTORDER, level, list_ctx, data);

			if (rc != 1)
				return rc;
		}
	}

	return 1;
}

int pt_walk(struct pt_node *n, pt_walk_action action, void *data)
{
	struct pt_walk_list_ctx ctx = { 0, 0, 0 };
	return pt_recurse(n, 0, action, action, 0, &ctx, data);
}

int pt_walk_preorder(struct pt_node *n, pt_walk_action action, void *data)
{
	struct pt_walk_list_ctx ctx = { 0, 0, 0 };
	return pt_recurse(n, 0, action, NULL, 0, &ctx, data);
}

int pt_walk_postorder(struct pt_node *n, pt_walk_action action, void *data)
{
	struct pt_walk_list_ctx ctx = { 0, 0, 0 };
	return pt_recurse(n, 0, NULL, action, 0, &ctx, data);
}

int pt_walk_descendants(struct pt_node *n, pt_walk_action action, void *data)
{
	struct pt_walk_list_ctx ctx = { 0, 0, 0 };
	return pt_recurse(n, 1, action, action, 0, &ctx, data);
}

int pt_walk_descendants_preorder(struct pt_node *n, pt_walk_action action, void *data)
{
	struct pt_walk_list_ctx ctx = { 0, 0, 0 };
	return pt_recurse(n, 1, action, NULL, 0, &ctx, data);
}

int pt_walk_descendants_postorder(struct pt_node *n, pt_walk_action action, void *data)
{
	struct pt_walk_list_ctx ctx = { 0, 0, 0 };
	return pt_recurse(n, 1, NULL, action, 0, &ctx, data);
}

int pt_is_leaf(const struct pt_node *n)
{
	return (n->left == NULL &&
		n->middle == NULL &&
		n->right == NULL &&
		n->arg4 == NULL &&
		n->arg5 == NULL &&
		(n->children.list == NULL || n->children.count == 0));
}

static void pt_print_node_str_escaped(FILE *stream, struct pt_node *n)
{
	size_t i = 0;

	if (n->type == PT_STRVAL) {
		fprintf(stream, "\"");

		for (i = 0; i < n->istr->len; i++) {
			switch (n->istr->str[i]) {
			case '\n':
				fputs("\\n", stream);
				break;
			case '\r':
				fputs("\\r", stream);
				break;
			case '\f':
				fputs("\\f", stream);
				break;
			case '\t':
				fputs("\\t", stream);
				break;
			case '\v':
				fputs("\\v", stream);
				break;
			case '"':
				fputs("\\\"", stream);
				break;
			default:
				fputc(n->istr->str[i], stream);
				break;
			}
		}
		fprintf(stream, "\"");
	} else {
		fprintf(stream, "%s", n->istr->str);
	}
}

static void pt_print_symbol_type(struct symbol *symbol, FILE *stream)
{
	struct symbol *type = symbol->symbol_type;
	int print_sep = 0;

	while (type != NULL) {
		fprintf(stream, "%s.%s[", type->scope->name, type->name->str);
		PRINT_LOC(stream, type->node_def->loc);
		type = type->symbol_type;
		print_sep = type != NULL;
		fprintf(stream, "]%s", print_sep ? " " : "");
	}
}

void pt_print_node(struct pt_node *n, FILE *stream, int indent)
{
	size_t i;

	if (n == NULL)
		return;

	if (indent >= 0)
		fprintf(stream, "%*s", -indent, "");
	else
		indent = -indent;

	fprintf(stream, "%s ", pt_type_str(n->type));

	if (n->loc.filename) {
		fputc('[', stream);
		PRINT_LOC(stream, n->loc);
		fputc(']', stream);
	}

	if (n->scope != NULL)
		fprintf(stream, " [scope=%s]", n->scope->name);

	if (n->eval_type != PT_UNDEFINED)
		fprintf(stream, " [eval_type=%s]", pt_type_str(n->eval_type));

	if (n->promote_to_type != PT_UNDEFINED)
		fprintf(stream, " [promote_to_type=%s]", pt_type_str(n->promote_to_type));

	fputc('\n', stream);

	if (n->istr != NULL) {
		fprintf(stream, "%*sstr: ", indent + 2, "");
		pt_print_node_str_escaped(stream, n);
		fputc('\n', stream);
	}

	if (n->symbol != NULL && n->symbol->symbol_type != NULL) {
		fprintf(stream, "%*stype: ", indent + 2, "");
		pt_print_symbol_type(n->symbol, stream);
		fputc('\n', stream);
	}

	if (n->left != NULL) {
		fprintf(stream, "%*sleft: ", indent + 2, "");
		pt_print_node(n->left, stream, -(indent + 4));
	}

	if (n->middle != NULL) {
		fprintf(stream, "%*smiddle: ", indent + 2, "");
		pt_print_node(n->middle, stream, -(indent + 4));
	}

	if (n->right != NULL) {
		fprintf(stream, "%*sright: ", indent + 2, "");
		pt_print_node(n->right, stream, -(indent + 4));
	}

	if (n->arg4 != NULL) {
		fprintf(stream, "%*sarg4: ", indent + 2, "");
		pt_print_node(n->arg4, stream, -(indent + 4));
	}

	if (n->arg5 != NULL) {
		fprintf(stream, "%*sarg5: ", indent + 2, "");
		pt_print_node(n->arg5, stream, -(indent + 4));
	}

	if (n->children.list != NULL) {
		fprintf(stream, "%*s[\n", indent + 2, "");
		for (i = 0; i < n->children.count; i++)
			pt_print_node(n->children.list[i], stream, indent + 4);
		fprintf(stream, "%*s]\n", indent + 2, "");
	}
}

void pt_splice_beginning(struct pt_node *dst, struct pt_node *src, unsigned int pos)
{
	size_t i;

	assert(src->type == PT_LIST);

	if (src->children.list != NULL) {
		for (i = 0; i < src->children.count; i++)
			pt_node_list_unshift(dst, src->children.list[i]);
		free(src->children.list);
		src->children.list = NULL;
		src->children.count = 0;
		src->children.capacity = 0;
	}
}
