/* Dynamic String Buffer */

/* Copyright (c) 2011,2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdio.h>
#include <stdarg.h>
#include <cafe/cafe.h>
#include <assert.h>
#include "strbuf.h"

/*
 * An ephemeral string which means it lasts for a very short time
 * (i.e., only valid until the buffer is grown).  This ensures that an
 * initialized strbuf is already NUL-terminated.
 */
char strbuf_ephemeral[1] = { '\0' };

void strbuf_init(struct cafe_byte_buffer *strbuf,
		 const struct cafe_allocator *allocator)
{
	assert(allocator != NULL);
	cafe_bbuf_init(strbuf, allocator, 0);
	strbuf->data = strbuf_ephemeral;
}

void strbuf_release(struct cafe_byte_buffer *strbuf)
{
	if (strbuf->data != strbuf_ephemeral)
		CAFE_FREE(strbuf->allocator, strbuf->data);
}

/* Take ownership of the data. */

char *strbuf_steal(struct cafe_byte_buffer *strbuf)
{
	char *result;

	assert(strbuf->data[strbuf->capacity - 1] == '\0');

	if (strbuf->len == 0)
		return NULL;

	result = strbuf->data;
	strbuf_init(strbuf, strbuf->allocator);

	return result;
}

void strbuf_reset(struct cafe_byte_buffer *strbuf)
{
	strbuf->len = 0;

	if (strbuf->data != strbuf_ephemeral) {
		assert(strbuf->capacity > 0);
		strbuf->data[0] = '\0';
	} else {
		assert(strbuf_ephemeral[0] == '\0');
	}
}

int strbuf_fmt(struct cafe_byte_buffer *strbuf, const char *fmt, ...)
{
	int rc;
	va_list ap;

	va_start(ap, fmt);
	rc = strbuf_vfmt(strbuf, fmt, ap);
	va_end(ap);

	return rc;
}

int strbuf_vfmt(struct cafe_byte_buffer *strbuf, const char *fmt, va_list ap)
{
	// Note: this va_copy may not be 100% valid.
#ifndef va_copy
# define va_copy(DST, SRC)	((DST) = (SRC))
#endif
	va_list ap_local;
	size_t size;
	int rc;
#if !defined(_WIN32)
	int n;
#endif
	char *write_head;

	va_copy(ap_local, ap);
#if defined(_WIN32)
	size = _vscprintf(fmt, ap_local);
#else
	size = vsnprintf(NULL, 0, fmt, ap_local);
#endif
	va_end(ap_local);

	if (size == 0)
		return CAFE_OK;
	if (strbuf->data == strbuf_ephemeral)
		strbuf->data = NULL;

	rc = cafe_bbuf_reserve(strbuf, size + 1, &write_head);
	if (rc != CAFE_OK)
		return rc;
	strbuf->data[strbuf->capacity - 1] = '\0';

#if defined(_WIN32)
	_vsnprintf(write_head, size + 1, fmt, ap);
#else
	n = vsnprintf(write_head, size + 1, fmt, ap);
	if (n > -1 && n < size)
		return CAFE_ERROR;
#endif
	--strbuf->len;

	return CAFE_OK;
}

int strbuf_addc(struct cafe_byte_buffer *strbuf, int c)
{
	int rc;

	if (strbuf->data == strbuf_ephemeral)
		strbuf->data = NULL;

	rc = cafe_bbuf_write(strbuf, c);
	if (rc != CAFE_OK)
		return rc;

	strbuf->data[strbuf->len] = '\0';

	return CAFE_OK;
}
