/* action_list.c - event action list allocation and locking */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
*/

#include <stdio.h>
#include <cafe/cafe.h>
#include <cafe/atomic.h>
#include "cafe_assert.h"

#define GEN_MASK			(0xfffff000)
#define REFCNT_MASK			(0x00000fff)
#define GEN_INC				(0x1000)

#define HANDLE_INDEX_MASK		REFCNT_MASK
#define HANDLE_INDEX(H)			((H) & HANDLE_INDEX_MASK)

struct tbl_entry {
	/* generation count + reference count */
	cafe_atomic_val gen_refcnt;
	struct cafe_event_action_list *list;
};

/* WRS_XXX this is too big.  How do we dynamically grow this? */
static struct tbl_entry table[HANDLE_INDEX_MASK];

static int eal_retain(const cafe_eal_handle_t *hp,
		      cafe_eal_handle_t *result_handle,
		      struct cafe_event_action_list **result_list)
{
	cafe_eal_handle_t h;
	struct tbl_entry *e;
	cafe_atomic_val old_gen_refcnt;
	cafe_atomic_val new_gen_refcnt;

	*result_list = NULL;
	*result_handle = CAFE_EAL_HANDLE_NULL;

reload_hp:
	h = cafe_atomic_get((cafe_atomic_val *)hp);

	if (HANDLE_INDEX(h) == CAFE_EAL_HANDLE_NULL)
		return CAFE_OK;

	e = &table[HANDLE_INDEX(h)];

	do {
		old_gen_refcnt = cafe_atomic_get(&e->gen_refcnt);

		/* If the result of XOR'ing the in-table generation
		 * count doesn't match the supplied handle's
		 * generation count then the handle we were passed is
		 * past its use by date. */

		if (((old_gen_refcnt ^ h) & GEN_MASK) != 0) {
			/* Maybe *hp was modified by another core. */
			if (cafe_atomic_get((cafe_atomic_val *)hp) != h)
				goto reload_hp;
			return CAFE_EAL_STALE_HANDLE;
		}

		new_gen_refcnt = (old_gen_refcnt + 1) & REFCNT_MASK;

		if (new_gen_refcnt == 0) {
			return CAFE_EAL_TOO_MANY_REFERENCES;
		}

		new_gen_refcnt |= h & GEN_MASK;
	} while (!cafe_atomic_cas(&e->gen_refcnt,
				  old_gen_refcnt, new_gen_refcnt));

	*result_list = e->list;
	*result_handle = h;

	return CAFE_OK;
}

static void eal_release(cafe_eal_handle_t h)
{
	struct tbl_entry *e;
	cafe_atomic_val old_gen_refcnt;
	cafe_atomic_val new_gen_refcnt;

	if (HANDLE_INDEX(h) == 0)
		return;

	e = &table[HANDLE_INDEX(h)];

	do {
		old_gen_refcnt = cafe_atomic_get(&e->gen_refcnt);

		if (((old_gen_refcnt ^ h) & GEN_MASK) != 0 ||
		    (old_gen_refcnt & REFCNT_MASK) == 0) {
			return;
		}

		new_gen_refcnt = old_gen_refcnt - 1;

		if ((new_gen_refcnt & REFCNT_MASK) == 0) {
			/* Bump the generation count to invalidate
			 * existing handles. */
			new_gen_refcnt += GEN_INC;
		}
	} while (!cafe_atomic_cas(&e->gen_refcnt,
				  old_gen_refcnt, new_gen_refcnt));
}

static int list_create(const struct cafe_ep_inst *inst,
		       struct cafe_event_descr *event,
		       struct cafe_event_action_list **result)
{
	struct cafe_event_action_list *actions;

	CAFE_ASSERT_LOCKED();
	assert(inst);
	assert(event);
	assert(event->num_types > 0);
	actions = CAFE_ALLOC(inst->allocator, sizeof *actions);

	if (actions == NULL)
		return CAFE_NOMEM;

	actions->allocator = inst->allocator;
	actions->event = event;
	actions->max = 0;
	actions->ind = 0;
	actions->list = NULL;
	*result = actions;
	return CAFE_OK;
}

static void list_delete(struct cafe_event_action_list *actions)
{
	if (actions != NULL) {
		CAFE_FREE(actions->allocator, actions->list);
		CAFE_FREE(actions->allocator, actions);
	}
}

int cafe_event_action_list_read_lock(
	const cafe_eal_handle_t *handle,
	cafe_eal_handle_t *unlock_handle,
	struct cafe_event_action_list **action_list_ptr)
{
	return eal_retain(handle, unlock_handle, action_list_ptr);
}

void cafe_event_action_list_read_unlock(cafe_eal_handle_t handle)
{
	eal_release(handle);
}

int cafe_event_action_list_alloc(
	const struct cafe_ep_inst *inst,
	struct cafe_event_descr *event,
	struct cafe_event_action_list **list_result,
	cafe_eal_handle_t *handle_result)
{
	size_t i;
	int rc;
	struct cafe_event_action_list *actions;

	assert(list_result != NULL);
	assert(handle_result != NULL);
	CAFE_ASSERT_LOCKED();

	*list_result = NULL;
	*handle_result = CAFE_EAL_HANDLE_NULL;
	rc = list_create(inst, event, &actions);

	if (rc != CAFE_OK)
		return rc;

	/* We start at 1 because slot 0 maps to CAFE_EAL_HANDLE_NULL. */
	for (i = 1; i < CAFE_NELEMENTS(table); ++i) {
		struct tbl_entry *e = &table[i];
		cafe_atomic_val old_gen_refcnt =
			cafe_atomic_get(&e->gen_refcnt);

		if ((old_gen_refcnt & REFCNT_MASK) != 0)
			continue;

		list_delete(e->list);
		e->list = actions;
		(void)cafe_atomic_set(&e->gen_refcnt, old_gen_refcnt + 1);
		*handle_result = (old_gen_refcnt & GEN_MASK) | i;
		break;
	}

	if (i == CAFE_NELEMENTS(table)) {
		*handle_result = CAFE_EAL_HANDLE_NULL;
		list_delete(actions);
		return CAFE_EAL_TABLE_FULL;
	}

	*list_result = actions;
	return CAFE_OK;
}

void cafe_event_action_list_release(cafe_eal_handle_t *handle)
{
	eal_release(*handle);
	*handle = CAFE_EAL_HANDLE_NULL;
}

void cafe_event_action_list_swap(cafe_eal_handle_t *a, cafe_eal_handle_t b)
{
	cafe_eal_handle_t old_handle;

	assert(sizeof *a == sizeof(cafe_atomic_val));
	old_handle = cafe_atomic_set((cafe_atomic_val *)a, b);
	eal_release(old_handle);
}

void cafe_event_action_list_gc(void)
{
	size_t i;

	CAFE_ASSERT_LOCKED();

	for (i = 1; i < CAFE_NELEMENTS(table); i++) {
		cafe_atomic_val gen_refcnt =
			cafe_atomic_get(&table[i].gen_refcnt);

		if ((gen_refcnt & REFCNT_MASK) != 0)
			continue;

		if (table[i].list != NULL) {
			list_delete(table[i].list);
			table[i].list = NULL;
			(void)cafe_atomic_val_add(&table[i].gen_refcnt, GEN_INC);
		}
	}
}
