/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
11mar14,sdt  Fix 'Checked Return' warnings


DESCRIPTION

*/

#include "cafe_os.h"
#include <string.h>		/* memset() */
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <cafe/cafe.h>
#include <cafe/cafe_print.h>
#include <cafe/errors.h>
#include <cafe/value_type.h>
#include <cafe/autogenerated/cafe.type.h>
#include "cafe_assert.h"
#include "cafe_trace.h"

#define SESSION_LIST_ENTRY(PTR)					\
	CAFE_LIST_ENTRY((PTR), struct cafe_instrument, session_list)

static struct cafe_list_head session_list_head =
	CAFE_LIST_INIT(session_list_head);

static cafe_instrument_delete_listener delete_listeners[16];
static int delete_listeners_cnt = 0;
long cafe_namespace_use_stdlib_allocator = 0;

#if !defined(__VXWORKS__)
static const char *parse_long(const char *fmt, long *result)
{
	char *str;

	errno = 0;
	*result = strtol(fmt, &str, 0);

	if (errno != 0 || fmt >= str)
		return NULL;

	return str;
}
#endif

static bool use_stdlib_allocator(void)
{
#if defined(__VXWORKS__)
	return (bool) cafe_namespace_use_stdlib_allocator;
#else
	char *env_val = getenv("CAFE_NAMESPACE_USE_STDLIB_ALLOCATOR");

	if (env_val != NULL) {
		long value;

		if (parse_long(env_val, &value) != NULL)
			cafe_namespace_use_stdlib_allocator = value;
	}

	return cafe_namespace_use_stdlib_allocator;
#endif
}

void cafe_instrument_add_delete_listener(
	cafe_instrument_delete_listener listener)
{
	CAFE_ASSERT_LOCKED();
	assert(delete_listeners_cnt < (int)(sizeof delete_listeners / sizeof *delete_listeners));
	delete_listeners[delete_listeners_cnt++] = listener;
}

static void notify_instrument_delete(struct cafe_instrument *session)
{
	int i;
	for (i = 0; i < delete_listeners_cnt; ++i)
		delete_listeners[i](session);
}

int cafe_instrument_create(
	const char *name,
	struct cafe_instrument **result)
{
	int rc;
	struct cafe_instrument *session;
	struct cafe_allocator *allocator;

	CAFE_ASSERT_LOCKED();
	assert(name);
	if (use_stdlib_allocator())
		allocator = cafe_stdlib_allocator;
	else
		allocator = cafe_allocator_create();
	if (allocator == NULL)
		return CAFE_NOMEM;
	*result = session = CAFE_ALLOC(allocator, sizeof *session);
	if (session == NULL)
		return CAFE_NOMEM;
	memset(session, 0, sizeof *session);

	session->allocator = allocator;
	session->name = CAFE_STRDUP(allocator, name);
	if (session->name == NULL) {
		CAFE_FREE(allocator, session);
		if (!use_stdlib_allocator())
			cafe_allocator_delete(allocator);
		*result = NULL;
		rc = CAFE_NOMEM;
		goto error;
	}

	rc = cafe_namespace_create(allocator, &session->ns);
	if (rc != CAFE_OK)
		goto error;

	rc = cafe_type_parse(session->ns, cafe_wildcard_VT_DEF,
			     &session->wildcard_type);
	if (rc != CAFE_OK)
		goto error;

	cafe_list_init(&session->inst_list_head);
	cafe_list_add_first(&session->session_list, &session_list_head);

	return CAFE_OK;
error:
	*result = NULL;
	cafe_namespace_delete(session->ns);
	CAFE_FREE(allocator, session->name);
	CAFE_FREE(allocator, session);
	if (!use_stdlib_allocator())
		cafe_allocator_delete(allocator);
	return rc;
}

int cafe_instrument_delete(struct cafe_instrument *session)
{
	int rc;
	unsigned int i;
	const struct cafe_allocator *allocator;

	CAFE_ASSERT_LOCKED();
	assert(session);
	(void)cafe_instrument_stop(session);
	rc = cafe_instrument_delete_inst(session);
	if (rc != CAFE_OK)
		return rc;
	if (!cafe_list_is_empty(&session->inst_list_head))
		return CAFE_SESSION_BUSY;
	notify_instrument_delete(session);
	allocator = session->allocator;
	for (i = 0; i < session->events.ind; i++) {
		CAFE_FREE(allocator, session->events.list[i]);
	}
	cafe_list_remove(&session->session_list);
	CAFE_FREE(allocator, session->events.list);
	cafe_namespace_delete(session->ns);
	CAFE_FREE(allocator, session->name);
	CAFE_FREE(allocator, session);
	if (!use_stdlib_allocator())
		cafe_allocator_delete(allocator);
	return CAFE_OK;
}

int cafe_instrument_start(struct cafe_instrument *session)
{
	struct cafe_list_head *list_item;

	CAFE_ASSERT_LOCKED();

	/* Start sinks first. */
	cafe_list_foreach_rev(list_item, &session->inst_list_head) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		int rc = cafe_ep_inst_start(inst);
		if (rc != CAFE_OK) {
			(void)cafe_instrument_stop(session);
			return rc;
		}
	}

	return CAFE_OK;
}

int cafe_instrument_stop(struct cafe_instrument *session)
{
	struct cafe_list_head *list_item;

	CAFE_ASSERT_LOCKED();
	assert(session);

	/* Stop root providers first. */
	cafe_list_foreach(list_item, &session->inst_list_head) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		int rc = cafe_ep_inst_stop(inst);

		if (rc != CAFE_OK) {
			cafe_error_set(NULL, rc, "Error stopping %s EP", inst->spec->name);
			return rc;
		}
	}
	return CAFE_OK;
}

void cafe_instrument_disconnect_inst(struct cafe_instrument *session)
{
	struct cafe_list_head *list_item;

	CAFE_ASSERT_LOCKED();
	assert(session);

	cafe_list_foreach(list_item, &session->inst_list_head) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		cafe_ep_inst_disconnect_all(inst);
	}
}

int cafe_instrument_delete_inst(struct cafe_instrument *session)
{
	int rc;
	struct cafe_list_head *list_item, *tmp;

	CAFE_ASSERT_LOCKED();
	assert(session);

	cafe_instrument_disconnect_inst(session);
	rc = cafe_instrument_refresh_output_event_actions(session);
	if (rc != CAFE_OK)
		return rc;

	/* WRS_TODO: need to wait for all action lists to be released
	 * before deleting instances. In the meantime we try and
	 * collect dead lists using the garbage collector after the
	 * instance delete callback has been called. */

	/* Delete instances in reverse order to make sure that lief
	 * contrinutions are removed before the instance it
	 * contributes to */

	/* Must use the _safe form of iteration as
	 * cafe_ep_inst_delete(), via inst_free(), removes this
	 * instance from the session list. */

	cafe_list_foreach_rev_safe(list_item, &session->inst_list_head, tmp) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		rc = cafe_ep_inst_delete(inst);
		if (rc != CAFE_OK)
			return rc;
	}

	cafe_event_action_list_gc();

	return CAFE_OK;
}


/**
 *  Assumes contributed types are never passed in the "types" parameter; that
 * it only contains the root event type optionally followed by arbitrary
 * orthogonal data types.
 */
int cafe_cfg_get_event(
	struct cafe_instrument *session,
	unsigned int num_types,
	struct cafe_value_type **types,
	struct cafe_event_descr **result)
{
	struct cafe_event_descr *event;
	unsigned int i;
	unsigned int j;

	CAFE_ASSERT_LOCKED();
	assert(session);
	assert(result);

	for (i = 0; i < session->events.ind; i++) {
		event = session->events.list[i];
		if (event->num_types != num_types)
			continue;
		for (j = 0; j < num_types; j++)
			if (event->types[j] != types[j])
				break;
		if (j >= num_types) {
			*result = event;
			return CAFE_OK;
		}
	}
	if (i == session->events.max) {
		if (i == 0) {
			session->events.max = 1;
		} else {
			session->events.max *= 2;
		}
		session->events.list = CAFE_REALLOC(
			session->allocator, session->events.list,
			session->events.max * sizeof *session->events.list);
		if (session->events.list == NULL)
			return CAFE_NOMEM;
	}
	event = CAFE_ALLOC(session->allocator, sizeof *event +
			   (num_types - 1) * sizeof *event->types);
	if (event == NULL)
		return CAFE_NOMEM;
	event->num_types = num_types;
	for (j = 0; j < num_types; j++)
		event->types[j] = types[j];
	session->events.list[i] = event;
	session->events.ind = i + 1;
	*result = event;
	return CAFE_OK;
}

struct cafe_instrument *cafe_instrument_iter_first(
	struct cafe_instrument_iter *iter)
{
	struct cafe_instrument *session;

	CAFE_ASSERT_LOCKED();
	assert(iter);

	if (cafe_list_is_empty(&session_list_head))
		return NULL;
	iter->current_session = session =
		SESSION_LIST_ENTRY(session_list_head.next);
	return session;
}

struct cafe_instrument *cafe_instrument_iter_next(
	struct cafe_instrument_iter *iter)
{
	struct cafe_instrument *session;

	CAFE_ASSERT_LOCKED();
	assert(iter);

	session = iter->current_session;

	if (session->session_list.next == &session_list_head)
		return NULL;
	iter->current_session = session =
		SESSION_LIST_ENTRY(session->session_list.next);
	return session;
}

int cafe_instrument_refresh_output_event_actions(
	struct cafe_instrument *session)
{
	struct cafe_list_head *list_item;

	CAFE_ASSERT_LOCKED();
	assert(session);

	cafe_list_foreach(list_item, &session->inst_list_head) {
		struct cafe_ep_inst *inst =
			SESSION_INST_LIST_ENTRY(list_item);
		int rc;

		if (!inst->spec->update_output_event_actions)
			continue;
		rc = inst->spec->update_output_event_actions(inst);
		if (rc != CAFE_OK)
			return rc;
	}
	return CAFE_OK;
}
