/* script_parser.c - Parser for cafe script files */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
11mar14,sdt  Fix 'Checked Return' warnings
24feb14,sdt  Disable unexpected NO_EFFECT warnings

DESCRIPTION
Parser for cafe script files

*/

#include "../src/cafe_os.h"
#include "../src/strtoll.h"
#include "../src/cafe_assert.h"
#include <cafe/ptr_array.h>
#include <cafe/array.h>
#include <cafe/stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <limits.h>
#include <cafe/cafe.h>
#include <cafe/errors.h>
#include <cafe/utils.h>
#include <cafe/ep/aggregator.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/script.h>

#define STR_REPR(A)	(((A) == NULL) ? "" : (A))

#if CAFE_OS_WINDOWS && !defined(PATH_MAX)
#define PATH_MAX	1024
#endif

#ifdef _WIN32
#define strtok_r(SRC, DELIM, PTR) strtok_s(SRC, DELIM, PTR)
#endif

#define TOK_EQL(T, X)  (strcmp((X), (T)->str) == 0)

#define CURR_FILE(SCRIPT) ((SCRIPT)->stream->bs_source((SCRIPT)->stream))
#define CURR_LINE(SCRIPT) ((SCRIPT)->stream->bs_lineno((SCRIPT)->stream))

#define EP_LIST_NODE(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct script_ep, list)

#define EP_EVENT_LIST_NODE(PTR)					\
	CAFE_LIST_ENTRY((PTR), struct script_event, list)

#define CONNECTION_LIST_NODE(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct script_connection, list)

#define EP_EDGE_SOURCE_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct script_ep_edge, source_list)

#define EP_EDGE_SINK_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct script_ep_edge, sink_list)

#define BYTE_STREAM_LIST_NODE_ENTRY(PTR)			\
	CAFE_LIST_ENTRY((PTR), struct byte_stream, list)

#define AGGR_TABLE_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct aggregator_table, list)

#define AGGR_ACTION_LIST_NODE_ENTRY(PTR)			\
	CAFE_LIST_ENTRY((PTR), struct aggregator_action, list)

#define AGGR_TYPE_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct aggregator_type, list)

#define TYPE_LIST_NODE_ENTRY(PTR)				\
	CAFE_LIST_ENTRY((PTR), struct type, list)

#ifdef __GNUC__
static void error(struct cafe_script *s, const char *file, int lineno, const char *fmt, ...) __attribute__((format(printf, 4, 5)));
static void error_append(struct cafe_script *s, const char *fmt, ...) __attribute__((format(printf, 2, 3)));
#endif

struct byte_stream {
	const char *(*bs_source)(struct byte_stream *);
	int (*bs_getc)(struct byte_stream *);
	int (*bs_ungetc)(int, struct byte_stream *);
	int (*bs_eof)(struct byte_stream *);
	int (*bs_lineno)(struct byte_stream *);
	ssize_t (*bs_getline)(const struct cafe_allocator *allocator,
			      char **strp, size_t *n, struct byte_stream *stream);
	int lineno;
	struct cafe_list_head list;
};

struct file_stream {
	struct byte_stream base; /* must come first */
	FILE *fp;
	char source[PATH_MAX];
};

struct string_stream {
	struct byte_stream base; /* must come first */
	char *str;
	char *curr;
};

enum ep_visited_state {
	NOT_VISITED   = 0,
	BEING_VISITED = 1,
	DONE_VISITED  = 2
};

struct token {
	char *str;		/* pointer to start of token */
	int lineno;		/* line number token declared on */
	const char *file;
};

struct script_setting {
	struct token name;
	struct token value;
	struct cafe_list_head list;
};

typedef enum {
	ENABLE_TYPE_REF = 1,
	ENABLE_EP = 2
} enable_type;

struct script_event {
	enable_type enable_type;
	unsigned int n_event_types;
	struct token event_type[CAFE_MAX_EVENT_TYPES];
	struct cafe_list_head list;
};

struct aggregator_action {
	struct token trigger;
	struct token filter;
	struct token cell;
	struct token operation;
	struct token args;
	struct cafe_list_head list;
};

struct aggregator_table {
	struct token name;
	struct cafe_list_head list;
};

struct type {
	struct token type;
	struct cafe_list_head list;
};

struct script_ep {
	enum ep_visited_state visited;
	struct token name;
	struct token alias;
	struct cafe_list_head list;
	struct cafe_list_head enabled_event_list_head;
	struct cafe_ep_spec *spec;
	struct cafe_ep_inst *inst;
	/* List head of predecessors. */
	struct cafe_list_head source_list_head;
	/* List head of successors. */
	struct cafe_list_head sink_list_head;
	struct cafe_value_type *workarea_type;
	struct cafe_vm_dynarray settings;
	struct {
		struct cafe_list_head action_list_head;
		struct cafe_list_head table_list_head;
	} aggr;
};

struct script_ep_edge {
	struct script_ep *source;
	struct script_ep *sink;
	struct cafe_list_head source_list;
	struct cafe_list_head sink_list;
};

struct script_connection {
	struct token src;
	struct token dst;
	struct script_ep *src_ep;
	struct script_ep *dst_ep;
	struct cafe_list_head list;
};

struct cafe_script {
	struct cafe_list_head ep_list_head;
	struct cafe_list_head ep_connection_list_head;
	struct cafe_list_head byte_stream_list_head;
	const struct cafe_allocator *allocator;
	struct cafe_script_error *err;
	unsigned int n_errors;
	struct cafe_instrument *session;
	struct script_ep *curr_ep;
	CAFE_SCRIPT_ERROR_HANDLER error_handler;
	const void *error_handler_client_data;
	char errbuf[512];	/* for cycle detection */
	struct byte_stream *stream;
	struct cafe_ptr_array tsorted_nodes;
	struct cafe_list_head type_list_head;
};

struct tokenizer {
	const struct cafe_allocator *allocator;
	char *wspace;		/* buffer for all words */
	char *word;		/* start of word */
	char *wptr;
	char *wspace_max;	/* upper bound of wspace */
	int quote;
	int comment;
	size_t argc;
	struct token argv[128];	/* WRS_XXX make this dynamic */
	struct byte_stream *stream;
};

static bool do_cmd(struct cafe_script *script,
		   size_t tokc, struct token tokv[],
		   int argc, const char *argv[]);

static int parse_from_file(struct cafe_script *script,
			   FILE *fp,
			   const char *filename,
			   int argc, const char *argv[]);

typedef bool (*CMD_HANDLER)(struct cafe_script *script,
			    size_t tokc, struct token tokv[]);

static int byte_stream_lineno(struct byte_stream *p)
{
	return p->lineno;
}

static ssize_t byte_stream_getline(const struct cafe_allocator *allocator,
				   char **linep, size_t *n, struct byte_stream *p)
{
	size_t len = 0;
	char *line;
	int c;

	assert(n);
	if (*linep == NULL || *n == 0) {
		*n = 1;
		errno = 0;
		if ((line = CAFE_ALLOC(allocator, *n)) == NULL)
			return -1;
		*linep = line;
	} else {
		line = *linep;
		assert(line);
	}

	line[len] = '\0';

	for (;;) {
		c = p->bs_getc(p);

		if (c == '\0')
			break;

		if (len + 1 >= *n) {
			size_t new_n = *n * 2 + 1;
			void *new_p;

			errno = 0;
			new_p = CAFE_REALLOC(allocator, line, new_n);
			if (new_p == NULL)
				return -1;
			line = *linep = new_p;
			*n = new_n;
		}

		line[len++] = (char) c;
		if (c == '\n')
			break;
	}
	line[len] = '\0';
	if (c != '\0')
		++p->lineno;
	return c == '\0' ? -1 : (ssize_t)len;
}

static inline int file_getc(struct byte_stream *p)
{
	return (p->bs_eof(p)) ? '\0' : fgetc_unlocked(((struct file_stream *)p)->fp);
}

static int file_ungetc(int c, struct byte_stream *p)
{
	return ungetc(c, ((struct file_stream *)p)->fp);
}

static inline int file_eof(struct byte_stream *p)
{
	return feof_unlocked(((struct file_stream *)p)->fp);
}

static const char *file_source(struct byte_stream *p)
{
	return ((struct file_stream *)p)->source;
}

static int string_stream_getc(struct byte_stream *p)
{
	return (p->bs_eof(p)) ? '\0' : *((struct string_stream *)p)->curr++;
}

static int string_stream_ungetc(int c, struct byte_stream *p)
{
	/*
	 * WRS_TODO: implement unget-buffer. Currently this only works for
	 * 1-deep ungetc's.
	 */
	struct string_stream *s = (struct string_stream *)p;

	if (c == '\0')
		return '\0';
	c = (unsigned char)c;
	*--s->curr = (char) c;
	return c;
}

static int string_stream_eof(struct byte_stream *p)
{
	return ((((struct string_stream *)p)->curr) == NULL ||
		*(((struct string_stream *)p)->curr) == '\0');
}

static const char *string_stream_source(struct byte_stream *p)
{
	return "";
}

static struct byte_stream *byte_stream_init_from_file(struct file_stream *s,
						      FILE *fp,
						      const char *filename)
{
	memset(s, 0, sizeof *s);
	s->base.bs_getc = file_getc;
	s->base.bs_ungetc = file_ungetc;
	s->base.bs_eof = file_eof;
	s->base.bs_lineno = byte_stream_lineno;
	s->base.bs_getline = byte_stream_getline;
	s->base.bs_source = file_source;
	s->fp = fp;
	s->source[0] = '\0';
	cafe_list_init(&s->base.list);
	if (filename != NULL)
		strncpy(s->source, filename, (sizeof(s->source) - 1));
	return &s->base;
}

static struct byte_stream *byte_stream_init_from_string(struct string_stream *s,
							char *str)
{
	memset(s, 0, sizeof *s);
	s->base.bs_getc = string_stream_getc;
	s->base.bs_ungetc = string_stream_ungetc;
	s->base.bs_eof = string_stream_eof;
	s->base.bs_lineno = byte_stream_lineno;
	s->base.bs_getline = byte_stream_getline;
	s->base.bs_source = string_stream_source;
	s->str = str;
	s->curr= s->str;
	cafe_list_init(&s->base.list);
	return &s->base;
}

static void error(struct cafe_script *s,
		  const char *file,
		  int lineno,
		  const char *fmt, ...)
{
	va_list ap;

	assert(file != NULL);
	s->n_errors++;
	va_start(ap, fmt);
	s->error_handler(s, s->error_handler_client_data,
			 file, lineno, fmt, ap);
	va_end(ap);
}

static void error_append(struct cafe_script *s, const char *fmt, ...)
{
	size_t len;

	len = strlen(s->errbuf);

	if (len +1 < sizeof s->errbuf) {
		va_list ap;
		va_start(ap, fmt);
		(void)vsnprintf(&s->errbuf[len], sizeof s->errbuf - len, fmt, ap);
		va_end(ap);
	}
}

static bool get_workarea_type(struct cafe_script *script,
			      struct script_ep *ep,
			      struct cafe_value_type **type)
{
	int rc;

	assert(ep->spec);

	if (ep->workarea_type != NULL) {
		*type = ep->workarea_type;
		return true;
	}

	if (ep->spec->types != NULL) {
		rc = cafe_type_parse(script->session->ns,
				     ep->spec->types, NULL);
		if (rc != CAFE_OK) {
			error(script,
			      CURR_FILE(script),
			      CURR_LINE(script),
			      "cannot parse types for EP '%s': %s",
			      ep->spec->name,
			      cafe_strerror(rc));
			return false;
		}
	}
	rc = cafe_type_parse(script->session->ns,
			     ep->spec->workarea,
			     &ep->workarea_type);
	if (rc != CAFE_OK) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "cannot parse workarea for EP '%s': %s",
		      ep->spec->name,
		      cafe_strerror(rc));
		return false;
	}
	*type = ep->workarea_type;
	return true;
}

static void default_error_handler(struct cafe_script *script,
				  const void *client_data,
				  const char *filename,
				  int lineno,
				  const char *fmt, va_list ap)
{
	if (lineno != -1) {
		(void)fprintf(stderr, "line %d: ", lineno);
	} else {
		(void)fputc(' ', stderr);
	}
	(void)vfprintf(stderr, fmt, ap);
	if (cafe_error_cur.rc != CAFE_OK) {
		(void)fputs(": ", stderr);
		cafe_error_print(NULL, stderr, ": ");
		cafe_error_clear(NULL);
	}
	(void)fputc('\n', stderr);
	(void)fflush(stderr);
}

static void *zmalloc(struct cafe_script *s, size_t n)
{
	void *p = CAFE_ALLOC(s->allocator, n);

	if (p == NULL) {
		error(s, CURR_FILE(s), -1, "out of memory!");
		return NULL;
	}
	memset(p, 0, n);
	return p;
}

static char *cp_str(struct cafe_script *s, const char *src)
{
	char *p;
	size_t len = strlen(src);

	if ((p = zmalloc(s, len + 1)) == NULL)
		return NULL;

	memcpy(p, src, len);
	p[len] = '\0';
	return p;
}

static char *cp_nstr(struct cafe_script *s, const char *src, size_t n)
{
	char *p;

	if ((p = zmalloc(s, n + 1)) == NULL)
		return NULL;

	memcpy(p, src, n);
	p[n] = '\0';
	return p;
}

static bool cp_token(struct cafe_script *script,
		     const struct token *src,
		     struct token *dst)
{
	if ((dst->str = cp_str(script, src->str)) == NULL)
		return false;
	dst->file = src->file;
	dst->lineno = src->lineno;
	assert(dst->file != NULL);
	return true;
}

static const char *ep_name(struct script_ep *ep)
{
	return (ep->alias.str != NULL) ? ep->alias.str : ep->name.str;
}

static bool grow_array_by_1(struct cafe_script *script,
			    struct cafe_vm_dynarray *array,
			    size_t element_size)
{
	if (array->l + 1 >= array->c) {
		char *newp;
		unsigned int newc;
		newc = array->c ? (unsigned int)array->c * 2 : 1;
		newp = CAFE_REALLOC(script->allocator, array->p, newc * element_size);
		if (newp == NULL) {
			error(script, CURR_FILE(script), -1, "out of memory!");
			return false;
		}
		memset(newp + array->c * element_size,
		       0, (newc - array->c) * element_size);
		array->c = newc;
		array->p = newp;
	}
	return true;
}

static struct type *new_type(struct cafe_script *s)
{
	struct type *p;

	if ((p = zmalloc(s, sizeof *p)) == NULL)
		return NULL;

	cafe_list_init(&p->list);

	return p;
}

static struct aggregator_action *new_aggr_action(struct cafe_script *s)
{
	struct aggregator_action *p;

	if ((p = zmalloc(s, sizeof *p)) == NULL)
		return NULL;

	cafe_list_init(&p->list);

	return p;
}

static struct aggregator_table *new_aggr_table(struct cafe_script *s)
{
	struct aggregator_table *p;

	if ((p = zmalloc(s, sizeof *p)) == NULL)
		return NULL;

	cafe_list_init(&p->list);

	return p;
}

static struct script_event *new_script_event(struct cafe_script *s)
{
	return zmalloc(s, sizeof(struct script_event));
}

static struct script_connection *new_script_connection(struct cafe_script *s)
{
	return zmalloc(s, sizeof(struct script_connection));
}

static struct script_ep *new_script_ep(struct cafe_script *s)
{
	struct script_ep *ep = zmalloc(s, sizeof(struct script_ep));

	if (ep != NULL) {
		cafe_list_init(&ep->enabled_event_list_head);
		cafe_list_init(&ep->source_list_head);
		cafe_list_init(&ep->sink_list_head);
		cafe_list_init(&ep->aggr.action_list_head);
		cafe_list_init(&ep->aggr.table_list_head);
	}

	return ep;
}

static const char *parse_ll(const char *fmt, long long *result)
{
	char *str;

	errno = 0;
	*result = strtoll(fmt, &str, 0);
	if (errno != 0 || fmt >= str)
		return NULL;
	return str;
}

static const char *parse_ull(const char *fmt, unsigned long long *result)
{
	char *str;

	errno = 0;
	*result = strtoull(fmt, &str, 0);
	if (errno != 0 || fmt >= str)
		return NULL;
	return str;
}

static struct script_ep *find_ep_by_alias(struct cafe_script *s,
					  const char *name)
{
	struct cafe_list_head *entry;

	cafe_list_foreach(entry, &s->ep_list_head) {
		struct script_ep *ep;

		ep = EP_LIST_NODE(entry);
		if (strcmp(ep->alias.str, name) == 0)
			return ep;
	}

	return NULL;
}

static void disconnect_edge(struct cafe_script *script,
			    struct script_ep_edge *edge)

{
	cafe_list_remove(&edge->source_list);
	cafe_list_remove(&edge->sink_list);
	CAFE_FREE(script->allocator, edge);
}

static void disconnect_all(struct cafe_script *script,
			   struct script_ep *ep)
{
	struct cafe_list_head *list_item, *tmp;

	cafe_list_foreach_safe(list_item, &ep->source_list_head, tmp) {
		struct script_ep_edge *edge;
		edge = EP_EDGE_SOURCE_LIST_NODE_ENTRY(list_item);
		disconnect_edge(script, edge);
	}

	cafe_list_foreach_safe(list_item, &ep->sink_list_head, tmp) {
		struct script_ep_edge *edge;
		edge = EP_EDGE_SINK_LIST_NODE_ENTRY(list_item);
		disconnect_edge(script, edge);
	}
}

static int connect_ep(struct cafe_script *script,
		      struct script_ep *source,
		      struct script_ep *sink)
{
	struct script_ep_edge *edge;

	assert(source);
	assert(sink);

	if ((edge = CAFE_ALLOC(script->allocator, sizeof *edge)) == NULL)
		return CAFE_NOMEM;

	edge->source = source;
	edge->sink = sink;
	cafe_list_add_last(&edge->sink_list, &source->sink_list_head);
	cafe_list_add_last(&edge->source_list, &sink->source_list_head);

	return CAFE_OK;
}

static struct script_ep *has_cycle(struct cafe_script *script,
				   struct script_ep *const ep)
{
	struct cafe_list_head *list_item;

	if (ep->visited == BEING_VISITED) {
		error_append(script, "connection cycle: %s", ep_name(ep));
		return ep;
	}

	if (ep->visited == DONE_VISITED)
		return NULL;

	ep->visited = BEING_VISITED;

	/* depth first search of all outgoing edges. */

	cafe_list_foreach(list_item, &ep->sink_list_head) {
		struct script_ep_edge *edge;
		struct script_ep *v2; /* vertex */

		edge = EP_EDGE_SINK_LIST_NODE_ENTRY(list_item);
		v2 = has_cycle(script, edge->sink);

		if (v2 != NULL) {
			error_append(script, " <- %s", ep_name(ep));
			return v2;
		}
	}

	ep->visited = DONE_VISITED;
	return NULL;
}

static bool graph_has_cycles(struct cafe_script *script)
{
	struct cafe_list_head *entry;

	cafe_list_foreach(entry, &script->ep_list_head) {
		struct script_ep *ep;

		ep = EP_LIST_NODE(entry);

		if (ep->visited == NOT_VISITED && has_cycle(script, ep)) {
			return true;
		}
	}

	return false;
}

/* This will only handle simple key/value pairs. */

static bool parse_set(struct cafe_script *script,
		      size_t tokc, struct token tokv[])
{
	struct cafe_cfg_setting *settings;
	struct cafe_value_type *workarea_type;
	struct cafe_vm_attr *attr;
	long long val;
	unsigned long long uval;
	cafe_any_value *value;

	if (script->curr_ep == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "settings must follow their "
		      "respective 'create' statements");
		return false;
	}

	if (tokc < 2) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'set'");
		return false;
	}

	if (script->curr_ep->spec->workarea == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "%s does not have any settings",
		      script->curr_ep->spec->name);
		return false;
	}

	if (!get_workarea_type(script, script->curr_ep, &workarea_type))
		return false;

	if (!grow_array_by_1(script, &script->curr_ep->settings,
			     sizeof(struct cafe_cfg_setting)))
		return false;

	attr = cafe_type_attr_by_name(script->curr_ep->workarea_type, tokv[0].str);

	if (attr == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "'%s' has no field named '%s'",
		      script->curr_ep->spec->name,
		      tokv[0].str);
		return false;
	}

	if (cafe_find_nature_str(attr->ident, cafe_setting_VT_DEF) == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "'%s.%s' has no nature (%s)",
		      script->curr_ep->spec->name,
		      tokv[0].str,
		      cafe_setting_VT_NAT);
		return false;
	}

	settings = script->curr_ep->settings.p;
	settings[script->curr_ep->settings.l].name = cp_str(script, tokv[0].str);
	value = &settings[script->curr_ep->settings.l].value;
	script->curr_ep->settings.l++;

	switch (cafe_attr_type(attr)->kind) {
	case CAFE_TYPE_UCHAR:
	case CAFE_TYPE_USHORT:
	case CAFE_TYPE_UINT:
	case CAFE_TYPE_ULONG:
	case CAFE_TYPE_ULONGLONG:
		if (!parse_ull(tokv[1].str, &uval)) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "'%s' is not a number",
			      tokv[1].str);
			return false;
		}
		value->ull = uval;
		break;
	case CAFE_TYPE_CHAR:
	case CAFE_TYPE_SHORT:
	case CAFE_TYPE_INT:
	case CAFE_TYPE_LONG:
	case CAFE_TYPE_LONGLONG:
		if (!parse_ll(tokv[1].str, &val)) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "'%s' is not a number",
			      tokv[1].str);
			return false;
		}
		value->ll = val;
		break;
	case CAFE_TYPE_BOOL:
		if (!parse_ull(tokv[1].str, &uval)) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "'%s' is not a valid number",
			      tokv[1].str);
			return false;
		}
		value->ull = (bool)uval;
		break;
	case CAFE_TYPE_STRING:
		value->str = cp_str(script, tokv[1].str);
		break;
	default:
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "cannot handle the type"
		      " of '%s.%s' (implement me @%s:%d)",
		      script->curr_ep->spec->name,
		      tokv[0].str,
		      __FILE__, __LINE__);
		return false;
	}

	return true;
}

static bool parse_enable(struct cafe_script *script,
			 size_t tokc, struct token tokv[])
{
	size_t i;
	struct script_event *p;
	enable_type type;

	if (tokc < 1) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'enable'");
		return false;
	}

	if (script->curr_ep == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "enable statements must follow "
		      "their respective 'create' statements");
		return false;
	}

	type = TOK_EQL(&tokv[0], "for") ? ENABLE_EP : ENABLE_TYPE_REF;

	if ((p = new_script_event(script)) == NULL)
		return false;

	p->enable_type = type;
	cafe_list_add_last(&p->list,
			   &script->curr_ep->enabled_event_list_head);

	for (i = (type == ENABLE_EP) ? 1 : 0; i < tokc; i++) {
		if (p->n_event_types == CAFE_NELEMENTS(p->event_type)) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "maximum number of types is %d",
			      (int)CAFE_NELEMENTS(p->event_type));
			return false;
		}
		if (!cp_token(script, &tokv[i],
			      &p->event_type[p->n_event_types]))
			return false;
		p->n_event_types++;
	}

	return true;
}

static bool parse_connect(struct cafe_script *script,
			  size_t tokc, struct token tokv[])
{
	struct script_connection *p;

	if (tokc != 3) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'connect'");
		return false;
	}

	if ((p = new_script_connection(script)) == NULL)
		return false;

	cafe_list_add_last(&p->list, &script->ep_connection_list_head);

	if (!cp_token(script, &tokv[0], &p->src))
		return false;

	if (!TOK_EQL(&tokv[1], "to")) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "unexpected token '%s', expected 'to'",
		      tokv[1].str);
		return false;
	}

	return cp_token(script, &tokv[2], &p->dst);
}

static int is_valid_identifier(int c)
{
	return ((c >= 'A' && c <= 'Z') ||
		(c >= 'a' && c <= 'z') ||
		(c >= '0' && c <= '9') || (c == '_'));
}

static void tokenizer_init(struct tokenizer *t,
			   struct byte_stream *stream,
			   struct cafe_script *script)
{
	t->allocator = script->allocator;
	t->wspace = NULL;	/* buffer for all accumulated words */
	t->word = NULL;		/* beginning of word */
	t->wptr = NULL;
	t->wspace_max = NULL;
	t->wptr = NULL;
	t->quote = 0;
	t->comment = 0;
	t->argc = 0;
	memset(t->argv, 0, sizeof t->argv);
	t->stream = stream;

}

static void tokenizer_reset(struct tokenizer *t)
{
	t->word = t->wspace;	/* beginning of word */
	t->wptr = t->wspace;
	t->quote = 0;
	t->comment = 0;
	t->argc = 0;
	memset(t->argv, 0, sizeof t->argv);
}

static void tokenizer_finish_word(struct tokenizer *t)
{
	*t->wptr = '\0';	/* terminate current word */

	if (t->wptr != t->word) {
		/* words must have at least 1 character. */
		t->argv[t->argc].str = t->word;
		t->argv[t->argc].file = t->stream->bs_source(t->stream);
		t->argv[t->argc].lineno = t->stream->bs_lineno(t->stream);
		t->argc++;
		t->word = ++t->wptr;
	}
}

static void tokenizer_release(struct tokenizer *t)
{
	CAFE_FREE(t->allocator, t->wspace);
}

/*
 * Returns:  0 success
 *          -1 on memory error
 *          -2 quoted word not terminated
 *          -3 escaped character is illegal
 *
 * This function assumes the input line is terminated by a [\r]\n.
 */
static int tokenize(struct cafe_script *script,
		    struct tokenizer *t,
		    const char *input)
{
	for (;;) {
		int c = *input++;

		if (t->wptr >= t->wspace_max) {
			size_t i;
			char *newp;
			size_t newsz;
			size_t oldsz = t->wspace_max - t->wspace;

			newsz = (oldsz == 0) ? 1 : oldsz * 2;
			newp = CAFE_REALLOC(script->allocator, t->wspace, newsz);
			if (newp == NULL) {
				error(script, CURR_FILE(script), -1, "out of memory!");
				return -1;
			}
			/* adjust existing word pointers. */
			for (i = 0; i < t->argc; i++) {
				t->argv[i].str = newp + (t->argv[i].str - t->wspace);
			}
			t->wptr = newp + (t->word - t->wspace) + (t->wptr - t->word);
			t->word = newp + (t->word - t->wspace);
			t->wspace_max = newp + newsz;
			t->wspace = newp;
		}

		if (c == '\r' || c == '\n' || c == '\0') {
			if (t->quote)
				return -2; /* quote not terminated */
			tokenizer_finish_word(t);
			return 0;
		}

		if (t->comment)
			continue;

		if (isspace(c) && !t->quote) {
			tokenizer_finish_word(t);
			continue;
		}

		if (!t->quote) {
			if (c == '#') {
				t->comment = 1;
				continue;
			}
		}

		if (c == '\\') {
			c = *input++;
			switch (c) {
			case '\r':
			case '\n':
				*t->wptr++ = '\\';
				continue;
			case '\\':
			case '"':
				break;
			default:
				return -3;
			}
			*t->wptr++ = (char) c;
			continue;
		}

		/* This is a hack to allow an input of "" to be
		 * treated as an empty token.  We have this case where
		 * we do something like:
		 *
		 *     action cafe_timer.timer "" snapshot
		 *
		 * Normal (POSIX) shell quoting rules would not make
		 * "" into a discrete token but we force the issue
		 * here rather than change the accepted practice and
		 * the existing scripts.
		 */
		if (t->quote == 0 && c == '"' &&
		    input[0] == '"' && isspace(input[1])) {
			*t->wptr++ = '\0';
		}

		if (c == '"') {
			t->quote = 1 - t->quote;
			continue;
		}

		*t->wptr++ = (char) c;
	}
}

static int parse_from_stream(struct cafe_script *script,
			     struct byte_stream *stream,
			     int argc, const char *argv[])
{
	char *line = NULL;
	size_t linesz = 0;
	ssize_t nbytes;
	struct byte_stream *last_stream = script->stream;
	struct tokenizer tokenizer;

	script->stream = stream;
	tokenizer_init(&tokenizer, script->stream, script);

	while ((nbytes = stream->bs_getline(script->allocator,
					    &line, &linesz,
					    stream)) != -1) {
		size_t len;
		int rc;

		rc = tokenize(script, &tokenizer, line);

		if (rc < 0) {
			const char *msg[] = {
				"success",
				"out of memory!",
				"quoted word not terminated",
				"illegal escaped character"
			};

			error(script, CURR_FILE(script), CURR_LINE(script),
			      "%s", msg[-rc]);
		}

		if (tokenizer.argc == 0) {
			tokenizer_reset(&tokenizer);
			continue;
		}

		len = strlen(tokenizer.argv[tokenizer.argc-1].str);

		if (tokenizer.argv[tokenizer.argc-1].str[len-1] == '\\') {
			tokenizer.argv[tokenizer.argc-1].str[len-1] = '\0';
			tokenizer.argc--;
			continue;
		}

		if (!do_cmd(script, tokenizer.argc, tokenizer.argv, argc, argv))
			break;

		if (script->n_errors > 0)
			break;

		tokenizer_reset(&tokenizer);
	}

	tokenizer_release(&tokenizer);
	CAFE_FREE(script->allocator, line);

	if (last_stream != NULL)
		script->stream = last_stream;

	return (script->n_errors > 0) ? CAFE_ERROR : CAFE_OK;
}

static int parse_from_string(struct cafe_script *script,
			     const char *content,
			     int argc, const char *argv[])
{
	struct string_stream *string_stream;
	struct byte_stream *stream;
	char *tmp_str;
	int rc;

	if ((tmp_str = cp_str(script, content)) == NULL) {
		return CAFE_NOMEM;
	}

	string_stream = CAFE_ALLOC(script->allocator, sizeof *string_stream);

	if (string_stream == NULL)
		return CAFE_NOMEM;

	stream = byte_stream_init_from_string(string_stream, tmp_str);
	cafe_list_add_last(&stream->list, &script->byte_stream_list_head);
	rc = parse_from_stream(script, stream, argc, argv);
	CAFE_FREE(script->allocator, tmp_str);
	return rc;
}

static int parse_from_file(struct cafe_script *script,
			   FILE *fp,
			   const char *filename,
			   int argc, const char *argv[])
{
	FILE *saved_fp = fp;
	int rc;
	struct file_stream *file_stream;
	struct byte_stream *stream;
#if CAFE_OS_UNIX
	char saved_working_dir[PATH_MAX + 1] = { '\0' };
	char tmp_filename[PATH_MAX + 1];
	char *dname;
	extern char *dirname(char *path);
#endif
	if (fp == NULL) {
		if ((fp = fopen(filename, "r")) == NULL) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "cannot open '%s': %s",
			      filename, strerror(errno));
			return CAFE_ERROR;
		}
	}

	file_stream = (struct file_stream *)CAFE_ALLOC(script->allocator, sizeof *file_stream);

	if (file_stream == NULL) {
		if (saved_fp == NULL)
			(void)fclose(fp);
		return CAFE_NOMEM;
	}

#if CAFE_OS_UNIX
	if (getcwd(saved_working_dir, PATH_MAX) == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "cannot get current working directory: %s",
		      strerror(errno));
		return CAFE_ERROR;
	}

	strncpy(tmp_filename, filename, PATH_MAX);

	if ((dname = dirname(tmp_filename)) == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "dirname() error: %s", strerror(errno));
		return CAFE_ERROR;
	}
#endif
#if CAFE_OS_UNIX
	if (chdir(dname) != 0) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "chdir failed: %s", strerror(errno));
		return CAFE_ERROR;
	}
#endif
	stream = byte_stream_init_from_file(file_stream, fp, filename);
	cafe_list_add_last(&stream->list, &script->byte_stream_list_head);
	rc = parse_from_stream(script, stream, argc, argv);
#if CAFE_OS_UNIX
	if (chdir(saved_working_dir) != 0) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "chdir failed: %s", strerror(errno));
		if (saved_fp == NULL)
			fclose(fp);
		return CAFE_ERROR;
	}
#endif
	if (saved_fp == NULL)	/* incoming was NULL */
		(void)fclose(fp);	/* then close locally opened file. */

	return rc;
}

static bool token_is_valid_identifier(struct token *tok)
{
	size_t i;
	size_t n = strlen(tok->str);

	for (i = 0; i < n; i++)
		if (!is_valid_identifier(tok->str[i]))
			return false;

	return true;
}

static bool parse_create(struct cafe_script *script,
			 size_t tokc, struct token tokv[])
{
	struct script_ep *ep, *duplicate_ep;

	if (tokc < 1) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'create'");
		return false;
	}

	if ((ep = new_script_ep(script)) == NULL)
		return false;

	script->curr_ep = ep;
	cafe_list_add_last(&script->curr_ep->list, &script->ep_list_head);

	if (!token_is_valid_identifier(&tokv[0])) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "'%s' is not a valid identifier",
		      tokv[0].str);
		return false;
	}

	if (!cp_token(script, &tokv[0], &ep->name))
		return false;

	/* the AS clause is optional. */

	if (tokc > 1) {
		if (!TOK_EQL(&tokv[1], "as")) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "unexpected token '%s', expected 'as'",
			      tokv[1].str);
			return false;
		}
		if (tokc < 3) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "missing argument after 'as'");
			return false;
		}
		if (!cp_token(script, &tokv[2], &ep->alias))
			return false;
	} else {
		if (!cp_token(script, &tokv[0], &ep->alias))
			return false;
	}

	/* Check for duplicate create statements which aren't
	 * disambiguated by an AS alias. */

	duplicate_ep = find_ep_by_alias(script, ep->alias.str);

	if (duplicate_ep != NULL && duplicate_ep != script->curr_ep) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "duplicate create statement for '%s'; "
		      "previously defined at %s:%d",
		      ep_name(ep),
		      duplicate_ep->alias.file,
		      duplicate_ep->alias.lineno);
		return false;
	}

	ep->spec = cafe_ep_spec_find(ep->name.str);

	if (ep->spec == NULL) {
		error(script, ep->name.file, ep->name.lineno,
		      "event processor '%s' not registered",
		      ep->name.str);
		return false;
	}

	return true;
}

static bool parse_aggregator_table(struct cafe_script *script,
				   size_t tokc, struct token tokv[])
{
	int rc;
	const char *name = "add_tables";
	size_t i;
	struct cafe_cfg_setting *setting;
	struct cafe_vm_attr *attr;
	struct cafe_value_type *workarea_type;
	cafe_any_value *value;
	struct cafe_value_type *attr_type;
	void *element_data;
	struct aggregator_table *aggr_table;

	if (tokc < 1) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'table'");
		return false;
	}

	if (script->curr_ep == NULL ||
	    strcmp(script->curr_ep->name.str, CAFE_AGGREGATOR_EP_NAME) != 0) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "table names must follow a "
		      "'create aggregator' statement");
		return false;
	}

	if (script->curr_ep->spec->workarea == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "%s does not have any settings",
		      script->curr_ep->spec->name);
		return false;
	}

	if (!get_workarea_type(script, script->curr_ep, &workarea_type))
		return false;

	attr = cafe_type_attr_by_name(script->curr_ep->workarea_type, name);

	if (attr == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "'%s' has no field named '%s'",
		      script->curr_ep->spec->name, name);
		return false;
	}

	attr_type = cafe_attr_type(attr);
	assert(attr_type->kind == CAFE_TYPE_DYNARRAY);
	setting = script->curr_ep->settings.p;

	for (i = 0; i < script->curr_ep->settings.l; i++) {
		if (strcmp(name, setting[i].name) == 0)
			break;
	}
	if (i == script->curr_ep->settings.l) {
		if (!grow_array_by_1(script, &script->curr_ep->settings,
				     sizeof(struct cafe_cfg_setting)))
			return false;
		setting = script->curr_ep->settings.p;
		setting[script->curr_ep->settings.l].name = cp_str(script, name);
		assert(i == script->curr_ep->settings.l);
		script->curr_ep->settings.l++;
	}

	value = &setting[i].value;
	rc =  cafe_vm_dynarray_elem_get(&value->dary, attr_type,
					value->dary.l, &element_data);
	if (rc != CAFE_OK) {
		error(script, CURR_FILE(script), CURR_LINE(script), "%s",
		      cafe_strerror(rc));
		return false;
	}
	rc = cafe_value_store_string(element_data, tokv[0].str,
				     script->session->ns->allocator);
	if (rc != CAFE_OK) {
		error(script, CURR_FILE(script), CURR_LINE(script), "%s",
		      cafe_strerror(rc));
		return false;
	}

	if ((aggr_table = new_aggr_table(script)) == NULL)
		return false;

	cafe_list_add_last(&aggr_table->list,
			   &script->curr_ep->aggr.table_list_head);

	if (!cp_token(script, tokv, &aggr_table->name))
		return false;

	return true;
}

static bool parse_aggregator_action(struct cafe_script *script,
				    size_t tokc, struct token tokv[])
{
	int rc;
	const char *name = "add_actions";
	size_t i;
	struct cafe_cfg_setting *setting;
	struct cafe_vm_attr *attr;
	struct cafe_value_type *workarea_type;
	cafe_any_value *value;
	struct cafe_value_type *attr_type;
	void *element_data;
	size_t n = 0;
	size_t remaining = tokc;
	struct token action[5];
	struct cafe_value_type *array_type;
	struct aggregator_action *aggr_action;

	if (tokc < 3) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'action'");
		return false;
	}

	if (script->curr_ep == NULL ||
	    strcmp(script->curr_ep->name.str, CAFE_AGGREGATOR_EP_NAME) != 0) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "aggregator actions must follow a "
		      "'create aggregator' statement");
		return false;
	}

	memset(&action, 0, sizeof action);

	if (TOK_EQL(&tokv[0], "-f")) { /* optional */
		assert(tokc >= 2);
		action[1] = tokv[1];
		remaining -= 2;
		n = 2;
	}

	if (remaining < 3) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "missing arguments for 'action'");
		return false;
	}

	/* Syntax is: ["-f" <filter>] <event_type-type> <cell> <op> [<op-args>] */

	action[0] = tokv[0+n];
	action[2] = tokv[1+n];
	action[3] = tokv[2+n];

	if (remaining - 3)
		action[4] = tokv[3+n];

	if (!get_workarea_type(script, script->curr_ep, &workarea_type))
		return false;

	attr = cafe_type_attr_by_name(script->curr_ep->workarea_type, name);

	if (attr == NULL) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "'%s' has no field named '%s'",
		      script->curr_ep->spec->name, name);
		return false;
	}

	attr_type = cafe_attr_type(attr);
	assert(attr_type->kind == CAFE_TYPE_DYNARRAY);
	setting = script->curr_ep->settings.p;

	for (i = 0; i < script->curr_ep->settings.l; i++) {
		if (strcmp(name, setting[i].name) == 0)
			break;
	}

	if (i == script->curr_ep->settings.l) {
		if (!grow_array_by_1(script, &script->curr_ep->settings,
				     sizeof(struct cafe_cfg_setting)))
			return false;
		setting = script->curr_ep->settings.p;
		setting[script->curr_ep->settings.l].name = cp_str(script, name);
		assert(i == script->curr_ep->settings.l);
		script->curr_ep->settings.l++;
	}

	value = &setting[i].value;
	rc = cafe_vm_dynarray_elem_get(&value->dary, attr_type,
				       value->dary.l, &element_data);
	if (rc != CAFE_OK) {
		error(script, CURR_FILE(script), CURR_LINE(script), "%s",
		      cafe_strerror(rc));
		return false;
	}
	value = element_data;
	array_type = cafe_element_type(attr_type);

	for (i = 0; i < CAFE_NELEMENTS(action); i++) {
		rc = cafe_vm_dynarray_elem_get(&value->dary,
					       array_type, i,
					       &element_data);
		if (rc != CAFE_OK) {
			error(script, CURR_FILE(script), CURR_LINE(script), "%s",
			      cafe_strerror(rc));
			return false;
		}
		rc = cafe_value_store_string(element_data,
					     action[i].str ? action[i].str : "",
					     script->session->ns->allocator);
		if (rc != CAFE_OK) {
			error(script, CURR_FILE(script), CURR_LINE(script), "%s",
			      cafe_strerror(rc));
			return false;
		}
	}

	if ((aggr_action = new_aggr_action(script)) == NULL)
		return false;

	cafe_list_add_last(&aggr_action->list,
			   &script->curr_ep->aggr.action_list_head);

	if (action[0].str != NULL) {
		if (!cp_token(script, &action[0], &aggr_action->trigger))
			return false;
	}
	if (action[1].str != NULL) {
		if (!cp_token(script, &action[1], &aggr_action->filter))
			return false;
	}
	if (action[2].str != NULL) {
		char *s;
		char t[1] = { '\0' };

		s = (strcmp(action[2].str, "-") == 0) ? t : action[2].str;

		if ((s = cp_str(script, s)) == NULL)
			return false;

		aggr_action->cell.str = s;
	}
	if (action[3].str != NULL) {
		if (!cp_token(script, &action[3], &aggr_action->operation))
			return false;
	}
	if (action[4].str != NULL) {
		if (!cp_token(script, &action[4], &aggr_action->args))
			return false;
	}

	return true;
}

static bool parse_typedef(struct cafe_script *script,
			  size_t tokc, struct token tokv[])
{
	struct type *type;

	if (tokc < 1) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'typedef'");
		return false;
	}

	if ((type = new_type(script)) == NULL)
		return false;

	cafe_list_add_last(&type->list, &script->type_list_head);

	return cp_token(script, tokv, &type->type);
}

static bool parse_source_cmd(struct cafe_script *script,
			     size_t tokc, struct token tokv[])
{
	int rc;
	size_t src_argc;
	char **src_argv;
	char filename[PATH_MAX];
	size_t strsz;

	if (tokc < 1) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "not enough arguments for 'source'");
		return false;
	}

	strsz = strlen(tokv[0].str);

	if (strsz + 1 >= PATH_MAX) {
		error(script, CURR_FILE(script), CURR_LINE(script),
		      "filename too long");
		return false;
	}

	strncpy(filename, tokv[0].str, PATH_MAX);
	filename[strsz] = '\0';
	src_argc = tokc - 1;
	src_argv = NULL;

	if (src_argc > 0) {
		size_t i;
		struct token *tp;

		src_argv = zmalloc(script, sizeof *src_argv * src_argc);

		if (src_argv == NULL)
			return false;

		for (tp = &tokv[1], i = 0; i < src_argc; i++, tp++) {
			char *s = cp_str(script, tp->str);
			if (s == NULL) {
				CAFE_FREE(script->allocator, s);
				return false;
			}
			src_argv[i] = s;
		}
	}

	rc = parse_from_file(script, NULL, filename,
			     (unsigned int) src_argc, (const char **)src_argv);

	if (src_argc > 0) {
		size_t i;

		for (i = 0; i < src_argc; i++) {
			CAFE_FREE(script->allocator, src_argv[i]);
		}
		CAFE_FREE(script->allocator, src_argv);
	}

	return rc == CAFE_OK;
}

static bool is_positional_arg(struct token *t, int *val)
{
	size_t i, n;
	size_t strsz = strlen(t->str);

	*val = 0;		/* invalid positional value */

	if (t->str[0] != '$')
		return false;

	for (i = 1, n = 0; i < strsz && isdigit(t->str[i]); i++) {
		n = 10 * n + (t->str[i] - '0');
	}
	if (i == strsz)
		*val = (int) n;
	return i == strsz;
}

static bool expand_positional_args(struct cafe_script *script,
				   size_t tokc,
				   struct token tokv[],
				   int argc, const char *argv[])
{
	unsigned int i;

	for (i = 0; i < tokc; i++) {
		int val;
		struct token t = tokv[i];

		if (!is_positional_arg(&t, &val))
			continue;
		if (val -1 < 0 || val -1 >= argc) {
			error(script, CURR_FILE(script), CURR_LINE(script),
			      "argument not supplied for '%s'",
			      tokv[i].str);
			return false;
		}

		tokv[i].str = (char *)argv[val-1];
	}

	return true;
}

static bool do_cmd(struct cafe_script *script,
		   size_t tokc,
		   struct token tokv[],
		   int argc, const char *argv[])
{
	static const struct {
		const char *name;
		CMD_HANDLER handler;
		bool clear_curr_ep;
		int expand_positional_args;
	} cmds[] = {
		{ "action", parse_aggregator_action, false, false },
		{ "connect", parse_connect, true, true},
		{ "create", parse_create, true, true },
		{ "enable", parse_enable, false, true },
		{ "set", parse_set, false, true },
		{ "source", parse_source_cmd, false, true },
		{ "table", parse_aggregator_table, false, true },
		{ "typedef", parse_typedef, false, false },
		{ "type", parse_typedef, false, false }
	};

	struct token *cmd = tokv;
	size_t i;

	/* skip lines that have no tokens; it may just contain spaces
	 * or be a comment that doesn't start @column 1. */

	if (tokc == 0)
		return true;

	for (i = 0; i < CAFE_NELEMENTS(cmds); i++) {
		if (!TOK_EQL(cmd, cmds[i].name))
			continue;
		if (cmds[i].clear_curr_ep)
			script->curr_ep = NULL;
		if (cmds[i].expand_positional_args)
			(void)expand_positional_args(script,
					       tokc -1, &tokv[1],
					       argc, argv);
		return cmds[i].handler(script, tokc - 1, &tokv[1]);
	}

	error(script, tokv[0].file, tokv[0].lineno,
	      "'%s': unknown command", tokv[0].str);

	return false;
}

static int enable_events_for_aggregator(struct cafe_script *script,
					struct script_ep *src_ep,
					struct script_event *e)
{
	int rc;
	unsigned int i;
	struct cafe_event_descr *event_descr;
	struct cafe_list_head *list_item;
	struct script_ep *ep;

	ep = find_ep_by_alias(script, e->event_type[0].str);

	if (ep == NULL || strcmp(ep->name.str, CAFE_AGGREGATOR_EP_NAME) != 0)
		return CAFE_OK;

	cafe_list_foreach(list_item, &ep->aggr.action_list_head) {
		struct aggregator_action *p;
		char *type, *word, *last;
		struct cafe_value_type *vt[CAFE_MAX_EVENT_TYPES];

		last = type = NULL;
		p = AGGR_ACTION_LIST_NODE_ENTRY(list_item);

		for (word = strtok_r(p->trigger.str, ", ", &last), i = 0;
		     word;
		     word = strtok_r(NULL, ", ", &last), i++) {
			assert(i != CAFE_MAX_EVENT_TYPES);
			type = cafe_mprintf(script->allocator, "t[%s]", word);
			if (type == NULL) {
				error(script, CURR_FILE(script), -1, "out of memory!");
				return CAFE_NOMEM;
			}
			rc = cafe_type_parse(script->session->ns, type, &vt[i]);
			if (rc != CAFE_OK) {
				error(script,
				      p->trigger.file,
				      p->trigger.lineno,
				      "cannot parse type '%s': %s",
				      type,
				      cafe_strerror(rc));
				CAFE_FREE(script->allocator, type);
				return rc;
			}
			CAFE_FREE(script->allocator, type);
		}

		if (i == 0) {
			error(script,
			      p->trigger.file,
			      p->trigger.lineno,
			      "no trigger event found '%s'",
			      p->trigger.str);
			return CAFE_SYNTAX;
		}
			
		rc = cafe_cfg_get_event(script->session, i, vt, &event_descr);

		if (rc != CAFE_OK) {
			error(script,
			      p->trigger.file,
			      p->trigger.lineno,
			      "unknown event '%s': %s",
			      type,
			      cafe_strerror(rc));
			return rc;
		}

		rc = cafe_ep_inst_enable_output_event(src_ep->inst,
						      event_descr);
		if (rc != CAFE_OK) {
			error(script,
			      p->trigger.file,
			      p->trigger.lineno,
			      "Unable to enable output event '%s': %s",
			      type,
			      cafe_strerror(rc));
			return rc;
		}
	}

	return CAFE_OK;
}

static int enable_events_for_ep(struct cafe_script *script,
				struct script_ep *ep,
				struct script_event *e)
{
	int rc;
	unsigned int i, j;
	struct script_ep *sink;
	struct cafe_event_descr *event_descr;

	sink = find_ep_by_alias(script, e->event_type[0].str);
	assert(sink);

	for (i = 0; i < sink->spec->n_input_events; i++) {
		struct cafe_ep_spec_input_evt *input_event;
		struct cafe_value_type *vt[CAFE_MAX_EVENT_TYPES] = {0};

		input_event = &sink->spec->input_events[i];

		for (j = 0; j < input_event->num_types; j++) {
			assert(j != CAFE_MAX_EVENT_TYPES);
			rc = cafe_type_parse(script->session->ns,
					     input_event->type[j],
					     &vt[j]);

			if (rc != CAFE_OK) {
				error(script,
				      e->event_type[j].file,
				      e->event_type[j].lineno,
				      "cannot parse type '%s': %s",
				      e->event_type[j].str,
				      cafe_strerror(rc));
				return rc;
			}
		}

		rc = cafe_cfg_get_event(script->session, j, vt, &event_descr);

		if (rc != CAFE_OK) {
			error(script,
			      e->event_type[0].file,
			      e->event_type[0].lineno,
			      "unknown event '%s': %s",
			      e->event_type[0].str,
			      cafe_strerror(rc));
			return rc;
		}

		rc = cafe_ep_inst_enable_output_event(ep->inst, event_descr);
		if (rc != CAFE_OK) {
			error(script,
			      e->event_type[0].file,
			      e->event_type[0].lineno,
			      "Unable to enable output event '%s': %s",
			      e->event_type[0].str,
			      cafe_strerror(rc));
			return rc;
		}
	}

	return enable_events_for_aggregator(script, ep, e);
}

static int enable_type_ref_events(struct cafe_script *s,
				  struct script_ep *ep,
				  struct script_event *e)
{
	int rc;
	struct cafe_value_type *vt[CAFE_MAX_EVENT_TYPES];
	struct cafe_event_descr *event_descr;
	unsigned int i;

	for (i = 0; i < e->n_event_types; i++) {
		rc = cafe_type_parse(s->session->ns,
				     e->event_type[i].str,
				     &vt[i]);
		if (rc != CAFE_OK) {
			error(s,
			      e->event_type[i].file,
			      e->event_type[i].lineno,
			      "cannot parse type '%s': %s",
			      e->event_type[i].str,
			      cafe_strerror(rc));
			return rc;
		}
	}

	rc = cafe_cfg_get_event(s->session,
				e->n_event_types,
				vt, &event_descr);
	if (rc != CAFE_OK) {
		error(s,
		      e->event_type[0].file,
		      e->event_type[0].lineno,
		      "unknown event '%s': %s",
		      e->event_type[0].str,
		      cafe_strerror(rc));
		return rc;
	}

	rc = cafe_ep_inst_enable_output_event(ep->inst, event_descr);

	if (rc != CAFE_OK) {
		error(s,
		      e->event_type[0].file,
		      e->event_type[0].lineno,
		      "cannot enable event '%s': %s",
		      e->event_type[0].str, cafe_strerror(rc));
		return rc;
	}

	return CAFE_OK;
}

static void delete_setting_value(const struct cafe_allocator *allocator,
				 struct cafe_value_type *type,
				 cafe_any_value *value)
{
	switch (type->kind) {
	case CAFE_TYPE_STRING:
		CAFE_FREE(allocator, value->str);
		break;
	case CAFE_TYPE_DYNARRAY:
		(void)cafe_vm_dynarray_set_size(&value->dary, 0, type);
		break;
	}
}

static void delete_ep_settings(struct cafe_script *script,
			       struct script_ep *ep)
{
	size_t i;
	void *tmp = ep->settings.p;
	struct cafe_cfg_setting *setting = ep->settings.p;

	for (i = 0; i < ep->settings.l; i++) {
		struct cafe_vm_attr *ident;
		struct cafe_value_type *workarea_type;

		if (!get_workarea_type(script, ep, &workarea_type))
			continue;

		ident = cafe_type_attr_by_name(ep->workarea_type,
					       setting[i].name);
		assert(ident != NULL);
		delete_setting_value(script->allocator,
				     cafe_attr_type(ident), &setting[i].value);
		CAFE_FREE(script->allocator, (char *)setting[i].name);
		memset(setting, 0, sizeof *setting);
	}
	memset(&ep->settings, 0, sizeof ep->settings);
	CAFE_FREE(script->allocator, tmp);
}

static void script_cleanup(struct cafe_script *script)
{
	struct cafe_list_head *list_item, *tmp;

	cafe_list_foreach_safe(list_item, &script->ep_list_head, tmp) {
		struct script_ep *ep;
		struct cafe_list_head *list_item2, *tmp2;

		ep = EP_LIST_NODE(list_item);
		delete_ep_settings(script, ep);
		disconnect_all(script, ep);

		cafe_list_foreach_safe(list_item2, &ep->enabled_event_list_head, tmp2) {
			struct script_event *event;
			unsigned int i;
			event = EP_EVENT_LIST_NODE(list_item2);
			for (i = 0; i < event->n_event_types; i++) {
				CAFE_FREE(script->allocator, event->event_type[i].str);
			}
			cafe_list_remove(&event->list);
			memset(event, 0, sizeof *event);
			CAFE_FREE(script->allocator, event);
		}
		assert(ep->settings.l == 0);
		assert(ep->settings.c == 0);
		assert(ep->settings.p == NULL);
		CAFE_FREE(script->allocator, ep->name.str);
		CAFE_FREE(script->allocator, ep->alias.str);
		assert(cafe_list_is_empty(&ep->enabled_event_list_head));
		assert(cafe_list_is_empty(&ep->sink_list_head));
		assert(cafe_list_is_empty(&ep->source_list_head));
		cafe_list_remove(&ep->list);

		cafe_list_foreach_safe(list_item2, &ep->aggr.table_list_head, tmp2) {
			struct aggregator_table *p;
			p = AGGR_TABLE_LIST_NODE_ENTRY(list_item2);
			CAFE_FREE(script->allocator, p->name.str);
			cafe_list_remove(&p->list);
			CAFE_FREE(script->allocator, p);
		}

		cafe_list_foreach_safe(list_item2, &ep->aggr.action_list_head, tmp2) {
			struct aggregator_action *p;
			p = AGGR_ACTION_LIST_NODE_ENTRY(list_item2);
			CAFE_FREE(script->allocator, p->filter.str);
			CAFE_FREE(script->allocator, p->trigger.str);
			CAFE_FREE(script->allocator, p->cell.str);
			CAFE_FREE(script->allocator, p->operation.str);
			CAFE_FREE(script->allocator, p->args.str);
			cafe_list_remove(&p->list);
			CAFE_FREE(script->allocator, p);
		}

		assert(cafe_list_is_empty(&ep->aggr.table_list_head));
		assert(cafe_list_is_empty(&ep->aggr.action_list_head));
		memset(ep, 0, sizeof *ep);
		CAFE_FREE(script->allocator, ep);
	}

	cafe_list_foreach_safe(list_item, &script->ep_connection_list_head, tmp) {
		struct script_connection *c;
		c = CONNECTION_LIST_NODE(list_item);
		CAFE_FREE(script->allocator, c->src.str);
		CAFE_FREE(script->allocator, c->dst.str);
		cafe_list_remove(&c->list);
		memset(c, 0, sizeof *c);
		CAFE_FREE(script->allocator, c);
	}
	assert(cafe_list_is_empty(&script->ep_connection_list_head));

	cafe_list_foreach_safe(list_item, &script->byte_stream_list_head, tmp) {
		struct byte_stream *bs;
		bs = BYTE_STREAM_LIST_NODE_ENTRY(list_item);
		cafe_list_remove(&bs->list);
		memset(bs, 0, sizeof *bs);
		CAFE_FREE(script->allocator, bs);
	}
	assert(cafe_list_is_empty(&script->byte_stream_list_head));

	cafe_list_foreach_safe(list_item, &script->type_list_head, tmp) {
		struct type *p;
		p = TYPE_LIST_NODE_ENTRY(list_item);
		CAFE_FREE(script->allocator, p->type.str);
		cafe_list_remove(&p->list);
		CAFE_FREE(script->allocator, p);
	}
	assert(cafe_list_is_empty(&script->type_list_head));

	cafe_ptr_array_release(&script->tsorted_nodes);
}

/* tsort: topological sort. */

static int tsort(struct cafe_script *script,
		 struct script_ep *const ep,
		 struct cafe_ptr_array *array)
{
	int rc;
	struct cafe_list_head *list_item;
	struct script_ep_edge *edge;

	if (ep->visited == DONE_VISITED)
		return CAFE_OK;

	ep->visited = DONE_VISITED;

	cafe_list_foreach(list_item, &ep->sink_list_head) {
		edge = EP_EDGE_SINK_LIST_NODE_ENTRY(list_item);
		rc = tsort(script, edge->sink, array);
		if (rc != CAFE_OK)
			return rc;
	}
	return cafe_ptr_array_append(array, ep);
}

static int add_aggregator_settings(struct cafe_script *script,
				   struct script_ep *ep)
{
	int rc;
	struct cafe_list_head *list_item;

	cafe_list_foreach(list_item, &ep->aggr.table_list_head) {
		struct aggregator_table *p;
		struct cafe_value_type *type;

		p = AGGR_TABLE_LIST_NODE_ENTRY(list_item);
		rc = cafe_type_parse(ep->inst->session->ns,
				     p->name.str, &type);
		if (rc != CAFE_OK) {
			error(script,
			      p->name.file,
			      p->name.lineno,
			      "cannot parse table type: %s", cafe_strerror(rc));
			return rc;
		}
		rc = cafe_aggregator_ep_table_create(ep->inst, type);
		if (rc != CAFE_OK) {
			error(script,
			      p->name.file,
			      p->name.lineno,
			      "cannot add table: %s", cafe_strerror(rc));
			return rc;
		}
	}

	cafe_list_foreach(list_item, &ep->aggr.action_list_head) {
		struct aggregator_action *p;

		p = AGGR_ACTION_LIST_NODE_ENTRY(list_item);
		rc = cafe_aggregator_ep_action_add(
			ep->inst,
			p->trigger.str,
			p->filter.str,
			p->cell.str,
			p->operation.str,
			p->args.str);
		if (rc != CAFE_OK) {
			error(script,
			      p->trigger.file,
			      p->trigger.lineno,
			      "cannot add action: %s", cafe_strerror(rc));
			return rc;
		}
	}

	return CAFE_OK;
}

static int build_connections(struct cafe_script *script)
{
	size_t i;
	int rc = CAFE_OK;
	struct cafe_list_head *list_item;
	struct cafe_list_head *entry;
	struct cafe_ptr_array roots;

	CAFE_ASSERT_LOCKED();

	cafe_ptr_array_init(&roots, script->allocator);
	assert(script->session);

	/* Add all types to the session. */

	cafe_list_foreach(list_item, &script->type_list_head) {
		struct type *p;
		struct cafe_value_type *type;

		p = TYPE_LIST_NODE_ENTRY(list_item);
		rc = cafe_type_parse(script->session->ns, p->type.str, &type);

		if (rc != CAFE_OK) {
			error(script,
			      p->type.file,
			      p->type.lineno,
			      "cannot parse type");
			return rc;
		}
	}

	cafe_list_foreach(entry, &script->ep_list_head) {
		bool is_aggregator;
		struct script_ep *ep;

		ep = EP_LIST_NODE(entry);
		ep->spec = cafe_ep_spec_find(ep->name.str);

		if (ep->spec == NULL) {
			error(script, ep->name.file, ep->name.lineno,
			      "event processor '%s' not registered",
			      ep->name.str);
			rc = CAFE_NAMESPACE_NOT_FOUND;
			goto error;
		}

		is_aggregator = strcmp(ep->name.str, CAFE_AGGREGATOR_EP_NAME) == 0;

		rc = cafe_ep_inst_create(script->session, ep->spec,
					 is_aggregator ? 0 : (unsigned int) ep->settings.l,
					 is_aggregator ? 0 : ep->settings.p,
					 &ep->inst);

		if (rc != CAFE_OK) {
			error(script, ep->name.file, ep->name.lineno,
			      "cannot create event processor '%s'",
			      ep->name.str);
			goto error;
		}

		if (ep->spec->dynamic_output_events) {
			char *type_str;
			char *tmp_str;
			const char *p;
			struct cafe_value_type *type;

			rc = ep->spec->dynamic_output_events(ep->inst, &type_str);

			if (rc != CAFE_OK) {
				error(script, ep->name.file, ep->name.lineno,
				      "cannot discover dynamic types: %s",
				      cafe_strerror(rc));
				CAFE_FREE(ep->inst->session->allocator, type_str);
				goto error;
			}

			p = type_str;

			/* Split the string into individual types to
			 * add error reporting if some individual type
			 * in the string fails to parse. */

			while (*p != '\0') {
				size_t n_brk;
				const char *s;

				assert(isascii(p[0]));
				assert(p[1] == '[');
				n_brk = 1;
				s = p + 2;

				while (*s != '\0' && n_brk) {
					if (*s == '[')
						n_brk++;
					else if (*s == ']')
						n_brk--;
					s++;
				}

				tmp_str = cp_nstr(script, p, s - p);

				if (tmp_str == NULL) {
					CAFE_FREE(ep->inst->session->allocator, type_str);
					goto error;
				}

				p = s;
				rc = cafe_type_parse(script->session->ns, tmp_str, &type);

				if (rc != CAFE_OK) {
					error(script, ep->name.file, ep->name.lineno,
					      "cannot parse type '%s': %s",
					      tmp_str,
					      cafe_strerror(rc));
					CAFE_FREE(ep->inst->session->allocator, tmp_str);
					CAFE_FREE(ep->inst->session->allocator, type_str);
					goto error;
				}
				CAFE_FREE(ep->inst->session->allocator, tmp_str);
			}

			CAFE_FREE(ep->inst->session->allocator, type_str);
		}

		if (is_aggregator) {
			rc = add_aggregator_settings(script, ep);
			if (rc != CAFE_OK)
				goto error;
		}

		if (!cafe_list_is_empty(&ep->enabled_event_list_head))
			continue;

		rc = cafe_inst_enable_all_output_events(ep->inst);

		if (rc != CAFE_OK) {
			error(script, ep->name.file, ep->name.lineno,
			      "%s: cannot enable output events: %s",
			      ep->name.str, cafe_strerror(rc));
			goto error;
		}
	}

	/* Build an array of nodes with no incoming edges. */

	cafe_list_foreach(entry, &script->ep_list_head) {
		struct script_ep *ep;
		ep = EP_LIST_NODE(entry);
		ep->visited = NOT_VISITED;
		if (cafe_list_is_empty(&ep->source_list_head)) {
			rc = cafe_ptr_array_append(&roots, ep);
			if (rc != CAFE_OK)
				goto error;
		}
	}

	/* Topologically sort the list of nodes with no incoming edges
	 * into the `sorted_nodes' array. */

	for (i = 0; i < cafe_ptr_array_len(&roots); i++) {
		struct script_ep *ep;

		ep = cafe_ptr_array_val(&roots, i);
		if (ep->visited == NOT_VISITED) {
			rc = tsort(script, ep, &script->tsorted_nodes);
			if (rc != CAFE_OK)
				goto error;
		}
	}

	/* Make all the connections based on the topological
	 * ordering. */

	for (i = cafe_ptr_array_len(&script->tsorted_nodes); i != 0; i--) {
		struct script_ep *ep;
		struct cafe_list_head *entry2;

		ep = cafe_ptr_array_val(&script->tsorted_nodes, i - 1);
		assert(ep->inst);

		cafe_list_foreach(entry2, &ep->sink_list_head) {
			struct script_ep_edge *edge;
			edge = EP_EDGE_SINK_LIST_NODE_ENTRY(entry2);
			assert(edge->sink->inst);
			rc = cafe_ep_inst_connect(ep->inst, edge->sink->inst);
			if (rc != CAFE_OK) {
				error(script, ep->name.file, ep->name.lineno,
				      "failed to connect "
				      "'%s' to '%s': %s",
				      ep->inst->spec->name,
				      edge->sink->inst->spec->name,
				      cafe_strerror(rc));
				goto error;
			}
		}
	}

	/* Enable explicit events on each EP that is connected. */

	cafe_list_foreach(entry, &script->ep_list_head) {
		struct script_ep *ep;
		struct cafe_list_head *entry2;

		ep = EP_LIST_NODE(entry);

		cafe_list_foreach(entry2, &ep->enabled_event_list_head) {
			struct script_event *e;

			e = EP_EVENT_LIST_NODE(entry2);

			if (e->enable_type == ENABLE_TYPE_REF)
				rc = enable_type_ref_events(script, ep, e);
			else if (e->enable_type == ENABLE_EP)
				rc = enable_events_for_ep(script, ep, e);
			if (rc != CAFE_OK)
				goto error;
		}
	}

	rc = cafe_instrument_refresh_output_event_actions(script->session);

	if (rc != CAFE_OK) {
		error(script, CURR_FILE(script), -1,
		      "cannot refresh event actions: %s",
		      cafe_strerror(rc));
		goto error;
	}
	cafe_ptr_array_release(&roots);
	assert(rc == CAFE_OK);
	return rc;
error:
	cafe_ptr_array_release(&roots);
	return rc;
}

static int validate_connections(struct cafe_script *script)
{
	struct cafe_list_head *entry;

	/* Validate connection list. */

	cafe_list_foreach(entry, &script->ep_connection_list_head) {
		struct script_connection *c;
		c = CONNECTION_LIST_NODE(entry);
		c->src_ep = find_ep_by_alias(script, c->src.str);
		if (c->src_ep == NULL) {
			error(script,
			      c->src.file,
			      c->src.lineno,
			      "missing 'create' statement for '%s'",
			      c->src.str);
			return CAFE_ERROR;
		}
		c->dst_ep = find_ep_by_alias(script, c->dst.str);
		if (c->dst_ep == NULL) {
			error(script,
			      c->src.file,
			      c->src.lineno,
			      "missing 'create' statement for '%s'",
			      c->dst.str);
			return CAFE_ERROR;
		}
	}

	/* Validate enable EP list. */

	cafe_list_foreach(entry, &script->ep_list_head) {
		struct script_ep *ep;
		struct cafe_list_head *entry2;

		ep = EP_LIST_NODE(entry);

		cafe_list_foreach(entry2, &ep->enabled_event_list_head) {
			size_t i;
			struct script_event *e;
			struct script_ep *src_ep;

			e = EP_EVENT_LIST_NODE(entry2);
			if (e->enable_type != ENABLE_EP)
				continue;

			for (i = 0; i < e->n_event_types; i++) {
				src_ep = find_ep_by_alias(script, e->event_type[i].str);
				if (src_ep == NULL) {
					error(script,
					      e->event_type[i].file,
					      e->event_type[i].lineno,
					      "missing 'create' statement for '%s'",
					      e->event_type[i].str);
					return CAFE_ERROR;
				}
			}
		}
	}

	cafe_list_foreach(entry, &script->ep_connection_list_head) {
		int rc;
		struct script_connection *c;
		c = CONNECTION_LIST_NODE(entry);
		rc = connect_ep(script, c->src_ep,
				find_ep_by_alias(script, c->dst.str));
		if (rc != CAFE_OK)
			return rc;
	}

	if (graph_has_cycles(script)) {
		error(script, CURR_FILE(script), -1, "%s", script->errbuf);
		return false;
	}

	return CAFE_OK;
}

static void script_init(struct cafe_script *script,
			const struct cafe_allocator *allocator,
			CAFE_SCRIPT_ERROR_HANDLER error_handler,
			void *error_handler_client_data)
{
	assert(allocator);
	memset(script, 0, sizeof *script);
	script->allocator = allocator;
	script->error_handler = error_handler;
	script->error_handler_client_data = error_handler_client_data;

	if (script->error_handler == NULL)
		script->error_handler = default_error_handler;

	cafe_list_init(&script->ep_list_head);
	cafe_list_init(&script->ep_connection_list_head);
	cafe_list_init(&script->byte_stream_list_head);
	cafe_ptr_array_init(&script->tsorted_nodes, script->allocator);
	cafe_list_init(&script->type_list_head);
}

#if CAFE_OS_UNIX || CAFE_OS_WINDOWS
static void print_str_escaped(FILE *fp, const char *s, bool enclose_in_quotes)
{
	size_t i;
	size_t j = strlen(s);

	if (enclose_in_quotes)
		fprintf(fp, "\"");

	for (i = 0; i < j; i++) {
		switch (s[i]) {
		case '"':
			fputs("\\\"", fp);
			break;
		default:
			fputc(s[i], fp);
			break;
		}
	}

	if (enclose_in_quotes)
		fprintf(fp, "\"");
}

static void dump_session_types_for_tcfsh_script(struct cafe_vm_ident *ident, FILE *fp)
{
	while (ident != NULL) {
		if (ident->type != NULL) {
			if (ident->type->type_str) {
				print_str_escaped(fp, ident->type->type_str, false);
			}
		}
		dump_session_types_for_tcfsh_script(ident->children, fp);
		ident = ident->peer;
	}
}

static void dump_as_tcfsh_script(struct cafe_script *script)
{
	size_t i;
	struct cafe_list_head *list_item;
	const char *filename;
	const char *host;
	const char *port;
	const char *add_types_from_session;
	FILE *fp;

	filename = getenv("CAFE_SCRIPT_TCFSH_FILENAME");
	host = getenv("CAFE_SCRIPT_TCFSH_HOST");
	port = getenv("CAFE_SCRIPT_TCFSH_PORT");
	add_types_from_session = getenv("CAFE_SCRIPT_TCFSH_ADD_TYPES_FROM_SESSION");

	if (filename == NULL)
		return;

	if (host == NULL)
		host = "127.0.0.1";

	if (port == NULL)
		port = "1534";

	if ((fp = fopen(filename, "w")) == NULL) {
		error(script,
		      CURR_FILE(script), -1,
		      "error: cannot open %s: %s",
		      filename,
		      strerror(errno));
		return;
	}

	fprintf(fp, "# -*-lua-*-\n"
		"connect TCP:%s:%s\n"
		"cafe_config_create test\n"
		"set session_id result\n", host, port);

	cafe_list_foreach(list_item, &script->ep_list_head) {
		struct script_ep *ep;
		ep = EP_LIST_NODE(list_item);
		fprintf(fp, "cafe_ep_create session_id %s {", ep->name.str);

		if (strcmp(ep->name.str, CAFE_AGGREGATOR_EP_NAME) == 0) {
			struct cafe_list_head *list_item2;

			fprintf(fp, "{\"add_types\",{");

			if (add_types_from_session != NULL) {
				fprintf(fp, "\"");
				dump_session_types_for_tcfsh_script(script->session->ns->children, fp);
				fprintf(fp, "\",");
			}

			cafe_list_foreach(list_item2, &script->type_list_head) {
				struct type *p;
				p = TYPE_LIST_NODE_ENTRY(list_item2);
				print_str_escaped(fp, p->type.str, true);
				if (list_item2->next != &script->type_list_head)
					fprintf(fp, ",");
			}
			fprintf(fp, "}},{\"add_tables\",{");
			cafe_list_foreach(list_item2, &ep->aggr.table_list_head) {
				struct aggregator_table *p;
				p = AGGR_TABLE_LIST_NODE_ENTRY(list_item2);
				print_str_escaped(fp, p->name.str, true);
				if (list_item2->next != &ep->aggr.table_list_head)
					fprintf(fp, ",");
			}
			fprintf(fp, "}},{\"add_actions\",{");
			cafe_list_foreach(list_item2, &ep->aggr.action_list_head) {
				struct aggregator_action *p;
				p = AGGR_ACTION_LIST_NODE_ENTRY(list_item2);
				fprintf(fp, "{");
				print_str_escaped(fp, STR_REPR(p->trigger.str), true);
				fprintf(fp, ",");
				print_str_escaped(fp, STR_REPR(p->filter.str), true);
				fprintf(fp, ",");
				print_str_escaped(fp, STR_REPR(p->cell.str), true);
				fprintf(fp, ",");
				print_str_escaped(fp, STR_REPR(p->operation.str), true);
				fprintf(fp, ",");
				print_str_escaped(fp, STR_REPR(p->args.str), true);
				fprintf(fp, "}");
				if (list_item2->next != &ep->aggr.action_list_head)
					fprintf(fp, ",");
			}
			fprintf(fp, "}}");
		} else {
			struct cafe_cfg_setting *setting = ep->settings.p;
			struct cafe_vm_attr *attr;
			cafe_any_value *value;

			for (i = 0; i < ep->settings.l; i++) {
				assert(ep->workarea_type);
				attr = cafe_type_attr_by_name(ep->workarea_type, setting[i].name);
				assert(attr);
				assert(cafe_find_nature_str(attr->ident, cafe_setting_VT_DEF));

				fprintf(fp, "{");
				print_str_escaped(fp, setting[i].name, true);
				fprintf(fp, ",");
				value = &setting[i].value;

				switch (cafe_attr_type(attr)->kind) {
				case CAFE_TYPE_UCHAR:
				case CAFE_TYPE_USHORT:
				case CAFE_TYPE_UINT:
				case CAFE_TYPE_ULONG:
				case CAFE_TYPE_ULONGLONG:
					fprintf(fp, "%llu", value->ull);
					break;
				case CAFE_TYPE_CHAR:
				case CAFE_TYPE_SHORT:
				case CAFE_TYPE_INT:
				case CAFE_TYPE_LONG:
				case CAFE_TYPE_LONGLONG:
					fprintf(fp, "%lld", value->ll);
					break;
				case CAFE_TYPE_BOOL:
					fprintf(fp, "%lld", value->ull);
					break;
				case CAFE_TYPE_STRING:
					print_str_escaped(fp, value->str, true);
					break;
				default:
					/* coverity[array_null] */
					assert(!"unexpected attribute type");
				}

				fprintf(fp, "}");
				if (i + 1 < ep->settings.l)
					fprintf(fp, ",");
			}
		}
		fprintf(fp, "}\n");
		fprintf(fp, "set %s_id result[1]\n", ep->name.str);
	}

	cafe_list_foreach(list_item, &script->ep_list_head) {
		struct script_ep *ep;
		struct cafe_list_head *list_item2;

		ep = EP_LIST_NODE(list_item);
		cafe_list_foreach(list_item2, &ep->enabled_event_list_head) {
			struct script_event *event;
			event = EP_EVENT_LIST_NODE(list_item2);
			fprintf(fp, "cafe_event_enable session_id %s_id {", ep->name.str);
			for (i = 0; i < event->n_event_types; i++) {
				fprintf(fp, "\"%s\"", event->event_type[i].str);
				if (i + 1 < event->n_event_types)
					fprintf(fp, ",");
			}
			fprintf(fp, "}\n");
		}
	}

	for (i = cafe_ptr_array_len(&script->tsorted_nodes); i != 0; i--) {
		struct script_ep *ep;
		struct cafe_list_head *entry2;

		ep = cafe_ptr_array_val(&script->tsorted_nodes, i - 1);

		cafe_list_foreach(entry2, &ep->sink_list_head) {
			struct script_ep_edge *edge;
			edge = EP_EDGE_SINK_LIST_NODE_ENTRY(entry2);
			assert(edge->sink->inst);
			fprintf(fp, "cafe_ep_connect session_id %s_id %s_id\n",
				ep->inst->spec->name,
				edge->sink->inst->spec->name);
		}
	}

	fprintf(fp, "cafe_config_start session_id\n");
	fclose(fp);
	exit(EXIT_SUCCESS);
}
#endif

int cafe_script_parse_from_file(const struct cafe_allocator *allocator,
				FILE *fp,
				const char *filename,
				int argc, const char *argv[],
				CAFE_SCRIPT_ERROR_HANDLER error_handler,
				void *error_handler_client_data,
				struct cafe_instrument **result)
{
	int rc;
	struct cafe_script script;

	script_init(&script, allocator, error_handler, error_handler_client_data);
	rc = cafe_instrument_create("name", &script.session);

	if (rc != CAFE_OK) {
		error(&script, CURR_FILE(&script), 1,
		      "cannot create session: %s", cafe_strerror(rc));
		return rc;
	}

	rc = parse_from_file(&script, fp, filename, argc, argv);

	if (rc == CAFE_OK && script.n_errors == 0)
		rc = validate_connections(&script);

	if (rc == CAFE_OK)
		rc = build_connections(&script);

	if (rc == CAFE_OK)
		*result = script.session;

#if CAFE_OS_UNIX || CAFE_OS_WINDOWS
	if (rc == CAFE_OK) {
		/* This dump function relies on a lot of state that
		 * has been built up in parse, validate, connect,
		 * etc. */
		dump_as_tcfsh_script(&script);
	}
#endif

	/* The order here is significant: the cleanup must come before
	 * cafe_instrument_delete(). */

	script_cleanup(&script);

	if (rc != CAFE_OK) {
		int rc2 = cafe_instrument_delete(script.session);
		if (rc2 != CAFE_OK)
			cafe_error_set(NULL, rc2,
				       "Error deleting instrument: %s",
				       cafe_strerror(rc2));
	}

	return rc;
}

int cafe_script_parse_from_string(const struct cafe_allocator *allocator,
				  const char *content,
				  int argc, const char *argv[],
				  CAFE_SCRIPT_ERROR_HANDLER error_handler,
				  void *error_handler_client_data,
				  struct cafe_instrument **result)
{
	int rc;
	struct cafe_script script;

	*result = NULL;
	script_init(&script, allocator, error_handler, error_handler_client_data);
	rc = cafe_instrument_create("name", &script.session);

	if (rc != CAFE_OK) {
		error(&script, CURR_FILE(&script), 1,
		      "cannot create session: %s", cafe_strerror(rc));
		return rc;
	}

	rc = parse_from_string(&script, content, argc, argv);

	if (rc == CAFE_OK && script.n_errors == 0)
		rc = validate_connections(&script);

	if (rc == CAFE_OK)
		rc = build_connections(&script);

	if (rc == CAFE_OK)
		*result = script.session;

	/* The order here is significant: the cleanup must come before
	 * cafe_instrument_delete(). */

	script_cleanup(&script);

	if (rc != CAFE_OK) {
		int rc2 = cafe_instrument_delete(script.session);
		if (rc2 != CAFE_OK)
			cafe_error_set(NULL, rc2,
				       "Error deleting instrument: %s",
				       cafe_strerror(rc2));
	}

	return rc;
}
