/* log_buffer_test.c - log buffer unit test suite */

/* Copyright (c) 2009-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
20jun14,sdt  Add missing reset init values that prevents to run test twice

DESCRIPTION
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <cafe/log_buffer.h>
#include "../log_buffer_diagnostics.h"
#include "../src/cafe_os.h"
#include <test/cafe_CUnitTest.h>


#define internal_log_write_reserve(buf, size, msg)		\
	do {							\
		msg = cafe_log_write_reserve(buf, size);	\
		CUT_ASSERT_NOT_NULL(msg);			\
	} while(0)

#define internal_log_read_reserve(buf, size, expected, msg)	\
	do {							\
		msg = cafe_log_read_reserve(buf, size);		\
		CUT_ASSERT_NOT_NULL(msg);			\
		CUT_ASSERT_EQUAL(*size, expected);		\
	} while(0)

#define internal_log_write_reserve_min(buf, minimum, size, expected, msg) \
	do {								\
		msg = cafe_log_write_reserve_min(buf, minimum, size);	\
		CUT_ASSERT_NOT_NULL(msg);				\
		CUT_ASSERT_EQUAL(*size, expected);			\
	} while(0)

/**
 *
 */
static
int test_log_create(void)
{
	struct cafe_log_buf *log = cafe_log_create(100, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}


static
int test_log_write(void)
{
	const int test = 0xdeadbeef;
	void *msg;
	struct cafe_log_buf *log =
		cafe_log_create(sizeof(int) + 1, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	internal_log_write_reserve(log, sizeof(int), msg);

	memcpy(msg, &test, sizeof(int));

	cafe_log_write_commit(log, sizeof(int));
	CUT_ASSERT(cafe_log_validate(log));

	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}


static
int test_log_read(void)
{
	const int test = 0xdeadbeef;
	void *msg;
	void *read_msg;
	size_t size;
	struct cafe_log_buf *log =
		cafe_log_create(sizeof(int) + 1, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	internal_log_write_reserve(log, sizeof(int), msg);

	memcpy(msg, &test, sizeof(int));
	cafe_log_write_commit(log, sizeof(int));

	read_msg = cafe_log_read_reserve(log, &size);
	CUT_ASSERT_NOT_NULL(read_msg);
	CUT_ASSERT_EQUAL(size, sizeof(int));

	CUT_ASSERT(*(int *)read_msg == test);

	cafe_log_read_commit(log, size);
	CUT_ASSERT(cafe_log_validate(log));

	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}


/**
 * Create a log with space for 4 words, then write 3, read 1, and write 2 more,
 * causing the buffer's write head to wrap around.
 */
static
int test_log_writewrap(void)
{
	struct cafe_log_buf *log =
		cafe_log_create(4 * sizeof(int) + 1, cafe_stdlib_allocator);
	void *msg;
	void *read_msg;
	size_t size;

	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	/* write 3 words */
	internal_log_write_reserve(log, 3 * sizeof(int), msg);

	CUT_ASSERT(cafe_log_validate(log));
	((int *)msg)[0] = 1;
	((int *)msg)[1] = 2;
	((int *)msg)[2] = 3;
	cafe_log_write_commit(log, 3 * sizeof(int));

	/* read two words */
	internal_log_read_reserve(log, &size, 3 * sizeof(int), read_msg);
	CUT_ASSERT_EQUAL(((int *)msg)[0], 1);
	CUT_ASSERT_EQUAL(((int *)msg)[1], 2);
	cafe_log_read_commit(log, 2 * sizeof(int));

	/* write two words separately */
	internal_log_write_reserve(log, sizeof(int), msg);
	CUT_ASSERT(cafe_log_validate(log));
	((int *)msg)[0] = 4;
	cafe_log_write_commit(log, sizeof(int));

	internal_log_write_reserve(log, sizeof(int), msg);
	CUT_ASSERT(cafe_log_validate(log));
	((int *)msg)[0] = 5;
	cafe_log_write_commit(log, sizeof(int));

	CUT_ASSERT(cafe_log_validate(log));
	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}


/**
 * Create a log with space for 4 bytes, then write 3, read 1, and write 2 more,
 * causing the buffer's write head to wrap around.  Then attempt to read
 * the entire buffer, causing the read head to wrap.
 */
static
int test_log_readwrap(void)
{
	void *msg;
	void *read_msg;
	size_t size;
	struct cafe_log_buf *log = cafe_log_create(5, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	/* write 4 bytes */
	internal_log_write_reserve(log, 4, msg);

	CUT_ASSERT(cafe_log_validate(log));
	((int8_t *) msg)[0] = 1;
	((int8_t *) msg)[1] = 2;
	((int8_t *) msg)[2] = 3;
	((int8_t *) msg)[3] = 4;
	cafe_log_write_commit(log, 4);

	/* read 2 bytes */
	internal_log_read_reserve(log, &size, 4, read_msg);
	CUT_ASSERT_EQUAL(((int8_t *) msg)[0], 1);
	CUT_ASSERT_EQUAL(((int8_t *) msg)[1], 2);
	cafe_log_read_commit(log, 2);

	/* write two bytes separately */
	internal_log_write_reserve(log, 1, msg);
	CUT_ASSERT(cafe_log_validate(log));
	((int8_t *) msg)[0] = 5;
	cafe_log_write_commit(log, 1);

	internal_log_write_reserve(log, 1, msg);
	CUT_ASSERT(cafe_log_validate(log));
	((int8_t *) msg)[0] = 6;
	cafe_log_write_commit(log, 1);

	/* read the rest of the current buffer */
	internal_log_read_reserve(log, &size, 3, read_msg);
	CUT_ASSERT_EQUAL(((int8_t *) read_msg)[0], 3);
	CUT_ASSERT_EQUAL(((int8_t *) read_msg)[1], 4);
	CUT_ASSERT_EQUAL(((int8_t *) read_msg)[2], 5);
	cafe_log_read_commit(log, size);

	/* read the rest of the wrapped content */
	internal_log_read_reserve(log, &size, 1, read_msg);
	CUT_ASSERT_EQUAL(((int8_t *) read_msg)[0], 6);
	cafe_log_read_commit(log, size);

	CUT_ASSERT(cafe_log_validate(log));
	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}


//td: update
/**
 *  When the buffer cannot fulfill a write reservation with the amount of space
 * left in the buffer without wrapping, it wraps immediately rather than return
 * a partial reservation.
 */
static
int test_log_writewrap_truncated(void)
{
	void *msg;
	void *read_msg;
	size_t size;
	struct cafe_log_buf *log =
		cafe_log_create(4 * sizeof(int) + 1, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	/* write 4 words */
	internal_log_write_reserve(log, 4 * sizeof(int), msg);

	((int *)msg)[0] = 1;
	((int *)msg)[1] = 2;
	((int *)msg)[2] = 3;
	((int *)msg)[3] = 4;
	cafe_log_write_commit(log, 4 * sizeof(int));

	/* read 2 words */
	internal_log_read_reserve(log, &size, 4 * sizeof(int), read_msg);
	CUT_ASSERT_EQUAL(((int *)msg)[0], 1);
	CUT_ASSERT_EQUAL(((int *)msg)[1], 2);
	cafe_log_read_commit(log, 2 * sizeof(int));

	/* write another word to wrap the write head */
	internal_log_write_reserve(log, sizeof(int), msg);
	CUT_ASSERT(cafe_log_validate(log));
	((int *)msg)[0] = 5;
	cafe_log_write_commit(log, sizeof(int));

	/* read the rest of the current buffer */
	internal_log_read_reserve(log, &size, 2 * sizeof(int), read_msg);
	CUT_ASSERT_EQUAL(((int *)read_msg)[0], 3);
	CUT_ASSERT_EQUAL(((int *)read_msg)[1], 4);
	cafe_log_read_commit(log, size);

	/* read the rest of the wrapped content */
	internal_log_read_reserve(log, &size, sizeof(int), read_msg);
	CUT_ASSERT_EQUAL(((int *)read_msg)[0], 5);
	cafe_log_read_commit(log, size);

	CUT_ASSERT(cafe_log_validate(log));
	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}


static
int test_log_simple_concurrency(void)
{
	void *msg;
	void *read_msg;
	size_t size;
	struct cafe_log_buf *log = cafe_log_create(4, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	/* write 3 bytes without committing */
	internal_log_write_reserve(log, 4, msg);

	((int8_t *) msg)[0] = 1;
	((int8_t *) msg)[1] = 2;
	((int8_t *) msg)[2] = 3;
	((int8_t *) msg)[3] = 4;

	internal_log_read_reserve(log, &size, 0, read_msg);

	/* now commit the outstanding write, and consume a portion */
	cafe_log_write_commit(log, 4);

	read_msg = cafe_log_read_reserve(log, &size);
	CUT_ASSERT_NOT_NULL(read_msg);
	CUT_ASSERT_EQUAL(size, 4);
	CUT_ASSERT_EQUAL(((int8_t *) msg)[0], 1);
	CUT_ASSERT_EQUAL(((int8_t *) msg)[1], 2);
	cafe_log_read_commit(log, 2);

	/* write another word to wrap the write head */
	internal_log_write_reserve(log, 1, msg);
	CUT_ASSERT(cafe_log_validate(log));
	((int8_t *) msg)[0] = 5;

	/* make sure the previously unconsumed 2 bytes are available, but not
	   the wrapped byte.  Then read the rest of the current buffer. */
	internal_log_read_reserve(log, &size, 2, read_msg);
	CUT_ASSERT_EQUAL(((int8_t *) read_msg)[0], 3);
	CUT_ASSERT_EQUAL(((int8_t *) read_msg)[1], 4);
	cafe_log_read_commit(log, size);

	/* read the rest of the wrapped content */
	cafe_log_write_commit(log, 1);
	internal_log_read_reserve(log, &size, 1, read_msg);
	CUT_ASSERT_EQUAL(((int8_t *) read_msg)[0], 5);
	cafe_log_read_commit(log, size);

	CUT_ASSERT(cafe_log_validate(log));
	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}


//td: may need barriers to work well on smp systems
static bool writer_done = false;
static bool reader_failed = false;
static size_t bytes_written = 0;
static int8_t *input_values;
static size_t input_size;


static
void dump_input(size_t mark)
{
	size_t i;

	printf("Input:\n");
	for (i = 0; i < input_size; ++i) {
		printf("\t%"PRIsize_t": %hhx", i, input_values[i]);
		if (i == mark)
			printf("  <-----");
		printf("\n");
	}
}


static
void *reader_thread(void *arg)
{
	struct cafe_log_buf *buf = (struct cafe_log_buf *) arg;
	int8_t *msg;
	size_t size;
	size_t bytes_read = 0;

	do {
		msg = (int8_t *) cafe_log_read_reserve(buf, &size);
		if (size > 0) {
			size_t i;

			for (i = 0; i < size; ++i) {
				if (msg[i] != input_values[bytes_read + i]) {
					cafe_dump_log(buf);
					dump_input(bytes_read + i);
					reader_failed = true;
				}
			}
			cafe_log_read_commit(buf, size);
			bytes_read += size;
		} else {
			(void)cafe_sched_yield();
		}
	} while ((!writer_done) || (bytes_read != bytes_written));

	return 0;
}


/**
 *
 */
static
int test_log_concurrency(void)
{
	struct cafe_log_buf *buf;
	void *msg;
	int8_t *trav;
	size_t i, size, write_size;
	pthread_t reader;
	struct timespec tv;

	/* reset shared values */
	writer_done = false;
	reader_failed = false;
	bytes_written = 0;

	/* seed the number generator */
	if (clock_gettime(CLOCK_REALTIME, &tv))
		perror("error seeding with time");
	else
		srand((unsigned int)tv.tv_nsec);

	do {
		/* limit ourselves to using ~ a page */
		size = rand() % getpagesize();
	} while (size <= 1);

	buf = cafe_log_create(size, cafe_stdlib_allocator);
	CUT_ASSERT_NOT_NULL(buf);
	CUT_ASSERT(cafe_log_validate(buf));

	/* create the test input */
	input_size = 10 * size;
	input_values = malloc(input_size);
	CUT_ASSERT_NOT_NULL(input_values);
	trav = input_values;
	for (i = 0; i < input_size; ++i)
		input_values[i] = (int8_t) (rand() % CHAR_MAX);

	/* spawn the reader thread */
	if (pthread_create(&reader, 0, reader_thread, buf)) {
		perror("failed to create reader thread");
		CUT_FAIL("failed to create reader thread");
	}

	/* write */
	while (bytes_written < input_size) {
		/* attempt to write between 1 and size/2 bytes */
		write_size = (rand() % (size >> 1)) + 1;
		if (write_size > (input_size - bytes_written)) {
			write_size = input_size - bytes_written;
		}

		/* spin until there's enough space */
		msg = cafe_log_write_reserve(buf, write_size);
		while (0 == msg) {
			(void)cafe_sched_yield();
			msg = cafe_log_write_reserve(buf, write_size);
		}

		for (i = 0; i < write_size; ++i) {
			trav[i] = (int8_t) (rand() % CHAR_MAX);
		}
		memcpy(msg, trav, write_size);
		trav += write_size;
		bytes_written += write_size;
		cafe_log_write_commit(buf, write_size);
	}

	writer_done = true;

	if (pthread_join(reader, 0)) {
		perror("Error waiting for the reader thread");
	}

	free(input_values);
	CAFE_FREE(cafe_stdlib_allocator, buf);
	CUT_ASSERT(!reader_failed);
	return 0;
}


/**
 *  Make sure the reserve_min() routine behaves as expected.
 */
static
int test_log_reserve_min(void)
{
	static const unsigned int log_size = 4 * sizeof(int) + 1;
	struct cafe_log_buf *log =
		cafe_log_create(log_size, cafe_stdlib_allocator);
	void *msg;
	size_t size;

	CUT_ASSERT_NOT_NULL(log);
	CUT_ASSERT(cafe_log_validate(log));

	/* first the simple cases */
	internal_log_write_reserve_min(log, 0, &size, log_size, msg);
	internal_log_write_reserve_min(log, log_size, &size, log_size, msg);
	internal_log_write_reserve_min(log, log_size, &size, log_size, msg);
	CUT_ASSERT_NULL(cafe_log_write_reserve_min(log, log_size + 1, &size));

	/* advance the write head and make sure the remaining space and
	   wrap-around space are used appropriately */
	cafe_log_write_commit(log, 3 * sizeof(int));
	internal_log_write_reserve_min(log, 0, &size,
				       log_size - 3 * sizeof(int), msg);
	CUT_ASSERT_NULL(cafe_log_write_reserve_min(log, 1 * sizeof(int) + 2,
						   &size));
	cafe_log_read_commit(log, 3 * sizeof(int));
	CUT_ASSERT_NULL(cafe_log_write_reserve_min(log, 3 * sizeof(int) + 1,
						   &size));
	internal_log_write_reserve_min(log, 1 * sizeof(int) + 2,
				       &size, 3 * sizeof(int) - 1, msg);

	CAFE_FREE(cafe_stdlib_allocator, log);
	return 0;
}

//td: test unaligned reads and writes

CAFE_BEGIN_SUITE_HARNESS(log_buffer_suite)
	CUT_RUN_TEST(test_log_create);
	CUT_RUN_TEST(test_log_write);
	CUT_RUN_TEST(test_log_read);
	CUT_RUN_TEST(test_log_writewrap);
	CUT_RUN_TEST(test_log_readwrap);
	CUT_RUN_TEST(test_log_writewrap_truncated);
	CUT_RUN_TEST(test_log_simple_concurrency);
	CUT_RUN_TEST(test_log_concurrency);
	CUT_RUN_TEST(test_log_reserve_min);
CAFE_END_SUITE_HARNESS
