/*
 * Copyright (c) 2010-2011, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
08jul14,sdt  Include cafe CUnitest header - WB4-3658
*/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <test/cafe_CUnitTest.h>
#include <cafe/hashtbl.h>
#include <cafe/cafe.h>

static int ht_size = 0;

struct test_val {
	int x[13];
	float f;
	char c;
	int v;
	long long int v64;
};

struct test_key {
	char *s[10];
	float fval;
	char c;
	int k;
};

static unsigned long key_hash(const void *k)
{
	return (unsigned long)((struct test_key *)k)->k;
}

static int key_equals(const void *a, const void *b)
{
	return ((struct test_key *)a)->k == ((struct test_key *)b)->k;
}

/* Test basic hash table creation/clear/size. */

static int test1(void)
{
	struct cafe_hashtbl *h;
	struct cafe_hashtbl_iter iter;

	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     key_hash, key_equals);
	CUT_ASSERT_NOT_NULL(h);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_clear(h, NULL, NULL, NULL, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_iter_init(h, &iter);
	CUT_ASSERT_FALSE(cafe_hashtbl_iter_next(h, &iter));
	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);
	return 0;
}

/* Test lookup of non-existent key. */

static int test2(void)
{
	struct test_key k;
	struct cafe_hashtbl *h;
	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     key_hash, key_equals);

	CUT_ASSERT_NOT_NULL(h);
	memset(&k, 0, sizeof(k));
	k.k = 2;
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_NULL(cafe_hashtbl_lookup(h, &k));
	cafe_hashtbl_clear(h, NULL, NULL, NULL, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);
	return 0;
}

/* Test lookup of key insert. */

static int test3(void)
{
	struct test_key k;
	struct test_val v;
	struct cafe_hashtbl *h;
	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     key_hash, key_equals);

	CUT_ASSERT_NOT_NULL(h);
	memset(&k, 0, sizeof(k));
	memset(&v, 0, sizeof(v));
	k.k = 3;
	v.v = 300;
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, &k, &v));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(&v, cafe_hashtbl_lookup(h, &k));
	CUT_ASSERT_EQUAL(300, ((struct test_val *)cafe_hashtbl_lookup(h, &k))->v);
	cafe_hashtbl_clear(h, NULL, NULL, NULL, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);
	return 0;
}

/* Test lookup of multiple key insert. */

static int test4(void)
{
	struct test_key k1, k2;
	struct test_val v1, v2;
	struct cafe_hashtbl *h;
	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     key_hash, key_equals);
	CUT_ASSERT_NOT_NULL(h);
	memset(&k1, 0, sizeof(k1));
	memset(&k2, 0, sizeof(k2));
	memset(&v1, 0, sizeof(v1));
	memset(&v2, 0, sizeof(v2));
	k1.k = 3;
	v1.v = 300;
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, &k1, &v1));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(&v1, cafe_hashtbl_lookup(h, &k1));
	CUT_ASSERT_EQUAL(300, ((struct test_val *)cafe_hashtbl_lookup(h, &k1))->v);

	cafe_hashtbl_clear(h, NULL, NULL, NULL, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));

	k2.k = 4;
	v2.v = 400;
	CUT_ASSERT_NOT_NULL(h);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, &k2, &v2));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(&v2, cafe_hashtbl_lookup(h, &k2));
	CUT_ASSERT_EQUAL(400, ((struct test_val *)cafe_hashtbl_lookup(h, &k2))->v);

	cafe_hashtbl_clear(h, NULL, NULL, NULL, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));

	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);
	return 0;
}

/* Test multiple key lookups. */

static int test5(void)
{
	struct test_key k1, k2;
	struct test_val v1, v2;
	struct cafe_hashtbl *h;
	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     key_hash, key_equals);
	CUT_ASSERT_NOT_NULL(h);

	memset(&k1, 0, sizeof(k1));
	memset(&k2, 0, sizeof(k2));
	memset(&k1, 0, sizeof(k1));
	memset(&k2, 0, sizeof(k2));

	k1.k = 3;
	v1.v = 300;
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, &k1, &v1));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(&v1, cafe_hashtbl_lookup(h, &k1));
	CUT_ASSERT_EQUAL(300, ((struct test_val *)cafe_hashtbl_lookup(h, &k1))->v);
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));

	k2.k = 4;
	v2.v = 400;
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, &k2, &v2));
	CUT_ASSERT_EQUAL(2, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(&v2, cafe_hashtbl_lookup(h, &k2));
	CUT_ASSERT_EQUAL(400, ((struct test_val *)cafe_hashtbl_lookup(h, &k2))->v);

	CUT_ASSERT_EQUAL(&v1, cafe_hashtbl_lookup(h, &k1));
	CUT_ASSERT_EQUAL(300, ((struct test_val *)cafe_hashtbl_lookup(h, &k1))->v);

	cafe_hashtbl_clear(h, NULL, NULL, NULL, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));

	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);
	return 0;
}

/* Test key/value remove(). */

static int test7(void)
{
	struct test_key k;
	struct test_val v;
	struct cafe_hashtbl *h;

	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     key_hash, key_equals);

	CUT_ASSERT_NOT_NULL(h);

	memset(&k, 0, sizeof(k));
	memset(&v, 0, sizeof(v));

	k.k = 3;
	v.v = 300;
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, &k, &v));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(&v, cafe_hashtbl_lookup(h, &k));
	CUT_ASSERT_EQUAL(300, ((struct test_val *)cafe_hashtbl_lookup(h, &k))->v);
	CUT_ASSERT_EQUAL(&v, cafe_hashtbl_remove(h, &k, NULL, NULL));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_remove(h, &k, NULL, NULL));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_clear(h, NULL, NULL, NULL, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);

	return 0;
}

static void hash_free(void *ptr, void *arg)
{
	free(ptr);
}

/* Test key insert and remove with malloc'ed keys and values. */

static int test8(void)
{
	struct cafe_hashtbl *h;
	struct test_key *k = malloc(sizeof(struct test_key));
	struct test_val *v1 = malloc(sizeof(struct test_val));
	struct test_val *v2 = malloc(sizeof(struct test_val));

	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     key_hash, key_equals);

	CUT_ASSERT_NOT_NULL(h);
	CUT_ASSERT_NOT_NULL(k);
	CUT_ASSERT_NOT_NULL(v1);
	CUT_ASSERT_NOT_NULL(v2);

	memset(k, 0, sizeof(*k));
	memset(v1, 0, sizeof(*v1));
	memset(v2, 0, sizeof(*v2));

	k->k = 3;
	v1->v = 300;
	v2->v = 600;

	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, k, v1));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(v1, cafe_hashtbl_lookup(h, k));
	CUT_ASSERT_EQUAL(300, ((struct test_val *)cafe_hashtbl_lookup(h, k))->v);

	/* Replace value for same key. */
	CUT_ASSERT_EQUAL(v1, cafe_hashtbl_replace(h, k, v2));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(v2, cafe_hashtbl_lookup(h, k));
	CUT_ASSERT_EQUAL(600, ((struct test_val *)cafe_hashtbl_lookup(h, k))->v);

	CUT_ASSERT_EQUAL(v2, cafe_hashtbl_remove(h, k, hash_free, NULL));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_clear(h, hash_free, NULL, hash_free, NULL);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	cafe_hashtbl_delete(h, hash_free, NULL, hash_free, NULL);
	free(v1);
	free(v2);
	return 0;
}

/* Test iterator. */

static int test9(void)
{
	int i, sum = 0;
	struct cafe_hashtbl *h;
	static int keys[] = { 100, 200, 300 };
	struct cafe_hashtbl_iter iter;

	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     cafe_hashtbl_direct_hash,
			     cafe_hashtbl_direct_equals);

	CUT_ASSERT_NOT_NULL(h);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));

	for (i = 0; i < (int)CAFE_NELEMENTS(keys); i++) {
		CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, &keys[i], NULL));
	}

	CUT_ASSERT_EQUAL(3, cafe_hashtbl_count(h));
	cafe_hashtbl_iter_init(h, &iter);
	CUT_ASSERT_TRUE(cafe_hashtbl_iter_next(h, &iter));
	sum += *(int *)iter.key;
	CUT_ASSERT_TRUE(cafe_hashtbl_iter_next(h, &iter));
	sum += *(int *)iter.key;
	CUT_ASSERT_TRUE(cafe_hashtbl_iter_next(h, &iter));
	sum += *(int *)iter.key;
	CUT_ASSERT_FALSE(cafe_hashtbl_iter_next(h, &iter));
	CUT_ASSERT_EQUAL(600, sum);

	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);
	return 0;
}

/*
 * Test hash function where the hash value is > uint32_t (which will
 * only validate on 64-bit builds).
 */

static unsigned long test10_hash(const void *k)
{
	return LONG_MAX;
}

static int test10_equals(const void *a, const void *b)
{
	return strcmp((const char *)a, (const char *)b) == 0;
}

static int test10(void)
{
	static const char *k[] = { "hello world" };
	struct cafe_hashtbl *h;

	h = cafe_hashtbl_new(ht_size,
			     cafe_stdlib_allocator,
			     test10_hash, test10_equals);
	CUT_ASSERT_NOT_NULL(h);
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, (void*)k, (void*)k));
	CUT_ASSERT_EQUAL(1, cafe_hashtbl_count(h));
	CUT_ASSERT_NOT_NULL(cafe_hashtbl_lookup(h, k));
	cafe_hashtbl_delete(h, NULL, NULL, NULL, NULL);
	return 0;
}

static int test11(void)
{
	size_t i;
	struct cafe_hashtbl *h;

	h = cafe_hashtbl_new(1,
			     cafe_stdlib_allocator,
			     cafe_hashtbl_int_hash,
			     cafe_hashtbl_int_equals);

	CUT_ASSERT_NOT_NULL(h);
	cafe_hashtbl_autoresize_enable(h, 75);

	for (i = 0 ; i < 1024; i++) {
		int *n;
		n = malloc(sizeof *n);
		CUT_ASSERT_NOT_NULL(n);
		*n = i;
		CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, n, n));
		CUT_ASSERT_EQUAL(i + 1, cafe_hashtbl_count(h));
	}

	CUT_ASSERT_TRUE(cafe_hashtbl_count(h) >= 11);
	CUT_ASSERT_TRUE(cafe_hashtbl_capacity(h) > 1);

	cafe_hashtbl_autoresize_disable(h);

	/* Shrink the table. */
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_resize(h, 1));
	CUT_ASSERT_EQUAL(1024, cafe_hashtbl_count(h));

	/* Check that post shrinkage the existing keys can still be
	 * found. */

	for (i = 0 ; i < 1024; i++) {
		int n = i;
		CUT_ASSERT_NOT_NULL(cafe_hashtbl_lookup(h, &n));
	}

	/* Add new keys with autoresize disabled. */

	for (i = 1024 ; i < 2048; i++) {
		int *n;
		n = malloc(sizeof *n);
		CUT_ASSERT_NOT_NULL(n);
		*n = i;
		CUT_ASSERT_EQUAL(0, cafe_hashtbl_insert(h, n, n));
		CUT_ASSERT_EQUAL(i + 1, cafe_hashtbl_count(h));
		CUT_ASSERT_EQUAL(1, cafe_hashtbl_capacity(h));
	}

	CUT_ASSERT_EQUAL(2048, cafe_hashtbl_count(h));

	for (i = 0 ; i < 2048; i++) {
		int *k = cafe_hashtbl_remove(h, &i, NULL, NULL);
		CUT_ASSERT_NOT_NULL(k);
		CUT_ASSERT_EQUAL(i, *k);
		free(k);
	}
	CUT_ASSERT_EQUAL(0, cafe_hashtbl_count(h));

	cafe_hashtbl_delete(h, hash_free, NULL, NULL, NULL);

	return 0;
}

CUT_BEGIN_TEST_HARNESS(hashtbl_suite)
CUT_RUN_TEST(test1);
CUT_RUN_TEST(test2);
CUT_RUN_TEST(test3);
CUT_RUN_TEST(test4);
CUT_RUN_TEST(test5);
CUT_RUN_TEST(test7);
CUT_RUN_TEST(test8);
CUT_RUN_TEST(test9);
CUT_RUN_TEST(test10);
CUT_RUN_TEST(test11);
CUT_END_TEST_HARNESS
