/* test_tokenizer.c - unit tests */

/* Copyright (c) 2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <test/cafe_CUnitTest.h>
#include <cafe/cafe.h>
#include <string.h>
#include "../tokenizer.h"

#define WORDCMP(T, N,EXPECTED) (strcmp((T)->wordv[(N)], (EXPECTED)) == 0)

struct test_allocator {
	struct cafe_allocator base;
	size_t nallocs;
};

static void *test_alloc(const struct cafe_allocator *a, size_t size)
{
	if (((struct test_allocator *)a)->nallocs++ > 1)
		return NULL;

	return malloc(size);
}

static void *test_realloc(const struct cafe_allocator *a, void *ptr,
			  size_t size)
{
	if (((struct test_allocator *)a)->nallocs++ > 1)
		return NULL;

	return realloc(ptr, size);
}

static void test_free(const struct cafe_allocator *a, void *ptr)
{
	if (ptr != NULL)
		free(ptr);
}

static struct test_allocator test_allocator = {
	{ test_alloc, test_realloc, test_free }
};

static int test1(void)
{
	int rc;
	const struct cafe_allocator *a = cafe_stdlib_allocator;
	struct cafe_tokenizer t;

	cafe_tokenizer_init(&t);

	rc = cafe_tokenize(&t, "", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(0, t.wordc);
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "a", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "aaa", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "aaa"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "a b", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(2, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	CUT_ASSERT_TRUE(WORDCMP(&t, 1, "b"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "a b c d e f g h i j k l m", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(13, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	CUT_ASSERT_TRUE(WORDCMP(&t, 12, "m"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, " a b ", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(2, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	CUT_ASSERT_TRUE(WORDCMP(&t, 1, "b"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "aaa bbb", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(2, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "aaa"));
	CUT_ASSERT_TRUE(WORDCMP(&t, 1, "bbb"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\f\n\r\t\va", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "a\f\n\r\t\v", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "a\f\n\r\t\v b", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(2, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	CUT_ASSERT_TRUE(WORDCMP(&t, 1, "b"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"Cafe\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "Cafe"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"Cafe Rocks!\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "Cafe Rocks!"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(0, t.wordc);
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_QUOTED_WORD_NOT_TERMINATED, rc);
	CUT_ASSERT_EQUAL(0, t.wordc);
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\\\t", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_ESCAPED_CHARACTER_IS_ILLEGAL, rc);
	CUT_ASSERT_EQUAL(0, t.wordc);
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"\t\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "\t"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"Cafe Rocks!\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "Cafe Rocks!"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"\\\"Cafe Rocks!\\\"\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "\"Cafe Rocks!\""));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"Cafe \\\"Rocks!\\\"\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "Cafe \"Rocks!\""));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "foo\"Cafe \"Rocks!\"\"bar", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "fooCafe Rocks!bar"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "\"Cafe \\\"Really\\\"\" Rocks!", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(2, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "Cafe \"Really\""));
	CUT_ASSERT_TRUE(WORDCMP(&t, 1, "Rocks!"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "typedef N[vx_mem_nature_def,,n[cafe.METRIC_CATEGORY,\\\"vx.mem\\\",\\\"Memory\\\",\\\"\\\",\\\"platform:/plugin/com.windriver.analysis.cafe.instruments/icons/memory.gif\\\"]]", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(2, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "typedef"));
	CUT_ASSERT_TRUE(WORDCMP(&t, 1, "N[vx_mem_nature_def,,n[cafe.METRIC_CATEGORY,\"vx.mem\",\"Memory\",\"\",\"platform:/plugin/com.windriver.analysis.cafe.instruments/icons/memory.gif\"]]"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "table \"S[summary_table__265,v[_vx_function_context,t[ULONG],n[cafe.SET]]v[vx_base_mallocd_bytes__cafe_SUM_METRIC_VARIANT,t[ULONGLONG],n[cafeui.UI_METRIC]n[cafe.SUM_METRIC_VARIANT,\\\"?\\\"]n[cafe.SUM]n[vx_mem_nature_def]n[cafeui.FORMATTER,\\\"formatter.byte\\\",\\\"\\\",\\\"\\\",\\\"2\\\"]n[cafe.LABEL,\\\"?Mallocd Bytes\\\"]n[cafeui.UI_VISIBLE]],n[cafe.TABLE_NAME,\\\"summary_table\\\"]]\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(2, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "table"));
	CUT_ASSERT_TRUE(WORDCMP(&t, 1, "S[summary_table__265,v[_vx_function_context,t[ULONG],n[cafe.SET]]v[vx_base_mallocd_bytes__cafe_SUM_METRIC_VARIANT,t[ULONGLONG],n[cafeui.UI_METRIC]n[cafe.SUM_METRIC_VARIANT,\"?\"]n[cafe.SUM]n[vx_mem_nature_def]n[cafeui.FORMATTER,\"formatter.byte\",\"\",\"\",\"2\"]n[cafe.LABEL,\"?Mallocd Bytes\"]n[cafeui.UI_VISIBLE]],n[cafe.TABLE_NAME,\"summary_table\"]]"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "vxWorks.memFree nBytesRequested n[cafe.COUNTER_METRIC_CLASS] \"n[cafe.LABEL,\\\"# \\\"free\\\"\\\"]\"", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(4, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 3, "n[cafe.LABEL,\"# \"free\"\"]"));
	cafe_tokenizer_reset(&t);

	cafe_tokenizer_release(&t, cafe_stdlib_allocator);

	return 0;
}

static int test2(void)
{
	int rc;
	const struct cafe_allocator *a = &test_allocator.base;
	struct cafe_tokenizer t;

	cafe_tokenizer_init(&t);

	rc = cafe_tokenize(&t, "", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(0, t.wordc);
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "this should fail with out of memory!", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_ENOMEM, rc);
	cafe_tokenizer_reset(&t);

	cafe_tokenizer_release(&t, a);

	return 0;
}

/* Test comment handling. */

static int test3(void)
{
	int rc;
	const struct cafe_allocator *a = cafe_stdlib_allocator;
	struct cafe_tokenizer t;

	cafe_tokenizer_init(&t);

	rc = cafe_tokenize(&t, "#", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(0, t.wordc);
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "# a b c", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(0, t.wordc);
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "a#", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	cafe_tokenizer_reset(&t);

	rc = cafe_tokenize(&t, "a #", a);
	CUT_ASSERT_EQUAL(CAFE_TOKENIZER_SUCCESS, rc);
	CUT_ASSERT_EQUAL(1, t.wordc);
	CUT_ASSERT_TRUE(WORDCMP(&t, 0, "a"));
	cafe_tokenizer_reset(&t);

	cafe_tokenizer_release(&t, cafe_stdlib_allocator);

	return 0;
}

CUT_BEGIN_TEST_HARNESS(tokenizer_suite)
CUT_RUN_TEST(test1);
CUT_RUN_TEST(test2);
CUT_RUN_TEST(test3);
CUT_END_TEST_HARNESS
