/* test_vm_dynarray.c - cafe dynamic array unit tests */

/* Copyright (c) 2010-2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <test/cafe_CUnitTest.h>
#include <string.h>

#define STREQ(A,B) (strcmp((A), (B)) == 0)

struct test_struct {
	int index;
	int foo;
	int bar;
	char *baz;
};

static int test_vm_dynarray(void)
{
	int rc;
	size_t i, length, capacity;
	struct cafe_vm_dynarray array;
	struct cafe_namespace *ns;
	struct cafe_allocator *allocator = cafe_stdlib_allocator;
	struct cafe_value_type *type;
	void *val;

	memset(&array, 0, sizeof array);
	CUT_ASSERT_CAFE_OK(cafe_namespace_create(allocator, &ns));
	CUT_ASSERT_CAFE_OK(cafe_type_parse(ns, "S[test_struct,v[index,i,,-1]v[foo,i]v[bar,i,,1000]v[baz,s,,\"Hello\"]]", &type));
	CUT_ASSERT_CAFE_OK(cafe_type_parse(ns, "a[t[test_struct]]", &type));

	/* Get 128 values - will grow the array. */

	for (i = 0; i < 128; i++) {
		struct test_struct *t;
		rc = cafe_vm_dynarray_elem_get(&array, type, i, &val);
		CUT_ASSERT_CAFE_OK(rc);
		t = (struct test_struct *)val;
		CUT_ASSERT_EQUAL(-1, t->index);
		CUT_ASSERT_EQUAL(0, t->foo);
		CUT_ASSERT_EQUAL(1000, t->bar);
		t->index = i;
		t->foo += 100;
		CUT_ASSERT_TRUE(array.l == i + 1);
		CUT_ASSERT_TRUE(STREQ(t->baz, "Hello"));
	}

	length = array.l;
	capacity = array.c;

	/* Get 128 values - will NOT grow the array as elements 1..128
	 * are already present. */

	for (i = 0; i < 128; i++) {
		struct test_struct *t;
		rc = cafe_vm_dynarray_elem_get(&array, type, i, &val);
		CUT_ASSERT_CAFE_OK(rc);
		t = (struct test_struct *)val;
		CUT_ASSERT_EQUAL(i, t->index);
		CUT_ASSERT_EQUAL(100, t->foo);
		CUT_ASSERT_EQUAL(1000, t->bar);
		CUT_ASSERT_TRUE(STREQ(t->baz, "Hello"));
		CUT_ASSERT_TRUE(array.l == length);
	}

	CUT_ASSERT_EQUAL(length, array.l);
	CUT_ASSERT_EQUAL(capacity, array.c);

	/*
	 * Add another value, but at an offset that is more than just
	 * +1 from the current length.
	 */
	length = array.l;
	CUT_ASSERT_EQUAL(length, 128);
	rc = cafe_vm_dynarray_elem_get(&array, type, 132, &val);
	CUT_ASSERT_CAFE_OK(rc);
	CUT_ASSERT_EQUAL(array.l, 133);
	length = array.l;

	for (i = 128; i <= 132; i++) {
		struct test_struct *t;
		rc = cafe_vm_dynarray_elem_get(&array, type, i, &val);
		CUT_ASSERT_CAFE_OK(rc);
		CUT_ASSERT_EQUAL(array.l, length);
		t = (struct test_struct *)val;
		CUT_ASSERT_EQUAL(-1, t->index);
		CUT_ASSERT_EQUAL(0, t->foo);
		CUT_ASSERT_EQUAL(1000, t->bar);
		CUT_ASSERT_TRUE(STREQ(t->baz, "Hello"));
		t->index = i;
	}
	CUT_ASSERT_EQUAL(133, array.l);

	/* Check existing values haven't changed. */

	length = array.l;
	
	for (i = 0; i < 128; i++) {
		struct test_struct *t;
		rc = cafe_vm_dynarray_elem_get(&array, type, i, &val);
		CUT_ASSERT_CAFE_OK(rc);
		t = (struct test_struct *)val;
		CUT_ASSERT_EQUAL(100, t->foo);
		CUT_ASSERT_EQUAL(1000, t->bar);
		CUT_ASSERT_TRUE(STREQ(t->baz, "Hello"));
		CUT_ASSERT_TRUE(array.l == length);
	}

	capacity = array.c;

	/* Get elements up to 256.
	 *
	 * The first 128 values should remain unchanged.
	 */
	for (i = 0; i < 256; i++) {
		struct test_struct *t;
		rc = cafe_vm_dynarray_elem_get(&array, type, i, &val);
		CUT_ASSERT_CAFE_OK(rc);
		t = (struct test_struct *)val;
		if (i < 128) {
			CUT_ASSERT_EQUAL(i, t->index);
			CUT_ASSERT_EQUAL(100, t->foo);
			CUT_ASSERT_EQUAL(1000, t->bar);
		} else if (i >= 128 && i < 133) {
			CUT_ASSERT_EQUAL(i, t->index);
			CUT_ASSERT_EQUAL(0, t->foo);
			CUT_ASSERT_EQUAL(1000, t->bar);
		} else {
			t->index = i;
			CUT_ASSERT_EQUAL(0, t->foo);
			CUT_ASSERT_EQUAL(1000, t->bar);
		}
		CUT_ASSERT_TRUE(STREQ(t->baz, "Hello"));
	}
	CUT_ASSERT_EQUAL(256, array.l);

	/* Get discrete elements that are far apart to force the
	 * underlying table to grow. */

	CUT_ASSERT_TRUE(array.l < 513);
	rc = cafe_vm_dynarray_elem_get(&array, type, 513, &val);
	CUT_ASSERT_CAFE_OK(rc);
	CUT_ASSERT_EQUAL(514, array.l);

	CUT_ASSERT_TRUE(array.l < 1029);
	CUT_ASSERT_TRUE(array.c < 1029);
	rc = cafe_vm_dynarray_elem_get(&array, type, 1029, &val);
	CUT_ASSERT_CAFE_OK(rc);
	CUT_ASSERT_EQUAL(1030, array.l);

	CUT_ASSERT_TRUE(array.l < 2049);
	CUT_ASSERT_TRUE(array.c < 2049);
	rc = cafe_vm_dynarray_elem_get(&array, type, 2049, &val);
	CUT_ASSERT_CAFE_OK(rc);
	CUT_ASSERT_EQUAL(2050, array.l);

	length = array.l;
	capacity = array.c;

	/* Change the string field in each element and update the
	 * index field for those elements where the value has not yet
	 * been set. */

	for (i = 0; i < array.l; i++) {
		char buf[64];
		struct test_struct *t;
		rc = cafe_vm_dynarray_elem_get(&array, type, i, &val);
		CUT_ASSERT_CAFE_OK(rc);
		t = (struct test_struct *)val;
		CUT_ASSERT_TRUE(length == array.l);
		CUT_ASSERT_TRUE(capacity == array.c);
		CUT_ASSERT_TRUE(STREQ(t->baz, "Hello"));
		sprintf(buf, "%u", (unsigned int)i);
		CUT_ASSERT_CAFE_OK(cafe_value_store_string(&t->baz, buf, allocator));
		if (i < 256) {
			CUT_ASSERT_EQUAL(i, t->index);
		} else {
			CUT_ASSERT_EQUAL(-1, t->index);
			t->index = i;
		}
	}

	/* Check new strings match and that all elements have a valid
	 * index field. */

	for (i = 0; i < array.l; i++) {
		char buf[64];
		struct test_struct *t;
		rc = cafe_vm_dynarray_elem_get(&array, type, i, &val);
		CUT_ASSERT_CAFE_OK(rc);
		t = (struct test_struct *)val;
		CUT_ASSERT_TRUE(length == array.l);
		CUT_ASSERT_TRUE(capacity == array.c);
		CUT_ASSERT_TRUE(!STREQ(t->baz, "Hello"));
		sprintf(buf, "%u", (unsigned int)i);
		CUT_ASSERT_TRUE(STREQ(t->baz, buf));
		CUT_ASSERT_EQUAL(i, t->index);
	}

	cafe_vm_dynarray_set_size(&array, 0, type);
	cafe_namespace_delete(ns);

	return 0;
}

CUT_BEGIN_TEST_HARNESS(vm_dynarray_suite)
CUT_RUN_TEST(test_vm_dynarray);
CUT_END_TEST_HARNESS
