/* tokenizer.c - tokenize lines into words similar to a UNIX shell. */

/* Copyright (c) 2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <cafe/cafe.h>
#include "tokenizer.h"

static int wordv_capacity_set(struct cafe_tokenizer *t,
			      size_t new_capacity,
			      const struct cafe_allocator *allocator)
{
	void *new_data;

	if (new_capacity < t->wordv_capacity)
		return 0;

	while (new_capacity <= t->wordv_capacity)
		new_capacity *= 2;

	assert(new_capacity > t->wordv_capacity);

	new_data = CAFE_REALLOC(allocator, t->wordv,
				new_capacity * sizeof *t->wordv);

	if (new_data == NULL)
		return -1;

	t->wordv_capacity = new_capacity;
	t->wordv = new_data;

	return 0;
}

static int terminate_word(struct cafe_tokenizer *t,
			  const struct cafe_allocator *allocator)
{
	int rc = 0;

	*t->wptr = '\0';	/* terminate current word */

	/* words must be at least 1 character in length. */

	if (t->wptr != t->word_start) {
		rc = wordv_capacity_set(t, 2 + t->wordc, allocator);
		if (rc == 0) {
			t->wordv[t->wordc++] = t->word_start;
			t->wordv[t->wordc] = NULL;
			t->word_start = ++t->wptr;
		}
	}

	return rc;
}

void cafe_tokenizer_init(struct cafe_tokenizer *t)
{
	t->wordc = 0;
	t->wordv = NULL;
	t->wordv_capacity = 0;
	t->word_start = NULL;
	t->wptr = NULL;
	t->buffer_max = NULL;
	t->quote = 0;
	t->comment = 0;
	t->buffer = NULL;
}

void cafe_tokenizer_reset(struct cafe_tokenizer *t)
{
	t->word_start = t->buffer; /* beginning of word */
	t->wptr = t->buffer;
	t->quote = 0;
	t->comment = 0;
	t->wordc = 0;
}

void cafe_tokenizer_release(struct cafe_tokenizer *t,
			    const struct cafe_allocator *a)
{
	CAFE_FREE(a, t->wordv);
	CAFE_FREE(a, t->buffer);
}

cafe_tokenizer_result_t cafe_tokenize(struct cafe_tokenizer *t,
				      const char *input,
				      const struct cafe_allocator *a)
{
	for (;;) {
		int c = *input++;

		if (t->wptr >= t->buffer_max) {
			size_t i;
			char *newp;
			size_t newsz;
			size_t oldsz = t->buffer_max - t->buffer;

			newsz = (oldsz == 0) ? 32 : oldsz * 2;
			newp = CAFE_REALLOC(a, t->buffer, newsz);
			if (newp == NULL)
				return CAFE_TOKENIZER_ENOMEM;
			for (i = 0; i < t->wordc; i++)
				t->wordv[i] = newp + (t->wordv[i] - t->buffer);
			t->wptr = newp + (t->word_start - t->buffer) + (t->wptr - t->word_start);
			t->word_start = newp + (t->word_start - t->buffer);
			t->buffer_max = newp + newsz;
			t->buffer = newp;
		}

		if (c == '\0') {
			if (t->quote)
				return CAFE_TOKENIZER_QUOTED_WORD_NOT_TERMINATED;
			if (terminate_word(t, a) != 0)
				return CAFE_TOKENIZER_ENOMEM;
			return CAFE_TOKENIZER_SUCCESS;
		}

		if (t->comment)
			continue;

		if (isspace(c) && !t->quote) {
			if (terminate_word(t, a) != 0)
				return CAFE_TOKENIZER_ENOMEM;
			continue;
		}

		if (!t->quote) {
			if (c == '#') {
				t->comment = 1;
				continue;
			}
		}

		if (c == '\\') {
			c = *input++;
			switch (c) {
			case '\r':
			case '\n':
				*t->wptr++ = '\\';
				continue;
			case '\\':
			case '"':
				break;
			default:
				return CAFE_TOKENIZER_ESCAPED_CHARACTER_IS_ILLEGAL;
			}
			*t->wptr++ = (char)c;
			continue;
		}

		if (c == '"') {
			t->quote = 1 - t->quote;
			continue;
		}

		*t->wptr++ = (char)c;
	}
}
