/* Copyright (c) 2010 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <hpc/hpc_drv/hpcHwDrvApi.h>
#include <hpc/drv_evt_cfg/hpcDrvEvtCfg.h>
#include <hpc/hpc_drv/ppc/e500Drv.h>
#include <hpc/drv_evt_cfg/ppc/e500DrvSet.h>
#include <hpc/drv_evt_cfg/ppc/e500DrvEvtCfg.h>

#include "../../hpc_log.h"

/******************************************************************************
* Function to decode the E500 event description string. Example descriptions
* are:
* "Com:45" - this is a common counter with event code 45, shared across counters
* "Ref:46" - this is a reference counter with event code 46, shared across
*            counters
* "C12:47" - this is a counter specific counter (counter 12) with event code 47.
* "C0,1,2,4:48" - this is a counter specific counter (counters 0, 1, 2 and 4)
*                 with event code 48.
*/

static int eventDescGet
    (
    const char *	pStr,	        /* event description string to parse */
    unsigned int *	pEvent,         /* out: event code */
    unsigned int *	pCtrMask,       /* out: counter bit mask */
    int *		pCtrSpecificEvt /* out: 1 = counter specific event, */
				        /* else 0. Param can be NULL. */
    )
    {
#define TOK_MAX 50
    char tok1[TOK_MAX]; /* string before the ":" */
    char tok2[TOK_MAX]; /* string following the ":" */
    char ctr[TOK_MAX];
    int out = 0, in = 0;
    int counter;

    *pEvent = *pCtrMask = 0;

    if (pCtrSpecificEvt != NULL)
        *pCtrSpecificEvt = 0;

    while (pStr[in] != ':')
        {
        tok1[out++] = pStr[in++];
        if (out == TOK_MAX)
            goto error;
        }
    tok1[out] = '\0';

    in++;
    out = 0;
    while (pStr[in] != '\0')
        {
        tok2[out++] = pStr[in++];
        if (out == TOK_MAX)
            goto error;
        }
    tok2[out] = '\0';

    if ((strcmp(tok1, "Com") == 0 || strcmp(tok1, "Ref") == 0))
        {
        *pCtrMask = ~0; /* All counters: PMC0-PMC31 */
        }
    else if (strcmp(tok1, "Ref1") == 0)
        {
        *pCtrMask = 0xfffffffe; /* not PMC0, PMC1-PMC31  */
        }
    else
        {
        if (pCtrSpecificEvt != NULL)
            *pCtrSpecificEvt = 1;
        in = 1;
        do
            {
            out = 0;
            while (tok1[in] != '\0' && tok1[in] != ',')
                ctr[out++] = tok1[in++];
            ctr[out] = '\0';
            if ((counter = atoi(ctr)) > 31)
                goto error;
            *pCtrMask |= 1 << counter;
            }
        while (tok1[in++] != '\0');
        }
    *pEvent = atoi(tok2);
    return CAFE_OK;
error:
    *pEvent = *pCtrMask = 0;
    return CAFE_ERROR;
    }

static int e500EventSettingsGet
    (
    struct cafe_value_type * pType,
    union eventSettings * pDrvSettings,
    struct hpcEventSettings * pHpcSettings,
    const struct hpcEventDescriptor *pEvtDesc,
    int *		pCtrSpecificEvt
    )
    {
    struct e500EventDescription *e500EvtDesc = (struct e500EventDescription *)pEvtDesc;
    const char *pEvtDescStr;
    int rc;

    /* get fixed settings from event description */
    pEvtDescStr = e500EvtDesc->eventDescStr;

    rc = eventDescGet(pEvtDescStr, &pDrvSettings->e500.evtCode,
		    &pDrvSettings->e500.ctrMask, pCtrSpecificEvt);

    if (rc != CAFE_OK)
        {
        hpc_error("Unable to decode e500 event description: %s %s\n", e500EvtDesc->eventVtStr, pEvtDescStr);
        return rc;
        }

    if (pHpcSettings->reload < e500EvtDesc->hpcEventSettingsConst.reloadMin)
        pHpcSettings->reload = e500EvtDesc->hpcEventSettingsConst.reloadMin;

    return CAFE_OK;
    }

int e500CoreEventSettingsGet
    (
    struct cafe_value_type * pType,
    union eventSettings * pDrvSettings,
    struct hpcEventSettings * pHpcSettings,
    const struct hpcEventDescriptor *pEvtDesc
    )
    {
    return e500EventSettingsGet(pType, pDrvSettings, pHpcSettings, pEvtDesc, NULL);
    }

int e500DevEventSettingsGet
    (
    struct cafe_value_type * pType,
    union eventSettings * pDrvSettings,
    struct hpcEventSettings * pHpcSettings,
    const struct hpcEventDescriptor *pEvtDesc
    )
    {
    int ctrSpecificEvt;
    int rc;

    if ((rc = e500EventSettingsGet(pType, pDrvSettings, pHpcSettings, pEvtDesc, &ctrSpecificEvt)) == CAFE_OK)
        {
        /* MPC peripheral device Counter Specific Events are offset by 64 */
        if (ctrSpecificEvt)
            pDrvSettings->e500.evtCode += 64;
        }

    return rc;
    }
