/* Copyright (c) 2010 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include <hpc/hpc_drv/hpcHwDrvApi.h>
#include <hpc/hpc_drv/i86/i86ArchPm1Drv.h>

/* Array index for a 2 dimensional array */
#define A2D(ROW,COLUMN,NCOLUMNS) (ROW * NCOLUMNS + COLUMN)

struct hpcHwCtrData
    {
    void *		pClient;
    HPC_COUNTER		countValue;
    int			counter;
    HPC_COUNTER		reloadValue;
    unsigned int	pmlcaValue;
    HPC_COUNTER		counterWrap;
    long long int	perfEvtSel;
    };

/* This macro is used to cast HPC_COUNTER type for use with pentiumMsrGet(). */
#define HPC_CAST(V) ((long long int *)&V)

/*
 * Architecture Performance Monitoring Version 1 registers.
 * Some events can only be counted on specific counters. For example event
 * FP_COMP_OPS_EXE.
 */

/* PERFEVTSEL register bit layout */
#define PERFEVTSEL_CMASK_MASK 0xff
#define PERFEVTSEL_UMASK_MASK 0xff
#define PERFEVTSEL_EVENT_MASK 0xff
#define PERFEVTSEL_INV_MASK 0x1

#define PERFEVTSEL_CMASK  24
#define PERFEVTSEL_INV    23
#define PERFEVTSEL_EN     22
#define PERFEVTSEL_ANY    21
#define PERFEVTSEL_INT    20
#define PERFEVTSEL_PC     19
#define PERFEVTSEL_E      18
#define PERFEVTSEL_OS     17
#define PERFEVTSEL_USR    16
#define PERFEVTSEL_UMASK  8
#define PERFEVTSEL_EVENT  0

static STATUS hpcHwDriverShow
    (
    void * pDrvInst
    )
    {
    struct driverInstArchV1 * pInst = pDrvInst;
    long long int value;
    int n;

    /* Display general counters */
    for (n = 0; n < pInst->numCounters; n++)
        {
	pentiumMsrGet(pInst->msrPmc + n, &value);
	printf("PMC%d\t\t\t0x%llx\n", n, value);
	}

    for (n = 0; n < pInst->numCounters; n++)
        {
	pentiumMsrGet(pInst->msrPerfEvtSel + n, &value);
	printf("PERFEVTSEL%d\t\t0x%llx\n", n, value);
	}

    return OK;
    }

/******************************************************************************
*/

static int perfIntHandler
    (
    struct driverInstArchV1 *	pInst,
    void * 			pRegs,
    int				cpu,
    long long int		perfGlobalStatus
    )
    {
    int eventsProcessed = 0;
    unsigned int counter;
    HPC_COUNTER pmcValue;
    int ctrPerCore;

    if (pInst->archPerfMonVer == I86_ARCH_PM_DRV1_UNCORE)
        ctrPerCore = 0;
    else
        ctrPerCore = cpu;

    if ((perfGlobalStatus & pInst->pmcCounterMask) == 0)
        return 0;

    for (counter = 0; counter < pInst->numCounters; counter++)
        {
	if ((1 << counter) & perfGlobalStatus)
	    {
            long long int perfEvtSel;
	    int idx2d = A2D(ctrPerCore, counter, pInst->numCounters);
	    struct hpcHwCtrData * pCtr;

	    pCtr = pInst->ppDrvHwCtrData[idx2d];
	    if (pCtr == NULL)
		continue;

            /* stop the counter */
            perfEvtSel = 0ULL;
            pentiumMsrSet (pInst->msrPerfEvtSel + pCtr->counter, &perfEvtSel);

            /* read counter and reload it */
            pentiumMsrGet (pInst->msrPmc + pCtr->counter, HPC_CAST(pmcValue));

	    pmcValue += pInst->pmcRollover - pCtr->reloadValue;
	    (*pInst->pHandler) (pCtr->pClient, pRegs, pmcValue, cpu);

	    /* Reload the counter */
            pentiumMsrSet(pInst->msrPmc + pCtr->counter, HPC_CAST(pCtr->reloadValue));

            /* restart the counter */
            pentiumMsrSet (pInst->msrPerfEvtSel + pCtr->counter, &pCtr->perfEvtSel);

	    eventsProcessed = 1;
	    }
        }

    return eventsProcessed;
    }

/******************************************************************************
*/

static STATUS hpcHwCtrStart
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr
    )
    {
    struct driverInstArchV1 * pInst = pDrvInst;
    long long int perfEvtSel;
    size_t idx2d;

    pentiumMsrGet(pInst->msrPerfEvtSel + pCtr->counter, &perfEvtSel);

    if (perfEvtSel & (1 << PERFEVTSEL_EN))
        {
	printf ("counter %u is already running, perfEvtSel 0x%llx %s\n",
	           pCtr->counter, perfEvtSel, pInst->pDriverName);
	return ERROR;
        }

    idx2d = A2D(vxCpuIndexGet(), pCtr->counter, pInst->numCounters);
    pInst->ppDrvHwCtrData[idx2d] = pCtr;
    pentiumMsrSet (pInst->msrPmc + pCtr->counter, HPC_CAST(pCtr->countValue));
    pentiumMsrSet (pInst->msrPerfEvtSel + pCtr->counter, &pCtr->perfEvtSel);

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwCtrStop
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr,
    HPC_COUNTER *	  pCount
    )
    {
    struct driverInstArchV1 * pInst = pDrvInst;
    long long int perfEvtSel;
    size_t idx2d;

    pentiumMsrGet(pInst->msrPerfEvtSel + pCtr->counter, &perfEvtSel);

    if ((perfEvtSel & (1 << PERFEVTSEL_EN)) == 0)
        {
	printf ("counter %u is not running, perfEvtSel 0x%llx %s\n",
	           pCtr->counter, perfEvtSel, pInst->pDriverName);
	return ERROR;
        }

    /* stop the counter */
    perfEvtSel = 0;
    pentiumMsrSet (pInst->msrPerfEvtSel + pCtr->counter, &perfEvtSel);

    /* save the counters value */
    pentiumMsrGet (pInst->msrPmc + pCtr->counter, HPC_CAST(pCtr->countValue));

    idx2d = A2D(vxCpuIndexGet(), pCtr->counter, pInst->numCounters);
    pInst->ppDrvHwCtrData[idx2d] = NULL;

    if (pCount != NULL)
        *pCount = pCtr->countValue - pCtr->reloadValue;

    return OK;
    }

/******************************************************************************
* This API can be called from ISR context
*/

static STATUS hpcHwCtrRead
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr,
    HPC_COUNTER	*	  pValue
    )
    {
    struct driverInstArchV1 * pInst = pDrvInst;
    long long int perfEvtSel;
    long long int pmcValue;

    /* stop the counter */
    perfEvtSel = 0ULL;
    pentiumMsrSet (pInst->msrPerfEvtSel + pCtr->counter, &perfEvtSel);
    /* read counter value */
    pentiumMsrGet (pInst->msrPmc + pCtr->counter, &pmcValue);
    /* restart the counter */
    pentiumMsrSet (pInst->msrPerfEvtSel + pCtr->counter, &pCtr->perfEvtSel);

    /* Check if the counter has rolled over and the ISR to reload it has not
     * run yet */
    if (pmcValue >= pCtr->reloadValue)
        *pValue = pmcValue - pCtr->reloadValue;
    else
	*pValue = pmcValue + (pInst->pmcRollover - pCtr->reloadValue);

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwDeleteCtrForEvent
    (
    void *		  pDrvInst,
    struct hpcHwCtrData * pCtr
    )
    {
    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwUpdateCtrReload
    (
    void * 		  pDrvInst,
    struct hpcHwCtrData * pCtr,
    HPC_COUNTER *	  pReload
    )
    {
    struct driverInstArchV1 * pInst = pDrvInst;

    if (*pReload > pInst->pmcRollover || *pReload == 0)
        {
        /*
         * set a value that will cause maximum number of perf counter increments
         * before counter interrupt occurs.
         */
	pCtr->reloadValue = 0;
        *pReload = pInst->pmcRollover;
        return OK;
        }
    pCtr->reloadValue = pInst->pmcRollover - *pReload;
    pCtr->countValue = pCtr->reloadValue;

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwCreateCtrForEvent
    (
    void *		  pDrvInst,
    void *		  pEvtSettings,
    HPC_COUNTER *	  pReload,
    struct hpcHwCtrData * pCtr,
    void *		  pClient,
    HPC_EVENT_ITER_FUNC	  pEventIterFunc,
    void *		  pIterArg
    )
    {
    STATUS status;
    struct driverInstArchV1 * pInst = pDrvInst;
    struct i86Arch1Settings * pSettings = pEvtSettings;
    struct hpcHwCtrData * pCtrIter;
    unsigned int ctrInUseMask = 0, ctrMask;
    int counter;

    if (pDrvInst == NULL || pEvtSettings == NULL)
	return ERROR;

    /* determine which counters are already in use */
    while ((pCtrIter = (*pEventIterFunc)(pIterArg)) != NULL)
	{
	ctrInUseMask |= 1 << pCtrIter->counter;
	}

    ctrMask = pSettings->restrictedCtrMask ? pSettings->restrictedCtrMask : NOT(0);
    ctrInUseMask &= ctrMask;
    counter = hpcGetFirstBitSet(NOT(ctrInUseMask));

    if (counter >= pInst->numCounters)
        {
		fprintf(stderr, "Unable to get counter, using 0x%x out of %u (%u)\n",
			ctrInUseMask, pInst->numCounters, counter);
	return ERROR;
        }

    pCtr->counter = counter;
    pCtr->pClient = pClient;

    if ((status = hpcHwUpdateCtrReload(pInst, pCtr, pReload)) != OK)
        {
        printf("unable to get hw reload value %s\n", pInst->pDriverName);
        return status;
        }

    /* Configure value for PERFEVTSEL register */
    pCtr->perfEvtSel = (pSettings->eventNum & PERFEVTSEL_EVENT_MASK) |
	(pSettings->eventUmask & PERFEVTSEL_UMASK_MASK) << PERFEVTSEL_UMASK |
	(pSettings->ctrMask & PERFEVTSEL_CMASK_MASK) << PERFEVTSEL_CMASK |
	(pSettings->inv & PERFEVTSEL_INV_MASK) << PERFEVTSEL_INV |
	(1 << PERFEVTSEL_INT) |
	(1 << PERFEVTSEL_EN);

    if (pInst->archPerfMonVer != I86_ARCH_PM_DRV1_UNCORE)
        pCtr->perfEvtSel |=  (1 << PERFEVTSEL_USR) |
	(1 << PERFEVTSEL_OS);

    return OK;
    }

/******************************************************************************
*/

static STATUS hpcHwDriverInit
    (
    void *			 pDrvInst,
    struct hpcHwDriverInitData * pDrvInstInitData,
    size_t *			 sizeOfHwPrivData
    )
    {
    STATUS rc = OK;
    struct driverInstArchV1 * pInst = pDrvInst;
    int bit;

    pInst->pHandler = pDrvInstInitData->pPerfIntCallback;
    *sizeOfHwPrivData = sizeof(struct hpcHwCtrData);

    for (bit = 0; bit < pInst->numCounterBits; bit++)
	pInst->pmcRollover |= (1ULL << bit);

    for (bit = 0; bit < pInst->numCounters; bit++)
	pInst->pmcCounterMask |= 1 << bit;

    if (pInst->enablePerfIsr != NULL)
        if ((rc = pInst->enablePerfIsr(perfIntHandler)) != OK)
            printf("failed to enable i86 v1 perf isr %u\n", rc);

    return rc;
    }

/******************************************************************************
*/

static STATUS hpcHwDriverFinit
    (
    void * pDrvInst
    )
    {
    STATUS rc = OK;
    struct driverInstArchV1 * pInst = pDrvInst;
    int force = 1; /* WRS_TODO add this as a parameter... */

    if (pInst->disablePerfIsr != NULL)
        if ((rc = pInst->disablePerfIsr(force)) != OK)
            printf("failed to disable i86 v1 perf isr %d, %s\n", rc,
                      pInst->pDriverName);

    return rc;
    }

/******************************************************************************
*/

static void resetRegisters
    (
    struct driverInstArchV1 * pInst
    )
    {
    int n;
    long long int value = 0ULL;

    for (n = 0; n < pInst->numCounters; n++)
        {
	pentiumMsrSet(pInst->msrPmc + n, &value);
	pentiumMsrSet(pInst->msrPerfEvtSel + n, &value);
	}
    }

/******************************************************************************
*/

static STATUS hpcHwPerCoreInit
    (
    void * pDrvInst
    )
    {
    struct driverInstArchV1 * pInst = pDrvInst;
    STATUS rc = OK;

    if (pInst->enablePerfIsrPerCore != NULL)
        if ((rc = pInst->enablePerfIsrPerCore(perfIntHandler)) != OK)
            printf("failed to enable i86 v1 perf isr per core %u\n", rc);

    resetRegisters(pInst);

    return rc;
    }

/******************************************************************************
*/

static STATUS hpcHwPerCoreFinit
    (
    void * pDrvInst,
    int    force
    )
    {
    struct driverInstArchV1 * pInst = pDrvInst;
    STATUS rc = OK;

    if (pInst->disablePerfIsrPerCore != NULL)
        if ((rc = pInst->disablePerfIsrPerCore(force)) != OK)
            printf("failed to disable i86 v1 perf isr per core %d, %s\n", rc,
                      pInst->pDriverName);

    resetRegisters(pInst);

    return rc;
    }

/******************************************************************************
*/

STATUS hpcHwI86ArchPm1DriverInit(void)
    {
    return OK;
    }

/******************************************************************************
*/

STATUS hpcHwI86ArchPm1DriverFinit(void)
    {
    return OK;
    }

/******************************************************************************
*/

struct hpcHwDriverFuncs hpcHwDrvI86Arch1Rtns =
    {
    hpcHwCtrStart,
    hpcHwCtrStop,
    hpcHwCtrRead,
    hpcHwUpdateCtrReload,
    hpcHwCreateCtrForEvent,
    hpcHwDeleteCtrForEvent,
    hpcHwDriverInit,
    hpcHwDriverFinit,
    hpcHwPerCoreInit,
    hpcHwPerCoreFinit,
    hpcHwDriverShow
    };
