/* hpc_ep.c - Cafe Hardware Performance Counter - Event Processor */

/* Copyright (c) 2010-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
modification history
--------------------
03aug14,sdt  Fixed checked-return warnings (US45196)
12mar14,tle  Disable use of alloca() on PPC and ARM (WB4-2840, WB4-3122)
11mar14,sdt  Fix 'Checked return' warning


DESCRIPTION
 * VC	Virtual Counter
 * VCS	Virtual Counter Slave
 *
 * HPC Events:
 * HPC counters are initialised to count a specified event. When the counter
 * reaches its maximum value, an interrupt is generated and the driver calls
 * hpc_perf_interrupt_handler_callback(). The callback updates the VC and
 * may call cafe_dispatch_event_unlocked() to inject the event to the event
 * stream.
 *
 * Virtual Counter Block:
 * The VC block is initialised when the EP is created. Initially only one block
 * is created, additional blocks are added when needed. Allocating VC from pre
 * allocated VC blocks means that no dynamic memory allocation is required
 * at event enable disable time.
 *
 * Virtual Counters:
 * VCs are used to support multiple EP instances and counting HPC events as
 * orthogonal data. For example, EP1 requires total_cycles event. To count this
 * event, a VC (VC1) and a VCS (VCS1) are allocated. EP2 is then created which
 * also requires total_cycles. Since total_cycles is already being collected
 * for EP1, the same hardware resource and VC is used for EP2, however a
 * different VCS (VCS2) is used. The second VCS is required for several
 * reasons:
 * 1) Events in EP1 and EP2 may have a different skip count value and so
 *    are injected at a different rate.
 * 2) The second event is enabled while the counter is running, so the counters
 *    latest value is used to offset the counter (see vcs_count_offset).
 * Now consider EP1 enables a second event: cache_miss per total_cycles. This
 * would allocate a new VC (VC2) and VCS (VCS3) for the cache miss event
 * (primary event). For the secondary counter (total_cycles), VC1 would be
 * used and a new VCS (VCS4) would be reserved. This would look like:
 *
 * VC1(total_cycles) -> VCS1 (EP1)
 *                      VCS2 (EP2)
 *                      VCS4 (EP1) secondary
 *
 * VC2(cache_miss)   -> VCS3 (EP1)
 *
 * IMPORTANT:
 * To prevent concurrency issues between the control task and the perf ISR,
 * the VC should be stopped to prevent the ISR from running when certain items
 * in the VC is updated.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <hpc/hpc_ep.h>
#include <cafe/errors.h>
#include <cafe/autogenerated/cafe.type.h>
#include <cafe/autogenerated/vxworks.type.h>
#include <cafe/autogenerated/hpc_common.type.h>
#include "../../cafe/src/cafe_assert.h"
#include <cafe/expr.h>

#include "hpc_utils.h"
#include "hpc_ep_priv.h"
#include "hpc_log.h"

#if defined(__VXWORKS__)
#  include <cafe/ep/vxworks_kernel.h>
#  include <isrLib.h>
#  include <taskLib.h>
#  if defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
#    include <st_stacktrace.h>
#  endif
#endif

#ifdef __VXWORKS__
#  ifdef _WRS_CONFIG_SMP
#    define MAX_NUM_CORES 		(VX_MAX_SMP_CPUS)
#  else
#    define MAX_NUM_CORES 		(1)
#  endif
#  if (_VX_CPU_FAMILY != _VX_PPC) && (_VX_CPU_FAMILY != _VX_ARM)
#    ifdef  __GNUC__
#      define CAFE_ALLOCA(size)   __builtin_alloca(size)
#    elif defined(__DCC__)
#      define CAFE_ALLOCA(size)   __alloca(size)
#    elif defined(__ICC)
#      define CAFE_ALLOCA(size)   _alloca(size)
#    endif
#  endif  /* if (_VX_CPU_FAMILY != _VX_PPC) */
#else /* __VXWORKS__ */
#  define MAX_NUM_CORES 		(32)
#endif

#if !defined(LLONG_MAX)
#define LLONG_MAX ((~(long long)0) >> 1)
#define LLONG_MIN (~((~(long long)0) >> 1))
#define ULLONG_MAX (~(unsigned long long)0)
#endif

/* Maximum number of low level hardware drivers */
#define HPC_NUM_DRIVER_INSTANCES (5)

/******************************************************************************/

#if !defined(ROUND_UP)
#  define ROUND_UP(x, align)	(((int) (x) + (align - 1)) & ~(align - 1))
#endif

/******************************************************************************/

#define EP_NAME  "perf_counter_provider"
#define WORKAREA  "S[" EP_NAME ".workarea,"		\
	"v[max_stack_depth,I," cafe_setting_VT_NAT "]"	\
	"v[ignore_int_context,?]"			\
	"v[tids_to_sample,a[L]]]"

struct hpc_workarea {
	unsigned int		max_stack_depth;
	bool			ignore_int_ctx;
	struct cafe_vm_dynarray	tids_to_sample;
};

/******************************************************************************/

/* HPC low level perf monitoring driver instance */

#define FOREACH_HPC_DRIVER(HW, DRV)			\
	for (DRV = 0, HW = &hw_drv[DRV];		\
	     DRV < HPC_NUM_DRIVER_INSTANCES;		\
	     HW++, DRV++)				\
		if (hw_drv[DRV].in_use)

struct hw_driver_inst {
	bool			in_use;
	struct hpcDriverConfig	drv;
};

/* hpc event implementation data */
struct hpc_evt_imp_data {
	struct hw_driver_inst		*hw;
	struct hpcEventDescriptor	*evt_desc;
};

/* event descriptor */
struct hpc_event_desc {
	struct hw_driver_inst		*hw;			/* hardware driver for event */
	struct hpcEventSettings		hpc_evt_settings;	/* generic setting used by EP */
	union eventSettings		drv_evt_settings;	/* hardware specific event settings, passed to low level driver */
};

/* HPC virtual counter */

struct vc_block_item;

#define VCS_SEC_LIST_ITEM(PTR) CAFE_LIST_ENTRY((PTR), struct hpc_virt_slave, vcs_sec_list)

struct hpc_virt_slave {
	volatile char			vcs_enabled;	/* flag to detect if event is still needed */
							/* This is only valid when UOEA() callback is called */
	HPC_COUNTER			vcs_count[MAX_NUM_CORES]; /* vcs count value per core */
	HPC_COUNTER			vcs_count_prev[MAX_NUM_CORES]; /* vcs previous count value per core */
	HPC_COUNTER			vcs_count_offset[MAX_NUM_CORES]; /* vcs count offset value per core */
	HPC_COUNTER			vcs_reload;	/* reload value requested by client */
	volatile char			vcs_running;	/* vcs started or stopped state. */
	struct hpc_inst_item		*vcs_hpc_inst;	/* instance vcs enabled in */
	cafe_eal_handle_t		vcs_sink_act_handle; /* event sink action handle */
	struct cafe_list_head 		vcs_sec_list_head; /* list head of secondary vcs */
	struct cafe_list_head 		vcs_sec_list;	/* list of secondary vcs */
	struct hpc_virt_counter		*vcs_vc;	/* back ptr to vc that vcs is in */
	int				vcs_odata_idx;	/* secondary odata index */
	struct hpc_virt_slave 		*vcs_pri;	/* for secondary vcs, this is pointer to the primary vcs */
	struct cafe_value_type		*vcs_type;
	int				vcs_hpc_od_offset; /* used to check that OD offset matches - check WRS_TODO notes */
};

/* This represents the numbers of instances that can share the same base event */
#define NUM_VC_SLAVES	5
#define VC_LIST_ITEM(PTR)	CAFE_LIST_ENTRY((PTR), struct hpc_virt_counter, vc_blk_list)
#define VC_NON_MUXED_LIST_ITEM(PTR) CAFE_LIST_ENTRY((PTR), struct hpc_virt_counter, vc_in_use_list)

/* driver API functions */
#define HW_DRV_CTR_STOP		vc->vc_evt_set.hw->drv.pFuncs->hpcHwCtrStop
#define HW_DRV_CTR_START	vc->vc_evt_set.hw->drv.pFuncs->hpcHwCtrStart
#define HW_DRV_CTR_READ		vc->vc_evt_set.hw->drv.pFuncs->hpcHwCtrRead
#define HW_DRV_CTR_RELOAD	vc->vc_evt_set.hw->drv.pFuncs->hpcHwUpdateCtrReload
#define HW_DRV_CTR_CREATE	vc->vc_evt_set.hw->drv.pFuncs->hpcHwCreateCtrForEvent

#define HW_DRV_DATA(VC, CORE)	(struct hpcHwCtrData *)&VC->vc_hw_drv.data[CORE * vc_blk_desc.sizeof_hw_data]

union dyn_array {
	char		       *align1;
	size_t			align2;
	unsigned long long	align3;
	char			data[1];
};

struct hpc_virt_counter {
	struct cafe_list_head 		vc_blk_list;		/* vc in vc block list */
	struct cafe_list_head 		vc_in_use_list;		/* vc in use by EP instances list */
	struct vc_block_item		*vc_block;		/* block that this vc belongs to */
	volatile char			vc_running;		/* for counter running/stopped */
	struct hpc_event_desc		vc_evt_set;		/* event configuration requested by client */
	struct hpc_virt_slave 		vc_slave[NUM_VC_SLAVES]; /* array of slave virtual counters */
	HPC_COUNTER			vc_stop_count[MAX_NUM_CORES];	/* current stopped count value, per core */
	HPC_COUNTER			vc_reload[MAX_NUM_CORES];	/* actual reload passed to low level driver */
	HPC_COUNTER			vc_reload_min[MAX_NUM_CORES];	/* min reload */
	volatile unsigned long long	vc_time_prev[MAX_NUM_CORES]; /* in units of cafe_vx_timestamp_freq64 */
	unsigned long long		vc_rate_prev[MAX_NUM_CORES]; /* previous rate (events/second) */

	/* variable size data structure which must be at the end of struct. It
	 * is used by the low-level hw perf driver */
	union dyn_array			vc_hw_drv;
};

#define VC_BLOCK_ITEM(PTR) CAFE_LIST_ENTRY((PTR), struct vc_block_item, blk_list)

struct vc_block_item {
	struct cafe_list_head		blk_list;
	int				alloc_count;
	struct hpc_virt_counter		vc_array;		/* dynamic allocation for vc_array follows here */
};								/* VCs are allocated from this memory */

/* VC Block data descripton */
struct vc_block_desc {
	struct cafe_list_head		vc_list_head;		/* memory block used for vc */
	struct cafe_list_head		vc_alloc_head;		/* in use vc */
	struct cafe_list_head		vc_free_head;		/* list of free vs */
	int				num_vc_per_block;
	size_t				sizeof_hw_data;
	int				max_cores;
	int				num_vc_blocks;
};

#define CFG_LIST_ITEM(PTR)	CAFE_LIST_ENTRY((PTR), struct hpc_inst_item, hpc_ep_inst_list)

/* data needed per EP instance */
struct hpc_inst_item {
	struct cafe_list_head		hpc_ep_inst_list;
	struct cafe_ep_inst		*ep_inst;
	unsigned long			change_count;
};

struct inst_desc {
	struct cafe_list_head	ep_inst_list_head;	/* list of ep instances */
	struct cafe_list_head	vc_in_use_list_head;	/* list of vc in ep instance */
};

#define FOREACH_HPC_DRIVER(HW, DRV)			\
	for (DRV = 0, HW = &hw_drv[DRV];		\
	     DRV < HPC_NUM_DRIVER_INSTANCES;		\
	     HW++, DRV++)				\
		if (hw_drv[DRV].in_use)

static struct hw_driver_inst hw_drv[HPC_NUM_DRIVER_INSTANCES];

static struct vc_block_desc vc_blk_desc; /* VC Blocks */
static struct inst_desc ep_inst_desc; /* EP Instances */
static void * (*hpc_malloc)(size_t);
static void (*hpc_free)(void *);
static unsigned long long hpc_callback_count;
static unsigned long long hpc_overflow_error_count;
static unsigned long long hpc_isr_error_count;
static unsigned long long hpc_counter_interrupted_count;
static unsigned long long timer_freq;

#define MAX_STACK_DEPTH 128

/* indices into orthogonal data items (hpc odata events are added dynamically) */

enum {
	ODATA_INDEX_TS = 0,
	ODATA_INDEX_CORE,
	ODATA_INDEX_SKIPCOUNT,
	ODATA_INDEX_SAMPLE_FREQ,
	ODATA_INDEX_CONTEXT,
#if defined(_WRS_CONFIG_COMPONENT_STACKTRACE) || !defined(__VXWORKS__)
	ODATA_INDEX_STACK,
#endif
	ODATA_INDEX_SKIPCOUNT_ACTUAL,
	NUM_ODATA_CAFE_ITEMS
};

#define NUM_ODATA_HPC_ITEMS 4
static struct hpc_evt_imp_data * hpc_events_output_odata[NUM_ODATA_HPC_ITEMS];

#define NUM_ODATA_ITEMS (NUM_ODATA_CAFE_ITEMS + NUM_ODATA_HPC_ITEMS)

static const char *output_odata[NUM_ODATA_ITEMS] = {
	cafe_timestamp_VT_REF,
	cafe_core_id_VT_REF,
	cafe_skipcount_VT_REF,
	cafe_sample_freq_VT_REF,
	vxWorks_context_VT_REF,
#if defined(_WRS_CONFIG_COMPONENT_STACKTRACE) || !defined(__VXWORKS__)
	cafe_stacktrace_VT_REF,
#endif
	cafe_skipcount_actual_limit_VT_REF,
	/* HPC events filled in dynamically */
};

static int perf_counter_provider_create(struct cafe_ep_inst *inst);
static void perf_counter_provider_delete(struct cafe_ep_inst *inst);
static int perf_counter_provider_start(struct cafe_ep_inst *inst);
static int perf_counter_provider_stop(struct cafe_ep_inst *inst);
static int perf_counter_update_output_event_actions(struct cafe_ep_inst *inst);

static struct cafe_ep_spec hpc_root_spec = {
	EP_NAME,
	perf_counter_provider_create,
	perf_counter_provider_delete,
	perf_counter_provider_start,
	perf_counter_provider_stop,
	NULL,
	perf_counter_update_output_event_actions,
	NULL,
	cafe_setting_VT_DEF
	cafe_timestamp_VT_DEF
	cafe_core_id_VT_DEF
	cafe_CURRENT_CORE_ID_VT_DEF
	cafe_skipcount_VT_DEF
	cafe_sample_freq_VT_DEF
	vxWorks_context_VT_DEF
#if defined(_WRS_CONFIG_COMPONENT_STACKTRACE) || !defined(__VXWORKS__)
	cafe_stacktrace_VT_DEF
#endif
	cafe_skipcount_actual_limit_VT_DEF,
	WORKAREA,
	0,		/* num input events */
	NULL,		/* input events */
	0,		/* num of output events */
	NULL,		/* output events */
	0,
	output_odata
};

#if defined(__VXWORKS__)
#if (_VX_CPU_FAMILY==_VX_I80X86)
unsigned long (*_func_cafe_isrTagToIrqNoConvert) (unsigned long) = NULL;
#endif /* (_VX_CPU_FAMILY==_VX_I80X86) */
#endif /* __VXWORKS__ */

/******************************************************************************/

void hpc_count_show(void)
{
	printf("hpc_callback_count            %llu\n", hpc_callback_count);
	printf("hpc_overflow_error_count      %llu\n", hpc_overflow_error_count);
	printf("hpc_isr_error_count           %llu\n", hpc_isr_error_count);
	printf("hpc_counter_interrupted_count %llu\n", hpc_counter_interrupted_count);
}

/******************************************************************************/

static __inline__ int cafe_list_item_is_valid(struct cafe_list_head *item)
{
	return item->prev->next == item && item->next->prev == item;
}

#if 0
/* this routine is called at ISR level */
static bool updateSlaveVcOffset(HPC_COUNTER *offset, HPC_COUNTER *count,
		HPC_COUNTER *acc_count)
{
	/* try and get rid of the offset, and return true if offset set to 0 */
	if (*count >= *offset) {
		*count -= *offset;
		*offset = 0;
		return true;
	}
	if (*acc_count >= *offset) {
		*acc_count -= *offset;
		*offset = 0;
		return true;
	}
	if (*count + *acc_count >= *offset) {
		*acc_count = (*count + *acc_count) - *offset;
		*count = 0;
		*offset = 0;
		return true;
	}
	return false;
}
#endif

/* this routine is called at ISR level */
static bool filter_vc_slave(struct hpc_workarea *wa, HPC_TASK_ID_TYPE tid,
				int int_count)
{
	unsigned int i;
	HPC_TASK_ID_TYPE *tids;

	if (int_count == 1) {
		if (wa->tids_to_sample.l == 0)
			return false;

		tids = wa->tids_to_sample.p;
		for (i = 0; i < wa->tids_to_sample.l; i++) {
			if (tids[i] == tid)
				return false;
		}
	} else {
		return wa->ignore_int_ctx;
	}

	return true;
}

/* this routine is called at ISR level */
static void read_hpc_counter_for_odata(struct hpc_virt_slave *pri_vcs,
	void *output_data[],
	HPC_COUNTER *hpc_od_count,
	int core)
{
	STATUS rc;
	struct cafe_list_head *item;
	HPC_COUNTER count;

	cafe_list_foreach(item, &pri_vcs->vcs_sec_list_head) {
		struct hpc_virt_slave *sec_vcs = VCS_SEC_LIST_ITEM(item);
		struct hpc_virt_counter *sec_vc = sec_vcs->vcs_vc;
		rc = sec_vc->vc_evt_set.hw->drv.pFuncs->hpcHwCtrRead(
				sec_vc->vc_evt_set.hw->drv.pDrvInst,
				HW_DRV_DATA(sec_vc, core), &count);
		if (rc != OK) {
			hpc_isr_error_count++;
			return;
		}
		*hpc_od_count = count + sec_vcs->vcs_count[core] -
				sec_vcs->vcs_count_prev[core];
		sec_vcs->vcs_count_prev[core] = count;
		sec_vcs->vcs_count[core] = 0;
		output_data[sec_vcs->vcs_odata_idx] = hpc_od_count++;
	}
}

static void update_freq_reload(struct hpc_virt_counter *vc, int core,
	HPC_COUNTER count_d,		/* count as a delta value */
	unsigned long long time_abs,	/* absolute time value */
	HPC_COUNTER *next_count_d)	/* next count (delta) */
{
	unsigned long long time_d; /* time delta */
	unsigned long long new_rate, ratio;
	unsigned int new_weight, prev_weight;

	time_d = time_abs - vc->vc_time_prev[core];
	vc->vc_time_prev[core] = time_abs;

	if (time_d == 0) {
		hpc_isr_error_count++;
		return;
	}
	new_rate = (vc->vc_evt_set.hpc_evt_settings.samplePeriod * count_d) / time_d;
	if (new_rate == 0) new_rate = 1;

	if (new_rate < vc->vc_rate_prev[core])
		ratio = vc->vc_rate_prev[core] / new_rate; /* exiting burst */
	else
		ratio = 1;

	if (ratio > 100) {
		/* exiting a burst */
		new_weight = 100;
		prev_weight = 0;
	} else {
		new_weight = 25;
		prev_weight = 75;
	}

	*next_count_d = ((new_weight * new_rate) + (prev_weight * vc->vc_rate_prev[core])) / 100;
	if (*next_count_d < 10)
		*next_count_d = 10;

	vc->vc_rate_prev[core] = new_rate;
}

/* Macro only meant to be used from the hpc_perf_interrupt_handler_callback()
   routine. Evaluates any filter orthogonal data entries, advancing the odata
   index past all consumed entries.  When a filter evaluates as false, move
   on to the next registered event. */
#define CHECK_FILTER()							\
	do {								\
		if (++i == num_odtypes)					\
			goto odata_done;				\
									\
		while (event->types[i]->kind == CAFE_TYPE_FILTER) {	\
			cafe_any_value result;				\
			struct cafe_expr_eval_context context;		\
									\
			cafe_expr_eval_context_create(&context, NULL);	\
			context.event =	event;				\
			context.event_data = output_data;		\
			rc = cafe_expr_eval_value(&context, event->types[i]->ident->expr, &result); \
			if (rc != CAFE_OK) {				\
				cafe_dispatch_error(inst, rc);		\
				goto unlock;				\
			}						\
			if (result.ll == 0) {				\
				goto unlock;				\
			}						\
			if (++i == event->num_types)			\
				goto odata_done;			\
		}							\
	} while(0)


/*****************************************************************************
 * this routine is called at ISR level
 */

static void hpc_perf_interrupt_handler_callback(void *c_data, void *data,
						HPC_COUNTER count, int core)
{
	int rc, coreid, i, c, num_odtypes, int_count;
	unsigned long long t_now;
	HPC_TASK_ID_TYPE tid;
	/* must allow for event param data - even though there aren't any for hpc events */
	void *output_data[NUM_ODATA_ITEMS + 1];
	HPC_COUNTER hpc_od_count[NUM_ODATA_HPC_ITEMS], count2, prev_reload;
	struct cafe_event_action_list *action_list;
	struct hpc_virt_counter *vc = (struct hpc_virt_counter *)c_data;
	struct hpc_virt_slave *vc_slave;
	struct cafe_event_descr *event;
	struct cafe_value_type **odata;
	cafe_eal_handle_t unlock_handle;
	struct hpc_workarea *wa;
	struct cafe_ep_inst *inst;
#if defined(_WRS_CONFIG_COMPONENT_STACKTRACE) || !defined(__VXWORKS__)
#  ifdef CAFE_ALLOCA
	void **stack = NULL;
#  else
	void *stack[MAX_STACK_DEPTH];
#  endif
	struct cafe_vm_dynarray stack_arg;
#endif
	hpc_callback_count++; /* no need for atomic inc, its just a debug counter... */
	/* dynamically update hw counter reload value to maintain constant
	 * event frequency */
	(void)cafe_vx_timestamp64(&t_now);
	/* if time goes in reverse, add on a time period */
	if (t_now <= vc->vc_time_prev[core])
		t_now += timer_freq;
	prev_reload = vc->vc_reload[core];
	if (vc->vc_evt_set.hpc_evt_settings.samplePeriod) {
		update_freq_reload(vc, core, count, t_now, &vc->vc_reload[core]);
		if (HW_DRV_CTR_RELOAD(vc->vc_evt_set.hw->drv.pDrvInst,
				HW_DRV_DATA(vc, core), &vc->vc_reload[core]) != OK)
			hpc_isr_error_count++;
	}
	if (vc->vc_reload[core] < vc->vc_reload_min[core])
		vc->vc_reload_min[core] = vc->vc_reload[core];
	tid = taskIdSelf();
	int_count = intCount();
	for (c = 0; c < NUM_VC_SLAVES; c++) {
		vc_slave = &vc->vc_slave[c];
		count2 = count;
		if (!vc_slave->vcs_running)
			continue;
		inst = vc_slave->vcs_hpc_inst->ep_inst;
		wa = (struct hpc_workarea *)inst->workarea->u.data;

		if (filter_vc_slave(wa, tid, int_count))
			continue;

		/* remove any offset */
		if (vc_slave->vcs_count_offset[core]) {
#if 1
			vc_slave->vcs_count_offset[core] = 0;
			continue;
#else
			if (!updateSlaveVcOffset(
					&vc_slave->vcs_count_offset[core],
					&count2,
					&vc_slave->vcs_count[core])) {
				vc_slave->vcs_count[core] += count2;
				continue;
			}
#endif
		}
		/* increment secondary vcs */
		if (vc_slave->vcs_sec_list.next) {
			vc_slave->vcs_count[core] += count2 - vc_slave->vcs_count_prev[core];
			vc_slave->vcs_count_prev[core] = 0;
			continue;
		}
		/* increment primary vcs */
		if (!vc->vc_evt_set.hpc_evt_settings.samplePeriod && (vc_slave->vcs_count[core] + count2 < vc_slave->vcs_reload)) {
			vc_slave->vcs_count[core] += count2;
			continue;
		}
		count2 += vc_slave->vcs_count[core];
		vc_slave->vcs_count[core] = 0;

		/* read any secondary HPC counters for odata */
		read_hpc_counter_for_odata(vc_slave, output_data, hpc_od_count, core);

		/* get a lock on the sink actions list for the event */
		if (vc_slave->vcs_sink_act_handle == CAFE_EAL_HANDLE_NULL)
			continue;

		rc = cafe_event_action_list_read_lock(&vc_slave->vcs_sink_act_handle,
						      &unlock_handle, &action_list);
		if (rc != CAFE_OK) {
			cafe_dispatch_error(inst, rc);/* WRS_TODO: make sure this is isr safe */
			return;
		}
		/* store any requested odata in the output event */
		event = action_list->event;

		if (event->num_types == 0)
			goto odata_done;
		odata = inst->output_odata;
		num_odtypes = event->num_types;
		i = 0;

		CHECK_FILTER();

		if (event->types[i] == odata[ODATA_INDEX_TS]) {
			output_data[i] = &t_now;
			CHECK_FILTER();
		}
		if (event->types[i] == odata[ODATA_INDEX_CORE]) {
			coreid = vxCpuIndexGet();
			output_data[i] = &coreid;
			CHECK_FILTER();
		}
		if (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SKIPCOUNT]) {
			output_data[i] = &count2;
			CHECK_FILTER();
		}
		if (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SAMPLE_FREQ]) {
			output_data[i] = &count2;
			CHECK_FILTER();
		}
#if defined(__VXWORKS__)
		if (event->types[i] == odata[ODATA_INDEX_CONTEXT]) {
			struct vxWorks_context ctx = {0};

			ctx.tid = (unsigned long) tid;
			ctx.pid = (unsigned long) RTP_ID_OF_TASK(ctx.tid);
			cafe_cfg_vxworks_cur_int_num_get (&ctx.intCount, &ctx.intNum);
			output_data[i] = &ctx;
			CHECK_FILTER();
		}
#endif
#if defined(_WRS_CONFIG_COMPONENT_STACKTRACE)
		if (event->types[i] == odata[ODATA_INDEX_STACK]) {
			REG_SET regs;
			ST_context context;

			if (wa->max_stack_depth > MAX_STACK_DEPTH ||
			    wa->max_stack_depth == 0) {
				wa->max_stack_depth = MAX_STACK_DEPTH;
			}
#  ifdef CAFE_ALLOCA
			if (stack == NULL)
				stack = (void **) CAFE_ALLOCA(wa->max_stack_depth * sizeof(void *));
#  endif

			output_data[i] = &stack_arg;
			stack_arg.p = stack;
			if (data) {
				memset(&regs, 0, sizeof(regs));
				hpcGetRegSet(data, &regs);
				if (!ST_context_create(&regs, (const unsigned long)tid,
							false, &context)) {
					stack[0] = ST_trace_error;
					stack_arg.l = 1;
				} else {
					stack_arg.l = ST_perform_stack_trace(&context, stack, wa->max_stack_depth);
					(void)ST_context_delete(&context);
				}
			} else
				stack_arg.l = 0;
			CHECK_FILTER();
		}
#elif !defined(__VXWORKS__)
		/* this is only for linux host test case: see hpc_test_orth_st.c */
		if (event->types[i] == odata[ODATA_INDEX_STACK]) {
			output_data[i] = &stack_arg;
			stack[0] = (void *)0x1234abcd;
			stack_arg.p = stack;
			stack_arg.l = 0;
			CHECK_FILTER();
		}
#endif
		if (event->types[i] == odata[ODATA_INDEX_SKIPCOUNT_ACTUAL]) {
			output_data[i] = &prev_reload;
		}
odata_done:
		cafe_dispatch_event_unlocked(action_list, output_data);
unlock:
		cafe_event_action_list_read_unlock(unlock_handle);
	}
}

#if 0
/* These routines can be used inject cafe/hpc events at task level for
 * easier debug on vxWorks */
void hpc_test_inject_cafe_event(void);

static void hpc_test_inject_cafe_event_vc(struct hpc_virt_counter *vc)
{
	HPC_COUNTER counter;

	counter = vc->vc_evt_set.hpc_evt_settings.reload + 100;
        hpc_perf_interrupt_handler_callback(vc, NULL, counter, vxCpuIndexGet());
}

void hpc_test_inject_cafe_event(void)
{
	struct cafe_list_head *item;

	cafe_list_foreach(item, &ep_inst_desc.vc_in_use_list_head) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);
		hpc_test_inject_cafe_event_vc(vc);
	}

}
#endif

/******************************************************************************/

static int alloc_vc_list(void)
{
	int counter;
	struct hpc_virt_counter *vc;
	struct vc_block_item *vc_blk_item;
	size_t size_vc_array;
	size_t size_vc_block;
	size_t size_vc_struct = sizeof(*vc) - sizeof(union dyn_array) +
			(vc_blk_desc.sizeof_hw_data * vc_blk_desc.max_cores);

	size_vc_array = size_vc_struct * vc_blk_desc.num_vc_per_block;
	size_vc_block = sizeof(*vc_blk_item) - sizeof(*vc) + size_vc_array;

	if ((vc_blk_item = (*hpc_malloc)(size_vc_block)) == NULL) {
		hpc_error("unable to alloc memory for virtual counters %"PRIsize_t"\n",
			  size_vc_block);
		return CAFE_NOMEM;
	}

	memset(vc_blk_item, 0, size_vc_block);
	vc_blk_desc.num_vc_blocks++;
	cafe_list_add_first(&vc_blk_item->blk_list, &vc_blk_desc.vc_list_head);
	vc = &vc_blk_item->vc_array;

	for (counter = 0; counter < vc_blk_desc.num_vc_per_block; counter++) {
		char *tmp = (char *)vc;
		vc->vc_block = vc_blk_item;
		cafe_list_add_first(&vc->vc_blk_list, &vc_blk_desc.vc_free_head);
		tmp += size_vc_struct;
		vc = (struct hpc_virt_counter *)tmp;
	}

	return CAFE_OK;
}

static struct hpc_virt_counter *alloc_vc(void)
{
	struct hpc_virt_counter *vc;
	struct vc_block_item *vc_blk_item;
	struct cafe_list_head *item = NULL;
	int rc, s;

	cafe_list_foreach(item, &vc_blk_desc.vc_free_head)
		break;

	if (item == &vc_blk_desc.vc_free_head) {
		hpc_debug(10, "empty free list, allocating more virtual counters\n");
		if ((rc = alloc_vc_list()) != CAFE_OK) {
			hpc_error("couldn't allocate virtual counters %u\n", rc);
			return NULL;
		}
		return alloc_vc();
	}

	cafe_list_remove(item);

	vc = VC_LIST_ITEM(item);
	vc_blk_item = vc->vc_block;
	vc_blk_item->alloc_count++;

	cafe_list_add_last(item, &vc_blk_desc.vc_alloc_head);

	for (s = 0; s < NUM_VC_SLAVES; s++)
		vc->vc_slave[s].vcs_hpc_inst = NULL;

	return vc;
}

static int free_vc(struct hpc_virt_counter *vc)
{
	struct cafe_list_head *item, *item_ptr;
	struct vc_block_item *vc_blk_item;
	struct hpc_virt_counter *vc_ptr;

	if (vc == NULL) {
		hpc_error("free_vc error, called with NULL\n");
		return CAFE_ERROR;
	}

	item = &vc->vc_blk_list;

	cafe_list_foreach(item_ptr, &vc_blk_desc.vc_alloc_head) {
		if (item_ptr == item)
			break;
	}

	if (item_ptr != item) {
		hpc_error("given invalid vc to free %p\n", vc);
		return CAFE_ERROR;
	}

	cafe_list_remove(item);
	cafe_list_add_first(item, &vc_blk_desc.vc_free_head);
	vc_blk_item = vc->vc_block;
	vc_blk_item->alloc_count--;

	if (vc_blk_item->alloc_count == 0 && vc_blk_desc.num_vc_blocks > 1) {
		/* remove items from the free list which are in this block */
		cafe_list_foreach(item_ptr, &vc_blk_desc.vc_free_head) {
			vc_ptr = VC_LIST_ITEM(item_ptr);
			if (vc_ptr->vc_block == vc_blk_item)
				cafe_list_remove(item_ptr);
		}
		/* free this block */
		cafe_list_remove(&vc_blk_item->blk_list);
		vc_blk_desc.num_vc_blocks--;
		(*hpc_free)(vc_blk_item);
	}

	return CAFE_OK;
}

static void init_vc_list_data(void)
{
	cafe_list_init(&vc_blk_desc.vc_alloc_head);
	cafe_list_init(&vc_blk_desc.vc_free_head);
	cafe_list_init(&vc_blk_desc.vc_list_head);
	vc_blk_desc.num_vc_blocks = 0;
}

static void sizeof_hw_data_set(size_t sizeof_hw_data)
{
	if (sizeof_hw_data > vc_blk_desc.sizeof_hw_data) {
		vc_blk_desc.sizeof_hw_data = ROUND_UP(sizeof_hw_data, sizeof(size_t));
	}
}

static void max_cores_set(int cores)
{
	if (cores > vc_blk_desc.max_cores) {
		vc_blk_desc.max_cores = cores;
	}
}

static void num_vc_per_block_set(int num_vc_per_block)
{
	vc_blk_desc.num_vc_per_block = num_vc_per_block;
}

static int init_vc_list(void)
{
	if (vc_blk_desc.num_vc_per_block == 0) {
		hpc_error("num_vc_per_block is zero\n");
		return CAFE_ERROR;
	}

	return (alloc_vc_list());
}

static int finit_vc_list(bool force_cleanup)
{
	struct vc_block_item *vc_blk_item;
	struct cafe_list_head *item, *tmp;

	if (!force_cleanup)
		if (!cafe_list_is_empty(&vc_blk_desc.vc_alloc_head)) {
			hpc_error("can't free vc as some are still in use\n");
			return CAFE_ERROR;
		}

	cafe_list_foreach_safe(item, &vc_blk_desc.vc_list_head, tmp) {
		vc_blk_item = VC_BLOCK_ITEM(item);
		cafe_list_remove(item);
		vc_blk_desc.num_vc_blocks--;
		(*hpc_free)(vc_blk_item);
	}

	if (vc_blk_desc.num_vc_blocks) {
		hpc_error("error with num_vc_blocks on cleanup %d",
			  vc_blk_desc.num_vc_blocks);

		if (!force_cleanup)
			return CAFE_ERROR;
	}

	init_vc_list_data();
	vc_blk_desc.num_vc_per_block = 0;
	vc_blk_desc.sizeof_hw_data = 0;
	vc_blk_desc.max_cores = 1;

	return CAFE_OK;
}

/******************************************************************************/

static unsigned int get_num_vc_free(void)
{
	unsigned int count = 0;
	struct cafe_list_head *item;

	cafe_list_foreach(item, &vc_blk_desc.vc_free_head) {
		count++;
	}

	return count;
}

static unsigned int get_num_vc_alloc(void)
{
	unsigned int count = 0;
	struct cafe_list_head *item;

	cafe_list_foreach(item, &vc_blk_desc.vc_alloc_head) {
		count++;
	}

	return count;
}

/******************************************************************************/

static void add_vc_to_in_use_list(struct hpc_virt_counter *vc)
{
	cafe_list_add_last(&vc->vc_in_use_list, &ep_inst_desc.vc_in_use_list_head);
}

/******************************************************************************/

static struct hpc_inst_item *alloc_ep_inst(void)
{
	struct hpc_inst_item *hpc_inst;

	if ((hpc_inst = (*hpc_malloc)(sizeof(*hpc_inst))) == NULL) {
		hpc_error("Can't allocate memory for configuration %"PRIsize_t"\n", sizeof(*hpc_inst));
		return NULL;
	}

	cafe_list_add_last(&hpc_inst->hpc_ep_inst_list, &ep_inst_desc.ep_inst_list_head);

	return hpc_inst;
}

static int free_ep_inst(struct hpc_inst_item *hpc_inst)
{
	struct cafe_list_head *item, *item_ptr;

	if (hpc_inst == NULL) {
		hpc_error("free_ep_inst error, called with NULL\n");
		return CAFE_ERROR;
	}

	item = &hpc_inst->hpc_ep_inst_list;

	cafe_list_foreach(item_ptr, &ep_inst_desc.ep_inst_list_head) {
		if (item_ptr == item)
			break;
	}

	if (item_ptr != item) {
		hpc_error("given invalid hpc_inst to free %p\n", hpc_inst);
		return CAFE_ERROR;
	}

	cafe_list_remove(item);
	(*hpc_free)(hpc_inst);

	return CAFE_OK;
}

static void init_ep_inst_list(void)
{
	cafe_list_init(&ep_inst_desc.ep_inst_list_head);
	cafe_list_init(&ep_inst_desc.vc_in_use_list_head);
}

static int finit_ep_inst_list(bool force_cleanup)
{
	struct hpc_inst_item *hpc_inst;
	struct cafe_list_head *item, *tmp;

	if (!force_cleanup)
		if (!cafe_list_is_empty(&ep_inst_desc.ep_inst_list_head)) {
			hpc_error("can't cleanup ep_inst_desc as some are still in the list\n");
			return CAFE_ERROR;
		}

	cafe_list_foreach_safe(item, &ep_inst_desc.ep_inst_list_head, tmp) {
		hpc_inst = CFG_LIST_ITEM(item);
		cafe_list_remove(item);
		(*hpc_free)(hpc_inst);
	}

	return CAFE_OK;
}

/******************************************************************************/

struct event_iter_data {
	struct hpc_virt_counter *vc;
	struct cafe_list_head	*item;
	struct cafe_list_head	*head;
	int			core;
};

static struct hpcHwCtrData *event_iter(void *iter_arg)
{
	struct event_iter_data *iter = (struct event_iter_data *)iter_arg;

again:
	if (iter->item != iter->head) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(iter->item);
		iter->item = iter->item->next;
		if (vc == iter->vc ||
		    vc->vc_evt_set.hw != iter->vc->vc_evt_set.hw)
			goto again;
		return HW_DRV_DATA(vc, iter->core);
	}
	return NULL;
}

static int configure_vc_for_event(struct hpc_inst_item *hpc_inst,
                                         struct hpc_virt_counter *vc,
                                         struct hpc_event_desc *evt_set)
{
        STATUS hpc_rc;
        int core;
        struct event_iter_data iter;

        /* copy event descriptor to the virtual counter */
        memcpy(&vc->vc_evt_set, evt_set, sizeof(*evt_set));

        iter.vc = vc;

        for (core = 0; core < vc->vc_evt_set.hw->drv.numCpuCores; core++) {
        	iter.head = &ep_inst_desc.vc_in_use_list_head;
        	iter.item = iter.head->next;
        	iter.core = core;
		hpc_rc = HW_DRV_CTR_CREATE(vc->vc_evt_set.hw->drv.pDrvInst,
				&vc->vc_evt_set.drv_evt_settings,
				&vc->vc_evt_set.hpc_evt_settings.reload,
				HW_DRV_DATA(vc, core), vc, event_iter, &iter);
		if (hpc_rc != OK)
			return CAFE_RESOURCE_UNAVAILABLE;

                vc->vc_reload[core] = vc->vc_evt_set.hpc_evt_settings.reload;
                vc->vc_reload_min[core] = ULLONG_MAX;
                vc->vc_time_prev[core] = 0;
                vc->vc_rate_prev[core] = 0;
                vc->vc_stop_count[core] = 0;
        }
        return CAFE_OK;
}

/******************************************************************************/

static struct hpc_virt_slave *get_vc_slave_for_inst(
		struct hpc_inst_item *hpc_inst, struct hpc_virt_counter *vc,
		int *c)
{
	struct hpc_virt_slave *vc_slave;

	while (*c < NUM_VC_SLAVES) {
		if (vc->vc_slave[*c].vcs_hpc_inst == hpc_inst) {
			vc_slave = &vc->vc_slave[*c];
			(*c)++;
			return vc_slave;
		}
		(*c)++;
	}
	return NULL;
}

static struct hpc_virt_slave *get_vc_slave(struct hpc_virt_counter *vc,
		int *c)
{
	while (*c < NUM_VC_SLAVES) {
		return &vc->vc_slave[(*c)++];
	}
	return NULL;
}

static bool vc_has_other_running_inst(struct hpc_inst_item *hpc_inst,
		struct hpc_virt_counter *vc)
{
	int c;

	for (c = 0; c < NUM_VC_SLAVES; c++) {
		if (vc->vc_slave[c].vcs_running && vc->vc_slave[c].vcs_hpc_inst != hpc_inst) {
			return true;
		}
	}
	return false;
}

struct vc_data_iter {
	int rc;
	struct hpc_virt_counter *vc;
};

static STATUS vc_stop_iter(int cpu, void *data)
{
	STATUS rc;
	struct vc_data_iter *iter_data = data;
	struct hpc_virt_counter *vc = iter_data->vc;

	rc = HW_DRV_CTR_STOP(vc->vc_evt_set.hw->drv.pDrvInst,
				HW_DRV_DATA(vc, cpu), &vc->vc_stop_count[cpu]);
	if (rc != OK) {
		hpc_error("error stopping counter on core %d\n", cpu);
		iter_data->rc = CAFE_ERROR;
		return ERROR;
	}

	return OK;
}

static int vc_suspend(struct hpc_virt_counter *vc)
{
	struct vc_data_iter iter_data;

	if (vc->vc_running == false) {
		hpc_error("vc not running so can't suspend\n");
		return CAFE_ERROR;
	}

	iter_data.rc = CAFE_OK;
	iter_data.vc = vc;

	(void)hpcIterateCpus(vc->vc_evt_set.hw->drv.numCpuCores, vc_stop_iter,
	                     &iter_data);
	if (iter_data.rc != CAFE_OK) {
		hpc_error("error stopping counters\n");
		return CAFE_ERROR;
	}
	vc->vc_running = false;

	return CAFE_OK;
}

static STATUS vc_start_iter(int cpu, void *data)
{
	STATUS rc;
	struct vc_data_iter *iter_data = data;
	struct hpc_virt_counter *vc = iter_data->vc;
	unsigned long long time;

	if (vc->vc_evt_set.hpc_evt_settings.samplePeriod) {
		(void)cafe_vx_timestamp64(&time);
		vc->vc_time_prev[cpu] = time;
	}
	rc = HW_DRV_CTR_START(vc->vc_evt_set.hw->drv.pDrvInst, HW_DRV_DATA(vc, cpu));

	if (rc != OK) {
		hpc_error("error starting counter on core %d\n", cpu);
		iter_data->rc = CAFE_ERROR;
		return ERROR;
	}

	return OK;
}

static int vc_start2(struct hpc_virt_counter *vc)
{
	struct vc_data_iter iter_data;

	vc->vc_running = true;

	iter_data.rc = CAFE_OK;
	iter_data.vc = vc;
	(void)hpcIterateCpus(vc->vc_evt_set.hw->drv.numCpuCores, vc_start_iter,
			     &iter_data);

	if (iter_data.rc != CAFE_OK) {
		hpc_error("error starting counters\n");
		/* WRS_TODO: stop any counters on prev cores */
		return CAFE_ERROR;
	}
	return CAFE_OK;
}

static int vc_stop(struct hpc_inst_item *hpc_inst, struct hpc_virt_counter *vc)
{
	int c, core, rc;
	struct vc_data_iter iter_data;
	bool vc_suspended = false, vc_slave_found = false;

	/* Check if vc is in this inst that we are stopping. */
	c = 0;
	while (get_vc_slave_for_inst(hpc_inst, vc, &c) != NULL) {
		vc_slave_found = true;
	}

	if (!vc_slave_found)
		/* This VC is not part of this inst */
		return CAFE_OK;

	if (vc->vc_running == false) {
		hpc_error("vc not running so can't stop\n");
		return CAFE_ERROR;
	}

	/* Stop the VC if need be. */
	if (!vc_has_other_running_inst(hpc_inst, vc)) {
		iter_data.rc = CAFE_OK;
		iter_data.vc = vc;
		(void)hpcIterateCpus(vc->vc_evt_set.hw->drv.numCpuCores,
				     vc_stop_iter, &iter_data);
		if (iter_data.rc != CAFE_OK) {
			hpc_error("error stopping counters\n");
			return CAFE_ERROR;
		}
		vc->vc_running = false;
	} else {
		vc_suspended = true;
		rc = vc_suspend(vc);
		if (rc != CAFE_OK) {
			hpc_error("unable to suspend vc\n");
			return rc;
		}
	}

	/* stop the slave counters in the requested instance and update the virtual
	 * counts */

	for (c = 0; c < NUM_VC_SLAVES; c++) {
		if (vc->vc_slave[c].vcs_hpc_inst == hpc_inst) {
			struct hpc_virt_slave *vc_slave = &vc->vc_slave[c];
			vc_slave->vcs_running = false;
			for (core = 0; core < vc->vc_evt_set.hw->drv.numCpuCores; core++) {
				vc_slave->vcs_count[core] += vc->vc_stop_count[core];
			}
		}
	}

	if (vc_suspended) {
		rc = vc_start2(vc);
		if (rc != CAFE_OK) {
			hpc_error("unable to restart suspended vc %d\n", rc);
			return rc;
		}
	}

	return CAFE_OK;
}

static HPC_COUNTER get_reload_value(struct hpc_inst_item *hpc_inst,
	struct hpc_virt_counter *vc)
{
	int c;
	struct hpc_virt_slave *vc_slave;
	HPC_COUNTER reload = ULLONG_MAX;

	for (c = 0; c < NUM_VC_SLAVES; c++) {
		vc_slave = &vc->vc_slave[c];
		if (vc_slave->vcs_hpc_inst != NULL) {
			if (vc_slave->vcs_reload < reload)
				reload = vc_slave->vcs_reload;
		}
	}

	return reload;
}

static int vc_start(struct hpc_inst_item *hpc_inst, struct hpc_virt_counter *vc)
{
	int rc, core, c;
	STATUS hpc_rc;
	HPC_COUNTER reload;
	struct hpc_virt_slave *vc_slave;
	bool vc_slave_found = false;
	bool vc_slave_stopped = false;
	bool vc_was_running = false;

	/* check if there are any vc_slaves to start in this VC */
	c = 0;
	while ((vc_slave = get_vc_slave_for_inst(hpc_inst, vc, &c)) != NULL) {
		vc_slave_found = true;
		if (!vc_slave->vcs_running) {
			vc_slave_stopped = true;
		}
	}

	if (!vc_slave_found)
		/* This VC is not part of this inst */
		return CAFE_OK;

	if (!vc_slave_stopped)
		return CAFE_OK; /* VC slave is already started */

	reload = get_reload_value(hpc_inst, vc);

	/* stop vc if its running - this prevents the counters ISR from
	 * executing */
	if (vc->vc_running) {
		rc = vc_suspend(vc);
		vc_was_running = true;
		if (rc != CAFE_OK) {
			hpc_error("unable to stop vc\n");
			return rc;
		}
	}

	/* update the reload, if necessary */
	if (vc->vc_evt_set.hpc_evt_settings.reload > reload) {
		vc->vc_evt_set.hpc_evt_settings.reload = reload;
			for (core = 0; core < vc->vc_evt_set.hw->drv.numCpuCores; core++) {
				hpc_rc = HW_DRV_CTR_RELOAD(vc->vc_evt_set.hw->drv.pDrvInst,
					HW_DRV_DATA(vc, core),
					&vc->vc_evt_set.hpc_evt_settings.reload);
				if (hpc_rc != OK) {
					hpc_error("unable to update reload %d\n", hpc_rc);
					return CAFE_ERROR;
				}
			}
		/* We need to account for the latest counter value for
		 * all slave counters that were running in this vc */
		c = 0;
		while ((vc_slave = get_vc_slave(vc, &c)) != NULL) {
			for (core = 0; core < vc->vc_evt_set.hw->drv.numCpuCores; core++) {
				if (vc_slave->vcs_hpc_inst == hpc_inst) {
					if (vc_was_running && vc_slave->vcs_running) {
						vc_slave->vcs_count[core] += vc->vc_stop_count[core];
					} else {
						vc_slave->vcs_count_offset[core] = 0;
					}
				} else {
					if (vc_was_running && vc_slave->vcs_running) {
						vc_slave->vcs_count[core] += vc->vc_stop_count[core];
					}
				}
			}
			if (vc_slave->vcs_hpc_inst == hpc_inst)
				vc_slave->vcs_running = true;
		}
	} else {
		c = 0;
		while ((vc_slave = get_vc_slave_for_inst(hpc_inst, vc, &c)) != NULL) {
			vc_slave->vcs_running = true;
			for (core = 0; core < vc->vc_evt_set.hw->drv.numCpuCores; core++) {
				if (vc_was_running)
					vc_slave->vcs_count_offset[core] = vc->vc_stop_count[core];
				else {
					vc_slave->vcs_count_offset[core] = 0;
				}
			}
		}
	}

	/* start the VC */
	rc = vc_start2(vc);
	if (rc != CAFE_OK) {
		hpc_error("unable to restart vc %d\n", rc);
		return rc;
	}

	return CAFE_OK;
}

/******************************************************************************/

static struct hpc_inst_item *hpc_create_inst(void)
{
	struct hpc_inst_item *hpc_inst;

	hpc_inst = alloc_ep_inst();

	return hpc_inst;
}

static int hpc_delete_inst(struct hpc_inst_item *hpc_inst)
{
	int c, rc;
	struct cafe_list_head *item, *tmp;
	struct hpc_virt_slave *vc_slave;

	cafe_list_foreach_safe(item, &ep_inst_desc.vc_in_use_list_head, tmp) {
		bool vc_still_in_use = false;
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);
		c = 0;
		while ((vc_slave = get_vc_slave(vc, &c)) != NULL) {
			if (vc_slave->vcs_hpc_inst == hpc_inst) {
				if (vc_slave->vcs_running) {
					hpc_error("Can't delete, instance is not stopped\n");
					return CAFE_ERROR;
				}
			} else if (vc_slave->vcs_hpc_inst != NULL) {
				vc_still_in_use = true;
			}
		}
		if (!vc_still_in_use) {
			cafe_list_remove(item);
			if ((rc = free_vc(vc)) != CAFE_OK) {
				hpc_error("Unable to free VC error %u\n", rc);
				return CAFE_ERROR;;
			}
		}
	}

	if ((rc = free_ep_inst(hpc_inst)) != CAFE_OK) {
		hpc_error("Unable to free the config %u\n", rc);
		return rc;
	}

	return CAFE_OK;
}

static int hpc_add_event_to_ep_inst(
	struct hpc_virt_counter **vc,
	struct hpc_inst_item *hpc_inst,
	struct hpc_event_desc *evt_set)
{
	int rc;

	if ((*vc = alloc_vc()) == NULL) {
		hpc_error("could not create virtual counter\n");
		return CAFE_NOMEM;
	}
	if ((rc = configure_vc_for_event(hpc_inst, *vc, evt_set)) != CAFE_OK)
		goto error;
	add_vc_to_in_use_list(*vc);

	return rc;
error:
	if (free_vc(*vc) != CAFE_OK)
		hpc_error("unable to free vc on error exit\n");
	return rc;
}

/******************************************************************************/

static int perf_counter_provider_create(struct cafe_ep_inst *inst)
{
	struct hpc_inst_item *hpc_inst;
	struct hpc_workarea *wa = (struct hpc_workarea *) inst->workarea->u.data;

	if ((hpc_inst = inst->client_data = hpc_create_inst()) == NULL) {
		hpc_error("failed to create hpc config\n");
		return CAFE_ERROR;
	}
	hpc_inst->change_count = inst->session->output_event_action_change_count - 1;
	hpc_inst->ep_inst = inst;

	if ((wa->max_stack_depth > MAX_STACK_DEPTH) || (wa->max_stack_depth == 0))
		wa->max_stack_depth = MAX_STACK_DEPTH;

	return CAFE_OK;
}

static void perf_counter_provider_delete(struct cafe_ep_inst *inst)
{
	int rc;
	struct hpc_inst_item *hpc_inst;

	if ((hpc_inst = (struct hpc_inst_item *)inst->client_data) == NULL) {
		hpc_error("delete given null client ptr\n");
		return;
	}
	if ((rc = hpc_delete_inst(hpc_inst)) != CAFE_OK) {
		hpc_error("failed to delete hpc config %u\n", rc);
	}
}

static int perf_counter_provider_start2(struct cafe_ep_inst *inst)
{
	int hpc_rc;
	struct hpc_inst_item *hpc_inst;
	struct cafe_list_head *item;

	if (inst == NULL) {
		hpc_error("start, given null inst\n");
		return CAFE_ERROR;
	}
	if ((hpc_inst = (struct hpc_inst_item *)inst->client_data) == NULL) {
		hpc_error("start given null client ptr\n");
		return CAFE_ERROR;
	}
	cafe_list_foreach(item, &ep_inst_desc.vc_in_use_list_head) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);
		if ((hpc_rc = vc_start(hpc_inst, vc)) != CAFE_OK) {
			hpc_error("error starting vc\n");
			return hpc_rc;
		}
	}
	return CAFE_OK;
}

static int perf_counter_provider_start(struct cafe_ep_inst *inst)
{
	int hpc_rc;

	cafe_ep_inst_state_started(inst);
	hpc_rc = perf_counter_provider_start2(inst);
	if (hpc_rc != CAFE_OK) {
		return hpc_rc;
	}
	return CAFE_OK;
}

static int perf_counter_provider_stop(struct cafe_ep_inst *inst)
{
	struct cafe_list_head *item;
	int hpc_rc;
	struct hpc_inst_item *hpc_inst;

	if (inst == NULL) {
		hpc_error("start, given null inst\n");
		return CAFE_ERROR;
	}
	if ((hpc_inst = (struct hpc_inst_item *)inst->client_data) == NULL) {
		hpc_error("start given null client ptr\n");
		return CAFE_ERROR;
	}
	cafe_list_foreach(item, &ep_inst_desc.vc_in_use_list_head) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);

		if ((hpc_rc = vc_stop(hpc_inst, vc)) != CAFE_OK) {
			hpc_error("error stopping vc\n");
			return hpc_rc;
		}
	}
	cafe_ep_inst_state_stopped(inst);
	return CAFE_OK;
}

/*
 * In order to figure out if we are enabling the same counter we must compare
 * the hardware settings of the event. This is because the hpc.total_cycles
 * event and the hpc_od.total_cycles event have a different value type and
 * so can't be compared, when inface they are the same physical evemt.
 */
static int cmp_vc_settings(struct hpc_event_desc *evt_set,
			   struct hpc_virt_counter *vc)
{
	if (memcmp(&vc->vc_evt_set.drv_evt_settings, &evt_set->drv_evt_settings,
							sizeof(union eventSettings)) == 0)
		return 1;

	return 0;
}

static struct hpc_virt_counter *check_for_existing_vc(
	struct hpc_event_desc *event_desc)
{
	struct cafe_list_head *item_vc;
	struct hpc_virt_counter *vc;

	cafe_list_foreach(item_vc, &ep_inst_desc.vc_in_use_list_head) {
		vc = VC_NON_MUXED_LIST_ITEM(item_vc);
		if (cmp_vc_settings(event_desc, vc)) {
			return vc;
		}
	}

	return NULL;
}

static HPC_COUNTER get_skipcount(struct cafe_value_type *type)
{
	const struct cafe_vm_expr *expr;
	struct cafe_value_type * vt = cafe_base_type(type);

	if ((vt == NULL) || (strcmp(vt->ident->name, "skipcount") != 0))
		return 0;

	expr = cafe_param_value_get(type, "limit");
	if (expr == NULL)
		return 0;
	return expr->u.ull;
}

static unsigned int get_sample_freq(struct cafe_value_type *type)
{
	const struct cafe_vm_expr *expr;
	struct cafe_value_type * vt = cafe_base_type(type);

	if ((vt == NULL) || (strcmp(vt->ident->name, "sample_freq") != 0))
		return 0;

	expr = cafe_param_value_get(type, "rate");
	if (expr == NULL)
		return 0;
	return (unsigned int)expr->u.ull;
}

static int event_odata(struct cafe_event_action_info *action_info,
	struct hpcEventSettings	*settings)
{
	struct cafe_event_descr *event;
	struct cafe_value_type **odata;
	int i;

	event = action_info->event;
	odata = action_info->inst->output_odata;
	for (i = 1; i < event->num_types; ++i) {
		if (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SKIPCOUNT]) {
			settings->reload = get_skipcount(event->types[i]);
			break;
		}
		if (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SAMPLE_FREQ]) {
			settings->samplePeriod = timer_freq / get_sample_freq(event->types[i]);
			break;
		}
	}

	return CAFE_OK;
}

static int get_hpc_od_index_offset(struct cafe_event_action_info *eai)
{
	struct cafe_event_descr *event;
	struct cafe_value_type **odata;
	int i, num_odtypes;

	event = eai->event;
	if (event->num_types == 0)
		return 0;
	odata = eai->inst->output_odata;
	num_odtypes = event->num_types - 1;
	i = 1;
	if (event->types[i] == odata[ODATA_INDEX_TS]) {
		if (i == num_odtypes)
			return i;
		i++;
	}
	if (event->types[i] == odata[ODATA_INDEX_CORE]) {
		if (i == num_odtypes)
			return i;
		i++;
	}
	if (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SKIPCOUNT]) {
		if (i == num_odtypes)
			return i;
		i++;
	}
	if (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SAMPLE_FREQ]) {
		if (i == num_odtypes)
			return i;
		i++;
	}

#if defined(__VXWORKS__)
	if (event->types[i] == odata[ODATA_INDEX_CONTEXT]) {
		if (i == num_odtypes)
			return i;
		i++;
	}
#endif
#if defined(_WRS_CONFIG_COMPONENT_STACKTRACE) || !defined(__VXWORKS__)
	if (event->types[i] == odata[ODATA_INDEX_STACK]) {
		if (i == num_odtypes)
			return i;
		i++;
	}
#endif
	if (event->types[i] == odata[ODATA_INDEX_SKIPCOUNT_ACTUAL]) {
		if (i == num_odtypes)
			return i;
		i++;
	}
	return i;
}

static bool matchHpcOd(struct hpc_virt_slave *pri_vcs,
	struct cafe_event_action_info *eai)
{
	int i, hpc_od, num_hpc_od_types1 = 0, num_hpc_od_types2 = 0;
	struct cafe_list_head *item;
	struct cafe_value_type **odata;
	struct cafe_value_type *base_type;
	struct cafe_event_descr *event;
	bool found;

	assert(eai != NULL);
	event = eai->event;
	odata = eai->inst->output_odata;
	for (i = 1; i < event->num_types; i++) {
		base_type = cafe_base_type(event->types[i]);
		for (hpc_od = 0; hpc_od < NUM_ODATA_HPC_ITEMS; hpc_od++) {
			if (base_type == odata[hpc_od + NUM_ODATA_CAFE_ITEMS]) {
				/* matched a requested type with hpc odata */
				/* now check if event is counting it. */
				/* iterate over secondary counters for this type */
				num_hpc_od_types1++;
				found = false;
				cafe_list_foreach(item, &pri_vcs->vcs_sec_list_head) {
					struct hpc_virt_slave *sec_vcs = VCS_SEC_LIST_ITEM(item);
					if (sec_vcs->vcs_type == base_type) {
						found = true;
						break;
					}
				}
				if (found)
					break;
				else {
					/* this vcs does not count the requested
					 * hpc OD */
					return false;
				}
			}
		}
	}
	/* Check that requested event and existing configured event have
	 * the same number of hpc od types */
	cafe_list_foreach(item, &pri_vcs->vcs_sec_list_head) {
		num_hpc_od_types2++;
	}
	if (num_hpc_od_types1 == num_hpc_od_types2) {
		return true;
	} else {
		return false;
	}
}

static int get_vc_output_evt_for_inst2(struct hpc_inst_item *hpc_inst,
	struct hpc_evt_imp_data *evt_impl_data,
	struct cafe_event_action_info *eai,
	unsigned int type_idx,
	struct hpc_virt_slave *pri_vcs, /* null means configure primary */
					/* non null means secondary vcs */
	struct hpc_virt_slave **vc_slave_param,
	bool *vcs_reused)
{
	struct hw_driver_inst *hw = evt_impl_data->hw;
	struct hpc_event_desc event_desc;
	struct hpc_virt_counter *vc;
	struct hpc_virt_slave *vcs;
	int rc, hpc_od_offset;

	/* search for existing event in the system */
	memset(&event_desc, 0, sizeof(event_desc));
	event_desc.hw = hw;

	if (pri_vcs == NULL) {
		rc = event_odata(eai, &event_desc.hpc_evt_settings);
		if (rc != CAFE_OK) {
			hpc_error("unable to get odata settings %s\n", eai->event->types[0]->type_str);
			return rc;
		}
	} else {
		event_desc.hpc_evt_settings.reload = ULLONG_MAX;
	}

	/* get event settings */
	rc = hw->drv.pEventSettingsGet(eai->event->types[type_idx],
					&event_desc.drv_evt_settings,
					&event_desc.hpc_evt_settings,
					evt_impl_data->evt_desc);

	if (rc != CAFE_OK) {
		hpc_error("unable to get settings %s\n", eai->event->types[0]->type_str);
		return rc;
	}
	if (pri_vcs == NULL)
		hpc_od_offset = get_hpc_od_index_offset(eai);
	else
		hpc_od_offset = -1;
	vc = check_for_existing_vc(&event_desc);
	if (vc != NULL) {
		/* the found vc counts this event */
		int c; /* slave counter */
		struct hpc_virt_slave *vcs_free = NULL;
		/* now check if this instance already counts this event */
		for (c = 0; c < NUM_VC_SLAVES; c++) {
			vcs = &vc->vc_slave[c];
			if ((pri_vcs == NULL && 		/* do check for primary vcs */
			    vcs->vcs_hpc_inst == hpc_inst &&
			    vcs->vcs_hpc_od_offset == hpc_od_offset &&
			    vcs->vcs_reload == event_desc.hpc_evt_settings.reload &&
			    matchHpcOd(vcs, eai) &&
			    vcs->vcs_pri == NULL) ||
			    (pri_vcs != NULL &&		/* do check for secondary vcs */
			    pri_vcs == vcs->vcs_pri &&
			    vcs->vcs_hpc_inst == hpc_inst)) {
				/* yes it does */
				vcs->vcs_enabled = true;
				*vc_slave_param = vcs;
				hpc_debug(1, "VC: reuse vc %p vcs %p\n", vc, vcs);
				*vcs_reused = true;
				return CAFE_OK;
			}
			if (vcs->vcs_hpc_inst == NULL) {
				vcs_free = &vc->vc_slave[c];
			}
		}
		/* no it doesn't, so use a free slave vc */
		if (vcs_free != NULL) {
			memset(vcs_free, 0, sizeof(*vcs_free));
			cafe_list_init(&vcs_free->vcs_sec_list_head);
			vcs_free->vcs_sink_act_handle = CAFE_EAL_HANDLE_NULL;
			vcs_free->vcs_enabled = true;
			vcs_free->vcs_reload = event_desc.hpc_evt_settings.reload;
			vcs_free->vcs_hpc_inst = hpc_inst;
			vcs_free->vcs_vc = vc;
			vcs_free->vcs_type = eai->event->types[type_idx];
			vcs_free->vcs_hpc_od_offset = hpc_od_offset;
			*vc_slave_param = vcs_free;
			*vcs_reused = false;
			hpc_debug(1, "VC: vc %p new vcs %p\n", vc, vcs_free);
			return CAFE_OK;
		} else {
			hpc_debug(1, "VC: no vcs available %d\n", NUM_VC_SLAVES);
			*vc_slave_param = NULL;
			return CAFE_RESOURCE_UNAVAILABLE;
		}
	}
	/* attempts to reuse an existing vc failed, create a new one */
	rc = hpc_add_event_to_ep_inst(&vc, hpc_inst, &event_desc);
	if (rc != CAFE_OK) {
		hpc_debug(1, "VC: no vc available\n");
		*vc_slave_param = NULL;
		return rc;
	}
	vcs = &vc->vc_slave[0];
	memset(vcs, 0, sizeof(*vcs));
	cafe_list_init(&vcs->vcs_sec_list_head);
	vcs->vcs_sink_act_handle = CAFE_EAL_HANDLE_NULL;
	vcs->vcs_enabled = true;
	vcs->vcs_reload = event_desc.hpc_evt_settings.reload;
	vcs->vcs_hpc_inst = hpc_inst;
	vcs->vcs_vc = vc;
	vcs->vcs_type = eai->event->types[type_idx];
	vcs->vcs_hpc_od_offset = hpc_od_offset;
	*vc_slave_param = vcs;
	*vcs_reused = false;
	hpc_debug(1, "VC: new vc %p new vcs %p\n", vc, vcs);

	return CAFE_OK;
}

static int get_sec_hpc_odata_event(
	int *e,
	struct cafe_event_action_info *eai,
	struct hpc_evt_imp_data **evt_impl_data)
{
	struct cafe_event_descr *event;
	struct cafe_value_type **odata;
	int item;
	event = eai->event;
	odata = eai->inst->output_odata;

	while (*e < event->num_types) {
		for (item = NUM_ODATA_CAFE_ITEMS; item < hpc_root_spec.n_output_odata; item++) {
			if (cafe_base_type(event->types[*e]) == odata[item]) {
				*evt_impl_data = hpc_events_output_odata[item - NUM_ODATA_CAFE_ITEMS];
				(*e)++;
				return *e - 1;
			}
		}
		(*e)++;
	}
	return 0;
}

static void cleanup_vc(struct hpc_virt_counter *vc)
{
	bool vc_in_use;
	int c, rc;

	/* check for other vcs, if none, free the vc */
	vc_in_use = false;
	for (c = 0; c < NUM_VC_SLAVES; c++) {
		struct hpc_virt_slave *vcs = &vc->vc_slave[c];
		if (vcs->vcs_hpc_inst != NULL) {
			vc_in_use = true;
			break;
		}
	}
	if (!vc_in_use) {
		cafe_list_remove(&vc->vc_in_use_list);
		if ((rc = free_vc(vc)) != CAFE_OK) {
			hpc_error("Unable to free VC error %u\n", rc);
		}
	}

}

static void cleanup_unstarted_pri_vcs(struct hpc_inst_item *hpc_inst,
	struct hpc_virt_slave *pri_vcs)
{
	struct cafe_list_head *item, *tmp;

	/* cleanup any secondary counters */
	cafe_list_foreach_safe(item, &pri_vcs->vcs_sec_list_head, tmp) {
		struct hpc_virt_slave *sec_vcs = VCS_SEC_LIST_ITEM(item);
		cafe_list_remove(item);
		sec_vcs->vcs_hpc_inst = NULL;
		cleanup_vc(sec_vcs->vcs_vc);
	}
	pri_vcs->vcs_hpc_inst = NULL;
	cleanup_vc(pri_vcs->vcs_vc);
}

static int get_vc_output_evt_for_inst(struct hpc_inst_item *hpc_inst,
	struct hpc_evt_imp_data *pri_eid, /* primary event implementation data */
	struct cafe_event_action_info *pri_eai, /* primary event action info */
	cafe_eal_handle_t sa_handle)
{
	int rc, e = 1, hpc_od_index, sec_type_idx;
	bool vcs_reused;
	struct hpc_virt_slave *pri_vcs, *sec_vcs; /*primary/secondary vcs */
	struct hpc_evt_imp_data *sec_eid; /* secondary event impl data */

	/* Configure a counter for the base event (primary counter) */
	rc = get_vc_output_evt_for_inst2(hpc_inst, pri_eid, pri_eai, 0, NULL, &pri_vcs, &vcs_reused);
	if (rc != CAFE_OK) {
		hpc_debug(1, "ERROR: get_vc_output_evt_for_inst1\n");
		return rc;
	}

	if (vcs_reused) {
		struct cafe_list_head *item;
		cafe_list_foreach(item, &pri_vcs->vcs_sec_list_head) {
			struct hpc_virt_slave *sec_vcs = VCS_SEC_LIST_ITEM(item);
			sec_vcs->vcs_enabled = true;
		}

		goto skip_configure_hpc_od;
	}

	/* Configure counter(s) for any HPC odata events (secondary counters) */
	hpc_od_index = get_hpc_od_index_offset(pri_eai);
	while ((sec_type_idx = get_sec_hpc_odata_event(&e, pri_eai, &sec_eid)) != 0) {
		rc = get_vc_output_evt_for_inst2(hpc_inst, sec_eid, pri_eai, sec_type_idx, pri_vcs, &sec_vcs, &vcs_reused);

		if (rc != CAFE_OK) {
			cleanup_unstarted_pri_vcs(hpc_inst, pri_vcs);
			hpc_debug(1, "ERROR: get_vc_output_evt_for_inst2\n");
			return rc;
		}
		if (sec_vcs->vcs_pri == NULL) {
			/* this is a new vcs */
			sec_vcs->vcs_pri = pri_vcs;
			sec_vcs->vcs_odata_idx = hpc_od_index++;
			cafe_list_add_last(&sec_vcs->vcs_sec_list, &pri_vcs->vcs_sec_list_head);
		}
	}
skip_configure_hpc_od:
	/* note: the counter could be running at this point */
	cafe_event_action_list_swap(&pri_vcs->vcs_sink_act_handle, sa_handle);
	return rc;
}

static void clear_vc_event_enabled_flag(struct hpc_inst_item *hpc_inst)
{
	struct cafe_list_head *item;

	cafe_list_foreach(item, &ep_inst_desc.vc_in_use_list_head) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);
		int c;
		struct hpc_virt_slave *vc_slave;
		for (c = 0; c < NUM_VC_SLAVES; c++) {
			vc_slave = &vc->vc_slave[c];
			if (vc_slave->vcs_hpc_inst == hpc_inst) {
				vc_slave->vcs_enabled = false;
			}
		}
	}
}

static int disable_unused_events_vc(struct hpc_inst_item *hpc_inst,
	struct hpc_virt_counter *vc,
	bool first_pass)
{
	int rc, c;
	struct hpc_virt_slave *vc_slave;
	bool vcs_to_disable = false;
	bool vc_in_use = false;

	for (c = 0; c < NUM_VC_SLAVES; c++) {
		vc_slave = &vc->vc_slave[c];
		if (vc_slave->vcs_hpc_inst == NULL)
			continue;
		if (vc_slave->vcs_hpc_inst != hpc_inst) {
			/* VC has a counter in another instance, can't disable */
			vc_in_use = true;
			continue;
		}
		/* we now have a vc_slave that is from this instance */
		if (first_pass) {
			/* check if enabled or if its a secondary vcs */
			if (vc_slave->vcs_enabled || vc_slave->vcs_sec_list.next != NULL) {
				/* This event is still required */
				vc_in_use = true;
			} else {
				/* This event is no longer required */
				vcs_to_disable = true;
			}
		} else {
			if (vc_slave->vcs_enabled) {
				/* This event is still required */
				vc_in_use = true;
			} else {
				/* This event is no longer required */
				vcs_to_disable = true;
			}
		}
	}
	if (vcs_to_disable) {
		bool vc_was_running = false;
		/* free the vc slave(s) that are no longer enabled */
		if (vc->vc_running) {
			vc_was_running = true;
			if ((rc = vc_suspend(vc)) != CAFE_OK) {
				hpc_error("unable to stop vc %p\n", vc);
				return rc;
			}
		}
		for (c = 0; c < NUM_VC_SLAVES; c++) {
			vc_slave = &vc->vc_slave[c];
			if (vc_slave->vcs_hpc_inst == hpc_inst && !vc_slave->vcs_enabled) {
				struct cafe_list_head *item, *tmp;
				cafe_list_foreach_safe(item, &vc_slave->vcs_sec_list_head, tmp) {
					cafe_list_remove(item);
				}
				vc_slave->vcs_hpc_inst = NULL;
				vc_slave->vcs_running = false;
				cafe_event_action_list_swap(&vc_slave->vcs_sink_act_handle, CAFE_EAL_HANDLE_NULL);
			}
		}
		if (vc_was_running && vc_in_use) {
			rc = vc_start2(vc);
			if (rc != CAFE_OK) {
				hpc_error("unable to restart suspended vc %d %p\n", rc, vc);
				return rc;
			}
		}
	}
	if (!vc_in_use) {
		/* stop the event and free the vc */
		if (vc->vc_running)
			if ((rc = vc_stop(hpc_inst, vc)) != CAFE_OK)
				hpc_error("could not stop vc %p\n", vc);
		cafe_list_remove(&vc->vc_in_use_list);
		if ((rc = free_vc(vc)) != CAFE_OK) {
			hpc_error("Unable to free VC error %u\n", rc);
			return rc;
		}
	}
	return CAFE_OK;
}

static int disable_unused_events(struct hpc_inst_item *hpc_inst)
{
	struct cafe_list_head *item, *tmp;
	int rc;

	/* there are two passes needed here. The first pass will not disable
	 * secondary vcs, it will only remove them from the primary vcs_sec
	 * list. The second pass will then disable them fully. */
	cafe_list_foreach_safe(item, &ep_inst_desc.vc_in_use_list_head, tmp) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);
		rc = disable_unused_events_vc(hpc_inst, vc, true);
		if (rc != CAFE_OK)
			return rc;
	}
	cafe_list_foreach_safe(item, &ep_inst_desc.vc_in_use_list_head, tmp) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);
		rc = disable_unused_events_vc(hpc_inst, vc, false);
		if (rc != CAFE_OK)
			return rc;
	}
	return CAFE_OK;
}

#if 0
static struct cafe_event_descr *rewrite_event(
	struct cafe_ep_inst *inst,
	struct cafe_event_descr *event)
{
	struct cafe_value_type *types[CAFE_MAX_EVENT_TYPES];
	struct cafe_value_type **odata;
	unsigned int i = 0, j = 0, rc;

	odata = inst->output_odata;
	types[j++] = event->types[i++];
	if (event->types[i] == odata[ODATA_INDEX_TS]) {
		types[j++] = event->types[i++];
	}
	if (event->types[i] == odata[ODATA_INDEX_CORE]) {
		types[j++] = event->types[i++];
	}
	/* always include skipcount in the event */
	types[j++] = odata[ODATA_INDEX_SKIPCOUNT];
	if (cafe_base_type(event->types[i]) == odata[ODATA_INDEX_SKIPCOUNT]) {
		i++;
	}
	while (i < event->num_types)
		types[j++] = event->types[i++];

	rc = cafe_cfg_get_event(inst->session, j, types, &event);
	if (rc != CAFE_OK)
		return NULL;

	return event;
}
#endif

static void dispatch_status(struct cafe_error_report *err,
                            struct cafe_ep_inst *inst)
{
	struct cafe_byte_buffer buf;

	(void)cafe_bbuf_init(&buf, inst->allocator, 128);
	(void)cafe_error_bbuf_write(err, &buf, NULL, ": ");
	cafe_dispatch_status(inst, CAFE_LOG_WARNING, buf.data);
	cafe_bbuf_release(&buf);
	cafe_error_clear(err);
}

static int perf_counter_update_output_event_actions2(
	struct cafe_ep_inst *inst, 
	bool warnReport)
{
	int i, rc;
	struct hpc_inst_item *hpc_inst = (struct hpc_inst_item *)inst->client_data;
	struct cafe_event_action_info *eai;
	struct cafe_event_descr *event;
	struct cafe_event_action_list *sa;
	struct hpc_evt_imp_data *evt_impl_data;
	cafe_eal_handle_t sa_handle;
	bool warnings_exist = false;

	for (i = inst->enabled_output_events.ind - 1; i >= 0; i--) {
		event = inst->enabled_output_events.list[i].event;

/*		event = rewrite_event(inst, event); */
		if (event == NULL) {
			hpc_error("hpc UOEA: unable to rewrite event\n");
			return CAFE_ERROR;
		}
		rc = cafe_ep_inst_event_action_info_get(inst, event, &eai);
		if (rc != CAFE_OK || eai == NULL) {
			hpc_error("hpc UOEA: can't get event info %p\n", eai);
			return rc;
		}
		assert(event == eai->event);
		assert(inst == eai->inst);
		/* make a new action list */
		rc = cafe_event_action_list_alloc(inst, event, &sa, &sa_handle);
		if (rc != CAFE_OK) {
			hpc_error("hpc UOEA: unable to alloc action list\n");
			return rc;
		}
		rc = cafe_ep_inst_get_sink_event_actions(inst, sa);
		if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sa_handle);
			hpc_error("hpc UOEA: can't get sink actions\n");
			return rc;
		}
		/* Check if the event has any action handlers */
		if (sa->ind == 0) {
			cafe_event_action_list_release(&sa_handle);
			continue;
		}
		evt_impl_data = (struct hpc_evt_imp_data *)
				inst->enabled_output_events.list[i].impl_data;

		rc = get_vc_output_evt_for_inst(hpc_inst, evt_impl_data, eai, sa_handle);
		if (rc == CAFE_RESOURCE_UNAVAILABLE) {
			/* No counter available for event */
			cafe_event_action_list_release(&sa_handle);
			if (warnReport)	{
				cafe_error_set (NULL, rc, "error setting event <%s>", event->types[0]->type_str);
				warnings_exist = true;
			}
			continue;
		} else if (rc != CAFE_OK) {
			cafe_event_action_list_release(&sa_handle);
			hpc_error("Unable to enable hpc event to %s\n",
						event->types[0]->ident->name);
			break;
		}
	}
	if (warnings_exist) {
		dispatch_status (&cafe_error_cur, inst);
	}
	return CAFE_OK;
}

static int perf_counter_update_output_event_actions(
	struct cafe_ep_inst *inst)
{
	struct hpc_inst_item *hpc_inst;
	int rc;

	if (inst == NULL) {
		hpc_error("given null inst ptr\n");
		return CAFE_ERROR;
	}
	if ((hpc_inst = (struct hpc_inst_item *)inst->client_data) == NULL) {
		hpc_error("given null client ptr\n");
		return CAFE_ERROR;
	}
	/* Check if anything changed since last time we updated. */
	if (hpc_inst->change_count == inst->session->output_event_action_change_count)
		return CAFE_OK;
	hpc_inst->change_count = inst->session->output_event_action_change_count;

	clear_vc_event_enabled_flag(hpc_inst);
	if ((rc = perf_counter_update_output_event_actions2(inst,false)) != CAFE_OK) {
		hpc_error("unable to enable some events 1\n");
		return rc;
	}
	if ((rc = disable_unused_events(hpc_inst)) != CAFE_OK) {
		hpc_error("unable to disable unused events\n");
		return rc;
	}
	/* Now, some events may have been disabled so it may be possible
	 * to enable more events */

	if ((rc = perf_counter_update_output_event_actions2(inst,true)) != CAFE_OK) {
		hpc_error("unable to enable some events 2\n");
		return rc;
	}
	if (inst->state == CAFE_EP_INST_STATE_STARTED) {
		if ((rc = perf_counter_provider_start2(inst)) != OK) {
			hpc_error("unable to start inst on UOEA\n");
			return rc;
		}
	}
	hpcChkEvtPeriodTaskResume();
	return CAFE_OK;
}

int hpc_unregister_configuration(void)
{
	int cafe_rc;

	if ((cafe_rc = cafe_ep_spec_unregister(&hpc_root_spec)) != CAFE_OK) {
		hpc_error("failed to unregister hpc config\n");
		return CAFE_ERROR;
	}
	if (hpc_root_spec.output_events != NULL) {
		if (hpc_root_spec.output_events[0].impl_data != NULL)
			free(hpc_root_spec.output_events[0].impl_data);
		(*hpc_free)(hpc_root_spec.output_events);
		hpc_root_spec.n_output_events = 0;
		hpc_root_spec.output_events = NULL;
	}
	hpc_root_spec.spec_list.next = NULL;
	hpc_root_spec.spec_list.prev = NULL;
	return CAFE_OK;
}

static const char *hpc_odata_type_defs[NUM_ODATA_HPC_ITEMS] = {
	hpc_od_total_cycles_VT_DEF,
	hpc_od_total_instructions_VT_DEF,
	hpc_od_load_instructions_VT_DEF,
	hpc_od_store_instructions_VT_DEF};

static void reg_config_hpc_odata(int *num, struct hpc_evt_imp_data *evt_imp_data)
{
	int e;
	const char *hpc_odata_def[NUM_ODATA_HPC_ITEMS] = {
		hpc_total_cycles_VT_DEF,
		hpc_total_instructions_VT_DEF,
		hpc_load_instructions_VT_DEF,
		hpc_store_instructions_VT_DEF
	};
	for (e = 0; e < NUM_ODATA_HPC_ITEMS; e++) {
		if (strcmp(hpc_odata_def[e], evt_imp_data->evt_desc->pEventNameDef) == 0) {
			output_odata[*num] = hpc_odata_type_defs[e];
			hpc_debug(1, "hpc root OD event %d %d evt:%s odata:%s\n", e, *num, hpc_odata_def[e], output_odata[*num]);
			hpc_events_output_odata[*num - NUM_ODATA_CAFE_ITEMS] = evt_imp_data;
			(*num)++;
			break;
		}
	}
}

int hpc_register_configuration(void)
{
	int rc = CAFE_OK;
	size_t output_event_sz, imp_data_sz;
	struct hw_driver_inst *hw;
	HPC_DRIVER_IDX driver;
	int n_evts = 0, o_evt = 0, evt, num_odata = NUM_ODATA_CAFE_ITEMS;
	struct cafe_ep_spec_output_evt *output_event;
	struct hpc_evt_imp_data *evt_imp_data;

	/* The VC list can only be created once all drivers have registered. */
	if ((rc = init_vc_list()) != CAFE_OK) {
		hpc_error("Unable to init virtual counters\n");
		return rc;
	}

	FOREACH_HPC_DRIVER(hw, driver) {
		n_evts += hw->drv.numEvents;
	}
	if (n_evts == 0) {
		return CAFE_OK;
	}
	output_event_sz = n_evts * sizeof(*output_event);
	output_event = (*hpc_malloc)(output_event_sz);
	if (output_event == NULL) {
		hpc_error("no memory for configuration event list %d\n", n_evts);
		return CAFE_NOMEM;
	}
	memset(output_event, 0, output_event_sz);

	imp_data_sz = n_evts * sizeof(struct hpc_evt_imp_data);
	evt_imp_data = (*hpc_malloc)(imp_data_sz);
	if (evt_imp_data == NULL) {
		hpc_error("no memory for imp event data list %d\n", n_evts);
		free(output_event);
		return CAFE_NOMEM;
	}
	memset(evt_imp_data, 0, imp_data_sz);

	FOREACH_HPC_DRIVER(hw, driver) {
	    struct hpcEventDescriptor *pEventDesc;
	    char * p = (char *)hw->drv.pEventDesc;
		for (evt = 0; evt < hw->drv.numEvents; evt++, o_evt++) {
			pEventDesc = (struct hpcEventDescriptor *)p;
			output_event[o_evt].type = pEventDesc->pEventNameDef;
			hpc_debug(4, "Reg Type %s impl_data %p\n", output_event[o_evt].type, evt_imp_data);
			output_event[o_evt].impl_data = evt_imp_data;
			evt_imp_data->hw = hw;
			evt_imp_data->evt_desc = pEventDesc;
			reg_config_hpc_odata(&num_odata, evt_imp_data);
			evt_imp_data++;
			p += hw->drv.eventSettingSize;
		}
	}
	hpc_root_spec.output_events = output_event;
	hpc_root_spec.n_output_events = n_evts;
	hpc_root_spec.n_output_odata = num_odata;

	rc = cafe_ep_spec_register(&hpc_root_spec);
	if (rc != CAFE_OK) {
		hpc_error("hpc spec failed to register\n");
		goto error;
	}
	if ((rc = cafe_vx_timestamp_freq64(&timer_freq)) != CAFE_OK) {
		hpc_error("unable to get timer frequency");
		timer_freq = 0;
	}
	if (hpcChkEvtPeriodTaskStart(hpc_vc_check_period) != OK)
		return CAFE_ERROR;

	return CAFE_OK;

error:
	if (hpc_root_spec.output_events != NULL) {
		(*hpc_free)(hpc_root_spec.output_events);
		hpc_root_spec.n_output_events = 0;
		hpc_root_spec.output_events = NULL;
	}
	if (hpc_root_spec.types != NULL) {
		(*hpc_free)((char *)hpc_root_spec.types);
		hpc_root_spec.types = NULL;
	}
	return rc;
}

/*****************************************************************************/

static int vc_check_period(struct hpc_virt_counter *vc, int core)
{
	int c, rc, attempts = 10;
	unsigned long long t_now, t_delta, t_prev1, t_prev2;
	HPC_COUNTER ctr;

	if (!vc->vc_running)
		return CAFE_OK;
	do {
		/* vc_time_prev cound be changing in perf isr */
		t_prev1 = vc->vc_time_prev[core];
		t_prev2 = vc->vc_time_prev[core];
	} while (t_prev1 != t_prev2 && --attempts);
	if (attempts == 0)
		return CAFE_OK;
	if ((rc = cafe_vx_timestamp64(&t_now)) != CAFE_OK)
		return rc;
	if (t_now <= t_prev1)
		t_now += timer_freq;
	t_delta = t_now - t_prev1;
	if (vc->vc_evt_set.hpc_evt_settings.samplePeriod == 0)
		return CAFE_OK;
	if (t_delta > vc->vc_evt_set.hpc_evt_settings.samplePeriod) {
		/* stop the counter */
		rc = HW_DRV_CTR_STOP(vc->vc_evt_set.hw->drv.pDrvInst,
				HW_DRV_DATA(vc, core), &ctr);
		if (rc != CAFE_OK)
			return CAFE_ERROR;
		hpc_counter_interrupted_count++;
		/* with the counter stopped, vc_time_prev will not update */
		cafe_vx_timestamp64(&t_now);
		if (t_now <= vc->vc_time_prev[core])
			t_now += timer_freq;
		t_delta = t_now - vc->vc_time_prev[core];
		vc->vc_time_prev[core] = t_now;
		for (c = 0; c < NUM_VC_SLAVES; c++) {
			struct hpc_virt_slave *vc_slave = &vc->vc_slave[c];
			if (!vc_slave->vcs_running)
				continue;
			vc_slave->vcs_count[core] += ctr;
		}
		/* update the reload */
		if (vc->vc_reload_min[core] != ULLONG_MAX)
			vc->vc_reload[core] = vc->vc_reload_min[core];
		else
			vc->vc_reload[core] = 1;
		vc->vc_rate_prev[core] = 0;
		rc = HW_DRV_CTR_RELOAD(vc->vc_evt_set.hw->drv.pDrvInst, HW_DRV_DATA(vc, core),
				&vc->vc_reload[core]);
		if (rc != CAFE_OK)
			return CAFE_ERROR;
		/* restart the counter */
		rc = HW_DRV_CTR_START(vc->vc_evt_set.hw->drv.pDrvInst, HW_DRV_DATA(vc, core));
		if (rc != CAFE_OK)
			return CAFE_ERROR;
	}
	return CAFE_OK;
}

static STATUS vc_check_period_per_core_iter(int core, void *data)
{
	int *work = data;
	struct cafe_list_head *item;
	int rc;

	*work = 0;
	cafe_list_foreach(item, &ep_inst_desc.vc_in_use_list_head) {
		struct hpc_virt_counter *vc = VC_NON_MUXED_LIST_ITEM(item);
		/* Check if this VC exists on this core */
		(*work)++;
		if (core >= vc->vc_evt_set.hw->drv.numCpuCores) {
			return ERROR;
		}
		rc = vc_check_period(vc, core);
		if (rc != CAFE_OK)
			fprintf(stderr, "Error checking HPC virtual counter period: %s\n",
				cafe_strerror(rc));
	}
	return OK;
}

int hpc_vc_check_period(void)
{
	int work;
	CAFE_LOCK();
	(void)hpcIterateCpus(vc_blk_desc.max_cores,
			     vc_check_period_per_core_iter, &work);
	CAFE_UNLOCK();
	return work;
}

/******************************************************************************/

static STATUS hpc_per_core_init_iter(int cpu, void *data)
{
	STATUS status;
	struct hw_driver_inst *drv_inst = data;

	status = drv_inst->drv.pFuncs->hpcHwPerCoreInit(drv_inst->drv.pDrvInst);

	if (status != OK) {
		hpc_error("hpcHwPerCoreInit error\n");
	}
	return status;
}

static int hpc_hw_driver_register2(struct hpcDriverConfig *driver_data,
	struct hw_driver_inst *drv_inst)
{
	STATUS status;
	size_t size;
	struct hpcHwDriverInitData drv_inst_init_data;
	STATUS rc;

	/* Take a copy of the driver configuration params */
	drv_inst->in_use = true;
	drv_inst->drv.numEvents = driver_data->numEvents;
	drv_inst->drv.numCpuCores = driver_data->numCpuCores;
	drv_inst->drv.pDrvInst = driver_data->pDrvInst;
	drv_inst->drv.pFuncs = driver_data->pFuncs;
	drv_inst->drv.pEventSettingsGet = driver_data->pEventSettingsGet;
	drv_inst->drv.eventSettingSize = driver_data->eventSettingSize;
	drv_inst->drv.pEventDesc = driver_data->pEventDesc;

	drv_inst_init_data.pPerfIntCallback = hpc_perf_interrupt_handler_callback;
	drv_inst_init_data.pMallocFunc = hpc_malloc;
	drv_inst_init_data.pFreeFunc = hpc_free;

	status = drv_inst->drv.pFuncs->hpcHwDriverInit(drv_inst->drv.pDrvInst,
						&drv_inst_init_data, &size);
	if (status != OK) {
		hpc_error("Low level driver init error\n");
		return CAFE_ERROR;
	}

	sizeof_hw_data_set(size);
	max_cores_set(drv_inst->drv.numCpuCores);

	rc = hpcIterateCpus(drv_inst->drv.numCpuCores, hpc_per_core_init_iter,
	                    drv_inst);
	if (rc != OK)
		return CAFE_ERROR;

	return OK;
}

int hpc_hw_driver_register(struct hpcDriverConfig *driver_data)
{
	int n;

	for (n = 0; n < HPC_NUM_DRIVER_INSTANCES; n++) {
		if (!hw_drv[n].in_use) {
			return hpc_hw_driver_register2(driver_data, &hw_drv[n]);
		}
	}
	hpc_error("HPC_NUM_DRIVER_INSTANCES limit exceeded %d\n",
		  HPC_NUM_DRIVER_INSTANCES);
	return CAFE_ERROR;
}

static STATUS hpc_per_core_finit_iter(int cpu, void *data)
{
	STATUS status;
	struct hw_driver_inst *drv_inst = data;

	status = drv_inst->drv.pFuncs->hpcHwPerCoreFinit(drv_inst->drv.pDrvInst, 0);

	if (status != OK) {
		hpc_error("hpcHwPerCoreInit error\n");
	}
	return status;
}

int hpc_hw_driver_unregister(struct hpcDriverConfig *driver_data)
{
	STATUS hpc_rc;
	int n;
	for (n = 0; n < HPC_NUM_DRIVER_INSTANCES; n++) {
		struct hw_driver_inst *d = &hw_drv[n];
		if (d->in_use == true && d->drv.pDrvInst == driver_data->pDrvInst) {
			d->in_use = false;
			(void)hpcIterateCpus(d->drv.numCpuCores,
			                     hpc_per_core_finit_iter, d);
			hpc_rc = d->drv.pFuncs->hpcHwDriverFinit(d->drv.pDrvInst);
			if (hpc_rc != OK) {
				hpc_error("Low level driver finit error\n");
				return CAFE_ERROR;
			}
			return CAFE_OK;
		}
	}
	return CAFE_ERROR;
}

/******************************************************************************/

int hpc_init(void *(*custom_malloc)(size_t), void (*custom_free)(void *),
		    int num_vc_per_block)
{
	if (custom_malloc == NULL) {
		hpc_malloc = malloc;
		hpc_free = free;
	} else {
		hpc_malloc = custom_malloc;
		hpc_free = custom_free;
	}

	memset(hw_drv, 0, sizeof(hw_drv));
	init_ep_inst_list();
	init_vc_list_data();
	vc_blk_desc.num_vc_per_block = 0;
	sizeof_hw_data_set(0);
	max_cores_set(1);
	num_vc_per_block_set(num_vc_per_block);

	return CAFE_OK;
}

int hpc_finit(void)
{
	int rc;

	hpcChkEvtPeriodTaskDelete();

	if ((rc = finit_vc_list(false)) != CAFE_OK) {
		hpc_error("unable to finalize vc list error code %u\n", rc);
	}

	return rc;
}

static STATUS hpc_force_shutdown_iter(int cpu, void *data)
{
	STATUS status;
	struct hw_driver_inst *hw = data;

	status = hw->drv.pFuncs->hpcHwPerCoreFinit(hw->drv.pDrvInst, 1);

	if (status != OK) {
		hpc_error("hpcHwPerCoreInit error\n");
	}
	return status;
}

void hpc_force_shutdown(void)
{
	struct hw_driver_inst *hw;
	HPC_DRIVER_IDX driver;

	FOREACH_HPC_DRIVER(hw, driver) {
		(void)hpcIterateCpus(hw->drv.numCpuCores,
				     hpc_force_shutdown_iter, hw);
	}
}

/******************************************************************************/

void hpc_driver_list_show(void)
{
	struct hw_driver_inst *hw;
	HPC_DRIVER_IDX driver;

	FOREACH_HPC_DRIVER(hw, driver) {
		printf("Driver Id\t\t%d\n", driver);
		printf(" num events\t\t%d\n", hw->drv.numEvents);
		printf(" num cores\t\t%d\n", hw->drv.numCpuCores);
	}
}

static STATUS hpc_driver_show_iter(int cpu, void *data)
{
	HPC_DRIVER_IDX *driver = data;
	return hw_drv[*driver].drv.pFuncs->hpcHwDriverShow(hw_drv[*driver].drv.pDrvInst);
}

void hpc_driver_show(HPC_DRIVER_IDX driver)
{
	if (hw_drv[driver].in_use)
		(void)hpcIterateCpus(hw_drv[driver].drv.numCpuCores,
					hpc_driver_show_iter, &driver);
}

void hpc_driver_show_all(void)
{
	struct hw_driver_inst *hw;
	HPC_DRIVER_IDX driver;

	FOREACH_HPC_DRIVER(hw, driver) {
		printf("Driver Id\t\t%d\n", driver);
		hpc_driver_show(driver);
	}
}

/*****************************************************************************/

int hpc_inst_show(void)
{
	struct cafe_list_head *item;

	cafe_list_foreach(item, &ep_inst_desc.ep_inst_list_head) {
		struct hpc_inst_item *hpc_inst = CFG_LIST_ITEM(item);
		printf("hpc_inst               %p\n", hpc_inst);
		printf("ep_inst                %p\n", hpc_inst->ep_inst);
		printf("change_count           %lu\n", hpc_inst->change_count);
	}
	return CAFE_OK;
}

static void vcs_show(struct hpc_virt_slave *vcs, int cores)
{
	int c;
	struct cafe_list_head *item;
	struct cafe_value_type *vt;

	printf("    vcs_hpc_inst           %p\n", vcs->vcs_hpc_inst);
	printf("    ep inst                %p\n", vcs->vcs_hpc_inst->ep_inst);
	printf("    vcs_running            %d\n", vcs->vcs_running);
	printf("    vcs_reload             %llu\n", vcs->vcs_reload);
	printf("    vcs_vc                 %p\n", vcs->vcs_vc);
	printf("    vcs_odata_idx          %d\n", vcs->vcs_odata_idx);
	printf("    vcs_sink_act_handle    0x%lx\n", vcs->vcs_sink_act_handle);
	printf("    vcs_sec_list_head      %p %p\n", vcs->vcs_sec_list_head.next, vcs->vcs_sec_list_head.prev);
	printf("    vcs_sec_list           %p %p\n", vcs->vcs_sec_list.next, vcs->vcs_sec_list.prev);
	cafe_list_foreach(item, &vcs->vcs_sec_list_head) {
		struct hpc_virt_slave *sec_vcs = VCS_SEC_LIST_ITEM(item);
		printf("        vcs_sec                %p\n", sec_vcs);
	}
	printf("    vcs_pri                %p\n", vcs->vcs_pri);
	printf("    vcs_type->type_str     %s\n", vcs->vcs_type->type_str);
	vt = cafe_base_type(vcs->vcs_type);
	assert(vt != NULL);
	printf("        base_type          %s\n", vt->type_str);
	printf("    hpc_od_offset          %d\n", vcs->vcs_hpc_od_offset);
	for (c = 0; c < cores; c++)
		printf("    vcs_count[%d]           %llu\n", c, vcs->vcs_count[c]);
	for (c = 0; c < cores; c++)
		printf("    vcs_count_prev[%d]      %llu\n", c, vcs->vcs_count_prev[c]);
	for (c = 0; c < cores; c++)
		printf("    vcs_count_offset[%d]    %llu\n", c, vcs->vcs_count_offset[c]);
}

static void vc_show(struct hpc_virt_counter *vc)
{
	int c;
	printf("vc                     %p\n", vc);
	printf("vc_running             %d\n", vc->vc_running);
	printf("reload                 %llu\n", vc->vc_evt_set.hpc_evt_settings.reload);
	printf("samplePeriod           %llu\n", vc->vc_evt_set.hpc_evt_settings.samplePeriod);
	printf("numCpuCores            %d\n", vc->vc_evt_set.hw->drv.numCpuCores);
	for (c = 0; c < vc->vc_evt_set.hw->drv.numCpuCores; c++) {
		printf("vc_stop_count[%d]        %llu\n", c, vc->vc_stop_count[c]);
		printf("vc_reload[%d]            %llu\n", c, vc->vc_reload[c]);
		printf("vc_reload_min[%d]        %llu\n", c, vc->vc_reload_min[c]);
		printf("vc_time_prev[%d]         %llu\n", c, vc->vc_time_prev[c]);
		printf("vc_rate_prev[%d]         %llu\n", c, vc->vc_rate_prev[c]);
	}
	for (c = 0; c < NUM_VC_SLAVES; c++) {
		printf("vc_slave[%d]            %p\n", c, &vc->vc_slave[c]);
		if (vc->vc_slave[c].vcs_hpc_inst != NULL)
			vcs_show(&vc->vc_slave[c], vc->vc_evt_set.hw->drv.numCpuCores);
	}
}

int hpc_vc_show(void)
{
	struct cafe_list_head *item;
	struct hpc_virt_counter *vc;

	printf("num vc blocks          %d\n", vc_blk_desc.num_vc_blocks);
	printf("num_vc_per_block       %d\n", vc_blk_desc.num_vc_per_block);
	printf("sizeof_hw_data         %"PRIsize_t"\n", vc_blk_desc.sizeof_hw_data);
	printf("max_cores              %d\n", vc_blk_desc.max_cores);
	printf("num free vc            %d\n", get_num_vc_free());
	printf("num alloc vc           %d\n", get_num_vc_alloc());
	printf("timer_freq             %llu\n", timer_freq);
	cafe_list_foreach(item, &ep_inst_desc.vc_in_use_list_head) {
		(void)puts("--------------------------------");
		vc = VC_NON_MUXED_LIST_ITEM(item);
		vc_show(vc);
	}
	return CAFE_OK;
}

/*****************************************************************************/

struct hpc_test_funcs hpc_test = {
	init_vc_list,
	finit_vc_list,
	alloc_vc,
	free_vc,
	get_num_vc_free,
	get_num_vc_alloc,
	sizeof_hw_data_set,
	num_vc_per_block_set,
	init_ep_inst_list,
	finit_ep_inst_list,
	alloc_ep_inst,
	free_ep_inst,
	get_skipcount
};
