/* Copyright (c) 2011 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

/*

This example shows how Cafe can identify areas of code which cause a problem, in
this case, excessive cache misses.


The example consists of three functions, initialize_source(),
transform_source(), and consume_destination()

1) initialize_source()

This function uses memset() to initialize a large buffer. The result of this is
that on exit, the cache contains addresses for the later part of the buffer
(assuming memset() writes from low address to high address)


2) transform_source()

This has two variants, written to demonstrate the effect of changing the data access
pattern.


Variant 1
---------

This copies its input to a workarea, does a transformation on it, and then
copies the workarea to its output. This is deliberately non-optimal.

Following the call to initialize_source(), the beginning of the input buffer
will not be in the cache, so a lot of misses will be incurred by the copy
operation.

Once the copy has been made, the tranformation is done, starting at the
beginning of the workarea. Since the copy resulted in the cache containing
addresses from the later part of the buffer, the transformation step will also
result in cache misses.

The copy to the output buffer results in yet more cache misses, for the same
reason.

Variant 2
---------

The function is modified so that the same process is followed, but the work is
broken into smaller chunks, which will fit in the cache.

Thus cache misses will occur when copying data into the workarea, and out of the
workarea, but when the transformation is done, fewer misses will occur.



3) consume_destination()

This uses memcpy() again to extract the result from the destination




Suggested workflow is as follows:

In the UI, select a total data cache miss count metric, with a function
context. This will display a count of cache misses,  with the function in which
they occur.


Use the UI 'clear' button to reset the data.

Run 'cafe_cache_demo 1', which is the baseline.

In the UI, sort on the cache miss count column, and you can see that the
majority of cache misses occur in memcpy(), called from two places in
transform_source().

This is because memcpy() is used to copy the large blocks of data in and out. A
more cache-friendly method is to break the data into smaller pieces: This is
enabled by setting variant=2 when calling the demo code.

To see this, reset the UI with the 'clear' button, and run 'cafe_cache_demo 2'

The results show that memcpy() still generates most cache-misses, but the count
is much lower.

In addition, the test case runs significantly faster.

*/

#define DEFAULT_NUM_WORK_ITEMS  (4*1024*1024)
#define NUM_RESULT_ITEMS	128

#define CHUNK_SIZE		64

#define DEFAULT_LOOPCOUNT	100
#define DEFAULT_SIZE		128
#define DEFAULT_TYPE		1

int initialize_source(int *src, size_t size);
int transform_source(int variant, int *src, int *workarea, int *dest, size_t size);
int consume_destination(int *dest, size_t size);

int initialize_source(int *src, size_t size)
{
	memset(src, 0x55, size * sizeof(*src));
	return 0;
}

int transform_source(int variant, int *src, int *workarea, int *dest, size_t size)
{
	size_t i;
	size_t offset;

	if (variant == 1) {
		memcpy (workarea, src, sizeof(*src) * size);

		for (i = 0; i < size; i++) {
			workarea[i] = workarea[i]+1;
		}

		memcpy (dest, workarea, sizeof(*src) * size);
	} else {
		for (offset = 0; offset < size; offset += CHUNK_SIZE) {

			memcpy (workarea+offset, src+offset, sizeof(*src) * CHUNK_SIZE);

			for (i = 0; i < CHUNK_SIZE; i++) {
				workarea[offset+i] = workarea[offset+i]+1;
			}

			memcpy (dest+offset, workarea+offset, sizeof(*dest) * CHUNK_SIZE);
		}
	}
	return 0;
}

int consume_destination(int *dest, size_t size)
{
	size_t i;
	int result[NUM_RESULT_ITEMS] = {0};
	for (i = 0; i < size; i+=NUM_RESULT_ITEMS) {
		memcpy (dest+i, result, sizeof(*dest) * NUM_RESULT_ITEMS);
	}
	return 0;
}

static
int do_cafe_memcpy_demo(long variant, long loops, long num_work_items)
{
	int *src = NULL;
	int *workarea = NULL;
	int *dest = NULL;
	int i;

	src = malloc(num_work_items * sizeof(*src));
	if (src == NULL)
		goto cleanup;

	dest = malloc(num_work_items * sizeof(*dest));
	if (dest == NULL)
		goto cleanup;

	workarea = malloc(num_work_items * sizeof(*workarea));
	if (workarea == NULL)
		goto cleanup;

	for (i = 0; i < loops || loops < 0; i++) {
		initialize_source(src, num_work_items);

		transform_source(variant, src, workarea, dest, num_work_items);

		consume_destination(dest, num_work_items);
	}

 cleanup:
	if (src != NULL)
		free(src);
	if (workarea != NULL)
		free(workarea);
	if (dest != NULL)
		free(dest);

	return 0;
}

static
void usage(void)
{
	printf("Usage: cafe-cache-demo [-t <type>] [-n loopcount] [-s size]\n"
	       "       type 1 - copy whole data set in one call to memcpy (default)\n"
	       "       type 2 - copy data in small stripes which fit in cache\n"
	       "\n"
	       "       size - number of integers in data (default %d)\n"
	       "\n"
	       "       loopcount - number of iterations: A negative value will cause the\n"
	       "                   test to run indefinitely (default %d)\n"
	       "\n"
	       "       To demonstrate different cache behaviour, count cache misses first\n"
	       "       for 'cafe-cache-demo -t 1' and then for 'cafe-cache-demo -t 2'\n",
	       DEFAULT_NUM_WORK_ITEMS, DEFAULT_LOOPCOUNT);
	exit(EXIT_SUCCESS);
}

int main(int argc, char *argv[])
{
	int c;
	char *endptr;
	long loopcount = DEFAULT_LOOPCOUNT;
	long size = DEFAULT_NUM_WORK_ITEMS;
	long type = DEFAULT_TYPE;

	while ((c = getopt(argc, argv, "hn:t:s:")) != -1) {
		switch (c) {
		case 'h':
			usage();
			break;
		case 'n':
			loopcount = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			break;
		case 's':
			size = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			break;
		case 't':
			type = strtol(optarg, &endptr, 0);
			if (endptr == optarg)
				usage();
			if (type != 1 && type != 2)
				usage();
			break;
		default:
			usage();
		}
	}
	if (optind != argc)
		usage();

	do_cafe_memcpy_demo(type, loopcount, size);
	return EXIT_SUCCESS;
}
