/* kernel_map_parser.c -  */

/* Copyright (c) 2003-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */

/*
  Garner kernel symbol values from System.map and the /proc filesystem.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <assert.h>
#include "../include/kernel_map_parser.h"
#include <cafe/cafe.h>


/* Location of System.map; the file which lists kernel symbols */
// TD:  add locations such as /boot/System.map-2.4.20 and
//      /lib/modules/2.4.20/System.map which incorporate kernel versions
static const char * const CommonMapfileLocs[] = {
	"/boot/System.map",
	"/System.map",
	"/usr/src/linux/System.map"
};


/**
 *  Locate the address of a kernel symbol.
 * symbolname - name of the symbol to look for.  Must be a valid null-
 *              terminated string.
 * mapFilename - pointer to a string which holds a suggested location of
 *               the system map.  If the mapFilename is 0, or the specified
 *               file isn't found, default locations are searched. mapFilename
 *               is set to point at the full filename of the actual map
 *               used, or zero if no maps were found.
 * returns true if address now corresponds to the given symbol
 * within the kernel, and false otherwise.
 */
// TD: update the interface to allow searching for multiple symbols at
//     once, that way we don't have to keep searching for, opening, and
//     iterating through the symbol map
bool kernel_symbol_find(const char *const symbolname, uintptr_t *address,
			size_t *size, const char **const map_filename)
{
	FILE *map_file = NULL;
	char type;
	char symbol[KSYM_NAME_LEN];
	unsigned int i = 0;
	bool found = false;

	assert(symbolname != NULL);
	assert(address != NULL);
	assert(size != NULL);
	/* if we were passed an explicit location, check that first */
	if (map_filename != NULL) {
		map_file = fopen(*map_filename, "r");
	}

	/* now search default locations */
	if (map_file == NULL) {
		for (i = 0; i < CAFE_NELEMENTS(CommonMapfileLocs); ++i) {
			// TD: also check <default location>-<uname -r>
			map_file = fopen(CommonMapfileLocs[i], "r");
			if (map_file != NULL) {
				if (map_filename != NULL) {
					*map_filename = CommonMapfileLocs[i];
				}
				break;
			}
		}
	}

	if (i == CAFE_NELEMENTS(CommonMapfileLocs)) {
		perror("Couldn't open any default System.map files");
		if (map_filename != NULL) {
			*map_filename = NULL;
		}
		return false;
	}

	while (fscanf(map_file, "%"SCNxPTR" %c %s", address, &type, symbol) == 3) {
		uintptr_t next_addr;

		if (strcmp(symbol, symbolname) == 0) {
			found = true;

			if (size) {
				if (fscanf(map_file, "%"SCNxPTR" %c %s", &next_addr, &type, symbol)
				    == 3) {
					*size = next_addr - *address;
				} else {
					*size = -1;
				}
			}
			break;
		}
	}

	if (fclose(map_file) != 0) {
		perror("Failed to close map file");
	}

	return found;
}


#if defined(CAFE_LINUX24)
/* 2.4 kernels use ksyms */
static const char * const ksym_file = "/proc/ksyms";
#else
/* 2.6 kernels use kallsyms */
static const char * const ksym_file = "/proc/kallsyms";
#endif


/**
 *  Remove the checksum wart from any kernel symbol string.  The pointer passed
 * must point at a valid null-terminated string.
 */
void kernel_symbol_strip_checksum(char *sym)
{
	char *ptr = rindex(sym, '_');

	if ( (ptr != NULL)
	     && ( (ptr + 2) < (sym + strlen(sym)) )  /* ensure we don't read off
							the end of the string */
	     && ('R' == *(ptr+1)) && (isxdigit(*(ptr+2))) )
	{
		*ptr = 0;
	}
}


/**
 *  This function reads a single entry in the public symbol map exported
 * via the procfs.
 *
 *   entry_num - the line number corresponding to the symbol requested.  Note
 *               that this function currently skips entries associated with
 *               modules.
 *   symbol - a pointer to space allocated for a ksym_entry_t.  This structure
 *            will be populated with the symbol data found, or left untouched
 *            if parsing fails.
 *
 * Returns - true iff symbol has been populated with the requested info.
 */
bool kernel_symbol_read_exported(const unsigned int entry_num,
				 ksym_entry_t *symbol)
{
	unsigned int line;
	FILE *map = fopen(ksym_file, "r");
	int c;

	assert(symbol != NULL);
	if (map == NULL) {
		printf("Error opening \"%s\": %s\n", ksym_file, strerror(errno));
		return false;
	}

	/* locate the line we were asked for */
	line = 0;
	/* 2.6 kernels have an entry for the symbol type, 2.4 kernels don't */
#if defined(CAFE_LINUX24)
	while (2 == fscanf(map, "%"SCNxPTR" %81s", &symbol->address, symbol->name)) {
#else
	while (3 == fscanf(map, "%"SCNxPTR" %c %81s",
			   &symbol->address, &symbol->type,
			   symbol->name))
	{
#endif

		/* if the symbol name is longer than the statically sized
		   buffers we use for the scanf calls, we need to advance the
		   stream pointer past the last character of the symbol name */
		do {
			c = fgetc_unlocked(map);
		} while ( (c != EOF) && (!isspace(c)) );

		/* skip entries for module symbols */
		if ('\t' == c) {
			/* read until the end of line */
			do {
				c = fgetc_unlocked(map);
			} while ( (c != EOF) && (c != '\n') );
			continue;
		}

		if (entry_num == line) {
			fclose(map);
			kernel_symbol_strip_checksum(symbol->name);
			return true;
		}

		++line;
	}

	fclose(map);
	return false;
}


/**
 * A qsort() callback used to compare two ksym_entry_t structures.
 */
static
int ksym_compare(const void *a1, const void *a2)
{
	const ksym_entry_t *k1 = a1;
	const ksym_entry_t *k2 = a2;

	if (k1->address < k2->address)
		return -1;
	if (k1->address > k2->address)
		return 1;
	return 0;
}


/**
 *  Reads kernel symbols exported via the proc filesystem, and returns them to
 * the caller.  This function allocates an array of symbol entries and returns
 * them in the "symbols" parameter.  This parameter must point to a valid
 * "ksym_entry_t *" when this function is called.  Further, the caller is
 * responsible for free()ing the returned pointer.
 *
 * Returns the number of symbols read, or -1 on error.
 */
int kernel_symbols_read(ksym_entry_t **symbols)
{
	char line_buf[2 * KSYM_NAME_LEN];
	unsigned int num_syms;
	unsigned int i;
	FILE *map_file = fopen(ksym_file, "r");

	assert(symbols != NULL);
	if (map_file == NULL) {
		printf("Error opening \"%s\": %s\n", ksym_file, strerror(errno));
		return -1;
	}

	/* first, count the number of symbols, so we know how much to
	   allocate */
	num_syms = 0;
	while(fgets_unlocked(line_buf, sizeof(line_buf), map_file) != 0) {
		++num_syms;
	}

	*symbols = (ksym_entry_t *) malloc(num_syms * sizeof(ksym_entry_t));

	if (*symbols == NULL) {
		printf("Error allocating buffer for reading kernel symbols (%zu bytes)\n", num_syms * sizeof(ksym_entry_t));
		fclose(map_file);
		return -1;
	}

	/* now read in the symbols */
	rewind(map_file);
	i = 0;
#if defined(CAFE_LINUX24)
	while (2 == fscanf(map_file, "%"SCNxPTR" %127s", 
			   &(*symbols)[i].address, (*symbols)[i].name))
#else
	while (3 == fscanf(map_file, "%"SCNxPTR" %c %127s", 
			   &(*symbols)[i].address, &(*symbols)[i].type,
			   (*symbols)[i].name))
#endif
	{
		if ('\t' == fgetc_unlocked(map_file)) {
			if (1 != fscanf(map_file, "%59s", (*symbols)[i].module))
				return -1;
		} else {
			(*symbols)[i].module[0] = 0;
		}

		++i;
	}

	if (fclose(map_file) != 0) {
		printf("Failed to close \"%s\": %s", ksym_file, strerror(errno));
	}

	qsort(*symbols, i, sizeof(**symbols), ksym_compare);
	return i;
}
