/**
 *  Utility functions for querying linux process memory maps.  Currently this
 * data is only captured from the /proc/<pid>/maps files.
 */

/* Copyright (c) 2003-2012 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */


#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "../include/memory_map.h"


/* states used by the custom memory region parser in getNextMemoryRegion() */
typedef enum {
	START_ADDR,
	END_ADDR,
	PERM,
	OFFSET,
	FILENAME_START,
	FILENAME_END
} mapstate_t;

/* use the fact the a > A > 0 to convert from hex->dec */
#define hex2dec(_c)							\
	(((_c)>='a')?((_c)-'a'+0xa):((_c)>='A')?((_c)-'A'+0xA):((_c)-'0'))

#define REGIONS_CONCATENATABLE(r1, r2)			\
	(r1.end == r2.start)				\
	&& (r2.file_offset == r1.end - r1.start)	\
	&& (r1.executable == r2.executable)		\
	&& (0 == strcmp(r1.name, r2.name))

#define MAX_MAP_LINE_LENGTH        (512)


void memory_region_print(FILE *out, const struct memory_region *const reg)
{
	fprintf(out, "0x%08" PRIxPTR " - 0x%08" PRIxPTR
		"  %d  0x%08" PRIxPTR "    %s\n",
		reg->start, reg->end, reg->executable,
		reg->file_offset, reg->name);
}

#if defined(__UCLIBC__)
static
int isXdigit(int x)
{
	// quick isxdigit for the moment. blech
	if ((x >= 48) && (x <= 57))
		return 1;
	if ((x >= 65) && (x <= 70))
		return 1;
	if ((x >= 97) && (x <= 102))
		return 1;
	return 0;
}
#else
#define isXdigit(x)  isxdigit(x)
#endif


void memory_map_print(FILE *out, const struct memory_map *const map)
{
	unsigned int i;

	fprintf(out, "MAP:\n");
	for (i = 0; i < map->region_count; ++i) {
		fprintf(out, "%u:\t", i);
		memory_region_print(out, &map->regions[i]);
	}

	fprintf(out, "\n\n");
}


/**
 *  Populates the given MemoryRegion object with the specification
 * beginning at the current read pointer of the given FILE pointer.
 */
bool memory_region_get_next(FILE *f, struct memory_region *r)
{
	char *cptr;
	mapstate_t state = START_ADDR;
	unsigned int len = 0;
	bool done = false;
	char line_buf[MAX_MAP_LINE_LENGTH];

	r->start = 0;
	r->end = 0;
	r->file_offset = 0;
	r->executable = false;
	r->name[0] = 0;

	cptr = fgets_unlocked(line_buf, sizeof(line_buf), f);
	if (cptr == NULL)
		return false;

	while (!done && *cptr) {
		switch (state) {
		case START_ADDR:
			if (isXdigit(*cptr)) {
				r->start = (r->start * 16) + hex2dec(*cptr);
			} else if (*cptr == '-') {
				state = END_ADDR;
			} else {
				return false;
			}
			break;
		case END_ADDR:
			if (isXdigit(*cptr)) {
				r->end = (r->end*16) + hex2dec(*cptr);
			} else if (*cptr == ' ') {
				state = PERM;
			} else {
				return false;
			}
			break;
		case PERM:
			if (*cptr == 'x') {
				r->executable = true;
			} else if (*cptr == ' ') {
				state = OFFSET;
			}
			break;
		case OFFSET:
			if (isXdigit(*cptr)) {
				r->file_offset = (r->file_offset * 16)
					+ hex2dec(*cptr);
			} else if (*cptr == ' ') {
				state = FILENAME_START;
			} else {
				return false;
			}
			break;
		case FILENAME_START:
			if ((*cptr == '/') || (*cptr == '[')){
				r->name[len] = *cptr;
				++len;
				state = FILENAME_END;
			} else if (*cptr == '\n')
				done = true;
			break;
		case FILENAME_END:
			if (*cptr == '\n') {
				done = true;
			} else {
				r->name[len] = *cptr;
				++len;
			}
			break;
		default:
			return false;
		}
		++cptr;
	}

	r->name[len] = 0;
	return true;
}


/**
 *  Populate the given MemoryMap with data corresponding to the current
 * process.  If it was previously populated, the MemoryMap object passed
 * should be deleted with MemoryMapDestroy() before calling this function to
 * avoid memory leaks.
 */
bool memory_map_get_self(struct memory_map * const map)
{
	unsigned int i;
	struct memory_region tmp;
	FILE *f = fopen("/proc/self/maps", "r");

	if (f == 0) {
		perror("Couldn't open maps file");
		return false;
	}

	map->region_count = 0;

	/* count the total regions */
	while (memory_region_get_next(f, &tmp)) {
		++(map->region_count);
	}

	map->regions = (struct memory_region *) malloc(map->region_count * sizeof(struct memory_region));

	rewind(f);

	/* now read the regions into our map */
	for (i = 0; i < map->region_count;) {
		if (!memory_region_get_next(f, &tmp)) {
			/* Means either we successfully concatenated some
			   regions, or less likely, the memory map has shrunk
			   since we counted the regions */
			map->region_count = i;
			break;
		}

		/* see if we can concatenate with a previous entry */
		if ( (i > 0) 
		     && (REGIONS_CONCATENATABLE(map->regions[i - 1], tmp)) )
		{
			map->regions[i - 1].end = tmp.end;
		} else {
			map->regions[i].start = tmp.start;
			map->regions[i].end = tmp.end;
			map->regions[i].file_offset = tmp.file_offset;
			map->regions[i].executable = tmp.executable;
			strncpy(map->regions[i].name, tmp.name, MAX_MAP_NAME_LENGTH);
			++i;
		}
	} /* for 0 to region_count */

	fclose(f);
	return true;
}


/**
 *  Discover which region within a given memory map contains the requested
 * address.  Returns a pointer to the appropriate region object if found,
 * and 0 if the address doesn not belong to a region in this map.
 */
struct memory_region *memory_map_find_address(
	const struct memory_map *const map,
	const unsigned long address)
{
    /* note that we should use signed instead of unsigned values here.
       The high variable in particular can go to -1 if the address
       to be located is in a lower address than the start of region[0]. */
    int mid;
    int low = 0;
    int high = map->region_count - 1;

    while (low <= high) {
        mid = low + ((high - low) >> 1);
        if (address >= map->regions[mid].start) {
            if (address <= map->regions[mid].end) {
                return &map->regions[mid];
            }

            low = mid + 1;
        } else {
            high = mid - 1;
        }
    }

    return 0;
}


/**
 *  Free all memory associated with a MemoryMap object.
 */
void memory_map_destroy(struct memory_map *map)
{
	map->region_count = 0;
	free(map->regions);
	map->regions = 0;
}


unsigned int memory_map_count_regions(FILE *f)
{
	char buf[MAX_MAP_LINE_LENGTH];
	unsigned int no = 0;

	rewind(f);
	while (fgets_unlocked(buf, sizeof(buf), f) != 0) {
		++no;
	}
	rewind(f);
	return no;
}
