/* wvr_converter.c - Converts a .wvr event stream into a cafe event stream */

/* Copyright (c) 2010-2012, 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software
 * may be licensed only pursuant to the terms of an applicable Wind River
 * license agreement.
 */


/*
modification history
--------------------
24feb14,sdt  Disable unexpected NO_EFFECT warnings

DESCRIPTION

Converts a WindView .wvr event stream into a phoenix leb128-encoded event
stream.
*/

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#define HAVE_STDINT_H  /*for wvr-types.h */
#include <wvr.h>
#include <wvr-bytebuf.h>
#include <wvr-event.h>
#include <stdio.h>
#include <assert.h>
#include <cafe/stdbool.h>

#include <cafe/cafe.h>
#include <cafe/event_map.h>
#include <cafe/log_stream.h>

#include "../cafe/ep/vxworks_events.h"
#include "logTimestamp.h"


static unsigned int cpuIdCurrent = 0;
static bool verbose = false;


/**
 * event_struct_size - On successful completion, <result_size>
 * will contain the size of the structure
 */
static
bool event_struct_size(struct wvr_evt *evt, int *result_size)
{
	int param_count;
	int param;

	if (result_size == NULL || evt == NULL)
		return false;

	param_count = wvr_evt_param_count(evt);

	/*
	 * calculate size of parameter array - currently, all scalar params are
	 * 32bit int
	 */

	*result_size = 0;

	for (param = 0; param < param_count; ++param) {

		switch (wvr_evt_param_type(evt, param)) {
		case VAL_INT8:
		case VAL_INT16:
		case VAL_INT32:
			*result_size += 4;
			break;
		case VAL_BLOB:
			/* blobs are length/data in the structure */
			*result_size += 4 * 2;
			break;

		case VAL_TEXT:
			*result_size += 4;
			break;

		default:
			printf("Unknown WVR parameter type %d\n",
			       wvr_evt_param_type(evt, param));
			/* coverity[array_null] */
			assert(!"unexpected type");
		}
	}
	return true;
}


/**
 * event_struct_populate - On successful completion, <result_size>
 * will contain the size of the structure
 */
static
bool event_struct_populate(struct wvr_evt *evt, uint32_t *event_data)
{
	unsigned int param_count;
	unsigned int param;

	if (evt == NULL)
		return false;

	param_count = wvr_evt_param_count(evt);

	/*
	 * currently, all scalar params are
	 * 32bit int
	 */

	for (param = 0; param < param_count; ++param) {

		switch (wvr_evt_param_type(evt, param)) {
		case VAL_INT8:
		case VAL_INT16:
		case VAL_INT32:
			*event_data = wvr_evt_param_int32(evt, param);
			break;

		case VAL_BLOB:
			*event_data = wvr_evt_param_bytes(evt, param);
			++event_data;
#if __WORDSIZE == 64
			assert("WRS_FIXME!");
#else
			*event_data = (uint32_t)wvr_evt_param_blob(evt, param);
#endif			
			break;

		case VAL_TEXT:
			/* for strings we assume the length was the
			 * previous param */
#if __WORDSIZE == 64
			assert("WRS_FIXME!");
#else
			*event_data = (uint32_t)wvr_evt_param_blob(evt, param);
#endif
			break;

		default:
			/* coverity[array_null] */
			assert(!"unexpected type");
		}
		++event_data;
	}
	return true;
}


/**
 *  Given a wvr event, return a string which represents its cafe valuetype.
 * The result needs to be free()ed.
 */
//td: this could certainly be more robust and performant
static
char *wvr_params_to_valuetype(struct wvr_evt *event)
{
	unsigned int i;
	char buf[4096];
	struct wvr_evt_desc *desc = wvr_evt_event_desc(event);

	if (desc->range_end == -1)
		sprintf(buf, "S[%s", desc->name);
	else
		/* return different event names for WV events which can span
		   multiple event ids */
		sprintf(buf, "S[%s_%d", desc->name, wvr_evt_event_type(event));

	if (desc->param_count > 0)
		strcat(buf, ",");

	for (i = 0; i < desc->param_count; ++i) {
		strcat(buf, "v[");
		strcat(buf, desc->param_names[i]);
		strcat(buf, ",");
		switch(wvr_evt_param_type(event, i)) {
		case VAL_INT8:
		case VAL_INT16:
		case VAL_INT32:
			strcat(buf, "i");
			break;
		case VAL_INT64:
			strcat(buf, "L");
			break;
		case VAL_BLOB:
			strcat(buf, "b");
			break;
		case VAL_TEXT:
			strcat(buf, "s");
			break;
		}
		strcat(buf, "]");
	}
	strcat(buf, "]");
	return strdup(buf);
}


/**
 *
 */
static
bool convert_wvr_file(const char *filename, const char *out_filename)
{
	int res;
	unsigned int eventCount = 0;
	struct wvr_file *wvrfile;
	struct wvr_bytebuf *buf;
	struct wvr_evt *event;
	int event_size = 0;
	struct cafe_event_map *event_map;
	void *output_event_data[2];
	unsigned long long ticks;
	uint32_t event_data[16];
	FILE *out;
	struct cafe_instrument *session;
	struct cafe_output_stream *ostream;
	struct cafe_slog_encoded_inst *slog;

	res = cafe_lib_init();
	if (res != CAFE_OK) {
		fprintf(stderr, "Error initializing cafe library: %s\n",
			cafe_strerror(res));
		return false;
	}

	CAFE_LOCK();
	res = cafe_instrument_create("convert_wvr_file", &session);
	if (res != CAFE_OK) {
		printf("error: cannot create session");
		return false;
	}

	res = wvr_file_open(filename, &wvrfile);
	if (res != WVR_OK) {
		printf("Failed to open file: %s\n", wvr_errmsg(res));
		return false;
	}

	out = fopen(out_filename, "w");
	if (!out) {
		wvr_file_close(wvrfile);
		perror("error opening output file");
		return false;
	}

	res = cafe_filestream_open_output(out, cafe_stdlib_allocator, &ostream);
	if (res != CAFE_OK) {
		wvr_file_close(wvrfile);
		printf("error creating file stream: %s\n", cafe_strerror(res));
		return false;
	}

	res = cafe_slog_encoded_create(session, ostream, &slog);
	if (res != CAFE_OK) {
		wvr_file_close(wvrfile);
		printf("error creating file encoder: %s\n", cafe_strerror(res));
		CAFE_FREE(cafe_stdlib_allocator, ostream);
		return false;
	}

	if ((buf = wvr_bytebuf_new(8)) == NULL) {
		wvr_file_close(wvrfile);
		perror("error: cannot create new byte buffer");
		return false;
	}

	if ((wvr_evt_new(&event)) != WVR_OK) {
		wvr_bytebuf_del(buf);
		wvr_file_close(wvrfile);
		perror("error: cannot create new wvr event");
		return false;
	}

	event_map = cafe_event_map_create(session);

	output_event_data[0] = event_data;
	output_event_data[1] = &ticks;
	while (1) {
		int evType;
		char *taskName = NULL;
		const char *taskNameData = NULL;
		struct cafe_remote_event *remote;

		wvr_evt_reset(event);

		/* read an event */
		res = wvr_file_read_event(wvrfile, event);
		if (res != WVR_DECODED_EVENT)
			break;

		/* print it for info */
		if (verbose) {
			wvr_evt_to_string(event, buf);
			printf("#%u: %s\n", eventCount, wvr_bytebuf_str(buf));
		}

		/* calculate the size required for the event params */
		event_struct_size(event, &event_size);

		if (event_size > sizeof(event_data)) {
			printf("Event too big!\n");
			continue;
		}

		/* build the event structure */
		event_struct_populate(event, event_data);

		evType = wvr_evt_event_type(event);
		if (evType == 0)
			evType = -1;

		remote = cafe_event_map_lookup(event_map, evType);
		if (!remote) {
			struct cafe_metadata md;
			const char *event_vt[2];
			int rc;

			event_vt[0] = wvr_params_to_valuetype(event);
			md.eid = evType;
			md.event_data.l = 1;
			md.event_data.p = event_vt;
			if (wvr_evt_event_desc(event)->has_timestamp) {
				++md.event_data.l;
				event_vt[1] = cafe_timestamp_VT_DEF;
			}

			rc = cafe_event_map_add_event(event_map, &md, NULL,
						      &remote);
			assert(rc == CAFE_OK);
			if (rc != CAFE_OK) {
				fprintf(stderr, "error: cannot add event: %s\n",
					cafe_strerror(rc));
				exit(EXIT_FAILURE);
			}
			
			free((char *)event_vt[0]);
		}
		switch (evType) {
		case EVENT_TIMESTAMP_CONFIG:
		{
			/* initialize the synthesized timestamp */
			logTimestampConfigSet
			    (wvr_evt_param_int32(event, 0),
			     wvr_evt_param_int32(event, 1));
			break;
		}
		case EVENT_CPU_ID:
			cpuIdCurrent = wvr_evt_param_int32(event, 0);
			break;
		}

		/*
		 * This is to allow us to read logs from vxWorks <
		 * 6.7, but pass them into top, which is build for 6.7
		 * only. If you didn't want to do this, we could have
		 * top-6.7, top-6.6, top-6.0 etc.
		 */

		if (evType == EVENT_TASKNAME) {
			unsigned int taskNameLength;
			switch (wvr_file_log_revision(wvrfile)) {
			case 0xB0B00001:	/* vxworks54 */
			case 0xB0B00002:	/* vxworks54 */
			case 0xB0B00005:	/* vxworks55 */
				taskNameLength = wvr_evt_param_bytes(event, 5);
				taskNameData = wvr_evt_param_text(event, 5);
				break;
			case 0xB1B10001:	/* vxworks60 */
			case 0xB1B10002:	/* vxworks63 */
			case 0xB1B50003:	/* vxworks67 */
			case 0xB1B40003:	/* vxworks67-smp */
				taskNameLength = wvr_evt_param_bytes(event, 7);
				taskNameData = wvr_evt_param_text(event, 7);
				break;
			case 0xB1B30003:	/* vxworks66 */
			case 0xB1B20003:	/* vxworks66-smp */
				taskNameLength = wvr_evt_param_bytes(event, 8);
				taskNameData = wvr_evt_param_text(event, 8);
				break;
			default:
				taskNameData = NULL;
				taskNameLength = 0;
				printf("Unknown log revision when reading "
				       "taskname\n");
			}

			if (taskNameData != NULL && taskNameLength != 0) {

				/* WRS_FIXME: there's no strndup on windows/solaris. */
				taskName = malloc(taskNameLength +1);
				
				if (!taskName) {
					perror("malloc");
					exit(EXIT_FAILURE);
				}

				memcpy(taskName, taskNameData, taskNameLength);
				taskName[taskNameLength] = '\0';

				/* this assumes top is built for vx-6.7 */
				/* WRS_XXX 64-bit issues here! */
				assert(sizeof(taskName) <=
				       sizeof(event_data[7]));
				event_data[7] = (unsigned int)taskNameLength;
#if __WORDSIZE == 64
				assert("WRS_FIXME!");
#else
				event_data[8] = (unsigned int)taskNameData;
#endif				
			}
		}

		/* get the event timestamp and update the synthesized one */
		ticks = wvr_evt_ticks(event);
		if (ticks != 0)
			logTimestampUpdate(ticks);

		/* log the event */
		cafe_slog_encoded_write(slog, remote->event,
				  output_event_data);

		if (taskName != NULL) {
			free(taskName);
			taskName = NULL;
		}
		++eventCount;
	}
	cafe_slog_encoded_delete(slog);
	fclose(out);
	cafe_event_map_destroy(event_map);
	res = cafe_instrument_delete(session);
	CAFE_UNLOCK();
	if (res != CAFE_OK)
		fprintf(stderr, "Error deleting instrument: %s\n",
			cafe_strerror(res));

	wvr_evt_del(event);
	wvr_bytebuf_del(buf);
	wvr_file_close(wvrfile);

	if (verbose)
		printf("%u events\n", eventCount);
	return true;
}


static
void usage(const char *name)
{
	fprintf(stderr, "Usage: %s [-v] <file.wvr> <output file>\n", name);
	exit(EXIT_FAILURE);
}


static
int parse_opts(int argc, char *argv[])
{
	int opt;
	int ind = 1;

	while (ind < argc) {
		char *s = argv[ind];
		if (*s++ != '-')
			break;
		ind++;
		if (s[0] == '-' && s[1] == '\0')
			break;
		while ((opt = *s++) != '\0') {
			switch (opt) {
			case 'v':
				verbose = 1;
				break;

			default:
				usage(argv[0]);
			}
		}
	}

	if (ind >= argc)
		usage(argv[0]);

	return ind;
}


int main(int argc, char *argv[])
{
	const int filename_index = parse_opts(argc, argv);

	if (argc < filename_index + 2)
		usage(argv[0]);

	convert_wvr_file(argv[filename_index], argv[filename_index + 1]);

	return EXIT_SUCCESS;
}
