/*
 * Copyright (c) 2013-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

#if !defined(_WRS_KERNEL)

#include "tools_common.h"
#include <assert.h>
#include <tcf/framework/channel.h>
#include <tcf/framework/proxy.h>
#include <tcf/framework/myalloc.h>
#include <tcf/framework/trace.h>
#include <tcf/framework/events.h>
#include <tcf/framework/protocol.h>

#ifdef ENABLE_WR_FILTERING_OPT
#include <tcf/main/logfilter.h>
#endif

#ifndef WIN32
#include <pthread.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#else
#include <windows.h>
#include <signal.h>
#endif

#ifdef ENABLE_WR_REDIRECT_OPT
static char *redirect_url = NULL;
#endif

#ifdef BUILD_VERSION_STRING
static void print_version() {
    fprintf(stdout, "%s\n", buildString);
    exit(0);
}
#endif

#ifdef ENABLE_WR_MONITOR_PID
#define MONITORED_DIED "Monitored process died. Suicide ourself !\n"
static unsigned int watch_pid;

#ifdef WIN32
static DWORD WINAPI monitor_pid ( LPVOID lpParam ) {
	HANDLE hdl = OpenProcess (SYNCHRONIZE, FALSE, watch_pid);
	WaitForSingleObject(hdl, INFINITE);
	fprintf (stderr, MONITORED_DIED);
	raise(SIGTERM);
	return 0;
}

 void monitor_pid_start (unsigned int pid) {
        watch_pid = pid;
	CreateThread (NULL, 0, monitor_pid, NULL, 0, NULL);
}

#else
static void * monitor_pid ( void *ptr ) {
	while (1) {
	    struct timeval delay;
	    delay.tv_sec = 0;
	    delay.tv_usec = 200 * 1000;

		if (kill (watch_pid, 0)) {
		/* Do a proper kill */
		fprintf (stderr, MONITORED_DIED);
		raise(SIGTERM);
		}
		select(0, NULL, NULL, NULL, &delay);
	}
	return NULL;
}

void monitor_pid_start (unsigned int pid) {
	pthread_t thread_monitor;
        watch_pid = pid;
	pthread_create (&thread_monitor, NULL, monitor_pid, NULL);
}
#endif /*WIN32 */

#endif /* ENABLE_WR_MONITOR_PID */

#ifdef ENABLE_WR_REDIRECT_OPT
typedef struct ConnectInfo {
    PeerServer * ps;
    Channel * c1;
} ConnectInfo;

static void connect_done(void * args, int error, Channel * c2) {
    ConnectInfo * info = (ConnectInfo *)args;
    Channel * c1 = info->c1;

    if (!is_channel_closed(c1)) {
        assert(c1->state == ChannelStateRedirectReceived);
        if (error) {
            fprintf(stderr, "cannot connect to peer: %s\n", redirect_url);
            channel_close(c1);
        }
        else {
            proxy_create(c1, c2);
        }
    }
    else if (!error) {
        channel_close(c2);
    }
    channel_unlock(c1);
    peer_server_free(info->ps);
    loc_free(info);
}

static void connect_dest(void * x) {
    Channel * c1 = (Channel *)x;
    PeerServer * ps;
    ConnectInfo * info;

    ps = channel_peer_from_url(redirect_url);
    if (ps == NULL) {
        trace(LOG_ALWAYS, "cannot parse peer url: %s", redirect_url);
        channel_close(c1);
        return;
    }
    channel_lock(c1);
    c1->state = ChannelStateRedirectReceived;
    info = (ConnectInfo *)loc_alloc_zero(sizeof(ConnectInfo));
    info->ps = ps;
    info->c1 = c1;
    channel_connect(ps, connect_done, info);
}

static void channel_server_connecting(Channel * c1) {
    assert(c1->state == ChannelStateStarted);
    /* Just fake that we sent hello message. */
    c1->state = ChannelStateHelloSent;
}

static void channel_server_connected(Channel * c1) {
    assert(c1->state == ChannelStateConnected);

    /* Connect to destination on next dispatch since we are limited in
     * what we can do in a callback, e.g. cannot close channel.
     */
    post_event(connect_dest, c1);
}

static void redirect_listener(Channel *c) {
    if (c->incoming) {
    /* Avoids "re-entrance" due to redirection.
     * Only do that on  incoming connect */
    c->connecting = channel_server_connecting;
    c->connected = channel_server_connected;
    }
}
#endif

int parse_args_hook_tools (char c, char **s, int *ind, char **argv, int argc, const char *progname) {
#ifdef BUILD_VERSION_STRING
    if (c == 'v') print_version();
#endif

#ifdef ENABLE_WR_MONITOR_PID
    if (c == 'w') {
        if (**s == '\0') {
        	*ind += 1;
            if (*ind >= argc) {
                fprintf(stderr, "%s: error: no argument given to option '%c'\n", progname, c);
                exit(1);
            }
            *s = argv[*ind];
        }
        monitor_pid_start(strtol(*s, 0, 0));
        *s = NULL;
        return 1;
    }
#endif

#ifdef ENABLE_WR_REDIRECT_OPT
    if (c == 'r') {
        if (**s == '\0') {
                *ind += 1;
            if (*ind >= argc) {
                fprintf(stderr, "%s: error: no argument given to option '%c'\n", progname, c);
                exit(1);
            }
            *s = argv[*ind];
        }
        redirect_url = *s;
        *s = NULL;
        /* Redirect when the channel will be created */
        add_channel_create_listener(redirect_listener);
        return 1;
    }
#endif

#ifdef ENABLE_WR_FILTERING_OPT
    if (c == 'f') {
        if (**s == '\0') {
                *ind += 1;
            if (*ind >= argc) {
                fprintf(stderr, "%s: error: no argument given to option '%c'\n", progname, c);
                exit(1);
            }
            *s = argv[*ind];
        }
        if (filter_add_message_filter(*s) != 0) {
            fprintf(stderr, "Cannot parse filter level: %s\n", *s);
            exit(1);
        }
        /* Got it, move to the next */
        *s = NULL;
        return 1;
    }
#endif

    return 0;
}

#ifdef ENABLE_WR_FILTERING_OPT
void wrs_init_default_filtering(void) {
    /* Default filters (use "-fi" to disable). */
    filter_add_message_filter("o,E,Locator,peerHeartBeat");
    filter_add_message_filter("o,E,Locator,peerAdded");
    filter_add_message_filter("o,E,Locator,peerRemoved");
    filter_add_message_filter("o,E,Locator,peerChanged");
    filter_add_message_filter("o,C,Locator,getPeers");
    filter_add_message_filter("tr10,C,FileSystem,read");
    filter_add_message_filter("t10,C,FileSystem,write");
    
    set_proxy_log_filter_listener2 (filter_is_log_filtered);
}
#endif


#endif
