/*******************************************************************************
 * Copyright (c) 2007, 2010, 2013-2014 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * You may elect to redistribute this code under either of these licenses.
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*
 * This module handles process/thread OS contexts and their state machine.
 */

#include <tcf/config.h>

#if defined(_WRS_KERNEL)

#if ENABLE_DebugContext && !ENABLE_ContextProxy

#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <tcf/framework/mdep-threads.h>
#include <tcf/framework/events.h>
#include <tcf/framework/errors.h>
#include <tcf/framework/trace.h>
#include <tcf/framework/myalloc.h>
#include <tcf/framework/waitpid.h>
#include <tcf/framework/signames.h>
#include <tcf/framework/json.h>
#include <tcf/services/breakpoints.h>
#include <tcf/services/memorymap.h>
#include <tcf/services/pathmap.h>
#include <tcf/services/runctrl.h>
#include <tcf/services/tcf_elf.h>
#include <system/VxWorks/tcf/context-vxworks.h>
#if SERVICE_WR_Module
#include <tcf/services/WR_Module.h>
#endif

#include <errnoLib.h>
#include <moduleLib.h>
#include <taskHookLib.h>
#include <spinLockLib.h>
#include <vxdbgLib.h>
#include <private/vxdbgLibP.h>
#include <private/vmLibP.h>
#include <private/moduleHookLibP.h>
#include <private/sigLibP.h>
#include <private/jobLibP.h>
#include <private/taskLibP.h>

#if defined(_WRS_CONFIG_RTP)
#include <rtpLib.h>
#include <ioLib.h>
#include <ptyDrv.h>
#include <private/rtpLibP.h>
#include <private/rtpUtilLibP.h>
#include <private/taskMemLibP.h>
#include <private/trcLibP.h>       /* Required for RTP syscall regs retrieval */
#endif	/* _WRS_CONFIG_RTP */

#if ENABLE_Symbols
#include <tcf/services/symbols.h>
#endif

/*
 * Here is the contexts hierarchy as returned by getChildren interface for
 * VxWorks:
 *
 *   Contexts Hierarchy               Name                           ID
 * -----------------------  -----------------------------  ---------------------
 * + VxWorks                VX_SYSTEM_FOLDER_NAME           VX_SYSTEM_FOLDER_ID
 *     + Kernel             VX_KERNEL_FOLDER_NAME           VX_KERNEL_FOLDER_ID
 *     |   + t1             Task Name                       Task ID
 *     |   + ...
 *     |   + tN
 *     |
 *     + RTPs               VX_RTPS_FOLDER_NAME             VX_RTPS_FOLDER_ID
 *         + rtp1           RTP Name                        RTP ID
 *         |   + t1
 *         |   + ...
 *         |   + tN
 *         |
 *         + ...
 *         + rtpN
 *             + t1
 *             + ...
 *             + tN
 */

/* Defines Names for the main containers of the hierarchy */

#define VX_SYSTEM_FOLDER_NAME   "VxWorks"
#define VX_KERNEL_FOLDER_NAME   "Kernel"
#define VX_RTPS_FOLDER_NAME     "RTPs"

#define ENABLE_KERNEL_MEM_CTX_ATTACH_AT_STARTUP

#define TRACE_EVENT_STEP        2
#define STOP_AT_EVENTPOINT      3

#define DEFAULT_TASK_PRIORITY   100	/* Task creation default priority */
#define DEFAULT_STACK_SIZE      0x4000	/* Task creation default stack size */
#define WORKER_STACK_SIZE	0x2000	/* Worker task stack size */
#define INV_REG_MARKER          0xbb    /* Marker for invalid register */

#define CTX_IS_TASK(ctx)	((EXT(ctx)->properties & IS_TASK) == IS_TASK)
#define CTX_IS_RTP(ctx)		((EXT(ctx)->properties & IS_RTP) == IS_RTP)
#define CTX_IS_CONTAINER(ctx)	((EXT(ctx)->properties & IS_CONTAINER) == IS_CONTAINER)
#define CTX_IS_DETACHING(ctx)	((EXT(ctx)->properties & DETACHING) == DETACHING)
#define CTX_IN_TERMINATE_REQUEST(ctx)	((EXT(ctx)->properties & TERMINATE_REQUEST) == TERMINATE_REQUEST)
#define CTX_START_OPTS_HANDLED(ctx)	((EXT(ctx)->properties & START_OPTS_HANDLED) == START_OPTS_HANDLED)

#if defined(_WRS_CONFIG_RTP)
#define CTX_IS_IN_SYSCALL(ctx)	((EXT(ctx)->properties & (IS_IN_SYSCALL)) == IS_IN_SYSCALL)
#else
#define CTX_IS_IN_SYSCALL(ctx)	(FALSE)
#endif

#define	VXDBG_CLIENT_ID		((vxdbg_clnt_id != NULL) ? vxdbg_clnt_id : vxdbg_client_initialize())

#define SET_TASK_UNBREAKABLE_OPTION(tid, options)	\
    taskOptionsGet(tid, &taskOptions);		\
    taskOptionsSet (tid, 0, VX_UNBREAKABLE);

#define RESET_TASK_UNBREAKABLE_OPTION(tid, options)	\
    if ((taskOptions & VX_UNBREAKABLE) == 0) taskOptionsSet (tid, VX_UNBREAKABLE, 0);

#define NOTIFY_CONTEXT_STOP(ctx) \
    { \
    send_context_stopped_event(ctx); \
    if ((EXT(ctx)->properties & NOTIFY_CTX_CHANGED) == NOTIFY_CTX_CHANGED) send_context_changed_event(ctx); \
    EXT(ctx)->no_auto_pending_intercept = FALSE; \
    EXT(ctx)->stop_retry_cnt = 0; \
    }


#define	PTY_ROOT_DEV_NAME	"/pty/tcf"
#define	PTY_DEV_NAME_STR_SIZE	28
#define PIPE_SIZE		0x400

#define MODULE_ID_KEY		"ModuleId"

#define CLEAR_STOP_REASON(ctx) \
    { \
    EXT(ctx)->event = 0; \
    ctx->signal = 0; \
    ctx->stopped_by_exception = 0; \
    ctx->stopped_by_bp = 0; \
    ctx->stopped_by_cb = 0; \
    }

#define CONTEXT_CAN_DETACH(ctx) \
    (((EXT(ctx)->attach_type & ALWAYS_ATTACH) == ALWAYS_ATTACH) ? 0 : 1)


int vx_terminate_timeout = 2;		/* Delete timeout in ticks */
const char * TERMINATE_ERROR = "Terminate aborted";
const char * RTP_TERMINATE_ERROR = "RTP Terminate aborted";

typedef enum {
    IMPLICIT_ATTACH	= 0x0,	/* Implicit attachment (child attachment) */
    EXPLICIT_ATTACH	= 0x1,	/* Explicit attachment (direct attachment) */
    ALWAYS_ATTACH	= 0x2,	/* Always attach (never detach) */
    COND_PARENT_ATTACH	= 0x4	/* Conditional attachment: Only if parent is attached */
} attach_type;

typedef enum {
    IS_TASK		= 0x1,	/* Context is a task */
    IS_RTP		= 0x2,	/* Context is an RTP */
    IS_CONTAINER	= 0x4,	/* Context is a container */
    DETACHING		= 0x8,	/* Context detachment in progress */
    START_OPTS_HANDLED	= 0x10,	/* start options already handled? */
    IS_IN_SYSCALL       = 0x20, /* RTP Task context is stopped in system call */
    NOTIFY_CTX_CHANGED  = 0x40, /* Must notify context change */
    TERMINATE_REQUEST   = 0x80  /* stopped_on_error & stopped_errno must be */
                                /* preserved on context_continue */
} ctx_props;


typedef struct ContextExtensionVxWorks {
    pid_t               pid;
    pid_t               bp_pid;         /* process or thread that hit breakpoint */
    int                 event;
    REG_SET *           regs;           /* copy of context registers, updated when context stops */
    ErrorReport *       regs_error;     /* if not NULL, 'regs' is invalid */
    int                 regs_dirty;     /* if not 0, 'regs' is modified and needs to be saved before context is continued */
#if 0   /* rlp - to be removed when VxDGB exception handling will be supported  */  
    BOOL                pass_signal;    /* pass signal to task ? */
#endif
    attach_type         attach_type;    /* Implicit or explicit attachment? */
    BOOL		no_auto_pending_intercept;
					/* Is automatic "pending_intercept" */
					/* handling disabled? */
    int                 stop_retry_cnt; /* stop retry count for safe tasks */
    int			attach_mode;	/* Attachment mode specified at creation */
    ctx_props		properties;	/* Context properties */
    BreakpointInfo *	main_ep;	/* StopAtMain event point */
    BreakpointInfo *	entry_ep;	/* StopAtEntry event point */
    void *		task_start_address; /* task entry point */
    BOOL		created_by_agent; /* Context created from debug agent */
    const char *        stopped_on_error; /* Context stopped on agent error */
    int                 stopped_errno; /* errno corresponding to agent error */
#if defined(_WRS_CONFIG_RTP) && ENABLE_Symbols
    ContextAddress      rtp_task_entry; /* address of the rtp task entry point */
    int                 rtp_task_entry_sym_size; /* size of the rtp task entry point */
#endif
} ContextExtensionVxWorks;

typedef struct task_creation_info {
    NODE	chain;		/* task chain */
    TASK_ID	tid;		/* task ID */
    void *	start_address;	/* task entry point */
    int		attach_mode;	/* task attachment mode */
    ContextAttachCallBack * attach_cb;	/* attachment callback routine */
    void *	attach_cb_data;	/* attachment callback routine argument */
} TASK_CREATION_INFO;

#if defined(_WRS_CONFIG_RTP)
typedef struct rtp_creation_info {
    NODE	chain;		/* RTP chain */
    RTP_ID	rtp_id;		/* RTP ID */
    RTP_SPAWN_REQ * pReq;	/* Pointer to RTP spawn request */
} RTP_CREATION_INFO;

typedef struct SHL_MAP_INFO {
    NODE	chain;		/* RTP chain */
    char * 	name;		/* SHL file name */
    SHL_ID	shl_id;		/* RTP ID */
    uint64_t	text_offs;	/* text offset in ELF file */
    uint64_t	data_offs;	/* data offset in ELF file */
    uint64_t    text_addr;	/* text address in memory */
    uint64_t    data_addr;	/* data address in memory */
    uint64_t    bss_addr;	/* bss address in memory */
    uint64_t    text_size;	/* text size */
    uint64_t    data_size;	/* data size */
    uint64_t    bss_size;	/* bss size */
} SHL_MAP_INFO;

typedef struct rtp_info {
    NODE	chain;		/* RTP chain */
    RTP_ID	rtp_id;		/* RTP ID */
    uint64_t	text_offs;	/* text offset in ELF file */
    uint64_t	data_offs;	/* data offset in ELF file */
    LIST	shlList;	/* List of shared libraries */
    int		attached;	/* is RTP attached? */
} RTP_INFO;

typedef struct rtp_io_info {
    NODE	chain;		/* RTP chain */
    RTP_ID      rtp_id;         /* RTP ID */
    char *	pipe_name;	/* Device name used for IOredirection */
    int		pipe_fds[2];	/* STD IO file descriptors */
} RTP_IO_INFO;
#endif	/* _WRS_CONFIG_RTP */

static size_t context_extension_offset = 0;

#define EXT(ctx) ((ContextExtensionVxWorks *)((char *)(ctx) + context_extension_offset))

#include <tcf/framework/pid-hash.h>

#define EVENT_HOOK_BREAKPOINT   2
#define EVENT_HOOK_STEP_DONE    3
#define EVENT_HOOK_STOP         4
#define EVENT_HOOK_TASK_ADD     5
#define EVENT_HOOK_EXCEPTION    6
#define EVENT_HOOK_TASK_DELETE  7
#define EVENT_HOOK_START        8

#define VX_CONTEXT_CREATE	1
#define VX_CONTEXT_DELETE	2
#define VX_CONTEXT_SELF_DELETE	3

struct event_info {
    int                 event;
    VXDBG_CTX_ID        current_ctx;    /* context that hit breakpoint */
    VXDBG_CTX_ID        stopped_ctx;    /* context stopped by the breakpoint */
    REG_SET             regs;           /* task registers before exception */
    void *              addr;           /* breakpoint addr */
    int			signal;		/* intercepted signal */
    ContextBreakpoint ** triggeredCb;	/* array of triggered breakpoints */
					/* (Must be NULL terminated) */
    int			triggeredCbCount; /* number of triggered breakpoints */
};

#if defined(_WRS_CONFIG_RTP)
typedef struct object_each_rtn_args {
    Channel * c;			/* TCF channel for objects listing */
    int count;				/* # of objects listed */
    int attached_only;			/* List only attached objects */
} object_each_rtn_args;

typedef struct rtp_delete_info {
    pid_t rtp_id;			/* ID of exited RTP */
    int exit_code;			/* RTP exit code */
} RTP_DELETE_INFO;

static void      (*_func_rtpOldShlOpenHook) (SHL_ID shlId, RTP_ID rtpId);
static void      (*_func_rtpOldShlCloseHook) (SHL_ID shlId, RTP_ID rtpId);
#endif	/* _WRS_CONFIG_RTP */

static VXDBG_CLNT_ID vxdbg_clnt_id = NULL;

static Context * system_ctx = NULL;
static Context * kernel_ctx = NULL;
static Context * rtps_ctx = NULL;
static MemoryErrorInfo mem_err_info;
static ContextBreakpoint ** triggeredCb = NULL;
static int triggeredCbMax = 0;
static BOOL kernelIsAttached = FALSE;   /* Kernel ctx is implicitly attached */

LOCAL LIST taskStartList;	/* List of tasks created from Processes */
				/* service */
LOCAL SEM_ID taskStartListLock;	/* taskStartList mutual exclusion */
LOCAL SEM_ID vxdbgClntIdLock;	/* VxDBG client ID mutual exclusion */
#ifdef _WRS_CONFIG_RTP
LOCAL LIST rtpStartList;	/* List of RTPs created from Processes */
				/* service */
LOCAL SEM_ID rtpListLock;	/* rtpStartList mutual exclusion */
LOCAL LIST rtpMapList;		/* List of all RTPs with their mapping */
LOCAL LIST rtpIoList;		/* List of RTPs with IO redirection info */
#endif

#ifdef _WRS_CONFIG_RTP          /* Required for RTP syscall regs retrieval */
extern char wrs_kernel_text_start[];
extern char wrs_kernel_text_end[];
extern char * memPoolEndAdrs;
#endif

/* Forward declaration */

static Context * system_folder_create(attach_type type);
static Context * rtps_folder_create(Context * parent_ctx, attach_type type);
static Context * task_folder_create(Context * parent_ctx, pid_t vxworksId, attach_type type);
static Context * task_create_context (TASK_ID taskId, Context * parent_ctx);
static BOOL rtp_task_create_context_each_rtn (RTP_ID rtpId, TASK_ID taskId, _Vx_usr_arg_t arg);
static void kernel_tasks_update_context(Context * parent_ctx, int update_type);
static Context * vx_create_context(pid_t pid, attach_type type, int * err);
static int vx_delete_context (Context * ctx, int self_exit);
static char * vx_obj_get_name (pid_t pid);
static int is_attached(pid_t pid, int explicit);
static int vxworks_mem_test (char * addr, int size, int mode, int width,
			     BOOL preserve, size_t * p_size_valid,
                             size_t * p_size_invalid, int * p_error);
#if defined(_WRS_CONFIG_RTP)
static STATUS vx_verify_rtp_id (RTP_ID rtp_id);
static STATUS rtp_sc_check (Context * ctx, REG_SET * pRegs, BOOL * pIsSyscall);
static STATUS rtp_sc_pcValidate (INSTR * pc, TRC_OS_CTX * pOsCtx);
#endif
static STATUS vx_verify_task_id (pid_t pid);
static void vx_rtp_create_reply (void * data);
static VXDBG_CLNT_ID vxdbg_client_initialize (void);
static int vx_detach_context (Context * ctx, int self_exit);
static void vx_set_errno (int vx_errno);

#if defined(_WRS_CONFIG_RTP)
/* 
 * Find info about specified RTP in stored RTP info.
 *
 * It is up to the caller to lock the RTP info list.
 *
 * Returns: pointer on RTP info or NULL if RTP is not found
 */
static RTP_INFO * find_rtp_info (RTP_ID rtpId) {
    NODE * pDll;
    RTP_INFO * rtp_info;
    for (pDll = lstFirst (&rtpMapList); pDll != NULL; pDll = lstNext (pDll)) {
	rtp_info = (RTP_INFO *)pDll;
	if (rtp_info->rtp_id == rtpId) {
	    return rtp_info;
	}
    }
    return NULL;
}
#endif

#if ENABLE_USER_DEFINED_id2ctx
/*
 * Define a VxWorks specific version of id2ctx() which supports contexts
 * hierarchy with more than 2 levels.
 */
Context * id2ctx(const char * id) {
    Context * ctx = NULL;
    pid_t parent = 0;
    pid_t pid = id2pid(id, &parent);
    if (pid == 0) return NULL;

    if ((ctx = context_find_from_pid(pid, 0)) != NULL) {
	return (ctx);
    } else {
	return context_find_from_pid(pid, 1);
    }
}
#endif	/* ENABLE_USER_DEFINED_id2ctx */

#if ENABLE_ContextStateProperties
int context_get_state_properties(Context * ctx, const char *** names, const char *** values,
        int * cnt) {
#if defined(_WRS_CONFIG_RTP)
    static const char * prop_names[] = {"Context"};
    static const char * prop_values[] = {"In System Call"};

    if (CTX_IS_IN_SYSCALL (ctx)) {
        *names = prop_names;
        *values = prop_values;
        *cnt = 1;
        return 0;
    }
#endif  /* _WRS_CONFIG_RTP */
    *cnt = 0;
    return 0;
}
#endif

void vxdbg_post_and_free_event(void * args) {
    POST_EVENT_ARGS * post_event_args = (POST_EVENT_ARGS *)args;
    post_event(post_event_args->call_back, post_event_args->arg);
    free (post_event_args);
}

void vxdbg_post_event(void * args) {
    POST_EVENT_ARGS * post_event_args = (POST_EVENT_ARGS *)args;
    post_event(post_event_args->call_back, post_event_args->arg);
}

const char * context_suspend_reason(Context * ctx) {
    static char reason[128];

    if (EXT(ctx)->stopped_on_error != 0) {
	snprintf(reason, sizeof(reason), "Error: %s : %s", EXT(ctx)->stopped_on_error, errno_to_str(EXT(ctx)->stopped_errno));
	return reason;
    }

    if (EXT(ctx)->event == TRACE_EVENT_STEP) return REASON_STEP;
    if (EXT(ctx)->event == STOP_AT_EVENTPOINT) return REASON_BREAKPOINT;

    if ((ctx->signal != 0) && (ctx->signal != SIGSTOP) && (ctx->signal != SIGTRAP))  {
	snprintf(reason, sizeof(reason), "Signal %d", ctx->signal);
	return reason;
    }	    
    return REASON_USER_REQUEST;
}

typedef struct AttachDoneArgs {
    pid_t pid;
    ContextAttachCallBack * done;
    void * data;
    int attach_mode;
} AttachDoneArgs;

static void event_attach_done(void * x) {
    AttachDoneArgs * args = (AttachDoneArgs *)x;
    int err = 0;
    Context * ctx = NULL;

    /*
     * Create TCF context for context to attach, and create implicit attachments
     * for parent context(s) if necessary.
     */
    if ((ctx = vx_create_context (args->pid, EXPLICIT_ATTACH, &err)) != NULL) {
	trace(LOG_CONTEXT, "context: attached: ctx %#lx, id %#lx", ctx, EXT(ctx)->pid);

	if ((args->attach_mode & CONTEXT_ATTACH_NO_STOP) == 0) {
	    ctx->pending_intercept = 1;
	}

	EXT(ctx)->attach_mode = args->attach_mode;
    }
    args->done(err, ctx, args->data);
    loc_free(x);
}

int context_attach(pid_t pid, ContextAttachCallBack * done, void * data, int mode) {
    AttachDoneArgs * args = (AttachDoneArgs *)loc_alloc(sizeof(AttachDoneArgs));

    assert(done != NULL);
    assert((mode & CONTEXT_ATTACH_SELF) == 0);
    args->pid = pid;
    args->done = done;
    args->data = data;
    args->attach_mode = mode;
    post_event(event_attach_done, args);

    return 0;
}

int context_has_state(Context * ctx) {
    if ((ctx == NULL) || CTX_IS_CONTAINER(ctx)) {
	return 0;
    }
    return 1;
}

int context_stop(Context * ctx) {
    ContextExtensionVxWorks * ext = EXT(ctx);
    struct event_info * info;
    VXDBG_CTX vxdbg_ctx;

    assert(is_dispatch_thread());
    assert(ctx->parent != NULL);
    assert(!ctx->stopped);
    assert(!ctx->exited);
    assert(!ext->regs_dirty);

    /*
     * Disable the automatic "pending_intercept" handling for this stop
     * request coming from the agent itself.
     */
    EXT(ctx)->no_auto_pending_intercept = TRUE;

    if (ctx->pending_intercept) {
        trace(LOG_CONTEXT, "context: stop ctx %#lx, id %#lx", ctx, ext->pid);
    } else {
        trace(LOG_CONTEXT, "context: temporary stop ctx %#lx, id %#lx", ctx, ext->pid);
    }

    vxdbg_ctx.ctxId = ext->pid;
    vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
    if (vxdbgStop(VXDBG_CLIENT_ID, &vxdbg_ctx, FALSE) != OK) {
        int error = errno;
        if (error == S_vxdbgLib_INVALID_CTX) return 0;
        trace(LOG_CONTEXT, "context: can't stop ctx %#lx, id %#lx: %s", ctx, ext->pid, errno_to_str(error));
        return -1;
    }
    assert(taskIsStopped(ext->pid));
    return 0;
}

int context_continue(Context * ctx) {
    ContextExtensionVxWorks * ext = EXT(ctx);
    VXDBG_CTX vxdbg_ctx;
    int signal = 0;
    unsigned n = 0;

    assert(is_dispatch_thread());
    assert(ctx->parent != NULL);
    assert(ctx->stopped);
    assert(!ctx->pending_intercept);
    assert(!ctx->exited);
    assert(taskIsStopped(ext->pid));

    if (ctx->parent && ctx->parent->exiting) return 0;

    trace(LOG_CONTEXT, "context: continue ctx %#lx, id %#lx", ctx, ext->pid);

    /* If context was stopped on agent error, clear the error before resuming */

    if (ext->stopped_on_error && !CTX_IN_TERMINATE_REQUEST(ctx)) {
	ext->stopped_on_error = NULL;
	ext->stopped_errno = 0;
    } else {
        EXT(ctx)->properties &= ~TERMINATE_REQUEST;
    }

    /*
     * Check for pending signals on this context.
     * If there is a pending signal, check if this signal must be passed
     * or not.
     */

    while (sigset_get_next(&ctx->pending_signals, &n)) {
	if (sigset_get(&ctx->sig_dont_pass, n)) {
	    sigset_set(&ctx->pending_signals, n, 0);
	}
	else {
	    signal = n;
	    break;
	}
    }

#if 0   /* rlp - to be removed when VxDGB exception handling will be supported  */
    /*
     * Indicate to the signal exception handler (sigExcIntercept()) if signal
     * must be passed or not.
     */
    ext->pass_signal = (signal != 0);
#endif

    if (ext->regs_dirty) {
        if (taskRegsSet((TASK_ID)(ext->pid), ext->regs) != OK) {
            int error = errno;
            trace(LOG_CONTEXT, "context: can't set regs ctx %#lx, id %#lx: %s",
                    ctx, ext->pid, errno_to_str(error));
            return -1;
        }
        ext->regs_dirty = 0;
    }

    vxdbg_ctx.ctxId = ext->pid;
    vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
    if (vxdbgCont(VXDBG_CLIENT_ID, &vxdbg_ctx) != OK) {
        int error = errnoGet();

	if (error == S_vxdbgLib_INVALID_CTX) {
	    return 0;	/* The context no longer exists (has been deleted) */
	}

        trace(LOG_CONTEXT, "context: can't continue ctx %#lx, id %#lx: %s",
                ctx, ext->pid, errno_to_str(error));
        return -1;
    }
    assert(!taskIsStopped(ext->pid));
    send_context_started_event(ctx);
    return 0;
}

int context_single_step(Context * ctx) {
    ContextExtensionVxWorks * ext = EXT(ctx);
    VXDBG_CTX vxdbg_ctx;

    assert(is_dispatch_thread());
    assert(context_has_state(ctx));
    assert(ctx->parent != NULL);
    assert(ctx->stopped);
    assert(!ctx->exited);
    assert(taskIsStopped(ext->pid));

#if defined(_WRS_CONFIG_RTP)
    if (CTX_IS_IN_SYSCALL (ctx)) {
        set_errno(ERR_OTHER, "Cannot step RTP task stopped in system call");
        return -1;
    }
#endif

    trace(LOG_CONTEXT, "context: single step ctx %#lx, id %#lx", ctx, ext->pid);

    if (ext->regs_dirty) {
        if (taskRegsSet((TASK_ID)(ext->pid), ext->regs) != OK) {
            int error = errno;
            trace(LOG_CONTEXT, "context: can't set regs ctx %#lx, id %#lx: %s",
                    ctx, ext->pid, errno_to_str(error));
            return -1;
        }
        ext->regs_dirty = 0;
    }

    vxdbg_ctx.ctxId = ext->pid;
    vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
    if (vxdbgStep(VXDBG_CLIENT_ID, &vxdbg_ctx, NULL, NULL) != OK) {
        int error = errno;
        trace(LOG_CONTEXT, "context: can't step ctx %#lx, id %#lx: %d",
                ctx, ext->pid, errno_to_str(error));
        return -1;
    }

    /*
     * Set no_auto_pending_intercept flag to indicate that single step
     * request comes from the debug agent, and does not need to be
     * intercepted in EVENT_HOOK_STEP_DONE event handler.
     */
    EXT(ctx)->no_auto_pending_intercept = TRUE;
    send_context_started_event(ctx);
    return 0;
}

static int context_terminate(Context * ctx) {
    ContextExtensionVxWorks * ext = EXT(ctx);

    assert(is_dispatch_thread());
    assert(ctx->parent != NULL);
    assert(ctx->stopped);
    assert(!ctx->pending_intercept);
    assert(!ctx->exited);
    assert(taskIsStopped(ext->pid));

    trace(LOG_CONTEXT, "context: terminate ctx %#lx, id %#lx", ctx, ext->pid);

    if (ctx->exiting) return 0;

    /* Do not try to delete system containers */
    if (CTX_IS_CONTAINER(ctx) && !CTX_IS_RTP(ctx)) return 0;

    if (ext->regs_dirty) {
        taskRegsSet((TASK_ID)(ext->pid), ext->regs);
        ext->regs_dirty = 0;
    }

    /*
     * Mark context and children contexts as exiting.
     * exiting flag will be reset if terminate fails.
     */
    ctx->exiting = 1;

    if (!list_is_empty(&ctx->children)) {
        LINK * l = ctx->children.next;
        while (l != &ctx->children) {
            Context * c = cldl2ctxp(l);
            if (!c->exited) {
		c->exiting = 1;
            }
            l = l->next;
        }
    }

    if (vx_context_terminate (ext->pid, NULL, NULL) == -1) {
	trace(LOG_CONTEXT, "context: can't kill ctx %#lx, id %#lx: %s",
		    ctx, ext->pid, errno_to_str(errnoGet()));
	return -1;
    }

    return 0;
}

/*******************************************************************************
*
* vx_delete_context_event - post_event routine for context deletion
*
* This routine is used form context_detach() to defer context deletion.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vx_delete_context_event (void * arg) {
    vx_delete_context ((Context *)arg, 0);
}

/*******************************************************************************
*
* context_detach - detach given context and children contexts
*
* This routine marks given context and children contexts as exiting.
* Contexts must not be deleted directly from context_detach(), the context
* deletion is defered via a post_event: See vx_delete_context_event().
*
* RETURNS: Always 0.
*
* ERRNO: N/A
*
*/

static int context_detach(Context * ctx) {

    /*
     * Mark context and children contexts as exiting (Only if
     * context can be detached).
     */
    if (CONTEXT_CAN_DETACH(ctx)) ctx->exiting = 1;

    if (!list_is_empty(&ctx->children)) {
        LINK * l = ctx->children.next;
        while (l != &ctx->children) {
            Context * c = cldl2ctxp(l);
            if (!c->exited) {
		if (CONTEXT_CAN_DETACH(c)) c->exiting = 1;
            }
            l = l->next;
        }
    }

    /*
     * Contexts must not be deleted directly from context_detach(), the context
     * context deletion is defered via post_event.
     */
    post_event (vx_delete_context_event, ctx);
    return 0;
}

int context_resume(Context * ctx, int mode, ContextAddress range_start, ContextAddress range_end) {
    switch (mode) {
    case RM_RESUME:
        return context_continue(ctx);
    case RM_STEP_INTO:
        return context_single_step(ctx);
    case RM_TERMINATE:
        /* Try to terminate task or RTP context */
        if (context_terminate(ctx) == 0) return 0;

        if (CTX_IS_TASK(ctx)) {
            /*
             * Failed to terminate task context: resume it and retry.
             * NOTE: This should not be done for RTP context because for now
             * on RM_TERMINATE, the agent does not stop RTP tasks.
             */
            if (context_continue(ctx) == -1) return -1;
            (void)vx_context_terminate(EXT(ctx)->pid, NULL, NULL);
        }
	return 0;
    case RM_DETACH:
	return context_detach(ctx);
    }
    errno = ERR_UNSUPPORTED;
    return -1;
}

int context_can_resume(Context * ctx, int mode) {
    switch (mode) {
    case RM_RESUME:
        return 1;
    case RM_STEP_INTO:
        /* Disable stepping for task stopped while executing a system call */
#if defined(_WRS_CONFIG_RTP)
        if (CTX_IS_IN_SYSCALL(ctx)) return 0;
#endif
        return context_has_state(ctx);
    case RM_TERMINATE:
	/* Can terminate tasks & RTPs contexts */
	if (CTX_IS_TASK(ctx) || CTX_IS_RTP(ctx)) {
	    return 1;
	}
    case RM_DETACH:
	return ctx != NULL;
    }
    return 0;
}

int context_read_mem(Context * ctx, ContextAddress address, void * buf, size_t size) {
    int status = 0;
    size_t size_valid = 0;
    size_t size_invalid = 0;
    int error = 0;
#if defined(_WRS_CONFIG_RTP)        
    RTP_ID old_rtp_id = NULL;
    RTP_ID rtp_id = (RTP_ID)EXT(ctx->mem)->pid;

    /*
     * If the context to read is not the kernel one, we must switch to RTP
     * memory ctx.
     */
    if ((rtp_id != (RTP_ID)VX_KERNEL_FOLDER_ID) && (_func_taskMemCtxSwitch != NULL)) {
	old_rtp_id = _func_taskMemCtxSwitch (rtp_id, 0);
    }
#endif	/* _WRS_CONFIG_RTP */

    mem_err_info.error = 0;

    /* Verify that this memory block can be read */

    if (vxworks_mem_test ((char *)address, size, VX_READ, 0, TRUE, &size_valid, &size_invalid, &error) < 0) {
        if (error != 0) {
            /*
             * The end of the block to read is not valid: Read the valid part
             * and return an error. mem_err_info structure will be used by TCF
             * to obtain additional information about this memory access error.
             */
#ifdef _WRS_PERSISTENT_SW_BP
            vxdbgMemRead((void *)address, buf, size_valid, 0);
#else
            bcopy((void *)address, buf, size_valid);
#endif

            if (check_breakpoints_on_memory_read(ctx, address, buf, size_valid) == 0) {
                /* Fill mem_err_info structure */
                mem_err_info.error = error;
                mem_err_info.size_valid = size_valid;
                mem_err_info.size_error = size_invalid;
            }
        }

	errno = ERR_INV_ADDRESS;
	status = -1;
	goto context_read_mem_end;
    }
    
#ifdef _WRS_PERSISTENT_SW_BP
    vxdbgMemRead((void *)address, buf, size, 0);
#else
    bcopy((void *)address, buf, size);
#endif
    if (check_breakpoints_on_memory_read(ctx, address, buf, size) < 0) {
	status = -1;
	goto context_read_mem_end;
    }

context_read_mem_end:
#if defined(_WRS_CONFIG_RTP)        
    if (old_rtp_id != NULL) _func_taskMemCtxSwitch (old_rtp_id, 0);
#endif
    return status;
}

int context_write_mem(Context * ctx, ContextAddress address, void * buf, size_t size) {
    int status = 0;
    size_t size_valid = 0;
    size_t size_invalid = 0;
    int error = 0;
#if defined(_WRS_CONFIG_RTP)        
    RTP_ID old_rtp_id = NULL;
    RTP_ID rtp_id = (RTP_ID)EXT(ctx->mem)->pid;

    /*
     * If the context to read is not the kernel one, we must switch to RTP
     * memory ctx.
     */
    if ((rtp_id != (RTP_ID)VX_KERNEL_FOLDER_ID) && (_func_taskMemCtxSwitch != NULL)) {
	old_rtp_id = _func_taskMemCtxSwitch (rtp_id, 0);
    }
#endif	/* _WRS_CONFIG_RTP */

    mem_err_info.error = 0;

    /* Verify that this memory block can be written */

    if (vxworks_mem_test ((char *)address, size, VX_WRITE, 0, TRUE, &size_valid, &size_invalid, &error) < 0) {
        if (error != 0) {
            /*
             * The end of the block to write is not valid: Write the valid part
             * and return an error. mem_err_info structure will be used by TCF
             * to obtain additional information about this memory access error.
             */

            if (check_breakpoints_on_memory_write(ctx, address, buf, size_valid) < 0) {
                errno = error;
                status = -1;
                goto context_write_mem_end;
            }

#ifdef _WRS_PERSISTENT_SW_BP
            vxdbgMemWrite(buf, (void *)address, size_valid, 0);
#else
            bcopy(buf, (void *)address, size_valid);
#endif

            /* Fill mem_err_info structure */
            mem_err_info.error = error;
            mem_err_info.size_valid = size_valid;
            mem_err_info.size_error = size_invalid;

            errno = error;
            status = -1;
            goto context_write_mem_end;
	}
    }

    if (check_breakpoints_on_memory_write(ctx, address, buf, size) < 0) {
	status = -1;
	goto context_write_mem_end;
    }
#ifdef _WRS_PERSISTENT_SW_BP
    vxdbgMemWrite(buf, (void *)address, size, 0);
#else
    bcopy(buf, (void *)address, size);
#endif

context_write_mem_end:
#if defined(_WRS_CONFIG_RTP)        
    if (old_rtp_id != NULL) _func_taskMemCtxSwitch (old_rtp_id, 0);
#endif
    return status;
}

#if ENABLE_ContextISA
int context_get_isa(Context * ctx, ContextAddress addr, ContextISA * isa) {
    memset(isa, 0, sizeof(ContextISA));
#if defined(__i386__)
    isa->def = "386";
    isa->max_instruction_size = sizeof(INSTR) * 15;
#elif defined(__x86_64__)
    isa->def = "X86_64";
    isa->max_instruction_size = sizeof(INSTR) * 15;
#elif defined(__arm__)
    isa->def = "ARM";
    isa->max_instruction_size = sizeof(INSTR) * 4;
#elif defined(__powerpc__)
    isa->def = "PPC";
    isa->max_instruction_size = sizeof(INSTR) * 4;
#else
    isa->def = NULL;
#endif
    isa->alignment = DBG_INST_ALIGN;
#if SERVICE_Symbols
    if (get_context_isa(ctx, addr, &isa->isa, &isa->addr, &isa->size) < 0) return -1;
#endif
    return 0;
}
#endif

/*******************************************************************************
*
* vxworks_mem_test - test VxWorks memory for valid read or write access
*
* This routine tests a range of addresses.
*
* RETURNS: 0 if memory is accessible, -1 otherwise.
*
* ERRNO: N/A
*
*/

static int vxworks_mem_test
    (
    char *	addr,		/* address to test */
    int		size,		/* number of bytes to test */
    int		mode,		/* VX_READ or VX_WRITE */
    int		width,		/* 1, 2, 4 or 8 (on VxWorks 64bit) */
    BOOL	preserve,	/* preserve memory on write test? */
    size_t *    p_size_valid,
    size_t *    p_size_invalid,
    int *       p_error
    )
    {
    ULONG	dummy;		/* dummy variable */
    char *	pageAddr;	/* address to probe */
    size_t	pageSize = VM_PAGE_SIZE_GET();	/* VxWork memory page size */

    if (size == 0)		/* memory length is null, test is done */
	return (0);

    /* Validate parameters */

    if (mode == VX_READ)
    	preserve = FALSE;

    if (width == 0)
    	width = 1;

    if (addr > addr + size)	/* test if we are not out of memory */
    	return (-1);

    /* Check width parameter */

    if ((width != 1) && (width != 2) && (width != 4)
#ifdef	_WRS_CONFIG_LP64
	&& (width != 8)
#endif	/* _WRS_CONFIG_LP64 */
       )
	return (-1);

    /* Check addr, size & width parameters coherency */

    if (((unsigned long) addr % width) || (size % width))
	return (-1);

    /* Check first address */

    if ((preserve && (vxMemProbe (addr, VX_READ, width,
    		      (char *)&dummy) != OK)) ||
	(vxMemProbe (addr, mode, width, (char *)&dummy) != OK))
	return (-1);

    /* Check if we have tested the whole memory */

    if (width == size)
    	return (0);

    /* Check if we must test intermediate pages */

    if ((pageSize != 0) && (pageSize != ERROR)) {
	/* Memory is paged; loop to check beginning of each page */

    	for (pageAddr = ROUND_UP (addr, pageSize);
    	     pageAddr < addr + size;
	     pageAddr += pageSize)
	    {
	    if ((preserve && (vxMemProbe (pageAddr, VX_READ, (int) width,
	    				(char *)&dummy) != OK)) ||
		(vxMemProbe (pageAddr, mode, (int) width,
					(char *)&dummy) != OK)) {
                /*
                 * The requested block is not entirely valid. Compute the
                 * size which can be read/written successfully.
                 */

                *p_error = ERR_INV_ADDRESS;
		*p_size_valid = pageAddr - addr;

                /* Now, we must determine the size of invalid data */

                for (pageAddr += pageSize;
                     pageAddr < addr + size;
                     pageAddr += pageSize) {
                     STATUS probeStatus = ERROR;

                    if (preserve) {
                        if ((vxMemProbe (pageAddr, VX_READ,
                                        (int) width, (char *)&dummy) == OK) &&
                            (vxMemProbe (pageAddr, mode, (int) width,
                                        (char *)&dummy) == OK)) {
                            probeStatus = OK;
                        }
                    } else if (vxMemProbe (pageAddr, mode, (int) width,
                                           (char *)&dummy) == OK) {
                            probeStatus = OK;
                    }

                    if (probeStatus == OK) {
                        /* The end of the requested data is valid */
                        *p_size_invalid = pageAddr - (addr + *p_size_valid);
                        return (-1);
                    }
                }

                /* The remaining of the requested data is invalid */
		*p_size_invalid = size - *p_size_valid;
                return (-1);
	    }
	}
    } else {
        /* Check last address */

        addr = addr + size - width;
        if ((preserve && (vxMemProbe (addr, VX_READ,
                          (int) width, (char *)&dummy) != OK)) ||
            (vxMemProbe (addr, mode, (int) width, (char *)&dummy) != OK)) {
            /*
             * Known limitation: In MMU less mode (not supported), we do not
             * return information to fill MemoryErrorInfo structure.
             */
            return (-1);
        }
    }

    return (0);
}

#if ENABLE_ExtendedMemoryErrorReports
int context_get_mem_error_info(MemoryErrorInfo * info) {
    if (mem_err_info.error == 0) {
        set_errno(ERR_OTHER, "Extended memory error info not available");
        return -1;
    }
    *info = mem_err_info;
    return 0;
}
#endif

int context_write_reg(Context * ctx, RegisterDefinition * def, unsigned offs, unsigned size, void * buf) {
    ContextExtensionVxWorks * ext = EXT(ctx);

    assert(is_dispatch_thread());
    assert(context_has_state(ctx));
    assert(ctx->stopped);
    assert(!ctx->exited);
    assert(offs + size <= def->size);

#if defined(_WRS_CONFIG_RTP)
    if (CTX_IS_IN_SYSCALL (ctx)) {
        set_errno(ERR_OTHER, "System Call registers are read-only");
        return -1;
    }
#endif

    if (ext->regs_error) {
        set_error_report_errno(ext->regs_error);
        return -1;
    }

    memcpy((uint8_t *)ext->regs + def->offset + offs, buf, size);
    ext->regs_dirty = 1;
    return 0;
}

int context_read_reg(Context * ctx, RegisterDefinition * def, unsigned offs, unsigned size, void * buf) {
    ContextExtensionVxWorks * ext = EXT(ctx);

    assert(is_dispatch_thread());
    assert(context_has_state(ctx));
    assert(ctx->stopped);
    assert(!ctx->exited);
    assert(offs + size <= def->size);

    if (ext->regs_error) {
        set_error_report_errno(ext->regs_error);
        return -1;
    }

#if defined(_WRS_CONFIG_RTP)
    if (CTX_IS_IN_SYSCALL (ctx)) {
        /*
         * Context is in a system call: we must return only valid registers.
         * Detect invalid registers using INV_REG_MARKER.
         */
        int idx;
        int marker_cnt = 0;

        for (idx = 0; idx != size; idx++) {
            if (((*((char *)ext->regs + def->offset + offs + idx)) & 0xff) != INV_REG_MARKER) break;
            marker_cnt++;
        }

        if (marker_cnt == size) {
            set_errno(ERR_OTHER, "Register not available in system call stack frame");
            return -1;
        }
    }
#endif  /* _WRS_CONFIG_RTP */

    memcpy(buf, (uint8_t *)ext->regs + def->offset + offs, size);
    return 0;
}

int context_get_canonical_addr(Context * ctx, ContextAddress addr,
        Context ** canonical_ctx, ContextAddress * canonical_addr,
        ContextAddress * block_addr, ContextAddress * block_size) {
    /* Direct mapping, page size is irrelevant */
    ContextAddress page_size = 0x100000;
    assert(is_dispatch_thread());
    *canonical_ctx = kernel_ctx;
    if (canonical_addr != NULL) *canonical_addr = addr;
    if (block_addr != NULL) *block_addr = addr & ~(page_size - 1);
    if (block_size != NULL) *block_size = page_size;
    return 0;
}

Context * context_get_group(Context * ctx, int group) {
    switch (group) {
    case CONTEXT_GROUP_STOP:
    case CONTEXT_GROUP_INTERCEPT:
        return ctx;
    case CONTEXT_GROUP_PROCESS:
    case CONTEXT_GROUP_CPU:
    case CONTEXT_GROUP_SYMBOLS:
    	/* NULL is not allowed for those context groups */
	if (ctx->mem == NULL) return ctx;
	else return ctx->mem;
    }

    /* Return NULL context if we have no memory access for this context */
    if (ctx->mem == NULL || ctx->mem->mem_access == 0) return NULL;
    else return ctx->mem;
}

int context_get_supported_bp_access_types(Context * ctx) {
    int access_types = 0;

    if (vxdbgHwBpTypeValidate (VXDBG_ACCESS_DATA_READ, 0) == OK) {
        access_types |= CTX_BP_ACCESS_DATA_READ;
    }
    if (vxdbgHwBpTypeValidate (VXDBG_ACCESS_DATA_WRITE, 0) == OK) {
        access_types |= CTX_BP_ACCESS_DATA_WRITE;
    }
    if (vxdbgHwBpTypeValidate (VXDBG_ACCESS_INSTRUCTION, 0) == OK) {
        access_types |= CTX_BP_ACCESS_INSTRUCTION;
    }

    return access_types;
}

static int context_plant_hw_breakpoint(ContextBreakpoint * bp, pid_t pid) {
    VXDBG_CTX vxdbg_ctx;
    VXDBG_BP_ID bp_id = 0;
    VXDBG_HWBP_ACCESS_TYPE vxdbg_access_type = 0;

    memset(&vxdbg_ctx, 0, sizeof(vxdbg_ctx));
    if (pid == VX_KERNEL_FOLDER_ID) {
        vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
    } else if (vx_verify_task_id (pid) == OK) {
        vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
        vxdbg_ctx.ctxId = (VXDBG_CTX_ID) pid;
#if defined(_WRS_CONFIG_RTP)        
    } else if (vx_verify_rtp_id((RTP_ID)pid) == OK) {
        vxdbg_ctx.ctxType = VXDBG_CTX_RTP;
        vxdbg_ctx.ctxId = (VXDBG_CTX_ID) pid;
#endif  /* _WRS_CONFIG_RTP */        
    } else {
        errno = ERR_INV_CONTEXT;
        return -1;
    }

    /* Convert TCF access type to VxDBG access type */

    if ((bp->access_types & CTX_BP_ACCESS_DATA_READ) == CTX_BP_ACCESS_DATA_READ) vxdbg_access_type |= VXDBG_ACCESS_DATA_READ;
    if ((bp->access_types & CTX_BP_ACCESS_DATA_WRITE) == CTX_BP_ACCESS_DATA_WRITE) vxdbg_access_type |= VXDBG_ACCESS_DATA_WRITE;
    if ((bp->access_types & CTX_BP_ACCESS_INSTRUCTION) == CTX_BP_ACCESS_INSTRUCTION) vxdbg_access_type |= VXDBG_ACCESS_INSTRUCTION;

    /* Add a VxDBG HW breapoint for this TCF ContextBreakpoint */

    if (vxdbgHwBpAdd(VXDBG_CLIENT_ID,
            &vxdbg_ctx, bp->length, vxdbg_access_type,
            BP_ACTION_STOP | BP_ACTION_NOTIFY,
            0, 0, (INSTR *)bp->address, 0, 0, &bp_id) != OK) {
        vx_set_errno(errnoGet());
        return -1;
    }

    /* Associate TCF ContextBreakpoint to VxDBG breakpoint */

    if (vxdbgBpUserKeySet (VXDBG_CLIENT_ID, bp_id, bp) == VXDBG_ERROR) {
	context_unplant_breakpoint (bp);
	return -1;
    }

    bp->id = bp_id;
    return 0;
}

static void is_hardware_bp(BreakpointInfo * bi, void * data) {
    BreakpointAttribute * attr = get_breakpoint_attributes(bi);
    int * p_hwbp = (int *) data;

    *p_hwbp = 0; /* By default, assume given BP is a software breakpoint */

    while (attr != NULL) {
        /* If an "AccessMode" is defined, then this is a watchpoint */
        if (strcmp(attr->name, "AccessMode") == 0) {
            *p_hwbp = 1;
            return;
        }

        /* If an "BreakpointType" is "Hardware" then this is an HW breakpoint */
        if (strcmp(attr->name, "BreakpointType") == 0) {
            *p_hwbp = strcmp(attr->value, "\"Hardware\"") == 0 ? 1 : 0;
            return;
        }
        attr = attr->next;
    }
}

int context_plant_breakpoint(ContextBreakpoint * bp) {
    pid_t pid;
    VXDBG_CTX vxdbg_ctx;
    VXDBG_BP_ID bp_id = 0;
    int hwbp = 0;
    Context * grp = context_get_group(bp->ctx, CONTEXT_GROUP_BREAKPOINT);

    if ((pid = id2pid (bp->ctx->id, NULL)) == 0) {
        errno = ERR_INV_CONTEXT;
        return -1;        
    }

    if (pid == VX_SYSTEM_FOLDER_ID) {
        errno = ERR_INV_CONTEXT;
        return -1;
    }

#if defined(_WRS_CONFIG_RTP)
    if (pid == VX_RTPS_FOLDER_ID) {
        errno = ERR_INV_CONTEXT;
        return -1;
    }
#endif

    /*
     * Iterate breakpoint list to determine if we're trying to plant
     * an hardware breakpoint.
     */
    iterate_context_breakpoint_links(grp, bp, is_hardware_bp, (void *) &hwbp);

    if (hwbp) {
        /* breakpoint to plant is an hardware breakpoint */
        return context_plant_hw_breakpoint(bp, pid);
    }

    if (bp->length != 1) {
        errno = ERR_INV_FORMAT;
        return -1;
    }
        
    memset(&vxdbg_ctx, 0, sizeof(vxdbg_ctx));
    if (pid == VX_KERNEL_FOLDER_ID) {
        vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
    } else if (vx_verify_task_id (pid) == OK) {
        vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
        vxdbg_ctx.ctxId = (VXDBG_CTX_ID) pid;
#if defined(_WRS_CONFIG_RTP)        
    } else if (vx_verify_rtp_id((RTP_ID)pid) == OK) {
        vxdbg_ctx.ctxType = VXDBG_CTX_RTP;
        vxdbg_ctx.ctxId = (VXDBG_CTX_ID) pid;
#endif  /* _WRS_CONFIG_RTP */        
    } else {
        errno = ERR_INV_CONTEXT;
        return -1;
    }
    
    /* Add a VxDBG breapoint for this TCF ContextBreakpoint */

    if (vxdbgBpAdd(VXDBG_CLIENT_ID,
            &vxdbg_ctx, 0, BP_ACTION_STOP | BP_ACTION_NOTIFY,
            0, 0, (INSTR *)bp->address, 0, 0, &bp_id) != OK) {
        vx_set_errno(errnoGet());
        return -1;
    }

    /* Associate TCF ContextBreakpoint to VxDBG breakpoint */

    if (vxdbgBpUserKeySet (VXDBG_CLIENT_ID, bp_id, bp) == VXDBG_ERROR) {
	context_unplant_breakpoint (bp);
	return -1;
    }

    bp->id = bp_id;
    return 0;
}

int context_unplant_breakpoint(ContextBreakpoint * bp) {
    VXDBG_BP_DEL_INFO info;
    memset(&info, 0, sizeof(info));
    info.pClnt = VXDBG_CLIENT_ID;
    info.type = BP_BY_ID_DELETE;
    info.info.id.bpId = bp->id;
    if ((vxdbgBpDelete(info) != OK) && (errno != S_vxdbgLib_INVALID_BP)) return -1;
    return 0;
}

#if ENABLE_ExtendedBreakpointStatus
int context_get_breakpoint_status(ContextBreakpoint * bp, const char *** names, const char *** values, int * cnt) {
    static const char * n[1];
    static const char * v[1];
    n[0] = "BreakpointType";

    if (bp->access_types & (CTX_BP_ACCESS_DATA_READ|CTX_BP_ACCESS_DATA_WRITE)) {
	v[0] = "\"Hardware\"";
    } else {
	v[0] = "\"Software\"";
    }

    *names = n;
    *values = v;
    *cnt = 1;
    return 0;
}
#endif

static void add_map_int_attr(MemoryRegion * r, char * name, long value) {
    MemoryRegionAttribute *attr = NULL;

    if ((name == NULL) || (r == NULL))
        return;

    attr =  (MemoryRegionAttribute *) loc_alloc_zero(sizeof(MemoryRegionAttribute));
    attr->name = loc_strdup (name);
    attr->value = loc_alloc (25);
    sprintf(attr->value,"\"0x%lx\"", value);
    if (r->attrs == NULL) {
        r->attrs = attr;
    } else {
        attr->next = r->attrs;
        r->attrs = attr;
    }
}

static MemoryRegion * add_map_region(MemoryMap * map, ContextAddress addr, ContextAddress size, uint64_t file_offs, unsigned flags, char * file, char * sect, int bss) {
    MemoryRegion * r = NULL;
    if (map->region_cnt >= map->region_max) {
        map->region_max += 8;
        map->regions = (MemoryRegion *)loc_realloc(map->regions, sizeof(MemoryRegion) * map->region_max);
    }
    r = map->regions + map->region_cnt++;
    memset(r, 0, sizeof(MemoryRegion));
    r->addr = addr;
    r->size = size;
    r->file_offs = file_offs;
    r->flags = flags;
    r->bss = bss;
    if (file != NULL) r->file_name = loc_strdup(file);
    if (sect != NULL) r->sect_name = loc_strdup(sect);
    return r;
}

static int module_list_proc(MODULE_ID id, _Vx_usr_arg_t args) {
    MODULE_INFO info;
    MemoryMap * map = (MemoryMap *)args;

    if ((id->flags & HIDDEN_MODULE) == HIDDEN_MODULE) {
	/* Skip hidden modules such as vxWorks.sym */
	return 1;
    }

    memset(&info, 0, sizeof(info));
    if (moduleInfoGet(id, &info) == OK) {
        char * file = id->nameWithPath;
        MemoryRegion *mr;
	SECTION_ID sectionId = moduleFirstSectionGet(id);
	unsigned int flags = 0;

	while (sectionId != NULL)
	    {
	    flags = 0;
	    if (sectionId->type == SECTION_RODATA) flags = MM_FLAG_R;
	    else if (sectionId->type == SECTION_TEXT) flags = MM_FLAG_R | MM_FLAG_X;
	    else if ((sectionId->type == SECTION_DATA) || (sectionId->type == SECTION_BSS)) flags = MM_FLAG_R | MM_FLAG_W;

	    mr = add_map_region(map, (ContextAddress) sectionId->address, sectionId->size, 0, flags, file, sectionId->name, (sectionId->type == SECTION_BSS) ? 1 : 0);
	    add_map_int_attr(mr, MODULE_ID_KEY, (long)id);
	    sectionId = moduleNextSectionGet(sectionId);
	    }
	}
    return 1;
}

static void module_load_event(void * args) {
    if (kernel_ctx != NULL) {
        memory_map_event_module_loaded(kernel_ctx);
    }
}

static void module_load_func(MODULE_ID  id, UINT32 options) {
    VXDBG_POST_EVENT(module_load_event, NULL, NULL);
}

static void module_unload_event(void * args) {
    if (kernel_ctx != NULL) {
        memory_map_event_module_unloaded (kernel_ctx);
    }
}

static void module_unload_func(MODULE_ID  id, UINT32 options) {
    VXDBG_POST_EVENT(module_unload_event, NULL, NULL);
}

/*******************************************************************************
*
* task_create_func - VxWorks task creation hook
*
* The purpose of this task creation hook is to stop & attach to a task as
* soon as it is created. This allow breakpoints to be planted by the debug
* agent in any situation.
*
* The task is stopped and attached only if it is a breakable task, and if the
* debugger is implicitly attached to its parent context (Either Kernel context
* or parent RTP context, or creator context).
*
* WARNING: For now, this routine does not handle the case where we're not
*          attached to parent memory context, but we're attached to creator
*          context, and creator context has the CONTEXT_ATTACH_CHILDREN
*          property (See VXW7-1624).
*          In this specific case, we may not hit breakpoints set for this task
*          until the debugger tasks are scheduled to create a TCF context for
*          this task.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void task_create_func(WIND_TCB * pTcb) {
    VXDBG_CTX vxdbg_ctx;

    /* Ignore unbreakable task */
    if ((pTcb->options & VX_UNBREAKABLE) != 0) return;

#if defined(_WRS_CONFIG_RTP)
    if (pTcb->rtpId != kernelIdGet()) {
        /*
         * We're handling an RTP task creation
         */
        RTP_INFO * pInfo;

        semTake (rtpListLock, WAIT_FOREVER);
        pInfo = find_rtp_info(pTcb->rtpId);
        semGive (rtpListLock);

        /* Ignore task if not attached to parent RTP */
        if ((pInfo == NULL) || !pInfo->attached) return;

        goto attach_new_task;
    }
#endif

    /*
     * We're handling a kernel task creation
     */

    if (!kernelIsAttached) {
        /*
         * We're not attached to kernel context: In this case, we must only
         * attach to this task if creator task has been started via Processes
         * Service with "AttachChildren"=TRUE.
         *
         * Here, we can only detect if debug client is attached to creator
         * task. The "AttachChildren"=TRUE condition is handled in
         * task_creation_handle().
         */
        vxdbg_ctx.ctxId = (VXDBG_CTX_ID)taskIdSelf();
        vxdbg_ctx.ctxType = VXDBG_CTX_TASK;

        if (!vxdbgIsAttached(VXDBG_CLIENT_ID, &vxdbg_ctx)) return;
    }

attach_new_task:
    vxdbg_ctx.ctxId = (VXDBG_CTX_ID)pTcb;
    vxdbg_ctx.ctxType = VXDBG_CTX_TASK;

    /* Attach VxDBG to this task: This will cause this task to be stopped */
    (void)vxdbgAttach(VXDBG_CLIENT_ID, &vxdbg_ctx, NULL);
}

#if defined(_WRS_CONFIG_RTP)
/*******************************************************************************
*
* rtp_initialize_context - initialize context for RTP
*
* This routine is invoked when the RTP init task is stopped due to a
* VxDBG attachment.
* The purpose of this handler is to:
* - Create a TCF context for new RTP and for its children tasks if debugger
*   is explicitly attached to RTPs folder, or if the creator context was
*   set up to automatically attach to its children.
* - If RTP has been created from Processes service:
*	- Set the attach_mode (StopAtMain/StopAtEntry).
*	- Reply to "Processes start" command.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void rtp_initialize_context(void * args) {
    RTP_ID rtpId = (RTP_ID)args;
    int err = 0;
    Context * rtp_ctx = NULL;
    NODE * pDll;
    RTP_SPAWN_REQ * pReq = NULL;
    Context * creator_ctx = NULL;
    RTP_ID creatorRtp = ((WIND_RTP *)rtpId)->parentRtpId;
    TASK_ID creatorTask = ((WIND_RTP *)rtpId)->parentTaskId;

    /* Only create TCF context if we're explicitly attached to RTPs folder */
    if ((rtps_ctx != NULL) && ((EXT(rtps_ctx)->attach_type & EXPLICIT_ATTACH) == EXPLICIT_ATTACH)) {
    
	/* Create TCF contexts for this RTP and children tasks */
	rtp_ctx = vx_create_context ((pid_t)rtpId, EXPLICIT_ATTACH, &err);
    }

    /*
     * Check if this RTP has been started from Processes service.
     * If yes, then update its attach_mode (StopAtMain/StopAtEntry).
     */
    semTake (rtpListLock, WAIT_FOREVER);	/* Lock RTP creation list */
    for (pDll = lstFirst (&rtpStartList); pDll != NULL; pDll = lstNext (pDll)) {
	RTP_CREATION_INFO * pInfo = (RTP_CREATION_INFO *)pDll;
	if (pInfo->rtp_id == rtpId) {

	    pReq = pInfo->pReq;

	    if ((rtp_ctx == NULL) && (err == 0) && (pReq->attach == 1)) {
		/*
		 * RTP has been created from Processes service with
		 * "Attach":true, this context must be attached even if
		 * parent folder "RTPs" is not attached.
		 */
		rtp_ctx = vx_create_context ((pid_t)rtpId, EXPLICIT_ATTACH, &err);
	    }

	    if (rtp_ctx != NULL) {
		/*
		 * This RTP context has been started from Processes service:
		 * Update its attach mode.
		 */
		EXT(rtp_ctx)->attach_mode = pReq->attach_mode;
		EXT(rtp_ctx)->created_by_agent = TRUE;
	    } else {
		/*
		 * Agent failed to attach to RTP context created with
		 * "Processes start" command. Report corresponding errno.
		 */
		pReq->err = err;
	    }

	    /* Remove/Free this entry */
	    lstDelete (&rtpStartList, &pInfo->chain);
	    loc_free (pInfo);
	    break;
	}
    }
    semGive (rtpListLock);	/* Unlock RTP creation list */

    /*
     * If RTP has been started from Processes service: it is time to
     * send "Processes start" reply.
     */

    if (pReq != NULL) {
	vx_rtp_create_reply (pReq);
    }

    if (rtp_ctx != NULL) return;

    /*
     * Check if RTP has been created by another RTP or by a task,
     * and check if a TCF context exists for the RTP creator.
     */
    if (creatorRtp != NULL) {
        /*
         * RTP has been created by another RTP: Check if debugger is attached
         * to the creator RTP.
         */
        creator_ctx = context_find_from_pid ((pid_t)creatorRtp, 1);
    }

    if ((creator_ctx == NULL) && (creatorTask != TASK_ID_NULL)) {
        /*
         * RTP has been created by a task: Check if debugger is attached to
         * this task.
         */
        creator_ctx = context_find_from_pid ((pid_t)creatorTask, 1);
    }

    /*
     * If debuger is attached to creator, check if its children must be
     * attached.
     */
    if ((creator_ctx != NULL) && ((EXT(creator_ctx)->attach_mode & CONTEXT_ATTACH_CHILDREN) == CONTEXT_ATTACH_CHILDREN)) {
        rtp_ctx = vx_create_context ((pid_t)rtpId, EXPLICIT_ATTACH, &err);

        /* new child context must inherit the children attachment flag */
        if (rtp_ctx != NULL) EXT(rtp_ctx)->attach_mode |= CONTEXT_ATTACH_CHILDREN;
    }
}

/* Read a RTP program header.
 * This routine is called whenever a RTP is loaded in order to read a 
 * RTP program header and register information that are not saved in the
 * RTP  structure (ELF information). Note that this routine cannot be
 * called from TCF dispatch thread in case the ELF module is accessed using
 * a hostfs device.
 */
static void vx_read_rtp_p_hdr(RTP_ID rtp_id) {
    unsigned pheader_cnt;
    ELF_PHeader * pheaders = NULL;
    ELF_PHeader * pheader;
    unsigned i;
    RTP_INFO * info;
    unsigned load_ix = 0;

    if (vx_elf_read_phdr(rtp_id->pPathName, &pheader_cnt, &pheaders) == -1) return;
    if (pheader_cnt == 0) return;
    info = loc_alloc_zero(sizeof(RTP_INFO));
    /* locate the corresponding program header */
    info->rtp_id = rtp_id;
    for (i = 0; i < pheader_cnt; ++i) {
	pheader = &pheaders[i];
	if (pheader->type != PT_LOAD) continue;
	if (load_ix == 0) {
	    info->text_offs = pheader->offset;
	}
	else if (load_ix == 1) {
	    info->data_offs = pheader->offset;
	}
	load_ix++;
    }
    loc_free(pheaders);
    lstInit (&info->shlList);
    semTake (rtpListLock, WAIT_FOREVER);
    lstAdd(&rtpMapList, &info->chain);
    semGive (rtpListLock);
}

/*******************************************************************************
*
* rtp_initialized_hook - RTP init complete hook
*
* This hook is invoked in the context of the RTP init task to notify the
* completion of the RTP initialization.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void rtp_initialized_hook (RTP_ID rtpId) {
    vx_read_rtp_p_hdr(rtpId);
    if (vxdbg_clnt_id != NULL) {
    VXDBG_CTX vxdbg_ctx;

    /*
     * When TCF agent is started, always attach and stop the init task of
     * all RTPs to handle various options. The RTP init task will be detached
     * and continued if needed.
     */

    vxdbg_ctx.ctxId = (VXDBG_CTX_ID) taskIdSelf();
    vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
    (void) vxdbgAttach(VXDBG_CLIENT_ID, &vxdbg_ctx, NULL);
    }
}

/*******************************************************************************
*
* rtp_delete_event - event handler for RTP deletion
*
* This event handler is invoked when RTP is deleted.
* The purpose of this handler is to:
* - Delete corresponding TCF context and its children contexts (RTP tasks).
* - Send "Processes exited" event.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void rtp_delete_event(void * args) {
    pid_t rtp_id = ((RTP_DELETE_INFO *)args)->rtp_id;
    int exit_code = ((RTP_DELETE_INFO *)args)->exit_code;
    Context * rtp_ctx = context_find_from_pid (rtp_id, 1);
    NODE * pDll;

    /*
     * Check if this RTP has been started with IO redirected.
     * If yes, then close corresponding file descriptors and ptys.
     */
    semTake (rtpListLock, WAIT_FOREVER);        /* Lock RTP list */
    for (pDll = lstFirst (&rtpIoList); pDll != NULL; pDll = lstNext (pDll)) {
        RTP_IO_INFO * pIoInfo = (RTP_IO_INFO *)pDll;
        if (pIoInfo->rtp_id == rtp_id) {
	    char pty_err[PTY_DEV_NAME_STR_SIZE];
	    char pty_in_out[PTY_DEV_NAME_STR_SIZE];
	    int bytes_in_out;
	    int bytes_err;

	    /*
	     * Check if the pseudo-terminals used for I/O redirection still
	     * contain un-handled data. If so, re-post the event to give a
	     * chance to read pseudo-terminal data.
	     */
	    if ((ioctl (pIoInfo->pipe_fds[0], FIONWRITE, (_Vx_ioctl_arg_t) &bytes_in_out) == OK) && (ioctl (pIoInfo->pipe_fds[1], FIONWRITE, (_Vx_ioctl_arg_t) &bytes_err) == OK)) {
		if ((bytes_in_out != 0) || (bytes_err != 0)) {
		    semGive (rtpListLock);
		    post_event_with_delay(rtp_delete_event, args, 10000);
		    return;
		}
	    }

	    if (pIoInfo->pipe_fds[0] != -1) (void) close (pIoInfo->pipe_fds[0]);
	    if (pIoInfo->pipe_fds[1] != -1) (void) close (pIoInfo->pipe_fds[1]);
	    snprintf(pty_err, sizeof(pty_err)-1, "%sErr", pIoInfo->pipe_name);
	    snprintf(pty_in_out, sizeof(pty_in_out)-1, "%sInOut", pIoInfo->pipe_name);

	    (void) ptyDevRemove (pty_in_out);
	    (void) ptyDevRemove (pty_err);

	    loc_free (pIoInfo->pipe_name);

            /* Remove/Free this entry */
            lstDelete (&rtpIoList, &pIoInfo->chain);
            loc_free (pIoInfo);
            break;
        }
    }
    semGive (rtpListLock);      /* Unlock RTP list */

    if (rtp_ctx != NULL) vx_delete_context (rtp_ctx, 1);

    start_process_send_exited_event (rtp_id, exit_code);

    loc_free (args);
}

/*******************************************************************************
*
* vxdbg_rtp_delete_hook - RTP delete hook
*
* Job is defered to TCF. See rtp_delete_event().
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vxdbg_rtp_delete_hook (void * pEvtData) {
    RTP_DELETE_INFO * pInfo;

    if (pEvtData == NULL) return;

    if ((pInfo = loc_alloc_zero(sizeof(RTP_DELETE_INFO))) == NULL) return;

    memcpy (pInfo, pEvtData, sizeof (RTP_DELETE_INFO));
    free (pEvtData);

    post_event(rtp_delete_event, pInfo);
}

/*******************************************************************************
*
* rtp_delete_hook - RTP deletion hook
*
* This hook is invoked to notify an RTP deletion.
* NOTE: Calling post_event() in this hook to defer rtp_delete_event() to
* TCF, would result in converting calling task (RTP init task) into a POSIX
* thread. To avoid this, we do inject a VxDBG event to defer post_event() call
* to VxDBG event task (Only the VxDBG event task will become a pthread).
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/
static void rtp_delete_hook (const RTP_ID rtpId, const int exitCode) {
    NODE * pDll;
    RTP_INFO * pInfo;
    semTake (rtpListLock, WAIT_FOREVER);
    pInfo = find_rtp_info(rtpId);
    if (pInfo != NULL) {
	for (pDll = lstFirst(&pInfo->shlList); pDll != NULL; ) {
	    SHL_MAP_INFO * pShlInfo = (SHL_MAP_INFO *)pDll;
	    pDll = lstNext(pDll);
	    lstDelete (&pInfo->shlList, &pShlInfo->chain);
	    loc_free(pShlInfo->name);
	    loc_free(pShlInfo);
	}
	lstDelete (&rtpMapList, &pInfo->chain);
	loc_free (pInfo);
    }
    semGive (rtpListLock);
    if (vxdbg_clnt_id != NULL) {
	RTP_DELETE_INFO * pInfo;
	if ((pInfo = (RTP_DELETE_INFO *)calloc (1, sizeof (RTP_DELETE_INFO))) == NULL) return;
	pInfo->rtp_id = (pid_t)rtpId;
	pInfo->exit_code = (int)exitCode;

	vxdbgEventInject(NULL, vxdbg_rtp_delete_hook, pInfo);
    }
}

/* 
 * Shared Library opened hook
 *
 * This shared library opened hook is called when the task opening
 * the shared library is stopped. This will allow planting breakpoints in 
 * the shared library before it is used.
 */
static void shl_opened_event_safe(void * args) {
    Context * rtp_ctx;
    unsigned i;
    TASK_ID tid = (TASK_ID) args;
    WIND_TCB * pTcb = (WIND_TCB *)args;
    pid_t pid;

    if (taskIdVerify(tid) != OK) return;

    assert (taskIsSuspended(tid));

    pid = (pid_t) pTcb->rtpId;
    rtp_ctx = context_find_from_pid(pid, 1);
    if (rtp_ctx != NULL) memory_map_event_mapping_changed (rtp_ctx);
done:
    taskResume(tid);
}

/* 
 * Shared Library opened hook
 *
 * This shared library opened hook is called with the task opening
 * the shared library about to be suspended. It will first wait for this
 * task to be suspended, and will post a safe event (event with the 
 * specified context stopped) to cause a refresh of memory mapping. 
 * Resuming of the task is deferred to this job (that will have first stop
 * the task).
 */
static void shl_opened_event(void * args) {
    Context * ctx;
    unsigned i;
    TASK_ID tid = (TASK_ID) args;
    WIND_TCB * pTcb = (WIND_TCB *)args;
    pid_t pid;
   
    if (taskIdVerify(tid) != OK) return;

    pid = (pid_t) tid;
    for (i = 0; i < 20; i ++) {
        if (taskIsSuspended(tid)) break;
	taskDelay (1);
    }
    if (!taskIsSuspended(tid)) return;

    ctx = context_find_from_pid((pid_t)tid, 1);
    if (ctx == NULL) ctx = context_find_from_pid((pid_t)pTcb->rtpId, 1);
    if (ctx != NULL) post_safe_event(ctx, shl_opened_event_safe, pTcb);
    else taskResume(tid);
}

/* Event handler to handle closing of a shared library */
static void shl_closed_event_safe(void * args) {
    Context * rtp_ctx = (Context *) args;;
    memory_map_event_mapping_changed (rtp_ctx);
}

static void shl_closed_event(void * args) {
    Context * rtp_ctx;
    unsigned i;
    RTP_ID rtp_id = (RTP_ID) args;
    pid_t pid;
   
    if (_func_rtpIdVerify == NULL || _func_rtpIdVerify(rtp_id) != OK) return;

    pid = (pid_t) rtp_id;
    rtp_ctx = context_find_from_pid(pid, 1);
    if (rtp_ctx == NULL) return;
    post_safe_event(rtp_ctx, shl_closed_event_safe, rtp_ctx);
}

/*******************************************************************************
*
* rtp_shl_open_hook - RTP Shared Library Open Hook
*
* This hook is invoked to notify when a shared library is opened in a RTP
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/
static void rtp_shl_open_hook (const SHL_ID shlId, const RTP_ID rtpId) {
    NODE * pDll = NULL;
    RTP_SHL_INFO shl_info;
    unsigned pheader_cnt;
    ELF_PHeader * pheaders = NULL;
    ELF_PHeader * pheader;
    RTP_INFO * rtp_info;
    SHL_MAP_INFO * shl_map_info = NULL;
    unsigned i;
    unsigned load_ix = 0;
    uint64_t data_mem_size = 0;
    TASK_ID tid = taskIdSelf();
    WIND_TCB * pTcb = (WIND_TCB *)tid;;
    int taskOptions;

    /* call previous hooks */
    if (_func_rtpOldShlOpenHook != NULL) _func_rtpOldShlOpenHook(shlId, rtpId);

    assert (pTcb->rtpId == rtpId);
    if (pTcb->rtpId != rtpId) return;

    if (_func_rtpShlInfoGet (shlId, rtpId, &shl_info) != OK) return;
    if (vx_elf_read_phdr(shl_info.fileName, &pheader_cnt, &pheaders) == -1 || pheader_cnt == 0) {
    	free(shl_info.fileName);
	return;
    }

    SET_TASK_UNBREAKABLE_OPTION(tid,taskOptions)
    semTake (rtpListLock, WAIT_FOREVER);
    rtp_info = find_rtp_info(rtpId);
    if (rtp_info == NULL) {
	semGive (rtpListLock);
	RESET_TASK_UNBREAKABLE_OPTION(tid,taskOptions);
	free (shl_info.fileName);
	return;
    }

    shl_map_info = loc_alloc_zero(sizeof(SHL_MAP_INFO));
    shl_map_info->name = loc_strdup(shl_info.fileName);
    shl_map_info->shl_id = shlId;
    free (shl_info.fileName);

    for (i = 0; i < pheader_cnt; ++i) {
	pheader = &pheaders[i];
	if (pheader->type != PT_LOAD) continue;
	if (load_ix == 0) {
	    shl_map_info->text_offs = pheader->offset;
	    shl_map_info->text_size = pheader->mem_size;
	}
	else if (load_ix == 1) {
	    shl_map_info->data_offs = pheader->offset;
	    shl_map_info->data_size = pheader->file_size;
	    data_mem_size = pheader->mem_size;
	}
	load_ix++;
    }
    loc_free(pheaders);

    if (data_mem_size == 0) {
	semGive (rtpListLock);
	RESET_TASK_UNBREAKABLE_OPTION(tid,taskOptions);
        loc_free(shl_map_info->name);
        loc_free(shl_map_info);
    	return;
    }
    shl_map_info->text_addr = shl_info.textBase;
    shl_map_info->data_addr = shl_info.dataBase;
    shl_map_info->bss_addr = shl_map_info->data_addr + shl_map_info->data_size;
    shl_map_info->bss_size = data_mem_size - shl_map_info->data_size;
    lstAdd(&rtp_info->shlList, &shl_map_info->chain);
    semGive (rtpListLock);
    RESET_TASK_UNBREAKABLE_OPTION(tid,taskOptions);
    if (rtp_info->attached) {
    	VXDBG_POST_EVENT(shl_opened_event, pTcb, NULL);

	/* Make sure execution of process is not resumed before we have 
	 * handled load of shared libraries (i.e. planting of breakpoints).
	 */
	taskSuspend(0);
    }
}

/*******************************************************************************
*
* rtp_shl_close_hook - RTP Shared Library Close Hook
*
* This hook is invoked to notify when a shared library is closed in a RTP
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/
static void rtp_shl_close_hook (const SHL_ID shlId, const RTP_ID rtpId) {
    NODE * pDll = NULL;
    RTP_INFO * rtp_info;
    SHL_MAP_INFO * shl_map_info = NULL;
    TASK_ID tid = taskIdSelf();
    WIND_TCB * pTcb = (WIND_TCB *)tid;;
    int taskOptions;

    if (_func_rtpOldShlCloseHook != NULL) _func_rtpOldShlCloseHook(shlId, rtpId);

    assert (pTcb->rtpId == rtpId);
    if (pTcb->rtpId != rtpId) return;

    SET_TASK_UNBREAKABLE_OPTION(tid,taskOptions)
    semTake (rtpListLock, WAIT_FOREVER);
    rtp_info = find_rtp_info(rtpId);
    if (rtp_info == NULL) {
	semGive (rtpListLock);
	RESET_TASK_UNBREAKABLE_OPTION(tid,taskOptions);
	return;
    }
    for (pDll = lstFirst(&rtp_info->shlList); pDll != NULL; ) {
	SHL_MAP_INFO * shl_info = (SHL_MAP_INFO *)pDll;
	pDll = lstNext(pDll);
	if (shl_info->shl_id != shlId) continue;
	lstDelete (&rtp_info->shlList, &shl_info->chain);
	loc_free(shl_info->name);
	loc_free(shl_info);
	break;
    }
    semGive (rtpListLock);
    RESET_TASK_UNBREAKABLE_OPTION(tid,taskOptions);
    if (rtp_info->attached) VXDBG_POST_EVENT(shl_closed_event, rtpId, NULL);
}

#endif	/* _WRS_CONFIG_RTP */

int context_get_memory_map(Context * ctx, MemoryMap * map) {
    BOOT_PARAMS bootParams;	/* VxWorks boot parameters */
    char * pModule;

    if (ctx == NULL) return 0;

    if (ctx == kernel_ctx) {
	/* Kernel context: register DKMs and VxWorks kernel image */

	moduleEach(module_list_proc, (_Vx_usr_arg_t)map);

	/* Add VxWorks module (From bootline) */
	bootStringToStruct (sysBootLine, &bootParams);

	pModule = bootParams.bootFile;

	add_map_region(map, 0, 0, 0 , 0, pModule, 0, 0);

	return 0;
    }

#if defined(_WRS_CONFIG_RTP)
    if (ctx->parent == rtps_ctx) {
    	/* RTP context: register RTP image segments */

	RTP_ID rtpId = (RTP_ID)EXT(ctx)->pid;
	RTP_INFO * pInfo;
	NODE * pDll;
	MemoryRegion * r;

	semTake (rtpListLock, WAIT_FOREVER);
	pInfo = find_rtp_info(rtpId);
	if (pInfo != NULL) {
	    r = add_map_region(map, (ContextAddress) rtpId->binaryInfo.textStart, (ContextAddress) rtpId->binaryInfo.textSize, pInfo->text_offs, MM_FLAG_R | MM_FLAG_X, rtpId->pPathName, NULL, 0);
	    add_map_int_attr(r, MODULE_ID_KEY, (long)rtpId);
	    r = add_map_region(map, (ContextAddress) rtpId->binaryInfo.dataStart, (ContextAddress) rtpId->binaryInfo.dataSize, pInfo->data_offs, MM_FLAG_R | MM_FLAG_W, rtpId->pPathName, NULL, 0);
	    add_map_int_attr(r, MODULE_ID_KEY, (long)rtpId);
	    r = add_map_region(map, (ContextAddress) rtpId->binaryInfo.bssStart, (ContextAddress) rtpId->binaryInfo.bssSize, pInfo->data_offs + (ContextAddress) rtpId->binaryInfo.dataSize, MM_FLAG_R | MM_FLAG_W, rtpId->pPathName, NULL, 1);
	    add_map_int_attr(r, MODULE_ID_KEY, (long)rtpId);

	    for (pDll = lstFirst(&pInfo->shlList); pDll != NULL; pDll = lstNext(pDll)) {
		SHL_MAP_INFO * pShlInfo = (SHL_MAP_INFO *)pDll;
		r = add_map_region(map, (ContextAddress) pShlInfo->text_addr, (ContextAddress) pShlInfo->text_size, pShlInfo->text_offs, MM_FLAG_R | MM_FLAG_X, pShlInfo->name, NULL, 0);
		add_map_int_attr(r, MODULE_ID_KEY, (long)pShlInfo->shl_id);
		r = add_map_region(map, (ContextAddress) pShlInfo->data_addr, (ContextAddress) pShlInfo->data_size, pShlInfo->data_offs, MM_FLAG_R | MM_FLAG_W, pShlInfo->name, NULL, 0);
		add_map_int_attr(r, MODULE_ID_KEY, (long)pShlInfo->shl_id);
		r = add_map_region(map, (ContextAddress) (ContextAddress) pShlInfo->bss_addr, (ContextAddress) pShlInfo->bss_size, pShlInfo->data_offs + (ContextAddress) pShlInfo->data_size, MM_FLAG_R | MM_FLAG_W, pShlInfo->name, NULL, 1);
		add_map_int_attr(r, MODULE_ID_KEY, (long)pShlInfo->shl_id);
	    }
	}
	semGive (rtpListLock);
    }
#endif

    return 0;
}

#if SERVICE_PathMap
static void event_path_map_changed(Channel * c, void * args) {
    if (kernel_ctx != NULL) {
        memory_map_event_mapping_changed (kernel_ctx);
        send_context_changed_event (kernel_ctx);
    }
}
#endif

unsigned context_word_size(Context * ctx) {
    return sizeof(void *);
}

/*******************************************************************************
*
* get_context_task_id - get VxWorks id from given context
*
* This routine returns the VxWorks ID associated to given TCF context.
*
* RETURNS: VxWorks object ID (Could be a fake ID created for system containers).
*
* ERRNO: N/A
*
*/

pid_t get_context_task_id(Context * ctx) {
    return EXT(ctx)->pid;
}

/*******************************************************************************
*
* eventpoint_at_main - StopAtMain eventpoint handler
*
* This "StopAtMain" eventpoint handler suspends the task context which hit the
* eventpoint, and then removes the eventpoint.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void eventpoint_at_main(Context * ctx, void * args) {
    EXT(ctx)->event = STOP_AT_EVENTPOINT;
    send_context_changed_event(ctx->mem);
    memory_map_event_mapping_changed(ctx->mem);
    suspend_debug_context(ctx);
    destroy_eventpoint (EXT(ctx)->main_ep);
}

/*******************************************************************************
*
* eventpoint_at_entry - StopAtEntry eventpoint handler
*
* This "StopAtEntry" eventpoint handler suspends the task context which hit the
* eventpoint, and then removes the eventpoint.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void eventpoint_at_entry(Context * ctx, void * args) {
    EXT(ctx)->event = STOP_AT_EVENTPOINT;
    send_context_changed_event(ctx->mem);
    memory_map_event_mapping_changed(ctx->mem);
    suspend_debug_context(ctx);
    destroy_eventpoint (EXT(ctx)->entry_ep);
}

/*******************************************************************************
*
* task_start_opts_handle - handle task startup options
*
* This routine is invoked when tasks are stopped.
* This routine handles "StopAtMain" & "StopAtEntry" options for kernel tasks
* created with Processes service (Processes start command). It uses eventpoints
* to stop tasks on their start routine (the entry point specified via
* "StartSymbol").
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void task_start_opts_handle(Context * ctx, int attach_mode) {

    /* Must ignore tasks not created by debug agent */
    if (!EXT(ctx)->created_by_agent) return;

    /* This code must only be excuted once per task */

    if (CTX_START_OPTS_HANDLED(ctx)) return;

    /* Handle "StopAtMain":true */

    if ((attach_mode & CONTEXT_ATTACH_NO_MAIN) == 0) {
	char entry_addr_str[22];	/* Main address in decimal */

        sprintf (entry_addr_str, "%lu", (unsigned long)EXT(ctx)->task_start_address);
	EXT(ctx)->main_ep = create_eventpoint(entry_addr_str, ctx, eventpoint_at_main, NULL);
    }

    /* Handle "StopAtEntry":true */

    if ((attach_mode & CONTEXT_ATTACH_NO_STOP) == 0) {
	/*
	 * Task is already stopped on vxTaskEntry: Task has been created, but
	 * not activated. The VxDBG attachment did stop the task, which caused
	 * this call of this hook.
	 * We simply need to indicate that this stop event must be intercepted.
	 */
    	ctx->pending_intercept = 1;
    } else {
    	ctx->pending_intercept = 0;
    }

    /* Task can now be activated */

    taskActivate((TASK_ID)EXT(ctx)->pid);

    EXT(ctx)->properties |= START_OPTS_HANDLED;
}

#if defined(_WRS_CONFIG_RTP)
/*******************************************************************************
*
* rtp_start_opts_handle - handle RTP startup options
*
* This routine handles "StopAtMain" & "StopAtEntry" options for RTPs created
* with Processes service (Processes start command). It uses eventpoints to stop
* the RTP init tasks on either "main" symbol for "StopAtMain" option, or the
* RTP entry point (usually "_start" symbol) for "StopAtEntry" option.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void rtp_start_opts_handle(Context * ctx, int attach_mode) {

    /* This code must only be excuted once per RTP */

    if (CTX_START_OPTS_HANDLED(ctx->mem)) return;

    /* Handle "StopAtMain":true */

    if ((attach_mode & CONTEXT_ATTACH_NO_MAIN) == 0) {
	EXT(ctx)->main_ep = create_eventpoint("main", ctx, eventpoint_at_main, NULL);
    }

    /* Handle "StopAtEntry":true */

    if ((attach_mode & CONTEXT_ATTACH_NO_STOP) == 0) {
	void * entry_addr = ((WIND_RTP *)(EXT(ctx->mem)->pid))->entrAddr;
	char entry_addr_str[22];	/* Entry address in decimal */

	sprintf (entry_addr_str, "%lu", (unsigned long)entry_addr);
	EXT(ctx)->entry_ep = create_eventpoint(entry_addr_str, ctx, eventpoint_at_entry, NULL);
    }

    EXT(ctx->mem)->properties |= START_OPTS_HANDLED;
}
#endif	/* _WRS_CONFIG_RTP */


/*******************************************************************************
*
* task_creation_handle - Handle VxWorks task creation
*
* This routine is invoked each time a VxWorks task is created.
* The purpose of this routine is to handle the attachment of this new task.
* If this task has been created from the debug agent (Using Processes start),
* then it creates a TCF context for the new task and record the attachment
* properties for later handling of StopAtMain & StopAtEntry.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void task_creation_handle (TASK_ID creatorId, TASK_ID taskId) {
    NODE * pDll;
    Context * task_ctx;
    BOOL task_created_by_agent = FALSE; /* Task created from Processes service? */
    int err = 0;

    /* Search <taskId> in the list of tasks created from Processes service */

    semTake (taskStartListLock, WAIT_FOREVER);	/* Lock task creation list */
    for (pDll = lstFirst (&taskStartList); pDll != NULL; pDll = lstNext (pDll)) {
	TASK_CREATION_INFO * pInfo = (TASK_CREATION_INFO *)pDll;
	if (pInfo->tid == taskId) {
	    task_created_by_agent = TRUE; /* created from Processes service */
	    task_ctx = vx_create_context ((pid_t)taskId, EXPLICIT_ATTACH, &err);

	    if (task_ctx != NULL) {
		/*
		 * Record the attachment properties for later handling
		 * of StopAtMain & StopAtEntry.
		 */
		EXT(task_ctx)->created_by_agent = TRUE;
		EXT(task_ctx)->attach_mode = pInfo->attach_mode;
		EXT(task_ctx)->task_start_address = pInfo->start_address;

		/* Send reply to "Processes start" command */
		if (pInfo->attach_cb != NULL && pInfo->attach_cb_data != NULL) {
		    pInfo->attach_cb (err, task_ctx, pInfo->attach_cb_data);
		}
	    }

	    /* Remove/Free this entry */
	    lstDelete (&taskStartList, &pInfo->chain);
	    loc_free (pInfo);
	    break;
	}
    }
    semGive (taskStartListLock);	/* Unlock task creation list */

    if (!task_created_by_agent) {
        Context * creator_ctx = context_find_from_pid ((pid_t)creatorId, 1);
        Context * new_ctx;

    	/*
	 * This task has not been created from Processes service:
	 * Only attach to this task if parent folder is explicitly attached,
         * or if creator task has CONTEXT_ATTACH_CHILDREN flag.
	 */
	
        if ((creator_ctx != NULL) && ((EXT(creator_ctx)->attach_mode & CONTEXT_ATTACH_CHILDREN) == CONTEXT_ATTACH_CHILDREN)) {
            /*
             * Debugger is attached to creator task, and this task has
             * CONTEXT_ATTACH_CHILDREN flag: We must attach to this new task.
             */

            new_ctx = vx_create_context ((pid_t)taskId, EXPLICIT_ATTACH, &err);

            /* new child context must inherit the children attachment flag */
            if (new_ctx != NULL) EXT(new_ctx)->attach_mode |= CONTEXT_ATTACH_CHILDREN;
        } else {
            /* Attach to new task if parent folder is explicitly attached. */
            vx_create_context ((pid_t)taskId, COND_PARENT_ATTACH, &err);
        }
    }
}

/*******************************************************************************
*
* event_handler - handle events posted from VxDBG event task
*
* This routine handles the events posted with post_event() from the VxDBG
* event task.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void event_handler(void * arg) {
    struct event_info * info = (struct event_info *)arg;
    int err;
#if defined(_WRS_CONFIG_RTP)
    pid_t rtpId;
    REG_SET rtpTaskRegSet;
#endif
    Context * current_ctx = context_find_from_pid(info->current_ctx, 1);
    Context * stopped_ctx = context_find_from_pid(info->stopped_ctx, 1);

    if (stopped_ctx != NULL) {
        if (CTX_IS_RTP(stopped_ctx)) {
            /*
             * stopped_ctx is an RTP context: This means that we've hit a
             * breakpoint installed with BP_ACTION_ALL_STOP (BP set from
             * kernel shell with bp -a).
             * In this specific case, we do not need to care about the
             * RTP context, but only about the task context which hit
             * the breakpoint: current_ctx.
             * The other stopped tasks of the RTP will be handled thanks
             * EVENT_HOOK_STOP handler.
             */
            assert (current_ctx == NULL);
            assert (current_ctx->parent != stopped_ctx);
            stopped_ctx = current_ctx; /* Stopped task context is current_ctx */
        }

        CLEAR_STOP_REASON(stopped_ctx);

#if defined(_WRS_CONFIG_RTP)
        if (info->event != EVENT_HOOK_TASK_DELETE) {
            /*
             * For RTP tasks, invoke rtp_sc_check() to set/unset IS_IN_SYSCALL
             * flag, and retrieve RTP task register set (At system call level if
             * task is stopped in system call).
             */
            if ((taskIdVerify ((TASK_ID)(EXT(stopped_ctx)->pid)) == OK) && CTX_IS_RTP(stopped_ctx->parent)) {
                (void)rtp_sc_check(stopped_ctx, &rtpTaskRegSet, NULL);
            }
        }
#endif
    }

    switch (info->event) {
    case EVENT_HOOK_BREAKPOINT:
        if (stopped_ctx == NULL) break;
        assert(!stopped_ctx->stopped);
        assert(!EXT(stopped_ctx)->regs_dirty);
        if (EXT(stopped_ctx)->regs_error) {
            release_error_report(EXT(stopped_ctx)->regs_error);
            EXT(stopped_ctx)->regs_error = NULL;
        }

        memcpy(EXT(stopped_ctx)->regs, &info->regs, sizeof(REG_SET));
        stopped_ctx->stopped = 1;
        assert(get_regs_PC(stopped_ctx) == info->addr);

	if (info->triggeredCb) {
	    if (triggeredCb == NULL) {
		/* No array for triggered CB: Allocate one */
		triggeredCb = loc_alloc_zero((info->triggeredCbCount + 1) * sizeof (ContextBreakpoint *));
		triggeredCbMax = info->triggeredCbCount;
	    } else if (info->triggeredCbCount > triggeredCbMax) {
		/* triggered CB breakpoint array too small: reallocated it */
		triggeredCb = loc_realloc (triggeredCb, (info->triggeredCbCount + 1) * sizeof (ContextBreakpoint *));
		triggeredCbMax = info->triggeredCbCount;
	    }

	    memcpy (triggeredCb, info->triggeredCb, (info->triggeredCbCount + 1) * sizeof (ContextBreakpoint *));
	    stopped_ctx->stopped_by_cb = triggeredCb;
	    loc_free(info->triggeredCb);
	} else {
	    stopped_ctx->stopped_by_bp = 1;
	}

        if (stopped_ctx->stopped_by_bp && !is_breakpoint_address(stopped_ctx, (ContextAddress)info->addr)) {
            /* Break instruction that is not planted by us */
	    stopped_ctx->signal = SIGTRAP;
            stopped_ctx->stopped_by_bp = 0;
            stopped_ctx->pending_intercept = 1;
        }
        if (current_ctx != NULL) EXT(stopped_ctx)->bp_pid = EXT(current_ctx)->pid;
        assert(taskIsStopped(EXT(stopped_ctx)->pid));
        trace(LOG_CONTEXT, "context: stopped by breakpoint: ctx %#lx, id %#lx",
                stopped_ctx, EXT(stopped_ctx)->pid);
        NOTIFY_CONTEXT_STOP(stopped_ctx);
        break;

    case EVENT_HOOK_STEP_DONE:
        if (current_ctx == NULL) break;
        assert(!current_ctx->stopped);
        assert(!EXT(current_ctx)->regs_dirty);
        if (EXT(current_ctx)->regs_error) {
            release_error_report(EXT(current_ctx)->regs_error);
            EXT(current_ctx)->regs_error = NULL;
        }
        memcpy(EXT(current_ctx)->regs, &info->regs, sizeof(REG_SET));
        EXT(current_ctx)->event = TRACE_EVENT_STEP;
        current_ctx->signal = SIGTRAP;
        current_ctx->stopped = 1;

	if (!(EXT(current_ctx)->no_auto_pending_intercept)) {
            /*
             * This step request does not come from the debug agent, but from
             * another client (like kernel shell): It must be intercepted.
             */
            current_ctx->pending_intercept = 1;
        }

        assert(taskIsStopped(EXT(current_ctx)->pid));
        trace(LOG_CONTEXT, "context: stopped by end of step: ctx %#lx, id %#lx",
                current_ctx, EXT(current_ctx)->pid);
        NOTIFY_CONTEXT_STOP(current_ctx);
        break;

    case EVENT_HOOK_STOP:
	if (taskIdVerify ((TASK_ID)(info->stopped_ctx)) != OK) break;
	if (stopped_ctx != NULL) {
	    TASK_ID tid = (TASK_ID)(info->stopped_ctx);
	    if (EXT(stopped_ctx)->no_auto_pending_intercept &&
	    			TASK_IS_SAFE(taskTcb(tid))) {
                /*
                 * We've received from debugger a suspend request on a task
                 * which is in delete safe state. As suspending a task holding
                 * a critical system resource is very dangerous and may block
                 * the debug agent itself if it needs this critical resource,
                 * we resume that context during 1 tick to let it a change to
                 * exit this critical section, and stop it again.
                 * We repeat this until the task is no longer in delete safe
                 * state, with a maximum of 10 retries.
                 */
		VXDBG_CTX vxdbg_ctx;
		vxdbg_ctx.ctxId = (VXDBG_CTX_ID) tid;
		vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
	        if (EXT(stopped_ctx)->stop_retry_cnt++ < 10) {
                    /*
                     * Before resuming this context, we must reset the
                     * IS_IN_SYSCALL flag as send_context_changed_event()
                     * (invoked via NOTIFY_CONTEXT_STOP()) won't be called
                     * in this specific case.
                     */

                    EXT(stopped_ctx)->properties &= ~IS_IN_SYSCALL;

                    (void) vxdbgCont(VXDBG_CLIENT_ID, &vxdbg_ctx);
                    taskDelay(1);
                    (void) vxdbgStop(VXDBG_CLIENT_ID, &vxdbg_ctx, FALSE);
                    return;
                }
	    }
	}
#if defined(_WRS_CONFIG_RTP)
        if (stopped_ctx == NULL) {
            /*
             * If the TCF context does not exist yet, the stop event may
             * have been generated by the VxDBG attachment of a RTP init
             * task.
             *
             * Check if it's an RTP init task. If so, create a TCF context
             * for this new RTP, and for its children tasks. If the RTP has
             * been created from Processes service then handles "create
             * options" and sends the "Processes start" reply.
             */
	    rtpId = (pid_t)(((WIND_TCB *)(info->stopped_ctx))->rtpId);
	    if (vx_verify_rtp_id ((RTP_ID)rtpId) != OK) break;

            if ((rtpId != (pid_t)kernelIdGet()) && (context_find_from_pid(rtpId, 1) == NULL)) rtp_initialize_context ((void *)rtpId);
            /*
             * If no TCF context has been created (either the RTP has been created
             * from the target shell or from the debugger without attachment), then detaches
             * the RTP init task.
             */
            if ((stopped_ctx = context_find_from_pid(info->stopped_ctx, 1)) == NULL) {
                VXDBG_CTX vxdbg_ctx;
                vxdbg_ctx.ctxId = (VXDBG_CTX_ID) info->stopped_ctx;
                vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
                (void) vxdbgDetach(VXDBG_CLIENT_ID, &vxdbg_ctx);
                break;
            }
        }
#else
        if (stopped_ctx == NULL) break;
#endif
        assert(!stopped_ctx->exited);
        if (stopped_ctx->stopped) break;
        if (EXT(stopped_ctx)->regs_error) {
            release_error_report(EXT(stopped_ctx)->regs_error);
            EXT(stopped_ctx)->regs_error = NULL;
        }

        /* Retrive register set for stopped task */

#if defined(_WRS_CONFIG_RTP)
        /*
         * For RTP tasks, simply copy the register set retrieved earlier with
         * rtp_sc_check().
         */
        if (CTX_IS_RTP(stopped_ctx->parent)) {
            memcpy (EXT(stopped_ctx)->regs, &rtpTaskRegSet, sizeof (REG_SET));
        } else
#endif

        if (taskRegsGet((TASK_ID)(EXT(stopped_ctx)->pid), EXT(stopped_ctx)->regs) != OK) {
            EXT(stopped_ctx)->regs_error = get_error_report(errno);
            assert(EXT(stopped_ctx)->regs_error != NULL);
        }
        stopped_ctx->stopped = 1;
	
	if (EXT(stopped_ctx)->no_auto_pending_intercept) {
	    /*
	     * The automatic "pending_intercept" handling is disabled:
	     * This stop event was expected: It must not be intercepted.
	     */
	    EXT(stopped_ctx)->no_auto_pending_intercept = FALSE;
	} else {
	    if (!EXT(stopped_ctx)->stopped_on_error) {
		stopped_ctx->signal = SIGSTOP;
	    }
	    stopped_ctx->pending_intercept = 1;
	}

	/*
	 * Handle Task & RTP startup options: StopAtMain/StopAtEntry.
	 */

	if (!CTX_IS_RTP(stopped_ctx->mem)) {
	    task_start_opts_handle(stopped_ctx, EXT(stopped_ctx)->attach_mode);
#if defined(_WRS_CONFIG_RTP)
	} else {
	    if (EXT(stopped_ctx->parent)->created_by_agent) rtp_start_opts_handle(stopped_ctx, EXT(stopped_ctx->mem)->attach_mode);
#endif	/* _WRS_CONFIG_RTP */
	}

        assert(taskIsStopped(EXT(stopped_ctx)->pid));
        trace(LOG_CONTEXT, "context: stopped by sofware request: ctx %#lx, id %#lx",
                stopped_ctx, EXT(stopped_ctx)->pid);
        NOTIFY_CONTEXT_STOP(stopped_ctx);
        break;

    case EVENT_HOOK_START:
	if (current_ctx == NULL || taskIdVerify ((TASK_ID)(info->current_ctx)) != OK) break;
	if (current_ctx->stopped == 0) break;

        trace(LOG_CONTEXT, "context: started by sofware request: ctx %#lx, id %#lx",
                current_ctx, EXT(current_ctx)->pid);
        current_ctx->pending_intercept = 0;
        send_context_started_event(current_ctx);
        break;

    case EVENT_HOOK_TASK_ADD:
	if (taskIdVerify ((TASK_ID)(info->stopped_ctx)) != OK) break;
#if defined(_WRS_CONFIG_RTP)
	/* Test if new task is an RTP init task */
	rtpId = (pid_t)(((WIND_TCB *)(info->stopped_ctx))->rtpId);
	if (vx_verify_rtp_id ((RTP_ID)rtpId) != OK) break;

	if ((rtpId != (pid_t)kernelIdGet()) && (context_find_from_pid(rtpId, 1) == NULL)) {
	    /*
	     * This task is an RTP task, and no TCF context has yet been
	     * created for its parent. Ignore this task, it will be handled
	     * by the rtp_initialize_context().
	     */
	    break;
	} else
#endif
	task_creation_handle((TASK_ID)info->current_ctx, (TASK_ID)info->stopped_ctx);
	break;

    case EVENT_HOOK_TASK_DELETE:
        if (stopped_ctx != NULL) {
            trace(LOG_CONTEXT, "context: exited ctx %#lx, id %#lx",
                  stopped_ctx, EXT(stopped_ctx)->pid);
            vx_delete_context (stopped_ctx, 1);
        }
	break;

    case EVENT_HOOK_EXCEPTION:
	if (taskIdVerify ((TASK_ID)(EXT(stopped_ctx)->pid)) != OK) break;
        while (!taskIsStopped ((TASK_ID)(EXT(stopped_ctx)->pid)))
	    taskDelay (1);
		
	memcpy(EXT(stopped_ctx)->regs, &info->regs, sizeof(REG_SET));
        stopped_ctx->signal = info->signal;
        stopped_ctx->stopped = 1;
        stopped_ctx->stopped_by_exception = 1;
	sigset_set(&stopped_ctx->pending_signals, info->signal, 1);

	if (sigset_get(&stopped_ctx->sig_dont_stop, info->signal) == 0) stopped_ctx->pending_intercept = 1;

        NOTIFY_CONTEXT_STOP(stopped_ctx);
        break;

    default:
        assert(0);
        break;
    }

    loc_free(info);
}

/*******************************************************************************
*
* vxdbg_evt_bp_hook - VxDBG EVT_BP notification hook
*
* This routine is invoked by VxDBG for breakpoint exception notification.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vxdbg_evt_bp_hook(VXDBG_CTX_ID ctx_id, VXDBG_CTX_ID stopped_ctx_id, REG_SET * pRegs, void * addr, VXDBG_BP_ID * bp_id) {
    struct event_info * info = loc_alloc_zero(sizeof(struct event_info));
    int bpCount = 0;
    ContextBreakpoint * cb = NULL;

    trace(LOG_CONTEXT, "vxdbg_evt_bp_hook ctx_id=%#lx / stopped_ctx_id=%#lx", ctx_id, stopped_ctx_id);

    info->event = EVENT_HOOK_BREAKPOINT;
    info->current_ctx = ctx_id;
    info->stopped_ctx = stopped_ctx_id;
    info->regs = *pRegs;
    info->addr = addr;
    info->triggeredCb = NULL;
    info->triggeredCbCount = 0;

    for (bpCount = 0; bp_id[bpCount] != 0; bpCount++);

    /* Allocate Array for triggered breakpoints (NULL terminated array) */

    if (bpCount != 0) {
	info->triggeredCb = loc_alloc_zero(sizeof(ContextBreakpoint *) * (bpCount + 1));
    }

    /* For each VXDBG_BP_ID, retrieve associated TCF ContextBreakpoint */

    for (bpCount = 0; bp_id[bpCount] != 0; bpCount++) {
	if (vxdbgBpUserKeyGet (VXDBG_CLIENT_ID, bp_id[bpCount], (void **)&cb) == OK) {
	    info->triggeredCb[info->triggeredCbCount] = cb;
	    info->triggeredCbCount++;
	}
    }

    /* Cleanup if no matching ContextBreakpoint */
    if (info->triggeredCbCount == 0) {
	loc_free(info->triggeredCb);
	info->triggeredCb = NULL;
    }

    post_event(event_handler, info);
}

/*******************************************************************************
*
* vxdbg_evt_trace_hook - VxDBG EVT_TRACE notification hook
*
* This routine is invoked by VxDBG for trace exception notification.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vxdbg_evt_trace_hook(VXDBG_CTX_ID ctx_id, VXDBG_CTX_ID stopped_ctx_id, REG_SET * pRegs, void * addr) {
    struct event_info * info = loc_alloc(sizeof(struct event_info));

    trace(LOG_CONTEXT, "vxdbg_evt_trace_hook ctx_id=%#lx / stopped_ctx_id=%#lx", ctx_id, stopped_ctx_id);

    info->event = EVENT_HOOK_STEP_DONE;
    info->current_ctx = ctx_id;
    info->stopped_ctx = stopped_ctx_id;
    info->regs = *pRegs;
    info->addr = addr;

    post_event(event_handler, info);
}

static void vxdbg_evt_ctx_stopped (VXDBG_CTX_ID	ctx_id) {
    struct event_info * info = loc_alloc(sizeof(struct event_info));

    trace(LOG_CONTEXT, "vxdbg_evt_ctx_stopped ctx_id=%#lx", ctx_id);

    info->event = EVENT_HOOK_STOP;
    info->current_ctx = ctx_id;
    info->stopped_ctx = ctx_id;

    post_event(event_handler, info);
}

static void vxdbg_evt_ctx_continued (VXDBG_CTX_ID	ctx_id) {
    struct event_info * info = loc_alloc(sizeof(struct event_info));

    trace(LOG_CONTEXT, "vxdbg_evt_ctx_continued ctx_id=%#lx", ctx_id);

    info->event = EVENT_HOOK_START;
    info->current_ctx = ctx_id;
    info->stopped_ctx = 0;

    post_event(event_handler, info);
}

static void vxdbg_evt_ctx_created (VXDBG_CTX_ID creator_id, VXDBG_CTX_ID created_id) {
    struct event_info * info = loc_alloc(sizeof(struct event_info));

    trace(LOG_CONTEXT, "vxdbg_evt_ctx_created current=%#lx created=%#lx",
            creator_id, created_id);

    info->event = EVENT_HOOK_TASK_ADD;
    info->current_ctx = creator_id;
    info->stopped_ctx = created_id;

    post_event(event_handler, info);    
}

static void vxdbg_evt_ctx_deleted (VXDBG_CTX_ID deleted_id) {
    struct event_info * info = loc_alloc(sizeof(struct event_info));

    trace(LOG_CONTEXT, "vxdbg_evt_ctx_deleted deleted=%#lx", deleted_id);

    info->event = EVENT_HOOK_TASK_DELETE;
    info->stopped_ctx = deleted_id;

    post_event(event_handler, info);    
}

static void vxdbg_evt_exc_hook (VXDBG_CTX_ID ctx_id, REG_SET * reg_set, int vector, int signum) {
    struct event_info * info = loc_alloc(sizeof(struct event_info));

#if 0   /* rlp - to be removed when VxDGB exception handling will be supported  */
    static	REG_SET regSetCopy;	/* registers of interrupted context */


    /*
     * context_find_from_pid() must not be called outside from the 
     * dispatch thread.
     * To move !!!
     */
    Context * stopped_ctx = context_find_from_pid (taskId, 1);


    if (stopped_ctx == NULL)
	{
	/* Not attached to this context: Invoke VxWorks default sigExcKill() */
	goto default_sig_handler;
	}

    
    /* Save register set to reload with _sigCtxLoad() */

    regSetCopy = *pRegSet;
#endif
    
    info->event = EVENT_HOOK_EXCEPTION;
    info->stopped_ctx = ctx_id;
    info->regs = *reg_set;
    info->signal = signum;

    post_event(event_handler, info);

#if 0   /* rlp - to be removed when VxDGB exception handling will be supported  */    
    /* Task is being resumed, check if signal must be passed */

    if (EXT(stopped_ctx)->pass_signal)
	goto default_sig_handler;

    /* Reload the context interrupted by the signal: Do not pass signal */

    _sigCtxLoad (&regSetCopy);

    return;

default_sig_handler:
   /* Invoke VxWorks default sigExcKill(): pass signal to task */

    _func_default_sigExcKill (faultType, signalNum, pRegSet);
    return;
#endif
}

/*******************************************************************************
*
* system_folder_create - create VxWorks System folder
*
* This routine creates the root folder for VxWorks system.
* If attachment <type> indicates an explicit attachment, then Kernel & RTPs
* folder must also be attached.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static Context * system_folder_create(attach_type type) {
    Context * ctx;

    /*
     * Check if VxWorks folder context already exists. This context may
     * already have been created for an implicit attachment: In this case,
     * we do not need to create the context, we must only handle the explicit
     * attachment.
     */
    if ((ctx = context_find_from_pid (VX_SYSTEM_FOLDER_ID, 0)) == NULL) {
	ctx = create_context (pid2id (VX_SYSTEM_FOLDER_ID, 0));

	if (ctx == NULL) return (NULL);

	ctx->name = loc_strdup (VX_SYSTEM_FOLDER_NAME);
	EXT(ctx)->pid = VX_SYSTEM_FOLDER_ID;
	EXT(ctx)->attach_type = type;
	EXT(ctx)->properties = IS_CONTAINER;

	ctx->parent = NULL;
	ctx->mem = ctx;			/* ctx->mem cannot be NULL */
	ctx->mem_access = 0;		/* Not a memory context */
	ctx->big_endian = big_endian_host();

	/* Add parent context to context list, and notify context creation */

	link_context(ctx);
	send_context_created_event(ctx);

	system_ctx = ctx;
    }

    /*
     * If this is an explicit attachment to System folder, then we must
     * recursively attach to Kernel & RTPs folder.
     */
    if ((type & EXPLICIT_ATTACH) == EXPLICIT_ATTACH) {
	task_folder_create(system_ctx, VX_KERNEL_FOLDER_ID, type);
#if defined(_WRS_CONFIG_RTP)
	if (_func_rtpSpawn != NULL) {
	    /* RTP support is enabled: Create RTPs folder */
	    rtps_folder_create(system_ctx, type);
	}
#endif
	EXT(ctx)->attach_type |= EXPLICIT_ATTACH;
    }

    return (ctx);
}

#if defined(_WRS_CONFIG_RTP)
/*******************************************************************************
*
* rtp_create_each_rtn - objEach routine for RTP contexts creation
*
* This routine is called via objEach for all RTPs of the system. It creates a
* task folder for each RTP.
*
* RETURNS: TRUE.
*
* ERRNO: N/A
*
*/

static BOOL rtp_create_each_rtn(OBJ_ID objId, void * pArg) {
    attach_type type = *((attach_type *)pArg);
    task_folder_create(rtps_ctx, (pid_t)objId, type);
    return (TRUE);
}

/*******************************************************************************
*
* rtp_delete_each_rtn - objEach routine for RTP contexts deletion
*
* This routine is called via objEach for all RTPs of the system. It deletes
* RTP folder for each RTP.
*
* RETURNS: TRUE.
*
* ERRNO: N/A
*
*/

static BOOL rtp_delete_each_rtn(OBJ_ID objId, void * pArg) {
    Context * ctx;
    int self_exit = *(int *)pArg;

    if ((ctx = context_find_from_pid((pid_t)objId, 1)) != NULL) {
	vx_delete_context(ctx, self_exit);
    }
    return (TRUE);
}

/*******************************************************************************
*
* rtps_folder_create - create folder for all RTPs
*
* This routine creates a folder that will hold all RTPs.
* If attachment <type> indicates an explicit attachment, then all RTPs must
* also be attached.
*
* RETURNS: Created context, or NULL if context not created.
*
* ERRNO: N/A
*
*/

static Context * rtps_folder_create(Context * parent_ctx, attach_type type) {
    Context * ctx;

    /*
     * Check if RTPs folder context already exists. This context may
     * already have been created for an implicit attachment: In this case,
     * we do not need to create the context, we must only handle the explicit
     * attachment.
     */
    if ((ctx = context_find_from_pid (VX_RTPS_FOLDER_ID, 1)) == NULL) {
	ctx = create_context (pid2id (VX_RTPS_FOLDER_ID, 0));

	if (ctx == NULL) return (NULL);

	ctx->name = loc_strdup (VX_RTPS_FOLDER_NAME);
	EXT(ctx)->pid = VX_RTPS_FOLDER_ID;
	EXT(ctx)->attach_type = type;
	EXT(ctx)->properties = IS_CONTAINER;

	ctx->parent = parent_ctx;
	ctx->mem = ctx;			/* ctx->mem cannot be NULL */
	ctx->mem_access = 0;		/* Not a memory context */
	ctx->big_endian = big_endian_host();

	context_lock (parent_ctx);
	list_add_last(&ctx->cldl, &parent_ctx->children);

	/* Add context to context list, and notify context creation */

	link_context(ctx);
	send_context_created_event(ctx);

	rtps_ctx = ctx;
    }

    if ((type & EXPLICIT_ATTACH) == EXPLICIT_ATTACH) {
	/* Must attach to all RTPs */

	if (_func_objRtpObjEachHook != NULL) {
	    _func_objRtpObjEachHook (windRtpClass, rtp_create_each_rtn, &type, FALSE);
	}
	EXT(ctx)->attach_type |= EXPLICIT_ATTACH;
    }

    return (ctx);
}
#endif /* _WRS_CONFIG_RTP */

/*******************************************************************************
*
* task_folder_create - create a TCF parent context for VxWorks tasks
*
* This routine creates a TCF parent context for VxWorks tasks. This could be
* either the Kernel folder or an RTP folder.
* If attachment <type> indicates an explicit attachment, then all childs of
* <vxworksId> must also be attached.
*
* RETURNS: Created context, or NULL if context not created.
*
* ERRNO: N/A
*
*/

static Context * task_folder_create(Context * parent_ctx, pid_t vxworksId, attach_type type) {
    Context * ctx;

    /*
     * Check if this task folder context already exists. This context may
     * already have been created for an implicit attachment: In this case,
     * we do not need to create the context, we must only handle the explicit
     * attachment.
     */
    if ((ctx = context_find_from_pid (vxworksId, 1)) == NULL) {

	ctx = create_context (pid2id (vxworksId, 0));

	if (ctx == NULL) return (NULL);

	if (vxworksId == VX_KERNEL_FOLDER_ID) {
	    ctx->name = loc_strdup (VX_KERNEL_FOLDER_NAME);
	    EXT(ctx)->properties = IS_CONTAINER;
	    kernel_ctx = ctx;
	} else {
	    char * obj_name = vx_obj_get_name (vxworksId);
	    ctx->name = loc_strdup (obj_name);
	    loc_free (obj_name);
	    EXT(ctx)->properties = IS_CONTAINER|IS_RTP;
	}

	EXT(ctx)->pid = vxworksId;
	EXT(ctx)->attach_type = type;

	ctx->parent = parent_ctx;
	ctx->mem = ctx;
	ctx->mem_access |= MEM_ACCESS_INSTRUCTION;
	ctx->mem_access |= MEM_ACCESS_DATA;
	ctx->big_endian = big_endian_host();

	if (ctx->parent != NULL) {
	    list_add_last(&ctx->cldl, &ctx->parent->children);
	    context_lock(ctx->parent);
	}

	/* Add parent context to context list, and notify context creation */

	link_context(ctx);
	send_context_created_event(ctx);
    }

    if ((type & EXPLICIT_ATTACH) == EXPLICIT_ATTACH) {
	/* Must attach to all childs */

	if (vxworksId == VX_KERNEL_FOLDER_ID) {
	    kernel_tasks_update_context(ctx, VX_CONTEXT_CREATE);
            kernelIsAttached = TRUE;
#if defined(_WRS_CONFIG_RTP)
	} else {
	    if (_func_rtpTaskEach != NULL) {
		_func_rtpTaskEach ((RTP_ID)vxworksId, rtp_task_create_context_each_rtn, (_Vx_usr_arg_t)ctx, NULL);
	    }
#endif
	}

	EXT(ctx)->attach_type |= EXPLICIT_ATTACH;
    }

    return (ctx);
    }

/*******************************************************************************
*
* kernel_tasks_update_context - create/delete task contexts for Kernel context
*
* This routine creates or deletes TCF contexts for Kernel tasks.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void kernel_tasks_update_context(Context * parent_ctx, int update_type) {
    int i = 0;
    int cnt = 0;
    int ids_cnt = 0;
    int ids_max = 100;
    TASK_ID * ids = (TASK_ID *)loc_alloc(ids_max * sizeof(TASK_ID));

    for (;;) {
	ids_cnt = taskIdListGet(ids, ids_max);
	if (ids_cnt < ids_max) break;
	loc_free(ids);
	ids_max *= 2;
	ids = (TASK_ID *)loc_alloc(ids_max * sizeof(TASK_ID));
    }

    for (i = 0; i < ids_cnt; i++) {
	if (IS_KERNEL_TASK(ids[i])) {
	    if (update_type == VX_CONTEXT_CREATE) {
		task_create_context(ids[i], parent_ctx);
	    } else {
	        Context * ctx = context_find_from_pid ((pid_t)ids[i], 1);
		if (ctx) {
		    if (update_type == VX_CONTEXT_SELF_DELETE) {
		    	vx_delete_context(ctx, 1);
		    } else {
		    	vx_delete_context(ctx, 0);
		    }
		}
	    }
	}
    }

    loc_free (ids);
}

#if defined(_WRS_CONFIG_RTP)
/*******************************************************************************
*
* rtp_task_create_context_each_rtn - each routine for RTP task contexts creation
*
* This routine is called via rtpTaskEach for all tasks of an RTP. It creates
* a TCF context for each RTP task.
*
* RETURNS: TRUE.
*
* ERRNO: N/A
*
*/

static BOOL rtp_task_create_context_each_rtn (RTP_ID rtpId, TASK_ID taskId, _Vx_usr_arg_t arg) {
    Context * parent_ctx = (Context *)arg;

    if ((context_find_from_pid ((pid_t)taskId, 1)) == NULL) {
	task_create_context (taskId, parent_ctx);
    }
    return (TRUE);
}

/*******************************************************************************
*
* rtp_task_delete_context_each_rtn - each routine for RTP task contexts deletion
*
* This routine is called via rtpTaskEach for all tasks of an RTP. It deletes
* the TCF context for each RTP task.
*
* RETURNS: TRUE.
*
* ERRNO: N/A
*
*/

static BOOL rtp_task_delete_context_each_rtn (RTP_ID rtpId, TASK_ID taskId, _Vx_usr_arg_t arg) {
    Context * ctx;
    int self_exit = *(int *)arg;

    if ((ctx = context_find_from_pid ((pid_t)taskId, 1)) != NULL) {
	vx_delete_context (ctx, self_exit);
    }
    return (TRUE);
}
#endif

/*******************************************************************************
*
* task_create_context - create a TCF context for a VxWorks task
*
* This routine creates a TCF context for a VxWorks task.
*
* RETURNS: Created context, or NULL if context not created.
*
* ERRNO: N/A
*
*/

static Context * task_create_context(TASK_ID taskId, Context * parent_ctx) {
    Context * ctx = NULL;
    VXDBG_CTX vxdbg_ctx;
    int	task_is_stopped = 0;

    if ((ctx = context_find_from_pid ((pid_t)taskId, 1)) != NULL) {
	return (ctx);
    }

    /* Do not create context for unbreakable tasks */

    if ((((WIND_TCB *)taskId)->options & VX_UNBREAKABLE) == VX_UNBREAKABLE)
	return (NULL);

    /* Create TCF context: Will be named "P<kernelId>.<taskId>" */

    ctx = create_context (pid2id ((pid_t)taskId, EXT(parent_ctx)->pid));
    ctx->name = loc_strdup (taskName (taskId));	/* Name to be displayed */

    EXT(ctx)->pid = (pid_t)taskId;
    EXT(ctx)->regs = (REG_SET *)loc_alloc(sizeof(REG_SET));
    EXT(ctx)->attach_type = EXPLICIT_ATTACH;
    EXT(ctx)->properties = IS_TASK;
    ctx->mem = parent_ctx;
    ctx->big_endian = parent_ctx->big_endian;
    ctx->parent = parent_ctx;
    context_lock(parent_ctx);

    /* Inherite signal settings from parent context */

    sigset_copy(&ctx->sig_dont_stop, &parent_ctx->sig_dont_stop);
    sigset_copy(&ctx->sig_dont_pass, &parent_ctx->sig_dont_pass);

    list_add_last(&ctx->cldl, &parent_ctx->children);

    vxdbg_ctx.ctxId = (VXDBG_CTX_ID)taskId;
    vxdbg_ctx.ctxType = VXDBG_CTX_TASK;

    /*
     * Disable the automatic "pending_intercept" handling for this stop
     * request (vxdbgAttach() does stop context) coming from the agent itself.
     */
    EXT(ctx)->no_auto_pending_intercept = TRUE;

    if (vxdbgAttach(VXDBG_CLIENT_ID, &vxdbg_ctx, &task_is_stopped) == VXDBG_ERROR) {
	/* VxDBG attachment failed: request ctx deletion */
	loc_free(EXT(ctx)->regs);
	EXT(ctx)->regs_error = NULL;
	EXT(ctx)->regs = NULL;
	return (NULL);
    }

    /*
     * If task was already stopped before attachment, don't ignore stop 
     * event.
     */
    if (task_is_stopped) EXT(ctx)->no_auto_pending_intercept = FALSE;

#if ENABLE_Symbols && defined(_WRS_CONFIG_RTP)
    /* initialize rtp_task_entry_sym_size */
    EXT(ctx)->rtp_task_entry_sym_size = -1;
#endif

    /* Add task context to context list, and notify context creation */

    link_context(ctx);
    send_context_created_event(ctx);

    trace(LOG_CONTEXT, "context: created: ctx %#lx, id %#lx, name=%s",
	  ctx, EXT(ctx)->pid, ctx->id);
    return (ctx);
    }

/*******************************************************************************
*
* vx_obj_get_name - get name of given VxWorks object ID
*
* This routine allocate buffer to return the name of given VxWorks object ID.
* This buffer must be freed by vx_obj_get_name() caller.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static char * vx_obj_get_name (pid_t pid) {
    int obj_name_len;
    char * obj_name;

    switch (pid) {
    case VX_SYSTEM_FOLDER_ID:
	return (loc_strdup (VX_SYSTEM_FOLDER_NAME));
    case VX_KERNEL_FOLDER_ID:
	return (loc_strdup (VX_KERNEL_FOLDER_NAME));
#if defined(_WRS_CONFIG_RTP)
    case VX_RTPS_FOLDER_ID:
	return (loc_strdup (VX_RTPS_FOLDER_NAME));
#endif
    default:
	obj_name_len = objNameLenGet ((OBJ_ID)pid) + 1;
	obj_name = loc_alloc_zero (obj_name_len);
	if (objNameGet ((OBJ_ID)pid, obj_name, obj_name_len) != OK) {
	    loc_free (obj_name);
	    return NULL;
	}
	return (obj_name);
    }
}

/*******************************************************************************
*
* vx_verify_task_id - check if given object ID is a VxWorks task
*
* This routine checks if given object ID is a VxWorks task or not.
*
* RETURNS: TRUE if given object is a system container, FALSE otherwise.
*
* ERRNO: N/A
*
*/

static STATUS vx_verify_task_id (pid_t pid) {
    switch (pid) {
    case VX_SYSTEM_FOLDER_ID:
    case VX_KERNEL_FOLDER_ID:
    case VX_RTPS_FOLDER_ID:
	return (ERROR);
    default:
	return (taskIdVerify ((TASK_ID)pid));
    }
}

/*******************************************************************************
*
* vx_get_id_from_obj_info - get TCF id from given VxWorks object info
*
* This routine converts given VxWorks object information structure in
* a TCF id.
*
* RETURNS: TCF id string or NULL if given object info is not valid
*
*/

char * vx_get_id_from_obj_info(VX_OBJ_INFO * obj_info) {
    if (obj_info->obj_type == VX_OBJ_SYS_FOLDER) {
	return (pid2id (VX_SYSTEM_FOLDER_ID,0));
    } else if (obj_info->obj_type == VX_OBJ_KERNEL) {
	return (pid2id (VX_KERNEL_FOLDER_ID,0));
#if defined(_WRS_CONFIG_RTP)
    } else if (obj_info->obj_type == VX_OBJ_RTPS_FOLDER) {
	return (pid2id (VX_RTPS_FOLDER_ID,0));
    } else if (obj_info->obj_type == VX_OBJ_RTP) {
	return (pid2id (obj_info->pid,0));
#endif
    } else if (obj_info->obj_type == VX_OBJ_TASK) {
	return (pid2id (obj_info->pid,(pid_t)obj_info->ppid));
    } else {
	return (NULL);
    }
}

/*******************************************************************************
*
* vx_get_id - get TCF id from given VxWorks Id
*
* This routine converts given VxWorks Id in a TCF id.
*
* RETURNS: TCF id string or NULL if given id is unknown
*
*/

char * vx_get_id(pid_t pid) {
    if (pid == VX_SYSTEM_FOLDER_ID) {
	return (pid2id (VX_SYSTEM_FOLDER_ID,0));
    } else if ((pid == (pid_t)kernelIdGet()) || (pid == VX_KERNEL_FOLDER_ID)) {
	return (pid2id (VX_KERNEL_FOLDER_ID,0));
#if defined(_WRS_CONFIG_RTP)
    } else if (pid == VX_RTPS_FOLDER_ID) {
	return (pid2id (VX_RTPS_FOLDER_ID,0));
#endif
    } else if (vx_verify_task_id (pid) == OK) {
	RTP_ID ppid = ((WIND_TCB *)(pid))->rtpId;
#if defined(_WRS_CONFIG_RTP)
	if (vx_verify_rtp_id (ppid) != OK) return (NULL);
#endif
	if (ppid == kernelIdGet()) ppid = (RTP_ID)VX_KERNEL_FOLDER_ID;
	return (pid2id (pid,(pid_t)ppid));
#if defined(_WRS_CONFIG_RTP)
    } else if (vx_verify_rtp_id((RTP_ID)pid) == OK) {
	return (pid2id (pid, 0));
#endif	/* _WRS_CONFIG_RTP */
    } else {
	return (NULL);
    }
}

/*******************************************************************************
*
* vx_get_obj_info - get VxWorks object information from given TCF id.
*
* This routine gets information on a VxWorks object from given TCF id.
*
* RETURNS: A VX_OBJ_INFO pointer which must be freed using vx_free_obj_info() or
*          NULL on invalid TCF id.
*
*/

VX_OBJ_INFO * vx_get_obj_info (char * id) {
    VX_OBJ_INFO * vx_obj_info;
    pid_t pid = id2pid (id, NULL);
    int priority;
    int taskOptions;

    /* Validate given id before going further */
    if ((pid == 0) || (vx_get_id(pid) == NULL)) return (NULL);

    if ((vx_obj_info = loc_alloc_zero(sizeof(VX_OBJ_INFO))) == NULL) return (NULL);
    
    vx_obj_info->pid = pid;
    vx_obj_info->ppid = 0;
    vx_obj_info->obj_name = vx_obj_get_name(pid);
    vx_obj_info->can_attach = TRUE;
    vx_obj_info->attached = is_attached(pid, 1);

    if (pid == VX_SYSTEM_FOLDER_ID) {
        vx_obj_info->obj_type = VX_OBJ_SYS_FOLDER;
	vx_obj_info->system_container = TRUE;
    } else if ((pid == (pid_t)kernelIdGet()) || (pid == VX_KERNEL_FOLDER_ID)) {
        vx_obj_info->obj_type = VX_OBJ_KERNEL;
	vx_obj_info->system_container = TRUE;
	vx_obj_info->task_creation = TRUE;
	vx_obj_info->max_task_args = MAX_TASK_ARGS;
	vx_obj_info->ppid = VX_SYSTEM_FOLDER_ID;
#if defined(_WRS_CONFIG_RTP)
    } else if (pid == VX_RTPS_FOLDER_ID) {
        vx_obj_info->obj_type = VX_OBJ_RTPS_FOLDER;
	vx_obj_info->system_container = TRUE;
	vx_obj_info->rtp_creation = TRUE;
	vx_obj_info->ppid = VX_SYSTEM_FOLDER_ID;
#endif
    } else if (vx_verify_task_id (pid) == OK) {
	RTP_ID ppid = ((WIND_TCB *)(pid))->rtpId;

        vx_obj_info->obj_type = VX_OBJ_TASK;

	if (taskPriorityGet ((TASK_ID)pid, &priority) == OK) {
	    vx_obj_info->task_priority = priority;
	} else {
	    vx_obj_info->task_priority = ERROR;
	}

        if (taskOptionsGet ((TASK_ID)pid, &taskOptions) == OK) {
	    if (taskOptions & VX_UNBREAKABLE) {
	        vx_obj_info->can_attach = FALSE;
            }
        }

	taskStatusString ((TASK_ID)pid, vx_obj_info->task_status);

	/* Get parent ID for this task (RTP ID or Kernel ID) */
	if (ppid == kernelIdGet ()) {
	    vx_obj_info->ppid = VX_KERNEL_FOLDER_ID;
	} else {
	    vx_obj_info->ppid = (pid_t)ppid;
	}
#if defined(_WRS_CONFIG_RTP)
    } else if (vx_verify_rtp_id((RTP_ID)pid) == OK) {
        vx_obj_info->obj_type = VX_OBJ_RTP;
	vx_obj_info->task_creation = FALSE;
	vx_obj_info->ppid = VX_RTPS_FOLDER_ID;
#endif	/* _WRS_CONFIG_RTP */
    } else {
        vx_obj_info->obj_type = VX_OBJ_UNKNOWN;
    }

    return (vx_obj_info);
}

/*******************************************************************************
*
* vx_free_obj_info - free a information structure returned by vx_get_obj_info()
*
* This routine frees a VxWorks object information structure returned by
* vx_get_obj_info().
*
* RETURNS: N/A
*
*/

void vx_free_obj_info (VX_OBJ_INFO * vx_obj_info) {

    if (vx_obj_info == NULL) return;

    if (vx_obj_info->obj_name != NULL) {
    	loc_free (vx_obj_info->obj_name);
    }
    loc_free (vx_obj_info);
}

/*******************************************************************************
*
* vx_create_context - create a TCF vxWorks context
*
* This routine creates a TCF context for given vxWorks context: Can be a task,
* an RTP or the kernel context.
* <type> argument specifies the type of attachment to perform. For example:
* when attaching to a kernel task while the debugger is not attached to the
* kernel context, then we must perform an implicit attachment to the kernel,
* which means that when this task will be detached, we will also detach from
* the kernel if no other kernel is attached.
*
* RETURNS: created TCF context.
*
* ERRNO: N/A
*
*/

static Context * vx_create_context(pid_t pid, attach_type type, int * err) {
    Context * ctx = NULL;
    Context * parent_ctx = NULL;

    if (is_attached (pid, 1)) {
	*err = ERR_ALREADY_ATTACHED;
	return (NULL);
    }

    if (vx_verify_task_id(pid) == OK) {
	/* Task attachment */

	/* First check if a context has been created for task parent */

	pid_t rtpId = (pid_t)(((WIND_TCB *)pid)->rtpId);

	if (rtpId == (pid_t)kernelIdGet()) {
	    rtpId = VX_KERNEL_FOLDER_ID;
        }

	parent_ctx = context_find_from_pid (rtpId, 1);

	if (parent_ctx == NULL) {
	    /* No parent context for this task */

	    if (type == COND_PARENT_ATTACH) {
		/* We must only attach if parent is already attach: exit */
                *err = ERR_INV_CONTEXT;
		return (NULL);
	    }

	    /* Create a parent context for this task (Implicit attachment) */

	    if ((parent_ctx = vx_create_context (rtpId, IMPLICIT_ATTACH, NULL)) == NULL) {
                *err = ERR_INV_CONTEXT;
		return (NULL);
            }
	}

	if ((type == COND_PARENT_ATTACH) && ((EXT(parent_ctx)->attach_type & EXPLICIT_ATTACH) != EXPLICIT_ATTACH)) {
	    /*
	     * Conditional attachment: We must only attach if parent context
	     * has been explicitly attached.
	     */
	    *err = ERR_INV_CONTEXT;
	    return (NULL);
	}

	/* Create Task context */
	if ((ctx = task_create_context ((TASK_ID)pid, parent_ctx)) == NULL) {
            *err = ERR_INV_CONTEXT;
            return (NULL);
	}
    }
#if defined(_WRS_CONFIG_RTP)
    else if (vx_verify_rtp_id((RTP_ID)pid) == OK) {
        /* RTP attachment */

	/* First check if a context has been created for the RTPs folder */

	parent_ctx = context_find_from_pid (VX_RTPS_FOLDER_ID, 1);

	if (parent_ctx == NULL) {
	    if ((parent_ctx = vx_create_context (VX_RTPS_FOLDER_ID, IMPLICIT_ATTACH, NULL)) == NULL) {
                *err = ERR_INV_CONTEXT;
		return (NULL);
	    }
	}

	/* Create a context for this RTP */

	ctx = task_folder_create(parent_ctx, pid, type);
    } else if (pid == VX_RTPS_FOLDER_ID) {
	/* RTPs folder attachment */

	/* First check if a context has been created for the System folder */

	parent_ctx = context_find_from_pid (VX_SYSTEM_FOLDER_ID, 0);

	if (parent_ctx == NULL) {
	    if ((parent_ctx = vx_create_context (VX_SYSTEM_FOLDER_ID, IMPLICIT_ATTACH, NULL)) == NULL) {
                *err = ERR_INV_CONTEXT;
		return (NULL);
	    }
	}

	/* Create RTPs context */

	ctx = rtps_folder_create(parent_ctx, type);
    }
#endif
    else if (pid == VX_KERNEL_FOLDER_ID) {
        /* Kernel folder attachment */

	/* First check if a context has been created for the System folder */

	parent_ctx = context_find_from_pid (VX_SYSTEM_FOLDER_ID, 0);

	if (parent_ctx == NULL) {
	    if ((parent_ctx = vx_create_context (VX_SYSTEM_FOLDER_ID, type & ~EXPLICIT_ATTACH, NULL)) == NULL) {
	        *err = ERR_INV_CONTEXT;
		return (NULL);
	    }
	}

	/* Create KERNEL context */
	ctx = task_folder_create(parent_ctx, VX_KERNEL_FOLDER_ID, type);
    } else if (pid == VX_SYSTEM_FOLDER_ID) {
	/* Create SYSTEM context */
	ctx = system_folder_create(type);
    }

    return (ctx);
}

/*******************************************************************************
*
* vx_detach_context - VxWorks context and children contexts detachment
*
* This routine handles VxWorks context and children contexts detachment.
*
* RETURNS: Always 0.
*
* ERRNO: N/A
*
*/

static int vx_detach_context (Context * ctx, int self_exit) {
    ContextExtensionVxWorks * ext = EXT(ctx);
    pid_t vxworksId = id2pid (ctx->id, NULL);
    int detach_children_only = ((ext->attach_type & ALWAYS_ATTACH) == ALWAYS_ATTACH);
    LINK * qp;

    if (CTX_IS_DETACHING(ctx)) {
	return 0;
    }

    if (CONTEXT_CAN_DETACH(ctx)) ext->properties |= DETACHING;

    /* Handle deletion of system contexts */

    if (vxworksId == VX_KERNEL_FOLDER_ID) {
	/* Kernel detachment */
	if (!detach_children_only) kernel_ctx = NULL;
        kernelIsAttached = FALSE;
#if defined(_WRS_CONFIG_RTP)
    } else if (vxworksId == VX_RTPS_FOLDER_ID) {
	/* RTPs folder detachment */
	if (!detach_children_only) rtps_ctx = NULL;
#endif
    } else if (vxworksId == VX_SYSTEM_FOLDER_ID) {
	/* System folder detachment */
	if (!detach_children_only) system_ctx = NULL;
    }

    if (detach_children_only) {
        /*
         * We're detaching the children, but we must remain attached to this
         * memory context. Remove the EXPLICIT_ATTACH flag to allow the
         * re-attachment to this context and its children (See is_attached()).
         */
	ext->attach_type = ext->attach_type & ~EXPLICIT_ATTACH;
    }

    /* Handle children contexts detachment */

    for (qp = ctx->children.next; qp != &ctx->children;) {
	Context * cctx = cldl2ctxp(qp);
	qp = qp->next;
	vx_detach_context(cctx, self_exit);
    }

    if (!CONTEXT_CAN_DETACH(ctx)) {
	/* Given context can't be detached: Nothing more to do */
	return 0;
    }

    /* Handle context detachment */

    unplant_breakpoints(ctx);
    ctx->exiting = 1;

    list_remove(&ctx->ctxl);
    if (ctx->parent != NULL) {
	list_remove(&ctx->cldl);
	context_unlock (ctx->parent);
	ctx->parent = NULL;
    }

    /* VxDBG detachment */

    if (!self_exit && (vx_verify_task_id (ext->pid) == OK)) {
	VXDBG_CTX vxdbg_ctx;
        vxdbg_ctx.ctxId = ext->pid;
        vxdbg_ctx.ctxType = VXDBG_CTX_TASK;
        vxdbgDetach(VXDBG_CLIENT_ID, &vxdbg_ctx);
    }

    /* NOTE: <ctx> must not be used after following command */

    send_context_exited_event(ctx);

    return 0;
}

/*******************************************************************************
*
* vx_delete_context - delete a TCF vxWorks context
*
* This routine cleanup a delete a TCF vxWorks context previously created by
* vx_create_context().
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static int vx_delete_context (Context * ctx, int self_exit) {
    ContextExtensionVxWorks * ext = EXT(ctx);
    Context * parent_ctx = ctx->parent;
    int detach_children_only = ((ext->attach_type & ALWAYS_ATTACH) == ALWAYS_ATTACH);

    /* Handle context and children contexts detachment */

    vx_detach_context (ctx, self_exit);

    /*
     * Handle parent context detachment:
     * If parent attachment was implicit, not flagged as ALWAYS_ATTACH and it
     * has no other child, then parent context must also be detached.
     */

    if ((parent_ctx != NULL) && ((EXT(parent_ctx)->attach_type & EXPLICIT_ATTACH) != EXPLICIT_ATTACH) && ((EXT(parent_ctx)->attach_type & ALWAYS_ATTACH) != ALWAYS_ATTACH)) {
        LINK * l = parent_ctx->children.next;
	int child_cnt = 0;
        while (l != &parent_ctx->children) {
            Context * c = cldl2ctxp(l);
	    child_cnt++;
            l = l->next;
        }

	if (child_cnt == 0) {
	    /* No more child: delete parent context */
	    vx_delete_context (parent_ctx, self_exit);
	}	
    }

    return 0;
}

static int is_attached(pid_t pid, int explicit) {
#if ENABLE_DebugContext
    Context * ctx;

    if (((ctx = context_find_from_pid(pid, 0)) != NULL) || ((ctx = context_find_from_pid(pid, 1)) != NULL)) {
	ContextExtensionVxWorks * ext = EXT(ctx);

	if (explicit == 0) return 1;

	if ((ext->attach_type & EXPLICIT_ATTACH) == EXPLICIT_ATTACH) {
	    return 1;
	}
    }
#endif
    return 0;
}

#if defined(_WRS_CONFIG_RTP)
/*******************************************************************************
*
* get_children_rtp_each_rtn - objEach routine for RTP contexts listing
*
* This routine is called via objEach for all RTPs of the system. It adds
* a element to getChildren json output for each RTPs.
*
* RETURNS: TRUE.
*
* ERRNO: N/A
*
*/

static BOOL get_children_rtp_each_rtn(OBJ_ID objId, void * pArg) {
    Channel * c = ((object_each_rtn_args *)pArg)->c;
    int rtpCount = ((object_each_rtn_args *)pArg)->count;
    int attached_only = ((object_each_rtn_args *)pArg)->attached_only;

    if (!attached_only || is_attached((pid_t)objId, 0)) {
	if (rtpCount > 0) {
	    write_stream(&c->out, ',');
	}

	json_write_string(&c->out, pid2id((pid_t)objId, 0));

	((object_each_rtn_args *)pArg)->count++;
    }

    return (TRUE);
}

/*******************************************************************************
*
* get_children_rtp_task_each_rtn - each routine for RTP task contexts listing
*
* This routine is called via rtpTaskEach for all tasks of an RTP. It adds
* a element to getChildren json output for each RTP task.
*
* RETURNS: TRUE.
*
* ERRNO: N/A
*
*/

static BOOL get_children_rtp_task_each_rtn (RTP_ID rtpId, TASK_ID taskId, _Vx_usr_arg_t arg) {
    Channel * c = ((object_each_rtn_args *)arg)->c;
    int rtpTaskCount = ((object_each_rtn_args *)arg)->count;
    int attached_only = ((object_each_rtn_args *)arg)->attached_only;

    if (!attached_only || is_attached((pid_t)taskId, 0)) {
	if (rtpTaskCount > 0) write_stream(&c->out, ',');

	json_write_string(&c->out, pid2id((pid_t)taskId, (pid_t)rtpId));
	((object_each_rtn_args *)arg)->count++;
    }

    return (TRUE);
}
#endif	/* _WRS_CONFIG_RTP */

/*******************************************************************************
*
* vx_get_children - get children of given <id>
*
* This routine returns children of given <id>. If <attached_only> is true,
* then only return contexts that are attached.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

void vx_get_children(Channel * c, char * id, int attached_only) {
    pid_t parent;
    pid_t vxId;

    if (id[0] == 0) {
	/* No id: Return System folder (Root folder) */
	write_errno(&c->out, 0);
	write_stream(&c->out, '[');

	if (!attached_only || is_attached((pid_t)VX_SYSTEM_FOLDER_ID, 0)) {
	    json_write_string(&c->out, pid2id(VX_SYSTEM_FOLDER_ID,0));
	}
	write_stream(&c->out, ']');
	write_stream(&c->out, 0);
	return;
    }

    vxId = id2pid (id, &parent);

    if (vxId == VX_SYSTEM_FOLDER_ID) {
	/* Get children of System folder */
	int cnt = 0;

	write_errno(&c->out, 0);
	write_stream(&c->out, '[');
	if (!attached_only || is_attached((pid_t)VX_KERNEL_FOLDER_ID, 0)) {
	    json_write_string(&c->out, pid2id(VX_KERNEL_FOLDER_ID,0));
	    cnt++;
	}

#if defined(_WRS_CONFIG_RTP)
	if ((_func_rtpSpawn != NULL) && (!attached_only || is_attached((pid_t)VX_RTPS_FOLDER_ID, 0))) {
	    if (cnt != 0) write_stream(&c->out, ',');
	    json_write_string(&c->out, pid2id(VX_RTPS_FOLDER_ID,0));
	}
#endif  /* _WRS_CONFIG_RTP */
	write_stream(&c->out, ']');
	write_stream(&c->out, 0);
	return;
    }

    if (vxId == VX_KERNEL_FOLDER_ID) {
	/* Get Kernel tasks list */
	int i = 0;
	int cnt = 0;
	int ids_cnt = 0;
	int ids_max = 100;

	TASK_ID * ids = (TASK_ID *)loc_alloc(ids_max * sizeof(TASK_ID));
	for (;;) {
	    ids_cnt = taskIdListGet(ids, ids_max);
	    if (ids_cnt < ids_max) break;
	    loc_free(ids);
	    ids_max *= 2;
	    ids = (TASK_ID *)loc_alloc(ids_max * sizeof(TASK_ID));
	}
	write_errno(&c->out, 0);
	write_stream(&c->out, '[');
	for (i = 0; i < ids_cnt; i++) {
	    if (IS_KERNEL_TASK(ids[i]) && (!attached_only || is_attached((pid_t)ids[i], 0))) {
		if (cnt > 0) write_stream(&c->out, ',');
		json_write_string(&c->out, pid2id((pid_t)ids[i], vxId));
		cnt++;
	    }
	}
	write_stream(&c->out, ']');
	write_stream(&c->out, 0);

	loc_free (ids);
	return;
    }

    if (vx_verify_task_id (vxId) == OK) {
	/* Tasks have no children */

	write_errno(&c->out, 0);
	write_string(&c->out, "[]");
	write_stream(&c->out, 0);
	return;
    }

#if defined(_WRS_CONFIG_RTP)
    if (vxId == VX_RTPS_FOLDER_ID) {
	object_each_rtn_args args;
	args.c = c;
	args.count = 0;
	args.attached_only = attached_only;

	/* Get RTPs list */

        write_errno(&c->out, 0);
	write_stream(&c->out, '[');

	if (_func_objRtpObjEachHook != NULL) {
	    _func_objRtpObjEachHook (windRtpClass, get_children_rtp_each_rtn, (void *)&args, FALSE);
	}
	write_stream(&c->out, ']');
	write_stream(&c->out, 0);
	return;
    }

    if (vx_verify_rtp_id((RTP_ID)vxId) == OK) {
	object_each_rtn_args args;
	args.c = c;
	args.count = 0;
	args.attached_only = attached_only;

	/* Get RTP tasks list */

	write_errno(&c->out, 0);
	write_stream(&c->out, '[');

	if (_func_rtpTaskEach != NULL) {
	    _func_rtpTaskEach ((RTP_ID)vxId, get_children_rtp_task_each_rtn, (_Vx_usr_arg_t)&args, NULL);
	}

	write_stream(&c->out, ']');
	write_stream(&c->out, 0);
	return;
    }
#endif	/* _WRS_CONFIG_RTP */

    /* Given id is not valid */
    write_errno(&c->out, ERR_INV_CONTEXT);
    write_stringz(&c->out, "null");
}

/*******************************************************************************
*
* vx_task_create - create a VxWorks task from "Processes start" parameters
*
* This routine parses "Processes start" parameters and create a VxWorks task.
*
* RETURNS: Id of created task, or if 0 failed to create.
*
* ERRNO: Errno is returned via p_err argument.
*
*/

pid_t vx_task_create (char * name, int priority, int task_options, ssize_t stack_size, ContextAddress start_address, char * args[], int attach, int attach_mode, int * p_err, ContextAttachCallBack * attach_cb, void * attach_cb_data) {
    TASK_ID task_id;
    _Vx_usr_arg_t task_args[10];
    static int taskCnt = 1;
    int ix = 0;
    TASK_CREATION_INFO * pInfo = NULL;

    *p_err = 0;

    if (attach)  {
	if (VXDBG_CLIENT_ID == NULL) {
	   *p_err = errnoGet();
	   return 0; 
	}
    }

    /* Task arguments */

    memset (task_args, 0, sizeof(_Vx_usr_arg_t) * 10);
    for (ix = 0; ix < 10; ix++) {
	if (args[ix] == NULL) break;	/* No more args */

	if (args[ix][0] == '"') {
	    /* Handle string argument */
	    int len = 0;
	    char * str_arg = loc_strdup(args[ix] + 1);
	    len = strlen (str_arg) - 1;
	    str_arg[len] = EOS;
	    task_args[ix] = (_Vx_usr_arg_t)str_arg;
	} else if (args[ix][0] == '0' && (args[ix][1] == 'x' || args[ix][1] == 'X')) {
	    /* Handle hexadecimal argument */
	    task_args[ix] = strtoul (args[ix], NULL, 16);
	} else if (args[ix][0] >= '0' && args[ix][0] <= '9') {
	    /* Handle decimal argument */
	    task_args[ix] = strtoul (args[ix], NULL, 10);
	}
    }

    if (priority == -1) priority = DEFAULT_TASK_PRIORITY;
    if (stack_size == -1) stack_size = DEFAULT_STACK_SIZE;
    if (task_options == -1) task_options = VX_COPROCS_ALL_TASK;

    if (attach) semTake (taskStartListLock, WAIT_FOREVER); /* Lock task creation list */

    task_id = taskCreate(name, priority, task_options, (size_t)stack_size, (FUNCPTR)start_address, task_args[0], task_args[1], task_args[2], task_args[3], task_args[4], task_args[5], task_args[6], task_args[7], task_args[8], task_args[9]);

    if (task_id == TASK_ID_NULL) {
	*p_err = errnoGet();
	if (attach) semGive (taskStartListLock); /* Unlock task creation list */
	return (0);
    }

    /*
     * If this task must be attached, then we must record attachment
     * properties in a list, those information will be handled when task
     * creation event for this task will be received.
     */

    if (attach) {
    	pInfo = loc_alloc_zero(sizeof(TASK_CREATION_INFO));
	pInfo->tid = task_id;
	pInfo->start_address = (void *)start_address;
	pInfo->attach_mode = attach_mode;
	pInfo->attach_cb = attach_cb;
	pInfo->attach_cb_data = attach_cb_data;
	lstAdd (&taskStartList, &pInfo->chain);
	semGive (taskStartListLock);	/* Unlock task creation list */
    } else {
	/* "Attach":false => Task must be activated now. */
	taskActivate(task_id);
    }

    return ((pid_t)task_id);
}

#ifdef _WRS_CONFIG_RTP
/*******************************************************************************
*
* vx_verify_rtp_id - verify if given ID is an RTP ID
*
* This routine checks if given ID is a fake ID (System folder, kernel folder,
* RTPs folder), if yes then it directly returns ERROR, otherwise it invokes
* rtpIdVerify() to check if it is an RTP ID.
*
* RETURNS: OK if given ID is a valid RTP identifier.
*
* ERRNO: rtpIdVerify() errnos.
*
*/

static STATUS vx_verify_rtp_id (RTP_ID id) {

    if (id == kernelIdGet()) return (OK);

    if ((id ==  (RTP_ID)VX_SYSTEM_FOLDER_ID) || (id == (RTP_ID)VX_KERNEL_FOLDER_ID) || (id == (RTP_ID)VX_RTPS_FOLDER_ID) || (_func_rtpIdVerify == NULL)) return (ERROR);

    return (_func_rtpIdVerify (id));
}

/*******************************************************************************
*
* vx_rtp_create_reply - reply to Processes start command
*
* This routine replies to "Processes start" command.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vx_rtp_create_reply (void * data) {
    RTP_SPAWN_REQ * pReq = (RTP_SPAWN_REQ *) data;

    start_process_send_reply (pReq);

    /* Answer is sent. We can discard pReq */
    channel_unlock (pReq->channel);

    loc_free (pReq->token);
    loc_free (pReq);
    }

/*******************************************************************************
*
* vx_rtp_create_worker - worker task for RTP creation
*
* This routine is the work task code used to handle RTP creation requests
* coming from the Processes service.
* The "Processes start" command must defer RTP creation to a worker thread to
* avoid potential agent dead lock when using blocking calls.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vx_rtp_create_worker (RTP_SPAWN_REQ * pReq) {
    RTP_ID rtp_id = RTP_ID_ERROR;
    static int pty_cnt = 0;
    int err = 0;
    char pty_in_out[PTY_DEV_NAME_STR_SIZE];
    char pty_err[PTY_DEV_NAME_STR_SIZE];
    char pty_name[PTY_DEV_NAME_STR_SIZE];
    RTP_IO_INFO * pIoInfo = NULL;

    pReq->pid = (pid_t)rtp_id;

    if (pReq->redirect_io) {
	if ((pReq->pipe_name = loc_alloc_zero(PTY_DEV_NAME_STR_SIZE)) == NULL) {
	    pReq->err = errnoGet();
	    VXDBG_POST_EVENT (vx_rtp_create_reply, pReq, NULL);
	    return;
	}
	snprintf(pReq->pipe_name, PTY_DEV_NAME_STR_SIZE, "%s%d", PTY_ROOT_DEV_NAME, ++pty_cnt);
	PIPE_INOUT_MASTER_FD(pReq) = -1;
	PIPE_INOUT_SLAVE_FD(pReq) = -1;
	PIPE_ERR_MASTER_FD(pReq) = -1;
	PIPE_ERR_SLAVE_FD(pReq) = -1;
	
	/* Create pty for IO redirection */
	snprintf(pty_err, sizeof(pty_err)-1, "%sErr", pReq->pipe_name);
	snprintf(pty_in_out, sizeof(pty_in_out)-1, "%sInOut", pReq->pipe_name);
	if ((ptyDevCreate(pty_in_out, PIPE_SIZE, PIPE_SIZE) == ERROR) || (ptyDevCreate(pty_err, PIPE_SIZE, PIPE_SIZE) == ERROR)) {
	    pReq->err = errnoGet();
	    goto vx_rtp_create_worker_err;
	}

	snprintf(pty_name, sizeof(pty_name), "%sM", pty_in_out);
	PIPE_INOUT_MASTER_FD(pReq) = open (pty_name, O_RDWR, 0);
	snprintf(pty_name, sizeof(pty_name), "%sS", pty_in_out);
	PIPE_INOUT_SLAVE_FD(pReq) = open (pty_name, O_RDWR, 0);
	snprintf(pty_name, sizeof(pty_name), "%sM", pty_err);
	PIPE_ERR_MASTER_FD(pReq) = open (pty_name, O_RDWR, 0);
	snprintf(pty_name, sizeof(pty_name), "%sS", pty_err);
	PIPE_ERR_SLAVE_FD(pReq) = open (pty_name, O_RDWR, 0);
	if ((PIPE_INOUT_MASTER_FD(pReq) == ERROR) || (PIPE_INOUT_SLAVE_FD(pReq) == ERROR) || (PIPE_ERR_MASTER_FD(pReq) == ERROR) || (PIPE_ERR_SLAVE_FD(pReq) == ERROR)) {
	    pReq->err = errnoGet();
	    goto vx_rtp_create_worker_err;
	} else {
	    if ((ioctl (PIPE_INOUT_SLAVE_FD(pReq), FIOSETOPTIONS, OPT_CRMOD|OPT_ECHO) == ERROR) || (ioctl (PIPE_ERR_SLAVE_FD(pReq), FIOSETOPTIONS, OPT_CRMOD|OPT_ECHO) == ERROR)) {
		pReq->err = errnoGet();
		goto vx_rtp_create_worker_err;
	    }
	    if ((ioTaskStdSet (0, 0, PIPE_INOUT_SLAVE_FD(pReq)) == ERROR) || (ioTaskStdSet (0, 1, PIPE_INOUT_SLAVE_FD(pReq)) == ERROR) || (ioTaskStdSet (0, 2, PIPE_ERR_SLAVE_FD(pReq)) == ERROR)) {
		pReq->err = errnoGet();
		goto vx_rtp_create_worker_err;
	    }
	}
    }

    if (pReq->err == 0) {
	semTake (rtpListLock, WAIT_FOREVER); /* Lock RTP creation list */

	rtp_id = _func_rtpSpawn (pReq->exe, pReq->args, pReq->envp, pReq->priority, pReq->stack_size, pReq->rtp_options, pReq->init_task_options);

	if (rtp_id == RTP_ID_ERROR) {
	    pReq->err = errnoGet();
	} else {
	    RTP_CREATION_INFO * pInfo = NULL;

	    pReq->pid = (pid_t)rtp_id;

	    /* Add RTP to RTP creation list */

	    pInfo = loc_alloc_zero(sizeof(RTP_CREATION_INFO));
	    pInfo->rtp_id = rtp_id;
	    pInfo->pReq = pReq;
	    lstAdd (&rtpStartList, &pInfo->chain);

            if (pReq->redirect_io) {
                pIoInfo = loc_alloc_zero(sizeof(RTP_IO_INFO));

                pIoInfo->rtp_id = rtp_id;
                pIoInfo->pipe_name = pReq->pipe_name;
                pIoInfo->pipe_fds[0] = PIPE_INOUT_SLAVE_FD(pReq);
                pIoInfo->pipe_fds[1] = PIPE_ERR_SLAVE_FD(pReq);
                lstAdd (&rtpIoList, &pIoInfo->chain);
            }
        }

	semGive (rtpListLock);	/* Unlock RTP creation list */
    }

    if (rtp_id != RTP_ID_ERROR) return;

vx_rtp_create_worker_err:
    if (pReq->redirect_io) {
        if (pReq->pipe_name != NULL) {
	    if (PIPE_INOUT_MASTER_FD(pReq) != -1) (void) close (PIPE_INOUT_MASTER_FD(pReq) );
	    if (PIPE_INOUT_SLAVE_FD(pReq) != -1) (void) close (PIPE_INOUT_SLAVE_FD(pReq));
	    if (PIPE_ERR_MASTER_FD(pReq) != -1) (void) close (PIPE_ERR_MASTER_FD(pReq));
	    if (PIPE_ERR_SLAVE_FD(pReq) != -1) (void) close (PIPE_ERR_SLAVE_FD(pReq));
	    (void) ptyDevRemove (pty_in_out);
	    (void) ptyDevRemove (pty_err);
	    loc_free (pReq->pipe_name);
            pReq->pipe_name = NULL;
	    pty_cnt--;
	}
    }

    /*
     * If rtpSpawn() failed, send reply to "Processes start" command now.
     * Otherwise, reply will be send from RTP init complete hook.
     */
    VXDBG_POST_EVENT (vx_rtp_create_reply, pReq, NULL);
}
#endif	/* _WRS_CONFIG_RTP */

/*******************************************************************************
*
* vx_rtp_create - create a VxWorks RTP from "Processes start" parameters
*
* This routine parses "Processes start" parameters and creates a worker
* task to handle RTP spawn request.
*
* The reply to "Processes start" command will be sent by the
* rtp_initialize_context().
*
* RETURNS: OK, or ERROR if failed to launch worker task for RTP spawn.
*
* ERRNO: Errno is returned via p_err argument.
*
*/

int vx_rtp_create (char * token, Channel * chnl,  void * prs, const char * rtp_file_name, const char * argv[], const char * envp[], int priority, ssize_t stack_size, int rtp_options, int init_task_options, int redirect_io, int attach, int attach_mode, int * p_err) {
#ifdef _WRS_CONFIG_RTP
    RTP_SPAWN_REQ * pReq;
    int rtpSpPriority;

    if (_func_rtpSpawn == NULL) {
	*p_err = ERR_UNSUPPORTED;	/* No RTP spawn support */
	return (-1);
    }

   if (VXDBG_CLIENT_ID == NULL) {
	*p_err = errnoGet();
	return -1; 
   }

    if (priority == -1) priority = DEFAULT_TASK_PRIORITY;
    if (stack_size == -1) stack_size = DEFAULT_STACK_SIZE;
    if (init_task_options == -1) init_task_options = VX_COPROCS_ALL_TASK;

    /*
     * Do not set the option to stop the RTP just before switching to user
     * land if either the "Stop at entry" or "Stop at main" option is set.
     */

    if (((attach_mode & CONTEXT_ATTACH_NO_STOP) == 0) || ((attach_mode & CONTEXT_ATTACH_NO_STOP) == 0)) rtp_options &= ~RTP_DEBUG;

    if ((pReq = loc_alloc_zero(sizeof(RTP_SPAWN_REQ))) == NULL) {
	*p_err = errnoGet();
	return (-1);
    }

    /* lock the channel (Must be unlocked sending command result) */

    channel_lock(chnl);

    /* Fill RTP spawn request */

    pReq->channel = chnl;
    pReq->token = loc_strdup(token);
    pReq->exe = rtp_file_name;
    pReq->args = argv;
    pReq->envp = envp;
    pReq->priority = priority;
    pReq->stack_size = (size_t)stack_size;
    pReq->rtp_options = rtp_options;
    pReq->init_task_options = init_task_options;
    pReq->redirect_io = redirect_io;
    pReq->attach = attach;
    pReq->attach_mode = attach_mode;
    pReq->prs = prs;
    
    /* Launch a worker task to execute RTP spawn request */
 
    if ((taskPriorityGet(taskIdSelf(), &rtpSpPriority) == ERROR) || (taskSpawn ("tTcfRtpSp", rtpSpPriority, VX_UNBREAKABLE|VX_COPROCS_ALL_TASK, WORKER_STACK_SIZE, (FUNCPTR) vx_rtp_create_worker,(_Vx_usr_arg_t) pReq,0,0,0,0,0,0,0,0,0) == TASK_ID_ERROR)) {
	*p_err = errnoGet();
	channel_unlock (pReq->channel);
	loc_free (pReq->token);
	loc_free (pReq);
	return (-1);
    }

    return (0);
#else
    *p_err = ERR_UNSUPPORTED;	/* No RTP support */
    return (-1);
#endif
}

/*******************************************************************************
*
* vx_delete_task - delete given task with a timeout
*
* This routine tries to delete given task, until timeout expires.
*
* RETURNS: OK, or ERROR if failed to delete task within given timeout.
*
*/

STATUS vx_delete_task (TASK_ID task_id, int timeout) {
    int ix;

    for (ix = 0; ix < timeout; ix++) {
	if (taskDestroy(task_id, TRUE, NO_WAIT, 0) == OK) {
	    /* We've successfully terminated task */

	    trace(LOG_CONTEXT, "context: killed VxWorks ctx id %#lx", task_id);
	    return (OK);
	}

	/* No need to loop if task is stopped */
	if (taskIsStopped(task_id)) break;

	taskDelay (1);
    }

    return (ERROR);
}

/*******************************************************************************
*
* vx_kill_task - kill given VxWorks task
*
* This routine tries to kill a VxWorks task using taskDestroy() and specified
* timeout.
* If the task deletion fail, we do update the corresponding TCF context with
* stopped_on_error/stopped_errno information. This error will only be visible
* in the debug view when suspending the task which can't be killed.
*
* NOTE: To get terminate error automatically reported in the debug view (without
* having the suspend the context), we would need to stop this context which
* could lead to severe troubles if this context is suspended while holding a
* critical resource of the system.
*
* RETURNS: 0, or -1 if failed to kill task within given timeout.
*
* ERRNO: N/A
*
*/

static int vx_kill_task(_Vx_usr_arg_t arg1, _Vx_usr_arg_t arg2, _Vx_usr_arg_t arg3, _Vx_usr_arg_t arg4, _Vx_usr_arg_t arg5, _Vx_usr_arg_t arg6) {
    pid_t vxworksId = (pid_t)arg1;
    _Vx_usr_arg_t timeout = arg2;
    EventCallBack * terminate_done = (EventCallBack *)arg3;
    TerminateDoneArgs * terminate_done_args = (TerminateDoneArgs *)arg4;
    Context * ctx;
    int terminate_error = 0;
    static char error_msg[128];

    ctx = context_find_from_pid ((pid_t)vxworksId, 1);

    if (vx_delete_task ((TASK_ID)vxworksId, timeout) == ERROR) {
	/*
	 * Failed to delete task with timeout:
	 * We must report terminate error.
	 */

	terminate_error = errnoGet();

	if (ctx != NULL) {
	    /*
	     * Found a TCF context for the task we failed to kill.
	     * If this context is not alredy stopped we must stop it,
	     * and record the the stop reason/errno.
	     */

	    EXT(ctx)->stopped_on_error = TERMINATE_ERROR;
	    EXT(ctx)->stopped_errno = terminate_error;
	}

	if (terminate_error) {
	    trace(LOG_CONTEXT, "context: can't kill VxWorks ctx id %#lx: %s", vxworksId,  errno_to_str(terminate_error));
	}
    }

    /*
     * If a callback routine has been set to handle "Processes terminate"
     * replies, we must invoke it here.
     */

    if (terminate_done != NULL) {
	if (terminate_error) {
	    snprintf(error_msg, sizeof(error_msg), "%s", TERMINATE_ERROR);
	    terminate_done_args->error = terminate_error;
	    terminate_done_args->msg = error_msg;
	} else {
	    terminate_done_args->error = 0;
	    terminate_done_args->msg = NULL;
	}
	post_event (terminate_done, terminate_done_args);
    }

    if (ctx != NULL && terminate_error != 0) {
	/* Context terminate has been aborted: reset exiting flag */
	ctx->exiting = 0;
    }

    return (terminate_error == 0 ? 0 : -1);
}

#if defined(_WRS_CONFIG_RTP)
/*******************************************************************************
*
* rtp_delete_task - delete an RTP task
*
* This routine deletes the given RTP task with a timeout.
* If the task deletion fail, we do update the corresponding TCF context with
* stopped_on_error/stopped_errno information. This error will only be visible
* in the debug view when suspending the task which can't be killed.
* 
* RETURNS: 0, or the errno corresponding to the terminate failure.
*
* ERRNO: N/A
*
*/

static int rtp_delete_task (TASK_ID taskId) {

    int terminate_error = 0;

    if (vx_delete_task (taskId, vx_terminate_timeout) != OK) {
	Context * ctx = context_find_from_pid ((pid_t)taskId, 1);

	terminate_error = errnoGet();

	if (ctx != NULL) {
	    /*
	     * Found a TCF context for this RTP task.
	     * Update this context with stop reason/errno.
	     * If this context is not already stopped we must stop it to get
	     * the terminate failure reported in the debug view.
	     */

	    EXT(ctx)->stopped_on_error = RTP_TERMINATE_ERROR;
	    EXT(ctx)->stopped_errno = terminate_error;
            EXT(ctx)->properties |= TERMINATE_REQUEST;
	}
    }
    
    return terminate_error;
}

/*******************************************************************************
*
* rtp_first_task_get - return first task of RTP
*
* This routine is an rtpTaskEach() which is used to simply get the first task
* of the RTP task list.
*
* RETURNS: Always FALSE to abort rtpTaskEach() on 1st task.
*
* ERRNO: N/A
*
*/

static BOOL rtp_first_task_get (RTP_ID rtpId, TASK_ID taskId, _Vx_usr_arg_t arg) {
    return FALSE;
}

/*******************************************************************************
*
* vx_kill_rtp - kill given VxWorks process
*
* This routine tries to kill a VxWorks RTP by deleting the RTP tasks one by
* one with timeout.
* If it fails to delete one of the RTP tasks (a task safe from deletion), then
* the RTP terminate operation is aborted, and the task we failed to kill is
* stopped with a specific stop reason describing the RTP terminate failure.
*
* If it fails to delete one of the RTP tasks (a task safe from deletion), then
* we do update the corresponding TCF context with stopped_on_error/stopped_errno
* information. This error will only be visible in the debug view when suspending* the task which can't be killed.
*
* NOTE: To get terminate error automatically reported in the debug view (without
* having the suspend the context), we would need to stop this context which
* could lead to severe troubles if this context is suspended while holding a
* critical resource of the system.
*
* RETURNS: 0, or -1 if failed to kill RTP.
*
* ERRNO: N/A
*
*/

static int vx_kill_rtp (pid_t pid, void * terminate_done, TerminateDoneArgs * terminate_done_args) {
    Context * ctx;
    int terminate_error = 0;
    static char error_msg[128];
    TASK_ID taskId;

    ctx = context_find_from_pid (pid, 1);

    /*
     * Parse RTP task list and try to delete them.
     * WARNING: The task deletion must not be performed in rtpTaskEach() to
     * avoid possible lock situation: For example, if an RTP task was holding
     * the taskClassList semaphore when we try to kill it, taskDestroy() will
     * fail as it also need to lock this list...
     * To avoid this situation, we use rtpTaskEach() to only retrieve the first
     * task of the RTP, and then delete it. This operation is repeated while
     * the RTP still exists. However, if we fail to delete a task, we abort
     * the deletion and report an error.
     */

    if (_func_rtpTaskEach != NULL) {
	while (vx_verify_rtp_id ((RTP_ID)pid) == OK) {
	    if (_func_rtpTaskEach ((RTP_ID)pid, rtp_first_task_get, 0, &taskId) == ERROR) {
		/* RTP no longer exist */
		break;
	    }

	    if ((taskId == TASK_ID_NULL) || ((terminate_error = rtp_delete_task (taskId)) != 0)) {
		/*
		 * If rtpTaskEach() failed to get RTP first task (This should
		 * never occur), or task deletion failed: Abort RTP termination.
		 */
		break;
	    }
	}
    }

    /*
     * If a callback routine has been set to handle "Processes terminate"
     * replies, we must invoke it here.
     */

    if (terminate_done != NULL) {
	if (terminate_error) {
	    snprintf(error_msg, sizeof(error_msg), "%s", RTP_TERMINATE_ERROR);
	    terminate_done_args->error = terminate_error;
	    terminate_done_args->msg = error_msg;
	} else {
	    terminate_done_args->error = 0;
	    terminate_done_args->msg = NULL;
	}
	post_event (terminate_done, terminate_done_args);
    }

    if (ctx != NULL && terminate_error != 0) {
	/* Context terminate has been aborted: reset exiting flag */

	ctx->exiting = 0;

	/* Also reset the exiting flag for all remaining children */
 
	if (!list_is_empty(&ctx->children)) {
	    LINK * l = ctx->children.next;
	    while (l != &ctx->children) {
		Context * c = cldl2ctxp(l);
		if (!c->exited) {
		    c->exiting = 0;
		}
		l = l->next;
	    }
	}
    }

    return ((terminate_error == 0) ? 0 : -1);
}

/*******************************************************************************
*
* rtp_sc_check - check if given context is executing a system call
*
* This routine checks if given context is executing a system call, and
* set/unset IS_IN_SYSCALL flag accordingly. The result of the check is returned
* via <pIsSyscall> boolean pointer if not NULL.
*
* If a register set pointer is specified, this routine also retrieves the
* registers at system call boundary for the given task context.
* If the given task context is not executing a system call, then this routine
* simply uses taskRegsGet() to retrieve user land register set.
* 
* RETURNS: OK, or ERROR if failed to get task register set.
*/

static STATUS rtp_sc_check (Context * ctx, REG_SET * pRegs, BOOL * pIsSyscall) {
#ifdef _WRS_PORTABLE_trcLib
    TRC_OS_CTX osCtx;
    TASK_ID taskId = (TASK_ID)(EXT(ctx)->pid);
    WIND_TCB * pTcb = taskTcb (taskId);

    /* Set up TRC_OS_CTX for syscallCheck call */

    if (pTcb->excCnt == 0) {
	osCtx.stackBase	= pTcb->pStackBase;
	osCtx.stackEnd	= pTcb->pStackEnd;
    } else {
	osCtx.stackBase	= pTcb->pExcStackBase;
	osCtx.stackEnd	= pTcb->pExcStackEnd;
    }

    osCtx.pRegs		= pRegs;	/* pointer to REG_SET to fill */
    osCtx.isSyscall	= FALSE;	/* default to FALSE */
    osCtx.pcValidateRtn	= rtp_sc_pcValidate;    /* PC validation routine */

    /*
     * If task is not in supervisor mode and excCnt not null, then we are may
     * be in a system call: Use syscallCheck() to check if this is true and
     * eventually retrieves the registers at system call level (Only the
     * registers that are required for stack back-tracing are filled by
     * syscallCheck() routine).
     */

    if (((pTcb->options & VX_SUPERVISOR_MODE) == 0) && (pTcb->excCnt != 0))
	{
	/* may be a system call */

	SYSCALL_ENTRY_STATE *	pScState;
        RTP_ID old_rtp_id = NULL;

	/*
	 * Fill REG_SET with a specific marker to easily identify invalid
         * registers), syscallCheck() will only retrieve registers required
         * for stack back-tracing.
	 */

	memset (osCtx.pRegs, INV_REG_MARKER, sizeof (REG_SET));

	pScState = (SYSCALL_ENTRY_STATE *)
	    (pTcb->pExcStackBase - sizeof (SYSCALL_ENTRY_STATE));

        /*
         * We must switch to RTP memory context to allow syscallCheck() to
         * validate PC using vxMemProbe().
         */
        if ((old_rtp_id = _func_taskMemCtxSwitch (pTcb->rtpId, 0)) != NULL) {

            trcLibFuncs.syscallCheck (pScState, &osCtx);

            /* Switch back to previous memory context */

            if (old_rtp_id != NULL) _func_taskMemCtxSwitch (old_rtp_id, 0);
	}
    }

    if (pIsSyscall != NULL) *pIsSyscall = osCtx.isSyscall;

    EXT(ctx)->properties &= ~NOTIFY_CTX_CHANGED;
    if (osCtx.isSyscall == FALSE) {
        if ((ctx != NULL) && CTX_IS_IN_SYSCALL(ctx)) {
            /*
             * On previous stop, task context was stopped in a system call.
             * As we're no longer in a system call, notify context change so
             * that the UI will send a getContext and be notified that step
             * actions must be re-enabled.
             */
            EXT(ctx)->properties &= ~IS_IN_SYSCALL;
            EXT(ctx)->properties |= NOTIFY_CTX_CHANGED;
        }

	 /*
	  * Task is not executing a system call: retrieve registers from TCB
	  * using taskRegsGet().
	  */
        if (pRegs != NULL) {
            if (taskRegsGet (taskId, pRegs) == ERROR) return (ERROR);
        }
        return (OK);
    } else {
        /* Task is executing a system call */

        if ((ctx != NULL) && !CTX_IS_IN_SYSCALL(ctx)) {
            /*
             * On previous stop, task context was not stopped in a system call.
             * As we're now in a system call, notify context change so
             * that the UI will send a getContext and be notified that step
             * actions must be disabled.
             */
            EXT(ctx)->properties |= IS_IN_SYSCALL|NOTIFY_CTX_CHANGED;
        }
    }

    return (OK);
#else	/* _WRS_PORTABLE_trcLib */
    /*
     * No way to retrieve registers at system call boundary.
     * This means that Workbench debugger will only display a one line stack
     * backtrace corresponding to current program counter in kernel area.
     */

    if (pIsSyscall != NULL) *pIsSyscall = FALSE; /* Assume not in syscall */

    return (taskRegsGet (taskId, pRegs));
#endif	/* _WRS_PORTABLE_trcLib */
}

#ifdef _WRS_PORTABLE_trcLib
/*******************************************************************************
*
* rtp_sc_pcValidate - PC validation routine for syscall detection
*
* This routine is the PC validation routine that will be used by
* rtp_sc_taskRegsGet().
* This routine is used by syscallCheck() routine to validate PC. Before a PC is 
* derefenced, this routine should be called to ensure there will be no invalid
* memory access.
*
* RETURNS: OK if PC is valid, ERROR otherwise
*
* ERRNO: N/A
*
* \NOMANUAL
*/

static STATUS rtp_sc_pcValidate
    (
    INSTR * 	 pc,		/* PC to validate */
    TRC_OS_CTX * pOsCtx		/* OS info structure */
    )
    {
    INSTR  instr;

    /* 
     * check if within valid range of the system memory; modules could be 
     * loaded by host tools or by the target loader are valid.
     */

    if (((char *) pc >= wrs_kernel_text_start) && 
	((char *) pc <  memPoolEndAdrs))
	return (ALIGNED (pc, sizeof (INSTR)) ? OK : ERROR);

    /* fall back to vxMemProbe() */

    return (vxMemProbe ((char *)pc, VX_READ, sizeof (INSTR), (char *)&instr));
    }
#endif /* _WRS_PORTABLE_trcLib */
#endif /* _WRS_CONFIG_RTP */

/*******************************************************************************
*
* vx_context_terminate - terminate a VxWorks context
*
* This routine terminates a VxWorks task or RTP context.
* Corresponding TCF context (if any) will be deleted via the exit hooks.
*
* RETURNS: 0, or -1 if failed to terminate given context.
*
* ERRNO: N/A
*
*/

int vx_context_terminate (pid_t pid, void * terminate_done, void * data) {
    Context * ctx;

    if (vx_verify_task_id (pid) == OK) {
	/* Try to terminate task with a small timeout */
	return (vx_kill_task ((_Vx_usr_arg_t)pid, (_Vx_usr_arg_t)vx_terminate_timeout, (_Vx_usr_arg_t)terminate_done, (_Vx_usr_arg_t)data, 0, 0));
    }
#if defined(_WRS_CONFIG_RTP)
    else if (vx_verify_rtp_id ((RTP_ID)pid) == OK) {
	return (vx_kill_rtp (pid, terminate_done, data));
    }
#endif
    else {
	errno = ERR_INV_CONTEXT;

	/*
	 * This context no longer exists, invoke terminate command callback
	 * to reply to Processes terminate command.
	 */
	if (terminate_done != NULL) {
	    ((TerminateDoneArgs *)data)->error = errno;
	    ((EventCallBack *)terminate_done) (data);
	}
	return -1;
    }

    return 0;
}

static VXDBG_CLNT_ID vxdbg_client_initialize (void) {
    semTake (vxdbgClntIdLock, WAIT_FOREVER);
    if (vxdbg_clnt_id != NULL) {
	semGive (vxdbgClntIdLock);
	return vxdbg_clnt_id;
    }

    /* Create a VxDBG debug client for TCF */

    if ((vxdbg_clnt_id = vxdbgClntRegister("VxWorks TCF Debug Agent", 0)) == NULL) {
	semGive (vxdbgClntIdLock);
        check_error(errno);
	return NULL;
    }

    semGive (vxdbgClntIdLock);

    /* Install VxDBG events hooks */

    if ((vxdbgHookAdd (vxdbg_clnt_id, EVT_BP, vxdbg_evt_bp_hook) == ERROR) || (vxdbgHookAdd (vxdbg_clnt_id, EVT_TRACE, vxdbg_evt_trace_hook) == ERROR) || (vxdbgHookAdd (vxdbg_clnt_id, EVT_CTX_STOP, vxdbg_evt_ctx_stopped) == ERROR) || (vxdbgHookAdd (vxdbg_clnt_id, EVT_CTX_CREATE, vxdbg_evt_ctx_created) == ERROR) || (vxdbgHookAdd (vxdbg_clnt_id, EVT_CTX_DELETE, vxdbg_evt_ctx_deleted) == ERROR) || (vxdbgHookAdd (vxdbg_clnt_id, EVT_EXC, vxdbg_evt_exc_hook) == ERROR) || (vxdbgHookAdd (vxdbg_clnt_id, EVT_CTX_CONT, vxdbg_evt_ctx_continued) == ERROR)) {
	goto vxdbg_client_initialize_error;
    }

    return vxdbg_clnt_id;

vxdbg_client_initialize_error:
    (void) vxdbgClntUnregister (vxdbg_clnt_id);
    vxdbg_clnt_id = NULL;
    check_error(errno);
    return NULL;
}

/*******************************************************************************
*
* vx_trace_stack_bottom_check - check stack bottom limit
*
* This routine check if the execution context stack reaches its limit.
*
* RETURNS: 1 if the limit is reached, 0 if it is not reached and -1 an error
* 	   or a cache miss occur.
*
* ERRNO: N/A
*
*/

#if ENABLE_Symbols
int vx_trace_stack_bottom_check(StackFrame * frame) {
    uint64_t pc_val = 0;
    Symbol * symbol = NULL;
    static int kerneltask_entry_symbol_size = -1;

    if (!CTX_IS_TASK(frame->ctx)) {
        return 0;
    }
    if (read_reg_value(frame, get_PC_definition(frame->ctx), &pc_val) < 0) {
        return -1;
    }
#if defined(_WRS_CONFIG_RTP)
    if (CTX_IS_RTP (frame->ctx->parent)) {
        ContextExtensionVxWorks * ext = EXT(frame->ctx);
        if ((((WIND_RTP *) (EXT(frame->ctx->mem)->pid))->initTaskId) == EXT(frame->ctx)->pid) {
            /* case for the main rtp task */
            if ((ext->rtp_task_entry == 0) || (ext->rtp_task_entry_sym_size == -1)) {
                /* retrieve _start symbol size if necessary */
                if (find_symbol_by_name(frame->ctx, STACK_NO_FRAME, 0, "_start", &symbol) < 0) {
                    return -1;
                }
                if (get_symbol_size(symbol, &(ext->rtp_task_entry_sym_size)) < 0) {
                    return -1;
                }
                if (get_symbol_address(symbol, &(ext->rtp_task_entry)) < 0) {
                    return -1;
                }
            }

            if ((ContextAddress) (((WIND_RTP *) (EXT(frame->ctx->mem)->pid))->entrAddr) != ext->rtp_task_entry) {
                return 0;
            }

            if ((pc_val >= ext->rtp_task_entry) &&
                 (pc_val < (ext->rtp_task_entry + ext->rtp_task_entry_sym_size))) {
                return 1;
            }
        } else {
            /* case for RTP task expected the main one */
            if ((ext->rtp_task_entry == 0 ) || (ext->rtp_task_entry_sym_size == -1)) {
                /* retrieve vxtaskEntry symbol size if necessary */
                if (find_symbol_by_name(frame->ctx, STACK_NO_FRAME, 0, "vxTaskEntry", &symbol) < 0) {
                    return -1;
                }
                if (get_symbol_size(symbol, &(ext->rtp_task_entry_sym_size)) < 0) {
                    return -1;
                }
                if (get_symbol_address(symbol, &(ext->rtp_task_entry)) < 0) {
                    return -1;
                }
            }
            if ((pc_val >= ext->rtp_task_entry) &&
                (pc_val < (ext->rtp_task_entry + ext->rtp_task_entry_sym_size))) {
                return 1;
            }
        }
    } else {
#endif	/* _WRS_CONFIG_RTP */
        /* case for kernel task */
        if (kerneltask_entry_symbol_size == -1) {
            /* retrieve vxtaskEntry symbol size if necessary */
            if (find_symbol_by_name(frame->ctx, STACK_NO_FRAME, 0, "vxTaskEntry", &symbol) < 0) {
                return -1;
            }
            if (get_symbol_size(symbol, &(kerneltask_entry_symbol_size)) < 0) {
                return -1;
            }
        }

        if ((pc_val >= &vxTaskEntry) &&
                (pc_val < (&vxTaskEntry + kerneltask_entry_symbol_size))) {
            return 1;
        }
#if defined(_WRS_CONFIG_RTP)
    }
#endif
    return 0;
}
#endif /* if ENABLE_Symbols */

/*******************************************************************************
*
* event_context_created - event context creation listener
*
* This routine is called whenever a TCF context is created (attached).
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void event_context_created(Context * ctx, void * x) {
#ifdef _WRS_CONFIG_RTP
    if (CTX_IS_RTP(ctx)) {
        RTP_INFO * info;
	semTake (rtpListLock, WAIT_FOREVER);
	info = find_rtp_info((RTP_ID) EXT(ctx)->pid);
	if (info != NULL) info->attached = 1;
	semGive (rtpListLock);
	assert (info != NULL);
    }
#endif
}

/*******************************************************************************
*
* event_context_deleted - event context deletion listener
*
* This routine is called whenever a TCF context is deleted (detached).
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void event_context_exited(Context * ctx, void * x) {
#ifdef _WRS_CONFIG_RTP
    if (CTX_IS_RTP(ctx)) {
        RTP_INFO * info;
	semTake (rtpListLock, WAIT_FOREVER);
	info = find_rtp_info((RTP_ID) EXT(ctx)->pid);
	if (info != NULL) info->attached = 0;
	semGive (rtpListLock);
	assert (info != NULL);
    }
#endif
}

#ifdef ENABLE_KERNEL_MEM_CTX_ATTACH_AT_STARTUP
/*******************************************************************************
*
* vx_attach_kernel_container - attach VxWorks Kernel container
*
* This routine is called at startup to attach VxWorks Kernel container
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vx_attach_kernel_container(void * arg) {
    int err = 0;
    /* Implicitly attach to kernel context and forbid detachment */
    vx_create_context (VX_KERNEL_FOLDER_ID, IMPLICIT_ATTACH|ALWAYS_ATTACH, &err);
}
#endif

/*******************************************************************************
*
* vx_set_errno - convert VxWorks errno to a TCF error string
*
* This routine converts given VxWorks errno to a TCF error string which will
* be reported to TCF clients.
*
* NOTE: This routine is not exhautive and only support errnos set by
*       vxdbgBpAdd() & vxdbgHwBpAdd() routines.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

static void vx_set_errno (int vx_errno) {

    switch (vx_errno) {
        case S_vxdbgLib_UNSUPPORTED_HW_BP_ACCESS_TYPE:
            set_errno(ERR_OTHER, "Unsupported hardware breakpoint access type.");
            break;
        case S_vxdbgLib_UNSUPPORTED_HW_BP_SIZE:
            set_errno(ERR_OTHER, "Unsupported hardware breakpoint access size.");
            break;
        case S_vxdbgLib_NO_HW_BP_SUPPORT:
            set_errno(ERR_OTHER, "No hardware breakpoint support.");
            break;
        case S_vxdbgLib_HW_REGS_FULL:
            set_errno(ERR_OTHER, "Too many hardware breakpoints set.");
            break;
        case S_vxdbgLib_INVALID_HW_BP:
            set_errno(ERR_OTHER, "Invalid hardware breakpoint type.");
            break;
        case S_vxdbgLib_INVALID_MEM_ACCESS:
            set_errno(ERR_OTHER, "Invalid address.");
            break;
        default:
            /*
             * This errno is not mapped to a TCF error string: The VxWorks
             * errno string will be reported to TCF clients.
             */
            break;
    }
}

/*******************************************************************************
*
* init_contexts_sys_dep - VxWorks TCF agent initialization
*
* This routine initializes VxWorks TCF agent:
*	- Create a VxDBG debug client for TCF.
*	- Add VxDBG events hooks.
*	- Attach to running VxWorks breakable tasks.
*
* RETURNS: N/A
*
* ERRNO: N/A
*
*/

void init_contexts_sys_dep(void) {
    static ContextEventListener listener = {
	event_context_created,
	event_context_exited,
	NULL,
	NULL,
	NULL
    };

    context_extension_offset = context_extension(sizeof(ContextExtensionVxWorks));
    ini_context_pid_hash();

#if SERVICE_PathMap
    {
    static PathMapEventListener listener = {event_path_map_changed};
    
    /* Track Path Map changes */

    add_path_map_event_listener (&listener, NULL);
    }
#endif
    add_context_event_listener(&listener, NULL);

    lstInit (&taskStartList);
    taskStartListLock = semMCreate (SEM_Q_FIFO);
    vxdbgClntIdLock = semMCreate (SEM_Q_FIFO);
#ifdef _WRS_CONFIG_RTP
    lstInit (&rtpStartList);
    lstInit (&rtpMapList);
    lstInit (&rtpIoList);
    rtpListLock = semMCreate (SEM_Q_FIFO);
    if (_func_rtpInitCompleteHookAdd != NULL) (void) _func_rtpInitCompleteHookAdd(rtp_initialized_hook, FALSE);
    if (_func_rtpDeleteHookAdd != NULL) (void) _func_rtpDeleteHookAdd(rtp_delete_hook, FALSE);
    if (_func_rtpShlInfoGet != NULL) {
	if (_func_rtpShlOpenHook != NULL) _func_rtpOldShlOpenHook = _func_rtpShlOpenHook;
	_func_rtpShlOpenHook = rtp_shl_open_hook;
	if (_func_rtpShlCloseHook != NULL) _func_rtpOldShlCloseHook = _func_rtpShlCloseHook;
	_func_rtpShlCloseHook = rtp_shl_close_hook;
    }
#endif

    /* Install Module load/unload hooks */
    moduleLoadHookAdd(module_load_func);
    moduleUnloadHookAdd(module_unload_func);

    /*
     * Install a task creation hook to stop & attach to new task as soon as
     * they are created. This allow breakpoints to be planted by the debug
     * agent in any situation.
     */
    taskCreateHookAdd(task_create_func);

#ifdef ENABLE_KERNEL_MEM_CTX_ATTACH_AT_STARTUP
    /*
     * Implicitly attach to kernel context and forbid detachment. Post this
     * to event thread to defer attachment until end of initialization
     * sequence.
     * The call to vx_attach_kernel_container() has to be deferred because we want
     * to make sure all TCF services have been initialized before we create any context;
     * this is because the init routine of a service may make use of context_extension()
     * API and this API cannot be called after any context has been created.
     */
    post_event(vx_attach_kernel_container, NULL);
#endif
}
#endif  /* if ENABLE_DebugContext */
#endif /* _WRS_KERNEL */
