/*******************************************************************************
 * Copyright (c) 2013-2014 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 * The Eclipse Public License is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 * You may elect to redistribute this code under either of these licenses.
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*
 * VxWorks local memory heap manager.
 */

#if !(USE_libc_malloc)

#include <vxWorks.h>
#include <memPartLib.h>
#include <stdlib.h>
#include <string.h>
#include <tcf/config.h>
#include <tcf/framework/trace.h>
#include <tcf/framework/myalloc.h>

/*
 * PARTITION_OVERHEAD - Partition overhead size
 *
 * From memPartAddToPool() documentation:
 * The size of the memory pool being added has to be large enough to accommodate
 * the section overhead consisting of a section header and some reserved blocks
 * that mark the beginning and the end of the section.
 * This overhead, approximately 64 bytes, is not available for allocation.
 *
 * As there is no published way to obtain the overhead size, this code uses
 * the following macro.
 */
#define	PARTITION_OVERHEAD	256
#define TCF_MEM_POOL_SIZE       (256 * 1024)    /* 256 Kbytes */

#define EXIT_MSG        "Exiting VxWorks Debug Agent."

PART_ID	tcfMemPartId = NULL;

static STATUS loc_add_mem_to_mem_partition (size_t size) {
    char * pTcfPool = NULL;     /* pointer to mem area */

    if ((pTcfPool = (char *) malloc(size)) == NULL) {
	/* Could not allocate memory.  malloc() returned NULL */
	perror("malloc error: Failed to add memory to Debug Agent memory partition");
	(void) memPartDelete (tcfMemPartId);
	tcfMemPartId = NULL;
	return ERROR;
    }

    /* add memory to created partition */
    if (memPartAddToPool(tcfMemPartId, pTcfPool, size) != OK) {
	perror("memPartAddToPool error: Failed to add memory to Debug Agent memory partition");
	free (pTcfPool);
	(void) memPartDelete (tcfMemPartId);
	tcfMemPartId = NULL;
	return ERROR;
    }
    return OK;
}

static STATUS loc_create_mem_partition (void) {
    char * pTcfPool = NULL;     /* pointer to mem area */
    UINT options = 0;

    tcfMemPartId = memPartCreate(pTcfPool, 0);
    if (tcfMemPartId == NULL) {
	/* Memory partition could not be created, returned NULL */
        perror("memPartCreate error: Failed to create Debug Agent memory partition.");
	return ERROR;
    }

    /* Modified memory partition options to remove messages for auto-growth */
    if (memPartOptionsGet(tcfMemPartId, &options) == ERROR) {
        perror("memPartOptionsGet error: Failed to create Debug Agent memory partition.");
	(void) memPartDelete (tcfMemPartId);
	tcfMemPartId = NULL;
	return ERROR;
    }
    options &= ~MEM_ALLOC_ERROR_LOG_FLAG;
    if (memPartOptionsSet(tcfMemPartId, options) == ERROR) {
	perror("memPartOptionsSet error: Failed to create Debug Agent memory partition.");
	(void) memPartDelete (tcfMemPartId);
	tcfMemPartId = NULL;
	return ERROR;
    }
    return (loc_add_mem_to_mem_partition(TCF_MEM_POOL_SIZE));
}

void * loc_alloc(size_t size) {
    void * p;
    int retry = 0;
    size_t tcfMemPoolSize = TCF_MEM_POOL_SIZE;

    if (tcfMemPartId == NULL) {
	if (loc_create_mem_partition() == ERROR) {
	    printErr (EXIT_MSG);
            exit(1);
        }
    }

    if (size == 0) size = 1;

again_alloc:
    if ((p = memPartAlloc(tcfMemPartId, size)) == NULL) {
        if (retry == 1) {
	    perror("memPartAlloc error: Failed to allocate memory from Debug Agent memory partition.");
        }

	retry += 1;

        if (size >= tcfMemPoolSize - PARTITION_OVERHEAD)
            tcfMemPoolSize = ROUND_UP(size + PARTITION_OVERHEAD, tcfMemPoolSize);

	if ((retry <= 1) && (loc_add_mem_to_mem_partition(tcfMemPoolSize) == OK)) goto again_alloc;
	else {
	    printErr (EXIT_MSG);
            exit(1);
	}
    }
    trace(LOG_ALLOC, "loc_alloc(%u) = %#lx", (unsigned)size, p);
    return p;
}

void * loc_alloc_zero(size_t size) {
    void * p;

    if (size == 0) size = 1;
    p = loc_alloc(size);
    memset(p, 0, size);
    trace(LOG_ALLOC, "loc_alloc_zero(%u) = %#lx", (unsigned)size, p);
    return p;
}

void * loc_realloc(void * ptr, size_t size) {
    void * p;
    int retry = 0;
    size_t tcfMemPoolSize = TCF_MEM_POOL_SIZE;

    if (tcfMemPartId == NULL) {
	if (loc_create_mem_partition() == ERROR) {
            printErr (EXIT_MSG);
            exit(1);
        }
    }

    if (size == 0) size = 1;

again_realloc:
    if ((p = memPartRealloc(tcfMemPartId, ptr, size)) == NULL) {
        if (retry == 1) {
	    perror("memPartAlloc error: Failed to allocate memory from Debug Agent memory partition.");
        }

	retry += 1;

        if (size >= tcfMemPoolSize - PARTITION_OVERHEAD)
            tcfMemPoolSize = ROUND_UP(size + PARTITION_OVERHEAD, tcfMemPoolSize);
	if ((retry <= 1) && (loc_add_mem_to_mem_partition(tcfMemPoolSize) == OK)) goto again_realloc;
	else {
	    printErr (EXIT_MSG);
	    exit(1);
	}
    }
    trace(LOG_ALLOC, "loc_realloc(%#lx, %u) = %#lx", ptr, (unsigned)size, p);
    return p;
}

void loc_free(const void * p) {
    trace(LOG_ALLOC, "loc_free %#lx", p);
    (void) memPartFree(tcfMemPartId, (char *)p);
}
#endif /* !USE_libc_malloc */
