/*
 * Copyright (c) 2013-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use of this
 * software may be licensed only pursuant to the terms of an applicable Wind
 * River license agreement.
 */

/*
 * This file handles the creation of a temporary task (a worker thread)
 * in order to load / unload a module.
 */

#include <vxWorks.h>
#include <loadLib.h>
#include <errnoLib.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#include "worker-module.h"
#include <tcf/framework/myalloc.h>
#include <tcf/framework/errors.h>
#include <tcf/framework/json.h>
#include <tcf/framework/streams.h>
#include <tcf/framework/channel.h>
#include <system/VxWorks/tcf/context-vxworks.h>

#define STACK_SIZE    0x2000


static void freeLoadReq(WR_MODULE_LOAD_LOAD_REQ *pReq) {
    loc_free (pReq->filename);
    loc_free (pReq->processId);
    loc_free (pReq->token);
}
   
static void dispatcherSendCommandResult (void *data) {
    WR_MODULE_LOAD_LOAD_REQ *pReq = (WR_MODULE_LOAD_LOAD_REQ *) data;
    wrModuleSendLoadResponse(pReq);
    /* Answer is sent. We can discard pReq */
    channel_unlock (pReq->channel);
    freeLoadReq (pReq);
}

/*
 * loadFnct - The worker thread that load the module
 *
 * This function is spawn by the TCF agent to do the load work
 * pReq - This parameter structure contain all information for
 * the load operation. This structure will be free at the end of
 * this operation.
 *
 */
static void loadFnct(WR_MODULE_LOAD_LOAD_REQ *pReq) {
    int fd;
    int error;
    
    pReq->loader_errno = 0;

    fd = open (pReq->filename,O_RDONLY,0);
    if (fd == -1) {
        /* The file is not accessible, so return immediately */
    	pReq->loader_errno = errnoGet();
    } else {
        /* The file exist, so try to load it */
        
        if (loadModuleEx(fd, pReq->loadOpts, &pReq->modId) == ERROR)
            {
	    /*
	     * If the module is already present on the target, the target
	     * load does not execute the load operation and the debug agent
	     * returns a NULL module ID as specified by the WR_Module
	     * service.
	     */

            if ((error = errnoGet()) != S_loadLib_ALREADY_LOADED)
                pReq->loader_errno = error;
            else
                pReq->modId = NULL;
            }

        close (fd);
    }

    /* Beware : after using dispatcherSendCommandResult, you can NOT use pReq anymore ! */
    VXDBG_POST_EVENT (dispatcherSendCommandResult, pReq, NULL);
}

void Wr_ModuleLoad_Worker (char * token, Channel *channel, char *ctxId,
                           char *fileName, int loadOpts) {    
    int priority;
    /* We need to lock the channel */
    channel_lock(channel);
    
    /* Create a task to do it. Since we will return immediately, we need to
     * create an object to store the request. Once the request is processed,
     * the object is destroyed.
     */
    
    WR_MODULE_LOAD_LOAD_REQ *pReq = loc_alloc_zero(sizeof(WR_MODULE_LOAD_LOAD_REQ));
    
    pReq->channel = channel;
    /* Copy the filename, options and channel to load */
    /* These one are allocated when parsing the command */
    pReq->filename = fileName;
    pReq->processId = ctxId;
    pReq->loadOpts = loadOpts;
    /* Since the reply will "sometime" after due to the worker-thread, I need
     * to duplicate token because the dispatcher command function might have
     * been exited...
     */
    pReq->token = loc_strdup(token);
    (void)taskPriorityGet(taskIdSelf(), &priority);
    TASK_ID tidLoader = taskSpawn ("tTcfLoad", priority,
            VX_UNBREAKABLE|VX_COPROCS_ALL_TASK, STACK_SIZE,
            (FUNCPTR) loadFnct,(_Vx_usr_arg_t) pReq,0,0,0,0,0,0,0,0,0);

    if (tidLoader == NULL) {
        /*
         * If the tTcfLoad task can't start, so send back
         * the TCF event notification
         */
        pReq->loader_errno = errnoGet();
        VXDBG_POST_EVENT (dispatcherSendCommandResult, pReq, NULL);
    }
}
