/*
 * Copyright (c) 2013-2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
 * TCF service filesystem - proxy version.
 *
 */

#include <tcf/config.h>

#if ENABLE_WR_FileSystemProxy
#include <assert.h>
#include <hostFsLib.h>
#include <tcf/framework/protocol.h>
#include <tcf/framework/events.h>
#include <tcf/framework/exceptions.h>
#include <tcf/framework/json.h>
#include <tcf/framework/myalloc.h>
#include <tcf/services/filesystem.h>
#include <tcf/framework/mdep-threads.h>
#include <system/VxWorks/tcf/context-vxworks.h>
#include "WR_FileSystem_proxy.h"


#ifdef  FSPROXY_DEBUG
#include <tcf/framework/trace.h>
#define FSPROXY_DEBUG_LOG(msg,args...)     \
        {                                  \
        trace (LOG_ALWAYS, msg, ## args);  \
        }
#else
#define FSPROXY_DEBUG_LOG(msg,args...)
#endif  /* DEBUG */

#define MAX_HOST_DRIVES 26 /* on Windows, max is 26 */

typedef struct CmdData CmdData;
typedef struct CmdArgs CmdArgs;

typedef int (*SendTcfFunc) (CmdArgs * args);
typedef int (*DecResFunc)  (CmdArgs * args);

typedef struct CmdArgs {
    Channel *   chan;		/* TCF channel */
    char *      cmd_name;	/* command name */
    int 	res_code;	/* TCF error code from current TCF query */
    SendTcfFunc	send_tcf_func;	/* TCF command formatter */
    DecResFunc  decode_res_func;/* TCF command answer decoder */
    SEM_ID	semid;		/* Binary semaphare to wait TCF response */

    /* common field used by several commands */
    union {
	struct {
	    char * file_name;    /* file name to open */
	    int    file_mode;    /* file mode for open command */ 
	    int    file_perm;    /* file permission for open command */
	    char * fd;		 /* TCF file descriptor provided by value-add */
	} open;
	struct {
	    char * fd;		/* TCF file descriptor to close */	
	} close;
	struct {
	    char * fd;		/* TCF file descriptor */
	    off_t  offset;	/* read offset */ 
	    size_t size;	/* size in bytes */
	    char * buf;	        /* outpout buffer */
	    size_t size_done;	/* num of bytes reported by TCF command */
	    int    eof;		/* end of file reached */
	} read;
	struct {
	    char * fd;		/* TCF file descriptor */
	    off_t  offset;	/* write offset */ 
	    char * buf;	        /* data to write */
	    size_t size;	/* max bytes to write */
	} write;
	struct {
	    char * file_name;	/* file name to remove */
	} remove;
	struct {
	    char * 		fd;	/* TCF file descriptor */
	    char * 		path;	/* file or directory path */
	    WrFsProxyStat * 	stat;	/* fstat result */
	} stat;
	struct {
	    char * path;    /* directory path to open */
	    char * fd;	    /* TCF directory descriptor provided by value-add */
        } opendir;
	struct {
	    char * 	fd;   		/* Directoy descriptor */
	    FileNode * 	file_list;      /* file list */
	    FileNode *  last_file_ptr;	/* points to last file of the list */
	    int 	eof;		/* end of file reached */	
	} readdir;
    } cmd;		 	/* data specific to the query type */
} CmdArgsStruct;
 
typedef struct {
    char * root_list[MAX_HOST_DRIVES + 1];  /* table of root names  */
    int nextroot;		/* points to next root name table row */
    char * devid;		/* hostfs device id */
    BOOL devcrea;		/* TRUE: devCreate, FALSE: autoMountRefresh */
    char token[256];		/* TCF token to send answer */
} RootsCmdArgs;


static const char * FILE_SYSTEM = "FileSystem";
static const char * OPEN_CMD 	= "open";
static const char * CLOSE_CMD 	= "close";
static const char * READ_CMD 	= "read";
static const char * WRITE_CMD 	= "write";
static const char * REMOVE_CMD 	= "remove";
static const char * FSTAT_CMD 	= "fstat";
static const char * STAT_CMD    = "stat";
static const char * OPENDIR_CMD = "opendir";
static const char * READDIR_CMD = "readdir";
static const char * ROOTS_CMD	= "roots";

static TASK_ID event_thread_tid = NULL;

/*
 * Common functions used to send TCF commands synchronously to the value-add.
 */

static void flush_channel(Channel * c) {
    /* flush TCF channel by reading remaining data */
    for (;;) {
        int ch = read_stream(&c->inp);
        if (ch == MARKER_EOM) break;
    }
}

static void validate_tcfcmd(Channel * c, CmdArgs * args, int error) {
    Trap trap;

    if (set_trap(&trap)) {
	if (error == 0) {
	    args->chan = c;
	    error = args->decode_res_func(args);
	} else {
            FSPROXY_DEBUG_LOG ("Error 0x%x reported by command %s",error, args->cmd_name); 
	    flush_channel(c);
	}
	clear_trap(&trap);
    } else {
        /* an Exception has been raised while decoding TCF reply*/

        FSPROXY_DEBUG_LOG ("Trap 0x%x when processing command %s", trap.error, args->cmd_name); 
	flush_channel(c);
    }

    args->res_code = error;
    if (args->semid != SEM_ID_NULL) semGive(args->semid);
}

static void send_sync_tcfcmd(CmdArgs * args) {
    
    /* Must be run from dispatch thread */
    assert(is_dispatch_thread());

    assert(args->chan);
    assert(strlen(args->cmd_name) > 0);

    /* Send command header and register reply handler */
    protocol_send_command(args->chan, FILE_SYSTEM, args->cmd_name, (ReplyHandlerCB) validate_tcfcmd, args);

    args->send_tcf_func(args);
    channel_unlock(args->chan);
}

static int send_tcfcmd(CmdArgs * args) {
    /* Cannot be executed in dispatch thread */
    /*
     * NOTE: Here we must not use is_dispatch_thread() which would result in
     * converting current task into a thread (as this routine is using
     * pthread_self()).
     */
    assert (taskIdSelf() != event_thread_tid);
    
    channel_lock(args->chan);

    /*
     * As this routine may be executed in the context of kernel or RTP
     * application tasks, we must not invoke post_event() directly from
     * this code. Calling post_event() which uses pthread_self() has the
     * side effect of converting the caller task into a pthread, and we
     * do not want to convert customers tasks to pthreads...
     * To avoid this, we do differ post_event() call to VxDBG event task.
     */
    VXDBG_POST_EVENT(send_sync_tcfcmd, args, NULL);

    /* Wait for reply (blocking caller that is Worker thread) */
   
    semTake(args->semid, WAIT_FOREVER);
    return args->res_code;
}


/* wr_fsProxy_open API */

static int send_cmd_open(CmdArgs * args) {
    json_write_string(&args->chan->out, args->cmd.open.file_name); 
    write_stream (&args->chan->out, MARKER_EOA);
    json_write_ulong (&args->chan->out, args->cmd.open.file_mode);
    write_stream (&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, '{');
    json_write_string(&args->chan->out, "Permissions");
    write_stream(&args->chan->out, ':');
    json_write_long(&args->chan->out, args->cmd.open.file_perm);
    write_stream(&args->chan->out, '}');
    write_stream(&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_open(CmdArgs * args) {
    int error = 0;

    error = read_errno(&args->chan->inp);

    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while opening %s", error, args->cmd.open.file_name);
	flush_channel(args->chan);
	return error;
    }

    args->cmd.open.fd = json_read_alloc_string(&args->chan->inp);
    json_test_char(&args->chan->inp, MARKER_EOA);
    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_open(Channel * c, SEM_ID semid, const char * name, int mode, int perm, char **fd) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Open a file, name = %s, mode %o, perm = %o\n", name, mode, perm);

    memset(&args, 0, sizeof(args));
    args.cmd_name = (char *)OPEN_CMD;
    args.send_tcf_func = send_cmd_open;
    args.decode_res_func = decode_res_open;
    args.chan = c;
    args.semid = semid;

    args.cmd.open.file_name  = (char *)name;
    args.cmd.open.file_mode = mode;
    args.cmd.open.file_perm = perm;

    if (send_tcfcmd(&args) == 0) {
        *fd = loc_strdup(args.cmd.open.fd);
        FSPROXY_DEBUG_LOG ("File opened sucessfully. File desc =  %s", *fd);
    }
    
    loc_free(args.cmd.open.fd);
    return  args.res_code;
}

/* wr_fsProxy_close API */

static int send_cmd_close(CmdArgs * args) {
    
    json_write_string(&args->chan->out, args->cmd.close.fd); 
    write_stream(&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_close(CmdArgs * args) {
    int error = 0;

    error = read_errno(&args->chan->inp);
    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while closing %s", args->cmd.close.fd);
	flush_channel(args->chan);
	return error;
    }

    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_close(Channel * c, SEM_ID semid, const char * fd) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG("Close File descriptor: %s", fd);

    memset(&args, 0, sizeof(args));
    args.cmd_name = (char *)CLOSE_CMD;
    args.send_tcf_func = send_cmd_close;
    args.decode_res_func = decode_res_close;
    args.chan = c;
    args.semid = semid;

    args.cmd.close.fd = (char *)fd;

    if (send_tcfcmd(&args) != 0) {
        FSPROXY_DEBUG_LOG("Error while closing %s", fd);
    }

    return args.res_code;
}

/* wr_fsProxy_read API */

static int send_cmd_read(CmdArgs * args) {
    json_write_string(&args->chan->out, args->cmd.read.fd); 
    write_stream (&args->chan->out, MARKER_EOA);
    json_write_long (&args->chan->out, args->cmd.read.offset);
    write_stream (&args->chan->out, MARKER_EOA);
    json_write_ulong (&args->chan->out, args->cmd.read.size);
    write_stream(&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_read(CmdArgs * args) {
    int error = 0;
    args->cmd.read.size_done = 0; 
    JsonReadBinaryState state;

    json_read_binary_start(&state, &args->chan->inp);
    args->cmd.read.size_done = json_read_binary_data(&state, args->cmd.read.buf, args->cmd.read.size);
    json_read_binary_end(&state);

    error = read_errno(&args->chan->inp);
    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while reading %s", error, args->cmd.read.fd);
	flush_channel(args->chan);
	return error;
    }
    json_test_char(&args->chan->inp, MARKER_EOA);
    args->cmd.read.eof = json_read_boolean(&args->chan->inp);
    json_test_char(&args->chan->inp, MARKER_EOA);
    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_read(Channel * c, SEM_ID semid, const char * fd, off_t offset, size_t size, char *buf, ssize_t * read_bytes) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Read, fd= %s, offset = %d, size = %d\n", fd, offset, size);

    memset(&args, 0, sizeof(args));
    args.cmd_name = (char *)READ_CMD;
    args.send_tcf_func = send_cmd_read;
    args.decode_res_func = decode_res_read;
    args.chan = c;
    args.semid = semid;

    args.cmd.read.fd = (char *)fd;
    args.cmd.read.offset= offset;
    args.cmd.read.size= size;
    args.cmd.read.buf = buf;

    if (send_tcfcmd(&args) == 0) {
	*read_bytes = args.cmd.read.size_done;
    }
    FSPROXY_DEBUG_LOG ("%d bytes read - error = 0x%x", *read_bytes, args.res_code);
    return args.res_code;
}

/* wr_fsProxy_write API */

static int send_cmd_write(CmdArgs * args) {
    JsonWriteBinaryState state;
    json_write_string(&args->chan->out, args->cmd.write.fd); 
    write_stream (&args->chan->out, MARKER_EOA);
    json_write_long (&args->chan->out, args->cmd.write.offset);
    write_stream (&args->chan->out, MARKER_EOA);
    json_write_binary_start(&state, &args->chan->out, args->cmd.write.size);
    json_write_binary_data(&state, args->cmd.write.buf, args->cmd.write.size);
    json_write_binary_end(&state);
    write_stream(&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_write(CmdArgs * args) {
    int error = 0;
    
    error = read_errno(&args->chan->inp);
    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while writing %s", error, args->cmd.write.fd);
	flush_channel(args->chan);
	return error;
    }

    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_write(Channel * c, SEM_ID semid, const char * fd, off_t offset, size_t size, char *buf) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Write, fd= %s, offset = %d, size = %d\n", fd, offset, size);

    memset(&args, 0, sizeof(args));
    args.cmd_name = (char *)WRITE_CMD;
    args.send_tcf_func = send_cmd_write;
    args.decode_res_func = decode_res_write;
    args.chan = c;
    args.semid = semid;

    args.cmd.write.fd = (char *)fd;
    args.cmd.write.offset= offset;
    args.cmd.write.size= size;
    args.cmd.write.buf= buf;

    return send_tcfcmd(&args);
}

/* wr_fsProxy_remove API */

static int send_cmd_remove(CmdArgs * args) {
    json_write_string(&args->chan->out, args->cmd.remove.file_name); 
    write_stream (&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_remove(CmdArgs * args) {
    int error = 0;
    
    error = read_errno(&args->chan->inp);
    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while removing %s", error, args->cmd.remove.file_name);
	flush_channel(args->chan);
	return error;
    }

    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_remove(Channel * c, SEM_ID semid, const char * file_name) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Remove file: %s", file_name);

    memset(&args, 0, sizeof(args));
    args.cmd_name= (char *)REMOVE_CMD;
    args.send_tcf_func = send_cmd_remove;
    args.decode_res_func = decode_res_remove;
    args.chan = c;
    args.semid = semid;

    args.cmd.remove.file_name = (char *)file_name;

    return send_tcfcmd(&args);
}

/* wr_fsProxy_fstat API */

static void read_file_stat(InputStream * inp, const char * nm, void * arg) {
    WrFsProxyStat * file_stat = (WrFsProxyStat*)arg;
    if (strcmp(nm, "Size") == 0) {
        file_stat->size = json_read_int64(inp);
    }
    else if (strcmp(nm, "UID") == 0) {
        file_stat->uid = (int)json_read_long(inp);
    }
    else if (strcmp(nm, "GID") == 0) {
        file_stat->gid = (int)json_read_long(inp);
    }
    else if (strcmp(nm, "Permissions") == 0) {
        file_stat->permissions = (int)json_read_long(inp);
    }
    else if (strcmp(nm, "ATime") == 0) {
        /* 
         * TCF FileSystem provides dates as milliseconds from midnight 
         * Jan 1, 1970 in UTC, but vxWorks assumes it is seconds and not
         * milliseconds.
         */	
        file_stat->atime = json_read_uint64(inp) / 1000;
    }
    else if (strcmp(nm, "MTime") == 0) {
        /* 
         * TCF FileSystem provides dates as milliseconds from midnight 
         * Jan 1, 1970 in UTC, but vxWorks assumes it is seconds and not 
         * milliseconds.
         */     
        file_stat->mtime = json_read_uint64(inp) / 1000;
    }
    else if (strcmp(nm, "Win32Attrs") == 0) {
    	json_read_ulong(inp);
    }
    else {
	FSPROXY_DEBUG_LOG ("File stat Unknown field : %s", nm);
        exception(ERR_JSON_SYNTAX);
    }
}

static int send_cmd_fstat(CmdArgs * args) {
    json_write_string(&args->chan->out, args->cmd.stat.fd); 
    write_stream (&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_fstat(CmdArgs * args) {
    int error = 0;
    
    error = read_errno(&args->chan->inp);
    if (error != 0) {
	FSPROXY_DEBUG_LOG("Fstat command Error 0x%x . FD =  %s", error, args->cmd.stat.fd);
	flush_channel(args->chan);
	return error;
    }

    json_read_struct(&args->chan->inp, read_file_stat, args->cmd.stat.stat);
    json_test_char(&args->chan->inp, MARKER_EOA);
    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_fstat(Channel * c, SEM_ID semid, const char * fd, WrFsProxyStat * proxy_stat) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Fstat fd: %s", fd);

    memset(&args, 0, sizeof(args));
    args.cmd_name= (char *)FSTAT_CMD;
    args.send_tcf_func = send_cmd_fstat;
    args.decode_res_func = decode_res_fstat;
    args.chan = c;
    args.semid = semid;

    args.cmd.stat.stat = proxy_stat;
    args.cmd.stat.fd = (char *)fd;

    return send_tcfcmd(&args);
}

/* wr_fsProxy_stat API */

static int send_cmd_stat(CmdArgs * args) {
    json_write_string(&args->chan->out, args->cmd.stat.path); 
    write_stream (&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_stat(CmdArgs * args) {
    int error = 0;
    
    error = read_errno(&args->chan->inp);
    if (error != 0) {
	FSPROXY_DEBUG_LOG("Stat command Error 0x%x . path =  %s", error, args->cmd.stat.path);
	flush_channel(args->chan);
	return error;
    }

    json_read_struct(&args->chan->inp, read_file_stat, args->cmd.stat.stat);
    json_test_char(&args->chan->inp, MARKER_EOA);
    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_stat(Channel * c, SEM_ID semid, const char * path, WrFsProxyStat * proxy_stat) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Stat path: %s", path);

    memset(&args, 0, sizeof(args));
    args.cmd_name= (char *)STAT_CMD;
    args.send_tcf_func = send_cmd_stat;
    args.decode_res_func = decode_res_stat;
    args.chan = c;
    args.semid = semid;

    args.cmd.stat.path = (char *)path;
    args.cmd.stat.stat = proxy_stat;
     
    return send_tcfcmd(&args);
}

/* wr_fsProxy_opendir API */

static int send_cmd_opendir(CmdArgs * args) {
    json_write_string(&args->chan->out, args->cmd.opendir.path); 
    write_stream (&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static int decode_res_opendir(CmdArgs * args) {
    int error = 0;

    error = read_errno(&args->chan->inp);
    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while opening directory %s", error, args->cmd.opendir.path);
	flush_channel(args->chan);
	return error;
    }

    args->cmd.opendir.fd = json_read_alloc_string(&args->chan->inp);
    json_test_char(&args->chan->inp, MARKER_EOA);
    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_opendir(Channel * c, SEM_ID semid, const char * path, char **fd){
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Open a directory, name = %s\n", path);

    memset(&args, 0, sizeof(args));
    args.cmd_name= (char *)OPENDIR_CMD;
    args.send_tcf_func = send_cmd_opendir;
    args.decode_res_func = decode_res_opendir;
    args.chan = c;
    args.semid = semid;

    args.cmd.opendir.path = (char *)path;

    if (send_tcfcmd(&args) == 0) {
	*fd = args.cmd.opendir.fd;
    }
    return args.res_code;
}

/* wr_fsProxy_readdir API */

static int send_cmd_readdir(CmdArgs * args) {
    json_write_string(&args->chan->out, args->cmd.readdir.fd); 
    write_stream (&args->chan->out, MARKER_EOA);
    write_stream(&args->chan->out, MARKER_EOM);
    return 0;
}

static void add_file(CmdArgs * args, char * file_name) {
    /* extend or initialize the file list with a new file  name */

    FileNode * newNode = loc_alloc_zero(sizeof(FileNode));
    newNode->file_name = file_name;
    
    if (args->cmd.readdir.file_list == NULL){
	args->cmd.readdir.file_list = newNode;
    } else {
	args->cmd.readdir.last_file_ptr->next = newNode;
    }
    args->cmd.readdir.last_file_ptr = newNode;
}

static void read_dir_one_file(InputStream * inp, const char * name, CmdArgs * args) {
    char * file_name = NULL; 
    if (strcmp(name, "FileName") == 0) {
	file_name = json_read_alloc_string(inp);
	add_file (args, file_name);
    } else if (strcmp(name, "LongName") == 0) {
	file_name = json_read_alloc_string(inp);
	add_file (args, file_name);
    } else if (strcmp(name, "Attrs") == 0) {
	WrFsProxyStat file_stat;
	json_read_struct (inp, read_file_stat, &file_stat);
    }
}

static void read_dir_array_cb (InputStream * inp, void * arg) {
    json_read_struct(inp, (JsonStructCallBack *)read_dir_one_file, arg);
}

static int decode_res_readdir(CmdArgs * args) {
    int error = 0;

    json_read_array(&args->chan->inp, read_dir_array_cb, args);
    json_test_char(&args->chan->inp, MARKER_EOA);
    error = read_errno(&args->chan->inp);

    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while reading directory %s", error, args->cmd.readdir.fd);
	flush_channel(args->chan);
	return error;
    }
    args->cmd.readdir.eof = json_read_boolean(&args->chan->inp);
    json_test_char(&args->chan->inp, MARKER_EOA);
    json_test_char(&args->chan->inp, MARKER_EOM);
    return 0;
}

int wr_fsproxy_readdir(Channel * c, SEM_ID semid, char *fd, FileNode ** file_list, FileNode ** last_node, int *oef) {
    CmdArgs args;

    FSPROXY_DEBUG_LOG ("Read a directory, FD = %s\n", fd);

    memset(&args, 0, sizeof(args));
    args.cmd_name= (char *)READDIR_CMD;

    args.send_tcf_func = send_cmd_readdir;
    args.decode_res_func = decode_res_readdir;
    args.chan = c;
    args.semid = semid;

    args.cmd.readdir.fd = fd;

    if (send_tcfcmd(&args) == 0) {
	*oef = args.cmd.readdir.eof;
        *file_list = args.cmd.readdir.file_list;
 	*last_node = args.cmd.readdir.last_file_ptr;	
        FSPROXY_DEBUG_LOG ("Directory sucessfully read.");
        }
    
    return  args.res_code;
}

/* wr_fsProxy_roots API */

static void read_one_root(InputStream * inp, const char * name, void * x) { 
    RootsCmdArgs * args = (RootsCmdArgs *)x;
    char * root_name = NULL; 
    if (strcmp(name, "FileName") == 0) {
	root_name = json_read_alloc_string(inp);
	args->root_list[args->nextroot++] = root_name;
    } else if (strcmp(name, "Attrs") == 0) {
	WrFsProxyStat file_stat;
	memset((void*)&file_stat, 0, sizeof(file_stat));
	json_read_struct (inp, read_file_stat, &file_stat);
    }
}

static void read_roots_array_cb(InputStream * inp, void * args) {
    json_read_struct(inp, (JsonStructCallBack *)read_one_root, args);
}

static void validate_res_roots(Channel * c, void * x, int error) {
    unsigned int ix = 0;

    RootsCmdArgs * args = (RootsCmdArgs *)x;
    json_read_array(&c->inp, read_roots_array_cb, args);
    json_test_char(&c->inp, MARKER_EOA);

    if (error == 0) 
	error = read_errno(&c->inp);

    if (error != 0) {
	FSPROXY_DEBUG_LOG("Error 0x%x while reading roots", error);
	flush_channel(c);
    }else {
        json_test_char(&c->inp, MARKER_EOM);

	/* refresh the hostfs mount with the generated host drives list */

        if (hostFsAutoMountRefresh(c, args->root_list) != OK) {
 	    if (args->devcrea) {
	        error = set_errno(ERR_OTHER, "Fail to automatically mount host drives.");
	    } else {
	        error = set_errno(ERR_OTHER, "Fail to refresh auto mounted host drives.");
	    }
        }
    }

    /* free the host drive list provided by the value-add */

    while (args->root_list[ix] != NULL)
	{
	loc_free ((void *) args->root_list[ix++]);
	}

    /* answer the WR_HostFs 'devCreate' or 'autoMountRefresh' TCF query */

    write_stringz(&c->out, "R");
    write_stringz(&c->out, args->token);
    write_errno(&c->out, error);
    if (args->devcrea) {
	/* case devCreate : device ID is provided on sucess, else null */

	if (args->devid != NULL) json_write_string(&c->out, args->devid);
	else write_stringz(&c->out, "null");
    }
    write_stream(&c->out, MARKER_EOM);
    loc_free(args);
}

void wr_fsproxy_roots (Channel * c, char * token, char * devid, BOOL devcrea){
    RootsCmdArgs * args  = NULL;
    assert(is_dispatch_thread());

    /* 
     * This non blocking client to 'FileSystem roots' command:
     *  - queries the value-add to determine host drives
     *  - updates hostfs auto-mouted mount point list accordingly
     *  - sends the response of the initial WR_HostFs 'devCreate' or 'autoMountRefresh' query
     */

    FSPROXY_DEBUG_LOG ("Update host roots\n");
    args = (RootsCmdArgs *)loc_alloc_zero(sizeof(RootsCmdArgs));
    args->devid = devid;
    args->devcrea = devcrea;
    args->nextroot = 0;
    strlcpy(args->token, token, sizeof(args->token));
    protocol_send_command(c, FILE_SYSTEM, (char *)ROOTS_CMD, (ReplyHandlerCB) validate_res_roots, args);
    write_stream(&c->out, MARKER_EOM);
} 

void ini_wr_fsproxy_lib(void) {
    event_thread_tid = taskIdSelf ();
}
#endif /* ENABLE_WR_FileSystemProxy*/
